// This file is generated by scripts/release.py
// Do not edit it manually.

// Include compatibility layer for cross-platform support
#include "compat.h"

#if defined(_WIN32)
// Stub implementations - instrumentation not supported on Windows
#include <stdlib.h>

#include "core.h"

struct InstrumentHooks {
  uint64_t _unused;
};

static struct InstrumentHooks stub_instance = {0};

InstrumentHooks *instrument_hooks_init(void) { return &stub_instance; }

void instrument_hooks_deinit(InstrumentHooks *hooks) { (void)hooks; }

bool instrument_hooks_is_instrumented(InstrumentHooks *hooks) {
  (void)hooks;
  return false;
}

uint8_t instrument_hooks_start_benchmark(InstrumentHooks *hooks) {
  (void)hooks;
  return 0;
}

uint8_t instrument_hooks_stop_benchmark(InstrumentHooks *hooks) {
  (void)hooks;
  return 0;
}

uint8_t instrument_hooks_set_executed_benchmark(InstrumentHooks *hooks,
                                                int32_t pid, const char *uri) {
  (void)hooks;
  (void)pid;
  (void)uri;
  return 0;
}

// Deprecated: use instrument_hooks_set_executed_benchmark instead
uint8_t instrument_hooks_executed_benchmark(InstrumentHooks *hooks, int32_t pid,
                                            const char *uri) {
  (void)hooks;
  (void)pid;
  (void)uri;
  return 0;
}

uint8_t instrument_hooks_set_integration(InstrumentHooks *hooks,
                                         const char *name,
                                         const char *version) {
  (void)hooks;
  (void)name;
  (void)version;
  return 0;
}

void instrument_hooks_set_feature(uint64_t feature, bool enabled) {
  (void)feature;
  (void)enabled;
}

uint64_t instrument_hooks_current_timestamp(void) { return 0; }

uint8_t instrument_hooks_add_marker(InstrumentHooks *hooks, int32_t pid,
                                    uint8_t marker_type, uint64_t timestamp) {
  (void)hooks;
  (void)pid;
  (void)marker_type;
  (void)timestamp;
  return 0;
}

uint8_t instrument_hooks_set_environment(InstrumentHooks *hooks,
                                         const char *section_name,
                                         const char *key, const char *value) {
  (void)hooks;
  (void)section_name;
  (void)key;
  (void)value;
  return 0;
}

uint8_t instrument_hooks_set_environment_list(InstrumentHooks *hooks,
                                              const char *section_name,
                                              const char *key,
                                              const char *const *values,
                                              uint32_t count) {
  (void)hooks;
  (void)section_name;
  (void)key;
  (void)values;
  (void)count;
  return 0;
}

uint8_t instrument_hooks_write_environment(InstrumentHooks *hooks,
                                           int32_t pid) {
  (void)hooks;
  (void)pid;
  return 0;
}

#elif defined(__APPLE__)
#include "core.h"
#define ZIG_TARGET_MAX_INT_ALIGNMENT 16
#include "zig.h"
struct anon__lazy_58 {
 uint8_t const *ptr;
 uintptr_t len;
};
struct fs_File__794;
typedef struct anon__lazy_87 nav__1280_39;
typedef struct anon__lazy_58 nav__1280_41;
struct fs_File__794 {
 int32_t handle;
};
struct anon__lazy_87 {
 struct fs_File__794 payload;
 uint16_t error;
};
struct fs_File_OpenFlags__1101;
struct fs_File_OpenFlags__1101 {
 uint8_t mode;
 uint8_t lock;
 bool lock_nonblocking;
 bool allow_ctty;
};
struct fifo_posix_Pipe_Writer__786;
typedef struct anon__lazy_95 nav__1282_39;
struct mem_Allocator__599;
typedef struct anon__lazy_58 nav__1282_42;
struct mem_Allocator_VTable__602;
struct mem_Allocator__599 {
 void *ptr;
 struct mem_Allocator_VTable__602 const *vtable;
};
struct array_list_ArrayListAligned_28u8_2cnull_29__806;
typedef struct anon__lazy_103 nav__1282_50;
struct anon__lazy_103 {
 uint8_t *ptr;
 uintptr_t len;
};
struct array_list_ArrayListAligned_28u8_2cnull_29__806 {
 struct anon__lazy_103 items;
 uintptr_t capacity;
 struct mem_Allocator__599 allocator;
};
struct fifo_posix_Pipe_Writer__786 {
 struct mem_Allocator__599 allocator;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 buffer;
 struct fs_File__794 file;
};
struct anon__lazy_95 {
 struct fifo_posix_Pipe_Writer__786 payload;
 uint16_t error;
};
typedef struct anon__lazy_87 nav__1282_59;
struct fifo_posix_Pipe_Reader__788;
typedef struct anon__lazy_110 nav__1281_39;
typedef struct anon__lazy_58 nav__1281_42;
typedef struct anon__lazy_103 nav__1281_50;
struct fifo_posix_Pipe_Reader__788 {
 struct mem_Allocator__599 allocator;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 buffer;
 struct fs_File__794 file;
};
struct anon__lazy_110 {
 struct fifo_posix_Pipe_Reader__788 payload;
 uint16_t error;
};
typedef struct anon__lazy_87 nav__1281_59;
struct runner_fifo_RunnerFifo__631;
typedef struct anon__lazy_116 nav__1129_39;
typedef struct anon__lazy_103 nav__1129_49;
struct runner_fifo_RunnerFifo__631 {
 struct mem_Allocator__599 allocator;
 struct fifo_posix_Pipe_Writer__786 writer;
 struct fifo_posix_Pipe_Reader__788 reader;
};
struct anon__lazy_116 {
 struct runner_fifo_RunnerFifo__631 payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__1129_64;
typedef struct anon__lazy_95 nav__1129_66;
typedef struct anon__lazy_110 nav__1129_71;
struct shared_Command__1178;
struct shared_Command__struct_1181__1181;
typedef struct anon__lazy_58 nav__1286_44;
struct shared_Command__struct_1181__1181 {
 struct anon__lazy_58 uri;
 int32_t pid;
};
struct shared_Command__struct_1182__1182;
struct shared_Command__struct_1182__1182 {
 struct anon__lazy_58 name;
 struct anon__lazy_58 version;
};
struct shared_Command__struct_1183__1183;
struct shared_MarkerType__1185;
struct shared_MarkerType__1185 {
 union {
  uint64_t SampleStart;
  uint64_t SampleEnd;
  uint64_t BenchmarkStart;
  uint64_t BenchmarkEnd;
 } payload;
 uint8_t tag;
};
struct shared_Command__struct_1183__1183 {
 struct shared_MarkerType__1185 marker;
 int32_t pid;
};
struct shared_Command__1178 {
 union {
  struct shared_Command__struct_1181__1181 ExecutedBenchmark;
  struct shared_Command__struct_1182__1182 SetIntegration;
  struct shared_Command__struct_1183__1183 AddMarker;
  uint64_t SetVersion;
  uint8_t IntegrationModeResponse;
 } payload;
 uint8_t tag;
};
typedef struct anon__lazy_103 nav__1286_65;
struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222;
struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *context;
};
struct c_pollfd__struct_1304__1304;
struct c_pollfd__struct_1304__1304 {
 int32_t fd;
 int16_t events;
 int16_t revents;
};
typedef struct anon__lazy_152 nav__1268_43;
struct anon__lazy_152 {
 struct c_pollfd__struct_1304__1304 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_155 nav__1268_46;
struct anon__lazy_155 {
 uintptr_t payload;
 uint16_t error;
};
typedef struct anon__lazy_157 nav__4193_38;
struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992;
struct anon__lazy_157 {
 uint32_t payload;
 uint16_t error;
};
struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968;
struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *context;
};
struct io_Reader__2141;
typedef struct anon__lazy_155 nav__4193_50;
typedef struct anon__lazy_103 nav__4193_52;
struct io_Reader__2141 {
 void const *context;
 struct anon__lazy_155 (*readFn)(void const *, struct anon__lazy_103);
};
typedef struct anon__lazy_177 nav__4193_64;
struct anon__lazy_177 {
 uint16_t error;
 uint8_t payload[4];
};
typedef struct anon__lazy_157 nav__4192_38;
typedef struct anon__lazy_181 nav__4235_38;
struct anon__lazy_181 {
 int32_t payload;
 uint16_t error;
};
typedef struct anon__lazy_155 nav__4235_50;
typedef struct anon__lazy_103 nav__4235_52;
typedef struct anon__lazy_177 nav__4235_64;
typedef struct anon__lazy_181 nav__4234_38;
typedef struct anon__lazy_103 nav__4237_39;
typedef struct anon__lazy_187 nav__4237_40;
struct anon__lazy_187 {
 struct anon__lazy_103 payload;
 uint16_t error;
};
typedef struct anon__lazy_155 nav__4237_59;
typedef struct anon__lazy_192 nav__4237_69;
struct anon__lazy_192 {
 uint64_t payload;
 uint16_t error;
};
typedef struct anon__lazy_194 nav__4237_72;
struct anon__lazy_194 {
 uint16_t error;
 uint8_t payload[8];
};
typedef struct anon__lazy_58 nav__4236_39;
typedef struct anon__lazy_197 nav__4236_40;
struct anon__lazy_197 {
 struct anon__lazy_58 payload;
 uint16_t error;
};
typedef struct anon__lazy_103 nav__4236_55;
typedef struct anon__lazy_187 nav__4236_57;
typedef struct anon__lazy_201 nav__4233_39;
typedef struct anon__lazy_58 nav__4233_44;
struct anon__lazy_201 {
 struct shared_Command__struct_1181__1181 payload;
 uint16_t error;
};
typedef struct anon__lazy_181 nav__4233_57;
typedef struct anon__lazy_197 nav__4233_60;
typedef struct anon__lazy_201 nav__4232_39;
typedef struct anon__lazy_58 nav__4232_44;
typedef struct anon__lazy_209 nav__4243_39;
typedef struct anon__lazy_58 nav__4243_44;
struct anon__lazy_209 {
 struct shared_Command__struct_1182__1182 payload;
 uint16_t error;
};
typedef struct anon__lazy_197 nav__4243_57;
typedef struct anon__lazy_209 nav__4242_39;
typedef struct anon__lazy_58 nav__4242_44;
typedef struct anon__lazy_192 nav__4249_38;
typedef struct anon__lazy_155 nav__4249_50;
typedef struct anon__lazy_103 nav__4249_52;
typedef struct anon__lazy_194 nav__4249_64;
typedef struct anon__lazy_192 nav__4248_38;
typedef struct anon__lazy_217 nav__4247_39;
struct anon__lazy_217 {
 struct shared_MarkerType__1185 payload;
 uint16_t error;
};
typedef struct anon__lazy_157 nav__4247_54;
typedef struct anon__lazy_192 nav__4247_56;
typedef struct anon__lazy_217 nav__4246_39;
typedef struct anon__lazy_221 nav__4245_39;
struct anon__lazy_221 {
 struct shared_Command__struct_1183__1183 payload;
 uint16_t error;
};
typedef struct anon__lazy_181 nav__4245_57;
typedef struct anon__lazy_217 nav__4245_60;
typedef struct anon__lazy_221 nav__4244_39;
typedef struct anon__lazy_226 nav__4251_38;
struct anon__lazy_226 {
 uint16_t error;
 uint8_t payload;
};
typedef struct anon__lazy_157 nav__4251_46;
typedef struct anon__lazy_226 nav__4250_38;
typedef struct anon__lazy_232 nav__4191_39;
typedef struct anon__lazy_58 nav__4191_45;
struct anon__lazy_232 {
 struct shared_Command__1178 payload;
 uint16_t error;
};
typedef struct anon__lazy_157 nav__4191_66;
typedef struct anon__lazy_201 nav__4191_68;
typedef struct anon__lazy_209 nav__4191_70;
typedef struct anon__lazy_221 nav__4191_72;
typedef struct anon__lazy_192 nav__4191_74;
typedef struct anon__lazy_226 nav__4191_76;
typedef struct anon__lazy_232 nav__4190_39;
typedef struct anon__lazy_58 nav__4190_45;
typedef struct anon__lazy_232 nav__1291_39;
typedef struct anon__lazy_58 nav__1291_45;
typedef struct anon__lazy_103 nav__1291_68;
typedef struct anon__lazy_155 nav__1291_76;
struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 {
 struct anon__lazy_103 buffer;
 uintptr_t pos;
};
typedef struct anon__lazy_232 nav__1292_39;
typedef struct anon__lazy_241 nav__1292_42;
typedef struct anon__lazy_58 nav__1292_46;
struct anon__lazy_241 {
 uint64_t payload;
 bool is_null;
};
typedef struct anon__lazy_103 nav__1292_69;
typedef struct anon__lazy_241 nav__1293_40;
typedef struct anon__lazy_58 nav__1293_48;
typedef struct anon__lazy_232 nav__1293_60;
typedef struct anon__lazy_103 nav__1293_70;
typedef struct anon__lazy_248 nav__1293_77;
struct anon__lazy_248 {
 struct shared_Command__1178 f0;
};
typedef struct anon__lazy_103 nav__1139_50;
typedef struct anon__lazy_58 nav__1139_63;
typedef struct anon__lazy_241 nav__1139_76;
typedef struct anon__lazy_58 nav__1130_43;
typedef struct anon__lazy_253 nav__1130_45;
struct anon__lazy_253 {
 struct anon__lazy_58 f0;
};
typedef struct anon__lazy_226 nav__1140_38;
typedef struct anon__lazy_103 nav__1140_53;
typedef struct anon__lazy_58 nav__1140_66;
typedef struct anon__lazy_241 nav__1140_79;
typedef struct anon__lazy_232 nav__1140_81;
struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591;
typedef struct anon__lazy_258 nav__820_39;
typedef struct anon__lazy_103 nav__820_50;
struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 {
 struct runner_fifo_RunnerFifo__631 fifo;
};
struct anon__lazy_258 {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 payload;
 uint16_t error;
};
typedef struct anon__lazy_116 nav__820_62;
typedef struct anon__lazy_226 nav__820_65;
struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582;
typedef struct anon__lazy_264 nav__807_39;
typedef struct anon__lazy_103 nav__807_50;
struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 {
 struct runner_fifo_RunnerFifo__631 fifo;
};
struct anon__lazy_264 {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 payload;
 uint16_t error;
};
typedef struct anon__lazy_116 nav__807_62;
typedef struct anon__lazy_226 nav__807_65;
struct instruments_root_Instrument__554;
typedef struct anon__lazy_270 nav__743_39;
struct instruments_valgrind_ValgrindInstrument__564;
struct instruments_valgrind_ValgrindInstrument__564 {
 struct mem_Allocator__599 allocator;
};
typedef struct anon__lazy_103 nav__743_53;
struct instruments_root_Instrument__554 {
 union {
  struct instruments_valgrind_ValgrindInstrument__564 valgrind;
  struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 walltime;
  struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 analysis;
 } payload;
 uint8_t tag;
};
struct anon__lazy_270 {
 struct instruments_root_Instrument__554 payload;
 uint16_t error;
};
typedef struct anon__lazy_258 nav__743_69;
typedef struct anon__lazy_264 nav__743_71;
struct InstrumentHooks;
typedef struct anon__lazy_279 nav__730_39;
typedef struct anon__lazy_103 nav__730_54;
struct environment_root_Environment__559;
struct environment_root_EnvironmentJson__833;
struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 {
 uint8_t *bytes;
 uintptr_t len;
 uintptr_t capacity;
};
struct array_hash_map_IndexHeader__880;
struct debug_SafetyLock__884;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 entries;
 struct array_hash_map_IndexHeader__880 *index_header;
};
struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 map;
};
struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 {
 uint8_t *bytes;
 uintptr_t len;
 uintptr_t capacity;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 entries;
 struct array_hash_map_IndexHeader__880 *index_header;
};
struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 map;
};
struct environment_root_EnvironmentJson__833 {
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 integration_environment;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 linked_libraries;
};
struct environment_linked_libraries_root_LinkedLibraries__837;
struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936;
struct environment_linked_libraries_root_LibraryInfo__934;
typedef struct anon__lazy_303 nav__730_90;
struct anon__lazy_303 {
 struct environment_linked_libraries_root_LibraryInfo__934 *ptr;
 uintptr_t len;
};
struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 {
 struct anon__lazy_303 items;
 uintptr_t capacity;
 struct mem_Allocator__599 allocator;
};
struct environment_linked_libraries_root_LinkedLibraries__837 {
 struct mem_Allocator__599 allocator;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 libraries;
};
struct environment_root_Environment__559 {
 struct mem_Allocator__599 allocator;
 struct environment_root_EnvironmentJson__833 data;
 struct environment_linked_libraries_root_LinkedLibraries__837 libs;
};
struct InstrumentHooks {
 struct instruments_root_Instrument__554 instrument;
 struct environment_root_Environment__559 environment;
};
struct anon__lazy_279 {
 struct InstrumentHooks payload;
 uint16_t error;
};
typedef struct anon__lazy_270 nav__730_100;
typedef struct anon__lazy_315 nav__238_46;
struct anon__lazy_315 {
 struct InstrumentHooks *payload;
 uint16_t error;
};
typedef struct anon__lazy_103 nav__238_51;
struct mem_Allocator_VTable__602 {
 uint8_t *(*alloc)(void *, uintptr_t, uint8_t, uintptr_t);
 bool (*resize)(void *, struct anon__lazy_103, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*remap)(void *, struct anon__lazy_103, uint8_t, uintptr_t, uintptr_t);
 void (*free)(void *, struct anon__lazy_103, uint8_t, uintptr_t);
};
typedef struct anon__lazy_303 nav__238_101;
typedef struct anon__lazy_279 nav__238_107;
typedef struct anon__lazy_315 nav__3706_40;
typedef struct anon__lazy_328 nav__3706_51;
struct anon__lazy_328 {
 uint8_t *payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__1348_39;
struct fs_Dir__1103;
struct fs_Dir__1103 {
 int32_t fd;
};
typedef struct anon__lazy_87 nav__1345_39;
typedef struct anon__lazy_58 nav__1345_41;
typedef struct anon__lazy_103 nav__1283_48;
typedef struct anon__lazy_103 nav__1288_48;
typedef struct anon__lazy_103 nav__1562_43;
typedef struct anon__lazy_58 nav__3956_46;
typedef struct anon__lazy_58 nav__1439_40;
typedef struct anon__lazy_155 nav__1439_47;
typedef struct anon__lazy_155 nav__1964_38;
typedef struct anon__lazy_152 nav__1964_41;
typedef struct anon__lazy_357 nav__1964_46;
struct anon__lazy_357 {
 uint32_t payload;
 bool is_null;
};
typedef struct anon__lazy_155 nav__1429_38;
typedef struct anon__lazy_103 nav__1429_41;
typedef struct anon__lazy_103 nav__1559_43;
typedef struct anon__lazy_103 nav__4156_40;
typedef struct anon__lazy_155 nav__4189_38;
typedef struct anon__lazy_103 nav__4189_41;
typedef struct anon__lazy_177 nav__4231_39;
typedef struct anon__lazy_155 nav__4231_44;
typedef struct anon__lazy_103 nav__4231_46;
typedef struct anon__lazy_194 nav__4238_39;
typedef struct anon__lazy_155 nav__4238_44;
typedef struct anon__lazy_103 nav__4238_46;
typedef struct anon__lazy_103 nav__4239_39;
typedef struct anon__lazy_187 nav__4239_40;
typedef struct anon__lazy_328 nav__4239_51;
typedef struct anon__lazy_155 nav__4196_38;
typedef struct anon__lazy_103 nav__4196_41;
typedef struct anon__lazy_58 nav__3829_43;
typedef struct anon__lazy_248 nav__4261_39;
typedef struct anon__lazy_58 nav__4261_43;
typedef struct anon__lazy_58 nav__4264_39;
typedef struct anon__lazy_253 nav__4264_40;
typedef struct anon__lazy_103 nav__1131_50;
typedef struct anon__lazy_303 nav__773_67;
typedef struct anon__lazy_328 nav__4265_39;
typedef struct anon__lazy_103 nav__4265_52;
typedef struct anon__lazy_58 nav__4300_39;
typedef struct anon__lazy_58 nav__3785_40;
typedef struct anon__lazy_402 nav__3785_49;
struct anon__lazy_402 {
 uint16_t error;
 uint8_t payload[1024];
};
typedef struct anon__lazy_87 nav__3721_39;
typedef struct anon__lazy_58 nav__3721_42;
typedef struct anon__lazy_402 nav__3721_53;
typedef struct anon__lazy_103 nav__1527_42;
typedef struct anon__lazy_103 nav__1564_43;
typedef struct anon__lazy_58 nav__4363_46;
typedef struct anon__lazy_155 nav__1438_38;
typedef struct anon__lazy_58 nav__1438_41;
typedef struct anon__lazy_357 nav__4364_38;
typedef struct anon__lazy_155 nav__1428_38;
typedef struct anon__lazy_103 nav__1428_41;
typedef struct anon__lazy_155 nav__4148_38;
typedef struct anon__lazy_103 nav__4148_42;
typedef struct anon__lazy_103 nav__4198_40;
typedef struct anon__lazy_155 nav__4198_43;
typedef struct anon__lazy_328 nav__4657_39;
typedef struct anon__lazy_155 nav__4657_49;
typedef struct anon__lazy_155 nav__4197_38;
typedef struct anon__lazy_103 nav__4197_41;
typedef struct anon__lazy_58 nav__4747_40;
typedef struct anon__lazy_103 nav__4747_51;
typedef struct anon__lazy_248 nav__4748_39;
typedef struct anon__lazy_58 nav__4748_43;
typedef struct anon__lazy_103 nav__4748_59;
typedef struct anon__lazy_187 nav__4748_63;
typedef struct anon__lazy_103 nav__4749_41;
typedef struct anon__lazy_187 nav__4749_45;
typedef struct anon__lazy_103 nav__4750_41;
typedef struct anon__lazy_187 nav__4750_45;
typedef struct anon__lazy_58 nav__4752_39;
typedef struct anon__lazy_253 nav__4752_40;
typedef struct anon__lazy_103 nav__4752_47;
typedef struct anon__lazy_187 nav__4752_51;
typedef struct anon__lazy_103 nav__1287_49;
typedef struct anon__lazy_103 nav__1294_53;
typedef struct anon__lazy_152 nav__1294_61;
typedef struct anon__lazy_155 nav__1294_64;
typedef struct anon__lazy_303 nav__2808_48;
typedef struct anon__lazy_58 nav__4306_39;
typedef struct anon__lazy_402 nav__2022_39;
typedef struct anon__lazy_58 nav__2022_41;
typedef struct anon__lazy_103 nav__2022_47;
typedef struct anon__lazy_87 nav__3722_39;
typedef struct anon__lazy_181 nav__3722_50;
typedef struct anon__lazy_155 nav__3722_52;
typedef struct anon__lazy_103 nav__1565_43;
typedef struct anon__lazy_187 nav__1565_55;
typedef struct anon__lazy_58 nav__4818_45;
typedef struct anon__lazy_58 nav__4820_45;
typedef struct anon__lazy_155 nav__1757_38;
typedef struct anon__lazy_58 nav__1757_40;
typedef struct anon__lazy_155 nav__1749_38;
typedef struct anon__lazy_103 nav__1749_40;
typedef struct anon__lazy_155 nav__4889_38;
typedef struct anon__lazy_469 nav__4889_42;
struct anon__lazy_469 {
 uintptr_t f0;
 uint8_t f1;
};
typedef struct anon__lazy_328 nav__4890_39;
typedef struct anon__lazy_103 nav__4890_52;
typedef struct anon__lazy_155 nav__4195_38;
typedef struct anon__lazy_103 nav__4195_41;
typedef struct anon__lazy_58 nav__4892_39;
struct io_Writer__3827;
typedef struct anon__lazy_248 nav__5177_40;
typedef struct anon__lazy_155 nav__5177_43;
typedef struct anon__lazy_58 nav__5177_45;
struct io_Writer__3827 {
 void const *context;
 struct anon__lazy_155 (*writeFn)(void const *, struct anon__lazy_58);
};
struct fmt_FormatOptions__4248;
typedef struct anon__lazy_484 nav__5177_67;
struct anon__lazy_484 {
 uintptr_t payload;
 bool is_null;
};
struct fmt_FormatOptions__4248 {
 struct anon__lazy_484 precision;
 struct anon__lazy_484 width;
 uint32_t fill;
 uint8_t alignment;
};
typedef struct anon__lazy_103 nav__5052_39;
typedef struct anon__lazy_187 nav__5052_40;
typedef struct anon__lazy_248 nav__5052_42;
typedef struct anon__lazy_58 nav__5052_48;
struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817;
struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *context;
};
typedef struct anon__lazy_155 nav__5052_71;
typedef struct anon__lazy_155 nav__5571_41;
typedef struct anon__lazy_58 nav__5571_43;
typedef struct anon__lazy_484 nav__5571_53;
typedef struct anon__lazy_103 nav__5054_39;
typedef struct anon__lazy_187 nav__5054_40;
typedef struct anon__lazy_155 nav__5054_54;
typedef struct anon__lazy_58 nav__5054_56;
typedef struct anon__lazy_155 nav__5573_41;
typedef struct anon__lazy_58 nav__5573_43;
typedef struct anon__lazy_103 nav__5055_39;
typedef struct anon__lazy_187 nav__5055_40;
typedef struct anon__lazy_155 nav__5055_54;
typedef struct anon__lazy_58 nav__5055_56;
typedef struct anon__lazy_58 nav__5574_40;
typedef struct anon__lazy_253 nav__5574_41;
typedef struct anon__lazy_155 nav__5574_44;
typedef struct anon__lazy_484 nav__5574_55;
typedef struct anon__lazy_103 nav__5056_39;
typedef struct anon__lazy_187 nav__5056_40;
typedef struct anon__lazy_58 nav__5056_42;
typedef struct anon__lazy_253 nav__5056_43;
typedef struct anon__lazy_155 nav__5056_59;
typedef struct anon__lazy_103 nav__1529_41;
typedef struct anon__lazy_303 nav__3512_43;
typedef struct anon__lazy_181 nav__1769_38;
typedef struct anon__lazy_155 nav__1928_38;
typedef struct anon__lazy_103 nav__1575_39;
typedef struct anon__lazy_103 nav__5136_39;
typedef struct anon__lazy_155 nav__5136_50;
typedef struct anon__lazy_103 nav__5137_39;
typedef struct anon__lazy_187 nav__5137_40;
typedef struct anon__lazy_328 nav__5137_51;
typedef struct anon__lazy_103 nav__5138_40;
typedef struct anon__lazy_155 nav__5139_48;
typedef struct anon__lazy_58 nav__5139_50;
typedef struct anon__lazy_58 nav__5140_45;
typedef struct anon__lazy_58 nav__5141_45;
typedef struct anon__lazy_155 nav__5143_48;
typedef struct anon__lazy_58 nav__5143_50;
typedef struct anon__lazy_155 nav__5160_38;
typedef struct anon__lazy_58 nav__5160_41;
typedef struct anon__lazy_58 nav__5168_40;
typedef struct anon__lazy_155 nav__5168_43;
typedef struct anon__lazy_58 nav__5585_44;
typedef struct anon__lazy_484 nav__5585_56;
typedef struct anon__lazy_155 nav__5585_60;
typedef struct anon__lazy_538 nav__5585_66;
struct anon__lazy_538 {
 int32_t f0;
 struct anon__lazy_58 f1;
};
typedef struct anon__lazy_543 nav__5585_76;
struct anon__lazy_543 {
 struct anon__lazy_58 f0;
 struct anon__lazy_58 f1;
};
typedef struct anon__lazy_545 nav__5585_78;
struct anon__lazy_545 {
 int32_t f0;
 struct shared_MarkerType__1185 f1;
};
typedef struct anon__lazy_547 nav__5585_80;
struct anon__lazy_547 {
 uint64_t f0;
};
typedef struct anon__lazy_550 nav__5585_84;
struct anon__lazy_550 {
 uint8_t f0;
};
typedef struct anon__lazy_58 nav__5570_44;
typedef struct anon__lazy_484 nav__5570_56;
typedef struct anon__lazy_155 nav__5570_60;
typedef struct anon__lazy_103 nav__4154_39;
typedef struct anon__lazy_484 nav__5572_41;
typedef struct anon__lazy_155 nav__5572_45;
typedef struct anon__lazy_58 nav__5572_47;
typedef struct anon__lazy_58 nav__5575_39;
typedef struct anon__lazy_484 nav__5575_44;
typedef struct anon__lazy_155 nav__5575_48;
typedef struct anon__lazy_103 nav__5580_39;
typedef struct anon__lazy_103 nav__5581_39;
typedef struct anon__lazy_155 nav__3955_38;
typedef struct anon__lazy_58 nav__3955_41;
typedef struct anon__lazy_58 nav__5583_40;
typedef struct anon__lazy_155 nav__4149_38;
typedef struct anon__lazy_58 nav__4149_42;
typedef struct anon__lazy_103 nav__4149_50;
typedef struct anon__lazy_155 nav__5167_38;
typedef struct anon__lazy_58 nav__5167_41;
typedef struct anon__lazy_58 nav__5586_40;
typedef struct anon__lazy_538 nav__5586_41;
typedef struct anon__lazy_155 nav__5586_44;
typedef struct anon__lazy_58 nav__5587_40;
typedef struct anon__lazy_543 nav__5587_41;
typedef struct anon__lazy_155 nav__5587_44;
typedef struct anon__lazy_545 nav__5588_40;
typedef struct anon__lazy_155 nav__5588_43;
typedef struct anon__lazy_58 nav__5588_45;
typedef struct anon__lazy_547 nav__5589_39;
typedef struct anon__lazy_155 nav__5589_42;
typedef struct anon__lazy_58 nav__5589_44;
typedef struct anon__lazy_550 nav__5590_39;
typedef struct anon__lazy_155 nav__5590_42;
typedef struct anon__lazy_58 nav__5590_44;
typedef struct anon__lazy_226 nav__5189_38;
typedef struct anon__lazy_155 nav__5205_38;
typedef struct anon__lazy_58 nav__5205_40;
typedef struct anon__lazy_226 nav__5205_48;
typedef struct anon__lazy_157 nav__5205_50;
typedef struct anon__lazy_58 nav__5592_39;
typedef struct anon__lazy_484 nav__5592_44;
typedef struct anon__lazy_155 nav__5592_48;
typedef struct anon__lazy_103 nav__5592_57;
typedef struct anon__lazy_226 nav__5592_60;
typedef struct anon__lazy_484 nav__5602_41;
typedef struct anon__lazy_155 nav__5602_45;
typedef struct anon__lazy_58 nav__5602_47;
typedef struct anon__lazy_595 nav__5602_56;
struct anon__lazy_595 {
 uint8_t array[2];
};
typedef struct anon__lazy_103 nav__5602_58;
typedef struct anon__lazy_484 nav__5601_41;
typedef struct anon__lazy_155 nav__5601_45;
typedef struct anon__lazy_58 nav__5601_47;
typedef struct anon__lazy_484 nav__5591_41;
typedef struct anon__lazy_155 nav__5591_45;
typedef struct anon__lazy_58 nav__5591_47;
typedef struct anon__lazy_155 nav__1553_38;
typedef struct anon__lazy_58 nav__1553_42;
typedef struct anon__lazy_155 nav__5593_48;
typedef struct anon__lazy_58 nav__5593_50;
typedef struct anon__lazy_58 nav__5594_40;
typedef struct anon__lazy_155 nav__5594_51;
typedef struct anon__lazy_58 nav__5596_40;
typedef struct anon__lazy_538 nav__5596_41;
typedef struct anon__lazy_155 nav__5596_44;
typedef struct anon__lazy_484 nav__5596_55;
typedef struct anon__lazy_58 nav__5597_40;
typedef struct anon__lazy_543 nav__5597_41;
typedef struct anon__lazy_155 nav__5597_44;
typedef struct anon__lazy_484 nav__5597_55;
typedef struct anon__lazy_545 nav__5598_40;
typedef struct anon__lazy_155 nav__5598_43;
typedef struct anon__lazy_58 nav__5598_45;
typedef struct anon__lazy_484 nav__5598_58;
typedef struct anon__lazy_547 nav__5599_39;
typedef struct anon__lazy_155 nav__5599_42;
typedef struct anon__lazy_58 nav__5599_44;
typedef struct anon__lazy_484 nav__5599_55;
typedef struct anon__lazy_550 nav__5600_39;
typedef struct anon__lazy_155 nav__5600_42;
typedef struct anon__lazy_58 nav__5600_44;
typedef struct anon__lazy_484 nav__5600_55;
typedef struct anon__lazy_595 nav__4953_39;
typedef struct anon__lazy_157 nav__5195_38;
typedef struct anon__lazy_58 nav__5195_40;
typedef struct anon__lazy_595 nav__5195_48;
typedef struct anon__lazy_615 nav__5195_52;
struct anon__lazy_615 {
 uint8_t array[3];
};
typedef struct anon__lazy_617 nav__5195_56;
struct anon__lazy_617 {
 uint8_t array[4];
};
typedef struct anon__lazy_226 nav__5188_38;
typedef struct anon__lazy_226 nav__5608_38;
typedef struct anon__lazy_103 nav__5608_40;
typedef struct anon__lazy_226 nav__5190_38;
typedef struct anon__lazy_103 nav__5190_40;
typedef struct anon__lazy_58 nav__5172_40;
typedef struct anon__lazy_155 nav__5172_43;
typedef struct anon__lazy_58 nav__1547_41;
typedef struct anon__lazy_484 nav__5603_41;
typedef struct anon__lazy_155 nav__5603_45;
typedef struct anon__lazy_58 nav__5603_47;
typedef struct anon__lazy_58 nav__5604_39;
typedef struct anon__lazy_484 nav__5604_44;
typedef struct anon__lazy_155 nav__5604_48;
typedef struct anon__lazy_484 nav__5610_44;
typedef struct anon__lazy_155 nav__5610_48;
typedef struct anon__lazy_58 nav__5610_50;
typedef struct anon__lazy_547 nav__5610_56;
typedef struct anon__lazy_484 nav__5605_44;
typedef struct anon__lazy_155 nav__5605_48;
typedef struct anon__lazy_58 nav__5605_50;
typedef struct anon__lazy_484 nav__5606_41;
typedef struct anon__lazy_155 nav__5606_45;
typedef struct anon__lazy_58 nav__5606_47;
typedef struct anon__lazy_484 nav__5607_41;
typedef struct anon__lazy_155 nav__5607_45;
typedef struct anon__lazy_58 nav__5607_47;
typedef struct anon__lazy_157 nav__5197_38;
typedef struct anon__lazy_595 nav__5197_40;
typedef struct anon__lazy_157 nav__5199_38;
typedef struct anon__lazy_615 nav__5199_40;
typedef struct anon__lazy_157 nav__5203_38;
typedef struct anon__lazy_617 nav__5203_40;
typedef struct anon__lazy_103 nav__1566_43;
typedef struct anon__lazy_155 nav__1566_52;
typedef struct anon__lazy_58 nav__1548_41;
typedef struct anon__lazy_103 nav__1548_47;
typedef struct anon__lazy_484 nav__5617_41;
typedef struct anon__lazy_155 nav__5617_45;
typedef struct anon__lazy_58 nav__5617_47;
typedef struct anon__lazy_595 nav__5617_56;
typedef struct anon__lazy_103 nav__5617_58;
typedef struct anon__lazy_484 nav__5616_41;
typedef struct anon__lazy_155 nav__5616_45;
typedef struct anon__lazy_58 nav__5616_47;
typedef struct anon__lazy_484 nav__5609_41;
typedef struct anon__lazy_155 nav__5609_45;
typedef struct anon__lazy_58 nav__5609_47;
typedef struct anon__lazy_547 nav__5611_39;
typedef struct anon__lazy_155 nav__5611_42;
typedef struct anon__lazy_58 nav__5611_44;
typedef struct anon__lazy_547 nav__5612_39;
typedef struct anon__lazy_155 nav__5612_42;
typedef struct anon__lazy_58 nav__5612_44;
typedef struct anon__lazy_547 nav__5613_39;
typedef struct anon__lazy_155 nav__5613_42;
typedef struct anon__lazy_58 nav__5613_44;
typedef struct anon__lazy_547 nav__5614_39;
typedef struct anon__lazy_155 nav__5614_42;
typedef struct anon__lazy_58 nav__5614_44;
typedef struct anon__lazy_484 nav__5622_41;
typedef struct anon__lazy_155 nav__5622_45;
typedef struct anon__lazy_58 nav__5622_47;
typedef struct anon__lazy_484 nav__5615_41;
typedef struct anon__lazy_155 nav__5615_45;
typedef struct anon__lazy_58 nav__5615_47;
typedef struct anon__lazy_157 nav__5201_38;
typedef struct anon__lazy_615 nav__5201_40;
typedef struct anon__lazy_155 nav__1491_38;
typedef struct anon__lazy_469 nav__1491_42;
typedef struct anon__lazy_547 nav__5618_39;
typedef struct anon__lazy_155 nav__5618_42;
typedef struct anon__lazy_58 nav__5618_44;
typedef struct anon__lazy_484 nav__5618_55;
typedef struct anon__lazy_547 nav__5619_39;
typedef struct anon__lazy_155 nav__5619_42;
typedef struct anon__lazy_58 nav__5619_44;
typedef struct anon__lazy_484 nav__5619_55;
typedef struct anon__lazy_547 nav__5620_39;
typedef struct anon__lazy_155 nav__5620_42;
typedef struct anon__lazy_58 nav__5620_44;
typedef struct anon__lazy_484 nav__5620_55;
typedef struct anon__lazy_547 nav__5621_39;
typedef struct anon__lazy_155 nav__5621_42;
typedef struct anon__lazy_58 nav__5621_44;
typedef struct anon__lazy_484 nav__5621_55;
typedef struct anon__lazy_103 nav__239_50;
typedef struct anon__lazy_103 nav__731_55;
typedef struct anon__lazy_303 nav__731_91;
typedef struct anon__lazy_103 nav__5623_52;
typedef struct anon__lazy_103 nav__808_51;
typedef struct anon__lazy_103 nav__821_51;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086;
typedef struct anon__lazy_58 nav__774_44;
struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 {
 struct anon__lazy_58 *keys;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *values;
 uint32_t len;
 uint32_t index;
};
typedef struct anon__lazy_303 nav__774_76;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 {
 struct anon__lazy_58 *key_ptr;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *value_ptr;
};
typedef struct anon__lazy_676 nav__774_87;
struct anon__lazy_676 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 payload;
 bool is_null;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107;
struct environment_root_EnvironmentValue__844;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 {
 struct anon__lazy_58 *keys;
 struct environment_root_EnvironmentValue__844 *values;
 uint32_t len;
 uint32_t index;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 {
 uint8_t *bytes;
 uintptr_t len;
 uintptr_t capacity;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 entries;
 struct array_hash_map_IndexHeader__880 *index_header;
};
struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 map;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6115;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6115 {
 struct anon__lazy_58 *key_ptr;
 struct environment_root_EnvironmentValue__844 *value_ptr;
};
typedef struct anon__lazy_693 nav__774_104;
struct anon__lazy_693 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6115 payload;
 bool is_null;
};
typedef struct anon__lazy_695 nav__774_109;
struct anon__lazy_695 {
 struct anon__lazy_58 const *ptr;
 uintptr_t len;
};
struct environment_root_EnvironmentValue__844 {
 union {
  struct anon__lazy_58 string;
  struct anon__lazy_695 list;
 } payload;
 uint8_t tag;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139;
struct environment_linked_libraries_root_LibraryEntry__852;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 {
 struct anon__lazy_58 *keys;
 struct environment_linked_libraries_root_LibraryEntry__852 *values;
 uint32_t len;
 uint32_t index;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147 {
 struct anon__lazy_58 *key_ptr;
 struct environment_linked_libraries_root_LibraryEntry__852 *value_ptr;
};
typedef struct anon__lazy_708 nav__774_122;
struct anon__lazy_708 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147 payload;
 bool is_null;
};
typedef struct anon__lazy_58 nav__2965_42;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 {
 uint8_t *ptrs[3];
 uintptr_t len;
 uintptr_t capacity;
};
typedef struct anon__lazy_719 nav__2965_62;
struct anon__lazy_719 {
 struct anon__lazy_58 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_722 nav__2965_65;
struct anon__lazy_722 {
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_676 nav__5624_39;
typedef struct anon__lazy_58 nav__5624_44;
typedef struct anon__lazy_58 nav__3176_42;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 {
 uint8_t *ptrs[3];
 uintptr_t len;
 uintptr_t capacity;
};
typedef struct anon__lazy_719 nav__3176_62;
typedef struct anon__lazy_735 nav__3176_65;
struct anon__lazy_735 {
 struct environment_root_EnvironmentValue__844 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_693 nav__5626_39;
typedef struct anon__lazy_58 nav__5626_44;
typedef struct anon__lazy_58 nav__777_43;
typedef struct anon__lazy_695 nav__777_46;
typedef struct anon__lazy_303 nav__777_79;
typedef struct anon__lazy_58 nav__3355_42;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 {
 uint8_t *ptrs[3];
 uintptr_t len;
 uintptr_t capacity;
};
typedef struct anon__lazy_719 nav__3355_62;
typedef struct anon__lazy_760 nav__3355_65;
struct anon__lazy_760 {
 struct environment_linked_libraries_root_LibraryEntry__852 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_708 nav__5628_39;
typedef struct anon__lazy_58 nav__5628_44;
typedef struct anon__lazy_303 nav__2809_50;
typedef struct anon__lazy_58 nav__2809_56;
struct environment_linked_libraries_root_LibraryInfo__934 {
 struct anon__lazy_58 path;
 struct anon__lazy_58 soname;
 struct anon__lazy_58 build_id;
};
typedef struct anon__lazy_58 nav__5643_39;
typedef struct anon__lazy_719 nav__5643_41;
typedef struct anon__lazy_722 nav__5644_40;
typedef struct anon__lazy_58 nav__5652_39;
typedef struct anon__lazy_719 nav__5652_41;
typedef struct anon__lazy_735 nav__5653_40;
typedef struct anon__lazy_58 nav__5653_51;
typedef struct anon__lazy_695 nav__5653_54;
typedef struct anon__lazy_58 nav__5654_40;
typedef struct anon__lazy_695 nav__5654_42;
typedef struct anon__lazy_103 nav__5654_54;
typedef struct anon__lazy_103 nav__3294_51;
struct array_hash_map_IndexHeader__880 {
 zig_align(4) uint8_t bit_index;
};
typedef struct anon__lazy_103 nav__3144_52;
typedef struct anon__lazy_103 nav__3100_51;
typedef struct anon__lazy_58 nav__5662_39;
typedef struct anon__lazy_719 nav__5662_41;
typedef struct anon__lazy_760 nav__5663_40;
typedef struct anon__lazy_58 nav__5663_51;
struct environment_linked_libraries_root_LibraryEntry__852 {
 struct anon__lazy_58 path;
 struct anon__lazy_58 build_id;
};
typedef struct anon__lazy_103 nav__3473_51;
typedef struct anon__lazy_58 nav__2810_46;
typedef struct anon__lazy_303 nav__3514_42;
typedef struct anon__lazy_58 nav__5735_39;
typedef struct anon__lazy_695 nav__5735_41;
typedef struct anon__lazy_103 nav__3326_39;
typedef struct anon__lazy_103 nav__5736_40;
typedef struct anon__lazy_103 nav__5737_40;
typedef struct anon__lazy_103 nav__3132_39;
typedef struct anon__lazy_103 nav__3505_39;
typedef struct anon__lazy_303 nav__3560_40;
typedef struct anon__lazy_58 nav__3560_55;
typedef struct anon__lazy_303 nav__5738_41;
typedef struct anon__lazy_103 nav__5738_51;
typedef struct anon__lazy_103 nav__5739_39;
typedef struct anon__lazy_103 nav__5755_39;
typedef struct anon__lazy_103 nav__5756_39;
typedef struct anon__lazy_303 nav__5756_42;
typedef struct anon__lazy_103 nav__240_55;
typedef struct anon__lazy_303 nav__240_91;
typedef struct anon__lazy_103 nav__1134_50;
typedef struct anon__lazy_58 nav__1134_63;
typedef struct anon__lazy_241 nav__1134_76;
typedef struct anon__lazy_103 nav__810_51;
typedef struct anon__lazy_103 nav__823_51;
typedef struct anon__lazy_103 nav__241_55;
typedef struct anon__lazy_303 nav__241_91;
typedef struct anon__lazy_103 nav__1135_50;
typedef struct anon__lazy_58 nav__1135_63;
typedef struct anon__lazy_241 nav__1135_76;
typedef struct anon__lazy_103 nav__811_51;
typedef struct anon__lazy_103 nav__824_51;
typedef struct anon__lazy_103 nav__242_55;
typedef struct anon__lazy_303 nav__242_91;
typedef struct anon__lazy_103 nav__1136_51;
typedef struct anon__lazy_58 nav__1136_61;
typedef struct anon__lazy_241 nav__1136_76;
typedef struct anon__lazy_103 nav__812_52;
typedef struct anon__lazy_103 nav__825_52;
typedef struct anon__lazy_103 nav__243_56;
typedef struct anon__lazy_303 nav__243_92;
typedef struct anon__lazy_58 nav__5771_39;
typedef struct anon__lazy_853 nav__5773_45;
struct anon__lazy_853 {
 bool array[16];
};
typedef struct anon__lazy_855 nav__5773_47;
struct anon__lazy_855 {
 bool is_null;
 uint8_t payload;
};
typedef struct anon__lazy_855 nav__5775_38;
typedef struct anon__lazy_853 nav__5775_40;
typedef struct anon__lazy_103 nav__1137_51;
typedef struct anon__lazy_58 nav__1137_61;
typedef struct anon__lazy_241 nav__1137_76;
typedef struct anon__lazy_103 nav__813_52;
typedef struct anon__lazy_103 nav__826_52;
typedef struct anon__lazy_103 nav__245_56;
typedef struct anon__lazy_303 nav__245_92;
typedef struct anon__lazy_870 nav__245_103;
struct anon__lazy_870 {
 uint8_t const *f0;
 uint8_t const *f1;
};
typedef struct anon__lazy_187 nav__245_105;
typedef struct anon__lazy_103 nav__5783_39;
typedef struct anon__lazy_187 nav__5783_40;
typedef struct anon__lazy_870 nav__5783_43;
typedef struct anon__lazy_103 nav__5784_40;
typedef struct anon__lazy_103 nav__5785_39;
typedef struct anon__lazy_187 nav__5785_40;
typedef struct anon__lazy_870 nav__5785_43;
typedef struct anon__lazy_484 nav__5785_54;
typedef struct anon__lazy_103 nav__5786_39;
typedef struct anon__lazy_870 nav__5828_40;
typedef struct anon__lazy_155 nav__5828_43;
typedef struct anon__lazy_58 nav__5828_44;
typedef struct anon__lazy_484 nav__5828_55;
typedef struct anon__lazy_870 nav__5787_39;
struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119;
struct io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29__7111;
struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 {
 uint64_t bytes_written;
};
struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133;
struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *context;
};
typedef struct anon__lazy_155 nav__5787_53;
typedef struct anon__lazy_58 nav__5787_54;
typedef struct anon__lazy_484 nav__5788_38;
typedef struct anon__lazy_103 nav__5789_39;
typedef struct anon__lazy_187 nav__5789_40;
typedef struct anon__lazy_870 nav__5789_42;
typedef struct anon__lazy_155 nav__5789_57;
typedef struct anon__lazy_58 nav__5789_58;
typedef struct anon__lazy_155 nav__5827_38;
typedef struct anon__lazy_58 nav__5827_41;
typedef struct anon__lazy_484 nav__5829_42;
typedef struct anon__lazy_155 nav__5829_46;
typedef struct anon__lazy_58 nav__5829_47;
typedef struct anon__lazy_155 nav__5812_38;
typedef struct anon__lazy_58 nav__5812_42;
typedef struct anon__lazy_155 nav__3933_38;
typedef struct anon__lazy_58 nav__3933_40;
typedef struct anon__lazy_103 nav__1138_54;
typedef struct anon__lazy_58 nav__1138_69;
typedef struct anon__lazy_241 nav__1138_77;
typedef struct anon__lazy_103 nav__814_55;
typedef struct anon__lazy_103 nav__827_55;
typedef struct anon__lazy_103 nav__250_58;
typedef struct anon__lazy_303 nav__250_94;
struct once_Once_28_28function_20_27init_27_29_29__7273;
struct Thread_Mutex__7278;
struct Thread_Mutex_DarwinImpl__7290;
struct c_darwin_os_unfair_lock__7296;
struct c_darwin_os_unfair_lock__7296 {
 uint32_t _os_unfair_lock_opaque;
};
struct Thread_Mutex_DarwinImpl__7290 {
 struct c_darwin_os_unfair_lock__7296 oul;
};
struct Thread_Mutex__7278 {
 struct Thread_Mutex_DarwinImpl__7290 impl;
};
struct once_Once_28_28function_20_27init_27_29_29__7273 {
 struct Thread_Mutex__7278 mutex;
 bool done;
};
struct c_MachTimebaseInfo__7314;
struct c_MachTimebaseInfo__7314 {
 uint32_t numer;
 uint32_t denom;
};
typedef struct anon__lazy_58 nav__778_41;
typedef struct anon__lazy_695 nav__778_46;
typedef struct anon__lazy_303 nav__778_79;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 {
 struct anon__lazy_58 *key_ptr;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *value_ptr;
 uintptr_t index;
 bool found_existing;
};
typedef struct anon__lazy_935 nav__778_93;
struct anon__lazy_935 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 payload;
 uint16_t error;
};
typedef struct anon__lazy_103 nav__778_96;
typedef struct anon__lazy_187 nav__778_98;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 {
 struct anon__lazy_58 *key_ptr;
 struct environment_root_EnvironmentValue__844 *value_ptr;
 uintptr_t index;
 bool found_existing;
};
typedef struct anon__lazy_940 nav__778_110;
struct anon__lazy_940 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__775_41;
typedef struct anon__lazy_303 nav__775_73;
typedef struct anon__lazy_103 nav__775_79;
typedef struct anon__lazy_187 nav__775_81;
typedef struct anon__lazy_695 nav__775_85;
typedef struct anon__lazy_103 nav__255_56;
typedef struct anon__lazy_303 nav__255_92;
typedef struct anon__lazy_58 nav__255_100;
typedef struct anon__lazy_103 nav__5934_39;
typedef struct anon__lazy_187 nav__5934_40;
typedef struct anon__lazy_58 nav__5934_43;
typedef struct anon__lazy_935 nav__2967_39;
typedef struct anon__lazy_58 nav__2967_44;
struct array_hash_map_StringContext__866;
typedef struct anon__lazy_940 nav__3178_39;
typedef struct anon__lazy_58 nav__3178_44;
typedef struct anon__lazy_935 nav__2968_39;
typedef struct anon__lazy_58 nav__2968_44;
typedef struct anon__lazy_940 nav__3179_39;
typedef struct anon__lazy_58 nav__3179_44;
typedef struct anon__lazy_935 nav__5935_39;
typedef struct anon__lazy_58 nav__5935_44;
typedef struct anon__lazy_484 nav__5935_67;
typedef struct anon__lazy_719 nav__5935_73;
typedef struct anon__lazy_722 nav__5935_76;
typedef struct anon__lazy_940 nav__5936_39;
typedef struct anon__lazy_58 nav__5936_44;
typedef struct anon__lazy_484 nav__5936_67;
typedef struct anon__lazy_719 nav__5936_73;
typedef struct anon__lazy_735 nav__5936_76;
typedef struct anon__lazy_226 nav__2977_60;
typedef struct anon__lazy_957 nav__2977_62;
struct anon__lazy_957 {
 struct array_hash_map_IndexHeader__880 *payload;
 uint16_t error;
};
typedef struct anon__lazy_484 nav__5937_38;
typedef struct anon__lazy_58 nav__5937_41;
typedef struct anon__lazy_961 nav__5937_61;
struct anon__lazy_961 {
 uint32_t *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_719 nav__5937_64;
typedef struct anon__lazy_58 nav__5938_42;
typedef struct anon__lazy_961 nav__5938_66;
typedef struct anon__lazy_719 nav__5938_69;
typedef struct anon__lazy_722 nav__5938_72;
typedef struct anon__lazy_226 nav__3188_60;
typedef struct anon__lazy_957 nav__3188_62;
typedef struct anon__lazy_484 nav__5939_38;
typedef struct anon__lazy_58 nav__5939_41;
typedef struct anon__lazy_961 nav__5939_61;
typedef struct anon__lazy_719 nav__5939_64;
typedef struct anon__lazy_58 nav__5940_42;
typedef struct anon__lazy_961 nav__5940_66;
typedef struct anon__lazy_719 nav__5940_69;
typedef struct anon__lazy_735 nav__5940_72;
typedef struct anon__lazy_103 nav__3124_52;
typedef struct anon__lazy_187 nav__3124_54;
typedef struct anon__lazy_961 nav__3124_62;
typedef struct anon__lazy_58 nav__3124_65;
typedef struct anon__lazy_719 nav__3124_67;
typedef struct anon__lazy_722 nav__3124_72;
typedef struct anon__lazy_226 nav__3142_38;
typedef struct anon__lazy_957 nav__3143_40;
typedef struct anon__lazy_103 nav__3143_51;
typedef struct anon__lazy_187 nav__3143_53;
typedef struct anon__lazy_58 nav__5941_40;
typedef struct anon__lazy_961 nav__5942_39;
typedef struct anon__lazy_58 nav__5943_40;
typedef struct anon__lazy_484 nav__5944_38;
typedef struct anon__lazy_58 nav__5944_41;
struct array_hash_map_Index_28u8_29__6669;
typedef struct anon__lazy_996 nav__5944_58;
struct anon__lazy_996 {
 struct array_hash_map_Index_28u8_29__6669 *ptr;
 uintptr_t len;
};
struct array_hash_map_Index_28u8_29__6669 {
 uint8_t entry_index;
 uint8_t distance_from_start_index;
};
typedef struct anon__lazy_484 nav__5945_38;
typedef struct anon__lazy_58 nav__5945_41;
struct array_hash_map_Index_28u16_29__6673;
typedef struct anon__lazy_1001 nav__5945_58;
struct anon__lazy_1001 {
 struct array_hash_map_Index_28u16_29__6673 *ptr;
 uintptr_t len;
};
struct array_hash_map_Index_28u16_29__6673 {
 uint16_t entry_index;
 uint16_t distance_from_start_index;
};
typedef struct anon__lazy_484 nav__5946_38;
typedef struct anon__lazy_58 nav__5946_41;
struct array_hash_map_Index_28u32_29__6677;
typedef struct anon__lazy_1006 nav__5946_58;
struct anon__lazy_1006 {
 struct array_hash_map_Index_28u32_29__6677 *ptr;
 uintptr_t len;
};
struct array_hash_map_Index_28u32_29__6677 {
 uint32_t entry_index;
 uint32_t distance_from_start_index;
};
typedef struct anon__lazy_58 nav__5947_42;
typedef struct anon__lazy_961 nav__5947_66;
typedef struct anon__lazy_719 nav__5947_69;
typedef struct anon__lazy_722 nav__5947_72;
typedef struct anon__lazy_996 nav__5947_77;
typedef struct anon__lazy_58 nav__5948_42;
typedef struct anon__lazy_961 nav__5948_66;
typedef struct anon__lazy_719 nav__5948_69;
typedef struct anon__lazy_722 nav__5948_72;
typedef struct anon__lazy_1001 nav__5948_77;
typedef struct anon__lazy_58 nav__5949_42;
typedef struct anon__lazy_961 nav__5949_66;
typedef struct anon__lazy_719 nav__5949_69;
typedef struct anon__lazy_722 nav__5949_72;
typedef struct anon__lazy_1006 nav__5949_77;
typedef struct anon__lazy_103 nav__3318_52;
typedef struct anon__lazy_187 nav__3318_54;
typedef struct anon__lazy_961 nav__3318_62;
typedef struct anon__lazy_58 nav__3318_65;
typedef struct anon__lazy_719 nav__3318_67;
typedef struct anon__lazy_735 nav__3318_72;
typedef struct anon__lazy_695 nav__3318_75;
typedef struct anon__lazy_58 nav__5950_40;
typedef struct anon__lazy_961 nav__5951_39;
typedef struct anon__lazy_58 nav__5952_40;
typedef struct anon__lazy_484 nav__5953_38;
typedef struct anon__lazy_58 nav__5953_41;
typedef struct anon__lazy_996 nav__5953_58;
typedef struct anon__lazy_484 nav__5954_38;
typedef struct anon__lazy_58 nav__5954_41;
typedef struct anon__lazy_1001 nav__5954_58;
typedef struct anon__lazy_484 nav__5955_38;
typedef struct anon__lazy_58 nav__5955_41;
typedef struct anon__lazy_1006 nav__5955_58;
typedef struct anon__lazy_58 nav__5956_42;
typedef struct anon__lazy_961 nav__5956_66;
typedef struct anon__lazy_719 nav__5956_69;
typedef struct anon__lazy_735 nav__5956_72;
typedef struct anon__lazy_996 nav__5956_77;
typedef struct anon__lazy_58 nav__5957_42;
typedef struct anon__lazy_961 nav__5957_66;
typedef struct anon__lazy_719 nav__5957_69;
typedef struct anon__lazy_735 nav__5957_72;
typedef struct anon__lazy_1001 nav__5957_77;
typedef struct anon__lazy_58 nav__5958_42;
typedef struct anon__lazy_961 nav__5958_66;
typedef struct anon__lazy_719 nav__5958_69;
typedef struct anon__lazy_735 nav__5958_72;
typedef struct anon__lazy_1006 nav__5958_77;
typedef struct anon__lazy_103 nav__5959_39;
typedef struct anon__lazy_187 nav__5959_40;
typedef struct anon__lazy_328 nav__5959_51;
typedef struct anon__lazy_103 nav__5961_39;
typedef struct anon__lazy_187 nav__5961_40;
typedef struct anon__lazy_328 nav__5961_51;
typedef struct anon__lazy_961 nav__5962_56;
typedef struct anon__lazy_996 nav__5962_60;
typedef struct anon__lazy_961 nav__5963_56;
typedef struct anon__lazy_1001 nav__5963_60;
typedef struct anon__lazy_961 nav__5964_56;
typedef struct anon__lazy_1006 nav__5964_60;
typedef struct anon__lazy_58 nav__2938_40;
typedef struct anon__lazy_58 nav__2939_40;
typedef struct anon__lazy_996 nav__5965_40;
typedef struct anon__lazy_484 nav__5966_38;
typedef struct anon__lazy_58 nav__5966_41;
typedef struct anon__lazy_996 nav__5966_47;
typedef struct anon__lazy_961 nav__5966_66;
typedef struct anon__lazy_719 nav__5966_69;
typedef struct anon__lazy_1001 nav__5967_40;
typedef struct anon__lazy_484 nav__5968_38;
typedef struct anon__lazy_58 nav__5968_41;
typedef struct anon__lazy_1001 nav__5968_47;
typedef struct anon__lazy_961 nav__5968_66;
typedef struct anon__lazy_719 nav__5968_69;
typedef struct anon__lazy_1006 nav__5969_40;
typedef struct anon__lazy_484 nav__5970_38;
typedef struct anon__lazy_58 nav__5970_41;
typedef struct anon__lazy_1006 nav__5970_47;
typedef struct anon__lazy_961 nav__5970_66;
typedef struct anon__lazy_719 nav__5970_69;
typedef struct anon__lazy_961 nav__5972_56;
typedef struct anon__lazy_996 nav__5972_60;
typedef struct anon__lazy_961 nav__5973_56;
typedef struct anon__lazy_1001 nav__5973_60;
typedef struct anon__lazy_961 nav__5974_56;
typedef struct anon__lazy_1006 nav__5974_60;
typedef struct anon__lazy_484 nav__5975_38;
typedef struct anon__lazy_58 nav__5975_41;
typedef struct anon__lazy_996 nav__5975_47;
typedef struct anon__lazy_961 nav__5975_66;
typedef struct anon__lazy_719 nav__5975_69;
typedef struct anon__lazy_484 nav__5976_38;
typedef struct anon__lazy_58 nav__5976_41;
typedef struct anon__lazy_1001 nav__5976_47;
typedef struct anon__lazy_961 nav__5976_66;
typedef struct anon__lazy_719 nav__5976_69;
typedef struct anon__lazy_484 nav__5977_38;
typedef struct anon__lazy_58 nav__5977_41;
typedef struct anon__lazy_1006 nav__5977_47;
typedef struct anon__lazy_961 nav__5977_66;
typedef struct anon__lazy_719 nav__5977_69;
typedef struct anon__lazy_328 nav__5978_39;
typedef struct anon__lazy_155 nav__5978_49;
typedef struct anon__lazy_328 nav__5980_39;
typedef struct anon__lazy_155 nav__5980_49;
typedef struct anon__lazy_58 nav__2901_39;
typedef struct anon__lazy_58 nav__2900_39;
typedef struct anon__lazy_328 nav__5981_39;
typedef struct anon__lazy_103 nav__5981_52;
typedef struct anon__lazy_58 nav__6041_39;
struct hash_wyhash_Wyhash__7909;
struct hash_wyhash_Wyhash__7909 {
 uint64_t a;
 uint64_t b;
 uint64_t state[3];
 uintptr_t total_len;
 uintptr_t buf_len;
 uint8_t buf[48];
};
typedef struct anon__lazy_58 nav__6042_39;
typedef struct anon__lazy_58 nav__856_39;
typedef struct anon__lazy_58 nav__776_41;
typedef struct anon__lazy_695 nav__776_43;
typedef struct anon__lazy_303 nav__776_76;
typedef struct anon__lazy_719 nav__776_83;
typedef struct anon__lazy_1080 nav__776_85;
struct anon__lazy_1080 {
 struct anon__lazy_719 payload;
 uint16_t error;
};
typedef struct anon__lazy_103 nav__776_88;
typedef struct anon__lazy_187 nav__776_90;
typedef struct anon__lazy_58 nav__256_45;
typedef struct anon__lazy_719 nav__256_47;
typedef struct anon__lazy_1080 nav__256_54;
typedef struct anon__lazy_103 nav__256_59;
typedef struct anon__lazy_303 nav__256_112;
typedef struct anon__lazy_695 nav__256_120;
typedef struct anon__lazy_58 nav__6046_39;
typedef struct anon__lazy_719 nav__6046_41;
typedef struct anon__lazy_1080 nav__6046_42;
typedef struct anon__lazy_328 nav__6046_54;
typedef struct anon__lazy_58 nav__6047_40;
typedef struct anon__lazy_719 nav__6047_42;
typedef struct anon__lazy_103 nav__6047_52;
typedef struct anon__lazy_328 nav__6048_39;
typedef struct anon__lazy_155 nav__6048_49;
typedef struct anon__lazy_103 nav__6049_39;
typedef struct anon__lazy_58 nav__6049_41;
typedef struct anon__lazy_719 nav__6049_43;
typedef struct anon__lazy_103 nav__257_55;
typedef struct anon__lazy_303 nav__257_91;
typedef struct anon__lazy_58 nav__779_44;
typedef struct anon__lazy_303 nav__779_76;
typedef struct anon__lazy_708 nav__779_87;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 {
 struct anon__lazy_58 *key_ptr;
 struct environment_linked_libraries_root_LibraryEntry__852 *value_ptr;
 uintptr_t index;
 bool found_existing;
};
typedef struct anon__lazy_1095 nav__779_95;
struct anon__lazy_1095 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 payload;
 uint16_t error;
};
typedef struct anon__lazy_103 nav__779_97;
typedef struct anon__lazy_187 nav__779_99;
typedef struct anon__lazy_303 nav__780_69;
typedef struct anon__lazy_58 nav__780_83;
typedef struct anon__lazy_103 nav__780_86;
typedef struct anon__lazy_1099 nav__780_89;
struct anon__lazy_1099 {
 struct anon__lazy_58 f0;
 int32_t f1;
};
typedef struct anon__lazy_187 nav__780_91;
struct json_stringify_StringifyOptions__8122;
struct json_stringify_StringifyOptions__8122 {
 uint8_t whitespace;
 bool emit_null_optional_fields;
 bool emit_strings_as_arrays;
 bool escape_unicode;
 bool emit_nonportable_numbers_as_strings;
};
struct fs_File_CreateFlags__8142;
struct fs_File_CreateFlags__8142 {
 uint16_t mode;
 bool read;
 bool truncate;
 bool exclusive;
 uint8_t lock;
 bool lock_nonblocking;
};
typedef struct anon__lazy_87 nav__780_100;
typedef struct anon__lazy_996 nav__3350_57;
typedef struct anon__lazy_1001 nav__3350_62;
typedef struct anon__lazy_1006 nav__3350_67;
typedef struct anon__lazy_1095 nav__3357_39;
typedef struct anon__lazy_58 nav__3357_44;
typedef struct anon__lazy_58 nav__6052_39;
typedef struct anon__lazy_58 nav__6081_40;
typedef struct anon__lazy_1099 nav__6081_41;
typedef struct anon__lazy_155 nav__6081_44;
typedef struct anon__lazy_484 nav__6081_53;
typedef struct anon__lazy_103 nav__6053_39;
typedef struct anon__lazy_187 nav__6053_40;
typedef struct anon__lazy_58 nav__6053_42;
typedef struct anon__lazy_1099 nav__6053_43;
typedef struct anon__lazy_155 nav__6053_59;
typedef struct anon__lazy_103 nav__6076_39;
typedef struct anon__lazy_187 nav__6076_40;
typedef struct anon__lazy_87 nav__1351_39;
typedef struct anon__lazy_58 nav__1351_41;
typedef struct anon__lazy_1095 nav__3358_39;
typedef struct anon__lazy_58 nav__3358_44;
typedef struct anon__lazy_103 nav__6079_41;
typedef struct anon__lazy_187 nav__6079_45;
typedef struct anon__lazy_103 nav__6082_41;
typedef struct anon__lazy_187 nav__6082_45;
struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244;
struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 stream;
 uintptr_t indent_level;
 struct json_stringify_StringifyOptions__8122 options;
 uint8_t next_punctuation;
};
typedef struct anon__lazy_103 nav__1533_39;
typedef struct anon__lazy_187 nav__1533_40;
typedef struct anon__lazy_103 nav__6113_41;
typedef struct anon__lazy_187 nav__6113_45;
typedef struct anon__lazy_87 nav__3724_39;
typedef struct anon__lazy_58 nav__3724_42;
typedef struct anon__lazy_402 nav__3724_53;
typedef struct anon__lazy_103 nav__6114_41;
typedef struct anon__lazy_187 nav__6114_45;
typedef struct anon__lazy_1095 nav__6115_39;
typedef struct anon__lazy_58 nav__6115_44;
typedef struct anon__lazy_484 nav__6115_67;
typedef struct anon__lazy_719 nav__6115_73;
typedef struct anon__lazy_760 nav__6115_76;
typedef struct anon__lazy_155 nav__6124_41;
typedef struct anon__lazy_58 nav__6124_43;
typedef struct anon__lazy_103 nav__6116_39;
typedef struct anon__lazy_187 nav__6116_40;
typedef struct anon__lazy_155 nav__6116_54;
typedef struct anon__lazy_58 nav__6116_56;
typedef struct anon__lazy_155 nav__6125_41;
typedef struct anon__lazy_58 nav__6125_43;
typedef struct anon__lazy_103 nav__6117_39;
typedef struct anon__lazy_187 nav__6117_40;
typedef struct anon__lazy_155 nav__6117_54;
typedef struct anon__lazy_58 nav__6117_56;
typedef struct anon__lazy_58 nav__6119_62;
typedef struct anon__lazy_103 nav__1563_43;
typedef struct anon__lazy_155 nav__6128_41;
typedef struct anon__lazy_58 nav__6128_43;
typedef struct anon__lazy_103 nav__6120_39;
typedef struct anon__lazy_187 nav__6120_40;
typedef struct anon__lazy_155 nav__6120_54;
typedef struct anon__lazy_58 nav__6120_56;
typedef struct anon__lazy_87 nav__3725_39;
typedef struct anon__lazy_181 nav__3725_50;
typedef struct anon__lazy_155 nav__3725_52;
typedef struct anon__lazy_155 nav__6129_41;
typedef struct anon__lazy_58 nav__6129_43;
typedef struct anon__lazy_103 nav__6121_39;
typedef struct anon__lazy_187 nav__6121_40;
typedef struct anon__lazy_155 nav__6121_54;
typedef struct anon__lazy_58 nav__6121_56;
typedef struct anon__lazy_226 nav__3367_60;
typedef struct anon__lazy_957 nav__3367_62;
typedef struct anon__lazy_484 nav__6122_38;
typedef struct anon__lazy_58 nav__6122_41;
typedef struct anon__lazy_961 nav__6122_61;
typedef struct anon__lazy_719 nav__6122_64;
typedef struct anon__lazy_58 nav__6123_42;
typedef struct anon__lazy_961 nav__6123_66;
typedef struct anon__lazy_719 nav__6123_69;
typedef struct anon__lazy_760 nav__6123_72;
typedef struct anon__lazy_155 nav__6096_56;
typedef struct anon__lazy_58 nav__6096_58;
typedef struct anon__lazy_155 nav__6099_56;
typedef struct anon__lazy_58 nav__6099_58;
typedef struct anon__lazy_1152 nav__6097_43;
struct anon__lazy_1152 {
 bool is_null;
 bool payload;
};
typedef struct anon__lazy_155 nav__6091_54;
typedef struct anon__lazy_58 nav__6091_56;
typedef struct anon__lazy_1152 nav__6098_43;
typedef struct anon__lazy_58 nav__6145_39;
typedef struct anon__lazy_484 nav__6145_44;
typedef struct anon__lazy_155 nav__6145_52;
typedef struct anon__lazy_103 nav__6145_67;
typedef struct anon__lazy_226 nav__6145_70;
typedef struct anon__lazy_484 nav__6144_41;
typedef struct anon__lazy_595 nav__6144_52;
typedef struct anon__lazy_103 nav__6144_54;
typedef struct anon__lazy_58 nav__6144_57;
typedef struct anon__lazy_484 nav__6143_41;
typedef struct anon__lazy_484 nav__6147_41;
typedef struct anon__lazy_595 nav__6147_52;
typedef struct anon__lazy_103 nav__6147_54;
typedef struct anon__lazy_58 nav__6147_57;
typedef struct anon__lazy_484 nav__6146_41;
typedef struct anon__lazy_155 nav__6142_48;
typedef struct anon__lazy_58 nav__6142_50;
typedef struct anon__lazy_484 nav__6142_64;
typedef struct anon__lazy_155 nav__6141_48;
typedef struct anon__lazy_58 nav__6141_50;
typedef struct anon__lazy_58 nav__6140_39;
typedef struct anon__lazy_155 nav__6140_54;
typedef struct anon__lazy_226 nav__6140_64;
typedef struct anon__lazy_157 nav__6140_66;
typedef struct anon__lazy_58 nav__6139_39;
typedef struct anon__lazy_155 nav__6139_53;
typedef struct anon__lazy_58 nav__6106_41;
typedef struct anon__lazy_58 nav__6148_56;
typedef struct anon__lazy_676 nav__6148_65;
typedef struct anon__lazy_58 nav__6150_56;
typedef struct anon__lazy_708 nav__6150_65;
typedef struct anon__lazy_155 nav__6093_55;
typedef struct anon__lazy_58 nav__6093_57;
typedef struct anon__lazy_103 nav__3497_52;
typedef struct anon__lazy_187 nav__3497_54;
typedef struct anon__lazy_961 nav__3497_62;
typedef struct anon__lazy_58 nav__3497_65;
typedef struct anon__lazy_719 nav__3497_67;
typedef struct anon__lazy_760 nav__3497_72;
typedef struct anon__lazy_58 nav__6130_40;
typedef struct anon__lazy_961 nav__6131_39;
typedef struct anon__lazy_58 nav__6132_40;
typedef struct anon__lazy_484 nav__6133_38;
typedef struct anon__lazy_58 nav__6133_41;
typedef struct anon__lazy_996 nav__6133_58;
typedef struct anon__lazy_484 nav__6134_38;
typedef struct anon__lazy_58 nav__6134_41;
typedef struct anon__lazy_1001 nav__6134_58;
typedef struct anon__lazy_484 nav__6135_38;
typedef struct anon__lazy_58 nav__6135_41;
typedef struct anon__lazy_1006 nav__6135_58;
typedef struct anon__lazy_58 nav__6136_42;
typedef struct anon__lazy_961 nav__6136_66;
typedef struct anon__lazy_719 nav__6136_69;
typedef struct anon__lazy_760 nav__6136_72;
typedef struct anon__lazy_996 nav__6136_77;
typedef struct anon__lazy_58 nav__6137_42;
typedef struct anon__lazy_961 nav__6137_66;
typedef struct anon__lazy_719 nav__6137_69;
typedef struct anon__lazy_760 nav__6137_72;
typedef struct anon__lazy_1001 nav__6137_77;
typedef struct anon__lazy_58 nav__6138_42;
typedef struct anon__lazy_961 nav__6138_66;
typedef struct anon__lazy_719 nav__6138_69;
typedef struct anon__lazy_760 nav__6138_72;
typedef struct anon__lazy_1006 nav__6138_77;
typedef struct anon__lazy_1152 nav__6101_38;
typedef struct anon__lazy_155 nav__5170_41;
typedef struct anon__lazy_58 nav__5170_43;
typedef struct anon__lazy_155 nav__5171_41;
typedef struct anon__lazy_58 nav__5171_43;
typedef struct anon__lazy_103 nav__5171_52;
typedef struct anon__lazy_58 nav__6158_56;
typedef struct anon__lazy_693 nav__6158_65;
typedef struct anon__lazy_695 nav__6158_69;
typedef struct anon__lazy_58 nav__6151_43;
typedef struct anon__lazy_961 nav__6152_56;
typedef struct anon__lazy_996 nav__6152_60;
typedef struct anon__lazy_961 nav__6153_56;
typedef struct anon__lazy_1001 nav__6153_60;
typedef struct anon__lazy_961 nav__6154_56;
typedef struct anon__lazy_1006 nav__6154_60;
typedef struct anon__lazy_484 nav__6155_38;
typedef struct anon__lazy_58 nav__6155_41;
typedef struct anon__lazy_996 nav__6155_47;
typedef struct anon__lazy_961 nav__6155_66;
typedef struct anon__lazy_719 nav__6155_69;
typedef struct anon__lazy_484 nav__6156_38;
typedef struct anon__lazy_58 nav__6156_41;
typedef struct anon__lazy_1001 nav__6156_47;
typedef struct anon__lazy_961 nav__6156_66;
typedef struct anon__lazy_719 nav__6156_69;
typedef struct anon__lazy_484 nav__6157_38;
typedef struct anon__lazy_58 nav__6157_41;
typedef struct anon__lazy_1006 nav__6157_47;
typedef struct anon__lazy_961 nav__6157_66;
typedef struct anon__lazy_719 nav__6157_69;
typedef struct anon__lazy_58 nav__6162_43;
typedef struct anon__lazy_695 nav__6162_46;
typedef struct anon__lazy_58 nav__6159_43;
typedef struct anon__lazy_695 nav__6159_46;
typedef struct anon__lazy_58 nav__6111_41;
typedef struct anon__lazy_58 nav__6160_41;
typedef struct anon__lazy_58 nav__6161_41;
typedef struct anon__lazy_58 nav__6163_41;
typedef struct anon__lazy_695 nav__6163_43;
typedef struct anon__lazy_58 nav__5206_39;
typedef struct anon__lazy_155 nav__6090_54;
typedef struct anon__lazy_58 nav__6090_56;
typedef struct anon__lazy_550 nav__6164_51;
typedef struct anon__lazy_155 nav__6164_57;
typedef struct anon__lazy_58 nav__6164_59;
typedef struct anon__lazy_155 nav__6092_55;
typedef struct anon__lazy_58 nav__6092_57;
typedef struct anon__lazy_155 nav__6165_54;
typedef struct anon__lazy_58 nav__6165_56;
typedef struct anon__lazy_58 nav__6166_39;
typedef struct anon__lazy_550 nav__6167_39;
typedef struct anon__lazy_155 nav__6167_42;
typedef struct anon__lazy_58 nav__6167_44;
typedef struct anon__lazy_550 nav__6168_39;
typedef struct anon__lazy_155 nav__6168_42;
typedef struct anon__lazy_58 nav__6168_44;
typedef struct anon__lazy_484 nav__6168_51;
typedef struct anon__lazy_484 nav__6169_41;
typedef struct anon__lazy_155 nav__6169_45;
typedef struct anon__lazy_58 nav__6169_47;
typedef struct anon__lazy_484 nav__6172_41;
typedef struct anon__lazy_155 nav__6172_45;
typedef struct anon__lazy_58 nav__6172_47;
typedef struct anon__lazy_595 nav__6172_56;
typedef struct anon__lazy_103 nav__6172_58;
typedef struct anon__lazy_484 nav__6171_41;
typedef struct anon__lazy_155 nav__6171_45;
typedef struct anon__lazy_58 nav__6171_47;
typedef struct anon__lazy_484 nav__6170_41;
typedef struct anon__lazy_155 nav__6170_45;
typedef struct anon__lazy_58 nav__6170_47;
struct Target_Os__639;
union Target_Os_VersionRange__695;
struct SemanticVersion_Range__700;
struct SemanticVersion__698;
typedef struct anon__lazy_58 nav__270_43;
struct SemanticVersion__698 {
 uintptr_t major;
 uintptr_t minor;
 uintptr_t patch;
 struct anon__lazy_58 pre;
 struct anon__lazy_58 build;
};
struct SemanticVersion_Range__700 {
 struct SemanticVersion__698 zig_e_min;
 struct SemanticVersion__698 zig_e_max;
};
struct Target_Os_HurdVersionRange__702;
struct Target_Os_HurdVersionRange__702 {
 struct SemanticVersion_Range__700 range;
 struct SemanticVersion__698 glibc;
};
struct Target_Os_LinuxVersionRange__704;
struct Target_Os_LinuxVersionRange__704 {
 struct SemanticVersion_Range__700 range;
 struct SemanticVersion__698 glibc;
 uint32_t android;
};
struct Target_Os_WindowsVersion_Range__760;
struct Target_Os_WindowsVersion_Range__760 {
 uint32_t zig_e_min;
 uint32_t zig_e_max;
};
union Target_Os_VersionRange__695 {
 struct SemanticVersion_Range__700 semver;
 struct Target_Os_HurdVersionRange__702 hurd;
 struct Target_Os_LinuxVersionRange__704 linux;
 struct Target_Os_WindowsVersion_Range__760 windows;
};
struct Target_Os__639 {
 union Target_Os_VersionRange__695 version_range;
 uint8_t tag;
};
struct Target_DynamicLinker__976;
struct Target_DynamicLinker__976 {
 uint8_t buffer[255];
 uint8_t len;
};
typedef struct anon__lazy_103 nav__3703_40;
typedef struct anon__lazy_103 nav__3704_40;
typedef struct anon__lazy_103 nav__3705_40;
typedef struct anon__lazy_103 nav__3667_46;
typedef struct anon__lazy_103 nav__235_46;
struct Target_Cpu_Feature_Set__993;
struct Target_Cpu_Feature_Set__993 {
 uintptr_t ints[5];
};
struct Target_Cpu__961;
struct Target_Cpu_Model__988;
struct Target_Cpu__961 {
 struct Target_Cpu_Model__988 const *model;
 struct Target_Cpu_Feature_Set__993 features;
 uint8_t arch;
};
typedef struct anon__lazy_58 nav__269_46;
struct Target_Cpu_Model__988 {
 struct anon__lazy_58 name;
 struct anon__lazy_58 llvm_name;
 struct Target_Cpu_Feature_Set__993 features;
};
struct Target__637;
typedef struct anon__lazy_58 nav__271_51;
struct Target__637 {
 struct Target_Cpu__961 cpu;
 struct Target_Os__639 os;
 uint8_t abi;
 uint8_t ofmt;
 struct Target_DynamicLinker__976 dynamic_linker;
};
struct builtin_CallingConvention__266;
struct builtin_CallingConvention_CommonOptions__268;
typedef struct anon__lazy_241 nav__470_40;
struct builtin_CallingConvention_CommonOptions__268 {
 struct anon__lazy_241 incoming_stack_alignment;
};
struct builtin_CallingConvention_X86RegparmOptions__270;
struct builtin_CallingConvention_X86RegparmOptions__270 {
 struct anon__lazy_241 incoming_stack_alignment;
 uint8_t register_params;
};
struct builtin_CallingConvention_ArmInterruptOptions__272;
struct builtin_CallingConvention_ArmInterruptOptions__272 {
 struct anon__lazy_241 incoming_stack_alignment;
 uint8_t type;
};
struct builtin_CallingConvention_MipsInterruptOptions__274;
struct builtin_CallingConvention_MipsInterruptOptions__274 {
 struct anon__lazy_241 incoming_stack_alignment;
 uint8_t mode;
};
struct builtin_CallingConvention_RiscvInterruptOptions__276;
struct builtin_CallingConvention_RiscvInterruptOptions__276 {
 struct anon__lazy_241 incoming_stack_alignment;
 uint8_t mode;
};
struct builtin_CallingConvention__266 {
 union {
  struct builtin_CallingConvention_CommonOptions__268 x86_64_sysv;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_regcall_v3_sysv;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_regcall_v4_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_vectorcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_interrupt;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_sysv;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_win;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_stdcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_fastcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_thiscall;
  struct builtin_CallingConvention_CommonOptions__268 x86_thiscall_mingw;
  struct builtin_CallingConvention_CommonOptions__268 x86_regcall_v3;
  struct builtin_CallingConvention_CommonOptions__268 x86_regcall_v4_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_vectorcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs_darwin;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs_win;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_vfabi;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_vfabi_sve;
  struct builtin_CallingConvention_CommonOptions__268 arm_aapcs;
  struct builtin_CallingConvention_CommonOptions__268 arm_aapcs_vfp;
  struct builtin_CallingConvention_ArmInterruptOptions__272 arm_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 mips64_n64;
  struct builtin_CallingConvention_CommonOptions__268 mips64_n32;
  struct builtin_CallingConvention_MipsInterruptOptions__274 mips64_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 mips_o32;
  struct builtin_CallingConvention_MipsInterruptOptions__274 mips_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 riscv64_lp64;
  struct builtin_CallingConvention_CommonOptions__268 riscv64_lp64_v;
  struct builtin_CallingConvention_RiscvInterruptOptions__276 riscv64_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 riscv32_ilp32;
  struct builtin_CallingConvention_CommonOptions__268 riscv32_ilp32_v;
  struct builtin_CallingConvention_RiscvInterruptOptions__276 riscv32_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 sparc64_sysv;
  struct builtin_CallingConvention_CommonOptions__268 sparc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf_altivec;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf_v2;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_sysv_altivec;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_aix;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_aix_altivec;
  struct builtin_CallingConvention_CommonOptions__268 wasm_mvp;
  struct builtin_CallingConvention_CommonOptions__268 arc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 bpf_std;
  struct builtin_CallingConvention_CommonOptions__268 csky_sysv;
  struct builtin_CallingConvention_CommonOptions__268 csky_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 hexagon_sysv;
  struct builtin_CallingConvention_CommonOptions__268 hexagon_sysv_hvx;
  struct builtin_CallingConvention_CommonOptions__268 lanai_sysv;
  struct builtin_CallingConvention_CommonOptions__268 loongarch64_lp64;
  struct builtin_CallingConvention_CommonOptions__268 loongarch32_ilp32;
  struct builtin_CallingConvention_CommonOptions__268 m68k_sysv;
  struct builtin_CallingConvention_CommonOptions__268 m68k_gnu;
  struct builtin_CallingConvention_CommonOptions__268 m68k_rtd;
  struct builtin_CallingConvention_CommonOptions__268 m68k_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 msp430_eabi;
  struct builtin_CallingConvention_CommonOptions__268 propeller_sysv;
  struct builtin_CallingConvention_CommonOptions__268 s390x_sysv;
  struct builtin_CallingConvention_CommonOptions__268 s390x_sysv_vx;
  struct builtin_CallingConvention_CommonOptions__268 ve_sysv;
  struct builtin_CallingConvention_CommonOptions__268 xcore_xs1;
  struct builtin_CallingConvention_CommonOptions__268 xcore_xs2;
  struct builtin_CallingConvention_CommonOptions__268 xtensa_call0;
  struct builtin_CallingConvention_CommonOptions__268 xtensa_windowed;
  struct builtin_CallingConvention_CommonOptions__268 amdgcn_device;
  struct builtin_CallingConvention_CommonOptions__268 amdgcn_cs;
 } payload;
 uint8_t tag;
};
struct std_Options__4261;
typedef struct anon__lazy_484 nav__97_39;
struct std_Options__4261 {
 struct anon__lazy_484 page_size_min;
 struct anon__lazy_484 page_size_max;
 uintptr_t fmt_max_depth;
 bool enable_segfault_handler;
 uint8_t log_level;
 bool crypto_always_getrandom;
 bool crypto_fork_safety;
 bool keep_sigpipe;
 bool http_disable_tls;
 bool http_enable_ssl_key_log_file;
 uint8_t side_channels_mitigations;
};
typedef struct anon__lazy_484 nav__3683_38;
struct json_stringify_WriteStream__union_8236__8236;
struct json_stringify_WriteStream__union_8236__8236 {
 union {
  uintptr_t checked_to_fixed_depth;
 } payload;
 uint8_t tag;
};
struct macho_mach_header_64__9281;
struct macho_mach_header_64__9281 {
 uint32_t magic;
 int cputype;
 int cpusubtype;
 uint32_t filetype;
 uint32_t ncmds;
 uint32_t sizeofcmds;
 uint32_t flags;
 uint32_t reserved;
};
typedef struct anon__lazy_103 nav__3694_43;
typedef struct anon__lazy_58 nav__4052_40;
static uint8_t const __anon_1087[21];
static uint8_t const __anon_1152[21];
static char const __anon_3329[33];
static uint8_t const __anon_2303[62];
static uint8_t const __anon_3411[34];
static uint8_t const __anon_3951[49];
static uint8_t const __anon_4335[1];
static uint8_t const __anon_4451[68];
static uint8_t const __anon_4574[82];
static uint8_t const __anon_4628[32];
static uint8_t const __anon_4761[22];
static uint8_t const __anon_2011[15];
static uint8_t const __anon_2016[14];
static uint8_t const __anon_2021[4];
static uint8_t const __anon_2026[13];
static uint8_t const __anon_2034[4];
static uint8_t const __anon_2049[19];
static uint8_t const __anon_5099[8];
static uint8_t const __anon_5278[3];
static uint8_t const __anon_5352[25];
static uint8_t const __anon_5418[7];
static uint8_t const __anon_5444[2];
static uint8_t const __anon_5497[23];
static uint8_t const __anon_5522[11];
static uint8_t const __anon_5578[17];
static uint8_t const __anon_5600[10];
static uint8_t const __anon_5670[31];
static uint8_t const __anon_5725[24];
static uint8_t const __anon_5841[23];
static uint8_t const __anon_5846[2];
static uint8_t const __anon_5938[12];
static uint8_t const __anon_5948[1];
static uint8_t const __anon_5967[10];
static uint8_t const __anon_5999[15];
static uint8_t const __anon_6027[13];
static uint8_t const __anon_7177[10];
static uint8_t const __anon_7192[1];
static uint8_t const __anon_7205[1];
static uint8_t const __anon_8092[24];
static uint8_t const __anon_8208[13];
static uint8_t const __anon_8225[5];
static uint8_t const __anon_8429[62];
static uint8_t const __anon_8496[55];
static uint8_t const __anon_8509[24];
static uint8_t const __anon_8517[17];
static uint8_t const __anon_8668[63];
static uint8_t const __anon_8741[59];
static uint8_t const __anon_8939[3];
static uint8_t const __anon_8898[3];
static uint8_t const __anon_8903[3];
static uint8_t const __anon_8907[3];
static uint8_t const __anon_8911[3];
static uint8_t const __anon_8915[3];
static uint8_t const __anon_8919[3];
static uint8_t const __anon_8923[3];
static uint8_t const __anon_9129[5];
static uint8_t const __anon_9133[9];
static uint8_t const __anon_9236[5];
static uint8_t const __anon_4937[4];
static uint8_t const __anon_9352[6];
#define c_instrument_hooks_set_feature__237 instrument_hooks_set_feature
zig_extern void instrument_hooks_set_feature(uint64_t, bool);
static void features_set_feature__330(uint64_t, bool);
static void bit_set_IntegerBitSet_2864_29_set__369(uint64_t *, uintptr_t);
static void bit_set_IntegerBitSet_2864_29_unset__371(uint64_t *, uintptr_t);
static void debug_assert__180(bool);
static uint64_t bit_set_IntegerBitSet_2864_29_maskBit__390(uintptr_t);
static nav__1280_39 fifo_posix_Pipe_openPipe__1280(nav__1280_41);
static nav__1282_39 fifo_posix_Pipe_openWrite__1282(struct mem_Allocator__599, nav__1282_42);
static nav__1281_39 fifo_posix_Pipe_openRead__1281(struct mem_Allocator__599, nav__1281_42);
static nav__1129_39 runner_fifo_RunnerFifo_init__1129(struct mem_Allocator__599);
static uint16_t fifo_posix_Pipe_Writer_sendCmd__1286(struct fifo_posix_Pipe_Writer__786 *, struct shared_Command__1178);
static uint16_t fifo_posix_waitReadable__1268(int32_t, uint64_t);
static nav__4193_38 bincode_deserializeInt__anon_2126__4193(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992);
static nav__4192_38 bincode_deserializeAlloc__anon_2115__4192(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4235_38 bincode_deserializeInt__anon_2235__4235(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992);
static nav__4234_38 bincode_deserializeAlloc__anon_2230__4234(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4237_40 bincode_deserializePointerAlloc__anon_2265__4237(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4236_40 bincode_deserializeAlloc__anon_2252__4236(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4233_39 bincode_deserializeStructAlloc__anon_2214__4233(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4232_39 bincode_deserializeAlloc__anon_2185__4232(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static uint16_t bincode_deserializeAlloc__anon_2325__4241(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4243_39 bincode_deserializeStructAlloc__anon_2391__4243(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4242_39 bincode_deserializeAlloc__anon_2372__4242(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4249_38 bincode_deserializeInt__anon_2513__4249(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992);
static nav__4248_38 bincode_deserializeAlloc__anon_2508__4248(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4247_39 bincode_deserializeUnionAlloc__anon_2495__4247(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4246_39 bincode_deserializeAlloc__anon_2462__4246(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4245_39 bincode_deserializeStructAlloc__anon_2444__4245(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4244_39 bincode_deserializeAlloc__anon_2430__4244(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4251_38 bincode_deserializeEnum__anon_2598__4251(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992);
static nav__4250_38 bincode_deserializeAlloc__anon_2568__4250(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4191_39 bincode_deserializeUnionAlloc__anon_2110__4191(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__4190_39 bincode_deserializeAlloc__anon_2000__4190(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992, struct mem_Allocator__599);
static nav__1291_39 fifo_posix_Pipe_Reader_recvCmd__1291(struct fifo_posix_Pipe_Reader__788 *);
static nav__1292_39 fifo_posix_Pipe_Reader_waitForResponse__1292(struct fifo_posix_Pipe_Reader__788 *, nav__1292_42);
static uint16_t fifo_posix_Pipe_Reader_waitForAck__1293(struct fifo_posix_Pipe_Reader__788 *, nav__1293_40);
static uint16_t runner_fifo_RunnerFifo_send_version__1139(struct runner_fifo_RunnerFifo__631 *, uint64_t);
static uint16_t runner_fifo_RunnerFifo_validate_protocol_version__1130(struct runner_fifo_RunnerFifo__631 *);
static nav__1140_38 runner_fifo_RunnerFifo_get_integration_mode__1140(struct runner_fifo_RunnerFifo__631 *);
static nav__820_39 instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__820(struct mem_Allocator__599);
static nav__807_39 instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_init__807(struct mem_Allocator__599);
static nav__743_39 instruments_root_Instrument_init__743(struct mem_Allocator__599);
static nav__730_39 instrument_hooks_InstrumentHooks_init__730(struct mem_Allocator__599);
#define c_instrument_hooks_init__238 instrument_hooks_init
zig_extern struct InstrumentHooks *instrument_hooks_init(void);
static nav__3706_40 mem_Allocator_create__anon_1039__3706(struct mem_Allocator__599);
static uint16_t fs_accessAbsolute__1348(nav__1348_39, struct fs_File_OpenFlags__1101);
static nav__1345_39 fs_openFileAbsolute__1345(nav__1345_41, struct fs_File_OpenFlags__1101);
static struct fifo_posix_Pipe_Writer__786 fifo_posix_Pipe_Writer_init__1283(struct fs_File__794, struct mem_Allocator__599);
static struct fifo_posix_Pipe_Reader__788 fifo_posix_Pipe_Reader_init__1288(struct fs_File__794, struct mem_Allocator__599);
static void array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1562(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *);
static struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *);
static uint16_t bincode_serialize__anon_1232__3956(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__1178);
static uint8_t const (*mem_asBytes__anon_1270__3957(uint32_t const *))[4];
static uint16_t fs_File_writeAll__1439(struct fs_File__794, nav__1439_40);
static nav__1964_38 posix_poll__1964(nav__1964_41, int32_t);
static nav__1429_38 fs_File_readAll__1429(struct fs_File__794, nav__1429_41);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_resize__1559(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, uintptr_t);
static struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(nav__4156_40);
static struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4145(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *);
static nav__4189_38 io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189(void const *, nav__4189_41);
static nav__4231_39 io_Reader_readBytesNoEof__anon_2172__4231(struct io_Reader__2141);
static nav__4238_39 io_Reader_readBytesNoEof__anon_2278__4238(struct io_Reader__2141);
static nav__4239_40 mem_Allocator_alloc__anon_2287__4239(struct mem_Allocator__599, uintptr_t);
static nav__4196_38 io_Reader_readAll__4196(struct io_Reader__2141, nav__4196_41);
static zig_noreturn void bincode_invalidProtocol__anon_2310__4240(void);
static void shared_Command_deinit__3829(struct shared_Command__1178, struct mem_Allocator__599);
static void logger_debug__anon_2632__4261(nav__4261_39);
static void logger_err__anon_2656__4262(void);
static void logger_err__anon_2665__4263(void);
static zig_noreturn void posix_exit__1747(uint8_t);
static void logger_err__anon_2679__4264(nav__4264_40);
static void runner_fifo_RunnerFifo_deinit__1131(struct runner_fifo_RunnerFifo__631 *);
static struct environment_root_Environment__559 environment_root_Environment_init__773(struct mem_Allocator__599);
static nav__4265_39 mem_Allocator_allocBytesWithAlignment__anon_2726__4265(struct mem_Allocator__599, uintptr_t, uintptr_t);
static bool fs_path_isAbsolute__4300(nav__4300_39);
static struct fs_Dir__1103 fs_cwd__1340(void);
static uint16_t fs_Dir_access__3785(struct fs_Dir__1103, nav__3785_40, struct fs_File_OpenFlags__1101);
static nav__3721_39 fs_Dir_openFile__3721(struct fs_Dir__1103, nav__3721_42, struct fs_File_OpenFlags__1101);
static struct array_list_ArrayListAligned_28u8_2cnull_29__806 array_list_ArrayListAligned_28u8_2cnull_29_init__1527(struct mem_Allocator__599);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, uintptr_t);
static uint16_t bincode_serializeUnion__anon_2768__4363(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__1178);
static nav__1438_38 fs_File_write__1438(struct fs_File__794, nav__1438_41);
static nav__4364_38 math_cast__anon_2787__4364(uintptr_t);
static uint16_t posix_errno__anon_2970__4656(int);
static uint16_t posix_unexpectedErrno__2021(uint16_t);
static nav__1428_38 fs_File_read__1428(struct fs_File__794, nav__1428_41);
static nav__4148_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4148(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *, nav__4148_42);
static uint16_t io_Reader_readNoEof__4198(struct io_Reader__2141, nav__4198_40);
static nav__4657_39 mem_Allocator_allocWithSizeAndAlignment__anon_3006__4657(struct mem_Allocator__599, uintptr_t, uintptr_t);
static nav__4197_38 io_Reader_readAtLeast__4197(struct io_Reader__2141, nav__4197_41, uintptr_t);
static void utils_print__anon_3028__4668(void);
static void mem_Allocator_free__anon_3042__4747(struct mem_Allocator__599, nav__4747_40);
static void logger_logWithPrefix__anon_3050__4748(nav__4748_39);
static void logger_logWithPrefix__anon_3052__4749(void);
static void logger_logWithPrefix__anon_3053__4750(void);
static void logger_logWithPrefix__anon_3058__4752(nav__4752_40);
static void fifo_posix_Pipe_Writer_deinit__1287(struct fifo_posix_Pipe_Writer__786 *);
static void fifo_posix_Pipe_Reader_deinit__1294(struct fifo_posix_Pipe_Reader__788 *);
static struct environment_linked_libraries_root_LinkedLibraries__837 environment_linked_libraries_root_LinkedLibraries_init__2808(struct mem_Allocator__599);
static uint8_t mem_Alignment_fromByteUnits__1113(uintptr_t);
static bool fs_path_isAbsolutePosix__4306(nav__4306_39);
static nav__2022_39 posix_toPosixPath__2022(nav__2022_41);
static uint16_t fs_Dir_accessZ__3786(struct fs_Dir__1103, uint8_t const *, struct fs_File_OpenFlags__1101);
static nav__3722_39 fs_Dir_openFileZ__3722(struct fs_Dir__1103, uint8_t const *, struct fs_File_OpenFlags__1101);
static uintptr_t array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__4816(uintptr_t, uintptr_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1565(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, uintptr_t);
static uint16_t bincode_serialize__anon_3165__4817(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, uint32_t);
static uint16_t bincode_serialize__anon_3167__4818(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__struct_1181__1181);
static uint16_t bincode_serialize__anon_3169__4819(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t bincode_serialize__anon_3171__4820(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__struct_1182__1182);
static uint16_t bincode_serialize__anon_3173__4821(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__struct_1183__1183);
static uint16_t bincode_serialize__anon_3175__4822(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, uint64_t);
static uint16_t bincode_serialize__anon_3177__4823(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, uint8_t);
static nav__1757_38 posix_write__1757(int32_t, nav__1757_40);
static nav__1749_38 posix_read__1749(int32_t, nav__1749_40);
static nav__4889_38 math_mul__anon_3220__4889(uintptr_t, uintptr_t);
static nav__4890_39 mem_Allocator_allocBytesWithAlignment__anon_3222__4890(struct mem_Allocator__599, uintptr_t, uintptr_t);
static nav__4195_38 io_Reader_read__4195(struct io_Reader__2141, nav__4195_41);
static nav__4892_39 mem_sliceAsBytes__anon_3353__4892(nav__4892_39);
static uint16_t fmt_format__anon_3841__5177(struct io_Writer__3827, nav__5177_40);
static nav__5052_40 fmt_bufPrint__anon_3407__5052(nav__5052_39, nav__5052_42);
static uint16_t fmt_format__anon_4351__5571(struct io_Writer__3827);
static nav__5054_40 fmt_bufPrint__anon_3429__5054(nav__5054_39);
static uint16_t fmt_format__anon_4468__5573(struct io_Writer__3827);
static nav__5055_40 fmt_bufPrint__anon_3442__5055(nav__5055_39);
static uint16_t fmt_format__anon_4580__5574(struct io_Writer__3827, nav__5574_41);
static nav__5056_40 fmt_bufPrint__anon_3455__5056(nav__5056_39, nav__5056_43);
static void array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(struct array_list_ArrayListAligned_28u8_2cnull_29__806);
static void fs_File_close__1385(struct fs_File__794);
static struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_init__3512(struct mem_Allocator__599);
static bool math_isPowerOfTwo__anon_3478__5057(uintptr_t);
static uint16_t posix_faccessatZ__1913(int32_t, uint8_t const *, uint32_t, uint32_t);
static nav__1769_38 posix_openatZ__1769(int32_t, uint8_t const *, uint32_t, uint16_t);
static nav__1928_38 posix_fcntl__1928(int32_t, int32_t, uintptr_t);
static void posix_close__1727(int32_t);
static nav__1575_39 array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(struct array_list_ArrayListAligned_28u8_2cnull_29__806);
static nav__5136_39 mem_Allocator_remap__anon_3648__5136(struct mem_Allocator__599, nav__5136_39, uintptr_t);
static nav__5137_40 mem_Allocator_alignedAlloc__anon_3653__5137(struct mem_Allocator__599, uintptr_t);
static void mem_Allocator_free__anon_3655__5138(struct mem_Allocator__599, nav__5138_40);
static uint16_t bincode_serializeInt__anon_3660__5139(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, uint32_t);
static uint16_t bincode_serializeStruct__anon_3665__5140(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__struct_1181__1181);
static uint16_t bincode_serializeStruct__anon_3666__5141(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__struct_1182__1182);
static uint16_t bincode_serializeStruct__anon_3667__5142(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_Command__struct_1183__1183);
static uint16_t bincode_serializeInt__anon_3668__5143(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, uint64_t);
static uint16_t bincode_serializeEnum__anon_3671__5144(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, uint8_t);
static uint16_t posix_errno__anon_3681__5146(intptr_t);
static struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *);
static nav__5160_38 io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160(void const *, nav__5160_41);
static uint16_t io_Writer_writeAll__5168(struct io_Writer__3827, nav__5168_40);
static uint16_t shared_Command_format__anon_4988__5585(struct shared_Command__1178, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatType__anon_4328__5570(struct shared_Command__1178, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static nav__4154_39 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968);
static uint16_t fmt_formatType__anon_4459__5572(uint64_t, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static uint16_t fmt_formatType__anon_4732__5575(nav__5575_39, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static nav__5580_39 mem_sliceAsBytes__anon_4876__5580(nav__5580_39);
static nav__5581_39 mem_bytesAsSlice__anon_4889__5581(nav__5581_39);
static nav__3955_38 io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955(void const *, nav__3955_41);
static uint16_t bincode_serialize__anon_4920__5582(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, int32_t);
static uint16_t bincode_serialize__anon_4922__5583(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, nav__5583_40);
static uint16_t bincode_serialize__anon_4924__5584(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_MarkerType__1185);
static nav__4149_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4149(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *, nav__4149_42);
static nav__5167_38 io_Writer_write__5167(struct io_Writer__3827, nav__5167_41);
static uint16_t io_Writer_print__anon_5008__5586(struct io_Writer__3827, nav__5586_41);
static uint16_t io_Writer_print__anon_5029__5587(struct io_Writer__3827, nav__5587_41);
static uint16_t io_Writer_print__anon_5041__5588(struct io_Writer__3827, nav__5588_40);
static uint16_t io_Writer_print__anon_5050__5589(struct io_Writer__3827, nav__5589_39);
static uint16_t io_Writer_print__anon_5063__5590(struct io_Writer__3827, nav__5590_39);
static nav__5189_38 unicode_utf8ByteSequenceLength__5189(uint8_t);
static nav__5205_38 unicode_utf8CountCodepoints__5205(nav__5205_40);
static uint16_t fmt_formatBuf__anon_5109__5592(nav__5592_39, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatInt__anon_5170__5602(uint64_t, uint8_t, uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatIntValue__anon_5153__5601(uint64_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatValue__anon_5084__5591(uint64_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static nav__1553_38 array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1553(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, nav__1553_42);
static uint16_t bincode_serializeInt__anon_5117__5593(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, int32_t);
static uint16_t bincode_serializePointer__anon_5122__5594(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, nav__5594_40);
static uint16_t bincode_serializeUnion__anon_5123__5595(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct shared_MarkerType__1185);
static uint16_t fmt_format__anon_5127__5596(struct io_Writer__3827, nav__5596_41);
static uint16_t fmt_format__anon_5131__5597(struct io_Writer__3827, nav__5597_41);
static uint16_t fmt_format__anon_5135__5598(struct io_Writer__3827, nav__5598_40);
static uint16_t fmt_format__anon_5139__5599(struct io_Writer__3827, nav__5599_39);
static uint16_t fmt_format__anon_5143__5600(struct io_Writer__3827, nav__5600_39);
static nav__4953_39 fmt_digits2__4953(uint8_t);
static uint8_t fmt_digitToChar__4969(uint8_t, uint8_t);
static nav__5195_38 unicode_utf8Decode__5195(nav__5195_40);
static nav__5188_38 unicode_utf8CodepointSequenceLength__5188(uint32_t);
static nav__5608_38 unicode_utf8EncodeImpl__anon_5756__5608(uint32_t, nav__5608_40);
static nav__5190_38 unicode_utf8Encode__5190(uint32_t, nav__5190_40);
static uint16_t io_Writer_writeBytesNTimes__5172(struct io_Writer__3827, nav__5172_40, uintptr_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1547(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, nav__1547_41);
static uint16_t fmt_formatType__anon_5404__5603(int32_t, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static uint16_t fmt_formatType__anon_5435__5604(nav__5604_39, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static uint16_t shared_MarkerType_format__anon_5800__5610(struct shared_MarkerType__1185, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatType__anon_5610__5605(struct shared_MarkerType__1185, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static uint16_t fmt_formatType__anon_5679__5606(uint64_t, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static uint16_t fmt_formatType__anon_5733__5607(uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static nav__5197_38 unicode_utf8Decode2__5197(nav__5197_40);
static nav__5199_38 unicode_utf8Decode3__5199(nav__5199_40);
static nav__5203_38 unicode_utf8Decode4__5203(nav__5203_40);
static bool unicode_isSurrogateCodepoint__5268(uint32_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1566(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, uintptr_t);
static void array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1548(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *, nav__1548_41);
static uint16_t fmt_formatInt__anon_5888__5617(int32_t, uint8_t, uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatIntValue__anon_5877__5616(int32_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatValue__anon_5781__5609(int32_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t io_Writer_print__anon_5814__5611(struct io_Writer__3827, nav__5611_39);
static uint16_t io_Writer_print__anon_5819__5612(struct io_Writer__3827, nav__5612_39);
static uint16_t io_Writer_print__anon_5826__5613(struct io_Writer__3827, nav__5613_39);
static uint16_t io_Writer_print__anon_5832__5614(struct io_Writer__3827, nav__5614_39);
static uint16_t fmt_formatIntValue__anon_5915__5622(uint64_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatValue__anon_5834__5615(uint64_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static nav__5201_38 unicode_utf8Decode3AllowSurrogateHalf__5201(nav__5201_40);
static nav__1491_38 array_list_addOrOom__1491(uintptr_t, uintptr_t);
static uint16_t fmt_format__anon_5899__5618(struct io_Writer__3827, nav__5618_39);
static uint16_t fmt_format__anon_5903__5619(struct io_Writer__3827, nav__5619_39);
static uint16_t fmt_format__anon_5907__5620(struct io_Writer__3827, nav__5620_39);
static uint16_t fmt_format__anon_5911__5621(struct io_Writer__3827, nav__5621_39);
#define c_instrument_hooks_deinit__239 instrument_hooks_deinit
zig_extern void instrument_hooks_deinit(struct InstrumentHooks *);
static void instrument_hooks_InstrumentHooks_deinit__731(struct InstrumentHooks *);
static void mem_Allocator_destroy__anon_6049__5623(struct mem_Allocator__599, struct InstrumentHooks *);
static void instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_deinit__808(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *);
static void instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__821(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *);
static void environment_root_Environment_deinit__774(struct environment_root_Environment__559 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2965(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868);
static nav__5624_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5624(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3176(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889);
static nav__5626_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5626(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 *);
static void environment_root_Environment_freeEnvironmentValue__777(struct environment_root_Environment__559 *, struct environment_root_EnvironmentValue__844);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3168(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct mem_Allocator__599);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2957(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct mem_Allocator__599);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3355(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922);
static nav__5628_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__5628(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3347(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct mem_Allocator__599);
static void environment_linked_libraries_root_LinkedLibraries_deinit__2809(struct environment_linked_libraries_root_LinkedLibraries__837 *);
static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877);
static nav__5643_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165);
static nav__5644_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165);
static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894);
static nav__5652_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314);
static nav__5653_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314);
static void mem_Allocator_free__anon_6360__5654(struct mem_Allocator__599, nav__5654_42);
static void debug_SafetyLock_assertUnlocked__3149(void);
static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3294(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *, struct mem_Allocator__599);
static void array_hash_map_IndexHeader_free__3144(struct array_hash_map_IndexHeader__880 *, struct mem_Allocator__599);
static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3100(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *, struct mem_Allocator__599);
static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927);
static nav__5662_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393);
static nav__5663_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393);
static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3473(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *, struct mem_Allocator__599);
static void environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__2810(struct mem_Allocator__599, struct environment_linked_libraries_root_LibraryInfo__934);
static void array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_deinit__3514(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936);
static nav__5735_39 mem_sliceAsBytes__anon_6577__5735(nav__5735_41);
static nav__3326_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3326(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894);
static void mem_Allocator_free__anon_6588__5736(struct mem_Allocator__599, nav__5736_40);
static uintptr_t array_hash_map_capacityIndexSize__2907(uint8_t);
static uintptr_t array_hash_map_IndexHeader_length__3140(struct array_hash_map_IndexHeader__880);
static void mem_Allocator_free__anon_6606__5737(struct mem_Allocator__599, nav__5737_40);
static nav__3132_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3132(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877);
static nav__3505_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3505(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927);
static nav__3560_40 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__3560(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936);
static void mem_Allocator_free__anon_6635__5738(struct mem_Allocator__599, nav__5738_41);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3325(uintptr_t);
static nav__5739_39 mem_sliceAsBytes__anon_6654__5739(nav__5739_39);
static uint8_t array_hash_map_capacityIndexType__2906(uint8_t);
static nav__5755_39 mem_sliceAsBytes__anon_6688__5755(nav__5755_39);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3131(uintptr_t);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3504(uintptr_t);
static nav__5756_39 mem_sliceAsBytes__anon_6701__5756(nav__5756_42);
#define c_instrument_hooks_is_instrumented__240 instrument_hooks_is_instrumented
zig_extern bool instrument_hooks_is_instrumented(struct InstrumentHooks *);
static zig_cold uint16_t runner_fifo_RunnerFifo_start_benchmark__1134(struct runner_fifo_RunnerFifo__631 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_start_benchmark__810(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__823(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *);
#define c_instrument_hooks_start_benchmark__241 instrument_hooks_start_benchmark
zig_extern uint8_t instrument_hooks_start_benchmark(struct InstrumentHooks *);
static bool features_is_feature_enabled__331(uint64_t);
static bool bit_set_IntegerBitSet_2864_29_isSet__366(uint64_t, uintptr_t);
static zig_cold uint16_t runner_fifo_RunnerFifo_stop_benchmark__1135(struct runner_fifo_RunnerFifo__631 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_stop_benchmark__811(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__824(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *);
#define c_instrument_hooks_stop_benchmark__242 instrument_hooks_stop_benchmark
zig_extern uint8_t instrument_hooks_stop_benchmark(struct InstrumentHooks *);
static uint16_t runner_fifo_RunnerFifo_set_executed_benchmark__1136(struct runner_fifo_RunnerFifo__631 *, int32_t, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_executed_benchmark__812(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *, int32_t, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__825(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *, int32_t, uint8_t const *);
#define c_instrument_hooks_set_executed_benchmark__243 instrument_hooks_set_executed_benchmark
zig_extern uint8_t instrument_hooks_set_executed_benchmark(struct InstrumentHooks *, int32_t, char const *);
static uint8_t const *c_toU8__236(char const *);
static nav__5771_39 mem_span__anon_6867__5771(uint8_t const *);
static uintptr_t mem_len__anon_6884__5772(uint8_t const *);
static uintptr_t mem_indexOfSentinel__anon_6891__5773(uint8_t const *);
static bool math_isPowerOfTwo__anon_6902__5774(void);
static nav__5775_38 simd_firstTrue__anon_6954__5775(nav__5775_40);
static uintptr_t mem_alignForward__anon_6957__5776(uintptr_t, uintptr_t);
static bool mem_isAligned__982(uintptr_t, uintptr_t);
static bool mem_isValidAlignGeneric__anon_6999__5777(uintptr_t);
static uintptr_t mem_alignBackward__anon_7000__5778(uintptr_t, uintptr_t);
static bool mem_isAlignedGeneric__anon_7004__5779(uint64_t, uint64_t);
static uint64_t mem_alignBackward__anon_7006__5780(uint64_t, uint64_t);
static bool mem_isValidAlignGeneric__anon_7008__5781(uint64_t);
static bool math_isPowerOfTwo__anon_7009__5782(uint64_t);
#define c_instrument_hooks_executed_benchmark__244 instrument_hooks_executed_benchmark
zig_extern uint8_t instrument_hooks_executed_benchmark(struct InstrumentHooks *, int32_t, char const *);
static uint16_t runner_fifo_RunnerFifo_set_integration__1137(struct runner_fifo_RunnerFifo__631 *, uint8_t const *, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_integration__813(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *, uint8_t const *, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__826(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *, uint8_t const *, uint8_t const *);
#define c_instrument_hooks_set_integration__245 instrument_hooks_set_integration
zig_extern uint8_t instrument_hooks_set_integration(struct InstrumentHooks *, char const *, char const *);
static nav__5783_40 fmt_allocPrintZ__anon_7045__5783(struct mem_Allocator__599, nav__5783_43);
static void mem_Allocator_free__anon_7047__5784(struct mem_Allocator__599, nav__5784_40);
static nav__5785_40 fmt_allocPrint__anon_7082__5785(struct mem_Allocator__599, nav__5785_43);
static nav__5786_39 mem_sliceAsBytes__anon_7089__5786(nav__5786_39);
static uint16_t fmt_format__anon_7152__5828(struct io_Writer__3827, nav__5828_40);
static uint64_t fmt_count__anon_7096__5787(nav__5787_39);
static nav__5788_38 math_cast__anon_7098__5788(uint64_t);
static nav__5789_40 fmt_bufPrint__anon_7101__5789(nav__5789_39, nav__5789_42);
static struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 io_counting_writer_countingWriter__anon_7122__5814(void);
static struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__5813(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *);
static nav__5827_38 io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__5827(void const *, nav__5827_41);
static uint16_t fmt_formatType__anon_7186__5829(uint8_t const *, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static nav__5812_38 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__5812(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *, nav__5812_42);
static nav__3933_38 io_dummyWrite__3933(nav__3933_40);
static uint16_t runner_fifo_RunnerFifo_add_marker__1138(struct runner_fifo_RunnerFifo__631 *, int32_t, struct shared_MarkerType__1185);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_add_marker__814(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *, int32_t, struct shared_MarkerType__1185);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__827(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *, int32_t, struct shared_MarkerType__1185);
#define c_instrument_hooks_add_marker__250 instrument_hooks_add_marker
zig_extern uint8_t instrument_hooks_add_marker(struct InstrumentHooks *, int32_t, uint8_t, uint64_t);
#define c_instrument_hooks_current_timestamp__254 instrument_hooks_current_timestamp
zig_extern uint64_t instrument_hooks_current_timestamp(void);
static void once_Once_28_28function_20_27init_27_29_29_call__5842(struct once_Once_28_28function_20_27init_27_29_29__7273 *);
static zig_cold void once_Once_28_28function_20_27init_27_29_29_callSlow__5843(struct once_Once_28_28function_20_27init_27_29_29__7273 *);
static void Thread_Mutex_lock__5905(struct Thread_Mutex__7278 *);
static void c_instrument_hooks_current_timestamp_S_init__5832(void);
static void Thread_Mutex_unlock__5906(struct Thread_Mutex__7278 *);
static void Thread_Mutex_DarwinImpl_lock__5919(struct Thread_Mutex_DarwinImpl__7290 *);
static void Thread_Mutex_DarwinImpl_unlock__5920(struct Thread_Mutex_DarwinImpl__7290 *);
static uint16_t environment_root_Environment_setIntegrationEnvironmentValue__778(struct environment_root_Environment__559 *, nav__778_41, nav__778_41, struct environment_root_EnvironmentValue__844);
static uint16_t environment_root_Environment_setIntegrationEnvironment__775(struct environment_root_Environment__559 *, nav__775_41, nav__775_41, nav__775_41);
#define c_instrument_hooks_set_environment__255 instrument_hooks_set_environment
zig_extern uint8_t instrument_hooks_set_environment(struct InstrumentHooks *, char const *, char const *, char const *);
static nav__5934_40 mem_Allocator_dupe__anon_7390__5934(struct mem_Allocator__599, nav__5934_43);
static nav__2967_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2967(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct mem_Allocator__599, nav__2967_44);
static nav__3178_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3178(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct mem_Allocator__599, nav__3178_44);
static nav__2968_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2968(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct mem_Allocator__599, nav__2968_44);
static nav__3179_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3179(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct mem_Allocator__599, nav__3179_44);
static nav__5935_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5935(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct mem_Allocator__599, nav__5935_44);
static nav__5936_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5936(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct mem_Allocator__599, nav__5936_44);
static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2977(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct mem_Allocator__599, uintptr_t);
static nav__5937_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5937(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5937_41);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5938(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, nav__5938_42);
static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3188(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct mem_Allocator__599, uintptr_t);
static nav__5939_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5939(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5939_41);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5940(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, nav__5940_42);
static void debug_SafetyLock_lock__3147(void *);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3124(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *, struct mem_Allocator__599, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3120(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *, struct mem_Allocator__599, uintptr_t);
static void debug_SafetyLock_unlock__3148(void *);
static uint32_t array_hash_map_IndexHeader_capacity__3139(struct array_hash_map_IndexHeader__880);
static nav__3142_38 array_hash_map_IndexHeader_findBitIndex__3142(uintptr_t);
static nav__3143_40 array_hash_map_IndexHeader_alloc__3143(struct mem_Allocator__599, uint8_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3069(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct array_hash_map_IndexHeader__880 *);
static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(nav__5941_40);
static nav__5942_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165);
static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(nav__5943_40, nav__5943_40, uintptr_t);
static uint8_t array_hash_map_IndexHeader_capacityIndexType__3138(struct array_hash_map_IndexHeader__880);
static nav__5944_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5944(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5944_41, struct array_hash_map_IndexHeader__880 *);
static nav__5945_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5945(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5945_41, struct array_hash_map_IndexHeader__880 *);
static nav__5946_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5946(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5946_41, struct array_hash_map_IndexHeader__880 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5947(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, nav__5947_42, struct array_hash_map_IndexHeader__880 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5948(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, nav__5948_42, struct array_hash_map_IndexHeader__880 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5949(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, nav__5949_42, struct array_hash_map_IndexHeader__880 *);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3318(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *, struct mem_Allocator__599, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3314(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *, struct mem_Allocator__599, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3280(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct array_hash_map_IndexHeader__880 *);
static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(nav__5950_40);
static nav__5951_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314);
static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(nav__5952_40, nav__5952_40, uintptr_t);
static nav__5953_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5953(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5953_41, struct array_hash_map_IndexHeader__880 *);
static nav__5954_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5954(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5954_41, struct array_hash_map_IndexHeader__880 *);
static nav__5955_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5955(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5955_41, struct array_hash_map_IndexHeader__880 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5956(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, nav__5956_42, struct array_hash_map_IndexHeader__880 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5957(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, nav__5957_42, struct array_hash_map_IndexHeader__880 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5958(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, nav__5958_42, struct array_hash_map_IndexHeader__880 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3122(uintptr_t, uintptr_t);
static nav__5959_40 mem_Allocator_alignedAlloc__anon_7601__5959(struct mem_Allocator__599, uintptr_t);
static uint8_t math_log2_int_ceil__anon_7751__5960(uintptr_t);
static nav__5961_40 mem_Allocator_alignedAlloc__anon_7756__5961(struct mem_Allocator__599, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5962(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct array_hash_map_IndexHeader__880 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5963(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct array_hash_map_IndexHeader__880 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5964(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *, struct array_hash_map_IndexHeader__880 *);
static uint32_t array_hash_map_StringContext_hash__2938(nav__2938_40);
static bool array_hash_map_StringContext_eql__2939(nav__2939_40, nav__2939_40, uintptr_t);
static nav__5965_40 array_hash_map_IndexHeader_indexes__anon_7788__5965(struct array_hash_map_IndexHeader__880 *);
static nav__5966_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5966(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5966_41, struct array_hash_map_IndexHeader__880 *, nav__5966_47);
static nav__5967_40 array_hash_map_IndexHeader_indexes__anon_7798__5967(struct array_hash_map_IndexHeader__880 *);
static nav__5968_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5968(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5968_41, struct array_hash_map_IndexHeader__880 *, nav__5968_47);
static nav__5969_40 array_hash_map_IndexHeader_indexes__anon_7804__5969(struct array_hash_map_IndexHeader__880 *);
static nav__5970_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5970(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868, nav__5970_41, struct array_hash_map_IndexHeader__880 *, nav__5970_47);
static uintptr_t array_hash_map_safeTruncate__anon_7812__5971(uint32_t);
static uintptr_t array_hash_map_IndexHeader_constrainIndex__3136(struct array_hash_map_IndexHeader__880, uintptr_t);
static bool array_hash_map_Index_28u8_29_isEmpty__5743(struct array_hash_map_Index_28u8_29__6669);
static bool array_hash_map_Index_28u16_29_isEmpty__5748(struct array_hash_map_Index_28u16_29__6673);
static bool array_hash_map_Index_28u32_29_isEmpty__5753(struct array_hash_map_Index_28u32_29__6677);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3316(uintptr_t, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5972(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct array_hash_map_IndexHeader__880 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5973(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct array_hash_map_IndexHeader__880 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5974(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *, struct array_hash_map_IndexHeader__880 *);
static nav__5975_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5975(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5975_41, struct array_hash_map_IndexHeader__880 *, nav__5975_47);
static nav__5976_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5976(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5976_41, struct array_hash_map_IndexHeader__880 *, nav__5976_47);
static nav__5977_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5977(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889, nav__5977_41, struct array_hash_map_IndexHeader__880 *, nav__5977_47);
static nav__5978_39 mem_Allocator_allocWithSizeAndAlignment__anon_7871__5978(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint8_t math_log2_int__anon_7874__5979(uintptr_t);
static nav__5980_39 mem_Allocator_allocWithSizeAndAlignment__anon_7878__5980(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint32_t array_hash_map_hashString__2901(nav__2901_39);
static bool array_hash_map_eqlString__2900(nav__2900_39, nav__2900_39);
static uint32_t array_hash_map_IndexHeader_mask__3141(struct array_hash_map_IndexHeader__880);
static nav__5981_39 mem_Allocator_allocBytesWithAlignment__anon_7904__5981(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint64_t hash_wyhash_Wyhash_hash__6041(uint64_t, nav__6041_39);
static bool mem_eql__anon_7916__6042(nav__6042_39, nav__6042_39);
static struct hash_wyhash_Wyhash__7909 hash_wyhash_Wyhash_init__6029(uint64_t);
static bool mem_eqlBytes__856(nav__856_39, nav__856_39);
static uint16_t environment_root_Environment_setIntegrationEnvironmentList__776(struct environment_root_Environment__559 *, nav__776_41, nav__776_41, nav__776_43);
#define c_instrument_hooks_set_environment_list__256 instrument_hooks_set_environment_list
zig_extern uint8_t instrument_hooks_set_environment_list(struct InstrumentHooks *, char const *, char const *, char const *const *, uint32_t);
static nav__6046_42 mem_Allocator_alloc__anon_8016__6046(struct mem_Allocator__599, uintptr_t);
static void mem_Allocator_free__anon_8019__6047(struct mem_Allocator__599, nav__6047_42);
static nav__6048_39 mem_Allocator_allocWithSizeAndAlignment__anon_8027__6048(struct mem_Allocator__599, uintptr_t, uintptr_t);
static nav__6049_39 mem_sliceAsBytes__anon_8034__6049(nav__6049_43);
#define c_instrument_hooks_write_environment__257 instrument_hooks_write_environment
zig_extern uint8_t instrument_hooks_write_environment(struct InstrumentHooks *, int32_t);
static uint16_t environment_linked_libraries_root_LinkedLibraries_collect__2812(struct environment_linked_libraries_root_LinkedLibraries__837 *);
static uint16_t environment_root_Environment_populateLinkedLibraries__779(struct environment_root_Environment__559 *);
static uint8_t environment_root_Environment_writeEnvironment__780(struct environment_root_Environment__559 *, int32_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3350(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *);
static nav__3357_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3357(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct mem_Allocator__599, nav__3357_44);
static void logger_err__anon_8078__6050(void);
static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2962(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868);
static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3352(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922);
static nav__6052_39 mem_span__anon_8102__6052(uint8_t const *);
static uint16_t fmt_format__anon_8173__6081(struct io_Writer__3827, nav__6081_41);
static nav__6053_40 fmt_bufPrint__anon_8111__6053(nav__6053_39, nav__6053_43);
static void logger_err__anon_8118__6054(void);
static nav__6076_40 json_stringify_stringifyAlloc__anon_8133__6076(struct mem_Allocator__599, struct environment_root_EnvironmentJson__833, struct json_stringify_StringifyOptions__8122);
static void logger_err__anon_8140__6077(void);
static nav__1351_39 fs_createFileAbsolute__1351(nav__1351_41, struct fs_File_CreateFlags__8142);
static void logger_err__anon_8159__6078(void);
static nav__3358_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3358(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct mem_Allocator__599, nav__3358_44);
static void logger_logWithPrefix__anon_8171__6079(void);
static uintptr_t mem_len__anon_8172__6080(uint8_t const *);
static void logger_logWithPrefix__anon_8231__6082(void);
static uint16_t json_stringify_stringifyArbitraryDepth__anon_8258__6112(struct mem_Allocator__599, struct environment_root_EnvironmentJson__833, struct json_stringify_StringifyOptions__8122, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static nav__1533_40 array_list_ArrayListAligned_28u8_2cnull_29_toOwnedSlice__1533(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *);
static void logger_logWithPrefix__anon_8263__6113(void);
static nav__3724_39 fs_Dir_createFile__3724(struct fs_Dir__1103, nav__3724_42, struct fs_File_CreateFlags__8142);
static void logger_logWithPrefix__anon_8268__6114(void);
static nav__6115_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6115(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct mem_Allocator__599, nav__6115_44);
static uint16_t fmt_format__anon_8359__6124(struct io_Writer__3827);
static nav__6116_40 fmt_bufPrint__anon_8285__6116(nav__6116_39);
static uint16_t fmt_format__anon_8435__6125(struct io_Writer__3827);
static nav__6117_40 fmt_bufPrint__anon_8294__6117(nav__6117_39);
static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 json_stringify_writeStreamArbitraryDepth__anon_8302__6118(struct mem_Allocator__599, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct json_stringify_StringifyOptions__8122);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6119(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, struct environment_root_EnvironmentJson__833);
static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6089(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static void array_list_ArrayListAligned_28u8_2cnull_29_clearAndFree__1563(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *);
static uint16_t fmt_format__anon_8599__6128(struct io_Writer__3827);
static nav__6120_40 fmt_bufPrint__anon_8325__6120(nav__6120_39);
static nav__3725_39 fs_Dir_createFileZ__3725(struct fs_Dir__1103, uint8_t const *, struct fs_File_CreateFlags__8142);
static uint16_t fmt_format__anon_8674__6129(struct io_Writer__3827);
static nav__6121_40 fmt_bufPrint__anon_8340__6121(nav__6121_39);
static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3367(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct mem_Allocator__599, uintptr_t);
static nav__6122_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6122(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6122_41);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6123(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, nav__6123_42);
static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6088(struct mem_Allocator__599, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222, struct json_stringify_StringifyOptions__8122);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6096(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6099(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6094(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, uint8_t);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6098(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t fmt_formatBuf__anon_8984__6145(nav__6145_39, struct fmt_FormatOptions__4248, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t fmt_formatInt__anon_8974__6144(uint32_t, uint8_t, uint8_t, struct fmt_FormatOptions__4248, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t fmt_formatIntValue__anon_8949__6143(uint32_t, struct fmt_FormatOptions__4248, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t fmt_formatInt__anon_9000__6147(uint16_t, uint8_t, uint8_t, struct fmt_FormatOptions__4248, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t fmt_formatIntValue__anon_8996__6146(uint16_t, struct fmt_FormatOptions__4248, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t json_stringify_outputUnicodeEscape__anon_8933__6142(uint32_t, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t json_stringify_outputSpecialEscape__anon_8893__6141(uint8_t, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t json_stringify_encodeJsonStringChars__anon_8883__6140(nav__6140_39, struct json_stringify_StringifyOptions__8122, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t json_stringify_encodeJsonString__anon_8873__6139(nav__6139_39, struct json_stringify_StringifyOptions__8122, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, nav__6106_41);
static uint16_t json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29_jsonStringif__6148(struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6126(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849);
static uint16_t json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29_jsonStringify__anon_9047__6150(struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6127(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3497(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *, struct mem_Allocator__599, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3493(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *, struct mem_Allocator__599, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3459(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct array_hash_map_IndexHeader__880 *);
static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(nav__6130_40);
static nav__6131_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393);
static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(nav__6132_40, nav__6132_40, uintptr_t);
static nav__6133_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6133(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6133_41, struct array_hash_map_IndexHeader__880 *);
static nav__6134_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6134(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6134_41, struct array_hash_map_IndexHeader__880 *);
static nav__6135_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6135(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6135_41, struct array_hash_map_IndexHeader__880 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6136(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, nav__6136_42, struct array_hash_map_IndexHeader__880 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6137(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, nav__6137_42, struct array_hash_map_IndexHeader__880 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6138(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, nav__6138_42, struct array_hash_map_IndexHeader__880 *);
static nav__6101_38 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6101(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *);
static bool json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6102(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *);
static uint16_t io_Writer_writeByte__5170(struct io_Writer__3827, uint8_t);
static uint16_t io_Writer_writeByteNTimes__5171(struct io_Writer__3827, uint8_t, uintptr_t);
static uint16_t json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_jsonStringify__anon_9121__6158(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6149(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6151(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, struct environment_linked_libraries_root_LibraryEntry__852);
static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6095(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, uint8_t);
static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3495(uintptr_t, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6152(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct array_hash_map_IndexHeader__880 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6153(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct array_hash_map_IndexHeader__880 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6154(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *, struct array_hash_map_IndexHeader__880 *);
static nav__6155_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6155(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6155_41, struct array_hash_map_IndexHeader__880 *, nav__6155_47);
static nav__6156_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6156(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6156_41, struct array_hash_map_IndexHeader__880 *, nav__6156_47);
static nav__6157_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6157(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922, nav__6157_41, struct array_hash_map_IndexHeader__880 *, nav__6157_47);
static uint16_t environment_root_EnvironmentValue_jsonStringify__anon_9194__6162(struct environment_root_EnvironmentValue__844, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6159(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, struct environment_root_EnvironmentValue__844);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6111(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, nav__6111_41);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6160(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, nav__6160_41);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6161(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, nav__6161_41);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6163(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, nav__6163_43);
static bool unicode_utf8ValidateSlice__5206(nav__5206_39);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6090(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6164(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *, uint8_t);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6092(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6165(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *);
static bool unicode_utf8ValidateSliceImpl__anon_9218__6166(nav__6166_39);
static uint16_t io_Writer_print__anon_9232__6167(struct io_Writer__3827, nav__6167_39);
static uint16_t fmt_format__anon_9241__6168(struct io_Writer__3827, nav__6168_39);
static uint16_t fmt_formatType__anon_9259__6169(uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827, uintptr_t);
static uint16_t fmt_formatInt__anon_9272__6172(uint8_t, uint8_t, uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatIntValue__anon_9268__6171(uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint16_t fmt_formatValue__anon_9264__6170(uint8_t, struct fmt_FormatOptions__4248, struct io_Writer__3827);
static uint64_t const builtin_zig_backend__262;
static bool const start_simplified_logic__109;
static uint8_t const builtin_output_mode__263;
static uint8_t const builtin_link_mode__264;
static uint64_t features_features__329;
static uintptr_t const bit_set_IntegerBitSet_2864_29_bit_length__360;
static struct Target_Os__639 const builtin_os__270;
static bool const builtin_link_libc__274;
static bool const posix_use_libc__1590;
static uint8_t const c_native_os__2032;
static uint8_t const builtin_mode__273;
static bool const debug_runtime_safety__159;
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2940;
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3330;
static struct Target_DynamicLinker__976 const Target_DynamicLinker_none__3578;
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3151;
static bool const builtin_is_test__266;
static uint8_t *heap_CAllocator_alloc__3702(void *, uintptr_t, uint8_t, uintptr_t);
static bool heap_CAllocator_resize__3703(void *, nav__3703_40, uint8_t, uintptr_t, uintptr_t);
static uint8_t *heap_CAllocator_remap__3704(void *, nav__3704_40, uint8_t, uintptr_t, uintptr_t);
static void heap_CAllocator_free__3705(void *, nav__3705_40, uint8_t, uintptr_t);
static uint8_t *heap_CAllocator_alignedAlloc__3699(uintptr_t, uint8_t);
static uintptr_t heap_CAllocator_alignedAllocSize__3701(uint8_t *);
static void heap_CAllocator_alignedFree__3700(uint8_t *);
static uintptr_t mem_Alignment_toByteUnits__1112(uint8_t);
static struct mem_Allocator__599 const heap_c_allocator__3667;
static struct mem_Allocator__599 const c_allocator__235;
static uint8_t const (*const shared_RUNNER_CTL_FIFO__801)[21];
static uint8_t const (*const shared_RUNNER_ACK_FIFO__802)[21];
static uint64_t const runner_fifo_PROTOCOL_VERSION__1125;
static struct Target_Cpu_Feature_Set__993 const Target_Cpu_Feature_Set_empty__3623;
static struct Target_Cpu__961 const builtin_cpu__269;
static uint8_t const mem_native_endian__836;
static uint8_t const posix_native_os__1588;
static uint8_t const builtin_abi__268;
static uint8_t const builtin_object_format__272;
static struct Target__637 const builtin_target__271;
static struct builtin_CallingConvention__266 const builtin_CallingConvention_c__470;
zig_extern int poll(struct c_pollfd__struct_1304__1304 *, uint32_t, int);
zig_extern zig_noreturn void exit(int);
static uint8_t const fs_path_native_os__4279;
static uint8_t const fs_native_os__1303;
static uint8_t const fs_Dir_native_os__3827;
static bool const fs_File_is_windows__1479;
zig_extern int *zig_e___error(void) zig_mangled(zig_e___error, "__error");
static bool const posix_unexpected_error_tracing__2019;
zig_extern int printf(char const *, ...);
static uint8_t logger_max_level__4255;
zig_extern int printf(char const *, ...);
zig_extern int flock(int32_t, int);
static bool const fs_Dir_have_flock__3828;
zig_extern intptr_t write(int32_t, uint8_t const *, uintptr_t);
zig_extern intptr_t read(int32_t, uint8_t *, uintptr_t);
static uint16_t const fmt_max_format_args__4915;
static bool const debug_default_enable_segfault_handler__205;
static uint8_t const log_default_level__5504;
static struct std_Options__4261 const std_options__97;
zig_extern int faccessat(int32_t, uint8_t const *, unsigned int, unsigned int);
static bool const posix_lfs64_abi__2018;
zig_extern int openat(int, uint8_t const *, uint32_t, ...);
zig_extern int fcntl(int32_t, int, ...);
zig_extern int close_24NOCANCEL(int32_t) zig_mangled(close_24NOCANCEL, "close$NOCANCEL");
static uint8_t const (*const fmt_ANY__4918)[4];
static uint8_t const unicode_native_endian__5186;
static uint32_t const unicode_replacement_character__5187;
zig_extern uint8_t running_on_valgrind(void);
zig_extern void callgrind_zero_stats(void);
zig_extern void callgrind_start_instrumentation(void);
zig_extern void callgrind_stop_instrumentation(void);
zig_extern void callgrind_dump_stats_at(uint8_t const *);
static bool const mem_backend_supports_vectors__864;
static bool const builtin_valgrind_support__277;
static nav__3683_38 const heap_page_size_min_default__3683;
static uintptr_t const heap_page_size_min__3662;
static uint8_t const c_MARKER_TYPE_SAMPLE_START__246;
static uint8_t const c_MARKER_TYPE_SAMPLE_END__247;
static uint8_t const c_MARKER_TYPE_BENCHMARK_START__248;
static uint8_t const c_MARKER_TYPE_BENCHMARK_END__249;
static bool const builtin_single_threaded__267;
static struct once_Once_28_28function_20_27init_27_29_29__7273 c_instrument_hooks_current_timestamp_S_once__5831;
zig_extern uint64_t mach_absolute_time(void);
static struct c_MachTimebaseInfo__7314 c_instrument_hooks_current_timestamp_S_cached__5830;
static uint8_t const Thread_native_os__5849;
static bool const Thread_use_pthreads__5860;
zig_extern int mach_timebase_info(struct c_MachTimebaseInfo__7314 *);
zig_extern void os_unfair_lock_lock(struct c_darwin_os_unfair_lock__7296 *);
zig_extern void os_unfair_lock_unlock(struct c_darwin_os_unfair_lock__7296 *);
static uint32_t const array_hash_map_index_capacities__2914[33];
static uint8_t const array_hash_map_Index_28u8_29_empty_sentinel__5741;
static uint16_t const array_hash_map_Index_28u16_29_empty_sentinel__5746;
static uint32_t const array_hash_map_Index_28u32_29_empty_sentinel__5751;
static uint64_t const hash_wyhash_Wyhash_secret__6028[4];
static bool const builtin_fuzz__279;
static bool const mem_eqlBytes_allowed__854;
zig_extern uint8_t const *getenv(uint8_t const *);
static struct array_hash_map_Index_28u8_29__6669 const array_hash_map_Index_28u8_29_empty__5742;
static struct array_hash_map_Index_28u16_29__6673 const array_hash_map_Index_28u16_29_empty__5747;
static struct array_hash_map_Index_28u32_29__6677 const array_hash_map_Index_28u32_29_empty__5752;
static bool const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6084;
static struct json_stringify_WriteStream__union_8236__8236 const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6085;
#define c_dummy_execute_header__2050 _mh_execute_header
zig_extern zig_weak_linkage struct macho_mach_header_64__9281 _mh_execute_header;
static struct mem_Allocator_VTable__602 const heap_CAllocator_vtable__3694;
zig_extern uintptr_t malloc_size(void const *);
static bool const heap_CAllocator_supports_malloc_size__3695;
static bool const heap_CAllocator_supports_posix_memalign__3697;
zig_extern int posix_memalign(void **, uintptr_t, uintptr_t);
zig_extern void free(void *);
static struct Target_Cpu_Model__988 const Target_x86_cpu_core2__4052;
enum {
 zig_error_ModeError = 1u,
 zig_error_OutOfMemory = 2u,
 zig_error_PermissionDenied = 3u,
 zig_error_FileNotFound = 4u,
 zig_error_NameTooLong = 5u,
 zig_error_InputOutput = 6u,
 zig_error_SystemResources = 7u,
 zig_error_BadPathName = 8u,
 zig_error_FileBusy = 9u,
 zig_error_SymLinkLoop = 10u,
 zig_error_ReadOnlyFileSystem = 11u,
 zig_error_InvalidUtf8 = 12u,
 zig_error_InvalidWtf8 = 13u,
 zig_error_Unexpected = 14u,
 zig_error_SharingViolation = 15u,
 zig_error_PathAlreadyExists = 16u,
 zig_error_AccessDenied = 17u,
 zig_error_PipeBusy = 18u,
 zig_error_NoDevice = 19u,
 zig_error_NetworkNotFound = 20u,
 zig_error_AntivirusInterference = 21u,
 zig_error_ProcessFdQuotaExceeded = 22u,
 zig_error_SystemFdQuotaExceeded = 23u,
 zig_error_FileTooBig = 24u,
 zig_error_IsDir = 25u,
 zig_error_NoSpaceLeft = 26u,
 zig_error_NotDir = 27u,
 zig_error_DeviceBusy = 28u,
 zig_error_FileLocksNotSupported = 29u,
 zig_error_WouldBlock = 30u,
 zig_error_DiskQuota = 31u,
 zig_error_InvalidArgument = 32u,
 zig_error_BrokenPipe = 33u,
 zig_error_OperationAborted = 34u,
 zig_error_NotOpenForWriting = 35u,
 zig_error_LockViolation = 36u,
 zig_error_ConnectionResetByPeer = 37u,
 zig_error_ProcessNotFound = 38u,
 zig_error_NetworkSubsystemFailed = 39u,
 zig_error_AckTimeout = 40u,
 zig_error_ConnectionTimedOut = 41u,
 zig_error_NotOpenForReading = 42u,
 zig_error_SocketNotConnected = 43u,
 zig_error_Canceled = 44u,
 zig_error_NotReady = 45u,
 zig_error_UnexpectedEof = 46u,
 zig_error_EndOfStream = 47u,
 zig_error_UnexpectedError = 48u,
 zig_error_UnexpectedResponse = 49u,
 zig_error_Overflow = 50u,
 zig_error_Locked = 51u,
 zig_error_DeadLock = 52u,
 zig_error_LockedRegionLimitExceeded = 53u,
 zig_error_Utf8ExpectedContinuation = 54u,
 zig_error_Utf8OverlongEncoding = 55u,
 zig_error_Utf8EncodesSurrogateHalf = 56u,
 zig_error_Utf8CodepointTooLarge = 57u,
 zig_error_Utf8InvalidStartByte = 58u,
 zig_error_TruncatedInput = 59u,
 zig_error_Utf8CannotEncodeSurrogateHalf = 60u,
 zig_error_CodepointTooLarge = 61u,
};
static uint8_t const zig_errorName_ModeError[10] = "ModeError";
static uint8_t const zig_errorName_OutOfMemory[12] = "OutOfMemory";
static uint8_t const zig_errorName_PermissionDenied[17] = "PermissionDenied";
static uint8_t const zig_errorName_FileNotFound[13] = "FileNotFound";
static uint8_t const zig_errorName_NameTooLong[12] = "NameTooLong";
static uint8_t const zig_errorName_InputOutput[12] = "InputOutput";
static uint8_t const zig_errorName_SystemResources[16] = "SystemResources";
static uint8_t const zig_errorName_BadPathName[12] = "BadPathName";
static uint8_t const zig_errorName_FileBusy[9] = "FileBusy";
static uint8_t const zig_errorName_SymLinkLoop[12] = "SymLinkLoop";
static uint8_t const zig_errorName_ReadOnlyFileSystem[19] = "ReadOnlyFileSystem";
static uint8_t const zig_errorName_InvalidUtf8[12] = "InvalidUtf8";
static uint8_t const zig_errorName_InvalidWtf8[12] = "InvalidWtf8";
static uint8_t const zig_errorName_Unexpected[11] = "Unexpected";
static uint8_t const zig_errorName_SharingViolation[17] = "SharingViolation";
static uint8_t const zig_errorName_PathAlreadyExists[18] = "PathAlreadyExists";
static uint8_t const zig_errorName_AccessDenied[13] = "AccessDenied";
static uint8_t const zig_errorName_PipeBusy[9] = "PipeBusy";
static uint8_t const zig_errorName_NoDevice[9] = "NoDevice";
static uint8_t const zig_errorName_NetworkNotFound[16] = "NetworkNotFound";
static uint8_t const zig_errorName_AntivirusInterference[22] = "AntivirusInterference";
static uint8_t const zig_errorName_ProcessFdQuotaExceeded[23] = "ProcessFdQuotaExceeded";
static uint8_t const zig_errorName_SystemFdQuotaExceeded[22] = "SystemFdQuotaExceeded";
static uint8_t const zig_errorName_FileTooBig[11] = "FileTooBig";
static uint8_t const zig_errorName_IsDir[6] = "IsDir";
static uint8_t const zig_errorName_NoSpaceLeft[12] = "NoSpaceLeft";
static uint8_t const zig_errorName_NotDir[7] = "NotDir";
static uint8_t const zig_errorName_DeviceBusy[11] = "DeviceBusy";
static uint8_t const zig_errorName_FileLocksNotSupported[22] = "FileLocksNotSupported";
static uint8_t const zig_errorName_WouldBlock[11] = "WouldBlock";
static uint8_t const zig_errorName_DiskQuota[10] = "DiskQuota";
static uint8_t const zig_errorName_InvalidArgument[16] = "InvalidArgument";
static uint8_t const zig_errorName_BrokenPipe[11] = "BrokenPipe";
static uint8_t const zig_errorName_OperationAborted[17] = "OperationAborted";
static uint8_t const zig_errorName_NotOpenForWriting[18] = "NotOpenForWriting";
static uint8_t const zig_errorName_LockViolation[14] = "LockViolation";
static uint8_t const zig_errorName_ConnectionResetByPeer[22] = "ConnectionResetByPeer";
static uint8_t const zig_errorName_ProcessNotFound[16] = "ProcessNotFound";
static uint8_t const zig_errorName_NetworkSubsystemFailed[23] = "NetworkSubsystemFailed";
static uint8_t const zig_errorName_AckTimeout[11] = "AckTimeout";
static uint8_t const zig_errorName_ConnectionTimedOut[19] = "ConnectionTimedOut";
static uint8_t const zig_errorName_NotOpenForReading[18] = "NotOpenForReading";
static uint8_t const zig_errorName_SocketNotConnected[19] = "SocketNotConnected";
static uint8_t const zig_errorName_Canceled[9] = "Canceled";
static uint8_t const zig_errorName_NotReady[9] = "NotReady";
static uint8_t const zig_errorName_UnexpectedEof[14] = "UnexpectedEof";
static uint8_t const zig_errorName_EndOfStream[12] = "EndOfStream";
static uint8_t const zig_errorName_UnexpectedError[16] = "UnexpectedError";
static uint8_t const zig_errorName_UnexpectedResponse[19] = "UnexpectedResponse";
static uint8_t const zig_errorName_Overflow[9] = "Overflow";
static uint8_t const zig_errorName_Locked[7] = "Locked";
static uint8_t const zig_errorName_DeadLock[9] = "DeadLock";
static uint8_t const zig_errorName_LockedRegionLimitExceeded[26] = "LockedRegionLimitExceeded";
static uint8_t const zig_errorName_Utf8ExpectedContinuation[25] = "Utf8ExpectedContinuation";
static uint8_t const zig_errorName_Utf8OverlongEncoding[21] = "Utf8OverlongEncoding";
static uint8_t const zig_errorName_Utf8EncodesSurrogateHalf[25] = "Utf8EncodesSurrogateHalf";
static uint8_t const zig_errorName_Utf8CodepointTooLarge[22] = "Utf8CodepointTooLarge";
static uint8_t const zig_errorName_Utf8InvalidStartByte[21] = "Utf8InvalidStartByte";
static uint8_t const zig_errorName_TruncatedInput[15] = "TruncatedInput";
static uint8_t const zig_errorName_Utf8CannotEncodeSurrogateHalf[30] = "Utf8CannotEncodeSurrogateHalf";
static uint8_t const zig_errorName_CodepointTooLarge[18] = "CodepointTooLarge";
static struct anon__lazy_58 const zig_errorName[62] = {{zig_errorName_ModeError, 9ul}, {zig_errorName_OutOfMemory, 11ul}, {zig_errorName_PermissionDenied, 16ul}, {zig_errorName_FileNotFound, 12ul}, {zig_errorName_NameTooLong, 11ul}, {zig_errorName_InputOutput, 11ul}, {zig_errorName_SystemResources, 15ul}, {zig_errorName_BadPathName, 11ul}, {zig_errorName_FileBusy, 8ul}, {zig_errorName_SymLinkLoop, 11ul}, {zig_errorName_ReadOnlyFileSystem, 18ul}, {zig_errorName_InvalidUtf8, 11ul}, {zig_errorName_InvalidWtf8, 11ul}, {zig_errorName_Unexpected, 10ul}, {zig_errorName_SharingViolation, 16ul}, {zig_errorName_PathAlreadyExists, 17ul}, {zig_errorName_AccessDenied, 12ul}, {zig_errorName_PipeBusy, 8ul}, {zig_errorName_NoDevice, 8ul}, {zig_errorName_NetworkNotFound, 15ul}, {zig_errorName_AntivirusInterference, 21ul}, {zig_errorName_ProcessFdQuotaExceeded, 22ul}, {zig_errorName_SystemFdQuotaExceeded, 21ul}, {zig_errorName_FileTooBig, 10ul}, {zig_errorName_IsDir, 5ul}, {zig_errorName_NoSpaceLeft, 11ul}, {zig_errorName_NotDir, 6ul}, {zig_errorName_DeviceBusy, 10ul}, {zig_errorName_FileLocksNotSupported, 21ul}, {zig_errorName_WouldBlock, 10ul}, {zig_errorName_DiskQuota, 9ul}, {zig_errorName_InvalidArgument, 15ul}, {zig_errorName_BrokenPipe, 10ul}, {zig_errorName_OperationAborted, 16ul}, {zig_errorName_NotOpenForWriting, 17ul}, {zig_errorName_LockViolation, 13ul}, {zig_errorName_ConnectionResetByPeer, 21ul}, {zig_errorName_ProcessNotFound, 15ul}, {zig_errorName_NetworkSubsystemFailed, 22ul}, {zig_errorName_AckTimeout, 10ul}, {zig_errorName_ConnectionTimedOut, 18ul}, {zig_errorName_NotOpenForReading, 17ul}, {zig_errorName_SocketNotConnected, 18ul}, {zig_errorName_Canceled, 8ul}, {zig_errorName_NotReady, 8ul}, {zig_errorName_UnexpectedEof, 13ul}, {zig_errorName_EndOfStream, 11ul}, {zig_errorName_UnexpectedError, 15ul}, {zig_errorName_UnexpectedResponse, 18ul}, {zig_errorName_Overflow, 8ul}, {zig_errorName_Locked, 6ul}, {zig_errorName_DeadLock, 8ul}, {zig_errorName_LockedRegionLimitExceeded, 25ul}, {zig_errorName_Utf8ExpectedContinuation, 24ul}, {zig_errorName_Utf8OverlongEncoding, 20ul}, {zig_errorName_Utf8EncodesSurrogateHalf, 24ul}, {zig_errorName_Utf8CodepointTooLarge, 21ul}, {zig_errorName_Utf8InvalidStartByte, 20ul}, {zig_errorName_TruncatedInput, 14ul}, {zig_errorName_Utf8CannotEncodeSurrogateHalf, 29ul}, {zig_errorName_CodepointTooLarge, 17ul}};
static struct anon__lazy_58 zig_tag_name_shared_IntegrationMode__573(uint8_t const tag) {
 switch (tag) {
  case UINT8_C(0): {
   static uint8_t const name[9] = "Walltime";
   return (struct anon__lazy_58){name, (uintptr_t)8ul};
  }
  case UINT8_C(1): {
   static uint8_t const name[11] = "Simulation";
   return (struct anon__lazy_58){name, (uintptr_t)10ul};
  }
  case UINT8_C(2): {
   static uint8_t const name[9] = "Analysis";
   return (struct anon__lazy_58){name, (uintptr_t)8ul};
  }
 }
 while (true) zig_breakpoint();
}

static uint8_t const __anon_1087[21] = "/tmp/runner.ctl.fifo";

static uint8_t const __anon_1152[21] = "/tmp/runner.ack.fifo";

static char const __anon_3329[33] = {73,110,118,97,108,105,100,32,112,114,111,116,111,99,111,108,32,100,101,116,101,99,116,101,100,58,32,37,46,42,115,10,0};

static uint8_t const __anon_2303[62] = "The stream end was found before all required bytes were read.";

static uint8_t const __anon_3411[34] = "[ERROR] logger formatting failed\n";

static uint8_t const __anon_3951[49] = "[DEBUG] waitForAck received unexpected response: ";

static uint8_t const __anon_4335[1] = "\n";

static uint8_t const __anon_4451[68] = "[ERROR] instrument-hooks: CodSpeed runner rejected protocol version ";

static uint8_t const __anon_4574[82] = "[ERROR] instrument-hooks: please update the CodSpeed action to the latest version\n";

static uint8_t const __anon_4628[32] = "[ERROR] instrument-hooks: error ";

static uint8_t const __anon_4761[22] = " during version check\n";

static uint8_t const __anon_2011[15] = "StartBenchmark";

static uint8_t const __anon_2016[14] = "StopBenchmark";

static uint8_t const __anon_2021[4] = "Ack";

static uint8_t const __anon_2026[13] = "PingProfiler";

static uint8_t const __anon_2034[4] = "Err";

static uint8_t const __anon_2049[19] = "GetIntegrationMode";

static uint8_t const __anon_5099[8] = "{ ... }";

static uint8_t const __anon_5278[3] = "\357\277\275";

static uint8_t const __anon_5352[25] = "ExecutedBenchmark { pid: ";

static uint8_t const __anon_5418[7] = ", uri: ";

static uint8_t const __anon_5444[2] = " }";

static uint8_t const __anon_5497[23] = "SetIntegration { name: ";

static uint8_t const __anon_5522[11] = ", version: ";

static uint8_t const __anon_5578[17] = "AddMarker { pid: ";

static uint8_t const __anon_5600[10] = ", marker: ";

static uint8_t const __anon_5670[31] = "SetVersion { protocol_version: ";

static uint8_t const __anon_5725[24] = "IntegrationModeResponse ";

static uint8_t const __anon_5841[23] = "shared.IntegrationMode";

static uint8_t const __anon_5846[2] = ".";

static uint8_t const __anon_5938[12] = "SampleStart(";

static uint8_t const __anon_5948[1] = ")";

static uint8_t const __anon_5967[10] = "SampleEnd(";

static uint8_t const __anon_5999[15] = "BenchmarkStart(";

static uint8_t const __anon_6027[13] = "BenchmarkEnd(";

static uint8_t const __anon_7177[10] = "Metadata: ";

static uint8_t const __anon_7192[1] = " ";

static uint8_t const __anon_7205[1] = "\000";

static uint8_t const __anon_8092[24] = "CODSPEED_PROFILE_FOLDER";

static uint8_t const __anon_8208[13] = "/environment-";

static uint8_t const __anon_8225[5] = ".json";

static uint8_t const __anon_8429[62] = "[ERROR] instrument-hooks: failed to populate linked libraries\n";

static uint8_t const __anon_8496[55] = "[ERROR] instrument-hooks: profile folder path too long\n";

static uint8_t const __anon_8509[24] = "integration_environment";

static uint8_t const __anon_8517[17] = "linked_libraries";

static uint8_t const __anon_8668[63] = "[ERROR] instrument-hooks: failed to serialize environment JSON\n";

static uint8_t const __anon_8741[59] = "[ERROR] instrument-hooks: failed to write environment.json\n";

static uint8_t const __anon_8939[3] = "\\u";

static uint8_t const __anon_8898[3] = "\\\\";

static uint8_t const __anon_8903[3] = "\\\"";

static uint8_t const __anon_8907[3] = "\\b";

static uint8_t const __anon_8911[3] = "\\f";

static uint8_t const __anon_8915[3] = "\\n";

static uint8_t const __anon_8919[3] = "\\r";

static uint8_t const __anon_8923[3] = "\\t";

static uint8_t const __anon_9129[5] = "path";

static uint8_t const __anon_9133[9] = "build_id";

static uint8_t const __anon_9236[5] = "null";

static uint8_t const __anon_4937[4] = "any";

static uint8_t const __anon_9352[6] = "core2";

void c_instrument_hooks_set_feature__237(uint64_t const a0, bool const a1) {
 uint64_t t0;
 t0 = a0;
 features_set_feature__330(t0, a1);
 return;
}

static void features_set_feature__330(uint64_t const a0, bool const a1) {
 uint64_t t0;
 uintptr_t t1;
 if (a1) {
  t0 = a0;
  t1 = t0;
  bit_set_IntegerBitSet_2864_29_set__369(&features_features__329, t1);
  goto zig_block_0;
 }
 t0 = a0;
 t1 = t0;
 bit_set_IntegerBitSet_2864_29_unset__371(&features_features__329, t1);
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void bit_set_IntegerBitSet_2864_29_set__369(uint64_t *const a0, uintptr_t const a1) {
 uint64_t *const *t1;
 uint64_t t2;
 uint64_t t5;
 uint64_t *t4;
 uint64_t *t0;
 bool t3;
 t0 = a0;
 t1 = (uint64_t *const *)&t0;
 t2 = a1;
 t3 = t2 < UINT64_C(64);
 debug_assert__180(t3);
 t4 = (*t1);
 t4 = (uint64_t *)((uint8_t *)t4 + (uintptr_t)0ul);
 t2 = (*t4);
 t5 = bit_set_IntegerBitSet_2864_29_maskBit__390(a1);
 t5 = t2 | t5;
 (*t4) = t5;
 return;
}

static void bit_set_IntegerBitSet_2864_29_unset__371(uint64_t *const a0, uintptr_t const a1) {
 uint64_t *const *t1;
 uint64_t t2;
 uint64_t t5;
 uint64_t *t4;
 uint64_t *t0;
 bool t3;
 t0 = a0;
 t1 = (uint64_t *const *)&t0;
 t2 = a1;
 t3 = t2 < UINT64_C(64);
 debug_assert__180(t3);
 t4 = (*t1);
 t4 = (uint64_t *)((uint8_t *)t4 + (uintptr_t)0ul);
 t2 = (*t4);
 t5 = bit_set_IntegerBitSet_2864_29_maskBit__390(a1);
 t5 = zig_not_u64(t5, UINT8_C(64));
 t5 = t2 & t5;
 (*t4) = t5;
 return;
}

static void debug_assert__180(bool const a0) {
 bool t0;
 t0 = !a0;
 if (t0) {
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static uint64_t bit_set_IntegerBitSet_2864_29_maskBit__390(uintptr_t const a0) {
 uint64_t t1;
 uint8_t t0;
 t0 = (uint8_t)a0;
 t1 = zig_shlw_u64(UINT64_C(1), t0, UINT8_C(64));
 return t1;
}

static nav__1280_39 fifo_posix_Pipe_openPipe__1280(nav__1280_41 const a0) {
 nav__1280_39 t1;
 struct fs_File__794 t2;
 uint16_t t0;
 t0 = fs_accessAbsolute__1348(a0, (struct fs_File_OpenFlags__1101){UINT8_C(2),UINT8_C(0),false,false});
 if (t0) {
  t1.payload = (struct fs_File__794){-INT32_C(0x55555556)};
  t1.error = t0;
  return t1;
 }
 t1 = fs_openFileAbsolute__1345(a0, (struct fs_File_OpenFlags__1101){UINT8_C(2),UINT8_C(0),false,false});
 if (t1.error) {
  t0 = t1.error;
  t1.payload = (struct fs_File__794){-INT32_C(0x55555556)};
  t1.error = t0;
  return t1;
 }
 t2 = t1.payload;
 t1.payload = t2;
 t1.error = UINT16_C(0);
 return t1;
}

static nav__1282_39 fifo_posix_Pipe_openWrite__1282(struct mem_Allocator__599 const a0, nav__1282_42 const a1) {
 nav__1282_39 t2;
 struct fifo_posix_Pipe_Writer__786 t4;
 nav__1282_59 t0;
 struct fs_File__794 t3;
 uint16_t t1;
 t0 = fifo_posix_Pipe_openPipe__1280(a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct fifo_posix_Pipe_Writer__786){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = fifo_posix_Pipe_Writer_init__1283(t3, a0);
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__1281_39 fifo_posix_Pipe_openRead__1281(struct mem_Allocator__599 const a0, nav__1281_42 const a1) {
 nav__1281_39 t2;
 struct fifo_posix_Pipe_Reader__788 t4;
 nav__1281_59 t0;
 struct fs_File__794 t3;
 uint16_t t1;
 t0 = fifo_posix_Pipe_openPipe__1280(a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct fifo_posix_Pipe_Reader__788){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = fifo_posix_Pipe_Reader_init__1288(t3, a0);
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__1129_39 runner_fifo_RunnerFifo_init__1129(struct mem_Allocator__599 const a0) {
 struct runner_fifo_RunnerFifo__631 *t1;
 struct mem_Allocator__599 *t2;
 struct fifo_posix_Pipe_Writer__786 *t3;
 nav__1129_66 t4;
 nav__1129_39 t6;
 nav__1129_39 t0;
 struct fifo_posix_Pipe_Writer__786 t7;
 struct fifo_posix_Pipe_Reader__788 *t8;
 nav__1129_71 t9;
 struct fifo_posix_Pipe_Reader__788 t10;
 uint16_t t5;
 t0.error = UINT16_C(0);
 t1 = &t0.payload;
 t2 = (struct mem_Allocator__599 *)&t1->allocator;
 (*t2) = a0;
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t1->writer;
 t4 = fifo_posix_Pipe_openWrite__1282(a0, (nav__1129_64){(uint8_t const *)&__anon_1087,(uintptr_t)20ul});
 if (t4.error) {
  t5 = t4.error;
  t6.payload = (struct runner_fifo_RunnerFifo__631){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}};
  t6.error = t5;
  return t6;
 }
 t7 = t4.payload;
 (*t3) = t7;
 t8 = (struct fifo_posix_Pipe_Reader__788 *)&t1->reader;
 t9 = fifo_posix_Pipe_openRead__1281(a0, (nav__1129_64){(uint8_t const *)&__anon_1152,(uintptr_t)20ul});
 if (t9.error) {
  t5 = t9.error;
  t6.payload = (struct runner_fifo_RunnerFifo__631){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}};
  t6.error = t5;
  return t6;
 }
 t10 = t9.payload;
 (*t8) = t10;
 return t0;
}

static uint16_t fifo_posix_Pipe_Writer_sendCmd__1286(struct fifo_posix_Pipe_Writer__786 *const a0, struct shared_Command__1178 const a1) {
 struct fifo_posix_Pipe_Writer__786 *const *t1;
 struct fifo_posix_Pipe_Writer__786 *t2;
 struct fifo_posix_Pipe_Writer__786 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t6;
 nav__1286_65 t7;
 struct fs_File__794 *t8;
 uintptr_t t10;
 uint32_t const *t13;
 uint8_t const (*t14)[4];
 nav__1286_44 t15;
 struct fs_File__794 t9;
 uint32_t t11;
 uint32_t t12;
 uint16_t t5;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Writer__786 *const *)&t0;
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t2->buffer;
 array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1562(t3);
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t2->buffer;
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(t3);
 t5 = bincode_serialize__anon_1232__3956(t4, a1);
 if (t5) {
  return t5;
 }
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&a0->buffer;
 t6 = (*t3);
 t7 = t6.items;
 t2 = (*t1);
 t8 = (struct fs_File__794 *)&t2->file;
 t9 = (*t8);
 t10 = t7.len;
 t11 = (uint32_t)t10;
 t12 = t11;
 t13 = (uint32_t const *)&t12;
 t14 = mem_asBytes__anon_1270__3957(t13);
 t15.ptr = &(*t14)[(uintptr_t)0ul];
 t15.len = (uintptr_t)4ul;
 t5 = fs_File_writeAll__1439(t9, t15);
 if (t5) {
  return t5;
 }
 t2 = (*t1);
 t8 = (struct fs_File__794 *)&t2->file;
 t9 = (*t8);
 memcpy(&t15, &t7, sizeof(nav__1286_44));
 t5 = fs_File_writeAll__1439(t9, t15);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t fifo_posix_waitReadable__1268(int32_t const a0, uint64_t const a1) {
 uint64_t t3;
 uintptr_t t7;
 uintptr_t t10;
 nav__1268_43 t8;
 nav__1268_46 t9;
 struct c_pollfd__struct_1304__1304 t1;
 struct c_pollfd__struct_1304__1304 t2[1];
 struct c_pollfd__struct_1304__1304 t0[1];
 int32_t t4;
 int32_t t6;
 bool t5;
 t1.fd = a0;
 t1.events = INT16_C(1);
 t1.revents = INT16_C(0);
 t2[0] = t1;
 memcpy((char *)&t0, t2, sizeof(struct c_pollfd__struct_1304__1304[1]));
 t3 = a1 / UINT64_C(1000000);
 t5 = t3 > UINT64_C(2147483647);
 if (t5) {
  t4 = INT32_MAX;
  goto zig_block_0;
 }
 t6 = (int32_t)t3;
 t4 = t6;
 goto zig_block_0;

 zig_block_0:;
 t8.ptr = (struct c_pollfd__struct_1304__1304 *)&t0[(uintptr_t)0ul];
 t8.len = (uintptr_t)1ul;
 t9 = posix_poll__1964(t8, t4);
 t5 = t9.error == UINT16_C(0);
 if (t5) {
  t10 = t9.payload;
  t7 = t10;
  goto zig_block_1;
 }
 return zig_error_AckTimeout;

 zig_block_1:;
 t3 = t7;
 t5 = t3 == UINT64_C(0);
 if (t5) {
  return zig_error_AckTimeout;
 }
 goto zig_block_2;

 zig_block_2:;
 return 0;
}

static nav__4193_38 bincode_deserializeInt__anon_2126__4193(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t8;
 void const *t9;
 nav__4193_50 (**t10)(void const *, nav__4193_52);
 struct io_Reader__2141 t11;
 struct io_Reader__2141 t6;
 struct io_Reader__2141 t12;
 struct io_Reader__2141 t15;
 struct io_Reader__2141 const *t13;
 uint8_t const (*t21)[4];
 nav__4193_38 t14;
 nav__4193_38 t18;
 uint32_t t22;
 nav__4193_64 t16;
 uint16_t t17;
 uint8_t t19[4];
 uint8_t t20[4];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__4193_50 (**)(void const *, nav__4193_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2141 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2141 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2172__4231(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = UINT32_C(0xaaaaaaaa);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[4]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[4]));
 t21 = (uint8_t const (*)[4])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[4]));
 memcpy(&t22, &t19, sizeof(uint32_t));
 t22 = zig_wrap_u32(t22, UINT8_C(32));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__4193_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = UINT32_C(0xaaaaaaaa);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4192_38 bincode_deserializeAlloc__anon_2115__4192(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4192_38 t0;
 uint32_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2126__4193(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT32_C(0xaaaaaaaa);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4235_38 bincode_deserializeInt__anon_2235__4235(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t8;
 void const *t9;
 nav__4235_50 (**t10)(void const *, nav__4235_52);
 struct io_Reader__2141 t11;
 struct io_Reader__2141 t6;
 struct io_Reader__2141 t12;
 struct io_Reader__2141 t15;
 struct io_Reader__2141 const *t13;
 uint8_t const (*t21)[4];
 nav__4235_38 t14;
 nav__4235_38 t18;
 int32_t t22;
 nav__4235_64 t16;
 uint16_t t17;
 uint8_t t19[4];
 uint8_t t20[4];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__4235_50 (**)(void const *, nav__4235_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2141 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2141 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2172__4231(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = -INT32_C(0x55555556);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[4]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[4]));
 t21 = (uint8_t const (*)[4])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[4]));
 memcpy(&t22, &t19, sizeof(int32_t));
 t22 = zig_wrap_i32(t22, UINT8_C(32));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__4235_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = -INT32_C(0x55555556);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4234_38 bincode_deserializeAlloc__anon_2230__4234(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4234_38 t0;
 int32_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2235__4235(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = -INT32_C(0x55555556);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4237_40 bincode_deserializePointerAlloc__anon_2265__4237(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t6;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t7;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t32;
 struct mem_Allocator__599 const *t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t5;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t31;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *t8;
 void const **t10;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t11;
 void const *t12;
 nav__4237_59 (**t13)(void const *, nav__4237_39);
 struct io_Reader__2141 t14;
 struct io_Reader__2141 t9;
 struct io_Reader__2141 t15;
 struct io_Reader__2141 t18;
 struct io_Reader__2141 t33;
 struct io_Reader__2141 t34;
 struct io_Reader__2141 const *t16;
 nav__4237_69 t17;
 nav__4237_69 t21;
 uint8_t const (*t24)[8];
 uint64_t t25;
 uint64_t t38;
 nav__4237_40 t26;
 uintptr_t t27;
 uintptr_t t37;
 struct mem_Allocator__599 t29;
 struct mem_Allocator__599 t2;
 nav__4237_39 t30;
 nav__4237_39 t28;
 nav__4237_59 t35;
 nav__4237_59 t36;
 nav__4237_72 t19;
 uint16_t t20;
 uint8_t t22[8];
 uint8_t t23[8];
 bool t39;
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t1);
 t5 = t4;
 t6 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t5;
 t7 = t6;
 t8 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *)&t7;
 t10 = (void const **)&t9.context;
 t6 = (*t8);
 t11 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t6->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__4237_59 (**)(void const *, nav__4237_39))&t9.readFn;
 (*t13) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189;
 t14 = t9;
 t15 = t14;
 t16 = (struct io_Reader__2141 const *)&t15;
 t14 = (*t16);
 t18 = t14;
 t16 = (struct io_Reader__2141 const *)&t18;
 t14 = (*t16);
 t19 = io_Reader_readBytesNoEof__anon_2278__4238(t14);
 if (t19.error) {
  t20 = t19.error;
  t21.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t21.error = t20;
  t17 = t21;
  goto zig_block_0;
 }
 memcpy(t22, t19.payload, sizeof(uint8_t[8]));
 memcpy((char *)&t23, t22, sizeof(uint8_t[8]));
 t24 = (uint8_t const (*)[8])&t23;
 memcpy(t22, (const char *)t24, sizeof(uint8_t[8]));
 memcpy(&t25, &t22, sizeof(uint64_t));
 t25 = zig_wrap_u64(t25, UINT8_C(64));
 t21.payload = t25;
 t21.error = UINT16_C(0);
 t17 = t21;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t21, &t17, sizeof(nav__4237_69));
 if (t21.error) {
  t20 = t21.error;
  t26.payload = (nav__4237_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t26.error = t20;
  return t26;
 }
 t25 = t21.payload;
 t27 = t25;
 t29 = (*t3);
 t26 = mem_Allocator_alloc__anon_2287__4239(t29, t27);
 if (t26.error) {
  t20 = t26.error;
  t26.payload = (nav__4237_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t26.error = t20;
  return t26;
 }
 t30 = t26.payload;
 t28 = t30;
 t4 = (*t1);
 t30 = t28;
 t31 = t4;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t31;
 t32 = t1;
 t8 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *)&t32;
 t10 = (void const **)&t33.context;
 t1 = (*t8);
 t11 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t1->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__4237_59 (**)(void const *, nav__4237_39))&t33.readFn;
 (*t13) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189;
 t14 = t33;
 t34 = t14;
 t16 = (struct io_Reader__2141 const *)&t34;
 t14 = (*t16);
 t35 = io_Reader_readAll__4196(t14, t30);
 memcpy(&t36, &t35, sizeof(nav__4237_59));
 t37 = t36.payload;
 t25 = t37;
 t38 = t27;
 t39 = t25 != t38;
 if (t39) {
  bincode_invalidProtocol__anon_2310__4240();
  zig_unreachable();
 }
 goto zig_block_1;

 zig_block_1:;
 t30 = t28;
 t26.payload = t30;
 t26.error = UINT16_C(0);
 return t26;
}

static nav__4236_40 bincode_deserializeAlloc__anon_2252__4236(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4236_57 t0;
 nav__4236_40 t2;
 nav__4236_55 t3;
 nav__4236_39 t4;
 uint16_t t1;
 t0 = bincode_deserializePointerAlloc__anon_2265__4237(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (nav__4236_39){(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 memcpy(&t4, &t3, sizeof(nav__4236_39));
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__4233_39 bincode_deserializeStructAlloc__anon_2214__4233(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 int32_t *t1;
 nav__4233_39 t4;
 nav__4233_44 *t6;
 nav__4233_60 t7;
 nav__4233_44 t8;
 struct shared_Command__struct_1181__1181 t9;
 struct shared_Command__struct_1181__1181 t0;
 nav__4233_57 t2;
 int32_t t5;
 uint16_t t3;
 t1 = (int32_t *)&t0.pid;
 t2 = bincode_deserializeAlloc__anon_2230__4234(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1181__1181){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = (nav__4233_44 *)&t0.uri;
 t7 = bincode_deserializeAlloc__anon_2252__4236(a0, a1);
 if (t7.error) {
  t3 = t7.error;
  t4.payload = (struct shared_Command__struct_1181__1181){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t8 = t7.payload;
 (*t6) = t8;
 t9 = t0;
 t4.payload = t9;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4232_39 bincode_deserializeAlloc__anon_2185__4232(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4232_39 t0;
 struct shared_Command__struct_1181__1181 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2214__4233(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_1181__1181){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static uint16_t bincode_deserializeAlloc__anon_2325__4241(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 (void)a0;
 (void)a1;
 return 0;
}

static nav__4243_39 bincode_deserializeStructAlloc__anon_2391__4243(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4243_44 *t1;
 nav__4243_57 t2;
 nav__4243_39 t4;
 nav__4243_44 t5;
 struct shared_Command__struct_1182__1182 t6;
 struct shared_Command__struct_1182__1182 t0;
 uint16_t t3;
 t1 = (nav__4243_44 *)&t0.name;
 t2 = bincode_deserializeAlloc__anon_2252__4236(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1182__1182){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t1 = (nav__4243_44 *)&t0.version;
 t2 = bincode_deserializeAlloc__anon_2252__4236(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1182__1182){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = t0;
 t4.payload = t6;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4242_39 bincode_deserializeAlloc__anon_2372__4242(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4242_39 t0;
 struct shared_Command__struct_1182__1182 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2391__4243(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_1182__1182){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4249_38 bincode_deserializeInt__anon_2513__4249(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t8;
 void const *t9;
 nav__4249_50 (**t10)(void const *, nav__4249_52);
 struct io_Reader__2141 t11;
 struct io_Reader__2141 t6;
 struct io_Reader__2141 t12;
 struct io_Reader__2141 t15;
 struct io_Reader__2141 const *t13;
 nav__4249_38 t14;
 nav__4249_38 t18;
 uint8_t const (*t21)[8];
 uint64_t t22;
 nav__4249_64 t16;
 uint16_t t17;
 uint8_t t19[8];
 uint8_t t20[8];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__4249_50 (**)(void const *, nav__4249_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2141 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2141 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2278__4238(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[8]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[8]));
 t21 = (uint8_t const (*)[8])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[8]));
 memcpy(&t22, &t19, sizeof(uint64_t));
 t22 = zig_wrap_u64(t22, UINT8_C(64));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__4249_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4248_38 bincode_deserializeAlloc__anon_2508__4248(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4248_38 t0;
 uint64_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2513__4249(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4247_39 bincode_deserializeUnionAlloc__anon_2495__4247(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4247_39 t2;
 nav__4247_56 t6;
 uint64_t t7;
 struct shared_MarkerType__1185 t8;
 nav__4247_54 t0;
 uint32_t t3;
 uint16_t t1;
 uint8_t t4;
 bool t5;
 t0 = bincode_deserializeAlloc__anon_2115__4192(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct shared_MarkerType__1185){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t5 = t4 == UINT8_C(0);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2508__4248(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1185){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(0);
  t8.payload.SampleStart = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t4 == UINT8_C(1);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2508__4248(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1185){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(1);
  t8.payload.SampleEnd = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = t4 == UINT8_C(2);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2508__4248(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1185){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(2);
  t8.payload.BenchmarkStart = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_2;

 zig_block_2:;
 t5 = t4 == UINT8_C(3);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2508__4248(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1185){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(3);
  t8.payload.BenchmarkEnd = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_3;

 zig_block_3:;
 zig_unreachable();
}

static nav__4246_39 bincode_deserializeAlloc__anon_2462__4246(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4246_39 t0;
 struct shared_MarkerType__1185 t2;
 uint16_t t1;
 t0 = bincode_deserializeUnionAlloc__anon_2495__4247(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_MarkerType__1185){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4245_39 bincode_deserializeStructAlloc__anon_2444__4245(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 int32_t *t1;
 nav__4245_39 t4;
 struct shared_MarkerType__1185 *t6;
 nav__4245_60 t7;
 struct shared_MarkerType__1185 t8;
 struct shared_Command__struct_1183__1183 t9;
 struct shared_Command__struct_1183__1183 t0;
 nav__4245_57 t2;
 int32_t t5;
 uint16_t t3;
 t1 = (int32_t *)&t0.pid;
 t2 = bincode_deserializeAlloc__anon_2230__4234(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1183__1183){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = (struct shared_MarkerType__1185 *)&t0.marker;
 t7 = bincode_deserializeAlloc__anon_2462__4246(a0, a1);
 if (t7.error) {
  t3 = t7.error;
  t4.payload = (struct shared_Command__struct_1183__1183){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t8 = t7.payload;
 (*t6) = t8;
 t9 = t0;
 t4.payload = t9;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4244_39 bincode_deserializeAlloc__anon_2430__4244(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4244_39 t0;
 struct shared_Command__struct_1183__1183 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2444__4245(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_1183__1183){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},-INT32_C(0x55555556)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4251_38 bincode_deserializeEnum__anon_2598__4251(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0) {
 nav__4251_46 t0;
 uint32_t t3;
 uint16_t t1;
 nav__4251_38 t2;
 uint8_t t4;
 t0 = bincode_deserializeInt__anon_2126__4193(a0);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = UINT8_C(0x2);
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__4250_38 bincode_deserializeAlloc__anon_2568__4250(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4250_38 t0;
 uint16_t t1;
 uint8_t t2;
 (void)a1;
 t0 = bincode_deserializeEnum__anon_2598__4251(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT8_C(0x2);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__4191_39 bincode_deserializeUnionAlloc__anon_2110__4191(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4191_39 t2;
 nav__4191_68 t6;
 struct shared_Command__struct_1181__1181 t7;
 struct shared_Command__1178 t8;
 nav__4191_70 t9;
 struct shared_Command__struct_1182__1182 t10;
 nav__4191_72 t11;
 struct shared_Command__struct_1183__1183 t12;
 nav__4191_74 t13;
 uint64_t t14;
 nav__4191_66 t0;
 uint32_t t3;
 uint16_t t1;
 nav__4191_76 t15;
 uint8_t t4;
 bool t5;
 t0 = bincode_deserializeAlloc__anon_2115__4192(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t5 = t4 == UINT8_C(0);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2185__4232(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(0);
  t8.payload.ExecutedBenchmark = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t4 == UINT8_C(1);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2325__4241(a0, a1);
  return (nav__4191_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(1)},0};
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = t4 == UINT8_C(2);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2325__4241(a0, a1);
  return (nav__4191_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(2)},0};
 }
 goto zig_block_2;

 zig_block_2:;
 t5 = t4 == UINT8_C(3);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2325__4241(a0, a1);
  return (nav__4191_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(3)},0};
 }
 goto zig_block_3;

 zig_block_3:;
 t5 = t4 == UINT8_C(4);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2325__4241(a0, a1);
  return (nav__4191_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(4)},0};
 }
 goto zig_block_4;

 zig_block_4:;
 t5 = t4 == UINT8_C(5);
 if (t5) {
  t9 = bincode_deserializeAlloc__anon_2372__4242(a0, a1);
  if (t9.error) {
   t1 = t9.error;
   t2.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t10 = t9.payload;
  t8.tag = UINT8_C(5);
  t8.payload.SetIntegration = t10;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_5;

 zig_block_5:;
 t5 = t4 == UINT8_C(6);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2325__4241(a0, a1);
  return (nav__4191_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(6)},0};
 }
 goto zig_block_6;

 zig_block_6:;
 t5 = t4 == UINT8_C(7);
 if (t5) {
  t11 = bincode_deserializeAlloc__anon_2430__4244(a0, a1);
  if (t11.error) {
   t1 = t11.error;
   t2.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t12 = t11.payload;
  t8.tag = UINT8_C(7);
  t8.payload.AddMarker = t12;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_7;

 zig_block_7:;
 t5 = t4 == UINT8_C(8);
 if (t5) {
  t13 = bincode_deserializeAlloc__anon_2508__4248(a0, a1);
  if (t13.error) {
   t1 = t13.error;
   t2.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t14 = t13.payload;
  t8.tag = UINT8_C(8);
  t8.payload.SetVersion = t14;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_8;

 zig_block_8:;
 t5 = t4 == UINT8_C(9);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2325__4241(a0, a1);
  return (nav__4191_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(9)},0};
 }
 goto zig_block_9;

 zig_block_9:;
 t5 = t4 == UINT8_C(10);
 if (t5) {
  t15 = bincode_deserializeAlloc__anon_2568__4250(a0, a1);
  if (t15.error) {
   t1 = t15.error;
   t2.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t4 = t15.payload;
  t8.tag = UINT8_C(10);
  t8.payload.IntegrationModeResponse = t4;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_10;

 zig_block_10:;
 zig_unreachable();
}

static nav__4190_39 bincode_deserializeAlloc__anon_2000__4190(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 const a0, struct mem_Allocator__599 const a1) {
 nav__4190_39 t0;
 struct shared_Command__1178 t2;
 uint16_t t1;
 t0 = bincode_deserializeUnionAlloc__anon_2110__4191(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__1291_39 fifo_posix_Pipe_Reader_recvCmd__1291(struct fifo_posix_Pipe_Reader__788 *const a0) {
 struct fifo_posix_Pipe_Reader__788 *const *t1;
 uintptr_t t3;
 uintptr_t t10;
 struct fifo_posix_Pipe_Reader__788 *t4;
 struct fifo_posix_Pipe_Reader__788 *t0;
 struct fs_File__794 *t5;
 nav__1291_68 t7;
 nav__1291_76 t8;
 nav__1291_39 t12;
 nav__1291_39 t13;
 uint64_t t14;
 uint64_t t20;
 uint8_t const (*t15)[4];
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t18;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t19;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t22;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t21;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t23;
 struct mem_Allocator__599 *t24;
 struct mem_Allocator__599 t25;
 struct shared_Command__1178 t26;
 struct fs_File__794 t6;
 uint32_t t17;
 uint16_t t11;
 bool t9;
 uint8_t t16[4];
 uint8_t t2[4];
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__788 *const *)&t0;
 t4 = (*t1);
 t5 = (struct fs_File__794 *)&t4->file;
 t6 = (*t5);
 t7.ptr = &t2[(uintptr_t)0ul];
 t7.len = (uintptr_t)4ul;
 t8 = fs_File_readAll__1429(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_0;
 }
 t11 = t8.error;
 switch (t11) {
  case zig_error_WouldBlock:
  case zig_error_BrokenPipe: {
   t12 = (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_NotReady};
   goto zig_block_1;
  }
  default: {
   t13.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t13.error = t11;
   t12 = t13;
   goto zig_block_1;
  }
 }

 zig_block_1:;
 return t12;

 zig_block_0:;
 t14 = t3;
 t9 = t14 < UINT64_C(4);
 if (t9) {
  return (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_UnexpectedEof};
 }
 goto zig_block_2;

 zig_block_2:;
 t15 = (uint8_t const (*)[4])&t2;
 memcpy(t16, (const char *)t15, sizeof(uint8_t[4]));
 memcpy(&t17, &t16, sizeof(uint32_t));
 t17 = zig_wrap_u32(t17, UINT8_C(32));
 t4 = (*t1);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t4->buffer;
 t3 = (uintptr_t)t17;
 t11 = array_list_ArrayListAligned_28u8_2cnull_29_resize__1559(t18, t3);
 if (t11) {
  t13.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t13.error = t11;
  return t13;
 }
 t4 = (*t1);
 t5 = (struct fs_File__794 *)&t4->file;
 t6 = (*t5);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&a0->buffer;
 t19 = (*t18);
 t7 = t19.items;
 t8 = fs_File_readAll__1429(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_3;
 }
 t11 = t8.error;
 switch (t11) {
  case zig_error_WouldBlock:
  case zig_error_BrokenPipe: {
   t13 = (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_NotReady};
   goto zig_block_4;
  }
  default: {
   t12.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t12.error = t11;
   t13 = t12;
   goto zig_block_4;
  }
 }

 zig_block_4:;
 return t13;

 zig_block_3:;
 t14 = t3;
 t20 = (uint64_t)t17;
 t9 = t14 < t20;
 if (t9) {
  return (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_UnexpectedEof};
 }
 goto zig_block_5;

 zig_block_5:;
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&a0->buffer;
 t19 = (*t18);
 t7 = t19.items;
 t22 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(t7);
 t21 = t22;
 t23 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4145(&t21);
 t24 = (struct mem_Allocator__599 *)&a0->allocator;
 t25 = (*t24);
 t13 = bincode_deserializeAlloc__anon_2000__4190(t23, t25);
 if (t13.error) {
  t11 = t13.error;
  t13.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t13.error = t11;
  return t13;
 }
 t26 = t13.payload;
 t13.payload = t26;
 t13.error = UINT16_C(0);
 return t13;
}

static nav__1292_39 fifo_posix_Pipe_Reader_waitForResponse__1292(struct fifo_posix_Pipe_Reader__788 *const a0, nav__1292_42 const a1) {
 struct fifo_posix_Pipe_Reader__788 *const *t1;
 uint64_t t2;
 uint64_t t4;
 struct fs_File__794 *t5;
 nav__1292_39 t9;
 nav__1292_39 t11;
 struct fifo_posix_Pipe_Reader__788 *t10;
 struct fifo_posix_Pipe_Reader__788 *t0;
 struct fs_File__794 t6;
 int32_t t7;
 uint16_t t8;
 bool t3;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__788 *const *)&t0;
 t3 = a1.is_null != true;
 if (t3) {
  t4 = a1.payload;
  t2 = t4;
  goto zig_block_0;
 }
 t2 = UINT64_C(1000000000);
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__794 *)&a0->file;
 t6 = (*t5);
 t7 = t6.handle;
 t8 = fifo_posix_waitReadable__1268(t7, t2);
 if (t8) {
  t9.payload = (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t9.error = t8;
  return t9;
 }
 t10 = (*t1);
 t9 = fifo_posix_Pipe_Reader_recvCmd__1291(t10);
 memcpy(&t11, &t9, sizeof(nav__1292_39));
 return t11;
}

static uint16_t fifo_posix_Pipe_Reader_waitForAck__1293(struct fifo_posix_Pipe_Reader__788 *const a0, nav__1293_40 const a1) {
 struct fifo_posix_Pipe_Reader__788 *const *t1;
 struct fifo_posix_Pipe_Reader__788 *t2;
 struct fifo_posix_Pipe_Reader__788 *t0;
 nav__1293_60 t3;
 struct shared_Command__1178 t5;
 struct shared_Command__1178 t6;
 struct shared_Command__1178 const *t7;
 struct mem_Allocator__599 *t9;
 struct mem_Allocator__599 t10;
 nav__1293_77 t11;
 uint16_t t4;
 uint8_t t8;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__788 *const *)&t0;
 t2 = (*t1);
 t3 = fifo_posix_Pipe_Reader_waitForResponse__1292(t2, a1);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 t6 = t5;
 t7 = (struct shared_Command__1178 const *)&t6;
 t8 = t5.tag;
 switch (t8) {
  case UINT8_C(3): {
   t5 = (*t7);
   t9 = (struct mem_Allocator__599 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__3829(t5, t10);
   return 0;
  }
  case UINT8_C(6): {
   t5 = (*t7);
   t9 = (struct mem_Allocator__599 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__3829(t5, t10);
   return zig_error_UnexpectedError;
  }
  default: {
   t11.f0 = t5;
   logger_debug__anon_2632__4261(t11);
   t5 = (*t7);
   t9 = (struct mem_Allocator__599 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__3829(t5, t10);
   return zig_error_UnexpectedResponse;
  }
 }
}

static uint16_t runner_fifo_RunnerFifo_send_version__1139(struct runner_fifo_RunnerFifo__631 *const a0, uint64_t const a1) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 struct shared_Command__1178 t4;
 struct fifo_posix_Pipe_Reader__788 *t6;
 uint16_t t5;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4.tag = UINT8_C(8);
 t4.payload.SetVersion = a1;
 t5 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t4);
 if (t5) {
  return t5;
 }
 t2 = (*t1);
 t6 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t5 = fifo_posix_Pipe_Reader_waitForAck__1293(t6, (nav__1139_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t runner_fifo_RunnerFifo_validate_protocol_version__1130(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 nav__1130_43 t5;
 nav__1130_45 t6;
 uint16_t t3;
 bool t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = runner_fifo_RunnerFifo_send_version__1139(t2, UINT64_C(2));
 t4 = t3 == UINT16_C(0);
 if (t4) {
  goto zig_block_0;
 }
 switch (t3) {
  case zig_error_AckTimeout: {
   return 0;
  }
  case zig_error_UnexpectedError: {
   logger_err__anon_2656__4262();
   logger_err__anon_2665__4263();
   posix_exit__1747(UINT8_C(1));
   zig_unreachable();
  }
  default: {
   t5 = zig_errorName[t3 - 1];
   t6.f0 = t5;
   logger_err__anon_2679__4264(t6);
   posix_exit__1747(UINT8_C(1));
   zig_unreachable();
  }
 }

 zig_block_0:;
 return 0;
}

static nav__1140_38 runner_fifo_RunnerFifo_get_integration_mode__1140(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 struct fifo_posix_Pipe_Reader__788 *t6;
 nav__1140_81 t7;
 struct shared_Command__1178 t8;
 struct shared_Command__1178 t9;
 struct shared_Command__1178 const *t10;
 struct mem_Allocator__599 *t13;
 struct mem_Allocator__599 t14;
 uint16_t t4;
 nav__1140_38 t5;
 uint8_t t11;
 bool t12;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(9)});
 if (t4) {
  t5.payload = UINT8_C(0x2);
  t5.error = t4;
  return t5;
 }
 t2 = (*t1);
 t6 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t7 = fifo_posix_Pipe_Reader_waitForResponse__1292(t6, (nav__1140_79){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t7.error) {
  t4 = t7.error;
  t5.payload = UINT8_C(0x2);
  t5.error = t4;
  return t5;
 }
 t8 = t7.payload;
 t9 = t8;
 t10 = (struct shared_Command__1178 const *)&t9;
 t11 = t8.tag;
 t12 = t11 == UINT8_C(10);
 if (t12) {
  t11 = t8.payload.IntegrationModeResponse;
  t8 = (*t10);
  t13 = (struct mem_Allocator__599 *)&a0->allocator;
  t14 = (*t13);
  shared_Command_deinit__3829(t8, t14);
  t5.payload = t11;
  t5.error = UINT16_C(0);
  return t5;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (*t10);
 t13 = (struct mem_Allocator__599 *)&a0->allocator;
 t14 = (*t13);
 shared_Command_deinit__3829(t8, t14);
 return (nav__1140_38){zig_error_UnexpectedResponse,UINT8_C(0x2)};
}

static nav__820_39 instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__820(struct mem_Allocator__599 const a0) {
 nav__820_62 t1;
 nav__820_39 t3;
 struct runner_fifo_RunnerFifo__631 t4;
 struct runner_fifo_RunnerFifo__631 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 t9;
 uint16_t t2;
 nav__820_65 t6;
 uint8_t t5;
 uint8_t t8;
 bool t7;
 t1 = runner_fifo_RunnerFifo_init__1129(a0);
 if (t1.error) {
  t2 = t1.error;
  t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
  t3.error = t2;
  return t3;
 }
 t4 = t1.payload;
 t0 = t4;
 (void)runner_fifo_RunnerFifo_validate_protocol_version__1130(&t0);
 t6 = runner_fifo_RunnerFifo_get_integration_mode__1140(&t0);
 t7 = t6.error == UINT16_C(0);
 if (t7) {
  t8 = t6.payload;
  t5 = t8;
  goto zig_block_0;
 }
 t2 = t6.error;
 runner_fifo_RunnerFifo_deinit__1131(&t0);
 t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
 t3.error = t2;
 return t3;

 zig_block_0:;
 t7 = t5 != UINT8_C(2);
 if (t7) {
  runner_fifo_RunnerFifo_deinit__1131(&t0);
  return (nav__820_39){{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}},zig_error_ModeError};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t0;
 t9.fifo = t4;
 t3.payload = t9;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__807_39 instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_init__807(struct mem_Allocator__599 const a0) {
 nav__807_62 t1;
 nav__807_39 t3;
 struct runner_fifo_RunnerFifo__631 t4;
 struct runner_fifo_RunnerFifo__631 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 t9;
 uint16_t t2;
 nav__807_65 t6;
 uint8_t t5;
 uint8_t t8;
 bool t7;
 t1 = runner_fifo_RunnerFifo_init__1129(a0);
 if (t1.error) {
  t2 = t1.error;
  t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
  t3.error = t2;
  return t3;
 }
 t4 = t1.payload;
 t0 = t4;
 (void)runner_fifo_RunnerFifo_validate_protocol_version__1130(&t0);
 t6 = runner_fifo_RunnerFifo_get_integration_mode__1140(&t0);
 t7 = t6.error == UINT16_C(0);
 if (t7) {
  t8 = t6.payload;
  t5 = t8;
  goto zig_block_0;
 }
 t2 = t6.error;
 runner_fifo_RunnerFifo_deinit__1131(&t0);
 t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
 t3.error = t2;
 return t3;

 zig_block_0:;
 t7 = t5 != UINT8_C(0);
 if (t7) {
  runner_fifo_RunnerFifo_deinit__1131(&t0);
  return (nav__807_39){{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}},zig_error_ModeError};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t0;
 t9.fifo = t4;
 t3.payload = t9;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__743_39 instruments_root_Instrument_init__743(struct mem_Allocator__599 const a0) {
 nav__743_69 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 t2;
 struct instruments_root_Instrument__554 t3;
 nav__743_39 t4;
 nav__743_71 t5;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 t6;
 bool t1;
 t0 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__820(a0);
 t1 = t0.error == UINT16_C(0);
 if (t1) {
  t2 = t0.payload;
  t3.tag = UINT8_C(2);
  t3.payload.analysis = t2;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_init__807(a0);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t6 = t5.payload;
  t3.tag = UINT8_C(1);
  t3.payload.walltime = t6;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_1;

 zig_block_1:;
 return (nav__743_39){{{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}}},UINT8_C(3)},0};
}

static nav__730_39 instrument_hooks_InstrumentHooks_init__730(struct mem_Allocator__599 const a0) {
 struct InstrumentHooks *t1;
 struct instruments_root_Instrument__554 *t2;
 nav__730_100 t3;
 struct instruments_root_Instrument__554 t4;
 struct environment_root_Environment__559 *t5;
 struct environment_root_Environment__559 t6;
 nav__730_39 t0;
 t0.error = UINT16_C(0);
 t1 = &t0.payload;
 t2 = (struct instruments_root_Instrument__554 *)&t1->instrument;
 t3 = instruments_root_Instrument_init__743(a0);
 t4 = t3.payload;
 (*t2) = t4;
 t5 = (struct environment_root_Environment__559 *)&t1->environment;
 t6 = environment_root_Environment_init__773(a0);
 (*t5) = t6;
 return t0;
}

struct InstrumentHooks *c_instrument_hooks_init__238(void) {
 struct InstrumentHooks *t0;
 struct InstrumentHooks *t3;
 nav__238_46 t1;
 nav__238_107 t4;
 struct InstrumentHooks t5;
 bool t2;
 t1 = mem_Allocator_create__anon_1039__3706((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)});
 t2 = t1.error == UINT16_C(0);
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 return NULL;

 zig_block_0:;
 t4 = instrument_hooks_InstrumentHooks_init__730((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)});
 t5 = t4.payload;
 (*t0) = t5;
 t3 = (struct InstrumentHooks *)t0;
 return t3;
}

static nav__3706_40 mem_Allocator_create__anon_1039__3706(struct mem_Allocator__599 const a0) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 uintptr_t t3;
 nav__3706_51 t4;
 nav__3706_40 t6;
 uint8_t *t7;
 struct InstrumentHooks *t8;
 uint16_t t5;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t4 = mem_Allocator_allocBytesWithAlignment__anon_2726__4265(t2, (uintptr_t)288ul, t3);
 if (t4.error) {
  t5 = t4.error;
  t6.payload = ((struct InstrumentHooks *)(uintptr_t)0xaaaaaaaaaaaaaaaaul);
  t6.error = t5;
  return t6;
 }
 t7 = t4.payload;
 t8 = (struct InstrumentHooks *)t7;
 t6.payload = t8;
 t6.error = UINT16_C(0);
 return t6;
}

static uint16_t fs_accessAbsolute__1348(nav__1348_39 const a0, struct fs_File_OpenFlags__1101 const a1) {
 struct fs_Dir__1103 const *t3;
 struct fs_Dir__1103 t1;
 struct fs_Dir__1103 t2;
 uint16_t t4;
 bool t0;
 t0 = fs_path_isAbsolute__4300(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1340();
 t2 = t1;
 t3 = (struct fs_Dir__1103 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_access__3785(t1, a0, a1);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__1345_39 fs_openFileAbsolute__1345(nav__1345_41 const a0, struct fs_File_OpenFlags__1101 const a1) {
 struct fs_Dir__1103 const *t3;
 struct fs_Dir__1103 t1;
 struct fs_Dir__1103 t2;
 nav__1345_39 t4;
 bool t0;
 t0 = fs_path_isAbsolute__4300(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1340();
 t2 = t1;
 t3 = (struct fs_Dir__1103 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_openFile__3721(t1, a0, a1);
 return t4;
}

static struct fifo_posix_Pipe_Writer__786 fifo_posix_Pipe_Writer_init__1283(struct fs_File__794 const a0, struct mem_Allocator__599 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t0;
 struct fs_File__794 *t5;
 struct mem_Allocator__599 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t7;
 struct fifo_posix_Pipe_Writer__786 t4;
 uint16_t t2;
 bool t3;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(a1);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(&t0, (uintptr_t)1024ul);
 t3 = t2 == UINT16_C(0);
 if (t3) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__794 *)&t4.file;
 (*t5) = a0;
 t6 = (struct mem_Allocator__599 *)&t4.allocator;
 (*t6) = a1;
 t7 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t4.buffer;
 t1 = t0;
 (*t7) = t1;
 return t4;
}

static struct fifo_posix_Pipe_Reader__788 fifo_posix_Pipe_Reader_init__1288(struct fs_File__794 const a0, struct mem_Allocator__599 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t0;
 struct fs_File__794 *t5;
 struct mem_Allocator__599 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t7;
 struct fifo_posix_Pipe_Reader__788 t4;
 uint16_t t2;
 bool t3;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(a1);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(&t0, (uintptr_t)1024ul);
 t3 = t2 == UINT16_C(0);
 if (t3) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__794 *)&t4.file;
 (*t5) = a0;
 t6 = (struct mem_Allocator__599 *)&t4.allocator;
 (*t6) = a1;
 t7 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t4.buffer;
 t1 = t0;
 (*t7) = t1;
 return t4;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1562(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 nav__1562_43 *t3;
 uintptr_t *t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__1562_43 *)&t2->items;
 t4 = &t3->len;
 (*t4) = (uintptr_t)0ul;
 return;
}

static struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 **t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static uint16_t bincode_serialize__anon_1232__3956(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__1178 const a1) {
 uint16_t t0;
 t0 = bincode_serializeUnion__anon_2768__4363(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint8_t const (*mem_asBytes__anon_1270__3957(uint32_t const *const a0))[4] {
 uint8_t const (*t0)[4];
 t0 = (uint8_t const (*)[4])a0;
 return t0;
}

static uint16_t fs_File_writeAll__1439(struct fs_File__794 const a0, nav__1439_40 const a1) {
 struct fs_File__794 const *t1;
 nav__1439_40 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 nav__1439_40 t11;
 nav__1439_40 t2;
 uint8_t const *t12;
 nav__1439_47 t14;
 struct fs_File__794 t10;
 struct fs_File__794 t0;
 uint16_t t15;
 bool t9;
 t0 = a0;
 t1 = (struct fs_File__794 const *)&t0;
 t2 = a1;
 t3 = (nav__1439_40 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 < t8;
 if (t9) {
  t6 = t4;
  t10 = (*t1);
  t5 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t const *)(((uintptr_t)t12) + (t5*sizeof(uint8_t)));
  t13 = t11.len;
  t5 = t13 - t5;
  t11.ptr = t12;
  t11.len = t5;
  t14 = fs_File_write__1438(t10, t11);
  if (t14.error) {
   t15 = t14.error;
   return t15;
  }
  t5 = t14.payload;
  t5 = t6 + t5;
  t4 = t5;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 return 0;
}

static nav__1964_38 posix_poll__1964(nav__1964_41 const a0, int32_t const a1) {
 uintptr_t t1;
 struct c_pollfd__struct_1304__1304 *t5;
 nav__1964_38 t8;
 uint32_t t0;
 uint32_t t4;
 nav__1964_46 t2;
 int t6;
 uint16_t t7;
 bool t3;
 zig_loop_5:
 t1 = a0.len;
 t2 = math_cast__anon_2787__4364(t1);
 t3 = t2.is_null != true;
 if (t3) {
  t4 = t2.payload;
  t0 = t4;
  goto zig_block_1;
 }
 return (nav__1964_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};

 zig_block_1:;
 t5 = a0.ptr;
 t6 = a1;
 t6 = poll(t5, t0, t6);
 t7 = posix_errno__anon_2970__4656(t6);
 switch (t7) {
  case UINT16_C(0): {
   t1 = (uintptr_t)t6;
   t8.payload = t1;
   t8.error = UINT16_C(0);
   return t8;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(12): {
   return (nav__1964_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  default: {
   t7 = posix_unexpectedErrno__2021(t7);
   t8.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t8.error = t7;
   return t8;
  }
 }

 zig_block_0:;
 goto zig_loop_5;
}

static nav__1429_38 fs_File_readAll__1429(struct fs_File__794 const a0, nav__1429_41 const a1) {
 struct fs_File__794 const *t1;
 nav__1429_41 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 nav__1429_41 t11;
 nav__1429_41 t2;
 uint8_t *t12;
 nav__1429_38 t13;
 struct fs_File__794 t10;
 struct fs_File__794 t0;
 uint16_t t14;
 bool t9;
 t0 = a0;
 t1 = (struct fs_File__794 const *)&t0;
 t2 = a1;
 t3 = (nav__1429_41 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 != t8;
 if (t9) {
  t10 = (*t1);
  t6 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t *)(((uintptr_t)t12) + (t6*sizeof(uint8_t)));
  t5 = t11.len;
  t6 = t5 - t6;
  t11.ptr = t12;
  t11.len = t6;
  t13 = fs_File_read__1428(t10, t11);
  if (t13.error) {
   t14 = t13.error;
   t13.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t13.error = t14;
   return t13;
  }
  t6 = t13.payload;
  t8 = t6;
  t9 = t8 == UINT64_C(0);
  if (t9) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t5 = t4;
  t6 = t5 + t6;
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 t6 = t4;
 t13.payload = t6;
 t13.error = UINT16_C(0);
 return t13;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_resize__1559(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 nav__1559_43 *t4;
 uintptr_t *t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(t2, a1);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (nav__1559_43 *)&t2->items;
 t5 = &t4->len;
 (*t5) = a1;
 return 0;
}

static struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(nav__4156_40 const a0) {
 nav__4156_40 *t1;
 uintptr_t *t2;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 t1 = (nav__4156_40 *)&t0.buffer;
 (*t1) = a0;
 t2 = (uintptr_t *)&t0.pos;
 (*t2) = (uintptr_t)0ul;
 return t0;
}

static struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4145(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 **t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__1992 t0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__4189_38 io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__4189(void const *const a0, nav__4189_41 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t0;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *t1;
 nav__4189_38 t2;
 nav__4189_38 t3;
 t0 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)a0;
 t1 = (*t0);
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4148(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__4189_38));
 return t3;
}

static nav__4231_39 io_Reader_readBytesNoEof__anon_2172__4231(struct io_Reader__2141 const a0) {
 struct io_Reader__2141 const *t1;
 struct io_Reader__2141 t3;
 struct io_Reader__2141 t0;
 nav__4231_46 t4;
 uint16_t t5;
 nav__4231_39 t6;
 uint8_t t7[4];
 uint8_t t2[4];
 t0 = a0;
 t1 = (struct io_Reader__2141 const *)&t0;
 t3 = (*t1);
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)4ul;
 t5 = io_Reader_readNoEof__4198(t3, t4);
 if (t5) {
  memcpy(t6.payload, "\252\252\252\252", sizeof(uint8_t[4]));
  t6.error = t5;
  return t6;
 }
 memcpy(t7, (const char *)&t2, sizeof(uint8_t[4]));
 memcpy(t6.payload, t7, sizeof(uint8_t[4]));
 t6.error = UINT16_C(0);
 return t6;
}

static nav__4238_39 io_Reader_readBytesNoEof__anon_2278__4238(struct io_Reader__2141 const a0) {
 struct io_Reader__2141 const *t1;
 struct io_Reader__2141 t3;
 struct io_Reader__2141 t0;
 nav__4238_46 t4;
 uint16_t t5;
 nav__4238_39 t6;
 uint8_t t7[8];
 uint8_t t2[8];
 t0 = a0;
 t1 = (struct io_Reader__2141 const *)&t0;
 t3 = (*t1);
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)8ul;
 t5 = io_Reader_readNoEof__4198(t3, t4);
 if (t5) {
  memcpy(t6.payload, "\252\252\252\252\252\252\252\252", sizeof(uint8_t[8]));
  t6.error = t5;
  return t6;
 }
 memcpy(t7, (const char *)&t2, sizeof(uint8_t[8]));
 memcpy(t6.payload, t7, sizeof(uint8_t[8]));
 t6.error = UINT16_C(0);
 return t6;
}

static nav__4239_40 mem_Allocator_alloc__anon_2287__4239(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__4239_40 t4;
 nav__4239_40 t8;
 nav__4239_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__4239_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_3006__4657(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__4239_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static nav__4196_38 io_Reader_readAll__4196(struct io_Reader__2141 const a0, nav__4196_41 const a1) {
 uintptr_t t0;
 nav__4196_38 t1;
 t0 = a1.len;
 t1 = io_Reader_readAtLeast__4197(a0, a1, t0);
 return t1;
}

static zig_noreturn void bincode_invalidProtocol__anon_2310__4240(void) {
 utils_print__anon_3028__4668();
 posix_exit__1747(UINT8_C(1));
 zig_unreachable();
}

static void shared_Command_deinit__3829(struct shared_Command__1178 const a0, struct mem_Allocator__599 const a1) {
 struct mem_Allocator__599 const *t1;
 struct shared_Command__struct_1182__1182 t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t0;
 nav__3829_43 t5;
 struct shared_Command__struct_1181__1181 t6;
 uint8_t t2;
 t0 = a1;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(5): {
   t3 = a0.payload.SetIntegration;
   t4 = (*t1);
   t5 = t3.name;
   mem_Allocator_free__anon_3042__4747(t4, t5);
   t4 = (*t1);
   t5 = t3.version;
   mem_Allocator_free__anon_3042__4747(t4, t5);
   goto zig_block_0;
  }
  case UINT8_C(0): {
   t6 = a0.payload.ExecutedBenchmark;
   t4 = (*t1);
   t5 = t6.uri;
   mem_Allocator_free__anon_3042__4747(t4, t5);
   goto zig_block_0;
  }
  case UINT8_C(8): {
   goto zig_block_0;
  }
  case UINT8_C(9): {
   goto zig_block_0;
  }
  case UINT8_C(10): {
   goto zig_block_0;
  }
  default: {
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return;
}

static void logger_debug__anon_2632__4261(nav__4261_39 const a0) {
 logger_logWithPrefix__anon_3050__4748(a0);
 return;
}

static void logger_err__anon_2656__4262(void) {
 logger_logWithPrefix__anon_3052__4749();
 return;
}

static void logger_err__anon_2665__4263(void) {
 logger_logWithPrefix__anon_3053__4750();
 return;
}

static zig_noreturn void posix_exit__1747(uint8_t const a0) {
 int t0;
 t0 = (int)a0;
 exit(t0);
 zig_unreachable();
}

static void logger_err__anon_2679__4264(nav__4264_40 const a0) {
 logger_logWithPrefix__anon_3058__4752(a0);
 return;
}

static void runner_fifo_RunnerFifo_deinit__1131(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 struct fifo_posix_Pipe_Reader__788 *t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 fifo_posix_Pipe_Writer_deinit__1287(t3);
 t2 = (*t1);
 t4 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 fifo_posix_Pipe_Reader_deinit__1294(t4);
 return;
}

static struct environment_root_Environment__559 environment_root_Environment_init__773(struct mem_Allocator__599 const a0) {
 struct mem_Allocator__599 *t1;
 struct environment_linked_libraries_root_LinkedLibraries__837 *t2;
 struct environment_linked_libraries_root_LinkedLibraries__837 t3;
 struct environment_root_EnvironmentJson__833 *t4;
 struct environment_root_Environment__559 t0;
 t1 = (struct mem_Allocator__599 *)&t0.allocator;
 (*t1) = a0;
 t2 = (struct environment_linked_libraries_root_LinkedLibraries__837 *)&t0.libs;
 t3 = environment_linked_libraries_root_LinkedLibraries_init__2808(a0);
 (*t2) = t3;
 t4 = (struct environment_root_EnvironmentJson__833 *)&t0.data;
 (*t4) = (struct environment_root_EnvironmentJson__833){{{{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul,(uintptr_t)0ul},NULL}},{{{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul,(uintptr_t)0ul},NULL}}};
 return t0;
}

static nav__4265_39 mem_Allocator_allocBytesWithAlignment__anon_2726__4265(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t7;
 struct mem_Allocator_VTable__602 const *const *t8;
 struct mem_Allocator_VTable__602 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__4265_52 t17;
 nav__4265_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__4265_39){(uint8_t *)(uintptr_t)0xfffffffffffffff8ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t7 = t5;
 t1 = (struct mem_Allocator__599 const *)&t7;
 t8 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__4265_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t14 = (uint8_t *)t4;
 t18.payload = t14;
 t18.error = UINT16_C(0);
 return t18;
}

static bool fs_path_isAbsolute__4300(nav__4300_39 const a0) {
 bool t0;
 t0 = fs_path_isAbsolutePosix__4306(a0);
 return t0;
}

static struct fs_Dir__1103 fs_cwd__1340(void) {
 struct fs_Dir__1103 t0;
 t0 = (struct fs_Dir__1103){-INT32_C(2)};
 return t0;
}

static uint16_t fs_Dir_access__3785(struct fs_Dir__1103 const a0, nav__3785_40 const a1, struct fs_File_OpenFlags__1101 const a2) {
 struct fs_Dir__1103 const *t1;
 uint8_t const (*t6)[1024];
 uint8_t const *t8;
 struct fs_Dir__1103 t7;
 struct fs_Dir__1103 t0;
 nav__3785_49 t2;
 uint16_t t3;
 uint8_t t4[1024];
 uint8_t t5[1024];
 t0 = a0;
 t1 = (struct fs_Dir__1103 const *)&t0;
 t2 = posix_toPosixPath__2022(a1);
 if (t2.error) {
  t3 = t2.error;
  return t3;
 }
 memcpy(t4, t2.payload, sizeof(uint8_t[1024]));
 memcpy((char *)&t5, t4, sizeof(uint8_t[1024]));
 t6 = (uint8_t const (*)[1024])&t5;
 t7 = (*t1);
 t8 = (uint8_t const *)t6;
 t3 = fs_Dir_accessZ__3786(t7, t8, a2);
 return t3;
}

static nav__3721_39 fs_Dir_openFile__3721(struct fs_Dir__1103 const a0, nav__3721_42 const a1, struct fs_File_OpenFlags__1101 const a2) {
 struct fs_Dir__1103 const *t1;
 uint8_t const (*t8)[1024];
 uint8_t const *t10;
 nav__3721_39 t5;
 struct fs_Dir__1103 t9;
 struct fs_Dir__1103 t0;
 nav__3721_53 t3;
 uint16_t t4;
 uint8_t t6[1024];
 uint8_t t7[1024];
 struct fs_File_OpenFlags__1101 t2;
 t0 = a0;
 t1 = (struct fs_Dir__1103 const *)&t0;
 t2 = a2;
 t3 = posix_toPosixPath__2022(a1);
 if (t3.error) {
  t4 = t3.error;
  t5.payload = (struct fs_File__794){-INT32_C(0x55555556)};
  t5.error = t4;
  return t5;
 }
 memcpy(t6, t3.payload, sizeof(uint8_t[1024]));
 memcpy((char *)&t7, t6, sizeof(uint8_t[1024]));
 t8 = (uint8_t const (*)[1024])&t7;
 t9 = (*t1);
 t10 = (uint8_t const *)t8;
 t5 = fs_Dir_openFileZ__3722(t9, t10, a2);
 return t5;
}

static struct array_list_ArrayListAligned_28u8_2cnull_29__806 array_list_ArrayListAligned_28u8_2cnull_29_init__1527(struct mem_Allocator__599 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t0;
 t0.items = (nav__1527_42){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 t0.capacity = (uintptr_t)0ul;
 t0.allocator = a0;
 return t0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__4816(t3, a1);
 t7 = (*t1);
 t8 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1565(t7, t3);
 return t8;
}

static uint16_t bincode_serializeUnion__anon_2768__4363(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__1178 const a1) {
 struct shared_Command__struct_1181__1181 t4;
 struct shared_Command__struct_1182__1182 t5;
 struct shared_Command__struct_1183__1183 t6;
 uint64_t t7;
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 bool t3;
 t0 = a1.tag;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3165__4817(a0, t1);
 if (t2) {
  return t2;
 }
 t0 = a1.tag;
 t3 = t0 == UINT8_C(0);
 if (t3) {
  t4 = a1.payload.ExecutedBenchmark;
  t2 = bincode_serialize__anon_3167__4818(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(1);
 if (t3) {
  t2 = bincode_serialize__anon_3169__4819(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(2);
 if (t3) {
  t2 = bincode_serialize__anon_3169__4819(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(3);
 if (t3) {
  t2 = bincode_serialize__anon_3169__4819(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(4);
 if (t3) {
  t2 = bincode_serialize__anon_3169__4819(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_4;
 }
 goto zig_block_4;

 zig_block_4:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(5);
 if (t3) {
  t5 = a1.payload.SetIntegration;
  t2 = bincode_serialize__anon_3171__4820(a0, t5);
  if (t2) {
   return t2;
  }
  goto zig_block_5;
 }
 goto zig_block_5;

 zig_block_5:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(6);
 if (t3) {
  t2 = bincode_serialize__anon_3169__4819(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_6;
 }
 goto zig_block_6;

 zig_block_6:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(7);
 if (t3) {
  t6 = a1.payload.AddMarker;
  t2 = bincode_serialize__anon_3173__4821(a0, t6);
  if (t2) {
   return t2;
  }
  goto zig_block_7;
 }
 goto zig_block_7;

 zig_block_7:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(8);
 if (t3) {
  t7 = a1.payload.SetVersion;
  t2 = bincode_serialize__anon_3175__4822(a0, t7);
  if (t2) {
   return t2;
  }
  goto zig_block_8;
 }
 goto zig_block_8;

 zig_block_8:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(9);
 if (t3) {
  t2 = bincode_serialize__anon_3169__4819(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_9;
 }
 goto zig_block_9;

 zig_block_9:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(10);
 if (t3) {
  t0 = a1.payload.IntegrationModeResponse;
  t2 = bincode_serialize__anon_3177__4823(a0, t0);
  if (t2) {
   return t2;
  }
  goto zig_block_10;
 }
 goto zig_block_10;

 zig_block_10:;
 return 0;
}

static nav__1438_38 fs_File_write__1438(struct fs_File__794 const a0, nav__1438_41 const a1) {
 nav__1438_38 t1;
 int32_t t0;
 t0 = a0.handle;
 t1 = posix_write__1757(t0, a1);
 return t1;
}

static nav__4364_38 math_cast__anon_2787__4364(uintptr_t const a0) {
 uint64_t t0;
 uint32_t t2;
 nav__4364_38 t3;
 bool t1;
 t0 = a0;
 t1 = t0 > UINT64_C(4294967295);
 if (t1) {
  return (nav__4364_38){UINT32_C(0xaaaaaaaa),true};
 }
 t2 = (uint32_t)a0;
 t3.is_null = false;
 t3.payload = t2;
 return t3;
}

static uint16_t posix_errno__anon_2970__4656(int const a0) {
 int *t3;
 int32_t t1;
 int t4;
 uint16_t t0;
 uint16_t t5;
 bool t2;
 t1 = a0;
 t2 = t1 == -INT32_C(1);
 if (t2) {
  t3 = zig_e___error();
  t4 = (*t3);
  t5 = (uint16_t)t4;
  t0 = t5;
  goto zig_block_0;
 }
 t0 = UINT16_C(0);
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static uint16_t posix_unexpectedErrno__2021(uint16_t const a0) {
 (void)a0;
 return zig_error_Unexpected;
}

static nav__1428_38 fs_File_read__1428(struct fs_File__794 const a0, nav__1428_41 const a1) {
 nav__1428_38 t1;
 int32_t t0;
 t0 = a0.handle;
 t1 = posix_read__1749(t0, a1);
 return t1;
}

static nav__4148_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4148(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const a0, nav__4148_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t1;
 nav__4148_42 const *t3;
 uintptr_t t4;
 uintptr_t t7;
 uintptr_t t9;
 nav__4148_42 *t5;
 nav__4148_42 t6;
 nav__4148_42 t13;
 nav__4148_42 t2;
 uintptr_t *t8;
 uint64_t t10;
 uint8_t *t11;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *t12;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *t0;
 nav__4148_38 t14;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t0;
 t2 = a1;
 t3 = (nav__4148_42 const *)&t2;
 t4 = a1.len;
 t5 = (nav__4148_42 *)&a0->buffer;
 t6 = (*t5);
 t7 = t6.len;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t9 = t7 - t9;
 t9 = (t4 < t9) ? t4 : t9;
 t10 = t9;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t4 = t10;
 t4 = t9 + t4;
 t6 = (*t3);
 t11 = t6.ptr;
 t11 = (uint8_t *)(((uintptr_t)t11) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t9 = t10;
 t6.ptr = t11;
 t6.len = t9;
 t12 = (*t1);
 t5 = (nav__4148_42 *)&t12->buffer;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t13 = (*t5);
 t11 = t13.ptr;
 t11 = (uint8_t *)(((uintptr_t)t11) + (t9*sizeof(uint8_t)));
 t9 = t4 - t9;
 t13.ptr = t11;
 t13.len = t9;
 t11 = t13.ptr;
 if (t6.len != 0) memcpy(t6.ptr, t11, t6.len * sizeof(uint8_t));
 t12 = (*t1);
 t8 = (uintptr_t *)&t12->pos;
 (*t8) = t4;
 t4 = t10;
 t14.payload = t4;
 t14.error = UINT16_C(0);
 return t14;
}

static uint16_t io_Reader_readNoEof__4198(struct io_Reader__2141 const a0, nav__4198_40 const a1) {
 struct io_Reader__2141 const *t1;
 struct io_Reader__2141 t2;
 struct io_Reader__2141 t0;
 nav__4198_43 t3;
 uintptr_t t5;
 uintptr_t t6;
 uint64_t t7;
 uint64_t t8;
 uint16_t t4;
 bool t9;
 t0 = a0;
 t1 = (struct io_Reader__2141 const *)&t0;
 t2 = (*t1);
 t3 = io_Reader_readAll__4196(t2, a1);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 < t8;
 if (t9) {
  return zig_error_EndOfStream;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static nav__4657_39 mem_Allocator_allocWithSizeAndAlignment__anon_3006__4657(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__4657_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__4657_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3220__4889((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__4657_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_3222__4890(t6, t2, a2);
 return t7;
}

static nav__4197_38 io_Reader_readAtLeast__4197(struct io_Reader__2141 const a0, nav__4197_41 const a1, uintptr_t const a2) {
 struct io_Reader__2141 const *t1;
 nav__4197_41 const *t3;
 uintptr_t t4;
 uintptr_t t12;
 uintptr_t t8;
 uint64_t t5;
 uint64_t t6;
 struct io_Reader__2141 t9;
 struct io_Reader__2141 t0;
 nav__4197_41 t10;
 nav__4197_41 t2;
 uint8_t *t11;
 nav__4197_38 t13;
 uint16_t t14;
 bool t7;
 t0 = a0;
 t1 = (struct io_Reader__2141 const *)&t0;
 t2 = a1;
 t3 = (nav__4197_41 const *)&t2;
 t4 = a1.len;
 t5 = a2;
 t6 = t4;
 t7 = t5 <= t6;
 debug_assert__180(t7);
 t8 = (uintptr_t)0ul;
 zig_loop_17:
 t4 = t8;
 t6 = t4;
 t5 = a2;
 t7 = t6 < t5;
 if (t7) {
  t9 = (*t1);
  t4 = t8;
  t10 = (*t3);
  t11 = t10.ptr;
  t11 = (uint8_t *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
  t12 = t10.len;
  t4 = t12 - t4;
  t10.ptr = t11;
  t10.len = t4;
  t13 = io_Reader_read__4195(t9, t10);
  if (t13.error) {
   t14 = t13.error;
   t13.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t13.error = t14;
   return t13;
  }
  t4 = t13.payload;
  t5 = t4;
  t7 = t5 == UINT64_C(0);
  if (t7) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t12 = t8;
  t4 = t12 + t4;
  t8 = t4;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_17;

 zig_block_0:;
 t12 = t8;
 t13.payload = t12;
 t13.error = UINT16_C(0);
 return t13;
}

static void utils_print__anon_3028__4668(void) {
 (void)printf((char const *)&__anon_3329, 61, (uint8_t const *)&__anon_2303);
 return;
}

static void mem_Allocator_free__anon_3042__4747(struct mem_Allocator__599 const a0, nav__4747_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__4747_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t const *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 nav__4747_51 t10;
 struct mem_Allocator__599 t11;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__4747_51, uint8_t, uintptr_t);
 void (*t17)(void *, nav__4747_51, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_3353__4892(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t11 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t11;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__4747_51, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t11.ptr;
 t17(t18, t10, t12, t3);
 return;
}

static void logger_logWithPrefix__anon_3050__4748(nav__4748_39 const a0) {
 nav__4748_59 t3;
 nav__4748_59 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__4748_63 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(0) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3407__5052(t6, a0);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void logger_logWithPrefix__anon_3052__4749(void) {
 nav__4749_41 t3;
 nav__4749_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__4749_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3429__5054(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void logger_logWithPrefix__anon_3053__4750(void) {
 nav__4750_41 t3;
 nav__4750_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__4750_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3442__5055(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void logger_logWithPrefix__anon_3058__4752(nav__4752_40 const a0) {
 nav__4752_47 t3;
 nav__4752_47 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__4752_51 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3455__5056(t6, a0);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void fifo_posix_Pipe_Writer_deinit__1287(struct fifo_posix_Pipe_Writer__786 *const a0) {
 struct fifo_posix_Pipe_Writer__786 *const *t1;
 struct fifo_posix_Pipe_Writer__786 *t2;
 struct fifo_posix_Pipe_Writer__786 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t3;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t4;
 struct fs_File__794 *t5;
 struct fs_File__794 t6;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Writer__786 *const *)&t0;
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t2->buffer;
 t4 = (*t3);
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t4);
 t2 = (*t1);
 t5 = (struct fs_File__794 *)&t2->file;
 t6 = (*t5);
 fs_File_close__1385(t6);
 return;
}

static void fifo_posix_Pipe_Reader_deinit__1294(struct fifo_posix_Pipe_Reader__788 *const a0) {
 struct fifo_posix_Pipe_Reader__788 *const *t1;
 struct fs_File__794 *t4;
 uintptr_t t9;
 uintptr_t t13;
 nav__1294_61 t10;
 nav__1294_64 t11;
 uint64_t t14;
 struct fifo_posix_Pipe_Reader__788 *t15;
 struct fifo_posix_Pipe_Reader__788 *t0;
 nav__1294_53 t16;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t17;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t18;
 struct fs_File__794 t5;
 int32_t t6;
 struct c_pollfd__struct_1304__1304 t7;
 struct c_pollfd__struct_1304__1304 t8[1];
 struct c_pollfd__struct_1304__1304 t3[1];
 bool t12;
 uint8_t t2[4096];
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__788 *const *)&t0;
 zig_loop_7:
 t4 = (struct fs_File__794 *)&a0->file;
 t5 = (*t4);
 t6 = t5.handle;
 t7.fd = t6;
 t7.events = INT16_C(1);
 t7.revents = INT16_C(0);
 t8[0] = t7;
 memcpy((char *)&t3, t8, sizeof(struct c_pollfd__struct_1304__1304[1]));
 t10.ptr = (struct c_pollfd__struct_1304__1304 *)&t3[(uintptr_t)0ul];
 t10.len = (uintptr_t)1ul;
 t11 = posix_poll__1964(t10, INT32_C(0));
 t12 = t11.error == UINT16_C(0);
 if (t12) {
  t13 = t11.payload;
  t9 = t13;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t14 = t9;
 t12 = t14 == UINT64_C(0);
 if (t12) {
  goto zig_block_0;
 }
 goto zig_block_2;

 zig_block_2:;
 t15 = (*t1);
 t4 = (struct fs_File__794 *)&t15->file;
 t5 = (*t4);
 t16.ptr = &t2[(uintptr_t)0ul];
 t16.len = (uintptr_t)4096ul;
 t11 = fs_File_read__1428(t5, t16);
 t12 = t11.error == UINT16_C(0);
 if (t12) {
  t13 = t11.payload;
  t9 = t13;
  goto zig_block_3;
 }
 goto zig_block_0;

 zig_block_3:;
 t14 = t9;
 t12 = t14 == UINT64_C(0);
 if (t12) {
  goto zig_block_0;
 }
 goto zig_block_4;

 zig_block_4:;
 goto zig_loop_7;

 zig_block_0:;
 t15 = (*t1);
 t17 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *)&t15->buffer;
 t18 = (*t17);
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t18);
 t15 = (*t1);
 t4 = (struct fs_File__794 *)&t15->file;
 t5 = (*t4);
 fs_File_close__1385(t5);
 return;
}

static struct environment_linked_libraries_root_LinkedLibraries__837 environment_linked_libraries_root_LinkedLibraries_init__2808(struct mem_Allocator__599 const a0) {
 struct mem_Allocator__599 *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 *t2;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t3;
 struct environment_linked_libraries_root_LinkedLibraries__837 t0;
 t1 = (struct mem_Allocator__599 *)&t0.allocator;
 (*t1) = a0;
 t2 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 *)&t0.libraries;
 t3 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_init__3512(a0);
 (*t2) = t3;
 return t0;
}

static uint8_t mem_Alignment_fromByteUnits__1113(uintptr_t const a0) {
 bool t0;
 uint8_t t1;
 t0 = math_isPowerOfTwo__anon_3478__5057(a0);
 debug_assert__180(t0);
 t1 = zig_ctz_u64(a0, UINT8_C(64));
 return t1;
}

static bool fs_path_isAbsolutePosix__4306(nav__4306_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 bool t2;
 bool t3;
 uint8_t t4;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 > UINT64_C(0);
 if (t2) {
  t4 = a0.ptr[(uintptr_t)0ul];
  t2 = t4 == UINT8_C(47);
  t3 = t2;
  goto zig_block_0;
 }
 t3 = false;
 goto zig_block_0;

 zig_block_0:;
 return t3;
}

static nav__2022_39 posix_toPosixPath__2022(nav__2022_41 const a0) {
 uintptr_t t1;
 uint64_t t2;
 uint8_t *t4;
 nav__2022_47 t5;
 uint8_t const *t6;
 nav__2022_39 t8;
 bool t3;
 uint8_t t7[1024];
 uint8_t t0[1024];
 t1 = a0.len;
 t2 = t1;
 t3 = t2 >= UINT64_C(1024);
 if (t3) {
  return (nav__2022_39){zig_error_NameTooLong,"\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252"};
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0.len;
 t4 = (uint8_t *)&t0;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t1;
 t6 = a0.ptr;
 if (t5.len != 0) memcpy(t5.ptr, t6, t5.len * sizeof(uint8_t));
 t1 = a0.len;
 t4 = (uint8_t *)&t0[t1];
 (*t4) = UINT8_C(0);
 memcpy(t7, (const char *)&t0, sizeof(uint8_t[1024]));
 memcpy(t8.payload, t7, sizeof(uint8_t[1024]));
 t8.error = UINT16_C(0);
 return t8;
}

static uint16_t fs_Dir_accessZ__3786(struct fs_Dir__1103 const a0, uint8_t const *const a1, struct fs_File_OpenFlags__1101 const a2) {
 uint32_t t2;
 int32_t t3;
 struct fs_Dir__1103 t0;
 uint16_t t4;
 uint8_t t1;
 t0 = a0;
 t1 = a2.mode;
 switch (t1) {
  case UINT8_C(0): {
   t2 = UINT32_C(0);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t2 = UINT32_C(2);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = UINT32_C(6);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t3 = a0.fd;
 t4 = posix_faccessatZ__1913(t3, a1, t2, UINT32_C(0));
 return t4;
}

static nav__3722_39 fs_Dir_openFileZ__3722(struct fs_Dir__1103 const a0, uint8_t const *const a1, struct fs_File_OpenFlags__1101 const a2) {
 uint32_t *t2;
 nav__3722_52 t12;
 uintptr_t t13;
 uintptr_t t14;
 uintptr_t t11;
 struct fs_File__794 *t16;
 int32_t *t17;
 int32_t t6;
 uint32_t t7;
 uint32_t t1;
 nav__3722_50 t8;
 nav__3722_39 t10;
 nav__3722_39 t15;
 struct fs_Dir__1103 t0;
 uint16_t t9;
 uint8_t t3;
 uint8_t t4;
 bool t5;
 t0 = a0;
 t2 = (uint32_t *)&t1;
 t3 = a2.mode;
 switch (t3) {
  case UINT8_C(0): {
   t4 = UINT8_C(0);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t4 = UINT8_C(1);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t4 = UINT8_C(2);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffc)), zig_shl_u32((uint32_t)t4, UINT8_C(0)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffb)), zig_shl_u32((uint32_t)false, UINT8_C(2)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffff7)), zig_shl_u32((uint32_t)false, UINT8_C(3)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffef)), zig_shl_u32((uint32_t)false, UINT8_C(4)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffdf)), zig_shl_u32((uint32_t)false, UINT8_C(5)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffbf)), zig_shl_u32((uint32_t)false, UINT8_C(6)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffff7f)), zig_shl_u32((uint32_t)false, UINT8_C(7)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)false, UINT8_C(8)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffdff)), zig_shl_u32((uint32_t)false, UINT8_C(9)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffbff)), zig_shl_u32((uint32_t)false, UINT8_C(10)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffff7ff)), zig_shl_u32((uint32_t)false, UINT8_C(11)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff8fff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(12)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff7fff)), zig_shl_u32((uint32_t)false, UINT8_C(15)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffeffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(16)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffdffff)), zig_shl_u32((uint32_t)false, UINT8_C(17)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff3ffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(18)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffefffff)), zig_shl_u32((uint32_t)false, UINT8_C(20)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffdfffff)), zig_shl_u32((uint32_t)false, UINT8_C(21)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffbfffff)), zig_shl_u32((uint32_t)false, UINT8_C(22)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xff7fffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(23)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfeffffff)), zig_shl_u32((uint32_t)false, UINT8_C(24)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xe1ffffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(25)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xdfffffff)), zig_shl_u32((uint32_t)false, UINT8_C(29)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xbfffffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(30)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0x7fffffff)), zig_shl_u32((uint32_t)false, UINT8_C(31)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfeffffff)), zig_shl_u32((uint32_t)true, UINT8_C(24)));
 t2 = (uint32_t *)&t1;
 t5 = a2.allow_ctty;
 t5 = !t5;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffdffff)), zig_shl_u32((uint32_t)t5, UINT8_C(17)));
 t4 = a2.lock;
 switch (t4) {
  case UINT8_C(0): {
   goto zig_block_1;
  }
  case UINT8_C(1): {
   t2 = (uint32_t *)&t1;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffef)), zig_shl_u32((uint32_t)true, UINT8_C(4)));
   t2 = (uint32_t *)&t1;
   t5 = a2.lock_nonblocking;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffb)), zig_shl_u32((uint32_t)t5, UINT8_C(2)));
   goto zig_block_1;
  }
  case UINT8_C(2): {
   t2 = (uint32_t *)&t1;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffdf)), zig_shl_u32((uint32_t)true, UINT8_C(5)));
   t2 = (uint32_t *)&t1;
   t5 = a2.lock_nonblocking;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffb)), zig_shl_u32((uint32_t)t5, UINT8_C(2)));
   goto zig_block_1;
  }
  default: zig_unreachable();
 }

 zig_block_1:;
 t6 = a0.fd;
 t7 = t1;
 t8 = posix_openatZ__1769(t6, a1, t7, UINT16_C(0));
 if (t8.error) {
  t9 = t8.error;
  t10.payload = (struct fs_File__794){-INT32_C(0x55555556)};
  t10.error = t9;
  return t10;
 }
 t6 = t8.payload;
 t5 = a2.lock_nonblocking;
 if (t5) {
  t12 = posix_fcntl__1928(t6, INT32_C(3), (uintptr_t)0ul);
  t5 = t12.error == UINT16_C(0);
  if (t5) {
   t14 = t12.payload;
   t13 = t14;
   goto zig_block_3;
  }
  t9 = t12.error;
  switch (t9) {
   case zig_error_FileBusy: {
    zig_unreachable();
   }
   case zig_error_Locked: {
    zig_unreachable();
   }
   case zig_error_PermissionDenied: {
    zig_unreachable();
   }
   case zig_error_DeadLock: {
    zig_unreachable();
   }
   case zig_error_LockedRegionLimitExceeded: {
    zig_unreachable();
   }
   default: {
    posix_close__1727(t6);
    t10.payload = (struct fs_File__794){-INT32_C(0x55555556)};
    t10.error = t9;
    return t10;
   }
  }

  zig_block_3:;
  t11 = t13;
  t13 = t11;
  t13 = t13 & (uintptr_t)18446744073709551611ul;
  t11 = t13;
  t13 = t11;
  t12 = posix_fcntl__1928(t6, INT32_C(4), t13);
  t5 = t12.error == UINT16_C(0);
  if (t5) {
   goto zig_block_4;
  }
  t9 = t12.error;
  switch (t9) {
   case zig_error_FileBusy: {
    zig_unreachable();
   }
   case zig_error_Locked: {
    zig_unreachable();
   }
   case zig_error_PermissionDenied: {
    zig_unreachable();
   }
   case zig_error_DeadLock: {
    zig_unreachable();
   }
   case zig_error_LockedRegionLimitExceeded: {
    zig_unreachable();
   }
   default: {
    posix_close__1727(t6);
    t10.payload = (struct fs_File__794){-INT32_C(0x55555556)};
    t10.error = t9;
    return t10;
   }
  }

  zig_block_4:;
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t15.error = UINT16_C(0);
 t16 = &t15.payload;
 t17 = (int32_t *)&t16->handle;
 (*t17) = t6;
 return t15;
}

static uintptr_t array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__4816(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)128ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1565(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t8;
 nav__1565_43 t9;
 nav__1565_43 t12;
 nav__1565_43 t20;
 nav__1565_43 t21;
 nav__1565_43 t18;
 struct mem_Allocator__599 *t10;
 struct mem_Allocator__599 t11;
 nav__1565_43 *t13;
 uint8_t **t14;
 uint8_t *t15;
 nav__1565_55 t16;
 nav__1565_43 const *t19;
 uint16_t t17;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t8 = (*t7);
 t9 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t8);
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 t12 = mem_Allocator_remap__anon_3648__5136(t11, t9, a1);
 t6 = t12.ptr != NULL;
 if (t6) {
  t9 = t12;
  t7 = (*t1);
  t13 = (nav__1565_43 *)&t7->items;
  t14 = &t13->ptr;
  t15 = t9.ptr;
  (*t14) = t15;
  t7 = (*t1);
  t2 = (uintptr_t *)&t7->capacity;
  t3 = t9.len;
  (*t2) = t3;
  goto zig_block_1;
 }
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 t16 = mem_Allocator_alignedAlloc__anon_3653__5137(t11, a1);
 if (t16.error) {
  t17 = t16.error;
  return t17;
 }
 t12 = t16.payload;
 t18 = t12;
 t19 = (nav__1565_43 const *)&t18;
 t13 = (nav__1565_43 *)&a0->items;
 t20 = (*t13);
 t3 = t20.len;
 t20 = (*t19);
 t15 = t20.ptr;
 t15 = (uint8_t *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t20.ptr = t15;
 t20.len = t3;
 t13 = (nav__1565_43 *)&a0->items;
 t21 = (*t13);
 t15 = t21.ptr;
 if (t20.len != 0) memcpy(t20.ptr, t15, t20.len * sizeof(uint8_t));
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 mem_Allocator_free__anon_3655__5138(t11, t9);
 t7 = (*t1);
 t13 = (nav__1565_43 *)&t7->items;
 t14 = &t13->ptr;
 t15 = t12.ptr;
 (*t14) = t15;
 t7 = (*t1);
 t2 = (uintptr_t *)&t7->capacity;
 t3 = t12.len;
 (*t2) = t3;
 goto zig_block_1;

 zig_block_1:;
 return 0;
}

static uint16_t bincode_serialize__anon_3165__4817(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, uint32_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_3660__5139(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3167__4818(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__struct_1181__1181 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3665__5140(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3169__4819(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0) {
 (void)a0;
 return 0;
}

static uint16_t bincode_serialize__anon_3171__4820(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__struct_1182__1182 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3666__5141(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3173__4821(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__struct_1183__1183 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3667__5142(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3175__4822(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, uint64_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_3668__5143(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3177__4823(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, uint8_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeEnum__anon_3671__5144(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static nav__1757_38 posix_write__1757(int32_t const a0, nav__1757_40 const a1) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t const *t3;
 intptr_t t5;
 nav__1757_38 t7;
 uint32_t t4;
 uint16_t t6;
 bool t2;
 t0 = a1.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__1757_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_16:
 t3 = a1.ptr;
 t0 = a1.len;
 t0 = ((uintptr_t)2147483647ul < t0) ? (uintptr_t)2147483647ul : t0;
 t4 = (uint32_t)t0;
 t0 = (uintptr_t)t4;
 t5 = write(a0, t3, t0);
 t6 = posix_errno__anon_3681__5146(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(4): {
   goto zig_block_1;
  }
  case UINT16_C(22): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InvalidArgument};
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(2): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessNotFound};
  }
  case UINT16_C(35): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_WouldBlock};
  }
  case UINT16_C(9): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NotOpenForWriting};
  }
  case UINT16_C(39): {
   zig_unreachable();
  }
  case UINT16_C(69): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DiskQuota};
  }
  case UINT16_C(27): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_FileTooBig};
  }
  case UINT16_C(5): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InputOutput};
  }
  case UINT16_C(28): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
  }
  case UINT16_C(13): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_AccessDenied};
  }
  case UINT16_C(1): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_AccessDenied};
  }
  case UINT16_C(32): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_BrokenPipe};
  }
  case UINT16_C(54): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionResetByPeer};
  }
  case UINT16_C(16): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DeviceBusy};
  }
  case UINT16_C(6): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoDevice};
  }
  default: {
   t6 = posix_unexpectedErrno__2021(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_1:;
 goto zig_loop_16;
}

static nav__1749_38 posix_read__1749(int32_t const a0, nav__1749_40 const a1) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3;
 intptr_t t5;
 nav__1749_38 t7;
 uint32_t t4;
 uint16_t t6;
 bool t2;
 t0 = a1.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__1749_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_16:
 t3 = a1.ptr;
 t0 = a1.len;
 t0 = ((uintptr_t)2147483647ul < t0) ? (uintptr_t)2147483647ul : t0;
 t4 = (uint32_t)t0;
 t0 = (uintptr_t)t4;
 t5 = read(a0, t3, t0);
 t6 = posix_errno__anon_3681__5146(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(4): {
   goto zig_block_1;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(2): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessNotFound};
  }
  case UINT16_C(35): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_WouldBlock};
  }
  case UINT16_C(89): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Canceled};
  }
  case UINT16_C(9): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NotOpenForReading};
  }
  case UINT16_C(5): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InputOutput};
  }
  case UINT16_C(21): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_IsDir};
  }
  case UINT16_C(55): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  case UINT16_C(12): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  case UINT16_C(57): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SocketNotConnected};
  }
  case UINT16_C(54): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionResetByPeer};
  }
  case UINT16_C(60): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionTimedOut};
  }
  default: {
   t6 = posix_unexpectedErrno__2021(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_1:;
 goto zig_loop_16;
}

static nav__4889_38 math_mul__anon_3220__4889(uintptr_t const a0, uintptr_t const a1) {
 nav__4889_42 t0;
 uintptr_t t3;
 nav__4889_38 t4;
 uint8_t t1;
 bool t2;
 t0.f1 = zig_mulo_u64(&t0.f0, a0, a1, UINT8_C(64));
 t1 = t0.f1;
 t2 = t1 != UINT8_C(0);
 if (t2) {
  return (nav__4889_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Overflow};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t0.f0;
 t4.payload = t3;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__4890_39 mem_Allocator_allocBytesWithAlignment__anon_3222__4890(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t7;
 struct mem_Allocator_VTable__602 const *const *t8;
 struct mem_Allocator_VTable__602 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__4890_52 t17;
 nav__4890_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__4890_39){(uint8_t *)UINTPTR_MAX,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t7 = t5;
 t1 = (struct mem_Allocator__599 const *)&t7;
 t8 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__4890_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t18.payload = t4;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__4195_38 io_Reader_read__4195(struct io_Reader__2141 const a0, nav__4195_41 const a1) {
 struct io_Reader__2141 const *t1;
 nav__4195_38 (*const *t2)(void const *, nav__4195_41);
 nav__4195_38 (*t3)(void const *, nav__4195_41);
 void const *t4;
 nav__4195_38 t5;
 struct io_Reader__2141 t0;
 t0 = a0;
 t1 = (struct io_Reader__2141 const *)&t0;
 t2 = (nav__4195_38 (*const *)(void const *, nav__4195_41))&t1->readFn;
 t3 = (*t2);
 t4 = a0.context;
 t5 = t3(t4, a1);
 return t5;
}

static nav__4892_39 mem_sliceAsBytes__anon_3353__4892(nav__4892_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t const *t4;
 uint8_t const *t5;
 uint8_t const *const *t6;
 nav__4892_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__4892_39){(uint8_t const *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t const *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t const *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint16_t fmt_format__anon_3841__5177(struct io_Writer__3827 const a0, nav__5177_40 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 struct shared_Command__1178 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5177_45){(uint8_t const *)&__anon_3951,(uintptr_t)49ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_4328__5570(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5177_45){(uint8_t const *)&__anon_4335,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5052_40 fmt_bufPrint__anon_3407__5052(nav__5052_39 const a0, nav__5052_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__5052_71 (**t11)(void const *, nav__5052_48);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__5052_39 t15;
 nav__5052_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5052_71 (**)(void const *, nav__5052_48))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_3841__5177(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5052_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4351__5571(struct io_Writer__3827 const a0) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5571_43){(uint8_t const *)&__anon_4451,(uintptr_t)68ul});
 if (t3) {
  return t3;
 }
 t3 = fmt_formatType__anon_4459__5572(UINT64_C(2), (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5571_43){(uint8_t const *)&__anon_4335,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5054_40 fmt_bufPrint__anon_3429__5054(nav__5054_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__5054_54 (**t11)(void const *, nav__5054_56);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__5054_39 t15;
 nav__5054_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5054_54 (**)(void const *, nav__5054_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_4351__5571(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5054_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4468__5573(struct io_Writer__3827 const a0) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5573_43){(uint8_t const *)&__anon_4574,(uintptr_t)82ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5055_40 fmt_bufPrint__anon_3442__5055(nav__5055_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__5055_54 (**t11)(void const *, nav__5055_56);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__5055_39 t15;
 nav__5055_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5055_54 (**)(void const *, nav__5055_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_4468__5573(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5055_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4580__5574(struct io_Writer__3827 const a0, nav__5574_41 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 nav__5574_40 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5574_40){(uint8_t const *)&__anon_4628,(uintptr_t)32ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_4732__5575(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5574_40){(uint8_t const *)&__anon_4761,(uintptr_t)22ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5056_40 fmt_bufPrint__anon_3455__5056(nav__5056_39 const a0, nav__5056_43 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__5056_59 (**t11)(void const *, nav__5056_42);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__5056_39 t15;
 nav__5056_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5056_59 (**)(void const *, nav__5056_42))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_4580__5574(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5056_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(struct array_list_ArrayListAligned_28u8_2cnull_29__806 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 const *t1;
 struct mem_Allocator__599 const *t2;
 struct mem_Allocator__599 t3;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t0;
 nav__1529_41 t5;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 const *)&t0;
 t2 = (struct mem_Allocator__599 const *)&t1->allocator;
 t3 = (*t2);
 t4 = (*t1);
 t5 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t4);
 mem_Allocator_free__anon_3655__5138(t3, t5);
 return;
}

static void fs_File_close__1385(struct fs_File__794 const a0) {
 int32_t t0;
 t0 = a0.handle;
 posix_close__1727(t0);
 return;
}

static struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_init__3512(struct mem_Allocator__599 const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t0;
 t0.items = (nav__3512_43){(struct environment_linked_libraries_root_LibraryInfo__934 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 t0.capacity = (uintptr_t)0ul;
 t0.allocator = a0;
 return t0;
}

static bool math_isPowerOfTwo__anon_3478__5057(uintptr_t const a0) {
 uint64_t t0;
 uintptr_t t2;
 bool t1;
 t0 = a0;
 t1 = t0 > UINT64_C(0);
 debug_assert__180(t1);
 t2 = a0 - (uintptr_t)1ul;
 t2 = a0 & t2;
 t0 = t2;
 t1 = t0 == UINT64_C(0);
 return t1;
}

static uint16_t posix_faccessatZ__1913(int32_t const a0, uint8_t const *const a1, uint32_t const a2, uint32_t const a3) {
 unsigned int t0;
 unsigned int t1;
 int t2;
 uint16_t t3;
 t0 = a2;
 t1 = a3;
 t2 = faccessat(a0, a1, t0, t1);
 t3 = posix_errno__anon_2970__4656(t2);
 switch (t3) {
  case UINT16_C(0): {
   return 0;
  }
  case UINT16_C(13): {
   return zig_error_PermissionDenied;
  }
  case UINT16_C(1): {
   return zig_error_PermissionDenied;
  }
  case UINT16_C(30): {
   return zig_error_ReadOnlyFileSystem;
  }
  case UINT16_C(62): {
   return zig_error_SymLinkLoop;
  }
  case UINT16_C(26): {
   return zig_error_FileBusy;
  }
  case UINT16_C(20): {
   return zig_error_FileNotFound;
  }
  case UINT16_C(2): {
   return zig_error_FileNotFound;
  }
  case UINT16_C(63): {
   return zig_error_NameTooLong;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(5): {
   return zig_error_InputOutput;
  }
  case UINT16_C(12): {
   return zig_error_SystemResources;
  }
  case UINT16_C(92): {
   t3 = posix_unexpectedErrno__2021(t3);
   return t3;
  }
  default: {
   t3 = posix_unexpectedErrno__2021(t3);
   return t3;
  }
 }
}

static nav__1769_38 posix_openatZ__1769(int32_t const a0, uint8_t const *const a1, uint32_t const a2, uint16_t const a3) {
 int t0;
 unsigned int t1;
 int32_t t3;
 nav__1769_38 t4;
 uint16_t t2;
 zig_loop_11:
 t0 = a0;
 t1 = (unsigned int)a3;
 t0 = openat(t0, a1, a2, t1);
 t2 = posix_errno__anon_2970__4656(t0);
 switch (t2) {
  case UINT16_C(0): {
   t3 = t0;
   t4.payload = t3;
   t4.error = UINT16_C(0);
   return t4;
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(22): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_BadPathName};
  }
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(13): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_AccessDenied};
  }
  case UINT16_C(27): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileTooBig};
  }
  case UINT16_C(84): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileTooBig};
  }
  case UINT16_C(21): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_IsDir};
  }
  case UINT16_C(62): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_SymLinkLoop};
  }
  case UINT16_C(24): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_ProcessFdQuotaExceeded};
  }
  case UINT16_C(63): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NameTooLong};
  }
  case UINT16_C(23): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_SystemFdQuotaExceeded};
  }
  case UINT16_C(19): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NoDevice};
  }
  case UINT16_C(2): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileNotFound};
  }
  case UINT16_C(12): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_SystemResources};
  }
  case UINT16_C(28): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NoSpaceLeft};
  }
  case UINT16_C(20): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NotDir};
  }
  case UINT16_C(1): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_AccessDenied};
  }
  case UINT16_C(17): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_PathAlreadyExists};
  }
  case UINT16_C(16): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_DeviceBusy};
  }
  case UINT16_C(45): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileLocksNotSupported};
  }
  case UINT16_C(35): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_WouldBlock};
  }
  case UINT16_C(26): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileBusy};
  }
  case UINT16_C(6): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NoDevice};
  }
  case UINT16_C(92): {
   t2 = posix_unexpectedErrno__2021(t2);
   t4.payload = -INT32_C(0x55555556);
   t4.error = t2;
   return t4;
  }
  default: {
   t2 = posix_unexpectedErrno__2021(t2);
   t4.payload = -INT32_C(0x55555556);
   t4.error = t2;
   return t4;
  }
 }

 zig_block_0:;
 goto zig_loop_11;
}

static nav__1928_38 posix_fcntl__1928(int32_t const a0, int32_t const a1, uintptr_t const a2) {
 unsigned long t1;
 uintptr_t t3;
 nav__1928_38 t4;
 int t0;
 uint16_t t2;
 zig_loop_4:
 t0 = a1;
 t1 = a2;
 t0 = fcntl(a0, t0, t1);
 t2 = posix_errno__anon_2970__4656(t0);
 switch (t2) {
  case UINT16_C(0): {
   t3 = (uintptr_t)t0;
   t4.payload = t3;
   t4.error = UINT16_C(0);
   return t4;
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(16): {
   return (nav__1928_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_FileBusy};
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(1): {
   return (nav__1928_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_PermissionDenied};
  }
  case UINT16_C(24): {
   return (nav__1928_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessFdQuotaExceeded};
  }
  case UINT16_C(20): {
   zig_unreachable();
  }
  case UINT16_C(11): {
   return (nav__1928_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DeadLock};
  }
  case UINT16_C(77): {
   return (nav__1928_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_LockedRegionLimitExceeded};
  }
  case UINT16_C(35):
  case UINT16_C(13): {
   return (nav__1928_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Locked};
  }
  default: {
   t2 = posix_unexpectedErrno__2021(t2);
   t4.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t4.error = t2;
   return t4;
  }
 }

 zig_block_0:;
 goto zig_loop_4;
}

static void posix_close__1727(int32_t const a0) {
 int t0;
 uint16_t t1;
 t0 = close_24NOCANCEL(a0);
 t1 = posix_errno__anon_2970__4656(t0);
 switch (t1) {
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   return;
  }
  default: {
   return;
  }
 }
}

static nav__1575_39 array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(struct array_list_ArrayListAligned_28u8_2cnull_29__806 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 const *t1;
 nav__1575_39 const *t2;
 uint8_t *const *t3;
 uintptr_t t4;
 uint8_t *t5;
 nav__1575_39 t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t0;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 const *)&t0;
 t2 = (nav__1575_39 const *)&t1->items;
 t3 = &t2->ptr;
 t4 = a0.capacity;
 t5 = (*t3);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t6.ptr = t5;
 t6.len = t4;
 return t6;
}

static nav__5136_39 mem_Allocator_remap__anon_3648__5136(struct mem_Allocator__599 const a0, nav__5136_39 const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 nav__5136_39 const *t3;
 uint64_t t4;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t15;
 nav__5136_39 t7;
 nav__5136_39 t10;
 nav__5136_39 t2;
 uint8_t *t8;
 uint8_t *t20;
 uint8_t *t21;
 uint8_t *t22;
 void *t9;
 uintptr_t t11;
 uintptr_t t13;
 nav__5136_50 t12;
 struct mem_Allocator_VTable__602 const *const *t16;
 struct mem_Allocator_VTable__602 const *t17;
 uint8_t *(*const *t18)(void *, nav__5136_39, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*t19)(void *, nav__5136_39, uint8_t, uintptr_t, uintptr_t);
 uint8_t *const *t23;
 bool t5;
 uint8_t t14;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = (nav__5136_39 const *)&t2;
 t4 = a2;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  t6 = (*t1);
  mem_Allocator_free__anon_3655__5138(t6, a1);
  t7 = (*t3);
  t8 = t7.ptr;
  t8 = (uint8_t *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t9 = (void *)t8;
  t7.ptr = t9;
  t7.len = (uintptr_t)0ul;
  t10 = t7;
  return t10;
 }
 goto zig_block_0;

 zig_block_0:;
 t11 = a1.len;
 t4 = t11;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return (nav__5136_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 }
 goto zig_block_1;

 zig_block_1:;
 t10 = mem_sliceAsBytes__anon_4876__5580(a1);
 t12 = math_mul__anon_3220__4889((uintptr_t)1ul, a2);
 t5 = t12.error == UINT16_C(0);
 if (t5) {
  t13 = t12.payload;
  t11 = t13;
  goto zig_block_2;
 }
 return (nav__5136_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_2:;
 t6 = (*t1);
 t14 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t13 = (uintptr_t)zig_return_address();
 t15 = t6;
 t1 = (struct mem_Allocator__599 const *)&t15;
 t16 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t17 = (*t16);
 t18 = (uint8_t *(*const *)(void *, nav__5136_39, uint8_t, uintptr_t, uintptr_t))&t17->remap;
 t19 = (*t18);
 t9 = t6.ptr;
 t20 = t19(t9, t10, t14, t11, t13);
 t5 = t20 != NULL;
 if (t5) {
  t21 = t20;
  t8 = t21;
  goto zig_block_3;
 }
 return (nav__5136_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_3:;
 t22 = t8;
 t23 = (uint8_t *const *)&t22;
 t8 = (*t23);
 t8 = (uint8_t *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t8;
 t10.len = t11;
 memcpy(&t7, &t10, sizeof(nav__5136_39));
 t7 = mem_bytesAsSlice__anon_4889__5581(t7);
 t10 = t7;
 return t10;
}

static nav__5137_40 mem_Allocator_alignedAlloc__anon_3653__5137(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__5137_40 t4;
 nav__5137_40 t8;
 nav__5137_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__5137_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_3006__4657(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__5137_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void mem_Allocator_free__anon_3655__5138(struct mem_Allocator__599 const a0, nav__5138_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5138_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t12;
 struct mem_Allocator_VTable__602 const *const *t13;
 struct mem_Allocator_VTable__602 const *t14;
 void (*const *t15)(void *, nav__5138_40, uint8_t, uintptr_t);
 void (*t16)(void *, nav__5138_40, uint8_t, uintptr_t);
 void *t17;
 bool t5;
 uint8_t t11;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_4876__5580(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t11 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t12 = t10;
 t1 = (struct mem_Allocator__599 const *)&t12;
 t13 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t14 = (*t13);
 t15 = (void (*const *)(void *, nav__5138_40, uint8_t, uintptr_t))&t14->free;
 t16 = (*t15);
 t17 = t10.ptr;
 t16(t17, t2, t11, t3);
 return;
}

static uint16_t bincode_serializeInt__anon_3660__5139(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, uint32_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t8;
 void const *t9;
 nav__5139_48 (**t10)(void const *, nav__5139_50);
 struct io_Writer__3827 t11;
 struct io_Writer__3827 t6;
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t14;
 struct io_Writer__3827 const *t13;
 nav__5139_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[4];
 uint8_t t15[4];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5139_48 (**)(void const *, nav__5139_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__3827 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__3827 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[4]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[4]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)4ul;
 t18 = io_Writer_writeAll__5168(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3665__5140(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__struct_1181__1181 const a1) {
 nav__5140_45 t2;
 int32_t t0;
 uint16_t t1;
 t0 = a1.pid;
 t1 = bincode_serialize__anon_4920__5582(a0, t0);
 if (t1) {
  return t1;
 }
 t2 = a1.uri;
 t1 = bincode_serialize__anon_4922__5583(a0, t2);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3666__5141(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__struct_1182__1182 const a1) {
 nav__5141_45 t0;
 uint16_t t1;
 t0 = a1.name;
 t1 = bincode_serialize__anon_4922__5583(a0, t0);
 if (t1) {
  return t1;
 }
 t0 = a1.version;
 t1 = bincode_serialize__anon_4922__5583(a0, t0);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3667__5142(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_Command__struct_1183__1183 const a1) {
 struct shared_MarkerType__1185 t2;
 int32_t t0;
 uint16_t t1;
 t0 = a1.pid;
 t1 = bincode_serialize__anon_4920__5582(a0, t0);
 if (t1) {
  return t1;
 }
 t2 = a1.marker;
 t1 = bincode_serialize__anon_4924__5584(a0, t2);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeInt__anon_3668__5143(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, uint64_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t8;
 void const *t9;
 nav__5143_48 (**t10)(void const *, nav__5143_50);
 struct io_Writer__3827 t11;
 struct io_Writer__3827 t6;
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t14;
 struct io_Writer__3827 const *t13;
 nav__5143_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[8];
 uint8_t t15[8];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5143_48 (**)(void const *, nav__5143_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__3827 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__3827 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[8]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[8]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)8ul;
 t18 = io_Writer_writeAll__5168(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializeEnum__anon_3671__5144(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, uint8_t const a1) {
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 t0 = a1;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3165__4817(a0, t1);
 if (t2) {
  return t2;
 }
 return 0;
}

static uint16_t posix_errno__anon_3681__5146(intptr_t const a0) {
 int64_t t1;
 int *t3;
 int t4;
 uint16_t t0;
 uint16_t t5;
 bool t2;
 t1 = a0;
 t2 = t1 == -INT64_C(1);
 if (t2) {
  t3 = zig_e___error();
  t4 = (*t3);
  t5 = (uint16_t)t4;
  t0 = t5;
  goto zig_block_0;
 }
 t0 = UINT16_C(0);
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 **t1;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__5160_38 io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160(void const *const a0, nav__5160_41 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t0;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *t1;
 nav__5160_38 t2;
 nav__5160_38 t3;
 t0 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)a0;
 t1 = (*t0);
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4149(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__5160_38));
 return t3;
}

static uint16_t io_Writer_writeAll__5168(struct io_Writer__3827 const a0, nav__5168_40 const a1) {
 struct io_Writer__3827 const *t1;
 nav__5168_40 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 struct io_Writer__3827 t10;
 struct io_Writer__3827 t0;
 nav__5168_40 t11;
 nav__5168_40 t2;
 uint8_t const *t12;
 nav__5168_43 t14;
 uint16_t t15;
 bool t9;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = a1;
 t3 = (nav__5168_40 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 != t8;
 if (t9) {
  t6 = t4;
  t10 = (*t1);
  t5 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t const *)(((uintptr_t)t12) + (t5*sizeof(uint8_t)));
  t13 = t11.len;
  t5 = t13 - t5;
  t11.ptr = t12;
  t11.len = t5;
  t14 = io_Writer_write__5167(t10, t11);
  if (t14.error) {
   t15 = t14.error;
   return t15;
  }
  t5 = t14.payload;
  t5 = t6 + t5;
  t4 = t5;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 return 0;
}

static uint16_t shared_Command_format__anon_4988__5585(struct shared_Command__1178 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 struct io_Writer__3827 const *t1;
 struct shared_Command__struct_1181__1181 t3;
 struct io_Writer__3827 t4;
 struct io_Writer__3827 t0;
 nav__5585_44 t6;
 nav__5585_44 t10;
 nav__5585_66 t7;
 struct shared_Command__struct_1182__1182 t9;
 nav__5585_76 t11;
 struct shared_Command__struct_1183__1183 t12;
 struct shared_MarkerType__1185 t13;
 nav__5585_78 t14;
 uint64_t t15;
 nav__5585_80 t16;
 int32_t t5;
 uint16_t t8;
 uint8_t t2;
 nav__5585_84 t17;
 (void)a1;
 t0 = a2;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.ExecutedBenchmark;
   t4 = (*t1);
   t5 = t3.pid;
   t6 = t3.uri;
   t7.f0 = t5;
   t7.f1 = t6;
   t8 = io_Writer_print__anon_5008__5586(t4, t7);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5168(t4, (nav__5585_44){(uint8_t const *)&__anon_2011,(uintptr_t)14ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5168(t4, (nav__5585_44){(uint8_t const *)&__anon_2016,(uintptr_t)13ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5168(t4, (nav__5585_44){(uint8_t const *)&__anon_2021,(uintptr_t)3ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5168(t4, (nav__5585_44){(uint8_t const *)&__anon_2026,(uintptr_t)12ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(5): {
   t9 = a0.payload.SetIntegration;
   t4 = (*t1);
   t6 = t9.name;
   t10 = t9.version;
   t11.f0 = t6;
   t11.f1 = t10;
   t8 = io_Writer_print__anon_5029__5587(t4, t11);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(6): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5168(t4, (nav__5585_44){(uint8_t const *)&__anon_2034,(uintptr_t)3ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(7): {
   t12 = a0.payload.AddMarker;
   t4 = (*t1);
   t5 = t12.pid;
   t13 = t12.marker;
   t14.f0 = t5;
   t14.f1 = t13;
   t8 = io_Writer_print__anon_5041__5588(t4, t14);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(8): {
   t15 = a0.payload.SetVersion;
   t4 = (*t1);
   t16.f0 = t15;
   t8 = io_Writer_print__anon_5050__5589(t4, t16);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(9): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5168(t4, (nav__5585_44){(uint8_t const *)&__anon_2049,(uintptr_t)18ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(10): {
   t2 = a0.payload.IntegrationModeResponse;
   t4 = (*t1);
   t17.f0 = t2;
   t8 = io_Writer_print__anon_5063__5590(t4, t17);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatType__anon_4328__5570(struct shared_Command__1178 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct shared_Command__1178 const *t1;
 struct shared_Command__1178 t3;
 struct shared_Command__1178 t0;
 struct io_Writer__3827 t2;
 uint16_t t4;
 (void)a3;
 t0 = a0;
 t1 = (struct shared_Command__1178 const *)&t0;
 t2 = a2;
 t3 = (*t1);
 t4 = shared_Command_format__anon_4988__5585(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__4154_39 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 const *t1;
 nav__4154_39 const *t2;
 uintptr_t t3;
 nav__4154_39 t4;
 uint8_t *t5;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 const *)&t0;
 t2 = (nav__4154_39 const *)&t1->buffer;
 t3 = a0.pos;
 t4 = (*t2);
 t5 = t4.ptr;
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t5;
 t4.len = t3;
 return t4;
}

static uint16_t fmt_formatType__anon_4459__5572(uint64_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 uint64_t t0;
 struct io_Writer__3827 t1;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5084__5591(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_4732__5575(nav__5575_39 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct io_Writer__3827 const *t2;
 uint64_t t3;
 struct io_Writer__3827 t5;
 struct io_Writer__3827 t1;
 nav__5575_39 t7;
 nav__5575_39 t0;
 uint16_t t6;
 uint16_t t8;
 bool t4;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__3827 const *)&t1;
 t3 = a3;
 t4 = t3 == UINT64_C(0);
 if (t4) {
  t5 = (*t2);
  t6 = io_Writer_writeAll__5168(t5, (nav__5575_39){(uint8_t const *)&__anon_5099,(uintptr_t)7ul});
  return t6;
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t7, &a0, sizeof(nav__5575_39));
 t6 = fmt_formatBuf__anon_5109__5592(t7, a1, a2);
 memcpy(&t8, &t6, sizeof(uint16_t));
 return t8;
}

static nav__5580_39 mem_sliceAsBytes__anon_4876__5580(nav__5580_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__5580_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5580_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static nav__5581_39 mem_bytesAsSlice__anon_4889__5581(nav__5581_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__5581_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5581_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t0 = t0 / (uintptr_t)1ul;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static nav__3955_38 io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955(void const *const a0, nav__3955_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t1;
 nav__3955_38 t2;
 nav__3955_38 t3;
 t0 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)a0;
 t1 = (*t0);
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1553(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__3955_38));
 return t3;
}

static uint16_t bincode_serialize__anon_4920__5582(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, int32_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_5117__5593(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_4922__5583(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, nav__5583_40 const a1) {
 uint16_t t0;
 t0 = bincode_serializePointer__anon_5122__5594(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_4924__5584(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_MarkerType__1185 const a1) {
 uint16_t t0;
 t0 = bincode_serializeUnion__anon_5123__5595(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static nav__4149_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4149(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const a0, nav__4149_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t1;
 nav__4149_42 const *t3;
 uintptr_t t4;
 uintptr_t t10;
 uint64_t t5;
 uint64_t t11;
 uintptr_t *t7;
 nav__4149_50 *t8;
 nav__4149_50 t9;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *t12;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *t0;
 uint8_t *t13;
 nav__4149_42 t14;
 nav__4149_42 t2;
 uint8_t const *t15;
 nav__4149_38 t16;
 bool t6;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t0;
 t2 = a1;
 t3 = (nav__4149_42 const *)&t2;
 t4 = a1.len;
 t5 = t4;
 t6 = t5 == UINT64_C(0);
 if (t6) {
  return (nav__4149_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (uintptr_t *)&a0->pos;
 t4 = (*t7);
 t8 = (nav__4149_50 *)&a0->buffer;
 t9 = (*t8);
 t10 = t9.len;
 t5 = t4;
 t11 = t10;
 t6 = t5 >= t11;
 if (t6) {
  return (nav__4149_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
 }
 goto zig_block_1;

 zig_block_1:;
 t8 = (nav__4149_50 *)&a0->buffer;
 t9 = (*t8);
 t10 = t9.len;
 t7 = (uintptr_t *)&a0->pos;
 t4 = (*t7);
 t4 = t10 - t4;
 t10 = a1.len;
 t10 = (t4 < t10) ? t4 : t10;
 t11 = t10;
 t12 = (*t1);
 t8 = (nav__4149_50 *)&t12->buffer;
 t7 = (uintptr_t *)&a0->pos;
 t10 = (*t7);
 t9 = (*t8);
 t13 = t9.ptr;
 t13 = (uint8_t *)(((uintptr_t)t13) + (t10*sizeof(uint8_t)));
 t10 = t11;
 t9.ptr = t13;
 t9.len = t10;
 t14 = (*t3);
 t15 = t14.ptr;
 t15 = (uint8_t const *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10 = t11;
 t14.ptr = t15;
 t14.len = t10;
 t15 = t14.ptr;
 if (t9.len != 0) memcpy(t9.ptr, t15, t9.len * sizeof(uint8_t));
 t12 = (*t1);
 t7 = (uintptr_t *)&t12->pos;
 t10 = (*t7);
 t4 = t11;
 t4 = t10 + t4;
 (*t7) = t4;
 t6 = t11 == UINT64_C(0);
 if (t6) {
  return (nav__4149_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
 }
 goto zig_block_2;

 zig_block_2:;
 t4 = t11;
 t16.payload = t4;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__5167_38 io_Writer_write__5167(struct io_Writer__3827 const a0, nav__5167_41 const a1) {
 struct io_Writer__3827 const *t1;
 nav__5167_38 (*const *t2)(void const *, nav__5167_41);
 nav__5167_38 (*t3)(void const *, nav__5167_41);
 void const *t4;
 nav__5167_38 t5;
 struct io_Writer__3827 t0;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (nav__5167_38 (*const *)(void const *, nav__5167_41))&t1->writeFn;
 t3 = (*t2);
 t4 = a0.context;
 t5 = t3(t4, a1);
 return t5;
}

static uint16_t io_Writer_print__anon_5008__5586(struct io_Writer__3827 const a0, nav__5586_41 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5127__5596(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5029__5587(struct io_Writer__3827 const a0, nav__5587_41 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5131__5597(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5041__5588(struct io_Writer__3827 const a0, nav__5588_40 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5135__5598(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5050__5589(struct io_Writer__3827 const a0, nav__5589_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5139__5599(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5063__5590(struct io_Writer__3827 const a0, nav__5590_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5143__5600(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__5189_38 unicode_utf8ByteSequenceLength__5189(uint8_t const a0) {
 nav__5189_38 t0;
 switch (a0) {
  default: if ((a0 >= UINT8_C(0) && a0 <= UINT8_C(127))) {
   t0 = (nav__5189_38){0,UINT8_C(1)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(192) && a0 <= UINT8_C(223))) {
   t0 = (nav__5189_38){0,UINT8_C(2)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(224) && a0 <= UINT8_C(239))) {
   t0 = (nav__5189_38){0,UINT8_C(3)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(240) && a0 <= UINT8_C(247))) {
   t0 = (nav__5189_38){0,UINT8_C(4)};
   goto zig_block_0;
  }{
   t0 = (nav__5189_38){zig_error_Utf8InvalidStartByte,UINT8_C(0x2)};
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return t0;
}

static nav__5205_38 unicode_utf8CountCodepoints__5205(nav__5205_40 const a0) {
 nav__5205_40 const *t1;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t2;
 uintptr_t t3;
 uint64_t t6;
 uint64_t t7;
 nav__5205_40 t9;
 nav__5205_40 t0;
 uint8_t const *t10;
 uint8_t const (*t11)[8];
 nav__5205_38 t16;
 nav__5205_50 t17;
 nav__5205_48 t14;
 uint16_t t15;
 bool t8;
 uint8_t t12[8];
 uint8_t t13;
 t0 = a0;
 t1 = (nav__5205_40 const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = (uintptr_t)0ul;
 zig_loop_9:
 t4 = t3;
 t5 = a0.len;
 t6 = t4;
 t7 = t5;
 t8 = t6 < t7;
 if (t8) {
  zig_loop_18:
  t5 = t3;
  t5 = t5 + (uintptr_t)8ul;
  t4 = a0.len;
  t7 = t5;
  t6 = t4;
  t8 = t7 <= t6;
  if (t8) {
   t4 = t3;
   t9 = (*t1);
   t10 = t9.ptr;
   t10 = (uint8_t const *)(((uintptr_t)t10) + (t4*sizeof(uint8_t)));
   t11 = (uint8_t const (*)[8])t10;
   memcpy(t12, (const char *)t11, sizeof(uint8_t[8]));
   memcpy(&t4, &t12, sizeof(uintptr_t));
   t4 = zig_wrap_u64(t4, UINT8_C(64));
   t4 = t4 & (uintptr_t)9259542123273814144ul;
   t6 = t4;
   t8 = t6 != UINT64_C(0);
   if (t8) {
    goto zig_block_2;
   }
   goto zig_block_4;

   zig_block_4:;
   t4 = t2;
   t4 = t4 + (uintptr_t)8ul;
   t2 = t4;
   t4 = t3;
   t4 = t4 + (uintptr_t)8ul;
   t3 = t4;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_18;

  zig_block_2:;
  t5 = t3;
  t4 = a0.len;
  t7 = t5;
  t6 = t4;
  t8 = t7 < t6;
  if (t8) {
   t4 = t3;
   t13 = a0.ptr[t4];
   t14 = unicode_utf8ByteSequenceLength__5189(t13);
   if (t14.error) {
    t15 = t14.error;
    t16.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
    t16.error = t15;
    return t16;
   }
   t13 = t14.payload;
   t4 = t3;
   t5 = (uintptr_t)t13;
   t5 = t4 + t5;
   t4 = a0.len;
   t6 = t5;
   t7 = t4;
   t8 = t6 > t7;
   if (t8) {
    return (nav__5205_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_TruncatedInput};
   }
   goto zig_block_6;

   zig_block_6:;
   switch (t13) {
    case UINT8_C(1): {
     goto zig_block_7;
    }
    default: {
     t4 = t3;
     t9 = (*t1);
     t10 = t9.ptr;
     t10 = (uint8_t const *)(((uintptr_t)t10) + (t4*sizeof(uint8_t)));
     t4 = (uintptr_t)t13;
     t9.ptr = t10;
     t9.len = t4;
     t17 = unicode_utf8Decode__5195(t9);
     if (t17.error) {
      t15 = t17.error;
      t16.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
      t16.error = t15;
      return t16;
     }
     goto zig_block_7;
    }
   }

   zig_block_7:;
   t4 = t3;
   t5 = (uintptr_t)t13;
   t5 = t4 + t5;
   t3 = t5;
   t5 = t2;
   t5 = t5 + (uintptr_t)1ul;
   t2 = t5;
   goto zig_block_5;
  }
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_9;

 zig_block_0:;
 t5 = t2;
 t16.payload = t5;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_formatBuf__anon_5109__5592(nav__5592_39 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 struct io_Writer__3827 const *t1;
 nav__5592_44 t2;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 nav__5592_48 t6;
 uint64_t t8;
 uint64_t t9;
 struct io_Writer__3827 t10;
 struct io_Writer__3827 t0;
 nav__5592_57 t15;
 nav__5592_39 t17;
 nav__5592_39 t20;
 uint8_t *t19;
 uint32_t t14;
 uint16_t t11;
 uint16_t t12;
 nav__5592_60 t16;
 bool t3;
 uint8_t t18;
 uint8_t t13[4];
 t0 = a2;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = a1.width;
 t3 = t2.is_null != true;
 if (t3) {
  t4 = t2.payload;
  t6 = unicode_utf8CountCodepoints__5205(a0);
  t3 = t6.error == UINT16_C(0);
  if (t3) {
   t7 = t6.payload;
   t5 = t7;
   goto zig_block_1;
  }
  t7 = a0.len;
  t5 = t7;
  goto zig_block_1;

  zig_block_1:;
  t8 = t5;
  t9 = t4;
  t3 = t8 < t9;
  if (t3) {
   t5 = t4 - t5;
   t7 = t5;
   goto zig_block_2;
  }
  t7 = (uintptr_t)0ul;
  goto zig_block_2;

  zig_block_2:;
  t9 = t7;
  t3 = t9 == UINT64_C(0);
  if (t3) {
   t10 = (*t1);
   t11 = io_Writer_writeAll__5168(t10, a0);
   memcpy(&t12, &t11, sizeof(uint16_t));
   return t12;
  }
  goto zig_block_3;

  zig_block_3:;
  t14 = a1.fill;
  t15.ptr = &t13[(uintptr_t)0ul];
  t15.len = (uintptr_t)4ul;
  t16 = unicode_utf8Encode__5190(t14, t15);
  t3 = t16.error == UINT16_C(0);
  if (t3) {
   t18 = t16.payload;
   t19 = (uint8_t *)&t13;
   t19 = (uint8_t *)(((uintptr_t)t19) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t5 = (uintptr_t)t18;
   t15.ptr = t19;
   t15.len = t5;
   memcpy(&t20, &t15, sizeof(nav__5592_39));
   t17 = t20;
   goto zig_block_4;

  }
  t12 = t16.error;
  switch (t12) {
   case zig_error_Utf8CannotEncodeSurrogateHalf:
   case zig_error_CodepointTooLarge: {
    t17 = (nav__5592_39){(uint8_t const *)&__anon_5278,(uintptr_t)3ul};
    goto zig_block_4;
   }
   default: zig_unreachable();
  }

  zig_block_4:;
  t18 = a1.alignment;
  switch (t18) {
   case UINT8_C(0): {
    t10 = (*t1);
    t12 = io_Writer_writeAll__5168(t10, a0);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5172(t10, t17, t7);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   case UINT8_C(1): {
    t5 = t7 / (uintptr_t)2ul;
    t7 = t7 + (uintptr_t)1ul;
    t7 = t7 / (uintptr_t)2ul;
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5172(t10, t17, t5);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeAll__5168(t10, a0);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5172(t10, t17, t7);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   case UINT8_C(2): {
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5172(t10, t17, t7);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeAll__5168(t10, a0);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   default: zig_unreachable();
  }

  zig_block_6:;
  goto zig_block_0;
 }
 t10 = (*t1);
 t12 = io_Writer_writeAll__5168(t10, a0);
 if (t12) {
  return t12;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatInt__anon_5170__5602(uint64_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4248 const a3, struct io_Writer__3827 const a4) {
 uint64_t t4;
 uint64_t t11;
 uint64_t t2;
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__5602_58 t12;
 nav__5602_47 t13;
 uint16_t t14;
 uint16_t t15;
 bool t0;
 uint8_t t8;
 nav__5602_56 t9;
 uint8_t t10[2];
 uint8_t t1[65];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)65ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_19:
  t4 = t2;
  t0 = t4 >= UINT64_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT64_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__4953(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT64_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_19;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT64_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__4953(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_74:
 t4 = t2;
 t11 = (uint64_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__4969(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint64_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT64_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_74;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)65ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__5602_47));
 t14 = fmt_formatBuf__anon_5109__5592(t13, a3, a4);
 memcpy(&t15, &t14, sizeof(uint16_t));
 return t15;
}

static uint16_t fmt_formatIntValue__anon_5153__5601(uint64_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5170__5602(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5084__5591(uint64_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5153__5601(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__1553_38 array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1553(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, nav__1553_42 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 nav__1553_38 t4;
 uintptr_t t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1547(t2, a1);
 if (t3) {
  t4.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
  t4.error = t3;
  return t4;
 }
 t5 = a1.len;
 t4.payload = t5;
 t4.error = UINT16_C(0);
 return t4;
}

static uint16_t bincode_serializeInt__anon_5117__5593(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, int32_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t8;
 void const *t9;
 nav__5593_48 (**t10)(void const *, nav__5593_50);
 struct io_Writer__3827 t11;
 struct io_Writer__3827 t6;
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t14;
 struct io_Writer__3827 const *t13;
 nav__5593_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[4];
 uint8_t t15[4];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5593_48 (**)(void const *, nav__5593_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__3827 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__3827 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[4]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[4]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)4ul;
 t18 = io_Writer_writeAll__5168(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializePointer__anon_5122__5594(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, nav__5594_40 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t23;
 uintptr_t t3;
 uint64_t t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t8;
 void const **t10;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t11;
 void const *t12;
 nav__5594_51 (**t13)(void const *, nav__5594_40);
 struct io_Writer__3827 t14;
 struct io_Writer__3827 t9;
 struct io_Writer__3827 t15;
 struct io_Writer__3827 t17;
 struct io_Writer__3827 t25;
 struct io_Writer__3827 t26;
 struct io_Writer__3827 const *t16;
 nav__5594_40 t20;
 uint16_t t21;
 uint16_t t22;
 uint8_t t19[8];
 uint8_t t18[8];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = t3;
 t5 = t2;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t5;
 t7 = t6;
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t7;
 t10 = (void const **)&t9.context;
 t6 = (*t8);
 t11 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t6->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__5594_51 (**)(void const *, nav__5594_40))&t9.writeFn;
 (*t13) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t14 = t9;
 t15 = t14;
 t16 = (struct io_Writer__3827 const *)&t15;
 t14 = (*t16);
 t17 = t14;
 t16 = (struct io_Writer__3827 const *)&t17;
 memcpy(&t19, &t4, sizeof(uint8_t[8]));
 memcpy((char *)&t18, t19, sizeof(uint8_t[8]));
 t14 = (*t16);
 t20.ptr = &t18[(uintptr_t)0ul];
 t20.len = (uintptr_t)8ul;
 t21 = io_Writer_writeAll__5168(t14, t20);
 memcpy(&t22, &t21, sizeof(uint16_t));
 if (t22) {
  return t22;
 }
 t2 = (*t1);
 t23 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t23;
 t24 = t1;
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t24;
 t10 = (void const **)&t25.context;
 t1 = (*t8);
 t11 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__5594_51 (**)(void const *, nav__5594_40))&t25.writeFn;
 (*t13) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t14 = t25;
 t26 = t14;
 t16 = (struct io_Writer__3827 const *)&t26;
 t14 = (*t16);
 t22 = io_Writer_writeAll__5168(t14, a1);
 memcpy(&t21, &t22, sizeof(uint16_t));
 if (t21) {
  return t21;
 }
 return 0;
}

static uint16_t bincode_serializeUnion__anon_5123__5595(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a0, struct shared_MarkerType__1185 const a1) {
 uint64_t t4;
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 bool t3;
 t0 = a1.tag;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3165__4817(a0, t1);
 if (t2) {
  return t2;
 }
 t0 = a1.tag;
 t3 = t0 == UINT8_C(0);
 if (t3) {
  t4 = a1.payload.SampleStart;
  t2 = bincode_serialize__anon_3175__4822(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(1);
 if (t3) {
  t4 = a1.payload.SampleEnd;
  t2 = bincode_serialize__anon_3175__4822(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(2);
 if (t3) {
  t4 = a1.payload.BenchmarkStart;
  t2 = bincode_serialize__anon_3175__4822(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(3);
 if (t3) {
  t4 = a1.payload.BenchmarkEnd;
  t2 = bincode_serialize__anon_3175__4822(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 return 0;
}

static uint16_t fmt_format__anon_5127__5596(struct io_Writer__3827 const a0, nav__5596_41 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 nav__5596_40 t5;
 int32_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5596_40){(uint8_t const *)&__anon_5352,(uintptr_t)25ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5404__5603(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5596_40){(uint8_t const *)&__anon_5418,(uintptr_t)7ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5435__5604(t5, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5596_40){(uint8_t const *)&__anon_5444,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5131__5597(struct io_Writer__3827 const a0, nav__5597_41 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 nav__5597_40 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5597_40){(uint8_t const *)&__anon_5497,(uintptr_t)23ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5435__5604(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5597_40){(uint8_t const *)&__anon_5522,(uintptr_t)11ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f1;
 t3 = fmt_formatType__anon_5435__5604(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5597_40){(uint8_t const *)&__anon_5444,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5135__5598(struct io_Writer__3827 const a0, nav__5598_40 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 struct shared_MarkerType__1185 t5;
 int32_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5598_45){(uint8_t const *)&__anon_5578,(uintptr_t)17ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5404__5603(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5598_45){(uint8_t const *)&__anon_5600,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5610__5605(t5, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5598_45){(uint8_t const *)&__anon_5444,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5139__5599(struct io_Writer__3827 const a0, nav__5599_39 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5599_44){(uint8_t const *)&__anon_5670,(uintptr_t)31ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5679__5606(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5599_44){(uint8_t const *)&__anon_5444,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5143__5600(struct io_Writer__3827 const a0, nav__5600_39 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 uint8_t t4;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5600_44){(uint8_t const *)&__anon_5725,(uintptr_t)24ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5733__5607(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__4953_39 fmt_digits2__4953(uint8_t const a0) {
 uint8_t *t1;
 uint8_t t2;
 nav__4953_39 t3;
 uint8_t t0[2];
 t1 = (uint8_t *)&t0[(uintptr_t)0ul];
 t2 = a0 / UINT8_C(10);
 t2 = UINT8_C(48) + t2;
 (*t1) = t2;
 t1 = (uint8_t *)&t0[(uintptr_t)1ul];
 t2 = a0 % UINT8_C(10);
 t2 = UINT8_C(48) + t2;
 (*t1) = t2;
 memcpy(t3.array, t0, sizeof(uint8_t[2]));
 return t3;
}

static uint8_t fmt_digitToChar__4969(uint8_t const a0, uint8_t const a1) {
 uint8_t t0;
 uint8_t t1;
 bool t2;
 switch (a0) {
  default: if ((a0 >= UINT8_C(0) && a0 <= UINT8_C(9))) {
   t1 = a0 + UINT8_C(48);
   t0 = t1;
   goto zig_block_0;
  }if ((a0 >= UINT8_C(10) && a0 <= UINT8_C(35))) {
   t2 = a1 == UINT8_C(1);
   if (t2) {
    t1 = UINT8_C(65);
    goto zig_block_1;
   }
   t1 = UINT8_C(97);
   goto zig_block_1;

   zig_block_1:;
   t1 = t1 - UINT8_C(10);
   t1 = a0 + t1;
   t0 = t1;
   goto zig_block_0;
  }{
   zig_unreachable();
  }
 }

 zig_block_0:;
 return t0;
}

static nav__5195_38 unicode_utf8Decode__5195(nav__5195_40 const a0) {
 nav__5195_40 const *t1;
 uintptr_t t2;
 nav__5195_40 t7;
 nav__5195_40 t0;
 uint8_t const *t8;
 uint8_t const (*t9)[2];
 uint8_t const (*t13)[3];
 uint8_t const (*t16)[4];
 nav__5195_38 t3;
 nav__5195_38 t6;
 nav__5195_38 t12;
 uint32_t t5;
 uint8_t t4;
 uint8_t t10[2];
 nav__5195_48 t11;
 uint8_t t14[3];
 nav__5195_52 t15;
 uint8_t t17[4];
 nav__5195_56 t18;
 t0 = a0;
 t1 = (nav__5195_40 const *)&t0;
 t2 = a0.len;
 switch (t2) {
  case (uintptr_t)1ul: {
   t4 = a0.ptr[(uintptr_t)0ul];
   t5 = (uint32_t)t4;
   t6.payload = t5;
   t6.error = UINT16_C(0);
   t3 = t6;
   goto zig_block_0;
  }
  case (uintptr_t)2ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t9 = (uint8_t const (*)[2])t8;
   memcpy(t10, (const char *)t9, sizeof(uint8_t[2]));
   memcpy(t11.array, t10, sizeof(nav__5195_48));
   t6 = unicode_utf8Decode2__5197(t11);
   memcpy(&t12, &t6, sizeof(nav__5195_38));
   t3 = t12;
   goto zig_block_0;
  }
  case (uintptr_t)3ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t13 = (uint8_t const (*)[3])t8;
   memcpy(t14, (const char *)t13, sizeof(uint8_t[3]));
   memcpy(t15.array, t14, sizeof(nav__5195_52));
   t12 = unicode_utf8Decode3__5199(t15);
   memcpy(&t6, &t12, sizeof(nav__5195_38));
   t3 = t6;
   goto zig_block_0;
  }
  case (uintptr_t)4ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t16 = (uint8_t const (*)[4])t8;
   memcpy(t17, (const char *)t16, sizeof(uint8_t[4]));
   memcpy(t18.array, t17, sizeof(nav__5195_56));
   t12 = unicode_utf8Decode4__5203(t18);
   memcpy(&t6, &t12, sizeof(nav__5195_38));
   t3 = t6;
   goto zig_block_0;
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 return t3;
}

static nav__5188_38 unicode_utf8CodepointSequenceLength__5188(uint32_t const a0) {
 bool t0;
 t0 = a0 < UINT32_C(128);
 if (t0) {
  return (nav__5188_38){0,UINT8_C(1)};
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 < UINT32_C(2048);
 if (t0) {
  return (nav__5188_38){0,UINT8_C(2)};
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a0 < UINT32_C(65536);
 if (t0) {
  return (nav__5188_38){0,UINT8_C(3)};
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a0 < UINT32_C(1114112);
 if (t0) {
  return (nav__5188_38){0,UINT8_C(4)};
 }
 goto zig_block_3;

 zig_block_3:;
 return (nav__5188_38){zig_error_CodepointTooLarge,UINT8_C(0x2)};
}

static nav__5608_38 unicode_utf8EncodeImpl__anon_5756__5608(uint32_t const a0, nav__5608_40 const a1) {
 nav__5608_40 const *t1;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 nav__5608_40 t9;
 nav__5608_40 t0;
 uint8_t *t10;
 uint32_t t12;
 nav__5608_38 t2;
 uint16_t t3;
 uint8_t t4;
 uint8_t t11;
 bool t8;
 t0 = a1;
 t1 = (nav__5608_40 const *)&t0;
 t2 = unicode_utf8CodepointSequenceLength__5188(a0);
 if (t2.error) {
  t3 = t2.error;
  t2.payload = UINT8_C(0x2);
  t2.error = t3;
  return t2;
 }
 t4 = t2.payload;
 t5 = a1.len;
 t6 = t5;
 t7 = (uint64_t)t4;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 switch (t4) {
  case UINT8_C(1): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t11 = (uint8_t)a0;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = UINT32_C(192) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t8 = unicode_isSurrogateCodepoint__5268(a0);
   if (t8) {
    return (nav__5608_38){zig_error_Utf8CannotEncodeSurrogateHalf,UINT8_C(0x2)};
   }
   goto zig_block_1;

   zig_block_1:;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(12));
   t12 = UINT32_C(224) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)2ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(18));
   t12 = UINT32_C(240) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = zig_shr_u32(a0, UINT8_C(12));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)2ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)3ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__5190_38 unicode_utf8Encode__5190(uint32_t const a0, nav__5190_40 const a1) {
 nav__5190_38 t0;
 nav__5190_38 t1;
 t0 = unicode_utf8EncodeImpl__anon_5756__5608(a0, a1);
 memcpy(&t1, &t0, sizeof(nav__5190_38));
 return t1;
}

static uint16_t io_Writer_writeBytesNTimes__5172(struct io_Writer__3827 const a0, nav__5172_40 const a1, uintptr_t const a2) {
 struct io_Writer__3827 const *t1;
 uintptr_t t3;
 uintptr_t t2;
 uint64_t t4;
 uint64_t t5;
 struct io_Writer__3827 t7;
 struct io_Writer__3827 t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (uintptr_t)0ul;
 zig_loop_9:
 t3 = t2;
 t4 = t3;
 t5 = a2;
 t6 = t4 < t5;
 if (t6) {
  t7 = (*t1);
  t8 = io_Writer_writeAll__5168(t7, a1);
  if (t8) {
   return t8;
  }
  t3 = t2;
  t3 = t3 + (uintptr_t)1ul;
  t2 = t3;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_9;

 zig_block_0:;
 return 0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1547(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, nav__1547_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 uintptr_t t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1566(t2, t3);
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1548(t2, a1);
 return 0;
}

static uint16_t fmt_formatType__anon_5404__5603(int32_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct io_Writer__3827 t1;
 int32_t t0;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5781__5609(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_5435__5604(nav__5604_39 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct io_Writer__3827 const *t2;
 uint64_t t3;
 struct io_Writer__3827 t5;
 struct io_Writer__3827 t1;
 nav__5604_39 t0;
 uint16_t t6;
 uint16_t t7;
 bool t4;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__3827 const *)&t1;
 t3 = a3;
 t4 = t3 == UINT64_C(0);
 if (t4) {
  t5 = (*t2);
  t6 = io_Writer_writeAll__5168(t5, (nav__5604_39){(uint8_t const *)&__anon_5099,(uintptr_t)7ul});
  return t6;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = fmt_formatBuf__anon_5109__5592(a0, a1, a2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t shared_MarkerType_format__anon_5800__5610(struct shared_MarkerType__1185 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 struct io_Writer__3827 const *t1;
 uint64_t t3;
 struct io_Writer__3827 t4;
 struct io_Writer__3827 t0;
 nav__5610_56 t5;
 uint16_t t6;
 uint8_t t2;
 (void)a1;
 t0 = a2;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.SampleStart;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5814__5611(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t3 = a0.payload.SampleEnd;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5819__5612(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t3 = a0.payload.BenchmarkStart;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5826__5613(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t3 = a0.payload.BenchmarkEnd;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5832__5614(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatType__anon_5610__5605(struct shared_MarkerType__1185 const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct shared_MarkerType__1185 const *t1;
 struct shared_MarkerType__1185 t3;
 struct shared_MarkerType__1185 t0;
 struct io_Writer__3827 t2;
 uint16_t t4;
 (void)a3;
 t0 = a0;
 t1 = (struct shared_MarkerType__1185 const *)&t0;
 t2 = a2;
 t3 = (*t1);
 t4 = shared_MarkerType_format__anon_5800__5610(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t fmt_formatType__anon_5679__5606(uint64_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 uint64_t t0;
 struct io_Writer__3827 t1;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5834__5615(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_5733__5607(uint8_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct io_Writer__3827 const *t2;
 struct io_Writer__3827 t3;
 struct io_Writer__3827 t1;
 nav__5607_47 t5;
 nav__5607_47 t6;
 uint16_t t4;
 uint8_t t0;
 (void)a1;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__3827 const *)&t1;
 t3 = (*t2);
 t4 = io_Writer_writeAll__5168(t3, (nav__5607_47){(uint8_t const *)&__anon_5841,(uintptr_t)22ul});
 if (t4) {
  return t4;
 }
 t3 = (*t2);
 t4 = io_Writer_writeAll__5168(t3, (nav__5607_47){(uint8_t const *)&__anon_5846,(uintptr_t)1ul});
 if (t4) {
  return t4;
 }
 t3 = (*t2);
 t5 = zig_tag_name_shared_IntegrationMode__573(a0);
 memcpy(&t6, &t5, sizeof(nav__5607_47));
 t4 = io_Writer_writeAll__5168(t3, t6);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__5197_38 unicode_utf8Decode2__5197(nav__5197_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__5197_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(224);
 t1 = t0 == UINT8_C(192);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(31);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5197_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t4 = t2;
 t1 = t4 < UINT32_C(128);
 if (t1) {
  return (nav__5197_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t5.payload = t4;
 t5.error = UINT16_C(0);
 return t5;
}

static nav__5199_38 unicode_utf8Decode3__5199(nav__5199_40 const a0) {
 nav__5199_38 t1;
 uint32_t t3;
 uint16_t t2;
 nav__5199_40 t0;
 bool t4;
 bool t5;
 memcpy(t0.array, a0.array, sizeof(nav__5199_40));
 t1 = unicode_utf8Decode3AllowSurrogateHalf__5201(t0);
 if (t1.error) {
  t2 = t1.error;
  t1.payload = UINT32_C(0xaaaaa);
  t1.error = t2;
  return t1;
 }
 t3 = t1.payload;
 t4 = UINT32_C(55296) <= t3;
 if (t4) {
  t4 = t3 <= UINT32_C(57343);
  t5 = t4;
  goto zig_block_1;
 }
 t5 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t5) {
  return (nav__5199_38){UINT32_C(0xaaaaa),zig_error_Utf8EncodesSurrogateHalf};
 }
 goto zig_block_0;

 zig_block_0:;
 t1.payload = t3;
 t1.error = UINT16_C(0);
 return t1;
}

static nav__5203_38 unicode_utf8Decode4__5203(nav__5203_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__5203_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(248);
 t1 = t0 == UINT8_C(240);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(7);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5203_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5203_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t4 = zig_shlw_u32(t4, UINT8_C(6), UINT8_C(21));
 t2 = t4;
 t4 = t2;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(63);
 t3 = (uint32_t)t0;
 t3 = t4 | t3;
 t2 = t3;
 t0 = a0.array[(uintptr_t)3ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5203_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)3ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t4 = t2;
 t1 = t4 < UINT32_C(65536);
 if (t1) {
  return (nav__5203_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_3;

 zig_block_3:;
 t4 = t2;
 t1 = t4 > UINT32_C(1114111);
 if (t1) {
  return (nav__5203_38){UINT32_C(0xaaaaa),zig_error_Utf8CodepointTooLarge};
 }
 goto zig_block_4;

 zig_block_4:;
 t4 = t2;
 t5.payload = t4;
 t5.error = UINT16_C(0);
 return t5;
}

static bool unicode_isSurrogateCodepoint__5268(uint32_t const a0) {
 bool t0;
 switch (a0) {
  default: if ((a0 >= UINT32_C(55296) && a0 <= UINT32_C(57343))) {
   t0 = true;
   goto zig_block_0;
  }{
   t0 = false;
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return t0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1566(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 nav__1566_43 *t3;
 nav__1566_43 t4;
 uintptr_t t5;
 nav__1566_52 t6;
 uint16_t t7;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__1566_43 *)&a0->items;
 t4 = (*t3);
 t5 = t4.len;
 t6 = array_list_addOrOom__1491(t5, a1);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t5 = t6.payload;
 t7 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(t2, t5);
 return t7;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1548(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0, nav__1548_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 nav__1548_47 *t2;
 nav__1548_47 t3;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 uintptr_t *t6;
 uint64_t t8;
 uint64_t t9;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 uint8_t *t12;
 uint8_t const *t13;
 bool t10;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (nav__1548_47 *)&a0->items;
 t3 = (*t2);
 t4 = t3.len;
 t5 = a1.len;
 t5 = t4 + t5;
 t6 = (uintptr_t *)&a0->capacity;
 t7 = (*t6);
 t8 = t5;
 t9 = t7;
 t10 = t8 <= t9;
 debug_assert__180(t10);
 t11 = (*t1);
 t2 = (nav__1548_47 *)&t11->items;
 t6 = &t2->len;
 (*t6) = t5;
 t11 = (*t1);
 t2 = (nav__1548_47 *)&t11->items;
 t5 = a1.len;
 t3 = (*t2);
 t12 = t3.ptr;
 t12 = (uint8_t *)(((uintptr_t)t12) + (t4*sizeof(uint8_t)));
 t3.ptr = t12;
 t3.len = t5;
 t13 = a1.ptr;
 if (t3.len != 0) memcpy(t3.ptr, t13, t3.len * sizeof(uint8_t));
 return;
}

static uint16_t fmt_formatInt__anon_5888__5617(int32_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4248 const a3, struct io_Writer__3827 const a4) {
 uintptr_t t5;
 uintptr_t t4;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__5617_41 t12;
 uint64_t t14;
 nav__5617_58 t15;
 nav__5617_47 t16;
 uint32_t t1;
 uint32_t t11;
 uint32_t t3;
 uint16_t t17;
 uint16_t t18;
 bool t0;
 bool t13;
 uint8_t t8;
 nav__5617_56 t9;
 uint8_t t10[2];
 uint8_t t2[33];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t1 = zig_abs_i32(a0);
 t3 = t1;
 t4 = (uintptr_t)33ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t1 = t3;
  t0 = t1 >= UINT32_C(100);
  if (t0) {
   t5 = t4;
   t5 = t5 - (uintptr_t)2ul;
   t4 = t5;
   t5 = t4;
   t6 = (uint8_t *)&t2;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t1 = t3;
   t1 = t1 % UINT32_C(100);
   t8 = (uint8_t)t1;
   t9 = fmt_digits2__4953(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t1 = t3;
   t1 = t1 / UINT32_C(100);
   t3 = t1;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t1 = t3;
  t0 = t1 < UINT32_C(10);
  if (t0) {
   t5 = t4;
   t5 = t5 - (uintptr_t)1ul;
   t4 = t5;
   t5 = t4;
   t6 = (uint8_t *)&t2[t5];
   t1 = t3;
   t8 = (uint8_t)t1;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t4;
  t5 = t5 - (uintptr_t)2ul;
  t4 = t5;
  t5 = t4;
  t6 = (uint8_t *)&t2;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t1 = t3;
  t8 = (uint8_t)t1;
  t9 = fmt_digits2__4953(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t1 = t3;
 t11 = (uint32_t)a1;
 t11 = t1 % t11;
 t5 = t4;
 t5 = t5 - (uintptr_t)1ul;
 t4 = t5;
 t5 = t4;
 t6 = (uint8_t *)&t2[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__4969(t8, a2);
 (*t6) = t8;
 t11 = t3;
 t1 = (uint32_t)a1;
 t1 = t11 / t1;
 t3 = t1;
 t1 = t3;
 t0 = t1 == UINT32_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 < INT32_C(0);
 if (t0) {
  t5 = t4;
  t5 = t5 - (uintptr_t)1ul;
  t4 = t5;
  t5 = t4;
  t6 = (uint8_t *)&t2[t5];
  (*t6) = UINT8_C(45);
  goto zig_block_6;
 }
 t12 = a3.width;
 t0 = t12.is_null == true;
 if (t0) {
  t13 = true;
  goto zig_block_8;
 }
 t12 = a3.width;
 t5 = t12.payload;
 t14 = t5;
 t0 = t14 == UINT64_C(0);
 t13 = t0;
 goto zig_block_8;

 zig_block_8:;
 if (t13) {
  goto zig_block_7;
 }
 t5 = t4;
 t5 = t5 - (uintptr_t)1ul;
 t4 = t5;
 t5 = t4;
 t6 = (uint8_t *)&t2[t5];
 (*t6) = UINT8_C(43);
 goto zig_block_7;

 zig_block_7:;
 goto zig_block_6;

 zig_block_6:;
 t5 = t4;
 t6 = (uint8_t *)&t2;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)33ul - t5;
 t15.ptr = t6;
 t15.len = t5;
 memcpy(&t16, &t15, sizeof(nav__5617_47));
 t17 = fmt_formatBuf__anon_5109__5592(t16, a3, a4);
 memcpy(&t18, &t17, sizeof(uint16_t));
 return t18;
}

static uint16_t fmt_formatIntValue__anon_5877__5616(int32_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5888__5617(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5781__5609(int32_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5877__5616(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5814__5611(struct io_Writer__3827 const a0, nav__5611_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5899__5618(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5819__5612(struct io_Writer__3827 const a0, nav__5612_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5903__5619(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5826__5613(struct io_Writer__3827 const a0, nav__5613_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5907__5620(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5832__5614(struct io_Writer__3827 const a0, nav__5614_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5911__5621(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatIntValue__anon_5915__5622(uint64_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5170__5602(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5834__5615(uint64_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5915__5622(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__5201_38 unicode_utf8Decode3AllowSurrogateHalf__5201(nav__5201_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__5201_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(240);
 t1 = t0 == UINT8_C(224);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(15);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5201_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5201_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t4 = zig_shlw_u32(t4, UINT8_C(6), UINT8_C(21));
 t2 = t4;
 t4 = t2;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(63);
 t3 = (uint32_t)t0;
 t3 = t4 | t3;
 t2 = t3;
 t3 = t2;
 t1 = t3 < UINT32_C(2048);
 if (t1) {
  return (nav__5201_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t5.payload = t3;
 t5.error = UINT16_C(0);
 return t5;
}

static nav__1491_38 array_list_addOrOom__1491(uintptr_t const a0, uintptr_t const a1) {
 nav__1491_42 t2;
 uintptr_t t3;
 uintptr_t t0;
 uintptr_t const *t5;
 uint8_t const *t6;
 nav__1491_38 t8;
 uint8_t t4;
 uint8_t t1;
 bool t7;
 t2.f1 = zig_addo_u64(&t2.f0, a0, a1, UINT8_C(64));
 t3 = t2.f0;
 t0 = t3;
 t4 = t2.f1;
 t1 = t4;
 t5 = (uintptr_t const *)&t0;
 t6 = (uint8_t const *)&t1;
 t4 = (*t6);
 t7 = t4 != UINT8_C(0);
 if (t7) {
  return (nav__1491_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_OutOfMemory};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t5);
 t8.payload = t3;
 t8.error = UINT16_C(0);
 return t8;
}

static uint16_t fmt_format__anon_5899__5618(struct io_Writer__3827 const a0, nav__5618_39 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5618_44){(uint8_t const *)&__anon_5938,(uintptr_t)12ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5679__5606(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5618_44){(uint8_t const *)&__anon_5948,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5903__5619(struct io_Writer__3827 const a0, nav__5619_39 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5619_44){(uint8_t const *)&__anon_5967,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5679__5606(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5619_44){(uint8_t const *)&__anon_5948,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5907__5620(struct io_Writer__3827 const a0, nav__5620_39 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5620_44){(uint8_t const *)&__anon_5999,(uintptr_t)15ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5679__5606(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5620_44){(uint8_t const *)&__anon_5948,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5911__5621(struct io_Writer__3827 const a0, nav__5621_39 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5621_44){(uint8_t const *)&__anon_6027,(uintptr_t)13ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5679__5606(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5621_44){(uint8_t const *)&__anon_5948,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

void c_instrument_hooks_deinit__239(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t4;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 bool t0;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t4 = (*t3);
  instrument_hooks_InstrumentHooks_deinit__731(t4);
  mem_Allocator_destroy__anon_6049__5623((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)}, t1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void instrument_hooks_InstrumentHooks_deinit__731(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *const *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *t0;
 struct instruments_root_Instrument__554 *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *const *t5;
 struct instruments_root_Instrument__554 t6;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t9;
 struct environment_root_Environment__559 *t10;
 uint8_t t7;
 t0 = a0;
 t1 = (struct InstrumentHooks *const *)&t0;
 t2 = (*t1);
 t3 = (struct instruments_root_Instrument__554 *)&t2->instrument;
 t4 = t3;
 t5 = (struct instruments_root_Instrument__554 *const *)&t4;
 t6 = (*t3);
 t7 = t6.tag;
 switch (t7) {
  case UINT8_C(0): {
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t3 = (*t5);
   t8 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t3->payload.walltime;
   instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_deinit__808(t8);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t3 = (*t5);
   t9 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t3->payload.analysis;
   instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__821(t9);
   goto zig_block_0;
  }
  case UINT8_C(3): {
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t2 = (*t1);
 t10 = (struct environment_root_Environment__559 *)&t2->environment;
 environment_root_Environment_deinit__774(t10);
 return;
}

static void mem_Allocator_destroy__anon_6049__5623(struct mem_Allocator__599 const a0, struct InstrumentHooks *const a1) {
 struct mem_Allocator__599 const *t1;
 uint8_t *t2;
 uint8_t *t3;
 uint8_t *const *t4;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t10;
 uint8_t (*t6)[288];
 nav__5623_52 t7;
 uintptr_t t9;
 struct mem_Allocator_VTable__602 const *const *t11;
 struct mem_Allocator_VTable__602 const *t12;
 void (*const *t13)(void *, nav__5623_52, uint8_t, uintptr_t);
 void (*t14)(void *, nav__5623_52, uint8_t, uintptr_t);
 void *t15;
 uint8_t t8;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (uint8_t *)a1;
 t3 = t2;
 t4 = (uint8_t *const *)&t3;
 t5 = (*t1);
 t2 = (*t4);
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t6 = (uint8_t (*)[288])t2;
 t7.ptr = &(*t6)[(uintptr_t)0ul];
 t7.len = (uintptr_t)288ul;
 t8 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t9 = (uintptr_t)zig_return_address();
 t10 = t5;
 t1 = (struct mem_Allocator__599 const *)&t10;
 t11 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t12 = (*t11);
 t13 = (void (*const *)(void *, nav__5623_52, uint8_t, uintptr_t))&t12->free;
 t14 = (*t13);
 t15 = t5.ptr;
 t14(t15, t7, t8, t9);
 return;
}

static void instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_deinit__808(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 runner_fifo_RunnerFifo_deinit__1131(t3);
 return;
}

static void instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__821(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 runner_fifo_RunnerFifo_deinit__1131(t3);
 return;
}

static void environment_root_Environment_deinit__774(struct environment_root_Environment__559 *const a0) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t3;
 struct environment_root_Environment__559 *t0;
 struct environment_root_EnvironmentJson__833 *t4;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *t5;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 t2;
 nav__774_87 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 t12;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 const *t13;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t15;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t17;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t18;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 t19;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 t14;
 nav__774_104 t20;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6115 t21;
 struct mem_Allocator__599 *t22;
 struct mem_Allocator__599 t23;
 nav__774_44 *t24;
 nav__774_44 t25;
 struct environment_root_EnvironmentValue__844 *t26;
 struct environment_root_EnvironmentValue__844 t27;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *t29;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t28;
 nav__774_122 t33;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147 t34;
 struct environment_linked_libraries_root_LinkedLibraries__837 *t35;
 bool t10;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *)&t4->integration_environment;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *)&t5->map;
 t7 = (*t6);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2965(t7);
 t2 = t8;
 zig_loop_14:
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5624(&t2);
 t10 = t9.is_null != true;
 if (t10) {
  t11 = t9.payload;
  t12 = t11;
  t13 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 const *)&t12;
  t15 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *)&t13->value_ptr;
  t16 = (*t15);
  t17 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *)&t16->map;
  t18 = (*t17);
  t19 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3176(t18);
  t14 = t19;
  zig_loop_32:
  t20 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5626(&t14);
  t10 = t20.is_null != true;
  if (t10) {
   t21 = t20.payload;
   t3 = (*t1);
   t22 = (struct mem_Allocator__599 *)&t3->allocator;
   t23 = (*t22);
   t24 = t21.key_ptr;
   t25 = (*t24);
   mem_Allocator_free__anon_3042__4747(t23, t25);
   t3 = (*t1);
   t26 = t21.value_ptr;
   t27 = (*t26);
   environment_root_Environment_freeEnvironmentValue__777(t3, t27);
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_32;

  zig_block_2:;
  t15 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *)&t13->value_ptr;
  t16 = (*t15);
  t17 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *)&t16->map;
  t22 = (struct mem_Allocator__599 *)&a0->allocator;
  t23 = (*t22);
  array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3168(t17, t23);
  t3 = (*t1);
  t22 = (struct mem_Allocator__599 *)&t3->allocator;
  t23 = (*t22);
  t24 = t11.key_ptr;
  t25 = (*t24);
  mem_Allocator_free__anon_3042__4747(t23, t25);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_14;

 zig_block_0:;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *)&t4->integration_environment;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *)&t5->map;
 t22 = (struct mem_Allocator__599 *)&a0->allocator;
 t23 = (*t22);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2957(t6, t23);
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
 t29 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *)&t4->linked_libraries;
 t30 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *)&t29->map;
 t31 = (*t30);
 t32 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3355(t31);
 t28 = t32;
 zig_loop_91:
 t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__5628(&t28);
 t10 = t33.is_null != true;
 if (t10) {
  t34 = t33.payload;
  t3 = (*t1);
  t22 = (struct mem_Allocator__599 *)&t3->allocator;
  t23 = (*t22);
  t24 = t34.key_ptr;
  t25 = (*t24);
  mem_Allocator_free__anon_3042__4747(t23, t25);
  goto zig_block_5;
 }
 goto zig_block_4;

 zig_block_5:;
 goto zig_loop_91;

 zig_block_4:;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
 t29 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *)&t4->linked_libraries;
 t30 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *)&t29->map;
 t22 = (struct mem_Allocator__599 *)&a0->allocator;
 t23 = (*t22);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3347(t30, t23);
 t3 = (*t1);
 t35 = (struct environment_linked_libraries_root_LinkedLibraries__837 *)&t3->libs;
 environment_linked_libraries_root_LinkedLibraries_deinit__2809(t35);
 return;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2965(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t6;
 nav__2965_42 **t8;
 nav__2965_62 t10;
 nav__2965_42 *t11;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **t12;
 nav__2965_65 t13;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t14;
 uint32_t *t15;
 uintptr_t t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 t7;
 uint32_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *)&t1->entries;
 t3 = (*t2);
 t4 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t3);
 t5 = t4;
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t5;
 t8 = (nav__2965_42 **)&t7.keys;
 t9 = (*t6);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t9);
 t11 = t10.ptr;
 (*t8) = t11;
 t12 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t7.values;
 t9 = (*t6);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t9);
 t14 = t13.ptr;
 (*t12) = t14;
 t15 = (uint32_t *)&t7.len;
 t16 = t4.len;
 t17 = (uint32_t)t16;
 (*t15) = t17;
 t15 = (uint32_t *)&t7.index;
 (*t15) = UINT32_C(0);
 return t7;
}

static nav__5624_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5624(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 *const *t1;
 uint32_t *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 *t0;
 nav__5624_44 **t7;
 uintptr_t t8;
 nav__5624_44 *t9;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **t10;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 t12;
 nav__5624_39 t13;
 uint32_t t3;
 uint32_t t4;
 bool t5;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 *const *)&t0;
 t2 = (uint32_t *)&a0->index;
 t3 = (*t2);
 t2 = (uint32_t *)&a0->len;
 t4 = (*t2);
 t5 = t3 >= t4;
 if (t5) {
  return (nav__5624_39){{((nav__5624_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},true};
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = (*t1);
 t7 = (nav__5624_44 **)&t6->keys;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t9 = (*t7);
 t9 = (nav__5624_44 *)&t9[t8];
 t6 = (*t1);
 t10 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t6->values;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t11 = (*t10);
 t11 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t11[t8];
 t12.key_ptr = t9;
 t12.value_ptr = t11;
 t6 = (*t1);
 t2 = (uint32_t *)&t6->index;
 t4 = (*t2);
 t4 = t4 + UINT32_C(1);
 (*t2) = t4;
 t13.is_null = false;
 t13.payload = t12;
 return t13;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3176(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t6;
 nav__3176_42 **t8;
 nav__3176_62 t10;
 nav__3176_42 *t11;
 struct environment_root_EnvironmentValue__844 **t12;
 nav__3176_65 t13;
 struct environment_root_EnvironmentValue__844 *t14;
 uint32_t *t15;
 uintptr_t t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 t7;
 uint32_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *)&t1->entries;
 t3 = (*t2);
 t4 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t3);
 t5 = t4;
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t5;
 t8 = (nav__3176_42 **)&t7.keys;
 t9 = (*t6);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t9);
 t11 = t10.ptr;
 (*t8) = t11;
 t12 = (struct environment_root_EnvironmentValue__844 **)&t7.values;
 t9 = (*t6);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t9);
 t14 = t13.ptr;
 (*t12) = t14;
 t15 = (uint32_t *)&t7.len;
 t16 = t4.len;
 t17 = (uint32_t)t16;
 (*t15) = t17;
 t15 = (uint32_t *)&t7.index;
 (*t15) = UINT32_C(0);
 return t7;
}

static nav__5626_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5626(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 *const *t1;
 uint32_t *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 *t0;
 nav__5626_44 **t7;
 uintptr_t t8;
 nav__5626_44 *t9;
 struct environment_root_EnvironmentValue__844 **t10;
 struct environment_root_EnvironmentValue__844 *t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6115 t12;
 nav__5626_39 t13;
 uint32_t t3;
 uint32_t t4;
 bool t5;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 *const *)&t0;
 t2 = (uint32_t *)&a0->index;
 t3 = (*t2);
 t2 = (uint32_t *)&a0->len;
 t4 = (*t2);
 t5 = t3 >= t4;
 if (t5) {
  return (nav__5626_39){{((nav__5626_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_root_EnvironmentValue__844 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},true};
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = (*t1);
 t7 = (nav__5626_44 **)&t6->keys;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t9 = (*t7);
 t9 = (nav__5626_44 *)&t9[t8];
 t6 = (*t1);
 t10 = (struct environment_root_EnvironmentValue__844 **)&t6->values;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t11 = (*t10);
 t11 = (struct environment_root_EnvironmentValue__844 *)&t11[t8];
 t12.key_ptr = t9;
 t12.value_ptr = t11;
 t6 = (*t1);
 t2 = (uint32_t *)&t6->index;
 t4 = (*t2);
 t4 = t4 + UINT32_C(1);
 (*t2) = t4;
 t13.is_null = false;
 t13.payload = t12;
 return t13;
}

static void environment_root_Environment_freeEnvironmentValue__777(struct environment_root_Environment__559 *const a0, struct environment_root_EnvironmentValue__844 const a1) {
 struct environment_root_Environment__559 *const *t1;
 nav__777_43 t3;
 struct environment_root_Environment__559 *t4;
 struct environment_root_Environment__559 *t0;
 struct mem_Allocator__599 *t5;
 struct mem_Allocator__599 t6;
 nav__777_46 t7;
 uintptr_t t9;
 uintptr_t t10;
 uintptr_t t8;
 uint64_t t11;
 uint64_t t12;
 uint8_t t2;
 bool t13;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = a1.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a1.payload.string;
   t4 = (*t1);
   t5 = (struct mem_Allocator__599 *)&t4->allocator;
   t6 = (*t5);
   mem_Allocator_free__anon_3042__4747(t6, t3);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t7 = a1.payload.list;
   t8 = (uintptr_t)0ul;
   t9 = t7.len;
   zig_loop_20:
   t10 = t8;
   t11 = t10;
   t12 = t9;
   t13 = t11 < t12;
   if (t13) {
    t3 = t7.ptr[t10];
    t4 = (*t1);
    t5 = (struct mem_Allocator__599 *)&t4->allocator;
    t6 = (*t5);
    mem_Allocator_free__anon_3042__4747(t6, t3);
    goto zig_block_2;
   }
   goto zig_block_1;

   zig_block_2:;
   t10 = t10 + (uintptr_t)1ul;
   t8 = t10;
   goto zig_loop_20;

   zig_block_1:;
   t4 = (*t1);
   t5 = (struct mem_Allocator__599 *)&t4->allocator;
   t6 = (*t5);
   mem_Allocator_free__anon_6360__5654(t6, t7);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3168(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t3;
 struct array_hash_map_IndexHeader__880 **t4;
 struct array_hash_map_IndexHeader__880 *t5;
 struct array_hash_map_IndexHeader__880 *t7;
 struct array_hash_map_IndexHeader__880 *t8;
 struct array_hash_map_IndexHeader__880 *const *t9;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 debug_SafetyLock_assertUnlocked__3149();
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t2->entries;
 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3294(t3, a1);
 t4 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t8 = t7;
  t9 = (struct array_hash_map_IndexHeader__880 *const *)&t8;
  t7 = (*t9);
  array_hash_map_IndexHeader_free__3144(t7, a1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2957(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t3;
 struct array_hash_map_IndexHeader__880 **t4;
 struct array_hash_map_IndexHeader__880 *t5;
 struct array_hash_map_IndexHeader__880 *t7;
 struct array_hash_map_IndexHeader__880 *t8;
 struct array_hash_map_IndexHeader__880 *const *t9;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 debug_SafetyLock_assertUnlocked__3149();
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t2->entries;
 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3100(t3, a1);
 t4 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t8 = t7;
  t9 = (struct array_hash_map_IndexHeader__880 *const *)&t8;
  t7 = (*t9);
  array_hash_map_IndexHeader_free__3144(t7, a1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3355(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t6;
 nav__3355_42 **t8;
 nav__3355_62 t10;
 nav__3355_42 *t11;
 struct environment_linked_libraries_root_LibraryEntry__852 **t12;
 nav__3355_65 t13;
 struct environment_linked_libraries_root_LibraryEntry__852 *t14;
 uint32_t *t15;
 uintptr_t t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t7;
 uint32_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *)&t1->entries;
 t3 = (*t2);
 t4 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t3);
 t5 = t4;
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t5;
 t8 = (nav__3355_42 **)&t7.keys;
 t9 = (*t6);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t9);
 t11 = t10.ptr;
 (*t8) = t11;
 t12 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t7.values;
 t9 = (*t6);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t9);
 t14 = t13.ptr;
 (*t12) = t14;
 t15 = (uint32_t *)&t7.len;
 t16 = t4.len;
 t17 = (uint32_t)t16;
 (*t15) = t17;
 t15 = (uint32_t *)&t7.index;
 (*t15) = UINT32_C(0);
 return t7;
}

static nav__5628_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__5628(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 *const *t1;
 uint32_t *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 *t0;
 nav__5628_44 **t7;
 uintptr_t t8;
 nav__5628_44 *t9;
 struct environment_linked_libraries_root_LibraryEntry__852 **t10;
 struct environment_linked_libraries_root_LibraryEntry__852 *t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147 t12;
 nav__5628_39 t13;
 uint32_t t3;
 uint32_t t4;
 bool t5;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 *const *)&t0;
 t2 = (uint32_t *)&a0->index;
 t3 = (*t2);
 t2 = (uint32_t *)&a0->len;
 t4 = (*t2);
 t5 = t3 >= t4;
 if (t5) {
  return (nav__5628_39){{((nav__5628_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_linked_libraries_root_LibraryEntry__852 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},true};
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = (*t1);
 t7 = (nav__5628_44 **)&t6->keys;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t9 = (*t7);
 t9 = (nav__5628_44 *)&t9[t8];
 t6 = (*t1);
 t10 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t6->values;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t11 = (*t10);
 t11 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t11[t8];
 t12.key_ptr = t9;
 t12.value_ptr = t11;
 t6 = (*t1);
 t2 = (uint32_t *)&t6->index;
 t4 = (*t2);
 t4 = t4 + UINT32_C(1);
 (*t2) = t4;
 t13.is_null = false;
 t13.payload = t12;
 return t13;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3347(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t3;
 struct array_hash_map_IndexHeader__880 **t4;
 struct array_hash_map_IndexHeader__880 *t5;
 struct array_hash_map_IndexHeader__880 *t7;
 struct array_hash_map_IndexHeader__880 *t8;
 struct array_hash_map_IndexHeader__880 *const *t9;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 debug_SafetyLock_assertUnlocked__3149();
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t2->entries;
 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3473(t3, a1);
 t4 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t8 = t7;
  t9 = (struct array_hash_map_IndexHeader__880 *const *)&t8;
  t7 = (*t9);
  array_hash_map_IndexHeader_free__3144(t7, a1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void environment_linked_libraries_root_LinkedLibraries_deinit__2809(struct environment_linked_libraries_root_LinkedLibraries__837 *const a0) {
 struct environment_linked_libraries_root_LinkedLibraries__837 *const *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 *t3;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t4;
 nav__2809_50 t5;
 uintptr_t t6;
 uintptr_t t7;
 uintptr_t t2;
 uint64_t t8;
 uint64_t t9;
 struct environment_linked_libraries_root_LibraryInfo__934 t11;
 struct mem_Allocator__599 *t12;
 struct mem_Allocator__599 t13;
 struct environment_linked_libraries_root_LinkedLibraries__837 *t14;
 struct environment_linked_libraries_root_LinkedLibraries__837 *t0;
 bool t10;
 t0 = a0;
 t1 = (struct environment_linked_libraries_root_LinkedLibraries__837 *const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 *)&a0->libraries;
 t4 = (*t3);
 t5 = t4.items;
 t6 = t5.len;
 zig_loop_11:
 t7 = t2;
 t8 = t7;
 t9 = t6;
 t10 = t8 < t9;
 if (t10) {
  t11 = t5.ptr[t7];
  t12 = (struct mem_Allocator__599 *)&a0->allocator;
  t13 = (*t12);
  environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__2810(t13, t11);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t7 = t7 + (uintptr_t)1ul;
 t2 = t7;
 goto zig_loop_11;

 zig_block_0:;
 t14 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 *)&t14->libraries;
 t4 = (*t3);
 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_deinit__3514(t4);
 return;
}

static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const a0) {
 static uintptr_t const t10[3] = {16ul,32ul,4ul};
 static uintptr_t const t12[3] = {1ul,2ul,0ul};
 uint8_t *(*t1)[3];
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t11;
 uintptr_t t13;
 uintptr_t t7;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *t4;
 uint64_t t8;
 uint8_t **t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t0;
 bool t9;
 t1 = (uint8_t *(*)[3])&t0.ptrs;
 t2 = (uintptr_t *)&t0.len;
 t3 = a0.len;
 (*t2) = t3;
 t2 = (uintptr_t *)&t0.capacity;
 t3 = a0.capacity;
 (*t2) = t3;
 t5 = a0.bytes;
 t6 = (uint8_t *)t5;
 t4 = t6;
 t7 = (uintptr_t)0ul;
 zig_loop_17:
 t3 = t7;
 t8 = t3;
 t9 = t8 < UINT64_C(3);
 if (t9) {
  t11 = t10[t3];
  t13 = t12[t3];
  t1 = (uint8_t *(*)[3])&t0.ptrs;
  t14 = (uint8_t **)&(*t1)[t13];
  t6 = t4;
  (*t14) = t6;
  t6 = t4;
  t13 = a0.capacity;
  t13 = t11 * t13;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t13*sizeof(uint8_t)));
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t7 = t3;
 goto zig_loop_17;

 zig_block_0:;
 t15 = t0;
 return t15;
}

static nav__5643_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 nav__5643_39 *t5;
 nav__5643_39 *t6;
 nav__5643_39 *const *t7;
 nav__5643_41 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5643_41){(nav__5643_39 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)1ul];
 t5 = (nav__5643_39 *)t4;
 t6 = t5;
 t7 = (nav__5643_39 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (nav__5643_39 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(nav__5643_39)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__5644_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t5;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t6;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t7;
 nav__5644_40 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5644_40){(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)2ul];
 t5 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)t4;
 t6 = t5;
 t7 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const a0) {
 static uintptr_t const t10[3] = {16ul,24ul,4ul};
 static uintptr_t const t12[3] = {1ul,2ul,0ul};
 uint8_t *(*t1)[3];
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t11;
 uintptr_t t13;
 uintptr_t t7;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *t4;
 uint64_t t8;
 uint8_t **t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t0;
 bool t9;
 t1 = (uint8_t *(*)[3])&t0.ptrs;
 t2 = (uintptr_t *)&t0.len;
 t3 = a0.len;
 (*t2) = t3;
 t2 = (uintptr_t *)&t0.capacity;
 t3 = a0.capacity;
 (*t2) = t3;
 t5 = a0.bytes;
 t6 = (uint8_t *)t5;
 t4 = t6;
 t7 = (uintptr_t)0ul;
 zig_loop_17:
 t3 = t7;
 t8 = t3;
 t9 = t8 < UINT64_C(3);
 if (t9) {
  t11 = t10[t3];
  t13 = t12[t3];
  t1 = (uint8_t *(*)[3])&t0.ptrs;
  t14 = (uint8_t **)&(*t1)[t13];
  t6 = t4;
  (*t14) = t6;
  t6 = t4;
  t13 = a0.capacity;
  t13 = t11 * t13;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t13*sizeof(uint8_t)));
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t7 = t3;
 goto zig_loop_17;

 zig_block_0:;
 t15 = t0;
 return t15;
}

static nav__5652_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 nav__5652_39 *t5;
 nav__5652_39 *t6;
 nav__5652_39 *const *t7;
 nav__5652_41 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5652_41){(nav__5652_39 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)1ul];
 t5 = (nav__5652_39 *)t4;
 t6 = t5;
 t7 = (nav__5652_39 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (nav__5652_39 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(nav__5652_39)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__5653_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 struct environment_root_EnvironmentValue__844 *t5;
 struct environment_root_EnvironmentValue__844 *t6;
 struct environment_root_EnvironmentValue__844 *const *t7;
 nav__5653_40 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5653_40){(struct environment_root_EnvironmentValue__844 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)2ul];
 t5 = (struct environment_root_EnvironmentValue__844 *)t4;
 t6 = t5;
 t7 = (struct environment_root_EnvironmentValue__844 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (struct environment_root_EnvironmentValue__844 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_root_EnvironmentValue__844)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static void mem_Allocator_free__anon_6360__5654(struct mem_Allocator__599 const a0, nav__5654_42 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5654_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t const *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 nav__5654_54 t10;
 nav__5654_54 t12;
 struct mem_Allocator__599 t11;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t14;
 struct mem_Allocator_VTable__602 const *const *t15;
 struct mem_Allocator_VTable__602 const *t16;
 void (*const *t17)(void *, nav__5654_54, uint8_t, uintptr_t);
 void (*t18)(void *, nav__5654_54, uint8_t, uintptr_t);
 void *t19;
 bool t5;
 uint8_t t13;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6577__5735(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t11 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 memcpy(&t12, &t10, sizeof(nav__5654_54));
 t13 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t14 = t11;
 t1 = (struct mem_Allocator__599 const *)&t14;
 t15 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t16 = (*t15);
 t17 = (void (*const *)(void *, nav__5654_54, uint8_t, uintptr_t))&t16->free;
 t18 = (*t17);
 t19 = t11.ptr;
 t18(t19, t12, t13, t3);
 return;
}

static void debug_SafetyLock_assertUnlocked__3149(void) {
 return;
}

static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3294(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const a0, struct mem_Allocator__599 const a1) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *t1;
 struct mem_Allocator__599 const *t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 nav__3294_51 t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t3);
 t5 = (*t1);
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3326(t6);
 mem_Allocator_free__anon_6588__5736(t4, t7);
 return;
}

static void array_hash_map_IndexHeader_free__3144(struct array_hash_map_IndexHeader__880 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_IndexHeader__880 *const *t1;
 struct mem_Allocator__599 const *t3;
 uint8_t *t4;
 uint8_t *t7;
 uintptr_t t6;
 uintptr_t t11;
 uint8_t *const *t8;
 struct array_hash_map_IndexHeader__880 *t9;
 struct array_hash_map_IndexHeader__880 *t0;
 nav__3144_52 t12;
 struct mem_Allocator__599 t13;
 struct mem_Allocator__599 t2;
 struct array_hash_map_IndexHeader__880 t10;
 uint8_t t5;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__880 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uint8_t *)&a0->bit_index;
 t5 = (*t4);
 t6 = array_hash_map_capacityIndexSize__2907(t5);
 t4 = (uint8_t *)a0;
 t7 = t4;
 t8 = (uint8_t *const *)&t7;
 t9 = (*t1);
 t10 = (*t9);
 t11 = array_hash_map_IndexHeader_length__3140(t10);
 t6 = t11 * t6;
 t6 = (uintptr_t)4ul + t6;
 t4 = (*t8);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t4;
 t12.len = t6;
 t13 = (*t3);
 mem_Allocator_free__anon_6606__5737(t13, t12);
 return;
}

static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3100(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const a0, struct mem_Allocator__599 const a1) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *t1;
 struct mem_Allocator__599 const *t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 nav__3100_51 t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t3);
 t5 = (*t1);
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3132(t6);
 mem_Allocator_free__anon_6588__5736(t4, t7);
 return;
}

static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const a0) {
 static uintptr_t const t10[3] = {16ul,32ul,4ul};
 static uintptr_t const t12[3] = {1ul,2ul,0ul};
 uint8_t *(*t1)[3];
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t11;
 uintptr_t t13;
 uintptr_t t7;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *t4;
 uint64_t t8;
 uint8_t **t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t0;
 bool t9;
 t1 = (uint8_t *(*)[3])&t0.ptrs;
 t2 = (uintptr_t *)&t0.len;
 t3 = a0.len;
 (*t2) = t3;
 t2 = (uintptr_t *)&t0.capacity;
 t3 = a0.capacity;
 (*t2) = t3;
 t5 = a0.bytes;
 t6 = (uint8_t *)t5;
 t4 = t6;
 t7 = (uintptr_t)0ul;
 zig_loop_17:
 t3 = t7;
 t8 = t3;
 t9 = t8 < UINT64_C(3);
 if (t9) {
  t11 = t10[t3];
  t13 = t12[t3];
  t1 = (uint8_t *(*)[3])&t0.ptrs;
  t14 = (uint8_t **)&(*t1)[t13];
  t6 = t4;
  (*t14) = t6;
  t6 = t4;
  t13 = a0.capacity;
  t13 = t11 * t13;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t13*sizeof(uint8_t)));
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t7 = t3;
 goto zig_loop_17;

 zig_block_0:;
 t15 = t0;
 return t15;
}

static nav__5662_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 nav__5662_39 *t5;
 nav__5662_39 *t6;
 nav__5662_39 *const *t7;
 nav__5662_41 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5662_41){(nav__5662_39 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)1ul];
 t5 = (nav__5662_39 *)t4;
 t6 = t5;
 t7 = (nav__5662_39 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (nav__5662_39 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(nav__5662_39)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__5663_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 struct environment_linked_libraries_root_LibraryEntry__852 *t5;
 struct environment_linked_libraries_root_LibraryEntry__852 *t6;
 struct environment_linked_libraries_root_LibraryEntry__852 *const *t7;
 nav__5663_40 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5663_40){(struct environment_linked_libraries_root_LibraryEntry__852 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)2ul];
 t5 = (struct environment_linked_libraries_root_LibraryEntry__852 *)t4;
 t6 = t5;
 t7 = (struct environment_linked_libraries_root_LibraryEntry__852 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (struct environment_linked_libraries_root_LibraryEntry__852 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryEntry__852)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3473(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const a0, struct mem_Allocator__599 const a1) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *t1;
 struct mem_Allocator__599 const *t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 nav__3473_51 t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t3);
 t5 = (*t1);
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3505(t6);
 mem_Allocator_free__anon_6588__5736(t4, t7);
 return;
}

static void environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__2810(struct mem_Allocator__599 const a0, struct environment_linked_libraries_root_LibraryInfo__934 const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 nav__2810_46 t3;
 nav__2810_46 t5;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = a1.path;
 mem_Allocator_free__anon_3042__4747(t2, t3);
 t3 = a1.soname;
 t4 = t3.ptr != NULL;
 if (t4) {
  t5 = t3;
  t2 = (*t1);
  mem_Allocator_free__anon_3042__4747(t2, t5);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = a1.build_id;
 t4 = t5.ptr != NULL;
 if (t4) {
  t3 = t5;
  t2 = (*t1);
  mem_Allocator_free__anon_3042__4747(t2, t3);
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 return;
}

static void array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_deinit__3514(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 const *t1;
 struct mem_Allocator__599 const *t2;
 struct mem_Allocator__599 t3;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t4;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t0;
 nav__3514_42 t5;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 const *)&t0;
 t2 = (struct mem_Allocator__599 const *)&t1->allocator;
 t3 = (*t2);
 t4 = (*t1);
 t5 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__3560(t4);
 mem_Allocator_free__anon_6635__5738(t3, t5);
 return;
}

static nav__5735_39 mem_sliceAsBytes__anon_6577__5735(nav__5735_41 const a0) {
 uintptr_t t0;
 uint64_t t1;
 nav__5735_39 const *t4;
 uint8_t const *t5;
 uint8_t const *t6;
 uint8_t const *const *t7;
 nav__5735_39 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5735_39){(uint8_t const *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (uint8_t const *)t4;
 t6 = t5;
 t7 = (uint8_t const *const *)&t6;
 t0 = a0.len;
 t0 = t0 * (uintptr_t)16ul;
 t5 = (*t7);
 t5 = (uint8_t const *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__3326_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3326(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *t1;
 uint8_t *const *t2;
 uintptr_t t3;
 uint8_t *t4;
 nav__3326_39 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t0;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *)&t0;
 t2 = (uint8_t *const *)&t1->bytes;
 t3 = a0.capacity;
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3325(t3);
 t4 = (*t2);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t3;
 return t5;
}

static void mem_Allocator_free__anon_6588__5736(struct mem_Allocator__599 const a0, nav__5736_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5736_40 t2;
 nav__5736_40 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__5736_40, uint8_t, uintptr_t);
 void (*t17)(void *, nav__5736_40, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6654__5739(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__5736_40));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__5736_40, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static uintptr_t array_hash_map_capacityIndexSize__2907(uint8_t const a0) {
 uint8_t t0;
 t0 = array_hash_map_capacityIndexType__2906(a0);
 switch (t0) {
  case UINT8_C(0): {
   return ((uintptr_t)(uintptr_t)0x2ul);
  }
  case UINT8_C(1): {
   return ((uintptr_t)(uintptr_t)0x4ul);
  }
  case UINT8_C(2): {
   return ((uintptr_t)(uintptr_t)0x8ul);
  }
  default: zig_unreachable();
 }
}

static uintptr_t array_hash_map_IndexHeader_length__3140(struct array_hash_map_IndexHeader__880 const a0) {
 uintptr_t t1;
 uint8_t t0;
 t0 = a0.bit_index;
 t1 = zig_shlw_u64((uintptr_t)1ul, t0, UINT8_C(64));
 return t1;
}

static void mem_Allocator_free__anon_6606__5737(struct mem_Allocator__599 const a0, nav__5737_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5737_40 t2;
 nav__5737_40 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__5737_40, uint8_t, uintptr_t);
 void (*t17)(void *, nav__5737_40, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6688__5755(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__5737_40));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)4ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__5737_40, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static nav__3132_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3132(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *t1;
 uint8_t *const *t2;
 uintptr_t t3;
 uint8_t *t4;
 nav__3132_39 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t0;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *)&t0;
 t2 = (uint8_t *const *)&t1->bytes;
 t3 = a0.capacity;
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3131(t3);
 t4 = (*t2);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t3;
 return t5;
}

static nav__3505_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3505(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *t1;
 uint8_t *const *t2;
 uintptr_t t3;
 uint8_t *t4;
 nav__3505_39 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t0;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *)&t0;
 t2 = (uint8_t *const *)&t1->bytes;
 t3 = a0.capacity;
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3504(t3);
 t4 = (*t2);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t3;
 return t5;
}

static nav__3560_40 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__3560(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 const *t1;
 nav__3560_40 const *t2;
 struct environment_linked_libraries_root_LibraryInfo__934 *const *t3;
 uintptr_t t4;
 struct environment_linked_libraries_root_LibraryInfo__934 *t5;
 nav__3560_40 t6;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t0;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 const *)&t0;
 t2 = (nav__3560_40 const *)&t1->items;
 t3 = &t2->ptr;
 t4 = a0.capacity;
 t5 = (*t3);
 t5 = (struct environment_linked_libraries_root_LibraryInfo__934 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryInfo__934)));
 t6.ptr = t5;
 t6.len = t4;
 return t6;
}

static void mem_Allocator_free__anon_6635__5738(struct mem_Allocator__599 const a0, nav__5738_41 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5738_51 t2;
 nav__5738_51 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__5738_51, uint8_t, uintptr_t);
 void (*t17)(void *, nav__5738_51, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6701__5756(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__5738_51));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__5738_51, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3325(uintptr_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)44ul * a0;
 return t0;
}

static nav__5739_39 mem_sliceAsBytes__anon_6654__5739(nav__5739_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__5739_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5739_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint8_t array_hash_map_capacityIndexType__2906(uint8_t const a0) {
 bool t0;
 t0 = a0 <= UINT8_C(8);
 if (t0) {
  return UINT8_C(0);
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 <= UINT8_C(16);
 if (t0) {
  return UINT8_C(1);
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a0 <= UINT8_C(32);
 debug_assert__180(t0);
 return UINT8_C(2);
}

static nav__5755_39 mem_sliceAsBytes__anon_6688__5755(nav__5755_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__5755_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5755_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3131(uintptr_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)52ul * a0;
 return t0;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3504(uintptr_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)52ul * a0;
 return t0;
}

static nav__5756_39 mem_sliceAsBytes__anon_6701__5756(nav__5756_42 const a0) {
 uintptr_t t0;
 uint64_t t1;
 struct environment_linked_libraries_root_LibraryInfo__934 *t4;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *const *t7;
 nav__5756_39 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5756_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (uint8_t *)t4;
 t6 = t5;
 t7 = (uint8_t *const *)&t6;
 t0 = a0.len;
 t0 = t0 * (uintptr_t)48ul;
 t5 = (*t7);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

bool c_instrument_hooks_is_instrumented__240(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 t5;
 bool t0;
 bool t7;
 uint8_t t6;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t5 = (*t4);
  t6 = t5.tag;
  switch (t6) {
   case UINT8_C(0): {
    t6 = running_on_valgrind();
    t7 = t6 > UINT8_C(0);
    t0 = t7;
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t0 = true;
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t0 = true;
    goto zig_block_1;
   }
   case UINT8_C(3): {
    t0 = false;
    goto zig_block_1;
   }
   default: zig_unreachable();
  }

  zig_block_1:;
  return t0;
 }
 goto zig_block_0;

 zig_block_0:;
 return false;
}

static zig_cold uint16_t runner_fifo_RunnerFifo_start_benchmark__1134(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 struct fifo_posix_Pipe_Reader__788 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(1)});
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 t5 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t4 = fifo_posix_Pipe_Reader_waitForAck__1293(t5, (nav__1134_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_start_benchmark__810(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_start_benchmark__1134(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__823(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_start_benchmark__1134(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_start_benchmark__241(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t6;
 struct instruments_root_Instrument__554 *const *t7;
 struct instruments_root_Instrument__554 t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t13;
 uint16_t t5;
 uint16_t t11;
 uint16_t t12;
 bool t0;
 uint8_t t9;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t6 = t4;
  t7 = (struct instruments_root_Instrument__554 *const *)&t6;
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(1);
  if (t0) {
   t4 = (*t7);
   t10 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
   t11 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_start_benchmark__810(t10);
   memcpy(&t12, &t11, sizeof(uint16_t));
   t5 = t12;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(0);
  if (t0) {
   t0 = features_is_feature_enabled__331(UINT64_C(0));
   t0 = !t0;
   if (t0) {
    callgrind_zero_stats();
    callgrind_start_instrumentation();
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t5 = 0;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(2);
  if (t0) {
   t4 = (*t7);
   t13 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
   t12 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__823(t13);
   memcpy(&t11, &t12, sizeof(uint16_t));
   t5 = t11;
   goto zig_block_2;
  }
  goto zig_block_6;

  zig_block_6:;
  goto zig_block_4;

  zig_block_4:;
  goto zig_block_3;

  zig_block_3:;
  t5 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t12, &t5, sizeof(uint16_t));
  t0 = t12 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static bool features_is_feature_enabled__331(uint64_t const a0) {
 uint64_t t0;
 uint64_t t1;
 uintptr_t t2;
 bool t3;
 t0 = (*&features_features__329);
 t1 = a0;
 t2 = t1;
 t3 = bit_set_IntegerBitSet_2864_29_isSet__366(t0, t2);
 return t3;
}

static bool bit_set_IntegerBitSet_2864_29_isSet__366(uint64_t const a0, uintptr_t const a1) {
 uint64_t t0;
 uint64_t t2;
 bool t1;
 t0 = a1;
 t1 = t0 < UINT64_C(64);
 debug_assert__180(t1);
 t0 = zig_wrap_u64((uint64_t)a0, UINT8_C(64));
 t2 = bit_set_IntegerBitSet_2864_29_maskBit__390(a1);
 t2 = t0 & t2;
 t1 = t2 != UINT64_C(0);
 return t1;
}

static zig_cold uint16_t runner_fifo_RunnerFifo_stop_benchmark__1135(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 struct fifo_posix_Pipe_Reader__788 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, (struct shared_Command__1178){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(2)});
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 t5 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t4 = fifo_posix_Pipe_Reader_waitForAck__1293(t5, (nav__1135_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_stop_benchmark__811(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_stop_benchmark__1135(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__824(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_stop_benchmark__1135(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_stop_benchmark__242(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t6;
 struct instruments_root_Instrument__554 *const *t7;
 struct instruments_root_Instrument__554 t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t13;
 uint16_t t5;
 uint16_t t11;
 uint16_t t12;
 bool t0;
 uint8_t t9;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t6 = t4;
  t7 = (struct instruments_root_Instrument__554 *const *)&t6;
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(0);
  if (t0) {
   t0 = features_is_feature_enabled__331(UINT64_C(0));
   t0 = !t0;
   if (t0) {
    callgrind_stop_instrumentation();
    goto zig_block_4;
   }
   goto zig_block_4;

   zig_block_4:;
   t5 = 0;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(1);
  if (t0) {
   t4 = (*t7);
   t10 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
   t11 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_stop_benchmark__811(t10);
   memcpy(&t12, &t11, sizeof(uint16_t));
   t5 = t12;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(2);
  if (t0) {
   t4 = (*t7);
   t13 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
   t12 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__824(t13);
   memcpy(&t11, &t12, sizeof(uint16_t));
   t5 = t11;
   goto zig_block_2;
  }
  goto zig_block_6;

  zig_block_6:;
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_3;

  zig_block_3:;
  t5 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t12, &t5, sizeof(uint16_t));
  t0 = t12 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint16_t runner_fifo_RunnerFifo_set_executed_benchmark__1136(struct runner_fifo_RunnerFifo__631 *const a0, int32_t const a1, uint8_t const *const a2) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 nav__1136_61 t4;
 nav__1136_61 t5;
 struct shared_Command__struct_1181__1181 t6;
 struct shared_Command__1178 t7;
 struct fifo_posix_Pipe_Reader__788 *t9;
 uint16_t t8;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4 = mem_span__anon_6867__5771(a2);
 memcpy(&t5, &t4, sizeof(nav__1136_61));
 t6.uri = t5;
 t6.pid = a1;
 t7.tag = UINT8_C(0);
 t7.payload.ExecutedBenchmark = t6;
 t8 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t7);
 if (t8) {
  return t8;
 }
 t2 = (*t1);
 t9 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t8 = fifo_posix_Pipe_Reader_waitForAck__1293(t9, (nav__1136_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t8) {
  return t8;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_executed_benchmark__812(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0, int32_t const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_executed_benchmark__1136(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__825(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0, int32_t const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_executed_benchmark__1136(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_executed_benchmark__243(struct InstrumentHooks *const a0, int32_t const a1, char const *const a2) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t7;
 uint8_t const *t5;
 uint8_t const *t11;
 struct instruments_root_Instrument__554 *const *t8;
 struct instruments_root_Instrument__554 t9;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t12;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t14;
 uint16_t t6;
 uint16_t t13;
 bool t0;
 uint8_t t10;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t5 = c_toU8__236(a2);
  t7 = t4;
  t8 = (struct instruments_root_Instrument__554 *const *)&t7;
  t9 = (*t4);
  t10 = t9.tag;
  switch (t10) {
   case UINT8_C(0): {
    t11 = (uint8_t const *)t5;
    callgrind_dump_stats_at(t11);
    goto zig_block_3;
   }
   case UINT8_C(1): {
    t4 = (*t8);
    t12 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
    t13 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_executed_benchmark__812(t12, a1, t5);
    if (t13) {
     t6 = t13;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t4 = (*t8);
    t14 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
    t13 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__825(t14, a1, t5);
    if (t13) {
     t6 = t13;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(3): {
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t6 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t13, &t6, sizeof(uint16_t));
  t0 = t13 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint8_t const *c_toU8__236(char const *const a0) {
 uint8_t const *t0;
 t0 = (uint8_t const *)a0;
 return t0;
}

static nav__5771_39 mem_span__anon_6867__5771(uint8_t const *const a0) {
 uint8_t const *const *t1;
 uintptr_t t2;
 uint8_t const *t3;
 uint8_t const *t0;
 nav__5771_39 t4;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = mem_len__anon_6884__5772(a0);
 t3 = (*t1);
 t3 = (uint8_t const *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t3;
 t4.len = t2;
 return t4;
}

static uintptr_t mem_len__anon_6884__5772(uint8_t const *const a0) {
 uint8_t const *t1;
 uintptr_t t2;
 bool t0;
 t0 = a0 != NULL;
 debug_assert__180(t0);
 t1 = (uint8_t const *)a0;
 t2 = mem_indexOfSentinel__anon_6891__5773(t1);
 return t2;
}

static uintptr_t mem_indexOfSentinel__anon_6891__5773(uint8_t const *const a0) {
 static uint8_t const t11[16] = "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
 static uint8_t const t18[16] = "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
 uint8_t const *const *t1;
 uintptr_t t4;
 uintptr_t t6;
 uintptr_t t16;
 uintptr_t t2;
 uint8_t const *t5;
 uint8_t const *t0;
 uint64_t t7;
 uint8_t const (*t8)[16];
 uint8_t const (*t17)[16];
 bool t3;
 uint8_t t9[16];
 uint8_t t10[16];
 bool t12[16];
 nav__5773_45 t13;
 nav__5773_47 t14;
 uint8_t t15;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = math_isPowerOfTwo__anon_6902__5774();
 if (t3) {
  debug_assert__180(true);
  t4 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t4];
  t4 = (uintptr_t)t5;
  t6 = t4 & (uintptr_t)4095ul;
  t7 = t6;
  t3 = t7 <= UINT64_C(4080);
  if (t3) {
   t6 = t2;
   t5 = (*t1);
   t5 = (uint8_t const *)(((uintptr_t)t5) + (t6*sizeof(uint8_t)));
   t8 = (uint8_t const (*)[16])t5;
   memcpy(t9, (const char *)t8, sizeof(uint8_t[16]));
   memcpy(&t10, &t9, sizeof(uint8_t[16]));
   for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
    t12[t6] = t10[t6] == t11[t6];
   }
   t3 = false;
   for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
    t3 |= t12[t6];
   }
   if (t3) {
    t4 = t2;
    memcpy(t13.array, t12, sizeof(nav__5773_45));
    t14 = simd_firstTrue__anon_6954__5775(t13);
    t15 = t14.payload;
    t6 = (uintptr_t)t15;
    t6 = t4 + t6;
    return t6;
   }
   goto zig_block_2;

   zig_block_2:;
   t6 = t2;
   t16 = mem_alignForward__anon_6957__5776(t4, (uintptr_t)16ul);
   t4 = t16 - t4;
   t4 = t4 / (uintptr_t)1ul;
   t4 = t6 + t4;
   t2 = t4;
   goto zig_block_1;
  }
  zig_loop_64:
  t4 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t4];
  t4 = (uintptr_t)t5;
  t4 = t4 & (uintptr_t)15ul;
  t7 = t4;
  t3 = t7 != UINT64_C(0);
  if (t3) {
   t4 = t2;
   t15 = a0[t4];
   t3 = t15 == UINT8_C(0);
   if (t3) {
    t4 = t2;
    return t4;
   }
   goto zig_block_5;

   zig_block_5:;
   t4 = t2;
   t4 = t4 + (uintptr_t)1ul;
   t2 = t4;
   goto zig_block_4;
  }
  goto zig_block_3;

  zig_block_4:;
  goto zig_loop_64;

  zig_block_3:;
  goto zig_block_1;

  zig_block_1:;
  t6 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t6];
  t6 = (uintptr_t)t5;
  t3 = mem_isAligned__982(t6, (uintptr_t)16ul);
  debug_assert__180(t3);
  zig_loop_99:
  t6 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)(((uintptr_t)t5) + (t6*sizeof(uint8_t)));
  t8 = (uint8_t const (*)[16])t5;
  t17 = (uint8_t const (*)[16])t8;
  memcpy(t10, (const char *)t17, sizeof(uint8_t[16]));
  for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
   t12[t6] = t10[t6] == t18[t6];
  }
  t3 = false;
  for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
   t3 |= t12[t6];
  }
  if (t3) {
   t6 = t2;
   memcpy(t13.array, t12, sizeof(nav__5773_45));
   t14 = simd_firstTrue__anon_6954__5775(t13);
   t15 = t14.payload;
   t4 = (uintptr_t)t15;
   t4 = t6 + t4;
   return t4;
  }
  goto zig_block_6;

  zig_block_6:;
  t6 = t2;
  t6 = t6 + (uintptr_t)16ul;
  t2 = t6;
  goto zig_loop_99;
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_130:
 t16 = t2;
 t15 = a0[t16];
 t3 = t15 != UINT8_C(0);
 if (t3) {
  t16 = t2;
  t16 = t16 + (uintptr_t)1ul;
  t2 = t16;
  goto zig_block_8;
 }
 goto zig_block_7;

 zig_block_8:;
 goto zig_loop_130;

 zig_block_7:;
 t16 = t2;
 return t16;
}

static bool math_isPowerOfTwo__anon_6902__5774(void) {
 debug_assert__180(true);
 return true;
}

static nav__5775_38 simd_firstTrue__anon_6954__5775(nav__5775_40 const a0) {
 static uint8_t const t2[16] = {UINT8_C(0),UINT8_C(1),UINT8_C(2),UINT8_C(3),UINT8_C(4),UINT8_C(5),UINT8_C(6),UINT8_C(7),UINT8_C(8),UINT8_C(9),UINT8_C(10),UINT8_C(11),UINT8_C(12),UINT8_C(13),UINT8_C(14),UINT8_C(15)};
 static uint8_t const t3[16] = {UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15)};
 uintptr_t t1;
 bool t0;
 uint8_t t4[16];
 uint8_t t5;
 nav__5775_38 t6;
 t0 = false;
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)16ul; t1 += (uintptr_t)1ul) {
  t0 |= a0.array[t1];
 }
 t0 = !t0;
 if (t0) {
  return (nav__5775_38){true,UINT8_C(0xa)};
 }
 goto zig_block_0;

 zig_block_0:;
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)16ul; t1 += (uintptr_t)1ul) {
  t4[t1] = a0.array[t1] ? t2[t1] : t3[t1];
 }
 t5 = UINT8_C(15);
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)16ul; t1 += (uintptr_t)1ul) {
  t5 = t5 < t4[t1] ? t5 : t4[t1];
 }
 t6.is_null = false;
 t6.payload = t5;
 return t6;
}

static uintptr_t mem_alignForward__anon_6957__5776(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_6999__5777(a1);
 debug_assert__180(t0);
 t1 = a1 - (uintptr_t)1ul;
 t1 = a0 + t1;
 t1 = mem_alignBackward__anon_7000__5778(t1, a1);
 return t1;
}

static bool mem_isAligned__982(uintptr_t const a0, uintptr_t const a1) {
 uint64_t t0;
 uint64_t t1;
 bool t2;
 t0 = a0;
 t1 = a1;
 t2 = mem_isAlignedGeneric__anon_7004__5779(t0, t1);
 return t2;
}

static bool mem_isValidAlignGeneric__anon_6999__5777(uintptr_t const a0) {
 uint64_t t0;
 bool t1;
 bool t2;
 t0 = a0;
 t1 = t0 > UINT64_C(0);
 if (t1) {
  t1 = math_isPowerOfTwo__anon_3478__5057(a0);
  t2 = t1;
  goto zig_block_0;
 }
 t2 = false;
 goto zig_block_0;

 zig_block_0:;
 return t2;
}

static uintptr_t mem_alignBackward__anon_7000__5778(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_6999__5777(a1);
 debug_assert__180(t0);
 t1 = a1 - (uintptr_t)1ul;
 t1 = zig_not_u64(t1, UINT8_C(64));
 t1 = a0 & t1;
 return t1;
}

static bool mem_isAlignedGeneric__anon_7004__5779(uint64_t const a0, uint64_t const a1) {
 uint64_t t0;
 bool t1;
 t0 = mem_alignBackward__anon_7006__5780(a0, a1);
 t1 = t0 == a0;
 return t1;
}

static uint64_t mem_alignBackward__anon_7006__5780(uint64_t const a0, uint64_t const a1) {
 uint64_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_7008__5781(a1);
 debug_assert__180(t0);
 t1 = a1 - UINT64_C(1);
 t1 = zig_not_u64(t1, UINT8_C(64));
 t1 = a0 & t1;
 return t1;
}

static bool mem_isValidAlignGeneric__anon_7008__5781(uint64_t const a0) {
 bool t0;
 bool t1;
 t0 = a0 > UINT64_C(0);
 if (t0) {
  t0 = math_isPowerOfTwo__anon_7009__5782(a0);
  t1 = t0;
  goto zig_block_0;
 }
 t1 = false;
 goto zig_block_0;

 zig_block_0:;
 return t1;
}

static bool math_isPowerOfTwo__anon_7009__5782(uint64_t const a0) {
 uint64_t t1;
 bool t0;
 t0 = a0 > UINT64_C(0);
 debug_assert__180(t0);
 t1 = a0 - UINT64_C(1);
 t1 = a0 & t1;
 t0 = t1 == UINT64_C(0);
 return t0;
}

uint8_t c_instrument_hooks_executed_benchmark__244(struct InstrumentHooks *const a0, int32_t const a1, char const *const a2) {
 uint8_t t0;
 t0 = c_instrument_hooks_set_executed_benchmark__243(a0, a1, a2);
 return t0;
}

static uint16_t runner_fifo_RunnerFifo_set_integration__1137(struct runner_fifo_RunnerFifo__631 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 nav__1137_61 t4;
 nav__1137_61 t5;
 nav__1137_61 t6;
 struct shared_Command__struct_1182__1182 t7;
 struct shared_Command__1178 t8;
 struct fifo_posix_Pipe_Reader__788 *t10;
 uint16_t t9;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4 = mem_span__anon_6867__5771(a1);
 t5 = mem_span__anon_6867__5771(a2);
 memcpy(&t6, &t4, sizeof(nav__1137_61));
 memcpy(&t4, &t5, sizeof(nav__1137_61));
 t7.name = t6;
 t7.version = t4;
 t8.tag = UINT8_C(5);
 t8.payload.SetIntegration = t7;
 t9 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t8);
 if (t9) {
  return t9;
 }
 t2 = (*t1);
 t10 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t9 = fifo_posix_Pipe_Reader_waitForAck__1293(t10, (nav__1137_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t9) {
  return t9;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_integration__813(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_integration__1137(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__826(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_integration__1137(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_integration__245(struct InstrumentHooks *const a0, char const *const a1, char const *const a2) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t8;
 uint8_t const *t5;
 uint8_t const *t6;
 struct instruments_root_Instrument__554 *const *t9;
 struct instruments_root_Instrument__554 t10;
 struct instruments_valgrind_ValgrindInstrument__564 *t12;
 struct instruments_valgrind_ValgrindInstrument__564 t13;
 struct instruments_valgrind_ValgrindInstrument__564 t15;
 struct instruments_valgrind_ValgrindInstrument__564 const *t16;
 struct mem_Allocator__599 t17;
 nav__245_103 t18;
 nav__245_105 t19;
 nav__245_56 t21;
 uint8_t *t22;
 struct mem_Allocator__599 const *t23;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t24;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t25;
 uint16_t t7;
 uint16_t t14;
 uint16_t t20;
 bool t0;
 uint8_t t11;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t5 = c_toU8__236(a1);
  t6 = c_toU8__236(a2);
  t8 = t4;
  t9 = (struct instruments_root_Instrument__554 *const *)&t8;
  t10 = (*t4);
  t11 = t10.tag;
  switch (t11) {
   case UINT8_C(0): {
    t4 = (*t9);
    t12 = (struct instruments_valgrind_ValgrindInstrument__564 *)&t4->payload.valgrind;
    t13 = (*t12);
    t15 = t13;
    t16 = (struct instruments_valgrind_ValgrindInstrument__564 const *)&t15;
    t17 = t13.allocator;
    t18.f0 = t5;
    t18.f1 = t6;
    t19 = fmt_allocPrintZ__anon_7045__5783(t17, t18);
    if (t19.error) {
     t20 = t19.error;
     t14 = t20;
     goto zig_block_4;
    }
    t21 = t19.payload;
    t22 = t21.ptr;
    t6 = (uint8_t const *)t22;
    callgrind_dump_stats_at(t6);
    t23 = (struct mem_Allocator__599 const *)&t16->allocator;
    t17 = (*t23);
    mem_Allocator_free__anon_7047__5784(t17, t21);
    t14 = 0;
    goto zig_block_4;

    zig_block_4:;
    memcpy(&t20, &t14, sizeof(uint16_t));
    if (t20) {
     t7 = t20;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(1): {
    t4 = (*t9);
    t24 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
    t20 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_integration__813(t24, t5, t6);
    if (t20) {
     t7 = t20;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t4 = (*t9);
    t25 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
    t20 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__826(t25, t5, t6);
    if (t20) {
     t7 = t20;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(3): {
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t7 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t20, &t7, sizeof(uint16_t));
  t0 = t20 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static nav__5783_40 fmt_allocPrintZ__anon_7045__5783(struct mem_Allocator__599 const a0, nav__5783_43 const a1) {
 nav__5783_40 t0;
 nav__5783_39 t2;
 nav__5783_39 t3;
 nav__5783_39 const *t4;
 uintptr_t t5;
 uint8_t *t6;
 uint16_t t1;
 t0 = fmt_allocPrint__anon_7082__5785(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (nav__5783_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t3 = t2;
 t4 = (nav__5783_39 const *)&t3;
 t5 = t2.len;
 t5 = t5 - (uintptr_t)1ul;
 t2 = (*t4);
 t6 = t2.ptr;
 t6 = (uint8_t *)(((uintptr_t)t6) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t6;
 t2.len = t5;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static void mem_Allocator_free__anon_7047__5784(struct mem_Allocator__599 const a0, nav__5784_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5784_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t12;
 struct mem_Allocator_VTable__602 const *const *t13;
 struct mem_Allocator_VTable__602 const *t14;
 void (*const *t15)(void *, nav__5784_40, uint8_t, uintptr_t);
 void (*t16)(void *, nav__5784_40, uint8_t, uintptr_t);
 void *t17;
 bool t5;
 uint8_t t11;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_7089__5786(a1);
 t3 = t2.len;
 t3 = t3 + (uintptr_t)1ul;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t11 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t12 = t10;
 t1 = (struct mem_Allocator__599 const *)&t12;
 t13 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t14 = (*t13);
 t15 = (void (*const *)(void *, nav__5784_40, uint8_t, uintptr_t))&t14->free;
 t16 = (*t15);
 t17 = t10.ptr;
 t16(t17, t2, t11, t3);
 return;
}

static nav__5785_40 fmt_allocPrint__anon_7082__5785(struct mem_Allocator__599 const a0, nav__5785_43 const a1) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t6;
 uint64_t t3;
 nav__5785_54 t4;
 struct mem_Allocator__599 t7;
 struct mem_Allocator__599 t0;
 nav__5785_40 t8;
 nav__5785_40 t11;
 nav__5785_39 t10;
 uint16_t t9;
 bool t5;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = fmt_count__anon_7096__5787(a1);
 t4 = math_cast__anon_7098__5788(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t6 = t4.payload;
  t2 = t6;
  goto zig_block_0;
 }
 return (nav__5785_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_OutOfMemory};

 zig_block_0:;
 t7 = (*t1);
 t8 = mem_Allocator_alloc__anon_2287__4239(t7, t2);
 if (t8.error) {
  t9 = t8.error;
  t8.payload = (nav__5785_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t9;
  return t8;
 }
 t10 = t8.payload;
 t8 = fmt_bufPrint__anon_7101__5789(t10, a1);
 t5 = t8.error == UINT16_C(0);
 if (t5) {
  t10 = t8.payload;
  t8.payload = t10;
  t8.error = UINT16_C(0);
  t11 = t8;
  goto zig_block_1;
 }
 t9 = t8.error;
 switch (t9) {
  case zig_error_NoSpaceLeft: {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_1:;
 return t11;
}

static nav__5786_39 mem_sliceAsBytes__anon_7089__5786(nav__5786_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__5786_39 t7;
 bool t2;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = a0.ptr;
 t4 = (uint8_t *)t3;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint16_t fmt_format__anon_7152__5828(struct io_Writer__3827 const a0, nav__5828_40 const a1) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint8_t const *t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5828_44){(uint8_t const *)&__anon_7177,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_7186__5829(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5828_44){(uint8_t const *)&__anon_7192,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f1;
 t3 = fmt_formatType__anon_7186__5829(t4, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__5828_44){(uint8_t const *)&__anon_7205,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint64_t fmt_count__anon_7096__5787(nav__5787_39 const a0) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 t1;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 t0;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 t2;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 t3;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 const *t4;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 const *t5;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 const *const *t6;
 void const **t8;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const *t9;
 void const *t10;
 nav__5787_53 (**t11)(void const *, nav__5787_54);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 uint64_t t15;
 uint16_t t13;
 bool t14;
 t1 = io_counting_writer_countingWriter__anon_7122__5814();
 t0 = t1;
 t2 = io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__5813(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5787_53 (**)(void const *, nav__5787_54))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__5827;
 t12 = t7;
 t13 = fmt_format__anon_7152__5828(t12, a0);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 zig_unreachable();

 zig_block_0:;
 t1 = t0;
 t15 = t1.bytes_written;
 return t15;
}

static nav__5788_38 math_cast__anon_7098__5788(uint64_t const a0) {
 uintptr_t t0;
 nav__5788_38 t1;
 t0 = a0;
 t1.is_null = false;
 t1.payload = t0;
 return t1;
}

static nav__5789_40 fmt_bufPrint__anon_7101__5789(nav__5789_39 const a0, nav__5789_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__5789_57 (**t11)(void const *, nav__5789_58);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__5789_39 t15;
 nav__5789_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5789_57 (**)(void const *, nav__5789_58))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_7152__5828(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5789_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 io_counting_writer_countingWriter__anon_7122__5814(void) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 t0;
 t0 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119){UINT64_C(0)};
 return t0;
}

static struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__5813(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const a0) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 **t1;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7133 t0;
 t1 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__5827_38 io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__5827(void const *const a0, nav__5827_41 const a1) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const *t0;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *t1;
 nav__5827_38 t2;
 nav__5827_38 t3;
 t0 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const *)a0;
 t1 = (*t0);
 t2 = io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__5812(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__5827_38));
 return t3;
}

static uint16_t fmt_formatType__anon_7186__5829(uint8_t const *const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 nav__5829_47 t2;
 nav__5829_47 t3;
 uint8_t const *t0;
 struct io_Writer__3827 t1;
 uint16_t t4;
 uint16_t t5;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = mem_span__anon_6867__5771(a0);
 memcpy(&t3, &t2, sizeof(nav__5829_47));
 t4 = fmt_formatBuf__anon_5109__5592(t3, a1, a2);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static nav__5812_38 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__5812(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const a0, nav__5812_42 const a1) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const *t1;
 nav__5812_38 t2;
 uintptr_t t3;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *t4;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *t0;
 uint64_t *t5;
 uint64_t t6;
 uint64_t t7;
 t0 = a0;
 t1 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7119 *const *)&t0;
 t2 = io_dummyWrite__3933(a1);
 t3 = t2.payload;
 t4 = (*t1);
 t5 = (uint64_t *)&t4->bytes_written;
 t6 = (*t5);
 t7 = t3;
 t7 = t6 + t7;
 (*t5) = t7;
 t2.payload = t3;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__3933_38 io_dummyWrite__3933(nav__3933_40 const a0) {
 uintptr_t t0;
 nav__3933_38 t1;
 t0 = a0.len;
 t1.payload = t0;
 t1.error = UINT16_C(0);
 return t1;
}

static uint16_t runner_fifo_RunnerFifo_add_marker__1138(struct runner_fifo_RunnerFifo__631 *const a0, int32_t const a1, struct shared_MarkerType__1185 const a2) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__786 *t3;
 struct shared_Command__struct_1183__1183 t4;
 struct shared_Command__1178 t5;
 struct fifo_posix_Pipe_Reader__788 *t7;
 uint16_t t6;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__786 *)&t2->writer;
 t4.marker = a2;
 t4.pid = a1;
 t5.tag = UINT8_C(7);
 t5.payload.AddMarker = t4;
 t6 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t5);
 if (t6) {
  return t6;
 }
 t2 = (*t1);
 t7 = (struct fifo_posix_Pipe_Reader__788 *)&t2->reader;
 t6 = fifo_posix_Pipe_Reader_waitForAck__1293(t7, (nav__1138_77){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t6) {
  return t6;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_add_marker__814(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0, int32_t const a1, struct shared_MarkerType__1185 const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_add_marker__1138(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__827(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0, int32_t const a1, struct shared_MarkerType__1185 const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_add_marker__1138(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_add_marker__250(struct InstrumentHooks *const a0, int32_t const a1, uint8_t const a2, uint64_t const a3) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct shared_MarkerType__1185 t4;
 struct shared_MarkerType__1185 t5;
 struct instruments_root_Instrument__554 *t6;
 struct instruments_root_Instrument__554 *t8;
 struct instruments_root_Instrument__554 *const *t9;
 struct instruments_root_Instrument__554 t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t12;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t15;
 uint16_t t7;
 uint16_t t13;
 uint16_t t14;
 bool t0;
 uint8_t t11;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  switch (a2) {
   case UINT8_C(0): {
    t5.tag = UINT8_C(0);
    t5.payload.SampleStart = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t5.tag = UINT8_C(1);
    t5.payload.SampleEnd = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t5.tag = UINT8_C(2);
    t5.payload.BenchmarkStart = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(3): {
    t5.tag = UINT8_C(3);
    t5.payload.BenchmarkEnd = a3;
    t4 = t5;
    goto zig_block_1;
   }
   default: {
    return UINT8_C(2);
   }
  }

  zig_block_1:;
  t1 = (*t3);
  t6 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t8 = t6;
  t9 = (struct instruments_root_Instrument__554 *const *)&t8;
  t10 = (*t6);
  t11 = t10.tag;
  t0 = t11 == UINT8_C(1);
  if (t0) {
   t6 = (*t9);
   t12 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t6->payload.walltime;
   t13 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_add_marker__814(t12, a1, t4);
   memcpy(&t14, &t13, sizeof(uint16_t));
   t7 = t14;
   goto zig_block_3;
  }
  t10 = (*t6);
  t11 = t10.tag;
  t0 = t11 == UINT8_C(2);
  if (t0) {
   t6 = (*t9);
   t15 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t6->payload.analysis;
   t14 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__827(t15, a1, t4);
   memcpy(&t13, &t14, sizeof(uint16_t));
   t7 = t13;
   goto zig_block_3;
  }
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_4;

  zig_block_4:;
  t7 = 0;
  goto zig_block_3;

  zig_block_3:;
  memcpy(&t14, &t7, sizeof(uint16_t));
  t0 = t14 == UINT16_C(0);
  if (t0) {
   goto zig_block_2;
  }
  return UINT8_C(1);

  zig_block_2:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

uint64_t c_instrument_hooks_current_timestamp__254(void) {
 uint64_t t0;
 uint64_t t3;
 struct c_MachTimebaseInfo__7314 t1;
 uint32_t t2;
 once_Once_28_28function_20_27init_27_29_29_call__5842(((struct once_Once_28_28function_20_27init_27_29_29__7273 *)&c_instrument_hooks_current_timestamp_S_once__5831));
 t0 = mach_absolute_time();
 t1 = (*((struct c_MachTimebaseInfo__7314 *)&c_instrument_hooks_current_timestamp_S_cached__5830));
 t2 = t1.numer;
 t3 = (uint64_t)t2;
 t3 = t0 * t3;
 t1 = (*((struct c_MachTimebaseInfo__7314 *)&c_instrument_hooks_current_timestamp_S_cached__5830));
 t2 = t1.denom;
 t0 = (uint64_t)t2;
 t0 = t3 / t0;
 return t0;
}

static void once_Once_28_28function_20_27init_27_29_29_call__5842(struct once_Once_28_28function_20_27init_27_29_29__7273 *const a0) {
 struct once_Once_28_28function_20_27init_27_29_29__7273 *const *t1;
 struct once_Once_28_28function_20_27init_27_29_29__7273 *t2;
 struct once_Once_28_28function_20_27init_27_29_29__7273 *t0;
 bool *t3;
 bool const *t4;
 bool t5;
 t0 = a0;
 t1 = (struct once_Once_28_28function_20_27init_27_29_29__7273 *const *)&t0;
 t2 = (*t1);
 t3 = (bool *)&t2->done;
 t4 = (bool const *)t3;
 zig_atomic_load(t5, (zig_atomic(bool) *)t4, zig_memory_order_acquire, u8, bool);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 once_Once_28_28function_20_27init_27_29_29_callSlow__5843(t2);
 return;
}

static zig_cold void once_Once_28_28function_20_27init_27_29_29_callSlow__5843(struct once_Once_28_28function_20_27init_27_29_29__7273 *const a0) {
 struct once_Once_28_28function_20_27init_27_29_29__7273 *const *t1;
 struct once_Once_28_28function_20_27init_27_29_29__7273 *t2;
 struct once_Once_28_28function_20_27init_27_29_29__7273 *t0;
 struct Thread_Mutex__7278 *t3;
 bool *t4;
 bool t5;
 t0 = a0;
 t1 = (struct once_Once_28_28function_20_27init_27_29_29__7273 *const *)&t0;
 t2 = (*t1);
 t3 = (struct Thread_Mutex__7278 *)&t2->mutex;
 Thread_Mutex_lock__5905(t3);
 t4 = (bool *)&a0->done;
 t5 = (*t4);
 t5 = !t5;
 if (t5) {
  c_instrument_hooks_current_timestamp_S_init__5832();
  t2 = (*t1);
  t4 = (bool *)&t2->done;
  t5 = true;
  zig_atomic_store((zig_atomic(bool) *)t4, t5, zig_memory_order_release, u8, bool);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t3 = (struct Thread_Mutex__7278 *)&t2->mutex;
 Thread_Mutex_unlock__5906(t3);
 return;
}

static void Thread_Mutex_lock__5905(struct Thread_Mutex__7278 *const a0) {
 struct Thread_Mutex__7278 *const *t1;
 struct Thread_Mutex__7278 *t2;
 struct Thread_Mutex__7278 *t0;
 struct Thread_Mutex_DarwinImpl__7290 *t3;
 t0 = a0;
 t1 = (struct Thread_Mutex__7278 *const *)&t0;
 t2 = (*t1);
 t3 = (struct Thread_Mutex_DarwinImpl__7290 *)&t2->impl;
 Thread_Mutex_DarwinImpl_lock__5919(t3);
 return;
}

static void c_instrument_hooks_current_timestamp_S_init__5832(void) {
 int t1;
 int32_t t2;
 struct c_MachTimebaseInfo__7314 t5;
 struct c_MachTimebaseInfo__7314 t0;
 uint32_t t6;
 bool t3;
 bool t4;
 t1 = mach_timebase_info(&t0);
 t2 = t1;
 t3 = t2 != INT32_C(0);
 if (t3) {
  t4 = true;
  goto zig_block_1;
 }
 t5 = t0;
 t6 = t5.denom;
 t3 = t6 == UINT32_C(0);
 t4 = t3;
 goto zig_block_1;

 zig_block_1:;
 if (t4) {
  t0 = (struct c_MachTimebaseInfo__7314){UINT32_C(1),UINT32_C(1)};
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t0;
 (*((struct c_MachTimebaseInfo__7314 *)&c_instrument_hooks_current_timestamp_S_cached__5830)) = t5;
 return;
}

static void Thread_Mutex_unlock__5906(struct Thread_Mutex__7278 *const a0) {
 struct Thread_Mutex__7278 *const *t1;
 struct Thread_Mutex__7278 *t2;
 struct Thread_Mutex__7278 *t0;
 struct Thread_Mutex_DarwinImpl__7290 *t3;
 t0 = a0;
 t1 = (struct Thread_Mutex__7278 *const *)&t0;
 t2 = (*t1);
 t3 = (struct Thread_Mutex_DarwinImpl__7290 *)&t2->impl;
 Thread_Mutex_DarwinImpl_unlock__5920(t3);
 return;
}

static void Thread_Mutex_DarwinImpl_lock__5919(struct Thread_Mutex_DarwinImpl__7290 *const a0) {
 struct Thread_Mutex_DarwinImpl__7290 *const *t1;
 struct Thread_Mutex_DarwinImpl__7290 *t2;
 struct Thread_Mutex_DarwinImpl__7290 *t0;
 struct c_darwin_os_unfair_lock__7296 *t3;
 t0 = a0;
 t1 = (struct Thread_Mutex_DarwinImpl__7290 *const *)&t0;
 t2 = (*t1);
 t3 = (struct c_darwin_os_unfair_lock__7296 *)&t2->oul;
 os_unfair_lock_lock(t3);
 return;
}

static void Thread_Mutex_DarwinImpl_unlock__5920(struct Thread_Mutex_DarwinImpl__7290 *const a0) {
 struct Thread_Mutex_DarwinImpl__7290 *const *t1;
 struct Thread_Mutex_DarwinImpl__7290 *t2;
 struct Thread_Mutex_DarwinImpl__7290 *t0;
 struct c_darwin_os_unfair_lock__7296 *t3;
 t0 = a0;
 t1 = (struct Thread_Mutex_DarwinImpl__7290 *const *)&t0;
 t2 = (*t1);
 t3 = (struct c_darwin_os_unfair_lock__7296 *)&t2->oul;
 os_unfair_lock_unlock(t3);
 return;
}

static uint16_t environment_root_Environment_setIntegrationEnvironmentValue__778(struct environment_root_Environment__559 *const a0, nav__778_41 const a1, nav__778_41 const a2, struct environment_root_EnvironmentValue__844 const a3) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t0;
 struct environment_root_EnvironmentJson__833 *t3;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t5;
 struct mem_Allocator__599 *t6;
 struct mem_Allocator__599 t7;
 nav__778_93 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 const *t12;
 nav__778_41 *t14;
 nav__778_98 t15;
 nav__778_96 t16;
 nav__778_41 t17;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t18;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t19;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t20;
 nav__778_110 t21;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t22;
 struct environment_root_EnvironmentValue__844 *t23;
 struct environment_root_EnvironmentValue__844 t24;
 uint16_t t9;
 bool t13;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (struct environment_root_EnvironmentJson__833 *)&t2->data;
 t4 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *)&t3->integration_environment;
 t5 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *)&t4->map;
 t6 = (struct mem_Allocator__599 *)&a0->allocator;
 t7 = (*t6);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2967(t5, t7, a1);
 if (t8.error) {
  t9 = t8.error;
  return t9;
 }
 t10 = t8.payload;
 t11 = t10;
 t12 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 const *)&t11;
 t13 = t10.found_existing;
 t13 = !t13;
 if (t13) {
  t14 = t10.key_ptr;
  t2 = (*t1);
  t6 = (struct mem_Allocator__599 *)&t2->allocator;
  t7 = (*t6);
  t15 = mem_Allocator_dupe__anon_7390__5934(t7, a1);
  if (t15.error) {
   t9 = t15.error;
   return t9;
  }
  t16 = t15.payload;
  memcpy(&t17, &t16, sizeof(nav__778_41));
  (*t14) = t17;
  t18 = t10.value_ptr;
  (*t18) = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846){{{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul,(uintptr_t)0ul},NULL}};
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t19 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *)&t12->value_ptr;
 t18 = (*t19);
 t20 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *)&t18->map;
 t6 = (struct mem_Allocator__599 *)&a0->allocator;
 t7 = (*t6);
 t21 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3178(t20, t7, a2);
 if (t21.error) {
  t9 = t21.error;
  return t9;
 }
 t22 = t21.payload;
 t13 = t22.found_existing;
 if (t13) {
  t2 = (*t1);
  t23 = t22.value_ptr;
  t24 = (*t23);
  environment_root_Environment_freeEnvironmentValue__777(t2, t24);
  goto zig_block_1;
 }
 t14 = t22.key_ptr;
 t2 = (*t1);
 t6 = (struct mem_Allocator__599 *)&t2->allocator;
 t7 = (*t6);
 t15 = mem_Allocator_dupe__anon_7390__5934(t7, a2);
 if (t15.error) {
  t9 = t15.error;
  return t9;
 }
 t16 = t15.payload;
 memcpy(&t17, &t16, sizeof(nav__778_41));
 (*t14) = t17;
 goto zig_block_1;

 zig_block_1:;
 t23 = t22.value_ptr;
 (*t23) = a3;
 return 0;
}

static uint16_t environment_root_Environment_setIntegrationEnvironment__775(struct environment_root_Environment__559 *const a0, nav__775_41 const a1, nav__775_41 const a2, nav__775_41 const a3) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t3;
 struct environment_root_Environment__559 *t0;
 struct mem_Allocator__599 *t4;
 struct mem_Allocator__599 t5;
 nav__775_81 t6;
 nav__775_79 t8;
 nav__775_41 t9;
 struct environment_root_EnvironmentValue__844 t10;
 uint16_t t7;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (*t1);
 t4 = (struct mem_Allocator__599 *)&t3->allocator;
 t5 = (*t4);
 t6 = mem_Allocator_dupe__anon_7390__5934(t5, a3);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t8 = t6.payload;
 memcpy(&t9, &t8, sizeof(nav__775_41));
 t10.tag = UINT8_C(0);
 t10.payload.string = t9;
 t7 = environment_root_Environment_setIntegrationEnvironmentValue__778(t2, a1, a2, t10);
 if (t7) {
  return t7;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_environment__255(struct InstrumentHooks *const a0, char const *const a1, char const *const a2, char const *const a3) {
 struct InstrumentHooks *t2;
 struct InstrumentHooks *t3;
 struct InstrumentHooks *const *t4;
 struct environment_root_Environment__559 *t5;
 uint8_t const *t6;
 nav__255_100 t7;
 nav__255_100 t8;
 nav__255_100 t9;
 nav__255_100 t10;
 uint16_t t11;
 bool t0;
 bool t1;
 t0 = a1 == NULL;
 if (t0) {
  t1 = true;
  goto zig_block_1;
 }
 t0 = a2 == NULL;
 t1 = t0;
 goto zig_block_1;

 zig_block_1:;
 if (t1) {
  t0 = true;
  goto zig_block_2;
 }
 t1 = a3 == NULL;
 t0 = t1;
 goto zig_block_2;

 zig_block_2:;
 if (t0) {
  return UINT8_C(1);
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0 != NULL;
 if (t1) {
  t2 = a0;
  t3 = t2;
  t4 = (struct InstrumentHooks *const *)&t3;
  t2 = (*t4);
  t5 = (struct environment_root_Environment__559 *)&t2->environment;
  t6 = c_toU8__236(a1);
  t7 = mem_span__anon_6867__5771(t6);
  memcpy(&t8, &t7, sizeof(nav__255_100));
  t6 = c_toU8__236(a2);
  t7 = mem_span__anon_6867__5771(t6);
  memcpy(&t9, &t7, sizeof(nav__255_100));
  t6 = c_toU8__236(a3);
  t7 = mem_span__anon_6867__5771(t6);
  memcpy(&t10, &t7, sizeof(nav__255_100));
  t11 = environment_root_Environment_setIntegrationEnvironment__775(t5, t8, t9, t10);
  t1 = t11 == UINT16_C(0);
  if (t1) {
   goto zig_block_4;
  }
  return UINT8_C(1);

  zig_block_4:;
  return UINT8_C(0);
 }
 goto zig_block_3;

 zig_block_3:;
 return UINT8_C(1);
}

static nav__5934_40 mem_Allocator_dupe__anon_7390__5934(struct mem_Allocator__599 const a0, nav__5934_43 const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 uintptr_t t3;
 nav__5934_40 t4;
 nav__5934_39 t6;
 uint8_t const *t7;
 uint16_t t5;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = mem_Allocator_alloc__anon_2287__4239(t2, t3);
 if (t4.error) {
  t5 = t4.error;
  t4.payload = (nav__5934_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t4.error = t5;
  return t4;
 }
 t6 = t4.payload;
 t7 = a1.ptr;
 if (t6.len != 0) memcpy(t6.ptr, t7, t6.len * sizeof(uint8_t));
 t4.payload = t6;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__2967_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2967(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct mem_Allocator__599 const a1, nav__2967_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 nav__2967_39 t3;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2968(t2, a1, a2);
 return t3;
}

static nav__3178_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3178(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct mem_Allocator__599 const a1, nav__3178_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 nav__3178_39 t3;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3179(t2, a1, a2);
 return t3;
}

static nav__2968_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2968(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct mem_Allocator__599 const a1, nav__2968_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 nav__2968_39 t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t5;
 nav__2968_44 *t7;
 uint16_t t4;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5935(t2, a1, a2);
 if (t3.error) {
  t4 = t3.error;
  t3.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392){((nav__2968_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
  t3.error = t4;
  return t3;
 }
 t5 = t3.payload;
 t6 = t5.found_existing;
 t6 = !t6;
 if (t6) {
  t7 = t5.key_ptr;
  (*t7) = a2;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3.payload = t5;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__3179_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3179(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct mem_Allocator__599 const a1, nav__3179_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 nav__3179_39 t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t5;
 nav__3179_44 *t7;
 uint16_t t4;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5936(t2, a1, a2);
 if (t3.error) {
  t4 = t3.error;
  t3.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403){((nav__3179_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_root_EnvironmentValue__844 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
  t3.error = t4;
  return t3;
 }
 t5 = t3.payload;
 t6 = t5.found_existing;
 t6 = !t6;
 if (t6) {
  t7 = t5.key_ptr;
  (*t7) = a2;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3.payload = t5;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__5935_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5935(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct mem_Allocator__599 const a1, nav__5935_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t4;
 uintptr_t t5;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t8;
 nav__5935_67 t9;
 nav__5935_39 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t14;
 nav__5935_73 t15;
 nav__5935_73 t16;
 nav__5935_73 const *t17;
 nav__5935_44 *t18;
 nav__5935_76 t19;
 nav__5935_76 t20;
 nav__5935_76 const *t21;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t22;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t23;
 uint16_t t6;
 bool t7;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&a0->entries;
 t4 = (*t3);
 t5 = t4.len;
 t5 = t5 + (uintptr_t)1ul;
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2977(t2, a1, t5);
 t7 = t6 == UINT16_C(0);
 if (t7) {
  goto zig_block_0;
 }
 t2 = (*t1);
 t8 = (*t2);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5937(t8, a2);
 t7 = t9.is_null != true;
 if (t7) {
  t10 = t9.payload;
  t5 = t10;
  goto zig_block_1;
 }
 t11.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392){((nav__5935_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
 t11.error = t6;
 return t11;

 zig_block_1:;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t2->entries;
 t4 = (*t3);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t4);
 t13 = t12;
 t14 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t13;
 t12 = (*t14);
 t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t12);
 t16 = t15;
 t17 = (nav__5935_73 const *)&t16;
 t15 = (*t17);
 t18 = &t15.ptr[t5];
 t12 = (*t14);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t12);
 t20 = t19;
 t21 = (nav__5935_76 const *)&t20;
 t19 = (*t21);
 t22 = &t19.ptr[t5];
 t23.key_ptr = t18;
 t23.value_ptr = t22;
 t23.index = t5;
 t23.found_existing = true;
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;

 zig_block_0:;
 t2 = (*t1);
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5938(t2, a2);
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;
}

static nav__5936_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5936(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct mem_Allocator__599 const a1, nav__5936_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t4;
 uintptr_t t5;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t8;
 nav__5936_67 t9;
 nav__5936_39 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t14;
 nav__5936_73 t15;
 nav__5936_73 t16;
 nav__5936_73 const *t17;
 nav__5936_44 *t18;
 nav__5936_76 t19;
 nav__5936_76 t20;
 nav__5936_76 const *t21;
 struct environment_root_EnvironmentValue__844 *t22;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t23;
 uint16_t t6;
 bool t7;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&a0->entries;
 t4 = (*t3);
 t5 = t4.len;
 t5 = t5 + (uintptr_t)1ul;
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3188(t2, a1, t5);
 t7 = t6 == UINT16_C(0);
 if (t7) {
  goto zig_block_0;
 }
 t2 = (*t1);
 t8 = (*t2);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5939(t8, a2);
 t7 = t9.is_null != true;
 if (t7) {
  t10 = t9.payload;
  t5 = t10;
  goto zig_block_1;
 }
 t11.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403){((nav__5936_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_root_EnvironmentValue__844 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
 t11.error = t6;
 return t11;

 zig_block_1:;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t2->entries;
 t4 = (*t3);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t4);
 t13 = t12;
 t14 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t13;
 t12 = (*t14);
 t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t12);
 t16 = t15;
 t17 = (nav__5936_73 const *)&t16;
 t15 = (*t17);
 t18 = &t15.ptr[t5];
 t12 = (*t14);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t12);
 t20 = t19;
 t21 = (nav__5936_76 const *)&t20;
 t19 = (*t21);
 t22 = &t19.ptr[t5];
 t23.key_ptr = t18;
 t23.value_ptr = t22;
 t23.index = t5;
 t23.found_existing = true;
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;

 zig_block_0:;
 t2 = (*t1);
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5940(t2, a2);
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;
}

static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2977(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 void *t3;
 uint64_t t4;
 uint64_t t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t6;
 struct array_hash_map_IndexHeader__880 **t8;
 struct array_hash_map_IndexHeader__880 *t9;
 struct array_hash_map_IndexHeader__880 *t10;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t20;
 struct array_hash_map_IndexHeader__880 *const *t12;
 nav__2977_62 t18;
 struct array_hash_map_IndexHeader__880 t13;
 uint32_t t14;
 uint16_t t7;
 nav__2977_60 t16;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3147(t3);
 t4 = a2;
 t5 = t4 <= UINT64_C(8);
 if (t5) {
  t2 = (*t1);
  t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t2->entries;
  t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3120(t6, a1, a2);
  if (t7) {
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3148(t3);
   return t7;
  }
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t10 = t9;
  t11 = t10;
  t12 = (struct array_hash_map_IndexHeader__880 *const *)&t11;
  t10 = (*t12);
  t13 = (*t10);
  t14 = array_hash_map_IndexHeader_capacity__3139(t13);
  t4 = a2;
  t15 = (uint64_t)t14;
  t5 = t4 <= t15;
  if (t5) {
   t2 = (*t1);
   t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t2->entries;
   t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3120(t6, a1, a2);
   if (t7) {
    t2 = (*t1);
    t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
    debug_SafetyLock_unlock__3148(t3);
    return t7;
   }
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3148(t3);
   return 0;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t2->entries;
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3120(t6, a1, a2);
 if (t7) {
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t16 = array_hash_map_IndexHeader_findBitIndex__3142(a2);
 if (t16.error) {
  t7 = t16.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t17 = t16.payload;
 t18 = array_hash_map_IndexHeader_alloc__3143(a1, t17);
 if (t18.error) {
  t7 = t18.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t10 = t18.payload;
 t8 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t19 = t9;
  t20 = t19;
  t12 = (struct array_hash_map_IndexHeader__880 *const *)&t20;
  t19 = (*t12);
  array_hash_map_IndexHeader_free__3144(t19, a1);
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t2 = (*t1);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3069(t2, t10);
 t2 = (*t1);
 t8 = (struct array_hash_map_IndexHeader__880 **)&t2->index_header;
 t19 = (struct array_hash_map_IndexHeader__880 *)t10;
 (*t8) = t19;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3148(t3);
 return 0;
}

static nav__5937_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5937(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5937_41 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_IndexHeader__880 *t3;
 struct array_hash_map_IndexHeader__880 *t5;
 struct array_hash_map_IndexHeader__880 *t24;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t11;
 nav__5937_61 t12;
 nav__5937_64 t13;
 uintptr_t t15;
 uintptr_t t16;
 uintptr_t t14;
 uint64_t t17;
 uint64_t t18;
 nav__5937_41 *t19;
 nav__5937_41 t22;
 nav__5937_38 t23;
 struct array_hash_map_IndexHeader__880 *const *t25;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 uint32_t t6;
 uint32_t t20;
 struct array_hash_map_IndexHeader__880 t26;
 bool t4;
 bool t21;
 uint8_t t27;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t3 = a0.index_header;
 t4 = t3 != NULL;
 if (t4) {
  t5 = t3;
  t2 = t5;
  goto zig_block_0;
 }
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t7 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *)&t1->entries;
 t8 = (*t7);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t8);
 t10 = t9;
 t11 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t10;
 t9 = (*t11);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t9);
 t9 = (*t11);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t9);
 t14 = (uintptr_t)0ul;
 t15 = t13.len;
 zig_loop_32:
 t16 = t14;
 t17 = t16;
 t18 = t15;
 t4 = t17 < t18;
 if (t4) {
  t19 = &t13.ptr[t16];
  t20 = t12.ptr[t16];
  t4 = t20 == t6;
  if (t4) {
   t22 = (*t19);
   t4 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t22, t16);
   t21 = t4;
   goto zig_block_4;
  }
  t21 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t21) {
   t23.is_null = false;
   t23.payload = t16;
   return t23;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t16 = t16 + (uintptr_t)1ul;
 t14 = t16;
 goto zig_loop_32;

 zig_block_1:;
 return (nav__5937_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t24 = t2;
 t25 = (struct array_hash_map_IndexHeader__880 *const *)&t24;
 t5 = (*t25);
 t26 = (*t5);
 t27 = array_hash_map_IndexHeader_capacityIndexType__3138(t26);
 switch (t27) {
  case UINT8_C(0): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5944(t28, a1, t2);
   return t23;
  }
  case UINT8_C(1): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5945(t28, a1, t2);
   return t23;
  }
  case UINT8_C(2): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5946(t28, a1, t2);
   return t23;
  }
  default: zig_unreachable();
 }
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5938(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, nav__5938_42 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t6;
 struct array_hash_map_IndexHeader__880 *t39;
 struct array_hash_map_IndexHeader__880 **t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t13;
 nav__5938_66 t14;
 nav__5938_66 t15;
 nav__5938_66 const *t16;
 nav__5938_69 t17;
 nav__5938_69 t18;
 nav__5938_69 const *t19;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t20;
 uint64_t t23;
 uint64_t t24;
 nav__5938_42 *t25;
 nav__5938_42 t28;
 nav__5938_72 t29;
 nav__5938_72 t30;
 nav__5938_72 t37;
 nav__5938_72 const *t31;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t33;
 uint32_t *const *t34;
 uint32_t *t35;
 nav__5938_42 *const *t36;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t38;
 struct array_hash_map_IndexHeader__880 *const *t40;
 uint32_t t7;
 uint32_t t26;
 struct array_hash_map_IndexHeader__880 t41;
 bool t5;
 bool t27;
 uint8_t t42;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t3 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t4 = (*t3);
 t5 = t4 != NULL;
 if (t5) {
  t6 = t4;
  t2 = t6;
  goto zig_block_0;
 }
 t7 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t8->entries;
 t10 = (*t9);
 t11 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t10);
 t12 = t11;
 t13 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t12;
 t11 = (*t13);
 t14 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t11);
 t15 = t14;
 t16 = (nav__5938_66 const *)&t15;
 t11 = (*t13);
 t17 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t11);
 t18 = t17;
 t19 = (nav__5938_69 const *)&t18;
 t20 = (uintptr_t)0ul;
 t21 = t17.len;
 zig_loop_39:
 t22 = t20;
 t23 = t22;
 t24 = t21;
 t5 = t23 < t24;
 if (t5) {
  t25 = &t17.ptr[t22];
  t26 = t14.ptr[t22];
  t5 = t26 == t7;
  if (t5) {
   t28 = (*t25);
   t5 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t28, t22);
   t27 = t5;
   goto zig_block_4;
  }
  t27 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t27) {
   t11 = (*t13);
   t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t11);
   t30 = t29;
   t31 = (nav__5938_72 const *)&t30;
   t29 = (*t31);
   t32 = &t29.ptr[t22];
   t33.key_ptr = t25;
   t33.value_ptr = t32;
   t33.index = t22;
   t33.found_existing = true;
   return t33;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t22 = t22 + (uintptr_t)1ul;
 t20 = t22;
 goto zig_loop_39;

 zig_block_1:;
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t8->entries;
 t21 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t9);
 t34 = &t16->ptr;
 t35 = (*t34);
 t35 = (uint32_t *)&t35[t21];
 (*t35) = t7;
 t36 = &t19->ptr;
 t25 = (*t36);
 t25 = (nav__5938_42 *)&t25[t21];
 t11 = (*t13);
 t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t11);
 t37 = t29;
 t31 = (nav__5938_72 const *)&t37;
 t38 = &t31->ptr;
 t32 = (*t38);
 t32 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t32[t21];
 t33.key_ptr = t25;
 t33.value_ptr = t32;
 t33.index = t21;
 t33.found_existing = false;
 return t33;

 zig_block_0:;
 t39 = t2;
 t40 = (struct array_hash_map_IndexHeader__880 *const *)&t39;
 t6 = (*t40);
 t41 = (*t6);
 t42 = array_hash_map_IndexHeader_capacityIndexType__3138(t41);
 switch (t42) {
  case UINT8_C(0): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5947(t8, a1, t2);
   return t33;
  }
  case UINT8_C(1): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5948(t8, a1, t2);
   return t33;
  }
  case UINT8_C(2): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5949(t8, a1, t2);
   return t33;
  }
  default: zig_unreachable();
 }
}

static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3188(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 void *t3;
 uint64_t t4;
 uint64_t t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t6;
 struct array_hash_map_IndexHeader__880 **t8;
 struct array_hash_map_IndexHeader__880 *t9;
 struct array_hash_map_IndexHeader__880 *t10;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t20;
 struct array_hash_map_IndexHeader__880 *const *t12;
 nav__3188_62 t18;
 struct array_hash_map_IndexHeader__880 t13;
 uint32_t t14;
 uint16_t t7;
 nav__3188_60 t16;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3147(t3);
 t4 = a2;
 t5 = t4 <= UINT64_C(8);
 if (t5) {
  t2 = (*t1);
  t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t2->entries;
  t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3314(t6, a1, a2);
  if (t7) {
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3148(t3);
   return t7;
  }
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t10 = t9;
  t11 = t10;
  t12 = (struct array_hash_map_IndexHeader__880 *const *)&t11;
  t10 = (*t12);
  t13 = (*t10);
  t14 = array_hash_map_IndexHeader_capacity__3139(t13);
  t4 = a2;
  t15 = (uint64_t)t14;
  t5 = t4 <= t15;
  if (t5) {
   t2 = (*t1);
   t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t2->entries;
   t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3314(t6, a1, a2);
   if (t7) {
    t2 = (*t1);
    t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
    debug_SafetyLock_unlock__3148(t3);
    return t7;
   }
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3148(t3);
   return 0;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t2->entries;
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3314(t6, a1, a2);
 if (t7) {
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t16 = array_hash_map_IndexHeader_findBitIndex__3142(a2);
 if (t16.error) {
  t7 = t16.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t17 = t16.payload;
 t18 = array_hash_map_IndexHeader_alloc__3143(a1, t17);
 if (t18.error) {
  t7 = t18.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t10 = t18.payload;
 t8 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t19 = t9;
  t20 = t19;
  t12 = (struct array_hash_map_IndexHeader__880 *const *)&t20;
  t19 = (*t12);
  array_hash_map_IndexHeader_free__3144(t19, a1);
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t2 = (*t1);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3280(t2, t10);
 t2 = (*t1);
 t8 = (struct array_hash_map_IndexHeader__880 **)&t2->index_header;
 t19 = (struct array_hash_map_IndexHeader__880 *)t10;
 (*t8) = t19;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3148(t3);
 return 0;
}

static nav__5939_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5939(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5939_41 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_IndexHeader__880 *t3;
 struct array_hash_map_IndexHeader__880 *t5;
 struct array_hash_map_IndexHeader__880 *t24;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t11;
 nav__5939_61 t12;
 nav__5939_64 t13;
 uintptr_t t15;
 uintptr_t t16;
 uintptr_t t14;
 uint64_t t17;
 uint64_t t18;
 nav__5939_41 *t19;
 nav__5939_41 t22;
 nav__5939_38 t23;
 struct array_hash_map_IndexHeader__880 *const *t25;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 uint32_t t6;
 uint32_t t20;
 struct array_hash_map_IndexHeader__880 t26;
 bool t4;
 bool t21;
 uint8_t t27;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t3 = a0.index_header;
 t4 = t3 != NULL;
 if (t4) {
  t5 = t3;
  t2 = t5;
  goto zig_block_0;
 }
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t7 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *)&t1->entries;
 t8 = (*t7);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t8);
 t10 = t9;
 t11 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t10;
 t9 = (*t11);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t9);
 t9 = (*t11);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t9);
 t14 = (uintptr_t)0ul;
 t15 = t13.len;
 zig_loop_32:
 t16 = t14;
 t17 = t16;
 t18 = t15;
 t4 = t17 < t18;
 if (t4) {
  t19 = &t13.ptr[t16];
  t20 = t12.ptr[t16];
  t4 = t20 == t6;
  if (t4) {
   t22 = (*t19);
   t4 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t22, t16);
   t21 = t4;
   goto zig_block_4;
  }
  t21 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t21) {
   t23.is_null = false;
   t23.payload = t16;
   return t23;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t16 = t16 + (uintptr_t)1ul;
 t14 = t16;
 goto zig_loop_32;

 zig_block_1:;
 return (nav__5939_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t24 = t2;
 t25 = (struct array_hash_map_IndexHeader__880 *const *)&t24;
 t5 = (*t25);
 t26 = (*t5);
 t27 = array_hash_map_IndexHeader_capacityIndexType__3138(t26);
 switch (t27) {
  case UINT8_C(0): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5953(t28, a1, t2);
   return t23;
  }
  case UINT8_C(1): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5954(t28, a1, t2);
   return t23;
  }
  case UINT8_C(2): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5955(t28, a1, t2);
   return t23;
  }
  default: zig_unreachable();
 }
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5940(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, nav__5940_42 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t6;
 struct array_hash_map_IndexHeader__880 *t39;
 struct array_hash_map_IndexHeader__880 **t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t13;
 nav__5940_66 t14;
 nav__5940_66 t15;
 nav__5940_66 const *t16;
 nav__5940_69 t17;
 nav__5940_69 t18;
 nav__5940_69 const *t19;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t20;
 uint64_t t23;
 uint64_t t24;
 nav__5940_42 *t25;
 nav__5940_42 t28;
 nav__5940_72 t29;
 nav__5940_72 t30;
 nav__5940_72 t37;
 nav__5940_72 const *t31;
 struct environment_root_EnvironmentValue__844 *t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t33;
 uint32_t *const *t34;
 uint32_t *t35;
 nav__5940_42 *const *t36;
 struct environment_root_EnvironmentValue__844 *const *t38;
 struct array_hash_map_IndexHeader__880 *const *t40;
 uint32_t t7;
 uint32_t t26;
 struct array_hash_map_IndexHeader__880 t41;
 bool t5;
 bool t27;
 uint8_t t42;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t3 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t4 = (*t3);
 t5 = t4 != NULL;
 if (t5) {
  t6 = t4;
  t2 = t6;
  goto zig_block_0;
 }
 t7 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t8->entries;
 t10 = (*t9);
 t11 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t10);
 t12 = t11;
 t13 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t12;
 t11 = (*t13);
 t14 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t11);
 t15 = t14;
 t16 = (nav__5940_66 const *)&t15;
 t11 = (*t13);
 t17 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t11);
 t18 = t17;
 t19 = (nav__5940_69 const *)&t18;
 t20 = (uintptr_t)0ul;
 t21 = t17.len;
 zig_loop_39:
 t22 = t20;
 t23 = t22;
 t24 = t21;
 t5 = t23 < t24;
 if (t5) {
  t25 = &t17.ptr[t22];
  t26 = t14.ptr[t22];
  t5 = t26 == t7;
  if (t5) {
   t28 = (*t25);
   t5 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t28, t22);
   t27 = t5;
   goto zig_block_4;
  }
  t27 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t27) {
   t11 = (*t13);
   t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t11);
   t30 = t29;
   t31 = (nav__5940_72 const *)&t30;
   t29 = (*t31);
   t32 = &t29.ptr[t22];
   t33.key_ptr = t25;
   t33.value_ptr = t32;
   t33.index = t22;
   t33.found_existing = true;
   return t33;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t22 = t22 + (uintptr_t)1ul;
 t20 = t22;
 goto zig_loop_39;

 zig_block_1:;
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t8->entries;
 t21 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t9);
 t34 = &t16->ptr;
 t35 = (*t34);
 t35 = (uint32_t *)&t35[t21];
 (*t35) = t7;
 t36 = &t19->ptr;
 t25 = (*t36);
 t25 = (nav__5940_42 *)&t25[t21];
 t11 = (*t13);
 t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t11);
 t37 = t29;
 t31 = (nav__5940_72 const *)&t37;
 t38 = &t31->ptr;
 t32 = (*t38);
 t32 = (struct environment_root_EnvironmentValue__844 *)&t32[t21];
 t33.key_ptr = t25;
 t33.value_ptr = t32;
 t33.index = t21;
 t33.found_existing = false;
 return t33;

 zig_block_0:;
 t39 = t2;
 t40 = (struct array_hash_map_IndexHeader__880 *const *)&t39;
 t6 = (*t40);
 t41 = (*t6);
 t42 = array_hash_map_IndexHeader_capacityIndexType__3138(t41);
 switch (t42) {
  case UINT8_C(0): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5956(t8, a1, t2);
   return t33;
  }
  case UINT8_C(1): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5957(t8, a1, t2);
   return t33;
  }
  case UINT8_C(2): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5958(t8, a1, t2);
   return t33;
  }
  default: zig_unreachable();
 }
}

static void debug_SafetyLock_lock__3147(void *const a0) {
 void *t0;
 t0 = a0;
 return;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3124(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *t1;
 struct mem_Allocator__599 const *t3;
 uintptr_t *t4;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t2;
 nav__3124_54 t10;
 nav__3124_52 t12;
 nav__3124_52 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t18;
 uint8_t **t16;
 uint8_t *t17;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t19;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t20;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t22;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t21;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t23;
 nav__3124_62 t24;
 nav__3124_62 t25;
 uint32_t *t26;
 nav__3124_67 t27;
 nav__3124_67 t28;
 nav__3124_65 *t29;
 nav__3124_72 t30;
 nav__3124_72 t31;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t32;
 uint16_t t11;
 bool t8;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t6 = a2;
 t7 = t5;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 t9 = (*t3);
 t5 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3131(a2);
 t10 = mem_Allocator_alignedAlloc__anon_7601__5959(t9, t5);
 if (t10.error) {
  t11 = t10.error;
  return t11;
 }
 t12 = t10.payload;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = (*t3);
  t13 = (*t1);
  t14 = (*t13);
  t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3132(t14);
  mem_Allocator_free__anon_6588__5736(t9, t15);
  t13 = (*t1);
  t16 = (uint8_t **)&t13->bytes;
  t17 = t12.ptr;
  (*t16) = t17;
  t13 = (*t1);
  t4 = (uintptr_t *)&t13->capacity;
  (*t4) = a2;
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t17 = t12.ptr;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t14.bytes = t17;
 t14.len = t5;
 t14.capacity = a2;
 t18 = t14;
 t13 = (*t1);
 t14 = (*t13);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t14);
 t20 = t19;
 t21 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t20;
 t14 = t18;
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t14);
 t22 = t19;
 t23 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t22;
 t19 = (*t23);
 t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t19);
 t19 = (*t21);
 t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t19);
 t26 = t25.ptr;
 if (t24.len != 0) memcpy(t24.ptr, t26, t24.len * sizeof(uint32_t));
 t19 = (*t23);
 t27 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t19);
 t19 = (*t21);
 t28 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t19);
 t29 = t28.ptr;
 if (t27.len != 0) memcpy(t27.ptr, t29, t27.len * sizeof(nav__3124_65));
 t19 = (*t23);
 t30 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t19);
 t19 = (*t21);
 t31 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t19);
 t32 = t31.ptr;
 if (t30.len != 0) memcpy(t30.ptr, t32, t30.len * sizeof(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846));
 t9 = (*t3);
 t13 = (*t1);
 t14 = (*t13);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3132(t14);
 mem_Allocator_free__anon_6588__5736(t9, t12);
 t14 = t18;
 (*a0) = t14;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3120(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t0;
 uint16_t t8;
 uint16_t t9;
 bool t6;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a2;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3122(t3, a2);
 t8 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3124(t7, a1, t3);
 memcpy(&t9, &t8, sizeof(uint16_t));
 return t9;
}

static void debug_SafetyLock_unlock__3148(void *const a0) {
 void *t0;
 t0 = a0;
 return;
}

static uint32_t array_hash_map_IndexHeader_capacity__3139(struct array_hash_map_IndexHeader__880 const a0) {
 static uint32_t const t2[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};
 uintptr_t t1;
 uint32_t t3;
 uint8_t t0;
 t0 = a0.bit_index;
 t1 = (uintptr_t)t0;
 t3 = t2[t1];
 return t3;
}

static nav__3142_38 array_hash_map_IndexHeader_findBitIndex__3142(uintptr_t const a0) {
 static uint32_t const t5[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};
 static uint32_t const t8[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};
 uint64_t t0;
 uint64_t t7;
 uintptr_t t4;
 uint32_t t6;
 nav__3142_38 t9;
 bool t1;
 uint8_t t3;
 uint8_t t2;
 t0 = a0;
 t1 = t0 > UINT64_C(4294967295);
 if (t1) {
  return (nav__3142_38){zig_error_OutOfMemory,UINT8_C(0xaa)};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = math_log2_int_ceil__anon_7751__5960(a0);
 t2 = t3;
 t3 = t2;
 t4 = (uintptr_t)t3;
 t6 = t5[t4];
 t0 = a0;
 t7 = (uint64_t)t6;
 t1 = t0 > t7;
 if (t1) {
  t3 = t2;
  t3 = t3 + UINT8_C(1);
  t2 = t3;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t3 = t2;
 t1 = t3 < UINT8_C(5);
 if (t1) {
  t2 = UINT8_C(5);
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t4 = (uintptr_t)t3;
 t6 = t8[t4];
 t7 = a0;
 t0 = (uint64_t)t6;
 t1 = t7 <= t0;
 debug_assert__180(t1);
 t3 = t2;
 t9.payload = t3;
 t9.error = UINT16_C(0);
 return t9;
}

static nav__3143_40 array_hash_map_IndexHeader_alloc__3143(struct mem_Allocator__599 const a0, uint8_t const a1) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t3;
 uintptr_t t4;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 nav__3143_53 t6;
 nav__3143_40 t8;
 nav__3143_51 t9;
 nav__3143_51 t12;
 nav__3143_51 t10;
 nav__3143_51 const *t11;
 uint8_t *t13;
 struct array_hash_map_IndexHeader__880 *t14;
 uint16_t t7;
 uint8_t t2;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = zig_shlw_u64((uintptr_t)1ul, t2, UINT8_C(64));
 t4 = array_hash_map_capacityIndexSize__2907(a1);
 t3 = t4 * t3;
 t3 = (uintptr_t)4ul + t3;
 t5 = (*t1);
 t6 = mem_Allocator_alignedAlloc__anon_7756__5961(t5, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = ((struct array_hash_map_IndexHeader__880 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul);
  t8.error = t7;
  return t8;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (nav__3143_51 const *)&t10;
 t12 = (*t11);
 t13 = t12.ptr;
 t13 = (uint8_t *)(((uintptr_t)t13) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = t12.len;
 t3 = t3 - (uintptr_t)4ul;
 t12.ptr = t13;
 t12.len = t3;
 memset(t12.ptr, UINT8_MAX, t12.len);
 t13 = t9.ptr;
 t14 = (struct array_hash_map_IndexHeader__880 *)t13;
 t13 = (uint8_t *)&t14->bit_index;
 (*t13) = a1;
 t8.payload = t14;
 t8.error = UINT16_C(0);
 return t8;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3069(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct array_hash_map_IndexHeader__880 t5;
 uint8_t t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = (*t4);
 t6 = array_hash_map_IndexHeader_capacityIndexType__3138(t5);
 switch (t6) {
  case UINT8_C(0): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5962(t7, a1);
   return;
  }
  case UINT8_C(1): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5963(t7, a1);
   return;
  }
  case UINT8_C(2): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5964(t7, a1);
   return;
  }
  default: zig_unreachable();
 }
}

static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(nav__5941_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_StringContext_hash__2938(a0);
 return t0;
}

static nav__5942_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 uint32_t *t5;
 uint32_t *t6;
 uint32_t *const *t7;
 nav__5942_39 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5942_39){(uint32_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)0ul];
 t5 = (uint32_t *)t4;
 t6 = t5;
 t7 = (uint32_t *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (uint32_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint32_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(nav__5943_40 const a0, nav__5943_40 const a1, uintptr_t const a2) {
 bool t0;
 t0 = array_hash_map_StringContext_eql__2939(a0, a1, a2);
 return t0;
}

static uint8_t array_hash_map_IndexHeader_capacityIndexType__3138(struct array_hash_map_IndexHeader__880 const a0) {
 uint8_t t0;
 t0 = a0.bit_index;
 t0 = array_hash_map_capacityIndexType__2906(t0);
 return t0;
}

static nav__5944_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5944(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5944_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5944_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 nav__5944_38 t8;
 bool t9;
 struct array_hash_map_Index_28u8_29__6669 t11;
 uint8_t t12;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5966(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__5944_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__5945_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5945(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5945_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5945_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 nav__5945_38 t8;
 struct array_hash_map_Index_28u16_29__6673 t11;
 uint16_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5968(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__5945_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__5946_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5946(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5946_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5946_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 nav__5946_38 t8;
 struct array_hash_map_Index_28u32_29__6677 t11;
 uint32_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5970(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__5946_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t0;
 bool t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *const *)&t0;
 t2 = (uintptr_t *)&a0->len;
 t3 = (*t2);
 t2 = (uintptr_t *)&a0->capacity;
 t4 = (*t2);
 t5 = t3;
 t6 = t4;
 t7 = t5 < t6;
 debug_assert__180(t7);
 t2 = (uintptr_t *)&a0->len;
 t4 = (*t2);
 t8 = (*t1);
 t2 = (uintptr_t *)&t8->len;
 t3 = (*t2);
 t3 = t3 + (uintptr_t)1ul;
 (*t2) = t3;
 return t4;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5947(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, nav__5947_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t9;
 nav__5947_66 t10;
 nav__5947_66 t11;
 nav__5947_66 const *t12;
 nav__5947_69 t13;
 nav__5947_69 t14;
 nav__5947_69 const *t15;
 nav__5947_72 t16;
 nav__5947_72 t17;
 nav__5947_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5947_77 t20;
 nav__5947_77 t55;
 nav__5947_77 t21;
 nav__5947_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u8_29__6669 *t35;
 uint8_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__5947_42 **t42;
 nav__5947_42 *const *t43;
 nav__5947_42 *t44;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **t45;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t46;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t47;
 uintptr_t *t48;
 nav__5947_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__880 t32;
 bool t30;
 bool t50;
 struct array_hash_map_Index_28u8_29__6669 t34;
 struct array_hash_map_Index_28u8_29__6669 t33;
 uint8_t t37;
 uint8_t t53;
 uint8_t t27;
 uint8_t t56;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t7);
 t11 = t10;
 t12 = (nav__5947_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t7);
 t14 = t13;
 t15 = (nav__5947_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t7);
 t17 = t16;
 t18 = (nav__5947_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t19);
 t21 = t20;
 t22 = (nav__5947_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT8_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u8_29_isEmpty__5743(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint8_t *)&t35->entry_index;
   t37 = (uint8_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__5947_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__5947_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__5947_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint8_t *)&t35->entry_index;
   t53 = (uint8_t)t24;
   (*t36) = t53;
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT8_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u8_29_isEmpty__5743(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint8_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__5947_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__5947_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint8_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT8_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT8_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5948(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, nav__5948_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t9;
 nav__5948_66 t10;
 nav__5948_66 t11;
 nav__5948_66 const *t12;
 nav__5948_69 t13;
 nav__5948_69 t14;
 nav__5948_69 const *t15;
 nav__5948_72 t16;
 nav__5948_72 t17;
 nav__5948_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5948_77 t20;
 nav__5948_77 t55;
 nav__5948_77 t21;
 nav__5948_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u16_29__6673 *t35;
 uint16_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__5948_42 **t42;
 nav__5948_42 *const *t43;
 nav__5948_42 *t44;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **t45;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t46;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t47;
 uintptr_t *t48;
 nav__5948_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__880 t32;
 struct array_hash_map_Index_28u16_29__6673 t34;
 struct array_hash_map_Index_28u16_29__6673 t33;
 uint16_t t37;
 uint16_t t53;
 uint16_t t27;
 uint16_t t56;
 bool t30;
 bool t50;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t7);
 t11 = t10;
 t12 = (nav__5948_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t7);
 t14 = t13;
 t15 = (nav__5948_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t7);
 t17 = t16;
 t18 = (nav__5948_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t19);
 t21 = t20;
 t22 = (nav__5948_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT16_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u16_29_isEmpty__5748(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint16_t *)&t35->entry_index;
   t37 = (uint16_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__5948_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__5948_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__5948_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint16_t *)&t35->entry_index;
   t53 = (uint16_t)t24;
   (*t36) = t53;
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT16_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u16_29_isEmpty__5748(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint16_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__5948_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__5948_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint16_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT16_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT16_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5949(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, nav__5949_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t9;
 nav__5949_66 t10;
 nav__5949_66 t11;
 nav__5949_66 const *t12;
 nav__5949_69 t13;
 nav__5949_69 t14;
 nav__5949_69 const *t15;
 nav__5949_72 t16;
 nav__5949_72 t17;
 nav__5949_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5949_77 t20;
 nav__5949_77 t53;
 nav__5949_77 t21;
 nav__5949_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t52;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u32_29__6677 *t35;
 uint32_t *t36;
 uint32_t *const *t38;
 bool *t40;
 nav__5949_42 **t41;
 nav__5949_42 *const *t42;
 nav__5949_42 *t43;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **t44;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *const *t45;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t46;
 uintptr_t *t47;
 nav__5949_42 t50;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t39;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7392 t55;
 uint32_t t23;
 uint32_t t37;
 uint32_t t48;
 uint32_t t27;
 uint32_t t54;
 struct array_hash_map_IndexHeader__880 t32;
 struct array_hash_map_Index_28u32_29__6677 t34;
 struct array_hash_map_Index_28u32_29__6677 t33;
 bool t30;
 bool t49;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t7);
 t11 = t10;
 t12 = (nav__5949_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t7);
 t14 = t13;
 t15 = (nav__5949_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5644(t7);
 t17 = t16;
 t18 = (nav__5949_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t19);
 t21 = t20;
 t22 = (nav__5949_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT32_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u32_29_isEmpty__5753(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint32_t *)&t35->entry_index;
   t37 = (uint32_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t25];
   (*t36) = t23;
   t40 = (bool *)&t39.found_existing;
   (*t40) = false;
   t41 = (nav__5949_42 **)&t39.key_ptr;
   t42 = &t15->ptr;
   t43 = (*t42);
   t43 = (nav__5949_42 *)&t43[t25];
   (*t41) = t43;
   t44 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t39.value_ptr;
   t45 = &t18->ptr;
   t46 = (*t45);
   t46 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t46[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t39.index;
   (*t47) = t25;
   return t39;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t48 = t10.ptr[t24];
  t30 = t23 == t48;
  if (t30) {
   t24 = (uintptr_t)t37;
   t50 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t50, t24);
   t49 = t30;
   goto zig_block_4;
  }
  t49 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t49) {
   t40 = (bool *)&t51.found_existing;
   (*t40) = true;
   t41 = (nav__5949_42 **)&t51.key_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t13 = (*t15);
   t43 = &t13.ptr[t25];
   (*t41) = t43;
   t44 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t51.value_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t16 = (*t18);
   t46 = &t16.ptr[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t51.index;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   (*t47) = t25;
   return t51;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t48 = t27;
  t30 = t37 < t48;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3109(t5);
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t24];
   (*t36) = t23;
   t52 = t31;
   t53 = (*t22);
   t35 = &t53.ptr[t52];
   t36 = (uint32_t *)&t35->entry_index;
   t23 = (uint32_t)t24;
   (*t36) = t23;
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t23 = t27;
   (*t36) = t23;
   t34 = t33;
   t23 = t34.distance_from_start_index;
   t27 = t23;
   t34 = t33;
   t23 = t34.entry_index;
   t54 = t23;
   t52 = t26;
   t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
   t26 = t52;
   t23 = t27;
   t23 = t23 + UINT32_C(1);
   t27 = t23;
   zig_loop_200:
   t52 = t26;
   t29 = t52;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t52 = t26;
    t52 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t52);
    t31 = t52;
    t52 = t31;
    t34 = t20.ptr[t52];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u32_29_isEmpty__5753(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint32_t *)&t35->entry_index;
     t23 = t54;
     (*t36) = t23;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t23 = t27;
     (*t36) = t23;
     t40 = (bool *)&t55.found_existing;
     (*t40) = false;
     t41 = (nav__5949_42 **)&t55.key_ptr;
     t42 = &t15->ptr;
     t43 = (*t42);
     t43 = (nav__5949_42 *)&t43[t24];
     (*t41) = t43;
     t44 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 **)&t55.value_ptr;
     t45 = &t18->ptr;
     t46 = (*t45);
     t46 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *)&t46[t24];
     (*t44) = t46;
     t47 = (uintptr_t *)&t55.index;
     (*t47) = t24;
     return t55;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t23 = t34.distance_from_start_index;
    t48 = t27;
    t30 = t23 < t48;
    if (t30) {
     t52 = t31;
     t53 = (*t22);
     t35 = &t53.ptr[t52];
     t36 = (uint32_t *)&t35->entry_index;
     t48 = t54;
     (*t36) = t48;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t48 = t27;
     (*t36) = t48;
     t34 = t33;
     t48 = t34.entry_index;
     t54 = t48;
     t34 = t33;
     t48 = t34.distance_from_start_index;
     t27 = t48;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t52 = t26;
    t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
    t26 = t52;
    t23 = t27;
    t23 = t23 + UINT32_C(1);
    t27 = t23;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t52 = t26;
  t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
  t26 = t52;
  t37 = t27;
  t37 = t37 + UINT32_C(1);
  t27 = t37;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3318(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *t1;
 struct mem_Allocator__599 const *t3;
 uintptr_t *t4;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t2;
 nav__3318_54 t10;
 nav__3318_52 t12;
 nav__3318_52 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t18;
 uint8_t **t16;
 uint8_t *t17;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t19;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t20;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t22;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t21;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t23;
 nav__3318_62 t24;
 nav__3318_62 t25;
 uint32_t *t26;
 nav__3318_67 t27;
 nav__3318_67 t28;
 nav__3318_65 *t29;
 nav__3318_72 t30;
 nav__3318_72 t31;
 struct environment_root_EnvironmentValue__844 *t32;
 uint16_t t11;
 bool t8;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t6 = a2;
 t7 = t5;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 t9 = (*t3);
 t5 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3325(a2);
 t10 = mem_Allocator_alignedAlloc__anon_7601__5959(t9, t5);
 if (t10.error) {
  t11 = t10.error;
  return t11;
 }
 t12 = t10.payload;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = (*t3);
  t13 = (*t1);
  t14 = (*t13);
  t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3326(t14);
  mem_Allocator_free__anon_6588__5736(t9, t15);
  t13 = (*t1);
  t16 = (uint8_t **)&t13->bytes;
  t17 = t12.ptr;
  (*t16) = t17;
  t13 = (*t1);
  t4 = (uintptr_t *)&t13->capacity;
  (*t4) = a2;
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t17 = t12.ptr;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t14.bytes = t17;
 t14.len = t5;
 t14.capacity = a2;
 t18 = t14;
 t13 = (*t1);
 t14 = (*t13);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t14);
 t20 = t19;
 t21 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t20;
 t14 = t18;
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t14);
 t22 = t19;
 t23 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t22;
 t19 = (*t23);
 t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t19);
 t19 = (*t21);
 t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t19);
 t26 = t25.ptr;
 if (t24.len != 0) memcpy(t24.ptr, t26, t24.len * sizeof(uint32_t));
 t19 = (*t23);
 t27 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t19);
 t19 = (*t21);
 t28 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t19);
 t29 = t28.ptr;
 if (t27.len != 0) memcpy(t27.ptr, t29, t27.len * sizeof(nav__3318_65));
 t19 = (*t23);
 t30 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t19);
 t19 = (*t21);
 t31 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t19);
 t32 = t31.ptr;
 if (t30.len != 0) memcpy(t30.ptr, t32, t30.len * sizeof(struct environment_root_EnvironmentValue__844));
 t9 = (*t3);
 t13 = (*t1);
 t14 = (*t13);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3326(t14);
 mem_Allocator_free__anon_6588__5736(t9, t12);
 t14 = t18;
 (*a0) = t14;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3314(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t0;
 uint16_t t8;
 uint16_t t9;
 bool t6;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a2;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3316(t3, a2);
 t8 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3318(t7, a1, t3);
 memcpy(&t9, &t8, sizeof(uint16_t));
 return t9;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3280(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct array_hash_map_IndexHeader__880 t5;
 uint8_t t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = (*t4);
 t6 = array_hash_map_IndexHeader_capacityIndexType__3138(t5);
 switch (t6) {
  case UINT8_C(0): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5972(t7, a1);
   return;
  }
  case UINT8_C(1): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5973(t7, a1);
   return;
  }
  case UINT8_C(2): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5974(t7, a1);
   return;
  }
  default: zig_unreachable();
 }
}

static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(nav__5950_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_StringContext_hash__2938(a0);
 return t0;
}

static nav__5951_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 uint32_t *t5;
 uint32_t *t6;
 uint32_t *const *t7;
 nav__5951_39 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__5951_39){(uint32_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)0ul];
 t5 = (uint32_t *)t4;
 t6 = t5;
 t7 = (uint32_t *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (uint32_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint32_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(nav__5952_40 const a0, nav__5952_40 const a1, uintptr_t const a2) {
 bool t0;
 t0 = array_hash_map_StringContext_eql__2939(a0, a1, a2);
 return t0;
}

static nav__5953_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5953(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5953_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5953_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 nav__5953_38 t8;
 bool t9;
 struct array_hash_map_Index_28u8_29__6669 t11;
 uint8_t t12;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5975(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__5953_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__5954_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5954(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5954_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5954_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 nav__5954_38 t8;
 struct array_hash_map_Index_28u16_29__6673 t11;
 uint16_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5976(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__5954_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__5955_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5955(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5955_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5955_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 nav__5955_38 t8;
 struct array_hash_map_Index_28u32_29__6677 t11;
 uint32_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5977(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__5955_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t0;
 bool t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *const *)&t0;
 t2 = (uintptr_t *)&a0->len;
 t3 = (*t2);
 t2 = (uintptr_t *)&a0->capacity;
 t4 = (*t2);
 t5 = t3;
 t6 = t4;
 t7 = t5 < t6;
 debug_assert__180(t7);
 t2 = (uintptr_t *)&a0->len;
 t4 = (*t2);
 t8 = (*t1);
 t2 = (uintptr_t *)&t8->len;
 t3 = (*t2);
 t3 = t3 + (uintptr_t)1ul;
 (*t2) = t3;
 return t4;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5956(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, nav__5956_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t9;
 nav__5956_66 t10;
 nav__5956_66 t11;
 nav__5956_66 const *t12;
 nav__5956_69 t13;
 nav__5956_69 t14;
 nav__5956_69 const *t15;
 nav__5956_72 t16;
 nav__5956_72 t17;
 nav__5956_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5956_77 t20;
 nav__5956_77 t55;
 nav__5956_77 t21;
 nav__5956_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u8_29__6669 *t35;
 uint8_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__5956_42 **t42;
 nav__5956_42 *const *t43;
 nav__5956_42 *t44;
 struct environment_root_EnvironmentValue__844 **t45;
 struct environment_root_EnvironmentValue__844 *const *t46;
 struct environment_root_EnvironmentValue__844 *t47;
 uintptr_t *t48;
 nav__5956_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__880 t32;
 bool t30;
 bool t50;
 struct array_hash_map_Index_28u8_29__6669 t34;
 struct array_hash_map_Index_28u8_29__6669 t33;
 uint8_t t37;
 uint8_t t53;
 uint8_t t27;
 uint8_t t56;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t7);
 t11 = t10;
 t12 = (nav__5956_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t7);
 t14 = t13;
 t15 = (nav__5956_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t7);
 t17 = t16;
 t18 = (nav__5956_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t19);
 t21 = t20;
 t22 = (nav__5956_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT8_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u8_29_isEmpty__5743(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint8_t *)&t35->entry_index;
   t37 = (uint8_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__5956_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__5956_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__844 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_root_EnvironmentValue__844 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__5956_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__844 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint8_t *)&t35->entry_index;
   t53 = (uint8_t)t24;
   (*t36) = t53;
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT8_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u8_29_isEmpty__5743(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint8_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__5956_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__5956_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_root_EnvironmentValue__844 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_root_EnvironmentValue__844 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint8_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT8_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT8_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5957(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, nav__5957_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t9;
 nav__5957_66 t10;
 nav__5957_66 t11;
 nav__5957_66 const *t12;
 nav__5957_69 t13;
 nav__5957_69 t14;
 nav__5957_69 const *t15;
 nav__5957_72 t16;
 nav__5957_72 t17;
 nav__5957_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5957_77 t20;
 nav__5957_77 t55;
 nav__5957_77 t21;
 nav__5957_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u16_29__6673 *t35;
 uint16_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__5957_42 **t42;
 nav__5957_42 *const *t43;
 nav__5957_42 *t44;
 struct environment_root_EnvironmentValue__844 **t45;
 struct environment_root_EnvironmentValue__844 *const *t46;
 struct environment_root_EnvironmentValue__844 *t47;
 uintptr_t *t48;
 nav__5957_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__880 t32;
 struct array_hash_map_Index_28u16_29__6673 t34;
 struct array_hash_map_Index_28u16_29__6673 t33;
 uint16_t t37;
 uint16_t t53;
 uint16_t t27;
 uint16_t t56;
 bool t30;
 bool t50;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t7);
 t11 = t10;
 t12 = (nav__5957_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t7);
 t14 = t13;
 t15 = (nav__5957_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t7);
 t17 = t16;
 t18 = (nav__5957_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t19);
 t21 = t20;
 t22 = (nav__5957_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT16_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u16_29_isEmpty__5748(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint16_t *)&t35->entry_index;
   t37 = (uint16_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__5957_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__5957_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__844 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_root_EnvironmentValue__844 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__5957_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__844 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint16_t *)&t35->entry_index;
   t53 = (uint16_t)t24;
   (*t36) = t53;
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT16_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u16_29_isEmpty__5748(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint16_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__5957_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__5957_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_root_EnvironmentValue__844 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_root_EnvironmentValue__844 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint16_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT16_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT16_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5958(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, nav__5958_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t9;
 nav__5958_66 t10;
 nav__5958_66 t11;
 nav__5958_66 const *t12;
 nav__5958_69 t13;
 nav__5958_69 t14;
 nav__5958_69 const *t15;
 nav__5958_72 t16;
 nav__5958_72 t17;
 nav__5958_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5958_77 t20;
 nav__5958_77 t53;
 nav__5958_77 t21;
 nav__5958_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t52;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u32_29__6677 *t35;
 uint32_t *t36;
 uint32_t *const *t38;
 bool *t40;
 nav__5958_42 **t41;
 nav__5958_42 *const *t42;
 nav__5958_42 *t43;
 struct environment_root_EnvironmentValue__844 **t44;
 struct environment_root_EnvironmentValue__844 *const *t45;
 struct environment_root_EnvironmentValue__844 *t46;
 uintptr_t *t47;
 nav__5958_42 t50;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t39;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7403 t55;
 uint32_t t23;
 uint32_t t37;
 uint32_t t48;
 uint32_t t27;
 uint32_t t54;
 struct array_hash_map_IndexHeader__880 t32;
 struct array_hash_map_Index_28u32_29__6677 t34;
 struct array_hash_map_Index_28u32_29__6677 t33;
 bool t30;
 bool t49;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t7);
 t11 = t10;
 t12 = (nav__5958_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t7);
 t14 = t13;
 t15 = (nav__5958_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5653(t7);
 t17 = t16;
 t18 = (nav__5958_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t19);
 t21 = t20;
 t22 = (nav__5958_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT32_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u32_29_isEmpty__5753(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint32_t *)&t35->entry_index;
   t37 = (uint32_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t25];
   (*t36) = t23;
   t40 = (bool *)&t39.found_existing;
   (*t40) = false;
   t41 = (nav__5958_42 **)&t39.key_ptr;
   t42 = &t15->ptr;
   t43 = (*t42);
   t43 = (nav__5958_42 *)&t43[t25];
   (*t41) = t43;
   t44 = (struct environment_root_EnvironmentValue__844 **)&t39.value_ptr;
   t45 = &t18->ptr;
   t46 = (*t45);
   t46 = (struct environment_root_EnvironmentValue__844 *)&t46[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t39.index;
   (*t47) = t25;
   return t39;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t48 = t10.ptr[t24];
  t30 = t23 == t48;
  if (t30) {
   t24 = (uintptr_t)t37;
   t50 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t50, t24);
   t49 = t30;
   goto zig_block_4;
  }
  t49 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t49) {
   t40 = (bool *)&t51.found_existing;
   (*t40) = true;
   t41 = (nav__5958_42 **)&t51.key_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t13 = (*t15);
   t43 = &t13.ptr[t25];
   (*t41) = t43;
   t44 = (struct environment_root_EnvironmentValue__844 **)&t51.value_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t16 = (*t18);
   t46 = &t16.ptr[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t51.index;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   (*t47) = t25;
   return t51;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t48 = t27;
  t30 = t37 < t48;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3303(t5);
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t24];
   (*t36) = t23;
   t52 = t31;
   t53 = (*t22);
   t35 = &t53.ptr[t52];
   t36 = (uint32_t *)&t35->entry_index;
   t23 = (uint32_t)t24;
   (*t36) = t23;
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t23 = t27;
   (*t36) = t23;
   t34 = t33;
   t23 = t34.distance_from_start_index;
   t27 = t23;
   t34 = t33;
   t23 = t34.entry_index;
   t54 = t23;
   t52 = t26;
   t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
   t26 = t52;
   t23 = t27;
   t23 = t23 + UINT32_C(1);
   t27 = t23;
   zig_loop_200:
   t52 = t26;
   t29 = t52;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t52 = t26;
    t52 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t52);
    t31 = t52;
    t52 = t31;
    t34 = t20.ptr[t52];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u32_29_isEmpty__5753(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint32_t *)&t35->entry_index;
     t23 = t54;
     (*t36) = t23;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t23 = t27;
     (*t36) = t23;
     t40 = (bool *)&t55.found_existing;
     (*t40) = false;
     t41 = (nav__5958_42 **)&t55.key_ptr;
     t42 = &t15->ptr;
     t43 = (*t42);
     t43 = (nav__5958_42 *)&t43[t24];
     (*t41) = t43;
     t44 = (struct environment_root_EnvironmentValue__844 **)&t55.value_ptr;
     t45 = &t18->ptr;
     t46 = (*t45);
     t46 = (struct environment_root_EnvironmentValue__844 *)&t46[t24];
     (*t44) = t46;
     t47 = (uintptr_t *)&t55.index;
     (*t47) = t24;
     return t55;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t23 = t34.distance_from_start_index;
    t48 = t27;
    t30 = t23 < t48;
    if (t30) {
     t52 = t31;
     t53 = (*t22);
     t35 = &t53.ptr[t52];
     t36 = (uint32_t *)&t35->entry_index;
     t48 = t54;
     (*t36) = t48;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t48 = t27;
     (*t36) = t48;
     t34 = t33;
     t48 = t34.entry_index;
     t54 = t48;
     t34 = t33;
     t48 = t34.distance_from_start_index;
     t27 = t48;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t52 = t26;
    t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
    t26 = t52;
    t23 = t27;
    t23 = t23 + UINT32_C(1);
    t27 = t23;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t52 = t26;
  t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
  t26 = t52;
  t37 = t27;
  t37 = t37 + UINT32_C(1);
  t27 = t37;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3122(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)4ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static nav__5959_40 mem_Allocator_alignedAlloc__anon_7601__5959(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__5959_40 t4;
 nav__5959_40 t8;
 nav__5959_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__5959_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_7871__5978(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__5959_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static uint8_t math_log2_int_ceil__anon_7751__5960(uintptr_t const a0) {
 uint64_t t0;
 uintptr_t t2;
 bool t1;
 uint8_t t3;
 t0 = a0;
 t1 = t0 != UINT64_C(0);
 debug_assert__180(t1);
 t0 = a0;
 t1 = t0 == UINT64_C(1);
 if (t1) {
  return UINT8_C(0);
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = a0 - (uintptr_t)1ul;
 t3 = math_log2_int__anon_7874__5979(t2);
 t3 = t3 + UINT8_C(1);
 return t3;
}

static nav__5961_40 mem_Allocator_alignedAlloc__anon_7756__5961(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__5961_40 t4;
 nav__5961_40 t8;
 nav__5961_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__5961_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_7878__5980(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__5961_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5962(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t9;
 nav__5962_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5962_60 t12;
 nav__5962_60 t33;
 nav__5962_60 t13;
 nav__5962_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u8_29__6669 const *t31;
 struct array_hash_map_Index_28u8_29__6669 *t34;
 uint8_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__880 t28;
 bool t20;
 uint8_t t26;
 uint8_t t36;
 uint8_t t25;
 uint8_t t27;
 struct array_hash_map_Index_28u8_29__6669 t29;
 struct array_hash_map_Index_28u8_29__6669 t32;
 struct array_hash_map_Index_28u8_29__6669 t30;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t11);
 t13 = t12;
 t14 = (nav__5962_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint8_t)t17;
  t25 = t26;
  t27 = UINT8_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u8_29__6669 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u8_29_isEmpty__5743(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint8_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint8_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT8_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5963(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t9;
 nav__5963_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5963_60 t12;
 nav__5963_60 t33;
 nav__5963_60 t13;
 nav__5963_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u16_29__6673 const *t31;
 struct array_hash_map_Index_28u16_29__6673 *t34;
 uint16_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__880 t28;
 uint16_t t26;
 uint16_t t36;
 uint16_t t25;
 uint16_t t27;
 struct array_hash_map_Index_28u16_29__6673 t29;
 struct array_hash_map_Index_28u16_29__6673 t32;
 struct array_hash_map_Index_28u16_29__6673 t30;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t11);
 t13 = t12;
 t14 = (nav__5963_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint16_t)t17;
  t25 = t26;
  t27 = UINT16_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u16_29__6673 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u16_29_isEmpty__5748(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint16_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint16_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT16_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5964(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t9;
 nav__5964_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5964_60 t12;
 nav__5964_60 t32;
 nav__5964_60 t13;
 nav__5964_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u32_29__6677 const *t30;
 struct array_hash_map_Index_28u32_29__6677 *t33;
 uint32_t *t34;
 uint32_t t21;
 uint32_t t35;
 uint32_t t25;
 uint32_t t26;
 struct array_hash_map_IndexHeader__880 t27;
 struct array_hash_map_Index_28u32_29__6677 t28;
 struct array_hash_map_Index_28u32_29__6677 t31;
 struct array_hash_map_Index_28u32_29__6677 t29;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t11);
 t13 = t12;
 t14 = (nav__5964_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t21 = (uint32_t)t17;
  t25 = t21;
  t26 = UINT32_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t27 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t27, t22);
   t28 = t12.ptr[t22];
   t29 = t28;
   t30 = (struct array_hash_map_Index_28u32_29__6677 const *)&t29;
   t31 = (*t30);
   t20 = array_hash_map_Index_28u32_29_isEmpty__5753(t31);
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t21 = t26;
    (*t34) = t21;
    t34 = (uint32_t *)&t33->entry_index;
    t21 = t25;
    (*t34) = t21;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t21 = t28.distance_from_start_index;
   t35 = t26;
   t20 = t21 < t35;
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t35 = t26;
    (*t34) = t35;
    t34 = (uint32_t *)&t33->entry_index;
    t35 = t25;
    (*t34) = t35;
    t35 = t28.distance_from_start_index;
    t26 = t35;
    t35 = t28.entry_index;
    t25 = t35;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t35 = t26;
   t35 = t35 + UINT32_C(1);
   t26 = t35;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static uint32_t array_hash_map_StringContext_hash__2938(nav__2938_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_hashString__2901(a0);
 return t0;
}

static bool array_hash_map_StringContext_eql__2939(nav__2939_40 const a0, nav__2939_40 const a1, uintptr_t const a2) {
 bool t0;
 (void)a2;
 t0 = array_hash_map_eqlString__2900(a0, a1);
 return t0;
}

static nav__5965_40 array_hash_map_IndexHeader_indexes__anon_7788__5965(struct array_hash_map_IndexHeader__880 *const a0) {
 struct array_hash_map_IndexHeader__880 *const *t1;
 uint8_t *t2;
 struct array_hash_map_Index_28u8_29__6669 *t3;
 struct array_hash_map_Index_28u8_29__6669 *t4;
 struct array_hash_map_Index_28u8_29__6669 *const *t5;
 struct array_hash_map_IndexHeader__880 *t6;
 struct array_hash_map_IndexHeader__880 *t0;
 uintptr_t t8;
 nav__5965_40 t9;
 struct array_hash_map_IndexHeader__880 t7;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__880 *const *)&t0;
 t2 = (uint8_t *)a0;
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = (struct array_hash_map_Index_28u8_29__6669 *)t2;
 t4 = t3;
 t5 = (struct array_hash_map_Index_28u8_29__6669 *const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_hash_map_IndexHeader_length__3140(t7);
 t3 = (*t5);
 t3 = (struct array_hash_map_Index_28u8_29__6669 *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(struct array_hash_map_Index_28u8_29__6669)));
 t9.ptr = t3;
 t9.len = t8;
 return t9;
}

static nav__5966_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5966(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5966_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__5966_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t8;
 nav__5966_66 t9;
 nav__5966_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u8_29__6669 const *t23;
 nav__5966_41 t30;
 nav__5966_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__880 t20;
 bool t18;
 bool t25;
 struct array_hash_map_Index_28u8_29__6669 t21;
 struct array_hash_map_Index_28u8_29__6669 t24;
 struct array_hash_map_Index_28u8_29__6669 t22;
 uint8_t t26;
 uint8_t t27;
 uint8_t t15;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT8_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u8_29__6669 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u8_29_isEmpty__5743(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__5966_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT8_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__5967_40 array_hash_map_IndexHeader_indexes__anon_7798__5967(struct array_hash_map_IndexHeader__880 *const a0) {
 struct array_hash_map_IndexHeader__880 *const *t1;
 uint8_t *t2;
 struct array_hash_map_Index_28u16_29__6673 *t3;
 struct array_hash_map_Index_28u16_29__6673 *t4;
 struct array_hash_map_Index_28u16_29__6673 *const *t5;
 struct array_hash_map_IndexHeader__880 *t6;
 struct array_hash_map_IndexHeader__880 *t0;
 uintptr_t t8;
 nav__5967_40 t9;
 struct array_hash_map_IndexHeader__880 t7;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__880 *const *)&t0;
 t2 = (uint8_t *)a0;
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = (struct array_hash_map_Index_28u16_29__6673 *)t2;
 t4 = t3;
 t5 = (struct array_hash_map_Index_28u16_29__6673 *const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_hash_map_IndexHeader_length__3140(t7);
 t3 = (*t5);
 t3 = (struct array_hash_map_Index_28u16_29__6673 *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(struct array_hash_map_Index_28u16_29__6673)));
 t9.ptr = t3;
 t9.len = t8;
 return t9;
}

static nav__5968_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5968(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5968_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__5968_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t8;
 nav__5968_66 t9;
 nav__5968_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u16_29__6673 const *t23;
 nav__5968_41 t30;
 nav__5968_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__880 t20;
 struct array_hash_map_Index_28u16_29__6673 t21;
 struct array_hash_map_Index_28u16_29__6673 t24;
 struct array_hash_map_Index_28u16_29__6673 t22;
 uint16_t t26;
 uint16_t t27;
 uint16_t t15;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT16_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u16_29__6673 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u16_29_isEmpty__5748(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__5968_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT16_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__5969_40 array_hash_map_IndexHeader_indexes__anon_7804__5969(struct array_hash_map_IndexHeader__880 *const a0) {
 struct array_hash_map_IndexHeader__880 *const *t1;
 uint8_t *t2;
 struct array_hash_map_Index_28u32_29__6677 *t3;
 struct array_hash_map_Index_28u32_29__6677 *t4;
 struct array_hash_map_Index_28u32_29__6677 *const *t5;
 struct array_hash_map_IndexHeader__880 *t6;
 struct array_hash_map_IndexHeader__880 *t0;
 uintptr_t t8;
 nav__5969_40 t9;
 struct array_hash_map_IndexHeader__880 t7;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__880 *const *)&t0;
 t2 = (uint8_t *)a0;
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = (struct array_hash_map_Index_28u32_29__6677 *)t2;
 t4 = t3;
 t5 = (struct array_hash_map_Index_28u32_29__6677 *const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_hash_map_IndexHeader_length__3140(t7);
 t3 = (*t5);
 t3 = (struct array_hash_map_Index_28u32_29__6677 *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(struct array_hash_map_Index_28u32_29__6677)));
 t9.ptr = t3;
 t9.len = t8;
 return t9;
}

static nav__5970_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5970(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0, nav__5970_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__5970_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *t8;
 nav__5970_66 t9;
 nav__5970_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u32_29__6677 const *t23;
 nav__5970_41 t29;
 nav__5970_38 t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t0;
 uint32_t t11;
 uint32_t t26;
 uint32_t t27;
 uint32_t t15;
 struct array_hash_map_IndexHeader__880 t20;
 struct array_hash_map_Index_28u32_29__6677 t21;
 struct array_hash_map_Index_28u32_29__6677 t24;
 struct array_hash_map_Index_28u32_29__6677 t22;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3102(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6165 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5942(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__5643(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5941(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT32_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u32_29__6677 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u32_29_isEmpty__5753(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__5970_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t26 = t9.ptr[t28];
  t25 = t11 == t26;
  if (t25) {
   t28 = (uintptr_t)t27;
   t29 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5943(a1, t29, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t30.is_null = false;
   t30.payload = t12;
   return t30;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT32_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static uintptr_t array_hash_map_safeTruncate__anon_7812__5971(uint32_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)a0;
 return t0;
}

static uintptr_t array_hash_map_IndexHeader_constrainIndex__3136(struct array_hash_map_IndexHeader__880 const a0, uintptr_t const a1) {
 struct array_hash_map_IndexHeader__880 const *t1;
 uintptr_t t4;
 struct array_hash_map_IndexHeader__880 t2;
 struct array_hash_map_IndexHeader__880 t0;
 uint32_t t3;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__880 const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_IndexHeader_mask__3141(t2);
 t4 = (uintptr_t)t3;
 t4 = a1 & t4;
 return t4;
}

static bool array_hash_map_Index_28u8_29_isEmpty__5743(struct array_hash_map_Index_28u8_29__6669 const a0) {
 uint8_t t0;
 bool t1;
 t0 = a0.entry_index;
 t1 = t0 == UINT8_MAX;
 return t1;
}

static bool array_hash_map_Index_28u16_29_isEmpty__5748(struct array_hash_map_Index_28u16_29__6673 const a0) {
 uint16_t t0;
 bool t1;
 t0 = a0.entry_index;
 t1 = t0 == UINT16_MAX;
 return t1;
}

static bool array_hash_map_Index_28u32_29_isEmpty__5753(struct array_hash_map_Index_28u32_29__6677 const a0) {
 uint32_t t0;
 bool t1;
 t0 = a0.entry_index;
 t1 = t0 == UINT32_MAX;
 return t1;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3316(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)5ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5972(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t9;
 nav__5972_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5972_60 t12;
 nav__5972_60 t33;
 nav__5972_60 t13;
 nav__5972_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u8_29__6669 const *t31;
 struct array_hash_map_Index_28u8_29__6669 *t34;
 uint8_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__880 t28;
 bool t20;
 uint8_t t26;
 uint8_t t36;
 uint8_t t25;
 uint8_t t27;
 struct array_hash_map_Index_28u8_29__6669 t29;
 struct array_hash_map_Index_28u8_29__6669 t32;
 struct array_hash_map_Index_28u8_29__6669 t30;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t11);
 t13 = t12;
 t14 = (nav__5972_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint8_t)t17;
  t25 = t26;
  t27 = UINT8_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u8_29__6669 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u8_29_isEmpty__5743(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint8_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint8_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT8_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5973(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t9;
 nav__5973_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5973_60 t12;
 nav__5973_60 t33;
 nav__5973_60 t13;
 nav__5973_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u16_29__6673 const *t31;
 struct array_hash_map_Index_28u16_29__6673 *t34;
 uint16_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__880 t28;
 uint16_t t26;
 uint16_t t36;
 uint16_t t25;
 uint16_t t27;
 struct array_hash_map_Index_28u16_29__6673 t29;
 struct array_hash_map_Index_28u16_29__6673 t32;
 struct array_hash_map_Index_28u16_29__6673 t30;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t11);
 t13 = t12;
 t14 = (nav__5973_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint16_t)t17;
  t25 = t26;
  t27 = UINT16_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u16_29__6673 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u16_29_isEmpty__5748(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint16_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint16_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT16_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5974(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t9;
 nav__5974_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__5974_60 t12;
 nav__5974_60 t32;
 nav__5974_60 t13;
 nav__5974_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u32_29__6677 const *t30;
 struct array_hash_map_Index_28u32_29__6677 *t33;
 uint32_t *t34;
 uint32_t t21;
 uint32_t t35;
 uint32_t t25;
 uint32_t t26;
 struct array_hash_map_IndexHeader__880 t27;
 struct array_hash_map_Index_28u32_29__6677 t28;
 struct array_hash_map_Index_28u32_29__6677 t31;
 struct array_hash_map_Index_28u32_29__6677 t29;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t11);
 t13 = t12;
 t14 = (nav__5974_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t21 = (uint32_t)t17;
  t25 = t21;
  t26 = UINT32_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t27 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t27, t22);
   t28 = t12.ptr[t22];
   t29 = t28;
   t30 = (struct array_hash_map_Index_28u32_29__6677 const *)&t29;
   t31 = (*t30);
   t20 = array_hash_map_Index_28u32_29_isEmpty__5753(t31);
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t21 = t26;
    (*t34) = t21;
    t34 = (uint32_t *)&t33->entry_index;
    t21 = t25;
    (*t34) = t21;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t21 = t28.distance_from_start_index;
   t35 = t26;
   t20 = t21 < t35;
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t35 = t26;
    (*t34) = t35;
    t34 = (uint32_t *)&t33->entry_index;
    t35 = t25;
    (*t34) = t35;
    t35 = t28.distance_from_start_index;
    t26 = t35;
    t35 = t28.entry_index;
    t25 = t35;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t35 = t26;
   t35 = t35 + UINT32_C(1);
   t26 = t35;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static nav__5975_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5975(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5975_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__5975_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t8;
 nav__5975_66 t9;
 nav__5975_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u8_29__6669 const *t23;
 nav__5975_41 t30;
 nav__5975_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__880 t20;
 bool t18;
 bool t25;
 struct array_hash_map_Index_28u8_29__6669 t21;
 struct array_hash_map_Index_28u8_29__6669 t24;
 struct array_hash_map_Index_28u8_29__6669 t22;
 uint8_t t26;
 uint8_t t27;
 uint8_t t15;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT8_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u8_29__6669 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u8_29_isEmpty__5743(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__5975_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT8_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__5976_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5976(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5976_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__5976_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t8;
 nav__5976_66 t9;
 nav__5976_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u16_29__6673 const *t23;
 nav__5976_41 t30;
 nav__5976_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__880 t20;
 struct array_hash_map_Index_28u16_29__6673 t21;
 struct array_hash_map_Index_28u16_29__6673 t24;
 struct array_hash_map_Index_28u16_29__6673 t22;
 uint16_t t26;
 uint16_t t27;
 uint16_t t15;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT16_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u16_29__6673 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u16_29_isEmpty__5748(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__5976_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT16_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__5977_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5977(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const a0, nav__5977_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__5977_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *t8;
 nav__5977_66 t9;
 nav__5977_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u32_29__6677 const *t23;
 nav__5977_41 t29;
 nav__5977_38 t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t0;
 uint32_t t11;
 uint32_t t26;
 uint32_t t27;
 uint32_t t15;
 struct array_hash_map_IndexHeader__880 t20;
 struct array_hash_map_Index_28u32_29__6677 t21;
 struct array_hash_map_Index_28u32_29__6677 t24;
 struct array_hash_map_Index_28u32_29__6677 t22;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__894 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__3296(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6314 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5951(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__5652(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5950(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT32_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u32_29__6677 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u32_29_isEmpty__5753(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__5977_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t26 = t9.ptr[t28];
  t25 = t11 == t26;
  if (t25) {
   t28 = (uintptr_t)t27;
   t29 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5952(a1, t29, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t30.is_null = false;
   t30.payload = t12;
   return t30;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT32_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__5978_39 mem_Allocator_allocWithSizeAndAlignment__anon_7871__5978(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__5978_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__5978_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3220__4889((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__5978_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_2726__4265(t6, t2, a2);
 return t7;
}

static uint8_t math_log2_int__anon_7874__5979(uintptr_t const a0) {
 uint64_t t0;
 uint16_t t3;
 bool t1;
 uint8_t t2;
 t0 = a0;
 t1 = t0 != UINT64_C(0);
 debug_assert__180(t1);
 t2 = zig_clz_u64(a0, UINT8_C(64));
 t3 = (uint16_t)t2;
 t3 = UINT16_C(63) - t3;
 t2 = (uint8_t)t3;
 return t2;
}

static nav__5980_39 mem_Allocator_allocWithSizeAndAlignment__anon_7878__5980(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__5980_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__5980_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3220__4889((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__5980_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_7904__5981(t6, t2, a2);
 return t7;
}

static uint32_t array_hash_map_hashString__2901(nav__2901_39 const a0) {
 uint64_t t0;
 uint32_t t1;
 t0 = hash_wyhash_Wyhash_hash__6041(UINT64_C(0), a0);
 t1 = (uint32_t)t0;
 return t1;
}

static bool array_hash_map_eqlString__2900(nav__2900_39 const a0, nav__2900_39 const a1) {
 bool t0;
 t0 = mem_eql__anon_7916__6042(a0, a1);
 return t0;
}

static uint32_t array_hash_map_IndexHeader_mask__3141(struct array_hash_map_IndexHeader__880 const a0) {
 struct array_hash_map_IndexHeader__880 const *t1;
 uintptr_t t3;
 struct array_hash_map_IndexHeader__880 t2;
 struct array_hash_map_IndexHeader__880 t0;
 uint32_t t4;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__880 const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_IndexHeader_length__3140(t2);
 t3 = t3 - (uintptr_t)1ul;
 t4 = (uint32_t)t3;
 return t4;
}

static nav__5981_39 mem_Allocator_allocBytesWithAlignment__anon_7904__5981(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t7;
 struct mem_Allocator_VTable__602 const *const *t8;
 struct mem_Allocator_VTable__602 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__5981_52 t17;
 nav__5981_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__5981_39){(uint8_t *)(uintptr_t)0xfffffffffffffffcul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1113((uintptr_t)4ul);
 t7 = t5;
 t1 = (struct mem_Allocator__599 const *)&t7;
 t8 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__5981_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t14 = (uint8_t *)t4;
 t18.payload = t14;
 t18.error = UINT16_C(0);
 return t18;
}

static uint64_t hash_wyhash_Wyhash_hash__6041(uint64_t const a0, nav__6041_39 const a1) {
 zig_u128 t41;
 zig_u128 t42;
 nav__6041_39 const *t1;
 nav__6041_39 const *t17;
 nav__6041_39 const *t61;
 struct hash_wyhash_Wyhash__7909 t3;
 struct hash_wyhash_Wyhash__7909 t2;
 uintptr_t t4;
 uintptr_t t10;
 uintptr_t t15;
 uintptr_t t26;
 uintptr_t t58;
 uint64_t t5;
 uint64_t t22;
 uint64_t t38;
 uint64_t t39;
 uint64_t t40;
 uint64_t t47;
 uint64_t t48;
 uint64_t t52;
 uint64_t t53;
 uint64_t t63;
 uint64_t t64;
 uint64_t t70;
 uint64_t t71;
 struct hash_wyhash_Wyhash__7909 *const *t8;
 struct hash_wyhash_Wyhash__7909 *t11;
 struct hash_wyhash_Wyhash__7909 *t7;
 struct hash_wyhash_Wyhash__7909 *t28;
 struct hash_wyhash_Wyhash__7909 *t54;
 struct hash_wyhash_Wyhash__7909 *t55;
 struct hash_wyhash_Wyhash__7909 *t68;
 uint64_t *t12;
 uint64_t *t69;
 nav__6041_39 t13;
 nav__6041_39 t59;
 nav__6041_39 t0;
 nav__6041_39 t9;
 nav__6041_39 t16;
 nav__6041_39 t21;
 nav__6041_39 t23;
 nav__6041_39 t24;
 nav__6041_39 t31;
 nav__6041_39 t35;
 nav__6041_39 t44;
 nav__6041_39 t46;
 nav__6041_39 t50;
 nav__6041_39 t51;
 nav__6041_39 t56;
 nav__6041_39 t57;
 nav__6041_39 t60;
 nav__6041_39 t62;
 nav__6041_39 t65;
 nav__6041_39 t66;
 uint8_t const *t14;
 uint8_t const (*t18)[4];
 uint8_t const (*t27)[48];
 uint8_t const (*t29)[48];
 uint8_t const (*const *t30)[48];
 uint8_t const (*t32)[8];
 uint8_t const (*t34)[40];
 uint64_t (*t36)[3];
 uint64_t t37[3];
 uint8_t const (*t43)[32];
 uint8_t const (*t45)[24];
 uint8_t const (*t49)[16];
 uintptr_t *t67;
 uint32_t t20;
 bool t6;
 uint8_t t19[4];
 uint8_t t25;
 uint8_t t33[8];
 t0 = a1;
 t1 = (nav__6041_39 const *)&t0;
 t3 = hash_wyhash_Wyhash_init__6029(a0);
 t2 = t3;
 t4 = a1.len;
 t5 = t4;
 t6 = t5 <= UINT64_C(16);
 if (t6) {
  t7 = &t2;
  t8 = (struct hash_wyhash_Wyhash__7909 *const *)&t7;
  t9 = a1;
  t1 = (nav__6041_39 const *)&t9;
  t4 = a1.len;
  t5 = t4;
  t6 = t5 <= UINT64_C(16);
  debug_assert__180(t6);
  t4 = a1.len;
  t5 = t4;
  t6 = t5 >= UINT64_C(4);
  if (t6) {
   t4 = a1.len;
   t4 = t4 - (uintptr_t)4ul;
   t10 = a1.len;
   t10 = zig_shr_u64(t10, UINT8_C(3));
   t10 = zig_shlw_u64(t10, UINT8_C(2), UINT8_C(64));
   t11 = (*t8);
   t12 = (uint64_t *)&t11->a;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t15 = t13.len;
   t13.ptr = t14;
   t13.len = t15;
   t16 = t13;
   t17 = (nav__6041_39 const *)&t16;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t5 = (uint64_t)t20;
   t5 = zig_shlw_u64(t5, UINT8_C(32), UINT8_C(64));
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
   t15 = t13.len;
   t15 = t15 - t10;
   t13.ptr = t14;
   t13.len = t15;
   t21 = t13;
   t17 = (nav__6041_39 const *)&t21;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t22 = (uint64_t)t20;
   t22 = t5 | t22;
   (*t12) = t22;
   t11 = (*t8);
   t12 = (uint64_t *)&t11->b;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t4*sizeof(uint8_t)));
   t15 = t13.len;
   t15 = t15 - t4;
   t13.ptr = t14;
   t13.len = t15;
   t23 = t13;
   t17 = (nav__6041_39 const *)&t23;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t22 = (uint64_t)t20;
   t22 = zig_shlw_u64(t22, UINT8_C(32), UINT8_C(64));
   t10 = t4 - t10;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
   t4 = t13.len;
   t10 = t4 - t10;
   t13.ptr = t14;
   t13.len = t10;
   t24 = t13;
   t1 = (nav__6041_39 const *)&t24;
   debug_assert__180(true);
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t5 = (uint64_t)t20;
   t5 = t22 | t5;
   (*t12) = t5;
   goto zig_block_1;
  }
  t15 = a1.len;
  t22 = t15;
  t6 = t22 > UINT64_C(0);
  if (t6) {
   t11 = (*t8);
   t12 = (uint64_t *)&t11->a;
   t25 = a1.ptr[(uintptr_t)0ul];
   t22 = (uint64_t)t25;
   t22 = zig_shlw_u64(t22, UINT8_C(16), UINT8_C(64));
   t15 = a1.len;
   t15 = zig_shr_u64(t15, UINT8_C(1));
   t25 = a1.ptr[t15];
   t5 = (uint64_t)t25;
   t5 = zig_shlw_u64(t5, UINT8_C(8), UINT8_C(64));
   t5 = t22 | t5;
   t15 = a1.len;
   t15 = t15 - (uintptr_t)1ul;
   t25 = a1.ptr[t15];
   t22 = (uint64_t)t25;
   t22 = t5 | t22;
   (*t12) = t22;
   t11 = (*t8);
   t12 = (uint64_t *)&t11->b;
   (*t12) = UINT64_C(0);
   goto zig_block_2;
  }
  t11 = (*t8);
  t12 = (uint64_t *)&t11->a;
  (*t12) = UINT64_C(0);
  t11 = (*t8);
  t12 = (uint64_t *)&t11->b;
  (*t12) = UINT64_C(0);
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;

  zig_block_1:;
  goto zig_block_0;
 }
 t26 = (uintptr_t)0ul;
 t15 = a1.len;
 t22 = t15;
 t6 = t22 >= UINT64_C(48);
 if (t6) {
  zig_loop_180:
  t15 = t26;
  t15 = t15 + (uintptr_t)48ul;
  t10 = a1.len;
  t22 = t15;
  t5 = t10;
  t6 = t22 < t5;
  if (t6) {
   t10 = t26;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
   t27 = (uint8_t const (*)[48])t14;
   t28 = &t2;
   t8 = (struct hash_wyhash_Wyhash__7909 *const *)&t28;
   t29 = t27;
   t30 = (uint8_t const (*const *)[48])&t29;
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t27 = (uint8_t const (*)[48])t14;
   t13.ptr = &(*t27)[(uintptr_t)0ul];
   t13.len = (uintptr_t)48ul;
   t31 = t13;
   t17 = (nav__6041_39 const *)&t31;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t5, &t33, sizeof(uint64_t));
   t5 = zig_wrap_u64(t5, UINT8_C(64));
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)8ul*sizeof(uint8_t)));
   t34 = (uint8_t const (*)[40])t14;
   t13.ptr = &(*t34)[(uintptr_t)0ul];
   t13.len = (uintptr_t)40ul;
   t35 = t13;
   t17 = (nav__6041_39 const *)&t35;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t22, &t33, sizeof(uint64_t));
   t22 = zig_wrap_u64(t22, UINT8_C(64));
   t11 = (*t8);
   t36 = (uint64_t (*)[3])&t11->state;
   t12 = (uint64_t *)&(*t36)[(uintptr_t)0ul];
   t5 = t5 ^ UINT64_C(16646288086500911323);
   t36 = (uint64_t (*)[3])&t2.state;
   memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
   t38 = t37[(uintptr_t)0ul];
   t38 = t22 ^ t38;
   t39 = t5;
   t40 = t38;
   t38 = t39;
   t41 = zig_make_u128(0, t38);
   t38 = t40;
   t42 = zig_make_u128(0, t38);
   t42 = zig_mulw_u128(t41, t42, UINT8_C(128));
   t38 = zig_lo_u128(t42);
   t39 = t38;
   t42 = zig_shr_u128(t42, UINT8_C(64));
   t38 = zig_lo_u128(t42);
   t40 = t38;
   t38 = t39;
   t5 = t40;
   t5 = t38 ^ t5;
   (*t12) = t5;
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)16ul*sizeof(uint8_t)));
   t43 = (uint8_t const (*)[32])t14;
   t13.ptr = &(*t43)[(uintptr_t)0ul];
   t13.len = (uintptr_t)32ul;
   t44 = t13;
   t17 = (nav__6041_39 const *)&t44;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t5, &t33, sizeof(uint64_t));
   t5 = zig_wrap_u64(t5, UINT8_C(64));
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)24ul*sizeof(uint8_t)));
   t45 = (uint8_t const (*)[24])t14;
   t13.ptr = &(*t45)[(uintptr_t)0ul];
   t13.len = (uintptr_t)24ul;
   t46 = t13;
   t17 = (nav__6041_39 const *)&t46;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t38, &t33, sizeof(uint64_t));
   t38 = zig_wrap_u64(t38, UINT8_C(64));
   t11 = (*t8);
   t36 = (uint64_t (*)[3])&t11->state;
   t12 = (uint64_t *)&(*t36)[(uintptr_t)1ul];
   t5 = t5 ^ UINT64_C(10285213230658275043);
   t36 = (uint64_t (*)[3])&t2.state;
   memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
   t22 = t37[(uintptr_t)1ul];
   t22 = t38 ^ t22;
   t47 = t5;
   t48 = t22;
   t22 = t47;
   t42 = zig_make_u128(0, t22);
   t22 = t48;
   t41 = zig_make_u128(0, t22);
   t41 = zig_mulw_u128(t42, t41, UINT8_C(128));
   t22 = zig_lo_u128(t41);
   t47 = t22;
   t41 = zig_shr_u128(t41, UINT8_C(64));
   t22 = zig_lo_u128(t41);
   t48 = t22;
   t22 = t47;
   t5 = t48;
   t5 = t22 ^ t5;
   (*t12) = t5;
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)32ul*sizeof(uint8_t)));
   t49 = (uint8_t const (*)[16])t14;
   t13.ptr = &(*t49)[(uintptr_t)0ul];
   t13.len = (uintptr_t)16ul;
   t50 = t13;
   t17 = (nav__6041_39 const *)&t50;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t5, &t33, sizeof(uint64_t));
   t5 = zig_wrap_u64(t5, UINT8_C(64));
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)40ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   t13.ptr = &(*t32)[(uintptr_t)0ul];
   t13.len = (uintptr_t)8ul;
   t51 = t13;
   t17 = (nav__6041_39 const *)&t51;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t22, &t33, sizeof(uint64_t));
   t22 = zig_wrap_u64(t22, UINT8_C(64));
   t11 = (*t8);
   t36 = (uint64_t (*)[3])&t11->state;
   t12 = (uint64_t *)&(*t36)[(uintptr_t)2ul];
   t5 = t5 ^ UINT64_C(6384245875588680899);
   t36 = (uint64_t (*)[3])&t2.state;
   memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
   t38 = t37[(uintptr_t)2ul];
   t38 = t22 ^ t38;
   t52 = t5;
   t53 = t38;
   t38 = t52;
   t41 = zig_make_u128(0, t38);
   t38 = t53;
   t42 = zig_make_u128(0, t38);
   t42 = zig_mulw_u128(t41, t42, UINT8_C(128));
   t38 = zig_lo_u128(t42);
   t52 = t38;
   t42 = zig_shr_u128(t42, UINT8_C(64));
   t38 = zig_lo_u128(t42);
   t53 = t38;
   t38 = t52;
   t5 = t53;
   t5 = t38 ^ t5;
   (*t12) = t5;
   t10 = t26;
   t10 = t10 + (uintptr_t)48ul;
   t26 = t10;
   goto zig_block_5;
  }
  goto zig_block_4;

  zig_block_5:;
  goto zig_loop_180;

  zig_block_4:;
  t54 = &t2;
  t8 = (struct hash_wyhash_Wyhash__7909 *const *)&t54;
  t11 = (*t8);
  t36 = (uint64_t (*)[3])&t11->state;
  t12 = (uint64_t *)&(*t36)[(uintptr_t)0ul];
  t38 = (*t12);
  t36 = (uint64_t (*)[3])&t2.state;
  memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
  t22 = t37[(uintptr_t)1ul];
  t36 = (uint64_t (*)[3])&t2.state;
  memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
  t5 = t37[(uintptr_t)2ul];
  t5 = t22 ^ t5;
  t5 = t38 ^ t5;
  (*t12) = t5;
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t15 = t26;
 t55 = &t2;
 t8 = (struct hash_wyhash_Wyhash__7909 *const *)&t55;
 t56 = a1;
 t1 = (nav__6041_39 const *)&t56;
 t10 = a1.len;
 t38 = t10;
 t6 = t38 >= UINT64_C(16);
 debug_assert__180(t6);
 t10 = a1.len;
 t10 = t10 - t15;
 t38 = t10;
 t6 = t38 <= UINT64_C(48);
 debug_assert__180(t6);
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t15*sizeof(uint8_t)));
 t10 = t13.len;
 t15 = t10 - t15;
 t13.ptr = t14;
 t13.len = t15;
 t57 = t13;
 t17 = (nav__6041_39 const *)&t57;
 t58 = (uintptr_t)0ul;
 zig_loop_483:
 t15 = t58;
 t15 = t15 + (uintptr_t)16ul;
 t10 = t13.len;
 t38 = t15;
 t22 = t10;
 t6 = t38 < t22;
 if (t6) {
  t11 = (*t8);
  t36 = (uint64_t (*)[3])&t11->state;
  t12 = (uint64_t *)&(*t36)[(uintptr_t)0ul];
  t10 = t58;
  t59 = (*t17);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
  t15 = t59.len;
  t10 = t15 - t10;
  t59.ptr = t14;
  t59.len = t10;
  t60 = t59;
  t61 = (nav__6041_39 const *)&t60;
  debug_assert__180(true);
  t59 = (*t61);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t32 = (uint8_t const (*)[8])t14;
  memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
  memcpy(&t22, &t33, sizeof(uint64_t));
  t22 = zig_wrap_u64(t22, UINT8_C(64));
  t22 = t22 ^ UINT64_C(16646288086500911323);
  t10 = t58;
  t10 = t10 + (uintptr_t)8ul;
  t59 = (*t17);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
  t15 = t59.len;
  t10 = t15 - t10;
  t59.ptr = t14;
  t59.len = t10;
  t62 = t59;
  t61 = (nav__6041_39 const *)&t62;
  debug_assert__180(true);
  t59 = (*t61);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t32 = (uint8_t const (*)[8])t14;
  memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
  memcpy(&t38, &t33, sizeof(uint64_t));
  t38 = zig_wrap_u64(t38, UINT8_C(64));
  t36 = (uint64_t (*)[3])&t2.state;
  memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
  t5 = t37[(uintptr_t)0ul];
  t5 = t38 ^ t5;
  t63 = t22;
  t64 = t5;
  t5 = t63;
  t42 = zig_make_u128(0, t5);
  t5 = t64;
  t41 = zig_make_u128(0, t5);
  t41 = zig_mulw_u128(t42, t41, UINT8_C(128));
  t5 = zig_lo_u128(t41);
  t63 = t5;
  t41 = zig_shr_u128(t41, UINT8_C(64));
  t5 = zig_lo_u128(t41);
  t64 = t5;
  t5 = t63;
  t22 = t64;
  t22 = t5 ^ t22;
  (*t12) = t22;
  t10 = t58;
  t10 = t10 + (uintptr_t)16ul;
  t58 = t10;
  goto zig_block_7;
 }
 goto zig_block_6;

 zig_block_7:;
 goto zig_loop_483;

 zig_block_6:;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->a;
 t15 = a1.len;
 t15 = t15 - (uintptr_t)16ul;
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t15*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 t13.ptr = &(*t32)[(uintptr_t)0ul];
 t13.len = (uintptr_t)8ul;
 t65 = t13;
 t17 = (nav__6041_39 const *)&t65;
 debug_assert__180(true);
 t13 = (*t17);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
 memcpy(&t38, &t33, sizeof(uint64_t));
 t38 = zig_wrap_u64(t38, UINT8_C(64));
 (*t12) = t38;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->b;
 t15 = a1.len;
 t15 = t15 - (uintptr_t)8ul;
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t15*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 t13.ptr = &(*t32)[(uintptr_t)0ul];
 t13.len = (uintptr_t)8ul;
 t66 = t13;
 t1 = (nav__6041_39 const *)&t66;
 debug_assert__180(true);
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
 memcpy(&t38, &t33, sizeof(uint64_t));
 t38 = zig_wrap_u64(t38, UINT8_C(64));
 (*t12) = t38;
 goto zig_block_0;

 zig_block_0:;
 t67 = (uintptr_t *)&t2.total_len;
 t15 = a1.len;
 (*t67) = t15;
 t68 = &t2;
 t8 = (struct hash_wyhash_Wyhash__7909 *const *)&t68;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->a;
 t38 = (*t12);
 t38 = t38 ^ UINT64_C(16646288086500911323);
 (*t12) = t38;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->b;
 t38 = (*t12);
 t36 = (uint64_t (*)[3])&t2.state;
 memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
 t22 = t37[(uintptr_t)0ul];
 t22 = t38 ^ t22;
 (*t12) = t22;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->a;
 t11 = (*t8);
 t69 = (uint64_t *)&t11->b;
 t22 = (*t12);
 t42 = zig_make_u128(0, t22);
 t22 = (*t69);
 t41 = zig_make_u128(0, t22);
 t41 = zig_mulw_u128(t42, t41, UINT8_C(128));
 t22 = zig_lo_u128(t41);
 (*t12) = t22;
 t41 = zig_shr_u128(t41, UINT8_C(64));
 t22 = zig_lo_u128(t41);
 (*t69) = t22;
 t69 = (uint64_t *)&t2.a;
 t22 = (*t69);
 t22 = t22 ^ UINT64_C(11562461410679940143);
 t67 = (uintptr_t *)&t2.total_len;
 t15 = (*t67);
 t38 = t15;
 t38 = t22 ^ t38;
 t69 = (uint64_t *)&t2.b;
 t22 = (*t69);
 t22 = t22 ^ UINT64_C(16646288086500911323);
 t70 = t38;
 t71 = t22;
 t22 = t70;
 t41 = zig_make_u128(0, t22);
 t22 = t71;
 t42 = zig_make_u128(0, t22);
 t42 = zig_mulw_u128(t41, t42, UINT8_C(128));
 t22 = zig_lo_u128(t42);
 t70 = t22;
 t42 = zig_shr_u128(t42, UINT8_C(64));
 t22 = zig_lo_u128(t42);
 t71 = t22;
 t22 = t70;
 t38 = t71;
 t38 = t22 ^ t38;
 return t38;
}

static bool mem_eql__anon_7916__6042(nav__6042_39 const a0, nav__6042_39 const a1) {
 nav__6042_39 t0;
 nav__6042_39 t1;
 bool t2;
 t0 = mem_sliceAsBytes__anon_3353__4892(a0);
 t1 = mem_sliceAsBytes__anon_3353__4892(a1);
 t2 = mem_eqlBytes__856(t0, t1);
 return t2;
}

static struct hash_wyhash_Wyhash__7909 hash_wyhash_Wyhash_init__6029(uint64_t const a0) {
 zig_u128 t6;
 zig_u128 t7;
 uint64_t (*t1)[3];
 uint64_t *t2;
 uint64_t t3;
 uint64_t t8;
 uint64_t t4;
 uint64_t t5;
 struct hash_wyhash_Wyhash__7909 t9;
 struct hash_wyhash_Wyhash__7909 t0;
 uint64_t t10[3];
 t0 = (struct hash_wyhash_Wyhash__7909){UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),{UINT64_C(0xaaaaaaaaaaaaaaaa), UINT64_C(0xaaaaaaaaaaaaaaaa), UINT64_C(0xaaaaaaaaaaaaaaaa)},(uintptr_t)0ul,(uintptr_t)0ul,"\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252"};
 t1 = (uint64_t (*)[3])&t0.state;
 t2 = (uint64_t *)&(*t1)[(uintptr_t)0ul];
 t3 = a0 ^ UINT64_C(11562461410679940143);
 t4 = t3;
 t5 = UINT64_C(16646288086500911323);
 t3 = t4;
 t6 = zig_make_u128(0, t3);
 t3 = t5;
 t7 = zig_make_u128(0, t3);
 t7 = zig_mulw_u128(t6, t7, UINT8_C(128));
 t3 = zig_lo_u128(t7);
 t4 = t3;
 t7 = zig_shr_u128(t7, UINT8_C(64));
 t3 = zig_lo_u128(t7);
 t5 = t3;
 t3 = t4;
 t8 = t5;
 t8 = t3 ^ t8;
 t8 = a0 ^ t8;
 (*t2) = t8;
 t1 = (uint64_t (*)[3])&t0.state;
 t2 = (uint64_t *)&(*t1)[(uintptr_t)1ul];
 t9 = t0;
 memcpy(t10, t9.state, sizeof(uint64_t[3]));
 t8 = t10[(uintptr_t)0ul];
 (*t2) = t8;
 t1 = (uint64_t (*)[3])&t0.state;
 t2 = (uint64_t *)&(*t1)[(uintptr_t)2ul];
 t9 = t0;
 memcpy(t10, t9.state, sizeof(uint64_t[3]));
 t8 = t10[(uintptr_t)0ul];
 (*t2) = t8;
 t9 = t0;
 return t9;
}

static bool mem_eqlBytes__856(nav__856_39 const a0, nav__856_39 const a1) {
 nav__856_39 const *t1;
 nav__856_39 const *t3;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t17;
 uintptr_t t18;
 uintptr_t t16;
 uintptr_t t26;
 uint64_t t6;
 uint64_t t7;
 uint8_t const *t10;
 uint8_t const *t11;
 uintptr_t t19[4];
 nav__856_39 t21;
 nav__856_39 t0;
 nav__856_39 t2;
 uint8_t const (*t22)[4];
 uint8_t const (*t27)[16];
 uint32_t t20;
 uint32_t t24;
 uint32_t t25;
 uint32_t t15;
 bool t8;
 bool t9;
 uint8_t t12;
 uint8_t t13;
 uint8_t t14;
 uint8_t t23[4];
 uint8_t t28[16];
 uint8_t t29[16];
 uint8_t t30[16];
 bool t31[16];
 t0 = a0;
 t1 = (nav__856_39 const *)&t0;
 t2 = a1;
 t3 = (nav__856_39 const *)&t2;
 t4 = a0.len;
 t5 = a1.len;
 t6 = t4;
 t7 = t5;
 t8 = t6 != t7;
 if (t8) {
  return false;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = a0.len;
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = true;
  goto zig_block_2;
 }
 t10 = a0.ptr;
 t11 = a1.ptr;
 t8 = t10 == t11;
 t9 = t8;
 goto zig_block_2;

 zig_block_2:;
 if (t9) {
  return true;
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = a0.len;
 t7 = t5;
 t9 = t7 <= UINT64_C(16);
 if (t9) {
  t5 = a0.len;
  t7 = t5;
  t9 = t7 < UINT64_C(4);
  if (t9) {
   t12 = a0.ptr[(uintptr_t)0ul];
   t13 = a1.ptr[(uintptr_t)0ul];
   t13 = t12 ^ t13;
   t5 = a0.len;
   t5 = t5 - (uintptr_t)1ul;
   t12 = a0.ptr[t5];
   t5 = a0.len;
   t5 = t5 - (uintptr_t)1ul;
   t14 = a1.ptr[t5];
   t14 = t12 ^ t14;
   t14 = t13 | t14;
   t5 = a0.len;
   t5 = t5 / (uintptr_t)2ul;
   t13 = a0.ptr[t5];
   t5 = a0.len;
   t5 = t5 / (uintptr_t)2ul;
   t12 = a1.ptr[t5];
   t12 = t13 ^ t12;
   t12 = t14 | t12;
   t9 = t12 == UINT8_C(0);
   return t9;
  }
  goto zig_block_4;

  zig_block_4:;
  t15 = UINT32_C(0);
  t16 = (uintptr_t)0ul;
  t5 = a0.len;
  t5 = t5 - (uintptr_t)4ul;
  t4 = a0.len;
  t4 = t4 / (uintptr_t)8ul;
  t4 = t4 * (uintptr_t)4ul;
  t17 = a0.len;
  t17 = t17 - (uintptr_t)4ul;
  t18 = a0.len;
  t18 = t18 / (uintptr_t)8ul;
  t18 = t18 * (uintptr_t)4ul;
  t18 = t17 - t18;
  t19[0] = (uintptr_t)0ul;
  t19[1] = t5;
  t19[2] = t4;
  t19[3] = t18;
  zig_loop_79:
  t18 = t16;
  t7 = t18;
  t9 = t7 < UINT64_C(4);
  if (t9) {
   t4 = t19[t18];
   t20 = t15;
   t21 = (*t1);
   t11 = t21.ptr;
   t11 = (uint8_t const *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
   t22 = (uint8_t const (*)[4])t11;
   memcpy(t23, (const char *)t22, sizeof(uint8_t[4]));
   memcpy(&t24, &t23, sizeof(uint32_t));
   t24 = zig_wrap_u32(t24, UINT8_C(32));
   t21 = (*t3);
   t11 = t21.ptr;
   t11 = (uint8_t const *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
   t22 = (uint8_t const (*)[4])t11;
   memcpy(t23, (const char *)t22, sizeof(uint8_t[4]));
   memcpy(&t25, &t23, sizeof(uint32_t));
   t25 = zig_wrap_u32(t25, UINT8_C(32));
   t25 = t24 ^ t25;
   t25 = t20 | t25;
   t15 = t25;
   goto zig_block_6;
  }
  goto zig_block_5;

  zig_block_6:;
  t18 = t18 + (uintptr_t)1ul;
  t16 = t18;
  goto zig_loop_79;

  zig_block_5:;
  t25 = t15;
  t9 = t25 == UINT32_C(0);
  return t9;
 }
 goto zig_block_3;

 zig_block_3:;
 t26 = (uintptr_t)0ul;
 t18 = a0.len;
 t18 = t18 - (uintptr_t)1ul;
 t18 = t18 / (uintptr_t)16ul;
 zig_loop_132:
 t17 = t26;
 t7 = t17;
 t6 = t18;
 t9 = t7 < t6;
 if (t9) {
  t5 = t17 * (uintptr_t)16ul;
  t21 = (*t1);
  t11 = t21.ptr;
  t11 = (uint8_t const *)(((uintptr_t)t11) + (t5*sizeof(uint8_t)));
  t27 = (uint8_t const (*)[16])t11;
  memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
  memcpy(&t29, &t28, sizeof(uint8_t[16]));
  t5 = t17 * (uintptr_t)16ul;
  t21 = (*t3);
  t11 = t21.ptr;
  t11 = (uint8_t const *)(((uintptr_t)t11) + (t5*sizeof(uint8_t)));
  t27 = (uint8_t const (*)[16])t11;
  memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
  memcpy(&t30, &t28, sizeof(uint8_t[16]));
  for (t5 = (uintptr_t)0ul; t5 < (uintptr_t)16ul; t5 += (uintptr_t)1ul) {
   t31[t5] = t29[t5] != t30[t5];
  }
  t9 = false;
  for (t5 = (uintptr_t)0ul; t5 < (uintptr_t)16ul; t5 += (uintptr_t)1ul) {
   t9 |= t31[t5];
  }
  if (t9) {
   return false;
  }
  goto zig_block_9;

  zig_block_9:;
  goto zig_block_8;
 }
 goto zig_block_7;

 zig_block_8:;
 t17 = t17 + (uintptr_t)1ul;
 t26 = t17;
 goto zig_loop_132;

 zig_block_7:;
 t18 = a0.len;
 t18 = t18 - (uintptr_t)16ul;
 t21 = (*t1);
 t11 = t21.ptr;
 t11 = (uint8_t const *)(((uintptr_t)t11) + (t18*sizeof(uint8_t)));
 t27 = (uint8_t const (*)[16])t11;
 memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
 memcpy(&t30, &t28, sizeof(uint8_t[16]));
 t18 = a0.len;
 t18 = t18 - (uintptr_t)16ul;
 t21 = (*t3);
 t11 = t21.ptr;
 t11 = (uint8_t const *)(((uintptr_t)t11) + (t18*sizeof(uint8_t)));
 t27 = (uint8_t const (*)[16])t11;
 memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
 memcpy(&t29, &t28, sizeof(uint8_t[16]));
 for (t18 = (uintptr_t)0ul; t18 < (uintptr_t)16ul; t18 += (uintptr_t)1ul) {
  t31[t18] = t30[t18] != t29[t18];
 }
 t9 = false;
 for (t18 = (uintptr_t)0ul; t18 < (uintptr_t)16ul; t18 += (uintptr_t)1ul) {
  t9 |= t31[t18];
 }
 t9 = !t9;
 return t9;
}

static uint16_t environment_root_Environment_setIntegrationEnvironmentList__776(struct environment_root_Environment__559 *const a0, nav__776_41 const a1, nav__776_41 const a2, nav__776_43 const a3) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t0;
 struct mem_Allocator__599 *t3;
 struct mem_Allocator__599 t4;
 uintptr_t t5;
 uintptr_t t12;
 uintptr_t t11;
 uintptr_t t20;
 uintptr_t t24;
 nav__776_85 t6;
 nav__776_83 t8;
 nav__776_83 t16;
 nav__776_83 t9;
 nav__776_83 const *t10;
 uint64_t t13;
 uint64_t t14;
 nav__776_41 *t17;
 nav__776_41 t18;
 nav__776_90 t19;
 nav__776_88 t21;
 nav__776_43 t22;
 struct environment_root_EnvironmentValue__844 t23;
 uint16_t t7;
 bool t15;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (struct mem_Allocator__599 *)&t2->allocator;
 t4 = (*t3);
 t5 = a3.len;
 t6 = mem_Allocator_alloc__anon_8016__6046(t4, t5);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t8 = t6.payload;
 t9 = t8;
 t10 = (nav__776_83 const *)&t9;
 t11 = (uintptr_t)0ul;
 zig_loop_25:
 t5 = t11;
 t12 = a3.len;
 t13 = t5;
 t14 = t12;
 t15 = t13 < t14;
 if (t15) {
  t12 = t11;
  t16 = (*t10);
  t17 = &t16.ptr[t12];
  t2 = (*t1);
  t3 = (struct mem_Allocator__599 *)&t2->allocator;
  t4 = (*t3);
  t12 = t11;
  t18 = a3.ptr[t12];
  t19 = mem_Allocator_dupe__anon_7390__5934(t4, t18);
  if (t19.error) {
   t7 = t19.error;
   t20 = (uintptr_t)0ul;
   t12 = t11;
   t16 = (*t10);
   t17 = t16.ptr;
   t17 = (nav__776_41 *)(((uintptr_t)t17) + ((uintptr_t)0ul*sizeof(nav__776_41)));
   t16.ptr = t17;
   t16.len = t12;
   t12 = t16.len;
   zig_loop_55:
   t5 = t20;
   t14 = t5;
   t13 = t12;
   t15 = t14 < t13;
   if (t15) {
    t18 = t16.ptr[t5];
    t2 = (*t1);
    t3 = (struct mem_Allocator__599 *)&t2->allocator;
    t4 = (*t3);
    mem_Allocator_free__anon_3042__4747(t4, t18);
    goto zig_block_3;
   }
   goto zig_block_2;

   zig_block_3:;
   t5 = t5 + (uintptr_t)1ul;
   t20 = t5;
   goto zig_loop_55;

   zig_block_2:;
   t2 = (*t1);
   t3 = (struct mem_Allocator__599 *)&t2->allocator;
   t4 = (*t3);
   mem_Allocator_free__anon_8019__6047(t4, t8);
   return t7;
  }
  t21 = t19.payload;
  memcpy(&t18, &t21, sizeof(nav__776_41));
  (*t17) = t18;
  t12 = t11;
  t12 = t12 + (uintptr_t)1ul;
  t11 = t12;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_25;

 zig_block_0:;
 t2 = (*t1);
 memcpy(&t22, &t8, sizeof(nav__776_43));
 t23.tag = UINT8_C(1);
 t23.payload.list = t22;
 t7 = environment_root_Environment_setIntegrationEnvironmentValue__778(t2, a1, a2, t23);
 if (t7) {
  t24 = (uintptr_t)0ul;
  t12 = t11;
  t16 = (*t10);
  t17 = t16.ptr;
  t17 = (nav__776_41 *)(((uintptr_t)t17) + ((uintptr_t)0ul*sizeof(nav__776_41)));
  t16.ptr = t17;
  t16.len = t12;
  t12 = t16.len;
  zig_loop_110:
  t5 = t24;
  t14 = t5;
  t13 = t12;
  t15 = t14 < t13;
  if (t15) {
   t18 = t16.ptr[t5];
   t2 = (*t1);
   t3 = (struct mem_Allocator__599 *)&t2->allocator;
   t4 = (*t3);
   mem_Allocator_free__anon_3042__4747(t4, t18);
   goto zig_block_5;
  }
  goto zig_block_4;

  zig_block_5:;
  t5 = t5 + (uintptr_t)1ul;
  t24 = t5;
  goto zig_loop_110;

  zig_block_4:;
  t2 = (*t1);
  t3 = (struct mem_Allocator__599 *)&t2->allocator;
  t4 = (*t3);
  mem_Allocator_free__anon_8019__6047(t4, t8);
  return t7;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_environment_list__256(struct InstrumentHooks *const a0, char const *const a1, char const *const a2, char const *const *const a3, uint32_t const a4) {
 struct InstrumentHooks *t2;
 struct InstrumentHooks *t3;
 struct InstrumentHooks *const *t4;
 nav__256_47 t5;
 nav__256_47 t8;
 nav__256_47 t9;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t12;
 nav__256_54 t7;
 nav__256_47 const *t10;
 uint8_t const *const *t11;
 uint64_t t14;
 uint64_t t15;
 uint8_t const *t16;
 nav__256_45 *t17;
 nav__256_45 t18;
 nav__256_45 t19;
 nav__256_45 t21;
 struct environment_root_Environment__559 *t20;
 nav__256_120 t22;
 uint16_t t23;
 bool t0;
 bool t1;
 t0 = a1 == NULL;
 if (t0) {
  t1 = true;
  goto zig_block_1;
 }
 t0 = a2 == NULL;
 t1 = t0;
 goto zig_block_1;

 zig_block_1:;
 if (t1) {
  t0 = true;
  goto zig_block_2;
 }
 t1 = a3 == NULL;
 t0 = t1;
 goto zig_block_2;

 zig_block_2:;
 if (t0) {
  return UINT8_C(1);
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0 != NULL;
 if (t1) {
  t2 = a0;
  t3 = t2;
  t4 = (struct InstrumentHooks *const *)&t3;
  t6 = (uintptr_t)a4;
  t7 = mem_Allocator_alloc__anon_8016__6046((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)}, t6);
  t1 = t7.error == UINT16_C(0);
  if (t1) {
   t8 = t7.payload;
   t5 = t8;
   goto zig_block_4;
  }
  return UINT8_C(1);

  zig_block_4:;
  t9 = t5;
  t10 = (nav__256_47 const *)&t9;
  t11 = (uint8_t const *const *)a3;
  t12 = (uintptr_t)0ul;
  t6 = (uintptr_t)a4;
  zig_loop_45:
  t13 = t12;
  t14 = t13;
  t15 = t6;
  t1 = t14 < t15;
  if (t1) {
   t16 = t11[t13];
   t1 = t16 == NULL;
   if (t1) {
    mem_Allocator_free__anon_8019__6047((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)}, t5);
    return UINT8_C(1);
   }
   goto zig_block_7;

   zig_block_7:;
   t8 = (*t10);
   t17 = &t8.ptr[t13];
   t16 = t11[t13];
   t18 = mem_span__anon_6867__5771(t16);
   memcpy(&t19, &t18, sizeof(nav__256_45));
   (*t17) = t19;
   goto zig_block_6;
  }
  goto zig_block_5;

  zig_block_6:;
  t13 = t13 + (uintptr_t)1ul;
  t12 = t13;
  goto zig_loop_45;

  zig_block_5:;
  t2 = (*t4);
  t20 = (struct environment_root_Environment__559 *)&t2->environment;
  t16 = c_toU8__236(a1);
  t19 = mem_span__anon_6867__5771(t16);
  memcpy(&t18, &t19, sizeof(nav__256_45));
  t16 = c_toU8__236(a2);
  t19 = mem_span__anon_6867__5771(t16);
  memcpy(&t21, &t19, sizeof(nav__256_45));
  memcpy(&t22, &t5, sizeof(nav__256_120));
  t23 = environment_root_Environment_setIntegrationEnvironmentList__776(t20, t18, t21, t22);
  t1 = t23 == UINT16_C(0);
  if (t1) {
   goto zig_block_8;
  }
  mem_Allocator_free__anon_8019__6047((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)}, t5);
  return UINT8_C(1);

  zig_block_8:;
  mem_Allocator_free__anon_8019__6047((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)}, t5);
  return UINT8_C(0);
 }
 goto zig_block_3;

 zig_block_3:;
 return UINT8_C(1);
}

static nav__6046_42 mem_Allocator_alloc__anon_8016__6046(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__6046_42 t4;
 nav__6046_42 t8;
 nav__6046_54 t6;
 uint8_t *t9;
 nav__6046_39 *t10;
 nav__6046_39 *t11;
 nav__6046_39 *const *t12;
 nav__6046_41 t13;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_8027__6048(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__6046_41){(nav__6046_39 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = (nav__6046_39 *)t9;
 t11 = t10;
 t12 = (nav__6046_39 *const *)&t11;
 t10 = (*t12);
 t10 = (nav__6046_39 *)(((uintptr_t)t10) + ((uintptr_t)0ul*sizeof(nav__6046_39)));
 t13.ptr = t10;
 t13.len = a1;
 t8.payload = t13;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void mem_Allocator_free__anon_8019__6047(struct mem_Allocator__599 const a0, nav__6047_42 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6047_52 t2;
 nav__6047_52 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__6047_52, uint8_t, uintptr_t);
 void (*t17)(void *, nav__6047_52, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_8034__6049(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__6047_52));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__6047_52, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static nav__6048_39 mem_Allocator_allocWithSizeAndAlignment__anon_8027__6048(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__6048_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__6048_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3220__4889((uintptr_t)16ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__6048_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_2726__4265(t6, t2, a2);
 return t7;
}

static nav__6049_39 mem_sliceAsBytes__anon_8034__6049(nav__6049_43 const a0) {
 uintptr_t t0;
 uint64_t t1;
 nav__6049_41 *t4;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *const *t7;
 nav__6049_39 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6049_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (uint8_t *)t4;
 t6 = t5;
 t7 = (uint8_t *const *)&t6;
 t0 = a0.len;
 t0 = t0 * (uintptr_t)16ul;
 t5 = (*t7);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

uint8_t c_instrument_hooks_write_environment__257(struct InstrumentHooks *const a0, int32_t const a1) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct environment_root_Environment__559 *t4;
 bool t0;
 uint8_t t5;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct environment_root_Environment__559 *)&t1->environment;
  t5 = environment_root_Environment_writeEnvironment__780(t4, a1);
  return t5;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint16_t environment_linked_libraries_root_LinkedLibraries_collect__2812(struct environment_linked_libraries_root_LinkedLibraries__837 *const a0) {
 struct environment_linked_libraries_root_LinkedLibraries__837 *t0;
 t0 = a0;
 return 0;
}

static uint16_t environment_root_Environment_populateLinkedLibraries__779(struct environment_root_Environment__559 *const a0) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t3;
 struct environment_root_Environment__559 *t0;
 struct environment_root_EnvironmentJson__833 *t4;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *t5;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t2;
 nav__779_87 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147 t11;
 struct mem_Allocator__599 *t12;
 struct mem_Allocator__599 t13;
 nav__779_44 *t14;
 nav__779_44 t15;
 nav__779_44 t26;
 nav__779_44 t27;
 struct environment_linked_libraries_root_LinkedLibraries__837 *t17;
 struct environment_linked_libraries_root_LinkedLibraries__837 t18;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__936 t19;
 nav__779_76 t20;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t16;
 uint64_t t23;
 uint64_t t24;
 struct environment_linked_libraries_root_LibraryInfo__934 t25;
 nav__779_95 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t30;
 nav__779_99 t31;
 nav__779_97 t32;
 struct environment_linked_libraries_root_LibraryEntry__852 *t33;
 uint16_t t29;
 bool t10;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *)&t4->linked_libraries;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *)&t5->map;
 t7 = (*t6);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3355(t7);
 t2 = t8;
 zig_loop_14:
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__5628(&t2);
 t10 = t9.is_null != true;
 if (t10) {
  t11 = t9.payload;
  t3 = (*t1);
  t12 = (struct mem_Allocator__599 *)&t3->allocator;
  t13 = (*t12);
  t14 = t11.key_ptr;
  t15 = (*t14);
  mem_Allocator_free__anon_3042__4747(t13, t15);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_14;

 zig_block_0:;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *)&t4->linked_libraries;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *)&t5->map;
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3350(t6);
 t16 = (uintptr_t)0ul;
 t17 = (struct environment_linked_libraries_root_LinkedLibraries__837 *)&a0->libs;
 t18 = (*t17);
 t19 = t18.libraries;
 t20 = t19.items;
 t21 = t20.len;
 zig_loop_46:
 t22 = t16;
 t23 = t22;
 t24 = t21;
 t10 = t23 < t24;
 if (t10) {
  t25 = t20.ptr[t22];
  t26 = t25.soname;
  t10 = t26.ptr != NULL;
  if (t10) {
   t27 = t26;
   t15 = t27;
   goto zig_block_4;
  }
  t26 = t25.path;
  t15 = t26;
  goto zig_block_4;

  zig_block_4:;
  t3 = (*t1);
  t4 = (struct environment_root_EnvironmentJson__833 *)&t3->data;
  t5 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *)&t4->linked_libraries;
  t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *)&t5->map;
  t12 = (struct mem_Allocator__599 *)&a0->allocator;
  t13 = (*t12);
  t28 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3357(t6, t13, t15);
  if (t28.error) {
   t29 = t28.error;
   return t29;
  }
  t30 = t28.payload;
  t14 = t30.key_ptr;
  t3 = (*t1);
  t12 = (struct mem_Allocator__599 *)&t3->allocator;
  t13 = (*t12);
  t31 = mem_Allocator_dupe__anon_7390__5934(t13, t15);
  if (t31.error) {
   t29 = t31.error;
   return t29;
  }
  t32 = t31.payload;
  memcpy(&t15, &t32, sizeof(nav__779_44));
  (*t14) = t15;
  t33 = t30.value_ptr;
  t14 = (nav__779_44 *)&t33->path;
  t15 = t25.path;
  (*t14) = t15;
  t14 = (nav__779_44 *)&t33->build_id;
  t15 = t25.build_id;
  (*t14) = t15;
  goto zig_block_3;
 }
 goto zig_block_2;

 zig_block_3:;
 t22 = t22 + (uintptr_t)1ul;
 t16 = t22;
 goto zig_loop_46;

 zig_block_2:;
 return 0;
}

static uint8_t environment_root_Environment_writeEnvironment__780(struct environment_root_Environment__559 *const a0, int32_t const a1) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t0;
 struct environment_linked_libraries_root_LinkedLibraries__837 *t3;
 struct environment_root_EnvironmentJson__833 *t6;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t9;
 uintptr_t t10;
 uint64_t t11;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *t13;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t14;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t15;
 uint8_t const *t16;
 uint8_t const *t17;
 uint8_t const *t18;
 nav__780_83 t19;
 nav__780_86 t21;
 nav__780_86 t22;
 nav__780_86 t28;
 nav__780_89 t23;
 nav__780_91 t24;
 struct mem_Allocator__599 *t25;
 struct mem_Allocator__599 t26;
 struct environment_root_EnvironmentJson__833 t27;
 struct fs_File__794 const *t33;
 struct fs_File__794 t29;
 struct fs_File__794 t31;
 struct fs_File__794 t32;
 nav__780_100 t30;
 uint16_t t4;
 bool t5;
 bool t12;
 uint8_t t20[512];
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (struct environment_linked_libraries_root_LinkedLibraries__837 *)&t2->libs;
 (void)environment_linked_libraries_root_LinkedLibraries_collect__2812(t3);
 t2 = (*t1);
 t4 = environment_root_Environment_populateLinkedLibraries__779(t2);
 t5 = t4 == UINT16_C(0);
 if (t5) {
  goto zig_block_0;
 }
 logger_err__anon_8078__6050();
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t6 = (struct environment_root_EnvironmentJson__833 *)&t2->data;
 t7 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 *)&t6->integration_environment;
 t8 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 *)&t7->map;
 t9 = (*t8);
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2962(t9);
 t11 = t10;
 t5 = t11 == UINT64_C(0);
 if (t5) {
  t2 = (*t1);
  t6 = (struct environment_root_EnvironmentJson__833 *)&t2->data;
  t13 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 *)&t6->linked_libraries;
  t14 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *)&t13->map;
  t15 = (*t14);
  t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3352(t15);
  t11 = t10;
  t5 = t11 == UINT64_C(0);
  t12 = t5;
  goto zig_block_2;
 }
 t12 = false;
 goto zig_block_2;

 zig_block_2:;
 if (t12) {
  return UINT8_C(0);
 }
 goto zig_block_1;

 zig_block_1:;
 t17 = getenv((uint8_t const *)&__anon_8092);
 t12 = t17 != NULL;
 if (t12) {
  t18 = t17;
  t16 = t18;
  goto zig_block_3;
 }
 return UINT8_C(0);

 zig_block_3:;
 t19 = mem_span__anon_8102__6052(t16);
 t22.ptr = &t20[(uintptr_t)0ul];
 t22.len = (uintptr_t)512ul;
 t23.f0 = t19;
 t23.f1 = a1;
 t24 = fmt_bufPrint__anon_8111__6053(t22, t23);
 t12 = t24.error == UINT16_C(0);
 if (t12) {
  t22 = t24.payload;
  t21 = t22;
  goto zig_block_4;
 }
 logger_err__anon_8118__6054();
 return UINT8_C(1);

 zig_block_4:;
 t25 = (struct mem_Allocator__599 *)&a0->allocator;
 t26 = (*t25);
 t6 = (struct environment_root_EnvironmentJson__833 *)&a0->data;
 t27 = (*t6);
 t24 = json_stringify_stringifyAlloc__anon_8133__6076(t26, t27, (struct json_stringify_StringifyOptions__8122){UINT8_C(2),true,false,false,false});
 t12 = t24.error == UINT16_C(0);
 if (t12) {
  t28 = t24.payload;
  t22 = t28;
  goto zig_block_5;
 }
 logger_err__anon_8140__6077();
 return UINT8_C(1);

 zig_block_5:;
 memcpy(&t19, &t21, sizeof(nav__780_83));
 t30 = fs_createFileAbsolute__1351(t19, (struct fs_File_CreateFlags__8142){UINT16_C(438),false,true,false,UINT8_C(0),false});
 t12 = t30.error == UINT16_C(0);
 if (t12) {
  t31 = t30.payload;
  t29 = t31;
  goto zig_block_6;
 }
 logger_err__anon_8159__6078();
 t2 = (*t1);
 t25 = (struct mem_Allocator__599 *)&t2->allocator;
 t26 = (*t25);
 mem_Allocator_free__anon_3655__5138(t26, t22);
 return UINT8_C(1);

 zig_block_6:;
 t32 = t29;
 t33 = (struct fs_File__794 const *)&t32;
 t29 = (*t33);
 memcpy(&t19, &t22, sizeof(nav__780_83));
 t4 = fs_File_writeAll__1439(t29, t19);
 t12 = t4 == UINT16_C(0);
 if (t12) {
  goto zig_block_7;
 }
 logger_err__anon_8159__6078();
 t29 = (*t33);
 fs_File_close__1385(t29);
 t2 = (*t1);
 t25 = (struct mem_Allocator__599 *)&t2->allocator;
 t26 = (*t25);
 mem_Allocator_free__anon_3655__5138(t26, t22);
 return UINT8_C(1);

 zig_block_7:;
 t29 = (*t33);
 fs_File_close__1385(t29);
 t2 = (*t1);
 t25 = (struct mem_Allocator__599 *)&t2->allocator;
 t26 = (*t25);
 mem_Allocator_free__anon_3655__5138(t26, t22);
 return UINT8_C(0);
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3350(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 void *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t4;
 uintptr_t *t5;
 struct array_hash_map_IndexHeader__880 **t6;
 struct array_hash_map_IndexHeader__880 *t7;
 struct array_hash_map_IndexHeader__880 *t9;
 struct array_hash_map_IndexHeader__880 *t10;
 struct array_hash_map_IndexHeader__880 *const *t11;
 nav__3350_57 t14;
 uintptr_t t15;
 nav__3350_62 t16;
 nav__3350_67 t17;
 struct array_hash_map_IndexHeader__880 t12;
 bool t8;
 uint8_t t13;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3147(t3);
 t2 = (*t1);
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t2->entries;
 t5 = (uintptr_t *)&t4->len;
 (*t5) = (uintptr_t)0ul;
 t6 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t7 = (*t6);
 t8 = t7 != NULL;
 if (t8) {
  t9 = t7;
  t10 = t9;
  t11 = (struct array_hash_map_IndexHeader__880 *const *)&t10;
  t9 = (*t11);
  t12 = (*t9);
  t13 = array_hash_map_IndexHeader_capacityIndexType__3138(t12);
  switch (t13) {
   case UINT8_C(0): {
    t9 = (*t11);
    t14 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t9);
    for (t15 = (uintptr_t)0ul; t15 != t14.len; ++t15) ((struct array_hash_map_Index_28u8_29__6669 *)t14.ptr)[t15] = (struct array_hash_map_Index_28u8_29__6669){UINT8_MAX,UINT8_C(0xaa)};
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t9 = (*t11);
    t16 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t9);
    for (t15 = (uintptr_t)0ul; t15 != t16.len; ++t15) ((struct array_hash_map_Index_28u16_29__6673 *)t16.ptr)[t15] = (struct array_hash_map_Index_28u16_29__6673){UINT16_MAX,UINT16_C(0xaaaa)};
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t9 = (*t11);
    t17 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t9);
    for (t15 = (uintptr_t)0ul; t15 != t17.len; ++t15) ((struct array_hash_map_Index_28u32_29__6677 *)t17.ptr)[t15] = (struct array_hash_map_Index_28u32_29__6677){UINT32_MAX,UINT32_C(0xaaaaaaaa)};
    goto zig_block_1;
   }
   default: zig_unreachable();
  }

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3148(t3);
 return;
}

static nav__3357_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3357(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct mem_Allocator__599 const a1, nav__3357_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 nav__3357_39 t3;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3358(t2, a1, a2);
 return t3;
}

static void logger_err__anon_8078__6050(void) {
 logger_logWithPrefix__anon_8171__6079();
 return;
}

static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2962(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__877 t0;
 uintptr_t t1;
 t0 = a0.entries;
 t1 = t0.len;
 return t1;
}

static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3352(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t0;
 uintptr_t t1;
 t0 = a0.entries;
 t1 = t0.len;
 return t1;
}

static nav__6052_39 mem_span__anon_8102__6052(uint8_t const *const a0) {
 uint8_t const *const *t1;
 uintptr_t t2;
 uint8_t const *t3;
 uint8_t const *t0;
 nav__6052_39 t4;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = mem_len__anon_8172__6080(a0);
 t3 = (*t1);
 t3 = (uint8_t const *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t3;
 t4.len = t2;
 return t4;
}

static uint16_t fmt_format__anon_8173__6081(struct io_Writer__3827 const a0, nav__6081_41 const a1) {
 struct io_Writer__3827 const *t1;
 nav__6081_40 t2;
 struct io_Writer__3827 t4;
 struct io_Writer__3827 t0;
 int32_t t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = a1.f0;
 t3 = fmt_formatType__anon_4732__5575(t2, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t4 = (*t1);
 t3 = io_Writer_writeAll__5168(t4, (nav__6081_40){(uint8_t const *)&__anon_8208,(uintptr_t)13ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5404__5603(t5, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t4 = (*t1);
 t3 = io_Writer_writeAll__5168(t4, (nav__6081_40){(uint8_t const *)&__anon_8225,(uintptr_t)5ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__6053_40 fmt_bufPrint__anon_8111__6053(nav__6053_39 const a0, nav__6053_43 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__6053_59 (**t11)(void const *, nav__6053_42);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__6053_39 t15;
 nav__6053_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6053_59 (**)(void const *, nav__6053_42))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_8173__6081(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6053_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static void logger_err__anon_8118__6054(void) {
 logger_logWithPrefix__anon_8231__6082();
 return;
}

static nav__6076_40 json_stringify_stringifyAlloc__anon_8133__6076(struct mem_Allocator__599 const a0, struct environment_root_EnvironmentJson__833 const a1, struct json_stringify_StringifyOptions__8122 const a2) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 nav__6076_40 t4;
 uint16_t t3;
 bool t5;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(a0);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(&t0);
 t3 = json_stringify_stringifyArbitraryDepth__anon_8258__6112(a0, a1, a2, t2);
 if (t3) {
  t1 = t0;
  array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t1);
  t4.payload = (nav__6076_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t4.error = t3;
  return t4;
 }
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_toOwnedSlice__1533(&t0);
 t5 = t4.error == UINT16_C(0);
 if (t5) {
  return t4;
 }
 t1 = t0;
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t1);
 return t4;
}

static void logger_err__anon_8140__6077(void) {
 logger_logWithPrefix__anon_8263__6113();
 return;
}

static nav__1351_39 fs_createFileAbsolute__1351(nav__1351_41 const a0, struct fs_File_CreateFlags__8142 const a1) {
 struct fs_Dir__1103 const *t3;
 struct fs_Dir__1103 t1;
 struct fs_Dir__1103 t2;
 nav__1351_39 t4;
 bool t0;
 t0 = fs_path_isAbsolute__4300(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1340();
 t2 = t1;
 t3 = (struct fs_Dir__1103 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_createFile__3724(t1, a0, a1);
 return t4;
}

static void logger_err__anon_8159__6078(void) {
 logger_logWithPrefix__anon_8268__6114();
 return;
}

static nav__3358_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3358(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct mem_Allocator__599 const a1, nav__3358_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 nav__3358_39 t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t5;
 nav__3358_44 *t7;
 uint16_t t4;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6115(t2, a1, a2);
 if (t3.error) {
  t4 = t3.error;
  t3.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062){((nav__3358_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_linked_libraries_root_LibraryEntry__852 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
  t3.error = t4;
  return t3;
 }
 t5 = t3.payload;
 t6 = t5.found_existing;
 t6 = !t6;
 if (t6) {
  t7 = t5.key_ptr;
  (*t7) = a2;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3.payload = t5;
 t3.error = UINT16_C(0);
 return t3;
}

static void logger_logWithPrefix__anon_8171__6079(void) {
 nav__6079_41 t3;
 nav__6079_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6079_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8285__6116(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static uintptr_t mem_len__anon_8172__6080(uint8_t const *const a0) {
 uintptr_t t0;
 t0 = mem_indexOfSentinel__anon_6891__5773(a0);
 return t0;
}

static void logger_logWithPrefix__anon_8231__6082(void) {
 nav__6082_41 t3;
 nav__6082_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6082_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8294__6117(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static uint16_t json_stringify_stringifyArbitraryDepth__anon_8258__6112(struct mem_Allocator__599 const a0, struct environment_root_EnvironmentJson__833 const a1, struct json_stringify_StringifyOptions__8122 const a2, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a3) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 t0;
 uint16_t t2;
 t1 = json_stringify_writeStreamArbitraryDepth__anon_8302__6118(a0, a3, a2);
 t0 = t1;
 t2 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6119(&t0, a1);
 if (t2) {
  json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6089(&t0);
  return t2;
 }
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6089(&t0);
 return 0;
}

static nav__1533_40 array_list_ArrayListAligned_28u8_2cnull_29_toOwnedSlice__1533(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct mem_Allocator__599 *t2;
 struct mem_Allocator__599 t3;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 const *t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t7;
 nav__1533_39 t8;
 nav__1533_39 t11;
 nav__1533_39 *t10;
 uintptr_t t12;
 nav__1533_40 t14;
 uint8_t *t16;
 uint16_t t15;
 bool t13;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (struct mem_Allocator__599 *)&a0->allocator;
 t3 = (*t2);
 t4 = t3;
 t5 = (struct mem_Allocator__599 const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t7);
 t9 = (*t5);
 t10 = (nav__1533_39 *)&a0->items;
 t11 = (*t10);
 t12 = t11.len;
 t8 = mem_Allocator_remap__anon_3648__5136(t9, t8, t12);
 t13 = t8.ptr != NULL;
 if (t13) {
  t11 = t8;
  t7 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(t3);
  (*a0) = t7;
  t14.payload = t11;
  t14.error = UINT16_C(0);
  return t14;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t5);
 t10 = (nav__1533_39 *)&a0->items;
 t8 = (*t10);
 t12 = t8.len;
 t14 = mem_Allocator_alignedAlloc__anon_3653__5137(t3, t12);
 if (t14.error) {
  t15 = t14.error;
  t14.payload = (nav__1533_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t14.error = t15;
  return t14;
 }
 t8 = t14.payload;
 t10 = (nav__1533_39 *)&a0->items;
 t11 = (*t10);
 t16 = t11.ptr;
 if (t8.len != 0) memcpy(t8.ptr, t16, t8.len * sizeof(uint8_t));
 t6 = (*t1);
 array_list_ArrayListAligned_28u8_2cnull_29_clearAndFree__1563(t6);
 t14.payload = t8;
 t14.error = UINT16_C(0);
 return t14;
}

static void logger_logWithPrefix__anon_8263__6113(void) {
 nav__6113_41 t3;
 nav__6113_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6113_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8325__6120(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static nav__3724_39 fs_Dir_createFile__3724(struct fs_Dir__1103 const a0, nav__3724_42 const a1, struct fs_File_CreateFlags__8142 const a2) {
 struct fs_Dir__1103 const *t1;
 uint8_t const (*t7)[1024];
 uint8_t const *t9;
 nav__3724_39 t4;
 struct fs_Dir__1103 t8;
 struct fs_Dir__1103 t0;
 nav__3724_53 t2;
 uint16_t t3;
 uint8_t t5[1024];
 uint8_t t6[1024];
 t0 = a0;
 t1 = (struct fs_Dir__1103 const *)&t0;
 t2 = posix_toPosixPath__2022(a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct fs_File__794){-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 memcpy(t5, t2.payload, sizeof(uint8_t[1024]));
 memcpy((char *)&t6, t5, sizeof(uint8_t[1024]));
 t7 = (uint8_t const (*)[1024])&t6;
 t8 = (*t1);
 t9 = (uint8_t const *)t7;
 t4 = fs_Dir_createFileZ__3725(t8, t9, a2);
 return t4;
}

static void logger_logWithPrefix__anon_8268__6114(void) {
 nav__6114_41 t3;
 nav__6114_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6114_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__4255);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8340__6121(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3411);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static nav__6115_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6115(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct mem_Allocator__599 const a1, nav__6115_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t4;
 uintptr_t t5;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t8;
 nav__6115_67 t9;
 nav__6115_39 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t14;
 nav__6115_73 t15;
 nav__6115_73 t16;
 nav__6115_73 const *t17;
 nav__6115_44 *t18;
 nav__6115_76 t19;
 nav__6115_76 t20;
 nav__6115_76 const *t21;
 struct environment_linked_libraries_root_LibraryEntry__852 *t22;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t23;
 uint16_t t6;
 bool t7;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&a0->entries;
 t4 = (*t3);
 t5 = t4.len;
 t5 = t5 + (uintptr_t)1ul;
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3367(t2, a1, t5);
 t7 = t6 == UINT16_C(0);
 if (t7) {
  goto zig_block_0;
 }
 t2 = (*t1);
 t8 = (*t2);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6122(t8, a2);
 t7 = t9.is_null != true;
 if (t7) {
  t10 = t9.payload;
  t5 = t10;
  goto zig_block_1;
 }
 t11.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062){((nav__6115_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_linked_libraries_root_LibraryEntry__852 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
 t11.error = t6;
 return t11;

 zig_block_1:;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t2->entries;
 t4 = (*t3);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t4);
 t13 = t12;
 t14 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t13;
 t12 = (*t14);
 t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t12);
 t16 = t15;
 t17 = (nav__6115_73 const *)&t16;
 t15 = (*t17);
 t18 = &t15.ptr[t5];
 t12 = (*t14);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t12);
 t20 = t19;
 t21 = (nav__6115_76 const *)&t20;
 t19 = (*t21);
 t22 = &t19.ptr[t5];
 t23.key_ptr = t18;
 t23.value_ptr = t22;
 t23.index = t5;
 t23.found_existing = true;
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;

 zig_block_0:;
 t2 = (*t1);
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6123(t2, a2);
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;
}

static uint16_t fmt_format__anon_8359__6124(struct io_Writer__3827 const a0) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__6124_43){(uint8_t const *)&__anon_8429,(uintptr_t)62ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__6116_40 fmt_bufPrint__anon_8285__6116(nav__6116_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__6116_54 (**t11)(void const *, nav__6116_56);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__6116_39 t15;
 nav__6116_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6116_54 (**)(void const *, nav__6116_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_8359__6124(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6116_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_8435__6125(struct io_Writer__3827 const a0) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__6125_43){(uint8_t const *)&__anon_8496,(uintptr_t)55ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__6117_40 fmt_bufPrint__anon_8294__6117(nav__6117_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__6117_54 (**t11)(void const *, nav__6117_56);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__6117_39 t15;
 nav__6117_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6117_54 (**)(void const *, nav__6117_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_8435__6125(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6117_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 json_stringify_writeStreamArbitraryDepth__anon_8302__6118(struct mem_Allocator__599 const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a1, struct json_stringify_StringifyOptions__8122 const a2) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 t0;
 t0 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6088(a0, a1, a2);
 return t0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6119(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, struct environment_root_EnvironmentJson__833 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 t7;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 t9;
 struct environment_root_EnvironmentJson__833 t2;
 uint16_t t4;
 bool t6;
 bool t5;
 bool t8;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(t3);
 if (t4) {
  return t4;
 }
 t5 = true;
 t6 = t5;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t3, (nav__6119_62){((uint8_t const *)&__anon_8509),(uintptr_t)23ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t7 = a1.integration_environment;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6126(t3, t7);
  if (t4) {
   return t4;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = true;
 t6 = t8;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t3, (nav__6119_62){((uint8_t const *)&__anon_8517),(uintptr_t)16ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t9 = a1.linked_libraries;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6127(t3, t9);
  if (t4) {
   return t4;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6089(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 t0 = a0;
 return;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_clearAndFree__1563(struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *t0;
 struct mem_Allocator__599 *t3;
 struct mem_Allocator__599 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 t5;
 nav__1563_43 t6;
 nav__1563_43 *t7;
 uintptr_t *t8;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t0;
 t2 = (*t1);
 t3 = (struct mem_Allocator__599 *)&t2->allocator;
 t4 = (*t3);
 t2 = (*t1);
 t5 = (*t2);
 t6 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t5);
 mem_Allocator_free__anon_3655__5138(t4, t6);
 t2 = (*t1);
 t7 = (nav__1563_43 *)&t2->items;
 t8 = &t7->len;
 (*t8) = (uintptr_t)0ul;
 t2 = (*t1);
 t8 = (uintptr_t *)&t2->capacity;
 (*t8) = (uintptr_t)0ul;
 return;
}

static uint16_t fmt_format__anon_8599__6128(struct io_Writer__3827 const a0) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__6128_43){(uint8_t const *)&__anon_8668,(uintptr_t)63ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__6120_40 fmt_bufPrint__anon_8325__6120(nav__6120_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__6120_54 (**t11)(void const *, nav__6120_56);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__6120_39 t15;
 nav__6120_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6120_54 (**)(void const *, nav__6120_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_8599__6128(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6120_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__3725_39 fs_Dir_createFileZ__3725(struct fs_Dir__1103 const a0, uint8_t const *const a1, struct fs_File_CreateFlags__8142 const a2) {
 uint32_t *t2;
 nav__3725_52 t11;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t10;
 struct fs_File__794 *t15;
 int32_t *t16;
 int32_t t5;
 uint32_t t6;
 uint32_t t1;
 nav__3725_50 t8;
 nav__3725_39 t9;
 nav__3725_39 t14;
 struct fs_Dir__1103 t0;
 uint16_t t7;
 uint8_t t3;
 bool t4;
 t0 = a0;
 t2 = (uint32_t *)&t1;
 t4 = a2.read;
 if (t4) {
  t3 = UINT8_C(2);
  goto zig_block_0;
 }
 t3 = UINT8_C(1);
 goto zig_block_0;

 zig_block_0:;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffc)), zig_shl_u32((uint32_t)t3, UINT8_C(0)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffdff)), zig_shl_u32((uint32_t)true, UINT8_C(9)));
 t2 = (uint32_t *)&t1;
 t4 = a2.truncate;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffbff)), zig_shl_u32((uint32_t)t4, UINT8_C(10)));
 t2 = (uint32_t *)&t1;
 t4 = a2.exclusive;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffff7ff)), zig_shl_u32((uint32_t)t4, UINT8_C(11)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffb)), zig_shl_u32((uint32_t)false, UINT8_C(2)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffff7)), zig_shl_u32((uint32_t)false, UINT8_C(3)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffef)), zig_shl_u32((uint32_t)false, UINT8_C(4)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffdf)), zig_shl_u32((uint32_t)false, UINT8_C(5)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffbf)), zig_shl_u32((uint32_t)false, UINT8_C(6)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffff7f)), zig_shl_u32((uint32_t)false, UINT8_C(7)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)false, UINT8_C(8)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff8fff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(12)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff7fff)), zig_shl_u32((uint32_t)false, UINT8_C(15)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffeffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(16)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffdffff)), zig_shl_u32((uint32_t)false, UINT8_C(17)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff3ffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(18)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffefffff)), zig_shl_u32((uint32_t)false, UINT8_C(20)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffdfffff)), zig_shl_u32((uint32_t)false, UINT8_C(21)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffbfffff)), zig_shl_u32((uint32_t)false, UINT8_C(22)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xff7fffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(23)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfeffffff)), zig_shl_u32((uint32_t)false, UINT8_C(24)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xe1ffffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(25)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xdfffffff)), zig_shl_u32((uint32_t)false, UINT8_C(29)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xbfffffff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(30)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0x7fffffff)), zig_shl_u32((uint32_t)false, UINT8_C(31)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfeffffff)), zig_shl_u32((uint32_t)true, UINT8_C(24)));
 t3 = a2.lock;
 switch (t3) {
  case UINT8_C(0): {
   goto zig_block_1;
  }
  case UINT8_C(1): {
   t2 = (uint32_t *)&t1;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffef)), zig_shl_u32((uint32_t)true, UINT8_C(4)));
   t2 = (uint32_t *)&t1;
   t4 = a2.lock_nonblocking;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffb)), zig_shl_u32((uint32_t)t4, UINT8_C(2)));
   goto zig_block_1;
  }
  case UINT8_C(2): {
   t2 = (uint32_t *)&t1;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffdf)), zig_shl_u32((uint32_t)true, UINT8_C(5)));
   t2 = (uint32_t *)&t1;
   t4 = a2.lock_nonblocking;
   (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffb)), zig_shl_u32((uint32_t)t4, UINT8_C(2)));
   goto zig_block_1;
  }
  default: zig_unreachable();
 }

 zig_block_1:;
 t5 = a0.fd;
 t6 = t1;
 t7 = a2.mode;
 t8 = posix_openatZ__1769(t5, a1, t6, t7);
 if (t8.error) {
  t7 = t8.error;
  t9.payload = (struct fs_File__794){-INT32_C(0x55555556)};
  t9.error = t7;
  return t9;
 }
 t5 = t8.payload;
 t4 = a2.lock_nonblocking;
 if (t4) {
  t11 = posix_fcntl__1928(t5, INT32_C(3), (uintptr_t)0ul);
  t4 = t11.error == UINT16_C(0);
  if (t4) {
   t13 = t11.payload;
   t12 = t13;
   goto zig_block_3;
  }
  t7 = t11.error;
  switch (t7) {
   case zig_error_FileBusy: {
    zig_unreachable();
   }
   case zig_error_Locked: {
    zig_unreachable();
   }
   case zig_error_PermissionDenied: {
    zig_unreachable();
   }
   case zig_error_DeadLock: {
    zig_unreachable();
   }
   case zig_error_LockedRegionLimitExceeded: {
    zig_unreachable();
   }
   default: {
    posix_close__1727(t5);
    t9.payload = (struct fs_File__794){-INT32_C(0x55555556)};
    t9.error = t7;
    return t9;
   }
  }

  zig_block_3:;
  t10 = t12;
  t12 = t10;
  t12 = t12 & (uintptr_t)18446744073709551611ul;
  t10 = t12;
  t12 = t10;
  t11 = posix_fcntl__1928(t5, INT32_C(4), t12);
  t4 = t11.error == UINT16_C(0);
  if (t4) {
   goto zig_block_4;
  }
  t7 = t11.error;
  switch (t7) {
   case zig_error_FileBusy: {
    zig_unreachable();
   }
   case zig_error_Locked: {
    zig_unreachable();
   }
   case zig_error_PermissionDenied: {
    zig_unreachable();
   }
   case zig_error_DeadLock: {
    zig_unreachable();
   }
   case zig_error_LockedRegionLimitExceeded: {
    zig_unreachable();
   }
   default: {
    posix_close__1727(t5);
    t9.payload = (struct fs_File__794){-INT32_C(0x55555556)};
    t9.error = t7;
    return t9;
   }
  }

  zig_block_4:;
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t14.error = UINT16_C(0);
 t15 = &t14.payload;
 t16 = (int32_t *)&t15->handle;
 (*t16) = t5;
 return t14;
}

static uint16_t fmt_format__anon_8674__6129(struct io_Writer__3827 const a0) {
 struct io_Writer__3827 const *t1;
 struct io_Writer__3827 t2;
 struct io_Writer__3827 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5168(t2, (nav__6129_43){(uint8_t const *)&__anon_8741,(uintptr_t)59ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__6121_40 fmt_bufPrint__anon_8340__6121(nav__6121_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *t9;
 void const *t10;
 nav__6121_54 (**t11)(void const *, nav__6121_56);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 nav__6121_39 t15;
 nav__6121_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_1971__4156(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4146(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3817 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__1968 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6121_54 (**)(void const *, nav__6121_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5160;
 t12 = t7;
 t13 = fmt_format__anon_8674__6129(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6121_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4154(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3367(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 void *t3;
 uint64_t t4;
 uint64_t t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t6;
 struct array_hash_map_IndexHeader__880 **t8;
 struct array_hash_map_IndexHeader__880 *t9;
 struct array_hash_map_IndexHeader__880 *t10;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t20;
 struct array_hash_map_IndexHeader__880 *const *t12;
 nav__3367_62 t18;
 struct array_hash_map_IndexHeader__880 t13;
 uint32_t t14;
 uint16_t t7;
 nav__3367_60 t16;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3147(t3);
 t4 = a2;
 t5 = t4 <= UINT64_C(8);
 if (t5) {
  t2 = (*t1);
  t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t2->entries;
  t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3493(t6, a1, a2);
  if (t7) {
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3148(t3);
   return t7;
  }
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t10 = t9;
  t11 = t10;
  t12 = (struct array_hash_map_IndexHeader__880 *const *)&t11;
  t10 = (*t12);
  t13 = (*t10);
  t14 = array_hash_map_IndexHeader_capacity__3139(t13);
  t4 = a2;
  t15 = (uint64_t)t14;
  t5 = t4 <= t15;
  if (t5) {
   t2 = (*t1);
   t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t2->entries;
   t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3493(t6, a1, a2);
   if (t7) {
    t2 = (*t1);
    t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
    debug_SafetyLock_unlock__3148(t3);
    return t7;
   }
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3148(t3);
   return 0;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t2->entries;
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3493(t6, a1, a2);
 if (t7) {
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t16 = array_hash_map_IndexHeader_findBitIndex__3142(a2);
 if (t16.error) {
  t7 = t16.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t17 = t16.payload;
 t18 = array_hash_map_IndexHeader_alloc__3143(a1, t17);
 if (t18.error) {
  t7 = t18.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3148(t3);
  return t7;
 }
 t10 = t18.payload;
 t8 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t19 = t9;
  t20 = t19;
  t12 = (struct array_hash_map_IndexHeader__880 *const *)&t20;
  t19 = (*t12);
  array_hash_map_IndexHeader_free__3144(t19, a1);
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t2 = (*t1);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3459(t2, t10);
 t2 = (*t1);
 t8 = (struct array_hash_map_IndexHeader__880 **)&t2->index_header;
 t19 = (struct array_hash_map_IndexHeader__880 *)t10;
 (*t8) = t19;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3148(t3);
 return 0;
}

static nav__6122_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6122(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6122_41 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_IndexHeader__880 *t3;
 struct array_hash_map_IndexHeader__880 *t5;
 struct array_hash_map_IndexHeader__880 *t24;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t11;
 nav__6122_61 t12;
 nav__6122_64 t13;
 uintptr_t t15;
 uintptr_t t16;
 uintptr_t t14;
 uint64_t t17;
 uint64_t t18;
 nav__6122_41 *t19;
 nav__6122_41 t22;
 nav__6122_38 t23;
 struct array_hash_map_IndexHeader__880 *const *t25;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 uint32_t t6;
 uint32_t t20;
 struct array_hash_map_IndexHeader__880 t26;
 bool t4;
 bool t21;
 uint8_t t27;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t3 = a0.index_header;
 t4 = t3 != NULL;
 if (t4) {
  t5 = t3;
  t2 = t5;
  goto zig_block_0;
 }
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t7 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *)&t1->entries;
 t8 = (*t7);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t8);
 t10 = t9;
 t11 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t10;
 t9 = (*t11);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t9);
 t9 = (*t11);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t9);
 t14 = (uintptr_t)0ul;
 t15 = t13.len;
 zig_loop_32:
 t16 = t14;
 t17 = t16;
 t18 = t15;
 t4 = t17 < t18;
 if (t4) {
  t19 = &t13.ptr[t16];
  t20 = t12.ptr[t16];
  t4 = t20 == t6;
  if (t4) {
   t22 = (*t19);
   t4 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t22, t16);
   t21 = t4;
   goto zig_block_4;
  }
  t21 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t21) {
   t23.is_null = false;
   t23.payload = t16;
   return t23;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t16 = t16 + (uintptr_t)1ul;
 t14 = t16;
 goto zig_loop_32;

 zig_block_1:;
 return (nav__6122_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t24 = t2;
 t25 = (struct array_hash_map_IndexHeader__880 *const *)&t24;
 t5 = (*t25);
 t26 = (*t5);
 t27 = array_hash_map_IndexHeader_capacityIndexType__3138(t26);
 switch (t27) {
  case UINT8_C(0): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6133(t28, a1, t2);
   return t23;
  }
  case UINT8_C(1): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6134(t28, a1, t2);
   return t23;
  }
  case UINT8_C(2): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6135(t28, a1, t2);
   return t23;
  }
  default: zig_unreachable();
 }
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6123(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, nav__6123_42 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t6;
 struct array_hash_map_IndexHeader__880 *t39;
 struct array_hash_map_IndexHeader__880 **t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t13;
 nav__6123_66 t14;
 nav__6123_66 t15;
 nav__6123_66 const *t16;
 nav__6123_69 t17;
 nav__6123_69 t18;
 nav__6123_69 const *t19;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t20;
 uint64_t t23;
 uint64_t t24;
 nav__6123_42 *t25;
 nav__6123_42 t28;
 nav__6123_72 t29;
 nav__6123_72 t30;
 nav__6123_72 t37;
 nav__6123_72 const *t31;
 struct environment_linked_libraries_root_LibraryEntry__852 *t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t33;
 uint32_t *const *t34;
 uint32_t *t35;
 nav__6123_42 *const *t36;
 struct environment_linked_libraries_root_LibraryEntry__852 *const *t38;
 struct array_hash_map_IndexHeader__880 *const *t40;
 uint32_t t7;
 uint32_t t26;
 struct array_hash_map_IndexHeader__880 t41;
 bool t5;
 bool t27;
 uint8_t t42;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t3 = (struct array_hash_map_IndexHeader__880 **)&a0->index_header;
 t4 = (*t3);
 t5 = t4 != NULL;
 if (t5) {
  t6 = t4;
  t2 = t6;
  goto zig_block_0;
 }
 t7 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t8->entries;
 t10 = (*t9);
 t11 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t10);
 t12 = t11;
 t13 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t12;
 t11 = (*t13);
 t14 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t11);
 t15 = t14;
 t16 = (nav__6123_66 const *)&t15;
 t11 = (*t13);
 t17 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t11);
 t18 = t17;
 t19 = (nav__6123_69 const *)&t18;
 t20 = (uintptr_t)0ul;
 t21 = t17.len;
 zig_loop_39:
 t22 = t20;
 t23 = t22;
 t24 = t21;
 t5 = t23 < t24;
 if (t5) {
  t25 = &t17.ptr[t22];
  t26 = t14.ptr[t22];
  t5 = t26 == t7;
  if (t5) {
   t28 = (*t25);
   t5 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t28, t22);
   t27 = t5;
   goto zig_block_4;
  }
  t27 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t27) {
   t11 = (*t13);
   t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t11);
   t30 = t29;
   t31 = (nav__6123_72 const *)&t30;
   t29 = (*t31);
   t32 = &t29.ptr[t22];
   t33.key_ptr = t25;
   t33.value_ptr = t32;
   t33.index = t22;
   t33.found_existing = true;
   return t33;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t22 = t22 + (uintptr_t)1ul;
 t20 = t22;
 goto zig_loop_39;

 zig_block_1:;
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t8->entries;
 t21 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t9);
 t34 = &t16->ptr;
 t35 = (*t34);
 t35 = (uint32_t *)&t35[t21];
 (*t35) = t7;
 t36 = &t19->ptr;
 t25 = (*t36);
 t25 = (nav__6123_42 *)&t25[t21];
 t11 = (*t13);
 t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t11);
 t37 = t29;
 t31 = (nav__6123_72 const *)&t37;
 t38 = &t31->ptr;
 t32 = (*t38);
 t32 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t32[t21];
 t33.key_ptr = t25;
 t33.value_ptr = t32;
 t33.index = t21;
 t33.found_existing = false;
 return t33;

 zig_block_0:;
 t39 = t2;
 t40 = (struct array_hash_map_IndexHeader__880 *const *)&t39;
 t6 = (*t40);
 t41 = (*t6);
 t42 = array_hash_map_IndexHeader_capacityIndexType__3138(t41);
 switch (t42) {
  case UINT8_C(0): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6136(t8, a1, t2);
   return t33;
  }
  case UINT8_C(1): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6137(t8, a1, t2);
   return t33;
  }
  case UINT8_C(2): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6138(t8, a1, t2);
   return t33;
  }
  default: zig_unreachable();
 }
}

static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6088(struct mem_Allocator__599 const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a1, struct json_stringify_StringifyOptions__8122 const a2) {
 struct json_stringify_StringifyOptions__8122 *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t2;
 uintptr_t *t3;
 uint8_t *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 t0;
 (void)a0;
 t1 = (struct json_stringify_StringifyOptions__8122 *)&t0.options;
 (*t1) = a2;
 t2 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t0.stream;
 (*t2) = a1;
 t3 = (uintptr_t *)&t0.indent_level;
 (*t3) = (uintptr_t)0ul;
 t4 = (uint8_t *)&t0.next_punctuation;
 (*t4) = UINT8_C(0);
 return t0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6096(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_StringifyOptions__8122 *t3;
 uintptr_t t6;
 uintptr_t t8;
 uintptr_t *t7;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t9;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t26;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t13;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t14;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t27;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t15;
 void const **t17;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t18;
 void const *t19;
 nav__6096_56 (**t20)(void const *, nav__6096_58);
 struct io_Writer__3827 t21;
 struct io_Writer__3827 t16;
 struct io_Writer__3827 t22;
 struct io_Writer__3827 t28;
 struct io_Writer__3827 t29;
 struct io_Writer__3827 const *t23;
 uint16_t t24;
 uint16_t t25;
 struct json_stringify_StringifyOptions__8122 t4;
 uint8_t t5;
 uint8_t t2;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = UINT8_C(32);
 t3 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
 t4 = (*t3);
 t5 = t4.whitespace;
 switch (t5) {
  case UINT8_C(0): {
   return 0;
  }
  case UINT8_C(1): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)2ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)3ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)4ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(5): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)8ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(6): {
   t2 = UINT8_C(9);
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t6 = t8;
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t9 = (*t1);
 t10 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t9->stream;
 t11 = (*t10);
 t12 = t11;
 t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t12;
 t14 = t13;
 t15 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t14;
 t17 = (void const **)&t16.context;
 t13 = (*t15);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t13->context;
 t19 = (void const *)t18;
 (*t17) = t19;
 t20 = (nav__6096_56 (**)(void const *, nav__6096_58))&t16.writeFn;
 (*t20) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t21 = t16;
 t22 = t21;
 t23 = (struct io_Writer__3827 const *)&t22;
 t21 = (*t23);
 t24 = io_Writer_writeByte__5170(t21, UINT8_C(10));
 memcpy(&t25, &t24, sizeof(uint16_t));
 if (t25) {
  return t25;
 }
 t9 = (*t1);
 t10 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t9->stream;
 t11 = (*t10);
 t5 = t2;
 t26 = t11;
 t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t26;
 t27 = t13;
 t15 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t27;
 t17 = (void const **)&t28.context;
 t13 = (*t15);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t13->context;
 t19 = (void const *)t18;
 (*t17) = t19;
 t20 = (nav__6096_56 (**)(void const *, nav__6096_58))&t28.writeFn;
 (*t20) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t21 = t28;
 t29 = t21;
 t23 = (struct io_Writer__3827 const *)&t29;
 t21 = (*t23);
 t25 = io_Writer_writeByteNTimes__5171(t21, t5, t6);
 memcpy(&t24, &t25, sizeof(uint16_t));
 if (t24) {
  return t24;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6099(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *t3;
 uint8_t *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t23;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t29;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t30;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t13;
 void const **t15;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t16;
 void const *t17;
 nav__6099_56 (**t18)(void const *, nav__6099_58);
 struct io_Writer__3827 t19;
 struct io_Writer__3827 t14;
 struct io_Writer__3827 t20;
 struct io_Writer__3827 t25;
 struct io_Writer__3827 t26;
 struct io_Writer__3827 t31;
 struct io_Writer__3827 t32;
 struct io_Writer__3827 const *t21;
 struct json_stringify_StringifyOptions__8122 *t27;
 uint16_t t7;
 uint16_t t22;
 bool t4;
 uint8_t t6;
 struct json_stringify_StringifyOptions__8122 t28;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *)t2;
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6102(t3);
 t4 = !t4;
 debug_assert__180(t4);
 t5 = (uint8_t *)&a0->next_punctuation;
 t6 = (*t5);
 switch (t6) {
  case UINT8_C(0): {
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t2 = (*t1);
   t7 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6096(t2);
   if (t7) {
    return t7;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = (*t1);
   t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
   t9 = (*t8);
   t10 = t9;
   t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t10;
   t12 = t11;
   t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t12;
   t15 = (void const **)&t14.context;
   t11 = (*t13);
   t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t11->context;
   t17 = (void const *)t16;
   (*t15) = t17;
   t18 = (nav__6099_56 (**)(void const *, nav__6099_58))&t14.writeFn;
   (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t19 = t14;
   t20 = t19;
   t21 = (struct io_Writer__3827 const *)&t20;
   t19 = (*t21);
   t7 = io_Writer_writeByte__5170(t19, UINT8_C(44));
   memcpy(&t22, &t7, sizeof(uint16_t));
   if (t22) {
    return t22;
   }
   t2 = (*t1);
   t22 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6096(t2);
   if (t22) {
    return t22;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t2 = (*t1);
   t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
   t9 = (*t8);
   t23 = t9;
   t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t23;
   t24 = t11;
   t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t24;
   t15 = (void const **)&t25.context;
   t11 = (*t13);
   t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t11->context;
   t17 = (void const *)t16;
   (*t15) = t17;
   t18 = (nav__6099_56 (**)(void const *, nav__6099_58))&t25.writeFn;
   (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t19 = t25;
   t26 = t19;
   t21 = (struct io_Writer__3827 const *)&t26;
   t19 = (*t21);
   t22 = io_Writer_writeByte__5170(t19, UINT8_C(58));
   memcpy(&t7, &t22, sizeof(uint16_t));
   if (t7) {
    return t7;
   }
   t27 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
   t28 = (*t27);
   t6 = t28.whitespace;
   t4 = t6 != UINT8_C(0);
   if (t4) {
    t2 = (*t1);
    t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
    t9 = (*t8);
    t29 = t9;
    t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t29;
    t30 = t11;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t30;
    t15 = (void const **)&t31.context;
    t11 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t11->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6099_56 (**)(void const *, nav__6099_58))&t31.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t31;
    t32 = t19;
    t21 = (struct io_Writer__3827 const *)&t32;
    t19 = (*t21);
    t7 = io_Writer_writeByte__5170(t19, UINT8_C(32));
    memcpy(&t22, &t7, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    goto zig_block_1;
   }
   goto zig_block_1;

   zig_block_1:;
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *t3;
 uint16_t t6;
 uint16_t t7;
 nav__6097_43 t4;
 bool t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *)t2;
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6101(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t5 = t4.payload;
  t5 = !t5;
  debug_assert__180(t5);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6099(t2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6094(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, uint8_t const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uintptr_t *t3;
 uintptr_t t4;
 (void)a1;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t *)&t2->indent_level;
 t4 = (*t3);
 t4 = t4 + (uintptr_t)1ul;
 (*t3) = t4;
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t9;
 void const **t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t12;
 void const *t13;
 nav__6091_54 (**t14)(void const *, nav__6091_56);
 struct io_Writer__3827 t15;
 struct io_Writer__3827 t10;
 struct io_Writer__3827 t16;
 struct io_Writer__3827 const *t17;
 uint8_t *t19;
 uint16_t t3;
 uint16_t t18;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(t2);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
 t5 = (*t4);
 t6 = t5;
 t7 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t6;
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t8;
 t11 = (void const **)&t10.context;
 t7 = (*t9);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t7->context;
 t13 = (void const *)t12;
 (*t11) = t13;
 t14 = (nav__6091_54 (**)(void const *, nav__6091_56))&t10.writeFn;
 (*t14) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t15 = t10;
 t16 = t15;
 t17 = (struct io_Writer__3827 const *)&t16;
 t15 = (*t17);
 t3 = io_Writer_writeByte__5170(t15, UINT8_C(123));
 memcpy(&t18, &t3, sizeof(uint16_t));
 if (t18) {
  return t18;
 }
 t2 = (*t1);
 (void)json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6094(t2, UINT8_C(0));
 t2 = (*t1);
 t19 = (uint8_t *)&t2->next_punctuation;
 (*t19) = UINT8_C(1);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6098(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *t3;
 uint16_t t6;
 uint16_t t7;
 nav__6098_43 t4;
 bool t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *)t2;
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6101(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t5 = t4.payload;
  debug_assert__180(t5);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6099(t2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t fmt_formatBuf__anon_8984__6145(nav__6145_39 const a0, struct fmt_FormatOptions__4248 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a2) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t33;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t34;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t38;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t42;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t46;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t50;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t54;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t58;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t62;
 nav__6145_44 t2;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 nav__6145_52 t6;
 uint64_t t8;
 uint64_t t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t32;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t37;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t41;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t45;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t49;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t53;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t57;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t61;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t13;
 void const **t15;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t16;
 void const *t17;
 nav__6145_52 (**t18)(void const *, nav__6145_39);
 struct io_Writer__3827 t19;
 struct io_Writer__3827 t14;
 struct io_Writer__3827 t20;
 struct io_Writer__3827 t35;
 struct io_Writer__3827 t36;
 struct io_Writer__3827 t39;
 struct io_Writer__3827 t40;
 struct io_Writer__3827 t43;
 struct io_Writer__3827 t44;
 struct io_Writer__3827 t47;
 struct io_Writer__3827 t48;
 struct io_Writer__3827 t51;
 struct io_Writer__3827 t52;
 struct io_Writer__3827 t55;
 struct io_Writer__3827 t56;
 struct io_Writer__3827 t59;
 struct io_Writer__3827 t60;
 struct io_Writer__3827 t63;
 struct io_Writer__3827 t64;
 struct io_Writer__3827 const *t21;
 nav__6145_67 t26;
 nav__6145_39 t28;
 nav__6145_39 t31;
 uint8_t *t30;
 uint32_t t25;
 uint16_t t22;
 uint16_t t23;
 nav__6145_70 t27;
 bool t3;
 uint8_t t29;
 uint8_t t24[4];
 t0 = a2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = a1.width;
 t3 = t2.is_null != true;
 if (t3) {
  t4 = t2.payload;
  t6 = unicode_utf8CountCodepoints__5205(a0);
  t3 = t6.error == UINT16_C(0);
  if (t3) {
   t7 = t6.payload;
   t5 = t7;
   goto zig_block_1;
  }
  t7 = a0.len;
  t5 = t7;
  goto zig_block_1;

  zig_block_1:;
  t8 = t5;
  t9 = t4;
  t3 = t8 < t9;
  if (t3) {
   t5 = t4 - t5;
   t7 = t5;
   goto zig_block_2;
  }
  t7 = (uintptr_t)0ul;
  goto zig_block_2;

  zig_block_2:;
  t9 = t7;
  t3 = t9 == UINT64_C(0);
  if (t3) {
   t10 = (*t1);
   t11 = t10;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t11;
   t12 = t1;
   t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t12;
   t15 = (void const **)&t14.context;
   t1 = (*t13);
   t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t17 = (void const *)t16;
   (*t15) = t17;
   t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t14.writeFn;
   (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t19 = t14;
   t20 = t19;
   t21 = (struct io_Writer__3827 const *)&t20;
   t19 = (*t21);
   t22 = io_Writer_writeAll__5168(t19, a0);
   memcpy(&t23, &t22, sizeof(uint16_t));
   memcpy(&t22, &t23, sizeof(uint16_t));
   return t22;
  }
  goto zig_block_3;

  zig_block_3:;
  t25 = a1.fill;
  t26.ptr = &t24[(uintptr_t)0ul];
  t26.len = (uintptr_t)4ul;
  t27 = unicode_utf8Encode__5190(t25, t26);
  t3 = t27.error == UINT16_C(0);
  if (t3) {
   t29 = t27.payload;
   t30 = (uint8_t *)&t24;
   t30 = (uint8_t *)(((uintptr_t)t30) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t5 = (uintptr_t)t29;
   t26.ptr = t30;
   t26.len = t5;
   memcpy(&t31, &t26, sizeof(nav__6145_39));
   t28 = t31;
   goto zig_block_4;

  }
  t23 = t27.error;
  switch (t23) {
   case zig_error_Utf8CannotEncodeSurrogateHalf:
   case zig_error_CodepointTooLarge: {
    t28 = (nav__6145_39){(uint8_t const *)&__anon_5278,(uintptr_t)3ul};
    goto zig_block_4;
   }
   default: zig_unreachable();
  }

  zig_block_4:;
  t29 = a1.alignment;
  switch (t29) {
   case UINT8_C(0): {
    t10 = (*t1);
    t32 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t32;
    t34 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t34;
    t15 = (void const **)&t35.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t35.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t35;
    t36 = t19;
    t21 = (struct io_Writer__3827 const *)&t36;
    t19 = (*t21);
    t23 = io_Writer_writeAll__5168(t19, a0);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    t10 = (*t1);
    t37 = t10;
    t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t37;
    t38 = t1;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t38;
    t15 = (void const **)&t39.context;
    t1 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t39.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t39;
    t40 = t19;
    t21 = (struct io_Writer__3827 const *)&t40;
    t19 = (*t21);
    t22 = io_Writer_writeBytesNTimes__5172(t19, t28, t7);
    memcpy(&t23, &t22, sizeof(uint16_t));
    if (t23) {
     return t23;
    }
    goto zig_block_6;
   }
   case UINT8_C(1): {
    t5 = t7 / (uintptr_t)2ul;
    t7 = t7 + (uintptr_t)1ul;
    t7 = t7 / (uintptr_t)2ul;
    t10 = (*t1);
    t41 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t41;
    t42 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t42;
    t15 = (void const **)&t43.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t43.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t43;
    t44 = t19;
    t21 = (struct io_Writer__3827 const *)&t44;
    t19 = (*t21);
    t23 = io_Writer_writeBytesNTimes__5172(t19, t28, t5);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    t10 = (*t1);
    t45 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t45;
    t46 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t46;
    t15 = (void const **)&t47.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t47.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t47;
    t48 = t19;
    t21 = (struct io_Writer__3827 const *)&t48;
    t19 = (*t21);
    t22 = io_Writer_writeAll__5168(t19, a0);
    memcpy(&t23, &t22, sizeof(uint16_t));
    if (t23) {
     return t23;
    }
    t10 = (*t1);
    t49 = t10;
    t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t49;
    t50 = t1;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t50;
    t15 = (void const **)&t51.context;
    t1 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t51.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t51;
    t52 = t19;
    t21 = (struct io_Writer__3827 const *)&t52;
    t19 = (*t21);
    t23 = io_Writer_writeBytesNTimes__5172(t19, t28, t7);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    goto zig_block_6;
   }
   case UINT8_C(2): {
    t10 = (*t1);
    t53 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t53;
    t54 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t54;
    t15 = (void const **)&t55.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t55.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t55;
    t56 = t19;
    t21 = (struct io_Writer__3827 const *)&t56;
    t19 = (*t21);
    t23 = io_Writer_writeBytesNTimes__5172(t19, t28, t7);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    t10 = (*t1);
    t57 = t10;
    t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t57;
    t58 = t1;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t58;
    t15 = (void const **)&t59.context;
    t1 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t59.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t19 = t59;
    t60 = t19;
    t21 = (struct io_Writer__3827 const *)&t60;
    t19 = (*t21);
    t22 = io_Writer_writeAll__5168(t19, a0);
    memcpy(&t23, &t22, sizeof(uint16_t));
    if (t23) {
     return t23;
    }
    goto zig_block_6;
   }
   default: zig_unreachable();
  }

  zig_block_6:;
  goto zig_block_0;
 }
 t10 = (*t1);
 t61 = t10;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t61;
 t62 = t1;
 t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t62;
 t15 = (void const **)&t63.context;
 t1 = (*t13);
 t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t17 = (void const *)t16;
 (*t15) = t17;
 t18 = (nav__6145_52 (**)(void const *, nav__6145_39))&t63.writeFn;
 (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t19 = t63;
 t64 = t19;
 t21 = (struct io_Writer__3827 const *)&t64;
 t19 = (*t21);
 t23 = io_Writer_writeAll__5168(t19, a0);
 memcpy(&t22, &t23, sizeof(uint16_t));
 if (t22) {
  return t22;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatInt__anon_8974__6144(uint32_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4248 const a3, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__6144_54 t12;
 nav__6144_57 t13;
 uint32_t t4;
 uint32_t t11;
 uint32_t t2;
 uint16_t t14;
 uint16_t t15;
 bool t0;
 uint8_t t8;
 nav__6144_52 t9;
 uint8_t t10[2];
 uint8_t t1[22];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)22ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t4 = t2;
  t0 = t4 >= UINT32_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT32_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__4953(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT32_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT32_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__4953(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t4 = t2;
 t11 = (uint32_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__4969(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint32_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT32_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)22ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__6144_57));
 t14 = fmt_formatBuf__anon_8984__6145(t13, a3, a4);
 memcpy(&t15, &t14, sizeof(uint16_t));
 return t15;
}

static uint16_t fmt_formatIntValue__anon_8949__6143(uint32_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_8974__6144(a0, UINT8_C(16), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatInt__anon_9000__6147(uint16_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4248 const a3, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__6147_54 t12;
 nav__6147_57 t13;
 uint16_t t4;
 uint16_t t11;
 uint16_t t2;
 bool t0;
 uint8_t t8;
 nav__6147_52 t9;
 uint8_t t10[2];
 uint8_t t1[17];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)17ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t4 = t2;
  t0 = t4 >= UINT16_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT16_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__4953(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT16_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT16_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__4953(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t4 = t2;
 t11 = (uint16_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__4969(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint16_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT16_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)17ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__6147_57));
 t11 = fmt_formatBuf__anon_8984__6145(t13, a3, a4);
 memcpy(&t4, &t11, sizeof(uint16_t));
 return t4;
}

static uint16_t fmt_formatIntValue__anon_8996__6146(uint16_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_9000__6147(a0, UINT8_C(16), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t json_stringify_outputUnicodeEscape__anon_8933__6142(uint32_t const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t19;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t20;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t26;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t18;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t25;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t6;
 void const **t8;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t9;
 void const *t10;
 nav__6142_48 (**t11)(void const *, nav__6142_50);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 struct io_Writer__3827 t13;
 struct io_Writer__3827 t21;
 struct io_Writer__3827 t22;
 struct io_Writer__3827 t27;
 struct io_Writer__3827 t28;
 struct io_Writer__3827 const *t14;
 uint32_t t17;
 uint16_t t15;
 uint16_t t16;
 uint16_t t23;
 uint16_t t24;
 bool t2;
 t0 = a1;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = a0 <= UINT32_C(65535);
 if (t2) {
  t3 = (*t1);
  t4 = t3;
  t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t4;
  t5 = t1;
  t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t5;
  t8 = (void const **)&t7.context;
  t1 = (*t6);
  t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
  t10 = (void const *)t9;
  (*t8) = t10;
  t11 = (nav__6142_48 (**)(void const *, nav__6142_50))&t7.writeFn;
  (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
  t12 = t7;
  t13 = t12;
  t14 = (struct io_Writer__3827 const *)&t13;
  t12 = (*t14);
  t15 = io_Writer_writeAll__5168(t12, (nav__6142_50){(uint8_t const *)&__anon_8939,(uintptr_t)2ul});
  memcpy(&t16, &t15, sizeof(uint16_t));
  if (t16) {
   return t16;
  }
  t16 = fmt_formatIntValue__anon_8949__6143(a0, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)4ul,false},UINT32_C(48),UINT8_C(2)}, a1);
  if (t16) {
   return t16;
  }
  goto zig_block_0;
 }
 t2 = a0 <= UINT32_C(1114111);
 debug_assert__180(t2);
 t17 = a0 - UINT32_C(65536);
 t17 = zig_shr_u32(t17, UINT8_C(10));
 t16 = (uint16_t)t17;
 t16 = t16 + UINT16_C(55296);
 t17 = a0 & UINT32_C(1023);
 t15 = (uint16_t)t17;
 t15 = t15 + UINT16_C(56320);
 t3 = (*t1);
 t18 = t3;
 t19 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t18;
 t20 = t19;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t20;
 t8 = (void const **)&t21.context;
 t19 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t19->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6142_48 (**)(void const *, nav__6142_50))&t21.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t12 = t21;
 t22 = t12;
 t14 = (struct io_Writer__3827 const *)&t22;
 t12 = (*t14);
 t23 = io_Writer_writeAll__5168(t12, (nav__6142_50){(uint8_t const *)&__anon_8939,(uintptr_t)2ul});
 memcpy(&t24, &t23, sizeof(uint16_t));
 if (t24) {
  return t24;
 }
 t16 = fmt_formatIntValue__anon_8996__6146(t16, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)4ul,false},UINT32_C(48),UINT8_C(2)}, a1);
 if (t16) {
  return t16;
 }
 t3 = (*t1);
 t25 = t3;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t25;
 t26 = t1;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t26;
 t8 = (void const **)&t27.context;
 t1 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6142_48 (**)(void const *, nav__6142_50))&t27.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t12 = t27;
 t28 = t12;
 t14 = (struct io_Writer__3827 const *)&t28;
 t12 = (*t14);
 t16 = io_Writer_writeAll__5168(t12, (nav__6142_50){(uint8_t const *)&__anon_8939,(uintptr_t)2ul});
 memcpy(&t24, &t16, sizeof(uint16_t));
 if (t24) {
  return t24;
 }
 t15 = fmt_formatIntValue__anon_8996__6146(t15, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)4ul,false},UINT32_C(48),UINT8_C(2)}, a1);
 if (t15) {
  return t15;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_outputSpecialEscape__anon_8893__6141(uint8_t const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t17;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t21;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t25;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t29;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t33;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t37;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t16;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t20;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t28;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t32;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t36;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t8;
 void const *t9;
 nav__6141_48 (**t10)(void const *, nav__6141_50);
 struct io_Writer__3827 t11;
 struct io_Writer__3827 t6;
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t18;
 struct io_Writer__3827 t19;
 struct io_Writer__3827 t22;
 struct io_Writer__3827 t23;
 struct io_Writer__3827 t26;
 struct io_Writer__3827 t27;
 struct io_Writer__3827 t30;
 struct io_Writer__3827 t31;
 struct io_Writer__3827 t34;
 struct io_Writer__3827 t35;
 struct io_Writer__3827 t38;
 struct io_Writer__3827 t39;
 struct io_Writer__3827 const *t13;
 uint32_t t40;
 uint16_t t14;
 uint16_t t15;
 t0 = a1;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 switch (a0) {
  case UINT8_C(92): {
   t2 = (*t1);
   t3 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t3;
   t4 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t4;
   t7 = (void const **)&t6.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t6.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t6;
   t12 = t11;
   t13 = (struct io_Writer__3827 const *)&t12;
   t11 = (*t13);
   t14 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8898,(uintptr_t)2ul});
   memcpy(&t15, &t14, sizeof(uint16_t));
   if (t15) {
    return t15;
   }
   goto zig_block_0;
  }
  case UINT8_C(34): {
   t2 = (*t1);
   t16 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t16;
   t17 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t17;
   t7 = (void const **)&t18.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t18.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t18;
   t19 = t11;
   t13 = (struct io_Writer__3827 const *)&t19;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8903,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(8): {
   t2 = (*t1);
   t20 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t20;
   t21 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t21;
   t7 = (void const **)&t22.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t22.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t22;
   t23 = t11;
   t13 = (struct io_Writer__3827 const *)&t23;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8907,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(12): {
   t2 = (*t1);
   t24 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t24;
   t25 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t25;
   t7 = (void const **)&t26.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t26.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t26;
   t27 = t11;
   t13 = (struct io_Writer__3827 const *)&t27;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8911,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(10): {
   t2 = (*t1);
   t28 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t28;
   t29 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t29;
   t7 = (void const **)&t30.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t30.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t30;
   t31 = t11;
   t13 = (struct io_Writer__3827 const *)&t31;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8915,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(13): {
   t2 = (*t1);
   t32 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t32;
   t33 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t33;
   t7 = (void const **)&t34.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t34.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t34;
   t35 = t11;
   t13 = (struct io_Writer__3827 const *)&t35;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8919,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(9): {
   t2 = (*t1);
   t36 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t36;
   t37 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t37;
   t7 = (void const **)&t38.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__6141_48 (**)(void const *, nav__6141_50))&t38.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
   t11 = t38;
   t39 = t11;
   t13 = (struct io_Writer__3827 const *)&t39;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5168(t11, (nav__6141_50){(uint8_t const *)&__anon_8923,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  default: {
   t40 = (uint32_t)a0;
   t15 = json_stringify_outputUnicodeEscape__anon_8933__6142(t40, a1);
   if (t15) {
    return t15;
   }
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_encodeJsonStringChars__anon_8883__6140(nav__6140_39 const a0, struct json_stringify_StringifyOptions__8122 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a2) {
 nav__6140_39 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t16;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t17;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t30;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t39;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t43;
 uintptr_t t7;
 uintptr_t t8;
 uintptr_t t4;
 uintptr_t t5;
 uint64_t t9;
 uint64_t t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t15;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t29;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t38;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t42;
 nav__6140_39 t13;
 nav__6140_39 t0;
 uint8_t const *t14;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t18;
 void const **t20;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t21;
 void const *t22;
 nav__6140_54 (**t23)(void const *, nav__6140_39);
 struct io_Writer__3827 t24;
 struct io_Writer__3827 t19;
 struct io_Writer__3827 t25;
 struct io_Writer__3827 t31;
 struct io_Writer__3827 t32;
 struct io_Writer__3827 t40;
 struct io_Writer__3827 t41;
 struct io_Writer__3827 t44;
 struct io_Writer__3827 t45;
 struct io_Writer__3827 const *t26;
 uint32_t t35;
 uint32_t t37;
 nav__6140_66 t36;
 uint16_t t27;
 uint16_t t28;
 nav__6140_64 t34;
 bool t6;
 uint8_t t11;
 uint8_t t33;
 t0 = a0;
 t1 = (nav__6140_39 const *)&t0;
 t2 = a2;
 t3 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t2;
 t4 = (uintptr_t)0ul;
 t5 = (uintptr_t)0ul;
 t6 = a1.escape_unicode;
 if (t6) {
  zig_loop_16:
  t7 = t5;
  t8 = a0.len;
  t9 = t7;
  t10 = t8;
  t6 = t9 < t10;
  if (t6) {
   t8 = t5;
   t11 = a0.ptr[t8];
   switch (t11) {
    default: if ((t11 >= UINT8_C(32) && t11 <= UINT8_C(33)) || (t11 >= UINT8_C(35) && t11 <= UINT8_C(91)) || (t11 >= UINT8_C(93) && t11 <= UINT8_C(126))) {
     goto zig_block_3;
    }if (t11 == UINT8_C(92) || t11 == UINT8_C(34) || (t11 >= UINT8_C(0) && t11 <= UINT8_C(31))) {
     t12 = (*t3);
     t8 = t4;
     t7 = t5;
     t13 = (*t1);
     t14 = t13.ptr;
     t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
     t8 = t7 - t8;
     t13.ptr = t14;
     t13.len = t8;
     t15 = t12;
     t16 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t15;
     t17 = t16;
     t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t17;
     t20 = (void const **)&t19.context;
     t16 = (*t18);
     t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t16->context;
     t22 = (void const *)t21;
     (*t20) = t22;
     t23 = (nav__6140_54 (**)(void const *, nav__6140_39))&t19.writeFn;
     (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
     t24 = t19;
     t25 = t24;
     t26 = (struct io_Writer__3827 const *)&t25;
     t24 = (*t26);
     t27 = io_Writer_writeAll__5168(t24, t13);
     memcpy(&t28, &t27, sizeof(uint16_t));
     if (t28) {
      return t28;
     }
     t8 = t5;
     t11 = a0.ptr[t8];
     t28 = json_stringify_outputSpecialEscape__anon_8893__6141(t11, a2);
     if (t28) {
      return t28;
     }
     t8 = t5;
     t8 = t8 + (uintptr_t)1ul;
     t4 = t8;
     goto zig_block_3;
    }if ((t11 >= UINT8_C(127) && t11 <= UINT8_MAX)) {
     t12 = (*t3);
     t8 = t4;
     t7 = t5;
     t13 = (*t1);
     t14 = t13.ptr;
     t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
     t8 = t7 - t8;
     t13.ptr = t14;
     t13.len = t8;
     t29 = t12;
     t16 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t29;
     t30 = t16;
     t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t30;
     t20 = (void const **)&t31.context;
     t16 = (*t18);
     t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t16->context;
     t22 = (void const *)t21;
     (*t20) = t22;
     t23 = (nav__6140_54 (**)(void const *, nav__6140_39))&t31.writeFn;
     (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
     t24 = t31;
     t32 = t24;
     t26 = (struct io_Writer__3827 const *)&t32;
     t24 = (*t26);
     t28 = io_Writer_writeAll__5168(t24, t13);
     memcpy(&t27, &t28, sizeof(uint16_t));
     if (t27) {
      return t27;
     }
     t8 = t5;
     t33 = a0.ptr[t8];
     t34 = unicode_utf8ByteSequenceLength__5189(t33);
     t6 = t34.error == UINT16_C(0);
     if (t6) {
      t33 = t34.payload;
      t11 = t33;
      goto zig_block_4;
     }
     zig_unreachable();

     zig_block_4:;
     t8 = t5;
     t13 = (*t1);
     t14 = t13.ptr;
     t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
     t8 = (uintptr_t)t11;
     t13.ptr = t14;
     t13.len = t8;
     t36 = unicode_utf8Decode__5195(t13);
     t6 = t36.error == UINT16_C(0);
     if (t6) {
      t37 = t36.payload;
      t35 = t37;
      goto zig_block_5;
     }
     zig_unreachable();

     zig_block_5:;
     t27 = json_stringify_outputUnicodeEscape__anon_8933__6142(t35, a2);
     if (t27) {
      return t27;
     }
     t8 = t5;
     t11 = t11 - UINT8_C(1);
     t7 = (uintptr_t)t11;
     t7 = t8 + t7;
     t5 = t7;
     t7 = t5;
     t7 = t7 + (uintptr_t)1ul;
     t4 = t7;
     goto zig_block_3;
    }zig_unreachable();
   }

   zig_block_3:;
   t8 = t5;
   t8 = t8 + (uintptr_t)1ul;
   t5 = t8;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_16;

  zig_block_1:;
  goto zig_block_0;
 }
 zig_loop_172:
 t8 = t5;
 t7 = a0.len;
 t10 = t8;
 t9 = t7;
 t6 = t10 < t9;
 if (t6) {
  t7 = t5;
  t33 = a0.ptr[t7];
  switch (t33) {
   default: if ((t33 >= UINT8_C(32) && t33 <= UINT8_C(33)) || (t33 >= UINT8_C(35) && t33 <= UINT8_C(91)) || (t33 >= UINT8_C(93) && t33 <= UINT8_MAX)) {
    goto zig_block_8;
   }if (t33 == UINT8_C(92) || t33 == UINT8_C(34) || (t33 >= UINT8_C(0) && t33 <= UINT8_C(31))) {
    t12 = (*t3);
    t7 = t4;
    t8 = t5;
    t13 = (*t1);
    t14 = t13.ptr;
    t14 = (uint8_t const *)(((uintptr_t)t14) + (t7*sizeof(uint8_t)));
    t7 = t8 - t7;
    t13.ptr = t14;
    t13.len = t7;
    t38 = t12;
    t16 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t38;
    t39 = t16;
    t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t39;
    t20 = (void const **)&t40.context;
    t16 = (*t18);
    t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t16->context;
    t22 = (void const *)t21;
    (*t20) = t22;
    t23 = (nav__6140_54 (**)(void const *, nav__6140_39))&t40.writeFn;
    (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
    t24 = t40;
    t41 = t24;
    t26 = (struct io_Writer__3827 const *)&t41;
    t24 = (*t26);
    t28 = io_Writer_writeAll__5168(t24, t13);
    memcpy(&t27, &t28, sizeof(uint16_t));
    if (t27) {
     return t27;
    }
    t7 = t5;
    t33 = a0.ptr[t7];
    t27 = json_stringify_outputSpecialEscape__anon_8893__6141(t33, a2);
    if (t27) {
     return t27;
    }
    t7 = t5;
    t7 = t7 + (uintptr_t)1ul;
    t4 = t7;
    goto zig_block_8;
   }zig_unreachable();
  }

  zig_block_8:;
  t7 = t5;
  t7 = t7 + (uintptr_t)1ul;
  t5 = t7;
  goto zig_block_7;
 }
 goto zig_block_6;

 zig_block_7:;
 goto zig_loop_172;

 zig_block_6:;
 goto zig_block_0;

 zig_block_0:;
 t12 = (*t3);
 t8 = t4;
 t7 = a0.len;
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
 t8 = t7 - t8;
 t13.ptr = t14;
 t13.len = t8;
 t42 = t12;
 t3 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t42;
 t43 = t3;
 t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t43;
 t20 = (void const **)&t44.context;
 t3 = (*t18);
 t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t3->context;
 t22 = (void const *)t21;
 (*t20) = t22;
 t23 = (nav__6140_54 (**)(void const *, nav__6140_39))&t44.writeFn;
 (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t24 = t44;
 t45 = t24;
 t26 = (struct io_Writer__3827 const *)&t45;
 t24 = (*t26);
 t28 = io_Writer_writeAll__5168(t24, t13);
 memcpy(&t27, &t28, sizeof(uint16_t));
 if (t27) {
  return t27;
 }
 return 0;
}

static uint16_t json_stringify_encodeJsonString__anon_8873__6139(nav__6139_39 const a0, struct json_stringify_StringifyOptions__8122 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const a2) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t18;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t17;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t6;
 void const **t8;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t9;
 void const *t10;
 nav__6139_53 (**t11)(void const *, nav__6139_39);
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t7;
 struct io_Writer__3827 t13;
 struct io_Writer__3827 t19;
 struct io_Writer__3827 t20;
 struct io_Writer__3827 const *t14;
 uint16_t t15;
 uint16_t t16;
 t0 = a2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6139_53 (**)(void const *, nav__6139_39))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t12 = t7;
 t13 = t12;
 t14 = (struct io_Writer__3827 const *)&t13;
 t12 = (*t14);
 t15 = io_Writer_writeByte__5170(t12, UINT8_C(34));
 memcpy(&t16, &t15, sizeof(uint16_t));
 if (t16) {
  return t16;
 }
 t16 = json_stringify_encodeJsonStringChars__anon_8883__6140(a0, a1, a2);
 if (t16) {
  return t16;
 }
 t2 = (*t1);
 t17 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t17;
 t18 = t1;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t18;
 t8 = (void const **)&t19.context;
 t1 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t1->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6139_53 (**)(void const *, nav__6139_39))&t19.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t12 = t19;
 t20 = t12;
 t14 = (struct io_Writer__3827 const *)&t20;
 t12 = (*t14);
 t16 = io_Writer_writeByte__5170(t12, UINT8_C(34));
 memcpy(&t15, &t16, sizeof(uint16_t));
 if (t15) {
  return t15;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, nav__6106_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_stringify_StringifyOptions__8122 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t7;
 uint8_t *t8;
 uint16_t t3;
 struct json_stringify_StringifyOptions__8122 t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6098(t2);
 if (t3) {
  return t3;
 }
 t4 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
 t5 = (*t4);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&a0->stream;
 t7 = (*t6);
 t3 = json_stringify_encodeJsonString__anon_8873__6139(a1, t5, t7);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t8 = (uint8_t *)&t2->next_punctuation;
 (*t8) = UINT8_C(3);
 return 0;
}

static uint16_t json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29_jsonStringif__6148(struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a1) {
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6086 t6;
 nav__6148_65 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6095 t12;
 nav__6148_56 *t13;
 nav__6148_56 t14;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 *t15;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 t16;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 t0;
 uint16_t t5;
 bool t11;
 t0 = a0;
 t1 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t2;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(t4);
 if (t5) {
  return t5;
 }
 t7 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const *)&t1->map;
 t8 = (*t7);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2965(t8);
 t6 = t9;
 zig_loop_22:
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__5624(&t6);
 t11 = t10.is_null != true;
 if (t11) {
  t12 = t10.payload;
  t4 = (*t3);
  t13 = t12.key_ptr;
  t14 = (*t13);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t4, t14);
  if (t5) {
   return t5;
  }
  t4 = (*t3);
  t15 = t12.value_ptr;
  t16 = (*t15);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6149(t4, t16);
  if (t5) {
   return t5;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_22;

 zig_block_0:;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(t4);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6126(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 const a1) {
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 const *t2;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 t3;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__849 const *)&t1;
 t3 = (*t2);
 t4 = json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29_jsonStringif__6148(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29_jsonStringify__anon_9047__6150(struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a1) {
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6139 t6;
 nav__6150_65 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6147 t12;
 nav__6150_56 *t13;
 nav__6150_56 t14;
 struct environment_linked_libraries_root_LibraryEntry__852 *t15;
 struct environment_linked_libraries_root_LibraryEntry__852 t16;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 t0;
 uint16_t t5;
 bool t11;
 t0 = a0;
 t1 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t2;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(t4);
 if (t5) {
  return t5;
 }
 t7 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t1->map;
 t8 = (*t7);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3355(t8);
 t6 = t9;
 zig_loop_22:
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__5628(&t6);
 t11 = t10.is_null != true;
 if (t11) {
  t12 = t10.payload;
  t4 = (*t3);
  t13 = t12.key_ptr;
  t14 = (*t13);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t4, t14);
  if (t5) {
   return t5;
  }
  t4 = (*t3);
  t15 = t12.value_ptr;
  t16 = (*t15);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6151(t4, t16);
  if (t5) {
   return t5;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_22;

 zig_block_0:;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(t4);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6127(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 const a1) {
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 const *t2;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 t3;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__854 const *)&t1;
 t3 = (*t2);
 t4 = json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29_jsonStringify__anon_9047__6150(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint8_t *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t11;
 void const **t13;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t14;
 void const *t15;
 nav__6093_55 (**t16)(void const *, nav__6093_57);
 struct io_Writer__3827 t17;
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t18;
 struct io_Writer__3827 const *t19;
 uint16_t t5;
 uint16_t t20;
 uint8_t t4;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6095(t2, UINT8_C(0));
 t3 = (uint8_t *)&a0->next_punctuation;
 t4 = (*t3);
 switch (t4) {
  case UINT8_C(1): {
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6096(t2);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  case UINT8_C(0):
  case UINT8_C(3): {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t2 = (*t1);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
 t7 = (*t6);
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t8;
 t10 = t9;
 t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t10;
 t13 = (void const **)&t12.context;
 t9 = (*t11);
 t14 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t9->context;
 t15 = (void const *)t14;
 (*t13) = t15;
 t16 = (nav__6093_55 (**)(void const *, nav__6093_57))&t12.writeFn;
 (*t16) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t17 = t12;
 t18 = t17;
 t19 = (struct io_Writer__3827 const *)&t18;
 t17 = (*t19);
 t5 = io_Writer_writeByte__5170(t17, UINT8_C(125));
 memcpy(&t20, &t5, sizeof(uint16_t));
 if (t20) {
  return t20;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(t2);
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3497(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *t1;
 struct mem_Allocator__599 const *t3;
 uintptr_t *t4;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t2;
 nav__3497_54 t10;
 nav__3497_52 t12;
 nav__3497_52 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t18;
 uint8_t **t16;
 uint8_t *t17;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t19;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t20;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t22;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t21;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t23;
 nav__3497_62 t24;
 nav__3497_62 t25;
 uint32_t *t26;
 nav__3497_67 t27;
 nav__3497_67 t28;
 nav__3497_65 *t29;
 nav__3497_72 t30;
 nav__3497_72 t31;
 struct environment_linked_libraries_root_LibraryEntry__852 *t32;
 uint16_t t11;
 bool t8;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t6 = a2;
 t7 = t5;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 t9 = (*t3);
 t5 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3504(a2);
 t10 = mem_Allocator_alignedAlloc__anon_7601__5959(t9, t5);
 if (t10.error) {
  t11 = t10.error;
  return t11;
 }
 t12 = t10.payload;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = (*t3);
  t13 = (*t1);
  t14 = (*t13);
  t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3505(t14);
  mem_Allocator_free__anon_6588__5736(t9, t15);
  t13 = (*t1);
  t16 = (uint8_t **)&t13->bytes;
  t17 = t12.ptr;
  (*t16) = t17;
  t13 = (*t1);
  t4 = (uintptr_t *)&t13->capacity;
  (*t4) = a2;
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t17 = t12.ptr;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t14.bytes = t17;
 t14.len = t5;
 t14.capacity = a2;
 t18 = t14;
 t13 = (*t1);
 t14 = (*t13);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t14);
 t20 = t19;
 t21 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t20;
 t14 = t18;
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t14);
 t22 = t19;
 t23 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t22;
 t19 = (*t23);
 t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t19);
 t19 = (*t21);
 t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t19);
 t26 = t25.ptr;
 if (t24.len != 0) memcpy(t24.ptr, t26, t24.len * sizeof(uint32_t));
 t19 = (*t23);
 t27 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t19);
 t19 = (*t21);
 t28 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t19);
 t29 = t28.ptr;
 if (t27.len != 0) memcpy(t27.ptr, t29, t27.len * sizeof(nav__3497_65));
 t19 = (*t23);
 t30 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t19);
 t19 = (*t21);
 t31 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t19);
 t32 = t31.ptr;
 if (t30.len != 0) memcpy(t30.ptr, t32, t30.len * sizeof(struct environment_linked_libraries_root_LibraryEntry__852));
 t9 = (*t3);
 t13 = (*t1);
 t14 = (*t13);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3505(t14);
 mem_Allocator_free__anon_6588__5736(t9, t12);
 t14 = t18;
 (*a0) = t14;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3493(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t0;
 uint16_t t8;
 uint16_t t9;
 bool t6;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a2;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3495(t3, a2);
 t8 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3497(t7, a1, t3);
 memcpy(&t9, &t8, sizeof(uint16_t));
 return t9;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3459(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct array_hash_map_IndexHeader__880 t5;
 uint8_t t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = (*t4);
 t6 = array_hash_map_IndexHeader_capacityIndexType__3138(t5);
 switch (t6) {
  case UINT8_C(0): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6152(t7, a1);
   return;
  }
  case UINT8_C(1): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6153(t7, a1);
   return;
  }
  case UINT8_C(2): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6154(t7, a1);
   return;
  }
  default: zig_unreachable();
 }
}

static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(nav__6130_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_StringContext_hash__2938(a0);
 return t0;
}

static nav__6131_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 uint32_t *t5;
 uint32_t *t6;
 uint32_t *const *t7;
 nav__6131_39 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6131_39){(uint32_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)0ul];
 t5 = (uint32_t *)t4;
 t6 = t5;
 t7 = (uint32_t *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (uint32_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint32_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(nav__6132_40 const a0, nav__6132_40 const a1, uintptr_t const a2) {
 bool t0;
 t0 = array_hash_map_StringContext_eql__2939(a0, a1, a2);
 return t0;
}

static nav__6133_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6133(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6133_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6133_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 nav__6133_38 t8;
 bool t9;
 struct array_hash_map_Index_28u8_29__6669 t11;
 uint8_t t12;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6155(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6133_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__6134_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6134(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6134_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6134_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 nav__6134_38 t8;
 struct array_hash_map_Index_28u16_29__6673 t11;
 uint16_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6156(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6134_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__6135_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6135(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6135_41 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_IndexHeader__880 *t4;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6135_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 nav__6135_38 t8;
 struct array_hash_map_Index_28u32_29__6677 t11;
 uint32_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6157(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6135_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t0;
 bool t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *const *)&t0;
 t2 = (uintptr_t *)&a0->len;
 t3 = (*t2);
 t2 = (uintptr_t *)&a0->capacity;
 t4 = (*t2);
 t5 = t3;
 t6 = t4;
 t7 = t5 < t6;
 debug_assert__180(t7);
 t2 = (uintptr_t *)&a0->len;
 t4 = (*t2);
 t8 = (*t1);
 t2 = (uintptr_t *)&t8->len;
 t3 = (*t2);
 t3 = t3 + (uintptr_t)1ul;
 (*t2) = t3;
 return t4;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6136(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, nav__6136_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t9;
 nav__6136_66 t10;
 nav__6136_66 t11;
 nav__6136_66 const *t12;
 nav__6136_69 t13;
 nav__6136_69 t14;
 nav__6136_69 const *t15;
 nav__6136_72 t16;
 nav__6136_72 t17;
 nav__6136_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6136_77 t20;
 nav__6136_77 t55;
 nav__6136_77 t21;
 nav__6136_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u8_29__6669 *t35;
 uint8_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__6136_42 **t42;
 nav__6136_42 *const *t43;
 nav__6136_42 *t44;
 struct environment_linked_libraries_root_LibraryEntry__852 **t45;
 struct environment_linked_libraries_root_LibraryEntry__852 *const *t46;
 struct environment_linked_libraries_root_LibraryEntry__852 *t47;
 uintptr_t *t48;
 nav__6136_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__880 t32;
 bool t30;
 bool t50;
 struct array_hash_map_Index_28u8_29__6669 t34;
 struct array_hash_map_Index_28u8_29__6669 t33;
 uint8_t t37;
 uint8_t t53;
 uint8_t t27;
 uint8_t t56;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t7);
 t11 = t10;
 t12 = (nav__6136_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t7);
 t14 = t13;
 t15 = (nav__6136_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t7);
 t17 = t16;
 t18 = (nav__6136_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t19);
 t21 = t20;
 t22 = (nav__6136_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT8_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u8_29_isEmpty__5743(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint8_t *)&t35->entry_index;
   t37 = (uint8_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__6136_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__6136_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__6136_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint8_t *)&t35->entry_index;
   t53 = (uint8_t)t24;
   (*t36) = t53;
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT8_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u8_29_isEmpty__5743(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint8_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__6136_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__6136_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint8_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT8_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT8_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6137(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, nav__6137_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t9;
 nav__6137_66 t10;
 nav__6137_66 t11;
 nav__6137_66 const *t12;
 nav__6137_69 t13;
 nav__6137_69 t14;
 nav__6137_69 const *t15;
 nav__6137_72 t16;
 nav__6137_72 t17;
 nav__6137_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6137_77 t20;
 nav__6137_77 t55;
 nav__6137_77 t21;
 nav__6137_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u16_29__6673 *t35;
 uint16_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__6137_42 **t42;
 nav__6137_42 *const *t43;
 nav__6137_42 *t44;
 struct environment_linked_libraries_root_LibraryEntry__852 **t45;
 struct environment_linked_libraries_root_LibraryEntry__852 *const *t46;
 struct environment_linked_libraries_root_LibraryEntry__852 *t47;
 uintptr_t *t48;
 nav__6137_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__880 t32;
 struct array_hash_map_Index_28u16_29__6673 t34;
 struct array_hash_map_Index_28u16_29__6673 t33;
 uint16_t t37;
 uint16_t t53;
 uint16_t t27;
 uint16_t t56;
 bool t30;
 bool t50;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t7);
 t11 = t10;
 t12 = (nav__6137_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t7);
 t14 = t13;
 t15 = (nav__6137_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t7);
 t17 = t16;
 t18 = (nav__6137_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t19);
 t21 = t20;
 t22 = (nav__6137_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT16_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u16_29_isEmpty__5748(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint16_t *)&t35->entry_index;
   t37 = (uint16_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__6137_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__6137_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__6137_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint16_t *)&t35->entry_index;
   t53 = (uint16_t)t24;
   (*t36) = t53;
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT16_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u16_29_isEmpty__5748(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint16_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__6137_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__6137_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint16_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT16_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT16_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6138(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, nav__6138_42 const a1, struct array_hash_map_IndexHeader__880 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t9;
 nav__6138_66 t10;
 nav__6138_66 t11;
 nav__6138_66 const *t12;
 nav__6138_69 t13;
 nav__6138_69 t14;
 nav__6138_69 const *t15;
 nav__6138_72 t16;
 nav__6138_72 t17;
 nav__6138_72 const *t18;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6138_77 t20;
 nav__6138_77 t53;
 nav__6138_77 t21;
 nav__6138_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t52;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u32_29__6677 *t35;
 uint32_t *t36;
 uint32_t *const *t38;
 bool *t40;
 nav__6138_42 **t41;
 nav__6138_42 *const *t42;
 nav__6138_42 *t43;
 struct environment_linked_libraries_root_LibraryEntry__852 **t44;
 struct environment_linked_libraries_root_LibraryEntry__852 *const *t45;
 struct environment_linked_libraries_root_LibraryEntry__852 *t46;
 uintptr_t *t47;
 nav__6138_42 t50;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t39;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8062 t55;
 uint32_t t23;
 uint32_t t37;
 uint32_t t48;
 uint32_t t27;
 uint32_t t54;
 struct array_hash_map_IndexHeader__880 t32;
 struct array_hash_map_Index_28u32_29__6677 t34;
 struct array_hash_map_Index_28u32_29__6677 t33;
 bool t30;
 bool t49;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t7);
 t11 = t10;
 t12 = (nav__6138_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t7);
 t14 = t13;
 t15 = (nav__6138_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5663(t7);
 t17 = t16;
 t18 = (nav__6138_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t19);
 t21 = t20;
 t22 = (nav__6138_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t24 = array_hash_map_safeTruncate__anon_7812__5971(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT32_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u32_29_isEmpty__5753(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint32_t *)&t35->entry_index;
   t37 = (uint32_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t25];
   (*t36) = t23;
   t40 = (bool *)&t39.found_existing;
   (*t40) = false;
   t41 = (nav__6138_42 **)&t39.key_ptr;
   t42 = &t15->ptr;
   t43 = (*t42);
   t43 = (nav__6138_42 *)&t43[t25];
   (*t41) = t43;
   t44 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t39.value_ptr;
   t45 = &t18->ptr;
   t46 = (*t45);
   t46 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t46[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t39.index;
   (*t47) = t25;
   return t39;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t48 = t10.ptr[t24];
  t30 = t23 == t48;
  if (t30) {
   t24 = (uintptr_t)t37;
   t50 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t50, t24);
   t49 = t30;
   goto zig_block_4;
  }
  t49 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t49) {
   t40 = (bool *)&t51.found_existing;
   (*t40) = true;
   t41 = (nav__6138_42 **)&t51.key_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t13 = (*t15);
   t43 = &t13.ptr[t25];
   (*t41) = t43;
   t44 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t51.value_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t16 = (*t18);
   t46 = &t16.ptr[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t51.index;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   (*t47) = t25;
   return t51;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t48 = t27;
  t30 = t37 < t48;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3482(t5);
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t24];
   (*t36) = t23;
   t52 = t31;
   t53 = (*t22);
   t35 = &t53.ptr[t52];
   t36 = (uint32_t *)&t35->entry_index;
   t23 = (uint32_t)t24;
   (*t36) = t23;
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t23 = t27;
   (*t36) = t23;
   t34 = t33;
   t23 = t34.distance_from_start_index;
   t27 = t23;
   t34 = t33;
   t23 = t34.entry_index;
   t54 = t23;
   t52 = t26;
   t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
   t26 = t52;
   t23 = t27;
   t23 = t23 + UINT32_C(1);
   t27 = t23;
   zig_loop_200:
   t52 = t26;
   t29 = t52;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t52 = t26;
    t52 = array_hash_map_IndexHeader_constrainIndex__3136(t32, t52);
    t31 = t52;
    t52 = t31;
    t34 = t20.ptr[t52];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u32_29_isEmpty__5753(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint32_t *)&t35->entry_index;
     t23 = t54;
     (*t36) = t23;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t23 = t27;
     (*t36) = t23;
     t40 = (bool *)&t55.found_existing;
     (*t40) = false;
     t41 = (nav__6138_42 **)&t55.key_ptr;
     t42 = &t15->ptr;
     t43 = (*t42);
     t43 = (nav__6138_42 *)&t43[t24];
     (*t41) = t43;
     t44 = (struct environment_linked_libraries_root_LibraryEntry__852 **)&t55.value_ptr;
     t45 = &t18->ptr;
     t46 = (*t45);
     t46 = (struct environment_linked_libraries_root_LibraryEntry__852 *)&t46[t24];
     (*t44) = t46;
     t47 = (uintptr_t *)&t55.index;
     (*t47) = t24;
     return t55;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t23 = t34.distance_from_start_index;
    t48 = t27;
    t30 = t23 < t48;
    if (t30) {
     t52 = t31;
     t53 = (*t22);
     t35 = &t53.ptr[t52];
     t36 = (uint32_t *)&t35->entry_index;
     t48 = t54;
     (*t36) = t48;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t48 = t27;
     (*t36) = t48;
     t34 = t33;
     t48 = t34.entry_index;
     t54 = t48;
     t34 = t33;
     t48 = t34.distance_from_start_index;
     t27 = t48;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t52 = t26;
    t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
    t26 = t52;
    t23 = t27;
    t23 = t23 + UINT32_C(1);
    t27 = t23;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t52 = t26;
  t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
  t26 = t52;
  t37 = t27;
  t37 = t37 + UINT32_C(1);
  t27 = t37;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static nav__6101_38 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6101(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *t0;
 t0 = a0;
 return (nav__6101_38){true,false};
}

static bool json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6102(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 const *const a0) {
 uintptr_t const *t0;
 uintptr_t t1;
 uint64_t t2;
 uint8_t const *t5;
 bool t3;
 bool t4;
 uint8_t t6;
 t0 = (uintptr_t const *)&a0->indent_level;
 t1 = (*t0);
 t2 = t1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  t5 = (uint8_t const *)&a0->next_punctuation;
  t6 = (*t5);
  t3 = t6 == UINT8_C(2);
  t4 = t3;
  goto zig_block_0;
 }
 t4 = false;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static uint16_t io_Writer_writeByte__5170(struct io_Writer__3827 const a0, uint8_t const a1) {
 struct io_Writer__3827 const *t1;
 uint8_t const (*t4)[1];
 struct io_Writer__3827 t5;
 struct io_Writer__3827 t0;
 nav__5170_43 t6;
 uint16_t t7;
 uint8_t t2[1];
 uint8_t t3[1];
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t2[0] = a1;
 memcpy((char *)&t3, t2, sizeof(uint8_t[1]));
 t4 = (uint8_t const (*)[1])&t3;
 t5 = (*t1);
 t6.ptr = &(*t4)[(uintptr_t)0ul];
 t6.len = (uintptr_t)1ul;
 t7 = io_Writer_writeAll__5168(t5, t6);
 return t7;
}

static uint16_t io_Writer_writeByteNTimes__5171(struct io_Writer__3827 const a0, uint8_t const a1, uintptr_t const a2) {
 struct io_Writer__3827 const *t1;
 uint8_t *t3;
 uint8_t (*t4)[256];
 uintptr_t t6;
 uintptr_t t14;
 uintptr_t t5;
 uint64_t t7;
 struct io_Writer__3827 t10;
 struct io_Writer__3827 t0;
 nav__5171_52 t11;
 nav__5171_43 t12;
 uint16_t t9;
 uint16_t t13;
 bool t8;
 uint8_t t2[256];
 t0 = a0;
 t1 = (struct io_Writer__3827 const *)&t0;
 t3 = (uint8_t *)&t2;
 t3 = (uint8_t *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4 = (uint8_t (*)[256])t3;
 memset(t4, a1, 256);
 t5 = a2;
 zig_loop_15:
 t6 = t5;
 t7 = t6;
 t8 = t7 > UINT64_C(0);
 if (t8) {
  t6 = t5;
  t6 = ((uintptr_t)256ul < t6) ? (uintptr_t)256ul : t6;
  t9 = (uint16_t)t6;
  t10 = (*t1);
  t3 = (uint8_t *)&t2;
  t3 = (uint8_t *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t6 = (uintptr_t)t9;
  t11.ptr = t3;
  t11.len = t6;
  memcpy(&t12, &t11, sizeof(nav__5171_43));
  t13 = io_Writer_writeAll__5168(t10, t12);
  if (t13) {
   return t13;
  }
  t6 = t5;
  t14 = (uintptr_t)t9;
  t14 = t6 - t14;
  t5 = t14;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_15;

 zig_block_0:;
 return 0;
}

static uint16_t json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_jsonStringify__anon_9121__6158(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a1) {
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6107 t6;
 nav__6158_65 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6115 t12;
 nav__6158_56 *t13;
 nav__6158_56 t14;
 struct environment_root_EnvironmentValue__844 *t15;
 struct environment_root_EnvironmentValue__844 t16;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 t0;
 uint16_t t5;
 bool t11;
 t0 = a0;
 t1 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t2;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(t4);
 if (t5) {
  return t5;
 }
 t7 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const *)&t1->map;
 t8 = (*t7);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3176(t8);
 t6 = t9;
 zig_loop_22:
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__5626(&t6);
 t11 = t10.is_null != true;
 if (t11) {
  t12 = t10.payload;
  t4 = (*t3);
  t13 = t12.key_ptr;
  t14 = (*t13);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t4, t14);
  if (t5) {
   return t5;
  }
  t4 = (*t3);
  t15 = t12.value_ptr;
  t16 = (*t15);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6159(t4, t16);
  if (t5) {
   return t5;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_22;

 zig_block_0:;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(t4);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6149(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 const a1) {
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 const *t2;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 t3;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__846 const *)&t1;
 t3 = (*t2);
 t4 = json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_jsonStringify__anon_9121__6158(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6151(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, struct environment_linked_libraries_root_LibraryEntry__852 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 nav__6151_43 t7;
 struct json_stringify_StringifyOptions__8122 *t9;
 struct environment_linked_libraries_root_LibraryEntry__852 t2;
 uint16_t t4;
 bool t6;
 bool t5;
 bool t8;
 struct json_stringify_StringifyOptions__8122 t10;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6091(t3);
 if (t4) {
  return t4;
 }
 t5 = true;
 t6 = t5;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t3, (nav__6151_43){((uint8_t const *)&__anon_9129),(uintptr_t)4ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t7 = a1.path;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6160(t3, t7);
  if (t4) {
   return t4;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = true;
 t9 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
 t10 = (*t9);
 t6 = t10.emit_null_optional_fields;
 t6 = !t6;
 if (t6) {
  t7 = a1.build_id;
  t6 = t7.ptr == NULL;
  if (t6) {
   t8 = false;
   goto zig_block_2;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t6 = t8;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6106(t3, (nav__6151_43){((uint8_t const *)&__anon_9133),(uintptr_t)8ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t7 = a1.build_id;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6161(t3, t7);
  if (t4) {
   return t4;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6093(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6095(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, uint8_t const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uintptr_t *t3;
 uintptr_t t4;
 (void)a1;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t *)&t2->indent_level;
 t4 = (*t3);
 t4 = t4 - (uintptr_t)1ul;
 (*t3) = t4;
 return;
}

static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint8_t *t3;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = (uint8_t *)&t2->next_punctuation;
 (*t3) = UINT8_C(2);
 return;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3495(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)4ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6152(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t9;
 nav__6152_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6152_60 t12;
 nav__6152_60 t33;
 nav__6152_60 t13;
 nav__6152_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u8_29__6669 const *t31;
 struct array_hash_map_Index_28u8_29__6669 *t34;
 uint8_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__880 t28;
 bool t20;
 uint8_t t26;
 uint8_t t36;
 uint8_t t25;
 uint8_t t27;
 struct array_hash_map_Index_28u8_29__6669 t29;
 struct array_hash_map_Index_28u8_29__6669 t32;
 struct array_hash_map_Index_28u8_29__6669 t30;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7788__5965(t11);
 t13 = t12;
 t14 = (nav__6152_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint8_t)t17;
  t25 = t26;
  t27 = UINT8_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u8_29__6669 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u8_29_isEmpty__5743(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint8_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint8_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT8_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6153(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t9;
 nav__6153_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6153_60 t12;
 nav__6153_60 t33;
 nav__6153_60 t13;
 nav__6153_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u16_29__6673 const *t31;
 struct array_hash_map_Index_28u16_29__6673 *t34;
 uint16_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__880 t28;
 uint16_t t26;
 uint16_t t36;
 uint16_t t25;
 uint16_t t27;
 struct array_hash_map_Index_28u16_29__6673 t29;
 struct array_hash_map_Index_28u16_29__6673 t32;
 struct array_hash_map_Index_28u16_29__6673 t30;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7798__5967(t11);
 t13 = t12;
 t14 = (nav__6153_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint16_t)t17;
  t25 = t26;
  t27 = UINT16_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u16_29__6673 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u16_29_isEmpty__5748(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint16_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint16_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT16_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6154(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const a0, struct array_hash_map_IndexHeader__880 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t9;
 nav__6154_56 t10;
 struct array_hash_map_IndexHeader__880 *t11;
 struct array_hash_map_IndexHeader__880 *t2;
 nav__6154_60 t12;
 nav__6154_60 t32;
 nav__6154_60 t13;
 nav__6154_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u32_29__6677 const *t30;
 struct array_hash_map_Index_28u32_29__6677 *t33;
 uint32_t *t34;
 uint32_t t21;
 uint32_t t35;
 uint32_t t25;
 uint32_t t26;
 struct array_hash_map_IndexHeader__880 t27;
 struct array_hash_map_Index_28u32_29__6677 t28;
 struct array_hash_map_Index_28u32_29__6677 t31;
 struct array_hash_map_Index_28u32_29__6677 t29;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7804__5969(t11);
 t13 = t12;
 t14 = (nav__6154_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7812__5971(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t21 = (uint32_t)t17;
  t25 = t21;
  t26 = UINT32_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t27 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3136(t27, t22);
   t28 = t12.ptr[t22];
   t29 = t28;
   t30 = (struct array_hash_map_Index_28u32_29__6677 const *)&t29;
   t31 = (*t30);
   t20 = array_hash_map_Index_28u32_29_isEmpty__5753(t31);
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t21 = t26;
    (*t34) = t21;
    t34 = (uint32_t *)&t33->entry_index;
    t21 = t25;
    (*t34) = t21;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t21 = t28.distance_from_start_index;
   t35 = t26;
   t20 = t21 < t35;
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t35 = t26;
    (*t34) = t35;
    t34 = (uint32_t *)&t33->entry_index;
    t35 = t25;
    (*t34) = t35;
    t35 = t28.distance_from_start_index;
    t26 = t35;
    t35 = t28.entry_index;
    t25 = t35;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t35 = t26;
   t35 = t35 + UINT32_C(1);
   t26 = t35;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static nav__6155_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6155(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6155_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__6155_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t8;
 nav__6155_66 t9;
 nav__6155_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u8_29__6669 const *t23;
 nav__6155_41 t30;
 nav__6155_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__880 t20;
 bool t18;
 bool t25;
 struct array_hash_map_Index_28u8_29__6669 t21;
 struct array_hash_map_Index_28u8_29__6669 t24;
 struct array_hash_map_Index_28u8_29__6669 t22;
 uint8_t t26;
 uint8_t t27;
 uint8_t t15;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT8_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u8_29__6669 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u8_29_isEmpty__5743(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6155_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT8_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6156_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6156(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6156_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__6156_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t8;
 nav__6156_66 t9;
 nav__6156_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u16_29__6673 const *t23;
 nav__6156_41 t30;
 nav__6156_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__880 t20;
 struct array_hash_map_Index_28u16_29__6673 t21;
 struct array_hash_map_Index_28u16_29__6673 t24;
 struct array_hash_map_Index_28u16_29__6673 t22;
 uint16_t t26;
 uint16_t t27;
 uint16_t t15;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT16_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u16_29__6673 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u16_29_isEmpty__5748(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6156_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT16_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6157_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6157(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const a0, nav__6157_41 const a1, struct array_hash_map_IndexHeader__880 *const a2, nav__6157_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *t1;
 struct array_hash_map_IndexHeader__880 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *t8;
 nav__6157_66 t9;
 nav__6157_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__880 *t19;
 struct array_hash_map_IndexHeader__880 *t2;
 struct array_hash_map_Index_28u32_29__6677 const *t23;
 nav__6157_41 t29;
 nav__6157_38 t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 t0;
 uint32_t t11;
 uint32_t t26;
 uint32_t t27;
 uint32_t t15;
 struct array_hash_map_IndexHeader__880 t20;
 struct array_hash_map_Index_28u32_29__6677 t21;
 struct array_hash_map_Index_28u32_29__6677 t24;
 struct array_hash_map_Index_28u32_29__6677 t22;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__880 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__927 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__3475(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6393 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6131(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__5662(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6130(a1);
 t12 = array_hash_map_safeTruncate__anon_7812__5971(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT32_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3136(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u32_29__6677 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u32_29_isEmpty__5753(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6157_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t26 = t9.ptr[t28];
  t25 = t11 == t26;
  if (t25) {
   t28 = (uintptr_t)t27;
   t29 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6132(a1, t29, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t30.is_null = false;
   t30.payload = t12;
   return t30;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT32_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static uint16_t environment_root_EnvironmentValue_jsonStringify__anon_9194__6162(struct environment_root_EnvironmentValue__844 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 nav__6162_43 t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 nav__6162_46 t6;
 uint16_t t5;
 uint8_t t2;
 t0 = a1;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.string;
   t4 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6160(t4, t3);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t6 = a0.payload.list;
   t4 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6163(t4, t6);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6159(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, struct environment_root_EnvironmentValue__844 const a1) {
 struct environment_root_EnvironmentValue__844 const *t2;
 struct environment_root_EnvironmentValue__844 t3;
 struct environment_root_EnvironmentValue__844 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct environment_root_EnvironmentValue__844 const *)&t1;
 t3 = (*t2);
 t4 = environment_root_EnvironmentValue_jsonStringify__anon_9194__6162(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6111(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, nav__6111_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_stringify_StringifyOptions__8122 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t7;
 uint16_t t3;
 struct json_stringify_StringifyOptions__8122 t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(t2);
 if (t3) {
  return t3;
 }
 t4 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
 t5 = (*t4);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&a0->stream;
 t7 = (*t6);
 t3 = json_stringify_encodeJsonString__anon_8873__6139(a1, t5, t7);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(t2);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6160(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, nav__6160_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_StringifyOptions__8122 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t7;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uintptr_t t11;
 uintptr_t t12;
 uintptr_t t10;
 uint64_t t13;
 uint64_t t14;
 nav__6160_41 t2;
 uint16_t t8;
 uint16_t t9;
 struct json_stringify_StringifyOptions__8122 t4;
 bool t5;
 bool t6;
 uint8_t t15;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
 t4 = (*t3);
 t5 = t4.emit_strings_as_arrays;
 t5 = !t5;
 if (t5) {
  t5 = unicode_utf8ValidateSlice__5206(a1);
  t6 = t5;
  goto zig_block_1;
 }
 t6 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t6) {
  t7 = (*t1);
  t8 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6111(t7, a1);
  memcpy(&t9, &t8, sizeof(uint16_t));
  return t9;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t9 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6090(t7);
 if (t9) {
  return t9;
 }
 t10 = (uintptr_t)0ul;
 t11 = a1.len;
 zig_loop_44:
 t12 = t10;
 t13 = t12;
 t14 = t11;
 t6 = t13 < t14;
 if (t6) {
  t15 = a1.ptr[t12];
  t7 = (*t1);
  t9 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6164(t7, t15);
  if (t9) {
   return t9;
  }
  goto zig_block_3;
 }
 goto zig_block_2;

 zig_block_3:;
 t12 = t12 + (uintptr_t)1ul;
 t10 = t12;
 goto zig_loop_44;

 zig_block_2:;
 t7 = (*t1);
 t9 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6092(t7);
 if (t9) {
  return t9;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6161(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, nav__6161_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 nav__6161_41 t4;
 nav__6161_41 t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t5;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint16_t t6;
 bool t3;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a1;
 t3 = a1.ptr != NULL;
 if (t3) {
  t4 = a1;
  t5 = (*t1);
  t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6160(t5, t4);
  if (t6) {
   return t6;
  }
  return 0;
 }
 t5 = (*t1);
 t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6165(t5);
 if (t6) {
  return t6;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6163(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, nav__6163_43 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uintptr_t t6;
 uintptr_t t7;
 uintptr_t t5;
 uint64_t t8;
 uint64_t t9;
 nav__6163_41 t11;
 nav__6163_43 t2;
 uint16_t t4;
 bool t10;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6090(t3);
 if (t4) {
  return t4;
 }
 t5 = (uintptr_t)0ul;
 t6 = a1.len;
 zig_loop_28:
 t7 = t5;
 t8 = t7;
 t9 = t6;
 t10 = t8 < t9;
 if (t10) {
  t11 = a1.ptr[t7];
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6160(t3, t11);
  if (t4) {
   return t4;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t7 = t7 + (uintptr_t)1ul;
 t5 = t7;
 goto zig_loop_28;

 zig_block_0:;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6092(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static bool unicode_utf8ValidateSlice__5206(nav__5206_39 const a0) {
 bool t0;
 t0 = unicode_utf8ValidateSliceImpl__anon_9218__6166(a0);
 return t0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6090(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t9;
 void const **t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t12;
 void const *t13;
 nav__6090_54 (**t14)(void const *, nav__6090_56);
 struct io_Writer__3827 t15;
 struct io_Writer__3827 t10;
 struct io_Writer__3827 t16;
 struct io_Writer__3827 const *t17;
 uint8_t *t19;
 uint16_t t3;
 uint16_t t18;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(t2);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
 t5 = (*t4);
 t6 = t5;
 t7 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t6;
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t8;
 t11 = (void const **)&t10.context;
 t7 = (*t9);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t7->context;
 t13 = (void const *)t12;
 (*t11) = t13;
 t14 = (nav__6090_54 (**)(void const *, nav__6090_56))&t10.writeFn;
 (*t14) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t15 = t10;
 t16 = t15;
 t17 = (struct io_Writer__3827 const *)&t16;
 t15 = (*t17);
 t3 = io_Writer_writeByte__5170(t15, UINT8_C(91));
 memcpy(&t18, &t3, sizeof(uint16_t));
 if (t18) {
  return t18;
 }
 t2 = (*t1);
 (void)json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6094(t2, UINT8_C(1));
 t2 = (*t1);
 t19 = (uint8_t *)&t2->next_punctuation;
 (*t19) = UINT8_C(1);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6164(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0, uint8_t const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct json_stringify_StringifyOptions__8122 *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t13;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t14;
 void const **t16;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t17;
 void const *t18;
 nav__6164_57 (**t19)(void const *, nav__6164_59);
 struct io_Writer__3827 t20;
 struct io_Writer__3827 t15;
 struct io_Writer__3827 t21;
 struct io_Writer__3827 const *t22;
 uint16_t t4;
 uint16_t t23;
 struct json_stringify_StringifyOptions__8122 t6;
 bool t7;
 nav__6164_51 t10;
 uint8_t t2;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(t3);
 if (t4) {
  return t4;
 }
 t5 = (struct json_stringify_StringifyOptions__8122 *)&a0->options;
 t6 = (*t5);
 t7 = t6.emit_nonportable_numbers_as_strings;
 if (t7) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t1);
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t3->stream;
 t9 = (*t8);
 t10.f0 = a1;
 t11 = t9;
 t12 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t11;
 t13 = t12;
 t14 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t13;
 t16 = (void const **)&t15.context;
 t12 = (*t14);
 t17 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t12->context;
 t18 = (void const *)t17;
 (*t16) = t18;
 t19 = (nav__6164_57 (**)(void const *, nav__6164_59))&t15.writeFn;
 (*t19) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t20 = t15;
 t21 = t20;
 t22 = (struct io_Writer__3827 const *)&t21;
 t20 = (*t22);
 t4 = io_Writer_print__anon_9232__6167(t20, t10);
 memcpy(&t23, &t4, sizeof(uint16_t));
 if (t23) {
  return t23;
 }
 t3 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(t3);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6092(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 uint8_t *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t11;
 void const **t13;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t14;
 void const *t15;
 nav__6092_55 (**t16)(void const *, nav__6092_57);
 struct io_Writer__3827 t17;
 struct io_Writer__3827 t12;
 struct io_Writer__3827 t18;
 struct io_Writer__3827 const *t19;
 uint16_t t5;
 uint16_t t20;
 uint8_t t4;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6095(t2, UINT8_C(1));
 t3 = (uint8_t *)&a0->next_punctuation;
 t4 = (*t3);
 switch (t4) {
  case UINT8_C(1): {
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6096(t2);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  case UINT8_C(0):
  case UINT8_C(3): {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t2 = (*t1);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
 t7 = (*t6);
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t8;
 t10 = t9;
 t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t10;
 t13 = (void const **)&t12.context;
 t9 = (*t11);
 t14 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t9->context;
 t15 = (void const *)t14;
 (*t13) = t15;
 t16 = (nav__6092_55 (**)(void const *, nav__6092_57))&t12.writeFn;
 (*t16) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t17 = t12;
 t18 = t17;
 t19 = (struct io_Writer__3827 const *)&t18;
 t17 = (*t19);
 t5 = io_Writer_writeByte__5170(t17, UINT8_C(93));
 memcpy(&t20, &t5, sizeof(uint16_t));
 if (t20) {
  return t20;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(t2);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6165(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *t9;
 void const **t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *t12;
 void const *t13;
 nav__6165_54 (**t14)(void const *, nav__6165_56);
 struct io_Writer__3827 t15;
 struct io_Writer__3827 t10;
 struct io_Writer__3827 t16;
 struct io_Writer__3827 const *t17;
 uint16_t t3;
 uint16_t t18;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8244 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6097(t2);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 *)&t2->stream;
 t5 = (*t4);
 t6 = t5;
 t7 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *)&t6;
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1222 const *const *)&t8;
 t11 = (void const **)&t10.context;
 t7 = (*t9);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__806 *const *)&t7->context;
 t13 = (void const *)t12;
 (*t11) = t13;
 t14 = (nav__6165_54 (**)(void const *, nav__6165_56))&t10.writeFn;
 (*t14) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___3955;
 t15 = t10;
 t16 = t15;
 t17 = (struct io_Writer__3827 const *)&t16;
 t15 = (*t17);
 t3 = io_Writer_writeAll__5168(t15, (nav__6165_56){(uint8_t const *)&__anon_9236,(uintptr_t)4ul});
 memcpy(&t18, &t3, sizeof(uint16_t));
 if (t18) {
  return t18;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6100(t2);
 return 0;
}

static bool unicode_utf8ValidateSliceImpl__anon_9218__6166(nav__6166_39 const a0) {
 static uint8_t const t9[16] = "\200\200\200\200\200\200\200\200\200\200\200\200\200\200\200\200";
 static uint8_t const t15[256] = "\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\023\003\003\003\003\003\003\003\003\003\003\003\003#\003\0034\004\004\004D\361\361\361\361\361\361\361\361\361\361\361";
 nav__6166_39 t1;
 nav__6166_39 t0;
 uintptr_t t2;
 uintptr_t t12;
 uintptr_t t17;
 uintptr_t t11;
 uint64_t t3;
 uint64_t t13;
 uint8_t const *t5;
 uint8_t const (*t6)[16];
 bool t4;
 bool t21;
 uint8_t t7[16];
 uint8_t t8[16];
 bool t10[16];
 uint8_t t14;
 uint8_t t16;
 uint8_t t20;
 uint8_t t18;
 uint8_t t19;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t1.len;
 t3 = t2;
 t4 = t3 >= UINT64_C(16);
 if (t4) {
  t1 = t0;
  t5 = t1.ptr;
  t5 = (uint8_t const *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t6 = (uint8_t const (*)[16])t5;
  memcpy(t7, (const char *)t6, sizeof(uint8_t[16]));
  memcpy(&t8, &t7, sizeof(uint8_t[16]));
  for (t2 = (uintptr_t)0ul; t2 < (uintptr_t)16ul; t2 += (uintptr_t)1ul) {
   t8[t2] = t8[t2] & t9[t2];
  }
  for (t2 = (uintptr_t)0ul; t2 < (uintptr_t)16ul; t2 += (uintptr_t)1ul) {
   t10[t2] = t8[t2] == t9[t2];
  }
  t4 = false;
  for (t2 = (uintptr_t)0ul; t2 < (uintptr_t)16ul; t2 += (uintptr_t)1ul) {
   t4 |= t10[t2];
  }
  if (t4) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t1 = t0;
  t5 = t1.ptr;
  t5 = (uint8_t const *)(((uintptr_t)t5) + ((uintptr_t)16ul*sizeof(uint8_t)));
  t2 = t1.len;
  t2 = t2 - (uintptr_t)16ul;
  t1.ptr = t5;
  t1.len = t2;
  t0 = t1;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_6;

 zig_block_0:;
 t1 = t0;
 t2 = t1.len;
 t11 = (uintptr_t)0ul;
 zig_loop_52:
 t12 = t11;
 t3 = t12;
 t13 = t2;
 t4 = t3 < t13;
 if (t4) {
  t1 = t0;
  t12 = t11;
  t14 = t1.ptr[t12];
  t4 = t14 < UINT8_C(128);
  if (t4) {
   t12 = t11;
   t12 = t12 + (uintptr_t)1ul;
   t11 = t12;
   goto zig_block_5;
  }
  goto zig_block_6;

  zig_block_6:;
  t12 = (uintptr_t)t14;
  t14 = t15[t12];
  t4 = t14 == UINT8_C(241);
  if (t4) {
   return false;
  }
  goto zig_block_7;

  zig_block_7:;
  t16 = t14 & UINT8_C(7);
  t12 = t11;
  t17 = (uintptr_t)t16;
  t17 = t12 + t17;
  t13 = t17;
  t3 = t2;
  t4 = t13 > t3;
  if (t4) {
   return false;
  }
  goto zig_block_8;

  zig_block_8:;
  t18 = UINT8_C(128);
  t19 = UINT8_C(191);
  t14 = zig_shr_u8(t14, UINT8_C(4));
  switch (t14) {
   case UINT8_C(0): {
    goto zig_block_9;
   }
   case UINT8_C(1): {
    t18 = UINT8_C(160);
    goto zig_block_9;
   }
   case UINT8_C(2): {
    t19 = UINT8_C(159);
    goto zig_block_9;
   }
   case UINT8_C(3): {
    t18 = UINT8_C(144);
    goto zig_block_9;
   }
   case UINT8_C(4): {
    t19 = UINT8_C(143);
    goto zig_block_9;
   }
   default: {
    zig_unreachable();
   }
  }

  zig_block_9:;
  t1 = t0;
  t17 = t11;
  t17 = t17 + (uintptr_t)1ul;
  t14 = t1.ptr[t17];
  t20 = t18;
  t4 = t14 < t20;
  if (t4) {
   t21 = true;
   goto zig_block_11;
  }
  t20 = t19;
  t4 = t20 < t14;
  t21 = t4;
  goto zig_block_11;

  zig_block_11:;
  if (t21) {
   return false;
  }
  goto zig_block_10;

  zig_block_10:;
  switch (t16) {
   case UINT8_C(2): {
    t17 = t11;
    t17 = t17 + (uintptr_t)2ul;
    t11 = t17;
    goto zig_block_12;
   }
   case UINT8_C(3): {
    t1 = t0;
    t17 = t11;
    t17 = t17 + (uintptr_t)2ul;
    t16 = t1.ptr[t17];
    t21 = t16 < UINT8_C(128);
    if (t21) {
     t4 = true;
     goto zig_block_14;
    }
    t21 = UINT8_C(191) < t16;
    t4 = t21;
    goto zig_block_14;

    zig_block_14:;
    if (t4) {
     return false;
    }
    goto zig_block_13;

    zig_block_13:;
    t17 = t11;
    t17 = t17 + (uintptr_t)3ul;
    t11 = t17;
    goto zig_block_12;
   }
   case UINT8_C(4): {
    t1 = t0;
    t17 = t11;
    t17 = t17 + (uintptr_t)2ul;
    t16 = t1.ptr[t17];
    t21 = t16 < UINT8_C(128);
    if (t21) {
     t4 = true;
     goto zig_block_16;
    }
    t21 = UINT8_C(191) < t16;
    t4 = t21;
    goto zig_block_16;

    zig_block_16:;
    if (t4) {
     return false;
    }
    goto zig_block_15;

    zig_block_15:;
    t1 = t0;
    t17 = t11;
    t17 = t17 + (uintptr_t)3ul;
    t16 = t1.ptr[t17];
    t21 = t16 < UINT8_C(128);
    if (t21) {
     t4 = true;
     goto zig_block_18;
    }
    t21 = UINT8_C(191) < t16;
    t4 = t21;
    goto zig_block_18;

    zig_block_18:;
    if (t4) {
     return false;
    }
    goto zig_block_17;

    zig_block_17:;
    t17 = t11;
    t17 = t17 + (uintptr_t)4ul;
    t11 = t17;
    goto zig_block_12;
   }
   default: {
    zig_unreachable();
   }
  }

  zig_block_12:;
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_4;
 }
 goto zig_block_3;

 zig_block_4:;
 goto zig_loop_52;

 zig_block_3:;
 return true;
}

static uint16_t io_Writer_print__anon_9232__6167(struct io_Writer__3827 const a0, nav__6167_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_9241__6168(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_format__anon_9241__6168(struct io_Writer__3827 const a0, nav__6168_39 const a1) {
 struct io_Writer__3827 t0;
 uint16_t t2;
 uint8_t t1;
 t0 = a0;
 t1 = a1.f0;
 t2 = fmt_formatType__anon_9259__6169(t1, (struct fmt_FormatOptions__4248){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t2) {
  return t2;
 }
 return 0;
}

static uint16_t fmt_formatType__anon_9259__6169(uint8_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2, uintptr_t const a3) {
 struct io_Writer__3827 t1;
 uint16_t t2;
 uint16_t t3;
 uint8_t t0;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_9264__6170(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatInt__anon_9272__6172(uint8_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4248 const a3, struct io_Writer__3827 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__6172_58 t10;
 nav__6172_47 t11;
 uint16_t t12;
 uint16_t t13;
 bool t0;
 uint8_t t4;
 uint8_t t2;
 nav__6172_56 t8;
 uint8_t t9[2];
 uint8_t t1[9];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)9ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t4 = t2;
  t0 = t4 >= UINT8_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT8_C(100);
   t8 = fmt_digits2__4953(t4);
   memcpy(t9, t8.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t9, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT8_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT8_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t4 = UINT8_C(48) + t4;
   (*t6) = t4;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = fmt_digits2__4953(t4);
  memcpy(t9, t8.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t9, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t4 = t2;
 t4 = t4 % a1;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t4 = fmt_digitToChar__4969(t4, a2);
 (*t6) = t4;
 t4 = t2;
 t4 = t4 / a1;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT8_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)9ul - t5;
 t10.ptr = t6;
 t10.len = t5;
 memcpy(&t11, &t10, sizeof(nav__6172_47));
 t12 = fmt_formatBuf__anon_5109__5592(t11, a3, a4);
 memcpy(&t13, &t12, sizeof(uint16_t));
 return t13;
}

static uint16_t fmt_formatIntValue__anon_9268__6171(uint8_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_9272__6172(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_9264__6170(uint8_t const a0, struct fmt_FormatOptions__4248 const a1, struct io_Writer__3827 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_9268__6171(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint64_t const builtin_zig_backend__262 = UINT64_C(3);

static bool const start_simplified_logic__109 = false;

static uint8_t const builtin_output_mode__263 = UINT8_C(1);

static uint8_t const builtin_link_mode__264 = UINT8_C(0);

static uint64_t features_features__329 = ((uint64_t)UINT64_C(0));

static uintptr_t const bit_set_IntegerBitSet_2864_29_bit_length__360 = 64ul;

static struct Target_Os__639 const builtin_os__270 = {{ .semver = {{13ul,0ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},{15ul,3ul,1ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}}} },UINT8_C(20)};

static bool const builtin_link_libc__274 = true;

static bool const posix_use_libc__1590 = true;

static uint8_t const c_native_os__2032 = UINT8_C(20);

static uint8_t const builtin_mode__273 = UINT8_C(3);

static bool const debug_runtime_safety__159 = false;

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__868 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__2940 = {{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),0ul,0ul},NULL};

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__922 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__3330 = {{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),0ul,0ul},NULL};

static struct Target_DynamicLinker__976 const Target_DynamicLinker_none__3578 = {"\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252",UINT8_C(0)};

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__889 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3151 = {{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),0ul,0ul},NULL};

static bool const builtin_is_test__266 = false;

static uint8_t *heap_CAllocator_alloc__3702(void *const a0, uintptr_t const a1, uint8_t const a2, uintptr_t const a3) {
 uint64_t t0;
 uint8_t *t2;
 bool t1;
 (void)a0;
 (void)a3;
 t0 = a1;
 t1 = t0 > UINT64_C(0);
 debug_assert__180(t1);
 t2 = heap_CAllocator_alignedAlloc__3699(a1, a2);
 return t2;
}

static bool heap_CAllocator_resize__3703(void *const a0, nav__3703_40 const a1, uint8_t const a2, uintptr_t const a3, uintptr_t const a4) {
 uintptr_t t0;
 uint64_t t1;
 uint64_t t2;
 uint8_t *t4;
 bool t3;
 (void)a0;
 (void)a2;
 (void)a4;
 t0 = a1.len;
 t1 = a3;
 t2 = t0;
 t3 = t1 <= t2;
 if (t3) {
  return true;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a1.ptr;
 t0 = heap_CAllocator_alignedAllocSize__3701(t4);
 t2 = a3;
 t1 = t0;
 t3 = t2 <= t1;
 if (t3) {
  return true;
 }
 goto zig_block_1;

 zig_block_1:;
 return false;
}

static uint8_t *heap_CAllocator_remap__3704(void *const a0, nav__3704_40 const a1, uint8_t const a2, uintptr_t const a3, uintptr_t const a4) {
 uint8_t *t0;
 uint8_t *t2;
 uint8_t *t3;
 bool t1;
 t1 = heap_CAllocator_resize__3703(a0, a1, a2, a3, a4);
 if (t1) {
  t2 = a1.ptr;
  t3 = (uint8_t *)t2;
  t0 = t3;
  goto zig_block_0;
 }
 t0 = NULL;
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static void heap_CAllocator_free__3705(void *const a0, nav__3705_40 const a1, uint8_t const a2, uintptr_t const a3) {
 uint8_t *t0;
 (void)a0;
 (void)a2;
 (void)a3;
 t0 = a1.ptr;
 heap_CAllocator_alignedFree__3700(t0);
 return;
}

static uint8_t *heap_CAllocator_alignedAlloc__3699(uintptr_t const a0, uint8_t const a1) {
 uint8_t const *t1;
 uintptr_t t3;
 uint64_t t4;
 void *t9;
 void *t5;
 uint8_t *t10;
 int t6;
 int32_t t7;
 uint8_t t2;
 uint8_t t0;
 bool t8;
 t0 = a1;
 t1 = (uint8_t const *)&t0;
 t2 = (*t1);
 t3 = mem_Alignment_toByteUnits__1112(t2);
 t3 = ((uintptr_t)8ul > t3) ? (uintptr_t)8ul : t3;
 t4 = t3;
 t3 = t4;
 t6 = posix_memalign(&t5, t3, a0);
 t7 = t6;
 t8 = t7 != INT32_C(0);
 if (t8) {
  return NULL;
 }
 goto zig_block_0;

 zig_block_0:;
 t9 = t5;
 t10 = (uint8_t *)t9;
 return t10;
}

static uintptr_t heap_CAllocator_alignedAllocSize__3701(uint8_t *const a0) {
 void const *t0;
 void const *t1;
 uintptr_t t2;
 t0 = (void const *)a0;
 t1 = t0;
 t2 = malloc_size(t1);
 return t2;
}

static void heap_CAllocator_alignedFree__3700(uint8_t *const a0) {
 void *t0;
 void *t1;
 t0 = (void *)a0;
 t1 = t0;
 free(t1);
 return;
}

static uintptr_t mem_Alignment_toByteUnits__1112(uint8_t const a0) {
 uintptr_t t1;
 uint8_t t0;
 t0 = a0;
 t1 = zig_shlw_u64((uintptr_t)1ul, t0, UINT8_C(64));
 return t1;
}

static struct mem_Allocator__599 const heap_c_allocator__3667 = {((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)};

static struct mem_Allocator__599 const c_allocator__235 = {((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__3694)};

static uint8_t const (*const shared_RUNNER_CTL_FIFO__801)[21] = &__anon_1087;

static uint8_t const (*const shared_RUNNER_ACK_FIFO__802)[21] = &__anon_1152;

static uint64_t const runner_fifo_PROTOCOL_VERSION__1125 = UINT64_C(2);

static struct Target_Cpu_Feature_Set__993 const Target_Cpu_Feature_Set_empty__3623 = {{0ul,0ul,0ul,0ul,0ul}};

static struct Target_Cpu__961 const builtin_cpu__269 = {((struct Target_Cpu_Model__988 const *)&Target_x86_cpu_core2__4052),{{14073748835532816ul,4506898179031040ul,1225016209309761536ul,0ul,0ul}},UINT8_C(42)};

static uint8_t const mem_native_endian__836 = UINT8_C(1);

static uint8_t const posix_native_os__1588 = UINT8_C(20);

static uint8_t const builtin_abi__268 = UINT8_C(0);

static uint8_t const builtin_object_format__272 = UINT8_C(0);

static struct Target__637 const builtin_target__271 = {{((struct Target_Cpu_Model__988 const *)&Target_x86_cpu_core2__4052),{{14073748835532816ul,4506898179031040ul,1225016209309761536ul,0ul,0ul}},UINT8_C(42)},{{ .semver = {{13ul,0ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},{15ul,3ul,1ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}}} },UINT8_C(20)},UINT8_C(0),UINT8_C(0),{"/usr/lib/dyld\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252",UINT8_C(13)}};

static struct builtin_CallingConvention__266 const builtin_CallingConvention_c__470 = {{ .x86_64_sysv = {{UINT64_C(0xaaaaaaaaaaaaaaaa),true}} },UINT8_C(4)};

static uint8_t const fs_path_native_os__4279 = UINT8_C(20);

static uint8_t const fs_native_os__1303 = UINT8_C(20);

static uint8_t const fs_Dir_native_os__3827 = UINT8_C(20);

static bool const fs_File_is_windows__1479 = false;

static bool const posix_unexpected_error_tracing__2019 = false;

static uint8_t logger_max_level__4255 = UINT8_C(0);

static bool const fs_Dir_have_flock__3828 = true;

static uint16_t const fmt_max_format_args__4915 = UINT16_C(32);

static bool const debug_default_enable_segfault_handler__205 = false;

static uint8_t const log_default_level__5504 = UINT8_C(0);

static struct std_Options__4261 const std_options__97 = {{0xaaaaaaaaaaaaaaaaul,true},{0xaaaaaaaaaaaaaaaaul,true},3ul,false,UINT8_C(0),false,true,false,false,false,UINT8_C(2)};

static bool const posix_lfs64_abi__2018 = false;

static uint8_t const (*const fmt_ANY__4918)[4] = &__anon_4937;

static uint8_t const unicode_native_endian__5186 = UINT8_C(1);

static uint32_t const unicode_replacement_character__5187 = UINT32_C(65533);

static bool const mem_backend_supports_vectors__864 = true;

static bool const builtin_valgrind_support__277 = false;

static nav__3683_38 const heap_page_size_min_default__3683 = {4096ul,false};

static uintptr_t const heap_page_size_min__3662 = 4096ul;

static uint8_t const c_MARKER_TYPE_SAMPLE_START__246 = UINT8_C(0);

static uint8_t const c_MARKER_TYPE_SAMPLE_END__247 = UINT8_C(1);

static uint8_t const c_MARKER_TYPE_BENCHMARK_START__248 = UINT8_C(2);

static uint8_t const c_MARKER_TYPE_BENCHMARK_END__249 = UINT8_C(3);

static bool const builtin_single_threaded__267 = false;

static struct once_Once_28_28function_20_27init_27_29_29__7273 c_instrument_hooks_current_timestamp_S_once__5831 = {{{{UINT32_C(0)}}},false};

static struct c_MachTimebaseInfo__7314 c_instrument_hooks_current_timestamp_S_cached__5830 = {UINT32_C(0),UINT32_C(0)};

static uint8_t const Thread_native_os__5849 = UINT8_C(20);

static bool const Thread_use_pthreads__5860 = true;

static uint32_t const array_hash_map_index_capacities__2914[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};

static uint8_t const array_hash_map_Index_28u8_29_empty_sentinel__5741 = UINT8_MAX;

static uint16_t const array_hash_map_Index_28u16_29_empty_sentinel__5746 = UINT16_MAX;

static uint32_t const array_hash_map_Index_28u32_29_empty_sentinel__5751 = UINT32_MAX;

static uint64_t const hash_wyhash_Wyhash_secret__6028[4] = {UINT64_C(11562461410679940143),UINT64_C(16646288086500911323),UINT64_C(10285213230658275043),UINT64_C(6384245875588680899)};

static bool const builtin_fuzz__279 = false;

static bool const mem_eqlBytes_allowed__854 = true;

static struct array_hash_map_Index_28u8_29__6669 const array_hash_map_Index_28u8_29_empty__5742 = {UINT8_MAX,UINT8_C(0xaa)};

static struct array_hash_map_Index_28u16_29__6673 const array_hash_map_Index_28u16_29_empty__5747 = {UINT16_MAX,UINT16_C(0xaaaa)};

static struct array_hash_map_Index_28u32_29__6677 const array_hash_map_Index_28u32_29_empty__5752 = {UINT32_MAX,UINT32_C(0xaaaaaaaa)};

static bool const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6084 = false;

static struct json_stringify_WriteStream__union_8236__8236 const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6085 = {{0xaaaaaaaaaaaaaaaaul},UINT8_C(2)};

struct macho_mach_header_64__9281 c_dummy_execute_header__2050 = {UINT32_C(0xaaaaaaaa),-0x55555556,-0x55555556,UINT32_C(0xaaaaaaaa),UINT32_C(0xaaaaaaaa),UINT32_C(0xaaaaaaaa),UINT32_C(0xaaaaaaaa),UINT32_C(0xaaaaaaaa)};

static struct mem_Allocator_VTable__602 const heap_CAllocator_vtable__3694 = {&heap_CAllocator_alloc__3702,&heap_CAllocator_resize__3703,&heap_CAllocator_remap__3704,&heap_CAllocator_free__3705};

static bool const heap_CAllocator_supports_malloc_size__3695 = true;

static bool const heap_CAllocator_supports_posix_memalign__3697 = true;

static struct Target_Cpu_Model__988 const Target_x86_cpu_core2__4052 = {{(uint8_t const *)&__anon_9352,5ul},{(uint8_t const *)&__anon_9352,5ul},{{5066549580791824ul,4506898179031040ul,1225014285164412928ul,0ul,0ul}}};

#else
#include "core.h"
#define ZIG_TARGET_MAX_INT_ALIGNMENT 16
#include "zig.h"
struct anon__lazy_58 {
 uint8_t const *ptr;
 uintptr_t len;
};
struct instruments_valgrind_ValgrindInstrument__564;
typedef struct anon__lazy_88 nav__786_39;
struct mem_Allocator__599;
struct mem_Allocator_VTable__602;
struct mem_Allocator__599 {
 void *ptr;
 struct mem_Allocator_VTable__602 const *vtable;
};
struct instruments_valgrind_ValgrindInstrument__564 {
 struct mem_Allocator__599 allocator;
};
struct anon__lazy_88 {
 struct instruments_valgrind_ValgrindInstrument__564 payload;
 uint16_t error;
};
struct fs_File__799;
typedef struct anon__lazy_99 nav__1280_39;
typedef struct anon__lazy_58 nav__1280_41;
struct fs_File__799 {
 int32_t handle;
};
struct anon__lazy_99 {
 struct fs_File__799 payload;
 uint16_t error;
};
struct fs_File_OpenFlags__1718;
struct fs_File_OpenFlags__1718 {
 uint8_t mode;
 uint8_t lock;
 bool lock_nonblocking;
 bool allow_ctty;
};
struct fifo_posix_Pipe_Writer__791;
typedef struct anon__lazy_107 nav__1282_39;
typedef struct anon__lazy_58 nav__1282_42;
struct array_list_ArrayListAligned_28u8_2cnull_29__811;
typedef struct anon__lazy_110 nav__1282_50;
struct anon__lazy_110 {
 uint8_t *ptr;
 uintptr_t len;
};
struct array_list_ArrayListAligned_28u8_2cnull_29__811 {
 struct anon__lazy_110 items;
 uintptr_t capacity;
 struct mem_Allocator__599 allocator;
};
struct fifo_posix_Pipe_Writer__791 {
 struct mem_Allocator__599 allocator;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 buffer;
 struct fs_File__799 file;
};
struct anon__lazy_107 {
 struct fifo_posix_Pipe_Writer__791 payload;
 uint16_t error;
};
typedef struct anon__lazy_99 nav__1282_59;
struct fifo_posix_Pipe_Reader__793;
typedef struct anon__lazy_117 nav__1281_39;
typedef struct anon__lazy_58 nav__1281_42;
typedef struct anon__lazy_110 nav__1281_50;
struct fifo_posix_Pipe_Reader__793 {
 struct mem_Allocator__599 allocator;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 buffer;
 struct fs_File__799 file;
};
struct anon__lazy_117 {
 struct fifo_posix_Pipe_Reader__793 payload;
 uint16_t error;
};
typedef struct anon__lazy_99 nav__1281_59;
struct runner_fifo_RunnerFifo__631;
typedef struct anon__lazy_123 nav__1129_39;
typedef struct anon__lazy_110 nav__1129_49;
struct runner_fifo_RunnerFifo__631 {
 struct mem_Allocator__599 allocator;
 struct fifo_posix_Pipe_Writer__791 writer;
 struct fifo_posix_Pipe_Reader__793 reader;
};
struct anon__lazy_123 {
 struct runner_fifo_RunnerFifo__631 payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__1129_64;
typedef struct anon__lazy_107 nav__1129_66;
typedef struct anon__lazy_117 nav__1129_71;
struct shared_Command__1795;
struct shared_Command__struct_1798__1798;
typedef struct anon__lazy_58 nav__1286_44;
struct shared_Command__struct_1798__1798 {
 struct anon__lazy_58 uri;
 int32_t pid;
};
struct shared_Command__struct_1799__1799;
struct shared_Command__struct_1799__1799 {
 struct anon__lazy_58 name;
 struct anon__lazy_58 version;
};
struct shared_Command__struct_1800__1800;
struct shared_MarkerType__1802;
struct shared_MarkerType__1802 {
 union {
  uint64_t SampleStart;
  uint64_t SampleEnd;
  uint64_t BenchmarkStart;
  uint64_t BenchmarkEnd;
 } payload;
 uint8_t tag;
};
struct shared_Command__struct_1800__1800 {
 struct shared_MarkerType__1802 marker;
 int32_t pid;
};
struct shared_Command__1795 {
 union {
  struct shared_Command__struct_1798__1798 ExecutedBenchmark;
  struct shared_Command__struct_1799__1799 SetIntegration;
  struct shared_Command__struct_1800__1800 AddMarker;
  uint64_t SetVersion;
  uint8_t IntegrationModeResponse;
 } payload;
 uint8_t tag;
};
typedef struct anon__lazy_110 nav__1286_65;
struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839;
struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *context;
};
struct os_linux_pollfd__1921;
struct os_linux_pollfd__1921 {
 int32_t fd;
 int16_t events;
 int16_t revents;
};
typedef struct anon__lazy_159 nav__1268_43;
struct anon__lazy_159 {
 struct os_linux_pollfd__1921 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_162 nav__1268_46;
struct anon__lazy_162 {
 uintptr_t payload;
 uint16_t error;
};
typedef struct anon__lazy_164 nav__5031_38;
struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089;
struct anon__lazy_164 {
 uint32_t payload;
 uint16_t error;
};
struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065;
struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *context;
};
struct io_Reader__2238;
typedef struct anon__lazy_162 nav__5031_50;
typedef struct anon__lazy_110 nav__5031_52;
struct io_Reader__2238 {
 void const *context;
 struct anon__lazy_162 (*readFn)(void const *, struct anon__lazy_110);
};
typedef struct anon__lazy_184 nav__5031_64;
struct anon__lazy_184 {
 uint16_t error;
 uint8_t payload[4];
};
typedef struct anon__lazy_164 nav__5030_38;
typedef struct anon__lazy_188 nav__5073_38;
struct anon__lazy_188 {
 int32_t payload;
 uint16_t error;
};
typedef struct anon__lazy_162 nav__5073_50;
typedef struct anon__lazy_110 nav__5073_52;
typedef struct anon__lazy_184 nav__5073_64;
typedef struct anon__lazy_188 nav__5072_38;
typedef struct anon__lazy_110 nav__5075_39;
typedef struct anon__lazy_194 nav__5075_40;
struct anon__lazy_194 {
 struct anon__lazy_110 payload;
 uint16_t error;
};
typedef struct anon__lazy_162 nav__5075_59;
typedef struct anon__lazy_199 nav__5075_69;
struct anon__lazy_199 {
 uint64_t payload;
 uint16_t error;
};
typedef struct anon__lazy_201 nav__5075_72;
struct anon__lazy_201 {
 uint16_t error;
 uint8_t payload[8];
};
typedef struct anon__lazy_58 nav__5074_39;
typedef struct anon__lazy_204 nav__5074_40;
struct anon__lazy_204 {
 struct anon__lazy_58 payload;
 uint16_t error;
};
typedef struct anon__lazy_110 nav__5074_55;
typedef struct anon__lazy_194 nav__5074_57;
typedef struct anon__lazy_208 nav__5071_39;
typedef struct anon__lazy_58 nav__5071_44;
struct anon__lazy_208 {
 struct shared_Command__struct_1798__1798 payload;
 uint16_t error;
};
typedef struct anon__lazy_188 nav__5071_57;
typedef struct anon__lazy_204 nav__5071_60;
typedef struct anon__lazy_208 nav__5070_39;
typedef struct anon__lazy_58 nav__5070_44;
typedef struct anon__lazy_216 nav__5081_39;
typedef struct anon__lazy_58 nav__5081_44;
struct anon__lazy_216 {
 struct shared_Command__struct_1799__1799 payload;
 uint16_t error;
};
typedef struct anon__lazy_204 nav__5081_57;
typedef struct anon__lazy_216 nav__5080_39;
typedef struct anon__lazy_58 nav__5080_44;
typedef struct anon__lazy_199 nav__5087_38;
typedef struct anon__lazy_162 nav__5087_50;
typedef struct anon__lazy_110 nav__5087_52;
typedef struct anon__lazy_201 nav__5087_64;
typedef struct anon__lazy_199 nav__5086_38;
typedef struct anon__lazy_224 nav__5085_39;
struct anon__lazy_224 {
 struct shared_MarkerType__1802 payload;
 uint16_t error;
};
typedef struct anon__lazy_164 nav__5085_54;
typedef struct anon__lazy_199 nav__5085_56;
typedef struct anon__lazy_224 nav__5084_39;
typedef struct anon__lazy_228 nav__5083_39;
struct anon__lazy_228 {
 struct shared_Command__struct_1800__1800 payload;
 uint16_t error;
};
typedef struct anon__lazy_188 nav__5083_57;
typedef struct anon__lazy_224 nav__5083_60;
typedef struct anon__lazy_228 nav__5082_39;
typedef struct anon__lazy_233 nav__5089_38;
struct anon__lazy_233 {
 uint16_t error;
 uint8_t payload;
};
typedef struct anon__lazy_164 nav__5089_46;
typedef struct anon__lazy_233 nav__5088_38;
typedef struct anon__lazy_239 nav__5029_39;
typedef struct anon__lazy_58 nav__5029_45;
struct anon__lazy_239 {
 struct shared_Command__1795 payload;
 uint16_t error;
};
typedef struct anon__lazy_164 nav__5029_66;
typedef struct anon__lazy_208 nav__5029_68;
typedef struct anon__lazy_216 nav__5029_70;
typedef struct anon__lazy_228 nav__5029_72;
typedef struct anon__lazy_199 nav__5029_74;
typedef struct anon__lazy_233 nav__5029_76;
typedef struct anon__lazy_239 nav__5028_39;
typedef struct anon__lazy_58 nav__5028_45;
typedef struct anon__lazy_239 nav__1291_39;
typedef struct anon__lazy_58 nav__1291_45;
typedef struct anon__lazy_110 nav__1291_68;
typedef struct anon__lazy_162 nav__1291_76;
struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 {
 struct anon__lazy_110 buffer;
 uintptr_t pos;
};
typedef struct anon__lazy_239 nav__1292_39;
typedef struct anon__lazy_248 nav__1292_42;
typedef struct anon__lazy_58 nav__1292_46;
struct anon__lazy_248 {
 uint64_t payload;
 bool is_null;
};
typedef struct anon__lazy_110 nav__1292_69;
typedef struct anon__lazy_248 nav__1293_40;
typedef struct anon__lazy_58 nav__1293_48;
typedef struct anon__lazy_239 nav__1293_60;
typedef struct anon__lazy_110 nav__1293_70;
typedef struct anon__lazy_255 nav__1293_77;
struct anon__lazy_255 {
 struct shared_Command__1795 f0;
};
typedef struct anon__lazy_110 nav__1139_50;
typedef struct anon__lazy_58 nav__1139_63;
typedef struct anon__lazy_248 nav__1139_76;
typedef struct anon__lazy_58 nav__1130_43;
typedef struct anon__lazy_260 nav__1130_45;
struct anon__lazy_260 {
 struct anon__lazy_58 f0;
};
typedef struct anon__lazy_233 nav__1140_38;
typedef struct anon__lazy_110 nav__1140_53;
typedef struct anon__lazy_58 nav__1140_66;
typedef struct anon__lazy_248 nav__1140_79;
typedef struct anon__lazy_239 nav__1140_81;
struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591;
typedef struct anon__lazy_265 nav__820_39;
typedef struct anon__lazy_110 nav__820_50;
struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 {
 struct runner_fifo_RunnerFifo__631 fifo;
};
struct anon__lazy_265 {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 payload;
 uint16_t error;
};
typedef struct anon__lazy_123 nav__820_62;
typedef struct anon__lazy_233 nav__820_65;
struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582;
typedef struct anon__lazy_271 nav__807_39;
typedef struct anon__lazy_110 nav__807_50;
struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 {
 struct runner_fifo_RunnerFifo__631 fifo;
};
struct anon__lazy_271 {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 payload;
 uint16_t error;
};
typedef struct anon__lazy_123 nav__807_62;
typedef struct anon__lazy_233 nav__807_65;
struct instruments_root_Instrument__554;
typedef struct anon__lazy_277 nav__743_39;
typedef struct anon__lazy_110 nav__743_53;
struct instruments_root_Instrument__554 {
 union {
  struct instruments_valgrind_ValgrindInstrument__564 valgrind;
  struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 walltime;
  struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 analysis;
 } payload;
 uint8_t tag;
};
struct anon__lazy_277 {
 struct instruments_root_Instrument__554 payload;
 uint16_t error;
};
typedef struct anon__lazy_88 nav__743_69;
typedef struct anon__lazy_265 nav__743_71;
typedef struct anon__lazy_271 nav__743_73;
struct InstrumentHooks;
typedef struct anon__lazy_284 nav__730_39;
typedef struct anon__lazy_110 nav__730_54;
struct environment_root_Environment__559;
struct environment_root_EnvironmentJson__845;
struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 {
 uint8_t *bytes;
 uintptr_t len;
 uintptr_t capacity;
};
struct array_hash_map_IndexHeader__892;
struct debug_SafetyLock__896;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 entries;
 struct array_hash_map_IndexHeader__892 *index_header;
};
struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 map;
};
struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 {
 uint8_t *bytes;
 uintptr_t len;
 uintptr_t capacity;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 entries;
 struct array_hash_map_IndexHeader__892 *index_header;
};
struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 map;
};
struct environment_root_EnvironmentJson__845 {
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 integration_environment;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 linked_libraries;
};
struct environment_linked_libraries_root_LinkedLibraries__849;
struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948;
struct environment_linked_libraries_root_LibraryInfo__946;
typedef struct anon__lazy_308 nav__730_90;
struct anon__lazy_308 {
 struct environment_linked_libraries_root_LibraryInfo__946 *ptr;
 uintptr_t len;
};
struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 {
 struct anon__lazy_308 items;
 uintptr_t capacity;
 struct mem_Allocator__599 allocator;
};
struct environment_linked_libraries_root_LinkedLibraries__849 {
 struct mem_Allocator__599 allocator;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 libraries;
};
struct environment_root_Environment__559 {
 struct mem_Allocator__599 allocator;
 struct environment_root_EnvironmentJson__845 data;
 struct environment_linked_libraries_root_LinkedLibraries__849 libs;
};
struct InstrumentHooks {
 struct instruments_root_Instrument__554 instrument;
 struct environment_root_Environment__559 environment;
};
struct anon__lazy_284 {
 struct InstrumentHooks payload;
 uint16_t error;
};
typedef struct anon__lazy_277 nav__730_100;
typedef struct anon__lazy_320 nav__238_46;
struct anon__lazy_320 {
 struct InstrumentHooks *payload;
 uint16_t error;
};
typedef struct anon__lazy_110 nav__238_51;
struct mem_Allocator_VTable__602 {
 uint8_t *(*alloc)(void *, uintptr_t, uint8_t, uintptr_t);
 bool (*resize)(void *, struct anon__lazy_110, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*remap)(void *, struct anon__lazy_110, uint8_t, uintptr_t, uintptr_t);
 void (*free)(void *, struct anon__lazy_110, uint8_t, uintptr_t);
};
typedef struct anon__lazy_308 nav__238_101;
typedef struct anon__lazy_284 nav__238_107;
typedef struct anon__lazy_320 nav__4541_40;
typedef struct anon__lazy_333 nav__4541_51;
struct anon__lazy_333 {
 uint8_t *payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__1348_39;
struct fs_Dir__1720;
struct fs_Dir__1720 {
 int32_t fd;
};
typedef struct anon__lazy_99 nav__1345_39;
typedef struct anon__lazy_58 nav__1345_41;
typedef struct anon__lazy_110 nav__1283_48;
typedef struct anon__lazy_110 nav__1288_48;
typedef struct anon__lazy_110 nav__1562_43;
typedef struct anon__lazy_58 nav__4913_46;
typedef struct anon__lazy_58 nav__1439_40;
typedef struct anon__lazy_162 nav__1439_47;
typedef struct anon__lazy_162 nav__1964_38;
typedef struct anon__lazy_159 nav__1964_41;
typedef struct anon__lazy_362 nav__1964_46;
struct anon__lazy_362 {
 uintptr_t payload;
 bool is_null;
};
typedef struct anon__lazy_162 nav__1429_38;
typedef struct anon__lazy_110 nav__1429_41;
typedef struct anon__lazy_110 nav__1559_43;
typedef struct anon__lazy_110 nav__4994_40;
typedef struct anon__lazy_162 nav__5027_38;
typedef struct anon__lazy_110 nav__5027_41;
typedef struct anon__lazy_184 nav__5069_39;
typedef struct anon__lazy_162 nav__5069_44;
typedef struct anon__lazy_110 nav__5069_46;
typedef struct anon__lazy_201 nav__5076_39;
typedef struct anon__lazy_162 nav__5076_44;
typedef struct anon__lazy_110 nav__5076_46;
typedef struct anon__lazy_110 nav__5077_39;
typedef struct anon__lazy_194 nav__5077_40;
typedef struct anon__lazy_333 nav__5077_51;
typedef struct anon__lazy_162 nav__5034_38;
typedef struct anon__lazy_110 nav__5034_41;
typedef struct anon__lazy_58 nav__4786_43;
typedef struct anon__lazy_255 nav__5099_39;
typedef struct anon__lazy_58 nav__5099_43;
typedef struct anon__lazy_58 nav__5102_39;
typedef struct anon__lazy_260 nav__5102_40;
typedef struct anon__lazy_110 nav__1131_50;
typedef struct anon__lazy_308 nav__773_67;
typedef struct anon__lazy_333 nav__5103_39;
typedef struct anon__lazy_110 nav__5103_52;
typedef struct anon__lazy_58 nav__5138_39;
typedef struct anon__lazy_58 nav__4742_40;
typedef struct anon__lazy_407 nav__4742_49;
struct anon__lazy_407 {
 uint16_t error;
 uint8_t payload[4096];
};
typedef struct anon__lazy_99 nav__4678_39;
typedef struct anon__lazy_58 nav__4678_42;
typedef struct anon__lazy_407 nav__4678_53;
typedef struct anon__lazy_110 nav__1527_42;
typedef struct anon__lazy_110 nav__1564_43;
typedef struct anon__lazy_58 nav__5201_46;
typedef struct anon__lazy_162 nav__1438_38;
typedef struct anon__lazy_58 nav__1438_41;
typedef struct anon__lazy_362 nav__5202_38;
typedef struct anon__lazy_162 nav__1428_38;
typedef struct anon__lazy_110 nav__1428_41;
typedef struct anon__lazy_162 nav__4986_38;
typedef struct anon__lazy_110 nav__4986_42;
typedef struct anon__lazy_110 nav__5036_40;
typedef struct anon__lazy_162 nav__5036_43;
typedef struct anon__lazy_333 nav__5206_39;
typedef struct anon__lazy_162 nav__5206_49;
typedef struct anon__lazy_162 nav__5035_38;
typedef struct anon__lazy_110 nav__5035_41;
typedef struct anon__lazy_58 nav__5296_40;
typedef struct anon__lazy_110 nav__5296_51;
typedef struct anon__lazy_255 nav__5297_39;
typedef struct anon__lazy_58 nav__5297_43;
typedef struct anon__lazy_110 nav__5297_59;
typedef struct anon__lazy_194 nav__5297_63;
typedef struct anon__lazy_110 nav__5298_41;
typedef struct anon__lazy_194 nav__5298_45;
typedef struct anon__lazy_110 nav__5299_41;
typedef struct anon__lazy_194 nav__5299_45;
typedef struct anon__lazy_58 nav__5301_39;
typedef struct anon__lazy_260 nav__5301_40;
typedef struct anon__lazy_110 nav__5301_47;
typedef struct anon__lazy_194 nav__5301_51;
typedef struct anon__lazy_110 nav__1287_49;
typedef struct anon__lazy_110 nav__1294_53;
typedef struct anon__lazy_159 nav__1294_61;
typedef struct anon__lazy_162 nav__1294_64;
typedef struct anon__lazy_308 nav__3643_48;
typedef struct anon__lazy_58 nav__5144_39;
typedef struct anon__lazy_407 nav__2022_39;
typedef struct anon__lazy_58 nav__2022_41;
typedef struct anon__lazy_110 nav__2022_47;
typedef struct anon__lazy_99 nav__4679_39;
typedef struct anon__lazy_188 nav__4679_50;
typedef struct anon__lazy_110 nav__1565_43;
typedef struct anon__lazy_194 nav__1565_55;
typedef struct anon__lazy_58 nav__5374_45;
typedef struct anon__lazy_58 nav__5376_45;
typedef struct anon__lazy_162 nav__1757_38;
typedef struct anon__lazy_58 nav__1757_40;
typedef struct anon__lazy_162 nav__1749_38;
typedef struct anon__lazy_110 nav__1749_40;
typedef struct anon__lazy_162 nav__5445_38;
typedef struct anon__lazy_473 nav__5445_42;
struct anon__lazy_473 {
 uintptr_t f0;
 uint8_t f1;
};
typedef struct anon__lazy_333 nav__5446_39;
typedef struct anon__lazy_110 nav__5446_52;
typedef struct anon__lazy_162 nav__5033_38;
typedef struct anon__lazy_110 nav__5033_41;
typedef struct anon__lazy_58 nav__5448_39;
struct io_Writer__3878;
typedef struct anon__lazy_255 nav__5741_40;
typedef struct anon__lazy_162 nav__5741_43;
typedef struct anon__lazy_58 nav__5741_45;
struct io_Writer__3878 {
 void const *context;
 struct anon__lazy_162 (*writeFn)(void const *, struct anon__lazy_58);
};
struct fmt_FormatOptions__4300;
typedef struct anon__lazy_362 nav__5741_67;
struct fmt_FormatOptions__4300 {
 struct anon__lazy_362 precision;
 struct anon__lazy_362 width;
 uint32_t fill;
 uint8_t alignment;
};
typedef struct anon__lazy_110 nav__5608_39;
typedef struct anon__lazy_194 nav__5608_40;
typedef struct anon__lazy_255 nav__5608_42;
typedef struct anon__lazy_58 nav__5608_48;
struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868;
struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *context;
};
typedef struct anon__lazy_162 nav__5608_71;
typedef struct anon__lazy_162 nav__6109_41;
typedef struct anon__lazy_58 nav__6109_43;
typedef struct anon__lazy_362 nav__6109_53;
typedef struct anon__lazy_110 nav__5610_39;
typedef struct anon__lazy_194 nav__5610_40;
typedef struct anon__lazy_162 nav__5610_54;
typedef struct anon__lazy_58 nav__5610_56;
typedef struct anon__lazy_162 nav__6111_41;
typedef struct anon__lazy_58 nav__6111_43;
typedef struct anon__lazy_110 nav__5611_39;
typedef struct anon__lazy_194 nav__5611_40;
typedef struct anon__lazy_162 nav__5611_54;
typedef struct anon__lazy_58 nav__5611_56;
typedef struct anon__lazy_58 nav__6112_40;
typedef struct anon__lazy_260 nav__6112_41;
typedef struct anon__lazy_162 nav__6112_44;
typedef struct anon__lazy_362 nav__6112_55;
typedef struct anon__lazy_110 nav__5612_39;
typedef struct anon__lazy_194 nav__5612_40;
typedef struct anon__lazy_58 nav__5612_42;
typedef struct anon__lazy_260 nav__5612_43;
typedef struct anon__lazy_162 nav__5612_59;
typedef struct anon__lazy_110 nav__1529_41;
typedef struct anon__lazy_308 nav__4347_43;
typedef struct anon__lazy_188 nav__1769_38;
typedef struct anon__lazy_110 nav__1575_39;
typedef struct anon__lazy_110 nav__5700_39;
typedef struct anon__lazy_162 nav__5700_50;
typedef struct anon__lazy_110 nav__5701_39;
typedef struct anon__lazy_194 nav__5701_40;
typedef struct anon__lazy_333 nav__5701_51;
typedef struct anon__lazy_110 nav__5702_40;
typedef struct anon__lazy_162 nav__5703_48;
typedef struct anon__lazy_58 nav__5703_50;
typedef struct anon__lazy_58 nav__5704_45;
typedef struct anon__lazy_58 nav__5705_45;
typedef struct anon__lazy_162 nav__5707_48;
typedef struct anon__lazy_58 nav__5707_50;
typedef struct anon__lazy_162 nav__5724_38;
typedef struct anon__lazy_58 nav__5724_41;
typedef struct anon__lazy_58 nav__5732_40;
typedef struct anon__lazy_162 nav__5732_43;
typedef struct anon__lazy_58 nav__6122_44;
typedef struct anon__lazy_362 nav__6122_56;
typedef struct anon__lazy_162 nav__6122_60;
typedef struct anon__lazy_539 nav__6122_66;
struct anon__lazy_539 {
 int32_t f0;
 struct anon__lazy_58 f1;
};
typedef struct anon__lazy_544 nav__6122_76;
struct anon__lazy_544 {
 struct anon__lazy_58 f0;
 struct anon__lazy_58 f1;
};
typedef struct anon__lazy_546 nav__6122_78;
struct anon__lazy_546 {
 int32_t f0;
 struct shared_MarkerType__1802 f1;
};
typedef struct anon__lazy_548 nav__6122_80;
struct anon__lazy_548 {
 uint64_t f0;
};
typedef struct anon__lazy_551 nav__6122_84;
struct anon__lazy_551 {
 uint8_t f0;
};
typedef struct anon__lazy_58 nav__6108_44;
typedef struct anon__lazy_362 nav__6108_56;
typedef struct anon__lazy_162 nav__6108_60;
typedef struct anon__lazy_110 nav__4992_39;
typedef struct anon__lazy_362 nav__6110_41;
typedef struct anon__lazy_162 nav__6110_45;
typedef struct anon__lazy_58 nav__6110_47;
typedef struct anon__lazy_58 nav__6113_39;
typedef struct anon__lazy_362 nav__6113_44;
typedef struct anon__lazy_162 nav__6113_48;
typedef struct anon__lazy_110 nav__6117_39;
typedef struct anon__lazy_110 nav__6118_39;
typedef struct anon__lazy_162 nav__4912_38;
typedef struct anon__lazy_58 nav__4912_41;
typedef struct anon__lazy_58 nav__6120_40;
typedef struct anon__lazy_162 nav__4987_38;
typedef struct anon__lazy_58 nav__4987_42;
typedef struct anon__lazy_110 nav__4987_50;
typedef struct anon__lazy_162 nav__5731_38;
typedef struct anon__lazy_58 nav__5731_41;
typedef struct anon__lazy_58 nav__6123_40;
typedef struct anon__lazy_539 nav__6123_41;
typedef struct anon__lazy_162 nav__6123_44;
typedef struct anon__lazy_58 nav__6124_40;
typedef struct anon__lazy_544 nav__6124_41;
typedef struct anon__lazy_162 nav__6124_44;
typedef struct anon__lazy_546 nav__6125_40;
typedef struct anon__lazy_162 nav__6125_43;
typedef struct anon__lazy_58 nav__6125_45;
typedef struct anon__lazy_548 nav__6126_39;
typedef struct anon__lazy_162 nav__6126_42;
typedef struct anon__lazy_58 nav__6126_44;
typedef struct anon__lazy_551 nav__6127_39;
typedef struct anon__lazy_162 nav__6127_42;
typedef struct anon__lazy_58 nav__6127_44;
typedef struct anon__lazy_233 nav__5753_38;
typedef struct anon__lazy_162 nav__5769_38;
typedef struct anon__lazy_58 nav__5769_40;
typedef struct anon__lazy_233 nav__5769_48;
typedef struct anon__lazy_164 nav__5769_50;
typedef struct anon__lazy_58 nav__6129_39;
typedef struct anon__lazy_362 nav__6129_44;
typedef struct anon__lazy_162 nav__6129_48;
typedef struct anon__lazy_110 nav__6129_57;
typedef struct anon__lazy_233 nav__6129_60;
typedef struct anon__lazy_362 nav__6139_41;
typedef struct anon__lazy_162 nav__6139_45;
typedef struct anon__lazy_58 nav__6139_47;
typedef struct anon__lazy_596 nav__6139_56;
struct anon__lazy_596 {
 uint8_t array[2];
};
typedef struct anon__lazy_110 nav__6139_58;
typedef struct anon__lazy_362 nav__6138_41;
typedef struct anon__lazy_162 nav__6138_45;
typedef struct anon__lazy_58 nav__6138_47;
typedef struct anon__lazy_362 nav__6128_41;
typedef struct anon__lazy_162 nav__6128_45;
typedef struct anon__lazy_58 nav__6128_47;
typedef struct anon__lazy_162 nav__1553_38;
typedef struct anon__lazy_58 nav__1553_42;
typedef struct anon__lazy_162 nav__6130_48;
typedef struct anon__lazy_58 nav__6130_50;
typedef struct anon__lazy_58 nav__6131_40;
typedef struct anon__lazy_162 nav__6131_51;
typedef struct anon__lazy_58 nav__6133_40;
typedef struct anon__lazy_539 nav__6133_41;
typedef struct anon__lazy_162 nav__6133_44;
typedef struct anon__lazy_362 nav__6133_55;
typedef struct anon__lazy_58 nav__6134_40;
typedef struct anon__lazy_544 nav__6134_41;
typedef struct anon__lazy_162 nav__6134_44;
typedef struct anon__lazy_362 nav__6134_55;
typedef struct anon__lazy_546 nav__6135_40;
typedef struct anon__lazy_162 nav__6135_43;
typedef struct anon__lazy_58 nav__6135_45;
typedef struct anon__lazy_362 nav__6135_58;
typedef struct anon__lazy_548 nav__6136_39;
typedef struct anon__lazy_162 nav__6136_42;
typedef struct anon__lazy_58 nav__6136_44;
typedef struct anon__lazy_362 nav__6136_55;
typedef struct anon__lazy_551 nav__6137_39;
typedef struct anon__lazy_162 nav__6137_42;
typedef struct anon__lazy_58 nav__6137_44;
typedef struct anon__lazy_362 nav__6137_55;
typedef struct anon__lazy_596 nav__5509_39;
typedef struct anon__lazy_164 nav__5759_38;
typedef struct anon__lazy_58 nav__5759_40;
typedef struct anon__lazy_596 nav__5759_48;
typedef struct anon__lazy_616 nav__5759_52;
struct anon__lazy_616 {
 uint8_t array[3];
};
typedef struct anon__lazy_618 nav__5759_56;
struct anon__lazy_618 {
 uint8_t array[4];
};
typedef struct anon__lazy_233 nav__5752_38;
typedef struct anon__lazy_233 nav__6145_38;
typedef struct anon__lazy_110 nav__6145_40;
typedef struct anon__lazy_233 nav__5754_38;
typedef struct anon__lazy_110 nav__5754_40;
typedef struct anon__lazy_58 nav__5736_40;
typedef struct anon__lazy_162 nav__5736_43;
typedef struct anon__lazy_58 nav__1547_41;
typedef struct anon__lazy_362 nav__6140_41;
typedef struct anon__lazy_162 nav__6140_45;
typedef struct anon__lazy_58 nav__6140_47;
typedef struct anon__lazy_58 nav__6141_39;
typedef struct anon__lazy_362 nav__6141_44;
typedef struct anon__lazy_162 nav__6141_48;
typedef struct anon__lazy_362 nav__6147_44;
typedef struct anon__lazy_162 nav__6147_48;
typedef struct anon__lazy_58 nav__6147_50;
typedef struct anon__lazy_548 nav__6147_56;
typedef struct anon__lazy_362 nav__6142_44;
typedef struct anon__lazy_162 nav__6142_48;
typedef struct anon__lazy_58 nav__6142_50;
typedef struct anon__lazy_362 nav__6143_41;
typedef struct anon__lazy_162 nav__6143_45;
typedef struct anon__lazy_58 nav__6143_47;
typedef struct anon__lazy_362 nav__6144_41;
typedef struct anon__lazy_162 nav__6144_45;
typedef struct anon__lazy_58 nav__6144_47;
typedef struct anon__lazy_164 nav__5761_38;
typedef struct anon__lazy_596 nav__5761_40;
typedef struct anon__lazy_164 nav__5763_38;
typedef struct anon__lazy_616 nav__5763_40;
typedef struct anon__lazy_164 nav__5767_38;
typedef struct anon__lazy_618 nav__5767_40;
typedef struct anon__lazy_110 nav__1566_43;
typedef struct anon__lazy_162 nav__1566_52;
typedef struct anon__lazy_58 nav__1548_41;
typedef struct anon__lazy_110 nav__1548_47;
typedef struct anon__lazy_362 nav__6154_41;
typedef struct anon__lazy_162 nav__6154_45;
typedef struct anon__lazy_58 nav__6154_47;
typedef struct anon__lazy_596 nav__6154_56;
typedef struct anon__lazy_110 nav__6154_58;
typedef struct anon__lazy_362 nav__6153_41;
typedef struct anon__lazy_162 nav__6153_45;
typedef struct anon__lazy_58 nav__6153_47;
typedef struct anon__lazy_362 nav__6146_41;
typedef struct anon__lazy_162 nav__6146_45;
typedef struct anon__lazy_58 nav__6146_47;
typedef struct anon__lazy_548 nav__6148_39;
typedef struct anon__lazy_162 nav__6148_42;
typedef struct anon__lazy_58 nav__6148_44;
typedef struct anon__lazy_548 nav__6149_39;
typedef struct anon__lazy_162 nav__6149_42;
typedef struct anon__lazy_58 nav__6149_44;
typedef struct anon__lazy_548 nav__6150_39;
typedef struct anon__lazy_162 nav__6150_42;
typedef struct anon__lazy_58 nav__6150_44;
typedef struct anon__lazy_548 nav__6151_39;
typedef struct anon__lazy_162 nav__6151_42;
typedef struct anon__lazy_58 nav__6151_44;
typedef struct anon__lazy_362 nav__6159_41;
typedef struct anon__lazy_162 nav__6159_45;
typedef struct anon__lazy_58 nav__6159_47;
typedef struct anon__lazy_362 nav__6152_41;
typedef struct anon__lazy_162 nav__6152_45;
typedef struct anon__lazy_58 nav__6152_47;
typedef struct anon__lazy_164 nav__5765_38;
typedef struct anon__lazy_616 nav__5765_40;
typedef struct anon__lazy_162 nav__1491_38;
typedef struct anon__lazy_473 nav__1491_42;
typedef struct anon__lazy_548 nav__6155_39;
typedef struct anon__lazy_162 nav__6155_42;
typedef struct anon__lazy_58 nav__6155_44;
typedef struct anon__lazy_362 nav__6155_55;
typedef struct anon__lazy_548 nav__6156_39;
typedef struct anon__lazy_162 nav__6156_42;
typedef struct anon__lazy_58 nav__6156_44;
typedef struct anon__lazy_362 nav__6156_55;
typedef struct anon__lazy_548 nav__6157_39;
typedef struct anon__lazy_162 nav__6157_42;
typedef struct anon__lazy_58 nav__6157_44;
typedef struct anon__lazy_362 nav__6157_55;
typedef struct anon__lazy_548 nav__6158_39;
typedef struct anon__lazy_162 nav__6158_42;
typedef struct anon__lazy_58 nav__6158_44;
typedef struct anon__lazy_362 nav__6158_55;
typedef struct anon__lazy_110 nav__239_50;
typedef struct anon__lazy_110 nav__731_55;
typedef struct anon__lazy_308 nav__731_91;
typedef struct anon__lazy_110 nav__6160_52;
typedef struct anon__lazy_110 nav__808_51;
typedef struct anon__lazy_110 nav__821_51;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152;
typedef struct anon__lazy_58 nav__774_44;
struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 {
 struct anon__lazy_58 *keys;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *values;
 uint32_t len;
 uint32_t index;
};
typedef struct anon__lazy_308 nav__774_76;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 {
 struct anon__lazy_58 *key_ptr;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *value_ptr;
};
typedef struct anon__lazy_677 nav__774_87;
struct anon__lazy_677 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 payload;
 bool is_null;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173;
struct environment_root_EnvironmentValue__856;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 {
 struct anon__lazy_58 *keys;
 struct environment_root_EnvironmentValue__856 *values;
 uint32_t len;
 uint32_t index;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 {
 uint8_t *bytes;
 uintptr_t len;
 uintptr_t capacity;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 entries;
 struct array_hash_map_IndexHeader__892 *index_header;
};
struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 map;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6181;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6181 {
 struct anon__lazy_58 *key_ptr;
 struct environment_root_EnvironmentValue__856 *value_ptr;
};
typedef struct anon__lazy_694 nav__774_104;
struct anon__lazy_694 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6181 payload;
 bool is_null;
};
typedef struct anon__lazy_696 nav__774_109;
struct anon__lazy_696 {
 struct anon__lazy_58 const *ptr;
 uintptr_t len;
};
struct environment_root_EnvironmentValue__856 {
 union {
  struct anon__lazy_58 string;
  struct anon__lazy_696 list;
 } payload;
 uint8_t tag;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205;
struct environment_linked_libraries_root_LibraryEntry__864;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 {
 struct anon__lazy_58 *keys;
 struct environment_linked_libraries_root_LibraryEntry__864 *values;
 uint32_t len;
 uint32_t index;
};
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213 {
 struct anon__lazy_58 *key_ptr;
 struct environment_linked_libraries_root_LibraryEntry__864 *value_ptr;
};
typedef struct anon__lazy_709 nav__774_122;
struct anon__lazy_709 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213 payload;
 bool is_null;
};
typedef struct anon__lazy_58 nav__3800_42;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 {
 uint8_t *ptrs[3];
 uintptr_t len;
 uintptr_t capacity;
};
typedef struct anon__lazy_720 nav__3800_62;
struct anon__lazy_720 {
 struct anon__lazy_58 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_723 nav__3800_65;
struct anon__lazy_723 {
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_677 nav__6161_39;
typedef struct anon__lazy_58 nav__6161_44;
typedef struct anon__lazy_58 nav__4011_42;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 {
 uint8_t *ptrs[3];
 uintptr_t len;
 uintptr_t capacity;
};
typedef struct anon__lazy_720 nav__4011_62;
typedef struct anon__lazy_736 nav__4011_65;
struct anon__lazy_736 {
 struct environment_root_EnvironmentValue__856 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_694 nav__6163_39;
typedef struct anon__lazy_58 nav__6163_44;
typedef struct anon__lazy_58 nav__777_43;
typedef struct anon__lazy_696 nav__777_46;
typedef struct anon__lazy_308 nav__777_79;
typedef struct anon__lazy_58 nav__4190_42;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459;
struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 {
 uint8_t *ptrs[3];
 uintptr_t len;
 uintptr_t capacity;
};
typedef struct anon__lazy_720 nav__4190_62;
typedef struct anon__lazy_761 nav__4190_65;
struct anon__lazy_761 {
 struct environment_linked_libraries_root_LibraryEntry__864 *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_709 nav__6165_39;
typedef struct anon__lazy_58 nav__6165_44;
typedef struct anon__lazy_308 nav__3644_50;
typedef struct anon__lazy_58 nav__3644_56;
struct environment_linked_libraries_root_LibraryInfo__946 {
 struct anon__lazy_58 path;
 struct anon__lazy_58 soname;
 struct anon__lazy_58 build_id;
};
typedef struct anon__lazy_58 nav__6180_39;
typedef struct anon__lazy_720 nav__6180_41;
typedef struct anon__lazy_723 nav__6181_40;
typedef struct anon__lazy_58 nav__6189_39;
typedef struct anon__lazy_720 nav__6189_41;
typedef struct anon__lazy_736 nav__6190_40;
typedef struct anon__lazy_58 nav__6190_51;
typedef struct anon__lazy_696 nav__6190_54;
typedef struct anon__lazy_58 nav__6191_40;
typedef struct anon__lazy_696 nav__6191_42;
typedef struct anon__lazy_110 nav__6191_54;
typedef struct anon__lazy_110 nav__4129_51;
struct array_hash_map_IndexHeader__892 {
 zig_align(4) uint8_t bit_index;
};
typedef struct anon__lazy_110 nav__3979_52;
typedef struct anon__lazy_110 nav__3935_51;
typedef struct anon__lazy_58 nav__6199_39;
typedef struct anon__lazy_720 nav__6199_41;
typedef struct anon__lazy_761 nav__6200_40;
typedef struct anon__lazy_58 nav__6200_51;
struct environment_linked_libraries_root_LibraryEntry__864 {
 struct anon__lazy_58 path;
 struct anon__lazy_58 build_id;
};
typedef struct anon__lazy_110 nav__4308_51;
typedef struct anon__lazy_58 nav__3645_46;
typedef struct anon__lazy_308 nav__4349_42;
typedef struct anon__lazy_58 nav__6272_39;
typedef struct anon__lazy_696 nav__6272_41;
typedef struct anon__lazy_110 nav__4161_39;
typedef struct anon__lazy_110 nav__6273_40;
typedef struct anon__lazy_110 nav__6274_40;
typedef struct anon__lazy_110 nav__3967_39;
typedef struct anon__lazy_110 nav__4340_39;
typedef struct anon__lazy_308 nav__4395_40;
typedef struct anon__lazy_58 nav__4395_55;
typedef struct anon__lazy_308 nav__6275_41;
typedef struct anon__lazy_110 nav__6275_51;
typedef struct anon__lazy_110 nav__6276_39;
typedef struct anon__lazy_110 nav__6292_39;
typedef struct anon__lazy_110 nav__6293_39;
typedef struct anon__lazy_308 nav__6293_42;
typedef struct anon__lazy_110 nav__240_55;
typedef struct anon__lazy_308 nav__240_91;
typedef struct anon__lazy_110 nav__1134_50;
typedef struct anon__lazy_58 nav__1134_63;
typedef struct anon__lazy_248 nav__1134_76;
typedef struct anon__lazy_110 nav__810_51;
typedef struct anon__lazy_110 nav__823_51;
typedef struct anon__lazy_110 nav__241_55;
typedef struct anon__lazy_308 nav__241_91;
typedef struct anon__lazy_110 nav__1135_50;
typedef struct anon__lazy_58 nav__1135_63;
typedef struct anon__lazy_248 nav__1135_76;
typedef struct anon__lazy_110 nav__811_51;
typedef struct anon__lazy_110 nav__824_51;
typedef struct anon__lazy_110 nav__242_55;
typedef struct anon__lazy_308 nav__242_91;
typedef struct anon__lazy_110 nav__1136_51;
typedef struct anon__lazy_58 nav__1136_61;
typedef struct anon__lazy_248 nav__1136_76;
typedef struct anon__lazy_110 nav__812_52;
typedef struct anon__lazy_110 nav__825_52;
typedef struct anon__lazy_110 nav__243_56;
typedef struct anon__lazy_308 nav__243_92;
typedef struct anon__lazy_58 nav__6298_39;
typedef struct anon__lazy_854 nav__6300_45;
struct anon__lazy_854 {
 bool array[16];
};
typedef struct anon__lazy_856 nav__6300_47;
struct anon__lazy_856 {
 bool is_null;
 uint8_t payload;
};
typedef struct anon__lazy_856 nav__6302_38;
typedef struct anon__lazy_854 nav__6302_40;
typedef struct anon__lazy_110 nav__1137_51;
typedef struct anon__lazy_58 nav__1137_61;
typedef struct anon__lazy_248 nav__1137_76;
typedef struct anon__lazy_110 nav__813_52;
typedef struct anon__lazy_110 nav__826_52;
typedef struct anon__lazy_110 nav__245_56;
typedef struct anon__lazy_308 nav__245_92;
typedef struct anon__lazy_871 nav__245_103;
struct anon__lazy_871 {
 uint8_t const *f0;
 uint8_t const *f1;
};
typedef struct anon__lazy_194 nav__245_105;
typedef struct anon__lazy_110 nav__6310_39;
typedef struct anon__lazy_194 nav__6310_40;
typedef struct anon__lazy_871 nav__6310_43;
typedef struct anon__lazy_110 nav__6311_40;
typedef struct anon__lazy_110 nav__6312_39;
typedef struct anon__lazy_194 nav__6312_40;
typedef struct anon__lazy_871 nav__6312_43;
typedef struct anon__lazy_362 nav__6312_54;
typedef struct anon__lazy_110 nav__6313_39;
typedef struct anon__lazy_871 nav__6355_40;
typedef struct anon__lazy_162 nav__6355_43;
typedef struct anon__lazy_58 nav__6355_44;
typedef struct anon__lazy_362 nav__6355_55;
typedef struct anon__lazy_871 nav__6314_39;
struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173;
struct io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29__7165;
struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 {
 uint64_t bytes_written;
};
struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187;
struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *context;
};
typedef struct anon__lazy_162 nav__6314_53;
typedef struct anon__lazy_58 nav__6314_54;
typedef struct anon__lazy_362 nav__6315_38;
typedef struct anon__lazy_110 nav__6316_39;
typedef struct anon__lazy_194 nav__6316_40;
typedef struct anon__lazy_871 nav__6316_42;
typedef struct anon__lazy_162 nav__6316_57;
typedef struct anon__lazy_58 nav__6316_58;
typedef struct anon__lazy_162 nav__6354_38;
typedef struct anon__lazy_58 nav__6354_41;
typedef struct anon__lazy_362 nav__6356_42;
typedef struct anon__lazy_162 nav__6356_46;
typedef struct anon__lazy_58 nav__6356_47;
typedef struct anon__lazy_162 nav__6339_38;
typedef struct anon__lazy_58 nav__6339_42;
typedef struct anon__lazy_162 nav__4890_38;
typedef struct anon__lazy_58 nav__4890_40;
typedef struct anon__lazy_110 nav__1138_54;
typedef struct anon__lazy_58 nav__1138_69;
typedef struct anon__lazy_248 nav__1138_77;
typedef struct anon__lazy_110 nav__814_55;
typedef struct anon__lazy_110 nav__827_55;
typedef struct anon__lazy_110 nav__250_58;
typedef struct anon__lazy_308 nav__250_94;
struct os_linux_timespec__struct_7321__7321;
struct os_linux_timespec__struct_7321__7321 {
 intptr_t sec;
 intptr_t nsec;
};
typedef struct anon__lazy_911 nav__254_41;
struct anon__lazy_911 {
 struct os_linux_timespec__struct_7321__7321 payload;
 uint16_t error;
};
typedef struct anon__lazy_911 nav__1938_39;
typedef struct anon__lazy_58 nav__778_41;
typedef struct anon__lazy_696 nav__778_46;
typedef struct anon__lazy_308 nav__778_79;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 {
 struct anon__lazy_58 *key_ptr;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *value_ptr;
 uintptr_t index;
 bool found_existing;
};
typedef struct anon__lazy_920 nav__778_93;
struct anon__lazy_920 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 payload;
 uint16_t error;
};
typedef struct anon__lazy_110 nav__778_96;
typedef struct anon__lazy_194 nav__778_98;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 {
 struct anon__lazy_58 *key_ptr;
 struct environment_root_EnvironmentValue__856 *value_ptr;
 uintptr_t index;
 bool found_existing;
};
typedef struct anon__lazy_925 nav__778_110;
struct anon__lazy_925 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__775_41;
typedef struct anon__lazy_308 nav__775_73;
typedef struct anon__lazy_110 nav__775_79;
typedef struct anon__lazy_194 nav__775_81;
typedef struct anon__lazy_696 nav__775_85;
typedef struct anon__lazy_110 nav__255_56;
typedef struct anon__lazy_308 nav__255_92;
typedef struct anon__lazy_58 nav__255_100;
typedef struct anon__lazy_110 nav__6358_39;
typedef struct anon__lazy_194 nav__6358_40;
typedef struct anon__lazy_58 nav__6358_43;
typedef struct anon__lazy_920 nav__3802_39;
typedef struct anon__lazy_58 nav__3802_44;
struct array_hash_map_StringContext__878;
typedef struct anon__lazy_925 nav__4013_39;
typedef struct anon__lazy_58 nav__4013_44;
typedef struct anon__lazy_920 nav__3803_39;
typedef struct anon__lazy_58 nav__3803_44;
typedef struct anon__lazy_925 nav__4014_39;
typedef struct anon__lazy_58 nav__4014_44;
typedef struct anon__lazy_920 nav__6359_39;
typedef struct anon__lazy_58 nav__6359_44;
typedef struct anon__lazy_362 nav__6359_67;
typedef struct anon__lazy_720 nav__6359_73;
typedef struct anon__lazy_723 nav__6359_76;
typedef struct anon__lazy_925 nav__6360_39;
typedef struct anon__lazy_58 nav__6360_44;
typedef struct anon__lazy_362 nav__6360_67;
typedef struct anon__lazy_720 nav__6360_73;
typedef struct anon__lazy_736 nav__6360_76;
typedef struct anon__lazy_233 nav__3812_60;
typedef struct anon__lazy_942 nav__3812_62;
struct anon__lazy_942 {
 struct array_hash_map_IndexHeader__892 *payload;
 uint16_t error;
};
typedef struct anon__lazy_362 nav__6361_38;
typedef struct anon__lazy_58 nav__6361_41;
typedef struct anon__lazy_946 nav__6361_61;
struct anon__lazy_946 {
 uint32_t *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_720 nav__6361_64;
typedef struct anon__lazy_58 nav__6362_42;
typedef struct anon__lazy_946 nav__6362_66;
typedef struct anon__lazy_720 nav__6362_69;
typedef struct anon__lazy_723 nav__6362_72;
typedef struct anon__lazy_233 nav__4023_60;
typedef struct anon__lazy_942 nav__4023_62;
typedef struct anon__lazy_362 nav__6363_38;
typedef struct anon__lazy_58 nav__6363_41;
typedef struct anon__lazy_946 nav__6363_61;
typedef struct anon__lazy_720 nav__6363_64;
typedef struct anon__lazy_58 nav__6364_42;
typedef struct anon__lazy_946 nav__6364_66;
typedef struct anon__lazy_720 nav__6364_69;
typedef struct anon__lazy_736 nav__6364_72;
typedef struct anon__lazy_110 nav__3959_52;
typedef struct anon__lazy_194 nav__3959_54;
typedef struct anon__lazy_946 nav__3959_62;
typedef struct anon__lazy_58 nav__3959_65;
typedef struct anon__lazy_720 nav__3959_67;
typedef struct anon__lazy_723 nav__3959_72;
typedef struct anon__lazy_233 nav__3977_38;
typedef struct anon__lazy_942 nav__3978_40;
typedef struct anon__lazy_110 nav__3978_51;
typedef struct anon__lazy_194 nav__3978_53;
typedef struct anon__lazy_58 nav__6365_40;
typedef struct anon__lazy_946 nav__6366_39;
typedef struct anon__lazy_58 nav__6367_40;
typedef struct anon__lazy_362 nav__6368_38;
typedef struct anon__lazy_58 nav__6368_41;
struct array_hash_map_Index_28u8_29__6735;
typedef struct anon__lazy_981 nav__6368_58;
struct anon__lazy_981 {
 struct array_hash_map_Index_28u8_29__6735 *ptr;
 uintptr_t len;
};
struct array_hash_map_Index_28u8_29__6735 {
 uint8_t entry_index;
 uint8_t distance_from_start_index;
};
typedef struct anon__lazy_362 nav__6369_38;
typedef struct anon__lazy_58 nav__6369_41;
struct array_hash_map_Index_28u16_29__6739;
typedef struct anon__lazy_986 nav__6369_58;
struct anon__lazy_986 {
 struct array_hash_map_Index_28u16_29__6739 *ptr;
 uintptr_t len;
};
struct array_hash_map_Index_28u16_29__6739 {
 uint16_t entry_index;
 uint16_t distance_from_start_index;
};
typedef struct anon__lazy_362 nav__6370_38;
typedef struct anon__lazy_58 nav__6370_41;
struct array_hash_map_Index_28u32_29__6743;
typedef struct anon__lazy_991 nav__6370_58;
struct anon__lazy_991 {
 struct array_hash_map_Index_28u32_29__6743 *ptr;
 uintptr_t len;
};
struct array_hash_map_Index_28u32_29__6743 {
 uint32_t entry_index;
 uint32_t distance_from_start_index;
};
typedef struct anon__lazy_58 nav__6371_42;
typedef struct anon__lazy_946 nav__6371_66;
typedef struct anon__lazy_720 nav__6371_69;
typedef struct anon__lazy_723 nav__6371_72;
typedef struct anon__lazy_981 nav__6371_77;
typedef struct anon__lazy_58 nav__6372_42;
typedef struct anon__lazy_946 nav__6372_66;
typedef struct anon__lazy_720 nav__6372_69;
typedef struct anon__lazy_723 nav__6372_72;
typedef struct anon__lazy_986 nav__6372_77;
typedef struct anon__lazy_58 nav__6373_42;
typedef struct anon__lazy_946 nav__6373_66;
typedef struct anon__lazy_720 nav__6373_69;
typedef struct anon__lazy_723 nav__6373_72;
typedef struct anon__lazy_991 nav__6373_77;
typedef struct anon__lazy_110 nav__4153_52;
typedef struct anon__lazy_194 nav__4153_54;
typedef struct anon__lazy_946 nav__4153_62;
typedef struct anon__lazy_58 nav__4153_65;
typedef struct anon__lazy_720 nav__4153_67;
typedef struct anon__lazy_736 nav__4153_72;
typedef struct anon__lazy_696 nav__4153_75;
typedef struct anon__lazy_58 nav__6374_40;
typedef struct anon__lazy_946 nav__6375_39;
typedef struct anon__lazy_58 nav__6376_40;
typedef struct anon__lazy_362 nav__6377_38;
typedef struct anon__lazy_58 nav__6377_41;
typedef struct anon__lazy_981 nav__6377_58;
typedef struct anon__lazy_362 nav__6378_38;
typedef struct anon__lazy_58 nav__6378_41;
typedef struct anon__lazy_986 nav__6378_58;
typedef struct anon__lazy_362 nav__6379_38;
typedef struct anon__lazy_58 nav__6379_41;
typedef struct anon__lazy_991 nav__6379_58;
typedef struct anon__lazy_58 nav__6380_42;
typedef struct anon__lazy_946 nav__6380_66;
typedef struct anon__lazy_720 nav__6380_69;
typedef struct anon__lazy_736 nav__6380_72;
typedef struct anon__lazy_981 nav__6380_77;
typedef struct anon__lazy_58 nav__6381_42;
typedef struct anon__lazy_946 nav__6381_66;
typedef struct anon__lazy_720 nav__6381_69;
typedef struct anon__lazy_736 nav__6381_72;
typedef struct anon__lazy_986 nav__6381_77;
typedef struct anon__lazy_58 nav__6382_42;
typedef struct anon__lazy_946 nav__6382_66;
typedef struct anon__lazy_720 nav__6382_69;
typedef struct anon__lazy_736 nav__6382_72;
typedef struct anon__lazy_991 nav__6382_77;
typedef struct anon__lazy_110 nav__6383_39;
typedef struct anon__lazy_194 nav__6383_40;
typedef struct anon__lazy_333 nav__6383_51;
typedef struct anon__lazy_110 nav__6385_39;
typedef struct anon__lazy_194 nav__6385_40;
typedef struct anon__lazy_333 nav__6385_51;
typedef struct anon__lazy_946 nav__6386_56;
typedef struct anon__lazy_981 nav__6386_60;
typedef struct anon__lazy_946 nav__6387_56;
typedef struct anon__lazy_986 nav__6387_60;
typedef struct anon__lazy_946 nav__6388_56;
typedef struct anon__lazy_991 nav__6388_60;
typedef struct anon__lazy_58 nav__3773_40;
typedef struct anon__lazy_58 nav__3774_40;
typedef struct anon__lazy_981 nav__6389_40;
typedef struct anon__lazy_362 nav__6390_38;
typedef struct anon__lazy_58 nav__6390_41;
typedef struct anon__lazy_981 nav__6390_47;
typedef struct anon__lazy_946 nav__6390_66;
typedef struct anon__lazy_720 nav__6390_69;
typedef struct anon__lazy_986 nav__6391_40;
typedef struct anon__lazy_362 nav__6392_38;
typedef struct anon__lazy_58 nav__6392_41;
typedef struct anon__lazy_986 nav__6392_47;
typedef struct anon__lazy_946 nav__6392_66;
typedef struct anon__lazy_720 nav__6392_69;
typedef struct anon__lazy_991 nav__6393_40;
typedef struct anon__lazy_362 nav__6394_38;
typedef struct anon__lazy_58 nav__6394_41;
typedef struct anon__lazy_991 nav__6394_47;
typedef struct anon__lazy_946 nav__6394_66;
typedef struct anon__lazy_720 nav__6394_69;
typedef struct anon__lazy_946 nav__6396_56;
typedef struct anon__lazy_981 nav__6396_60;
typedef struct anon__lazy_946 nav__6397_56;
typedef struct anon__lazy_986 nav__6397_60;
typedef struct anon__lazy_946 nav__6398_56;
typedef struct anon__lazy_991 nav__6398_60;
typedef struct anon__lazy_362 nav__6399_38;
typedef struct anon__lazy_58 nav__6399_41;
typedef struct anon__lazy_981 nav__6399_47;
typedef struct anon__lazy_946 nav__6399_66;
typedef struct anon__lazy_720 nav__6399_69;
typedef struct anon__lazy_362 nav__6400_38;
typedef struct anon__lazy_58 nav__6400_41;
typedef struct anon__lazy_986 nav__6400_47;
typedef struct anon__lazy_946 nav__6400_66;
typedef struct anon__lazy_720 nav__6400_69;
typedef struct anon__lazy_362 nav__6401_38;
typedef struct anon__lazy_58 nav__6401_41;
typedef struct anon__lazy_991 nav__6401_47;
typedef struct anon__lazy_946 nav__6401_66;
typedef struct anon__lazy_720 nav__6401_69;
typedef struct anon__lazy_333 nav__6402_39;
typedef struct anon__lazy_162 nav__6402_49;
typedef struct anon__lazy_333 nav__6404_39;
typedef struct anon__lazy_162 nav__6404_49;
typedef struct anon__lazy_58 nav__3736_39;
typedef struct anon__lazy_58 nav__3735_39;
typedef struct anon__lazy_333 nav__6405_39;
typedef struct anon__lazy_110 nav__6405_52;
typedef struct anon__lazy_58 nav__6465_39;
struct hash_wyhash_Wyhash__7880;
struct hash_wyhash_Wyhash__7880 {
 uint64_t a;
 uint64_t b;
 uint64_t state[3];
 uintptr_t total_len;
 uintptr_t buf_len;
 uint8_t buf[48];
};
typedef struct anon__lazy_58 nav__6466_39;
typedef struct anon__lazy_58 nav__856_39;
typedef struct anon__lazy_58 nav__776_41;
typedef struct anon__lazy_696 nav__776_43;
typedef struct anon__lazy_308 nav__776_76;
typedef struct anon__lazy_720 nav__776_83;
typedef struct anon__lazy_1066 nav__776_85;
struct anon__lazy_1066 {
 struct anon__lazy_720 payload;
 uint16_t error;
};
typedef struct anon__lazy_110 nav__776_88;
typedef struct anon__lazy_194 nav__776_90;
typedef struct anon__lazy_58 nav__256_45;
typedef struct anon__lazy_720 nav__256_47;
typedef struct anon__lazy_1066 nav__256_54;
typedef struct anon__lazy_110 nav__256_59;
typedef struct anon__lazy_308 nav__256_112;
typedef struct anon__lazy_696 nav__256_120;
typedef struct anon__lazy_58 nav__6470_39;
typedef struct anon__lazy_720 nav__6470_41;
typedef struct anon__lazy_1066 nav__6470_42;
typedef struct anon__lazy_333 nav__6470_54;
typedef struct anon__lazy_58 nav__6471_40;
typedef struct anon__lazy_720 nav__6471_42;
typedef struct anon__lazy_110 nav__6471_52;
typedef struct anon__lazy_333 nav__6472_39;
typedef struct anon__lazy_162 nav__6472_49;
typedef struct anon__lazy_110 nav__6473_39;
typedef struct anon__lazy_58 nav__6473_41;
typedef struct anon__lazy_720 nav__6473_43;
typedef struct anon__lazy_110 nav__257_55;
typedef struct anon__lazy_308 nav__257_91;
typedef struct anon__lazy_308 nav__3647_50;
typedef struct anon__lazy_58 nav__3647_56;
struct os_linux_dl_phdr_info__8025;
typedef struct anon__lazy_58 nav__779_44;
typedef struct anon__lazy_308 nav__779_76;
typedef struct anon__lazy_709 nav__779_87;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075;
struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 {
 struct anon__lazy_58 *key_ptr;
 struct environment_linked_libraries_root_LibraryEntry__864 *value_ptr;
 uintptr_t index;
 bool found_existing;
};
typedef struct anon__lazy_1085 nav__779_95;
struct anon__lazy_1085 {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 payload;
 uint16_t error;
};
typedef struct anon__lazy_110 nav__779_97;
typedef struct anon__lazy_194 nav__779_99;
typedef struct anon__lazy_308 nav__780_69;
typedef struct anon__lazy_58 nav__780_83;
typedef struct anon__lazy_110 nav__780_86;
typedef struct anon__lazy_1089 nav__780_89;
struct anon__lazy_1089 {
 struct anon__lazy_58 f0;
 int32_t f1;
};
typedef struct anon__lazy_194 nav__780_91;
struct json_stringify_StringifyOptions__8133;
struct json_stringify_StringifyOptions__8133 {
 uint8_t whitespace;
 bool emit_null_optional_fields;
 bool emit_strings_as_arrays;
 bool escape_unicode;
 bool emit_nonportable_numbers_as_strings;
};
struct fs_File_CreateFlags__8153;
struct fs_File_CreateFlags__8153 {
 uintptr_t mode;
 bool read;
 bool truncate;
 bool exclusive;
 uint8_t lock;
 bool lock_nonblocking;
};
typedef struct anon__lazy_99 nav__780_100;
typedef struct anon__lazy_308 nav__4382_44;
struct elf_Elf64_Phdr__8033;
struct os_linux_dl_phdr_info__8025 {
 uintptr_t addr;
 uint8_t const *name;
 struct elf_Elf64_Phdr__8033 *phdr;
 uint16_t phnum;
};
typedef struct anon__lazy_58 nav__6929_50;
struct environment_linked_libraries_elf_view__8184;
typedef struct anon__lazy_1105 nav__6929_59;
struct anon__lazy_1105 {
 struct elf_Elf64_Phdr__8033 const *ptr;
 uintptr_t len;
};
struct environment_linked_libraries_elf_view__8184 {
 uintptr_t base_addr;
 struct anon__lazy_1105 phdrs;
};
typedef struct anon__lazy_308 nav__6929_70;
typedef struct anon__lazy_1108 nav__6929_75;
struct anon__lazy_1108 {
 struct environment_linked_libraries_root_LibraryInfo__946 payload;
 bool is_null;
};
typedef struct anon__lazy_981 nav__4185_57;
typedef struct anon__lazy_986 nav__4185_62;
typedef struct anon__lazy_991 nav__4185_67;
typedef struct anon__lazy_1085 nav__4192_39;
typedef struct anon__lazy_58 nav__4192_44;
typedef struct anon__lazy_58 nav__6933_39;
typedef struct anon__lazy_58 nav__6990_40;
typedef struct anon__lazy_1089 nav__6990_41;
typedef struct anon__lazy_162 nav__6990_44;
typedef struct anon__lazy_362 nav__6990_53;
typedef struct anon__lazy_110 nav__6934_39;
typedef struct anon__lazy_194 nav__6934_40;
typedef struct anon__lazy_58 nav__6934_42;
typedef struct anon__lazy_1089 nav__6934_43;
typedef struct anon__lazy_162 nav__6934_59;
typedef struct anon__lazy_110 nav__6957_39;
typedef struct anon__lazy_194 nav__6957_40;
typedef struct anon__lazy_99 nav__1351_39;
typedef struct anon__lazy_58 nav__1351_41;
typedef struct anon__lazy_58 nav__6960_39;
typedef struct anon__lazy_1105 nav__6966_42;
struct elf_Elf64_Phdr__8033 {
 uint32_t p_type;
 uint32_t p_flags;
 uint64_t p_offset;
 uint64_t p_vaddr;
 uint64_t p_paddr;
 uint64_t p_filesz;
 uint64_t p_memsz;
 uint64_t p_align;
};
typedef struct anon__lazy_1108 nav__3646_39;
typedef struct anon__lazy_58 nav__3646_43;
typedef struct anon__lazy_1105 nav__3646_54;
struct elf_Elf64_Dyn__8329;
typedef struct anon__lazy_1136 nav__3646_62;
struct anon__lazy_1136 {
 struct elf_Elf64_Dyn__8329 const *ptr;
 uintptr_t len;
};
typedef struct anon__lazy_110 nav__3646_67;
typedef struct anon__lazy_194 nav__3646_69;
typedef struct anon__lazy_58 nav__4363_43;
typedef struct anon__lazy_1140 nav__4363_49;
struct anon__lazy_1140 {
 struct environment_linked_libraries_root_LibraryInfo__946 *payload;
 uint16_t error;
};
typedef struct anon__lazy_110 nav__6987_41;
typedef struct anon__lazy_194 nav__6987_45;
typedef struct anon__lazy_1085 nav__4193_39;
typedef struct anon__lazy_58 nav__4193_44;
typedef struct anon__lazy_110 nav__6988_41;
typedef struct anon__lazy_194 nav__6988_45;
typedef struct anon__lazy_110 nav__6991_41;
typedef struct anon__lazy_194 nav__6991_45;
struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291;
struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 stream;
 uintptr_t indent_level;
 struct json_stringify_StringifyOptions__8133 options;
 uint8_t next_punctuation;
};
typedef struct anon__lazy_110 nav__1533_39;
typedef struct anon__lazy_194 nav__1533_40;
typedef struct anon__lazy_110 nav__7022_41;
typedef struct anon__lazy_194 nav__7022_45;
typedef struct anon__lazy_99 nav__4681_39;
typedef struct anon__lazy_58 nav__4681_42;
typedef struct anon__lazy_407 nav__4681_53;
typedef struct anon__lazy_110 nav__7023_41;
typedef struct anon__lazy_194 nav__7023_45;
typedef struct anon__lazy_58 nav__6975_39;
typedef struct anon__lazy_1105 nav__6975_45;
typedef struct anon__lazy_58 nav__3638_39;
typedef struct anon__lazy_110 nav__3638_50;
typedef struct anon__lazy_194 nav__3638_52;
typedef struct anon__lazy_1136 nav__6970_40;
typedef struct anon__lazy_1105 nav__6970_46;
typedef struct anon__lazy_1157 nav__6970_52;
struct anon__lazy_1157 {
 struct elf_Elf64_Phdr__8033 payload;
 bool is_null;
};
struct elf_Elf64_Dyn__8329 {
 int64_t d_tag;
 uint64_t d_val;
};
typedef struct anon__lazy_1136 nav__6972_41;
typedef struct anon__lazy_362 nav__6972_45;
typedef struct anon__lazy_58 nav__6974_39;
typedef struct anon__lazy_1136 nav__6974_42;
typedef struct anon__lazy_362 nav__6974_47;
typedef struct anon__lazy_1140 nav__4388_40;
typedef struct anon__lazy_308 nav__4388_47;
typedef struct anon__lazy_162 nav__7032_41;
typedef struct anon__lazy_58 nav__7032_43;
typedef struct anon__lazy_110 nav__7024_39;
typedef struct anon__lazy_194 nav__7024_40;
typedef struct anon__lazy_162 nav__7024_54;
typedef struct anon__lazy_58 nav__7024_56;
typedef struct anon__lazy_1085 nav__7025_39;
typedef struct anon__lazy_58 nav__7025_44;
typedef struct anon__lazy_362 nav__7025_67;
typedef struct anon__lazy_720 nav__7025_73;
typedef struct anon__lazy_761 nav__7025_76;
typedef struct anon__lazy_162 nav__7035_41;
typedef struct anon__lazy_58 nav__7035_43;
typedef struct anon__lazy_110 nav__7026_39;
typedef struct anon__lazy_194 nav__7026_40;
typedef struct anon__lazy_162 nav__7026_54;
typedef struct anon__lazy_58 nav__7026_56;
typedef struct anon__lazy_162 nav__7036_41;
typedef struct anon__lazy_58 nav__7036_43;
typedef struct anon__lazy_110 nav__7027_39;
typedef struct anon__lazy_194 nav__7027_40;
typedef struct anon__lazy_162 nav__7027_54;
typedef struct anon__lazy_58 nav__7027_56;
typedef struct anon__lazy_58 nav__7029_62;
typedef struct anon__lazy_110 nav__1563_43;
typedef struct anon__lazy_162 nav__7039_41;
typedef struct anon__lazy_58 nav__7039_43;
typedef struct anon__lazy_110 nav__7030_39;
typedef struct anon__lazy_194 nav__7030_40;
typedef struct anon__lazy_162 nav__7030_54;
typedef struct anon__lazy_58 nav__7030_56;
typedef struct anon__lazy_99 nav__4682_39;
typedef struct anon__lazy_188 nav__4682_50;
typedef struct anon__lazy_162 nav__7040_41;
typedef struct anon__lazy_58 nav__7040_43;
typedef struct anon__lazy_110 nav__7031_39;
typedef struct anon__lazy_194 nav__7031_40;
typedef struct anon__lazy_162 nav__7031_54;
typedef struct anon__lazy_58 nav__7031_56;
typedef struct anon__lazy_58 nav__6968_39;
typedef struct anon__lazy_1105 nav__6968_45;
typedef struct anon__lazy_1157 nav__6969_39;
typedef struct anon__lazy_1105 nav__6969_45;
typedef struct anon__lazy_362 nav__6971_38;
typedef struct anon__lazy_1136 nav__6971_41;
typedef struct anon__lazy_58 nav__6973_39;
typedef struct anon__lazy_308 nav__4384_44;
typedef struct anon__lazy_308 nav__4389_44;
typedef struct anon__lazy_233 nav__4202_60;
typedef struct anon__lazy_942 nav__4202_62;
typedef struct anon__lazy_362 nav__7033_38;
typedef struct anon__lazy_58 nav__7033_41;
typedef struct anon__lazy_946 nav__7033_61;
typedef struct anon__lazy_720 nav__7033_64;
typedef struct anon__lazy_58 nav__7034_42;
typedef struct anon__lazy_946 nav__7034_66;
typedef struct anon__lazy_720 nav__7034_69;
typedef struct anon__lazy_761 nav__7034_72;
typedef struct anon__lazy_162 nav__7005_56;
typedef struct anon__lazy_58 nav__7005_58;
typedef struct anon__lazy_162 nav__7008_56;
typedef struct anon__lazy_58 nav__7008_58;
typedef struct anon__lazy_1200 nav__7006_43;
struct anon__lazy_1200 {
 bool is_null;
 bool payload;
};
typedef struct anon__lazy_162 nav__7000_54;
typedef struct anon__lazy_58 nav__7000_56;
typedef struct anon__lazy_1200 nav__7007_43;
typedef struct anon__lazy_58 nav__7115_39;
typedef struct anon__lazy_362 nav__7115_44;
typedef struct anon__lazy_162 nav__7115_52;
typedef struct anon__lazy_110 nav__7115_67;
typedef struct anon__lazy_233 nav__7115_70;
typedef struct anon__lazy_362 nav__7114_41;
typedef struct anon__lazy_596 nav__7114_52;
typedef struct anon__lazy_110 nav__7114_54;
typedef struct anon__lazy_58 nav__7114_57;
typedef struct anon__lazy_362 nav__7113_41;
typedef struct anon__lazy_362 nav__7117_41;
typedef struct anon__lazy_596 nav__7117_52;
typedef struct anon__lazy_110 nav__7117_54;
typedef struct anon__lazy_58 nav__7117_57;
typedef struct anon__lazy_362 nav__7116_41;
typedef struct anon__lazy_162 nav__7112_48;
typedef struct anon__lazy_58 nav__7112_50;
typedef struct anon__lazy_362 nav__7112_64;
typedef struct anon__lazy_162 nav__7111_48;
typedef struct anon__lazy_58 nav__7111_50;
typedef struct anon__lazy_58 nav__7110_39;
typedef struct anon__lazy_162 nav__7110_54;
typedef struct anon__lazy_233 nav__7110_64;
typedef struct anon__lazy_164 nav__7110_66;
typedef struct anon__lazy_58 nav__7109_39;
typedef struct anon__lazy_162 nav__7109_53;
typedef struct anon__lazy_58 nav__7015_41;
typedef struct anon__lazy_58 nav__7118_56;
typedef struct anon__lazy_677 nav__7118_65;
typedef struct anon__lazy_58 nav__7120_56;
typedef struct anon__lazy_709 nav__7120_65;
typedef struct anon__lazy_162 nav__7002_55;
typedef struct anon__lazy_58 nav__7002_57;
typedef struct anon__lazy_308 nav__4385_44;
typedef struct anon__lazy_1232 nav__4385_56;
struct anon__lazy_1232 {
 struct anon__lazy_308 payload;
 uint16_t error;
};
typedef struct anon__lazy_58 nav__4385_60;
typedef struct anon__lazy_110 nav__4332_52;
typedef struct anon__lazy_194 nav__4332_54;
typedef struct anon__lazy_946 nav__4332_62;
typedef struct anon__lazy_58 nav__4332_65;
typedef struct anon__lazy_720 nav__4332_67;
typedef struct anon__lazy_761 nav__4332_72;
typedef struct anon__lazy_58 nav__7100_40;
typedef struct anon__lazy_946 nav__7101_39;
typedef struct anon__lazy_58 nav__7102_40;
typedef struct anon__lazy_362 nav__7103_38;
typedef struct anon__lazy_58 nav__7103_41;
typedef struct anon__lazy_981 nav__7103_58;
typedef struct anon__lazy_362 nav__7104_38;
typedef struct anon__lazy_58 nav__7104_41;
typedef struct anon__lazy_986 nav__7104_58;
typedef struct anon__lazy_362 nav__7105_38;
typedef struct anon__lazy_58 nav__7105_41;
typedef struct anon__lazy_991 nav__7105_58;
typedef struct anon__lazy_58 nav__7106_42;
typedef struct anon__lazy_946 nav__7106_66;
typedef struct anon__lazy_720 nav__7106_69;
typedef struct anon__lazy_761 nav__7106_72;
typedef struct anon__lazy_981 nav__7106_77;
typedef struct anon__lazy_58 nav__7107_42;
typedef struct anon__lazy_946 nav__7107_66;
typedef struct anon__lazy_720 nav__7107_69;
typedef struct anon__lazy_761 nav__7107_72;
typedef struct anon__lazy_986 nav__7107_77;
typedef struct anon__lazy_58 nav__7108_42;
typedef struct anon__lazy_946 nav__7108_66;
typedef struct anon__lazy_720 nav__7108_69;
typedef struct anon__lazy_761 nav__7108_72;
typedef struct anon__lazy_991 nav__7108_77;
typedef struct anon__lazy_1200 nav__7010_38;
typedef struct anon__lazy_162 nav__5734_41;
typedef struct anon__lazy_58 nav__5734_43;
typedef struct anon__lazy_162 nav__5735_41;
typedef struct anon__lazy_58 nav__5735_43;
typedef struct anon__lazy_110 nav__5735_52;
typedef struct anon__lazy_58 nav__7130_56;
typedef struct anon__lazy_694 nav__7130_65;
typedef struct anon__lazy_696 nav__7130_69;
typedef struct anon__lazy_58 nav__7121_43;
typedef struct anon__lazy_308 nav__7122_40;
typedef struct anon__lazy_58 nav__7122_52;
typedef struct anon__lazy_110 nav__7122_56;
typedef struct anon__lazy_162 nav__7122_58;
typedef struct anon__lazy_308 nav__7123_40;
typedef struct anon__lazy_1232 nav__7123_41;
typedef struct anon__lazy_333 nav__7123_53;
typedef struct anon__lazy_58 nav__7123_57;
typedef struct anon__lazy_946 nav__7124_56;
typedef struct anon__lazy_981 nav__7124_60;
typedef struct anon__lazy_946 nav__7125_56;
typedef struct anon__lazy_986 nav__7125_60;
typedef struct anon__lazy_946 nav__7126_56;
typedef struct anon__lazy_991 nav__7126_60;
typedef struct anon__lazy_362 nav__7127_38;
typedef struct anon__lazy_58 nav__7127_41;
typedef struct anon__lazy_981 nav__7127_47;
typedef struct anon__lazy_946 nav__7127_66;
typedef struct anon__lazy_720 nav__7127_69;
typedef struct anon__lazy_362 nav__7128_38;
typedef struct anon__lazy_58 nav__7128_41;
typedef struct anon__lazy_986 nav__7128_47;
typedef struct anon__lazy_946 nav__7128_66;
typedef struct anon__lazy_720 nav__7128_69;
typedef struct anon__lazy_362 nav__7129_38;
typedef struct anon__lazy_58 nav__7129_41;
typedef struct anon__lazy_991 nav__7129_47;
typedef struct anon__lazy_946 nav__7129_66;
typedef struct anon__lazy_720 nav__7129_69;
typedef struct anon__lazy_58 nav__7136_43;
typedef struct anon__lazy_696 nav__7136_46;
typedef struct anon__lazy_58 nav__7131_43;
typedef struct anon__lazy_696 nav__7131_46;
typedef struct anon__lazy_58 nav__7020_41;
typedef struct anon__lazy_58 nav__7132_41;
typedef struct anon__lazy_58 nav__7133_41;
typedef struct anon__lazy_308 nav__7134_40;
typedef struct anon__lazy_110 nav__7134_42;
typedef struct anon__lazy_58 nav__7134_50;
typedef struct anon__lazy_333 nav__7135_39;
typedef struct anon__lazy_162 nav__7135_49;
typedef struct anon__lazy_58 nav__7137_41;
typedef struct anon__lazy_696 nav__7137_43;
typedef struct anon__lazy_58 nav__5770_39;
typedef struct anon__lazy_162 nav__6999_54;
typedef struct anon__lazy_58 nav__6999_56;
typedef struct anon__lazy_551 nav__7138_51;
typedef struct anon__lazy_162 nav__7138_57;
typedef struct anon__lazy_58 nav__7138_59;
typedef struct anon__lazy_162 nav__7001_55;
typedef struct anon__lazy_58 nav__7001_57;
typedef struct anon__lazy_162 nav__7139_54;
typedef struct anon__lazy_58 nav__7139_56;
typedef struct anon__lazy_58 nav__7140_39;
typedef struct anon__lazy_551 nav__7141_39;
typedef struct anon__lazy_162 nav__7141_42;
typedef struct anon__lazy_58 nav__7141_44;
typedef struct anon__lazy_551 nav__7142_39;
typedef struct anon__lazy_162 nav__7142_42;
typedef struct anon__lazy_58 nav__7142_44;
typedef struct anon__lazy_362 nav__7142_51;
typedef struct anon__lazy_362 nav__7143_41;
typedef struct anon__lazy_162 nav__7143_45;
typedef struct anon__lazy_58 nav__7143_47;
typedef struct anon__lazy_362 nav__7146_41;
typedef struct anon__lazy_162 nav__7146_45;
typedef struct anon__lazy_58 nav__7146_47;
typedef struct anon__lazy_596 nav__7146_56;
typedef struct anon__lazy_110 nav__7146_58;
typedef struct anon__lazy_362 nav__7145_41;
typedef struct anon__lazy_162 nav__7145_45;
typedef struct anon__lazy_58 nav__7145_47;
typedef struct anon__lazy_362 nav__7144_41;
typedef struct anon__lazy_162 nav__7144_45;
typedef struct anon__lazy_58 nav__7144_47;
struct Target_Os__639;
union Target_Os_VersionRange__695;
struct SemanticVersion_Range__700;
struct SemanticVersion__698;
typedef struct anon__lazy_58 nav__270_43;
struct SemanticVersion__698 {
 uintptr_t major;
 uintptr_t minor;
 uintptr_t patch;
 struct anon__lazy_58 pre;
 struct anon__lazy_58 build;
};
struct SemanticVersion_Range__700 {
 struct SemanticVersion__698 zig_e_min;
 struct SemanticVersion__698 zig_e_max;
};
struct Target_Os_HurdVersionRange__702;
struct Target_Os_HurdVersionRange__702 {
 struct SemanticVersion_Range__700 range;
 struct SemanticVersion__698 glibc;
};
struct Target_Os_LinuxVersionRange__704;
struct Target_Os_LinuxVersionRange__704 {
 struct SemanticVersion_Range__700 range;
 struct SemanticVersion__698 glibc;
 uint32_t android;
};
struct Target_Os_WindowsVersion_Range__760;
struct Target_Os_WindowsVersion_Range__760 {
 uint32_t zig_e_min;
 uint32_t zig_e_max;
};
union Target_Os_VersionRange__695 {
 struct SemanticVersion_Range__700 semver;
 struct Target_Os_HurdVersionRange__702 hurd;
 struct Target_Os_LinuxVersionRange__704 linux;
 struct Target_Os_WindowsVersion_Range__760 windows;
};
struct Target_Os__639 {
 union Target_Os_VersionRange__695 version_range;
 uint8_t tag;
};
struct Target_DynamicLinker__988;
struct Target_DynamicLinker__988 {
 uint8_t buffer[255];
 uint8_t len;
};
typedef struct anon__lazy_110 nav__4538_40;
typedef struct anon__lazy_110 nav__4539_40;
typedef struct anon__lazy_110 nav__4540_40;
typedef struct anon__lazy_110 nav__4502_46;
typedef struct anon__lazy_110 nav__235_46;
struct Target_Cpu_Feature_Set__1005;
struct Target_Cpu_Feature_Set__1005 {
 uintptr_t ints[5];
};
struct Target_Cpu__973;
struct Target_Cpu_Model__1000;
struct Target_Cpu__973 {
 struct Target_Cpu_Model__1000 const *model;
 struct Target_Cpu_Feature_Set__1005 features;
 uint8_t arch;
};
typedef struct anon__lazy_58 nav__269_46;
struct Target_Cpu_Model__1000 {
 struct anon__lazy_58 name;
 struct anon__lazy_58 llvm_name;
 struct Target_Cpu_Feature_Set__1005 features;
};
struct Target__637;
typedef struct anon__lazy_58 nav__271_51;
struct Target__637 {
 struct Target_Cpu__973 cpu;
 struct Target_Os__639 os;
 uint8_t abi;
 uint8_t ofmt;
 struct Target_DynamicLinker__988 dynamic_linker;
};
struct builtin_CallingConvention__266;
struct builtin_CallingConvention_CommonOptions__268;
typedef struct anon__lazy_248 nav__470_40;
struct builtin_CallingConvention_CommonOptions__268 {
 struct anon__lazy_248 incoming_stack_alignment;
};
struct builtin_CallingConvention_X86RegparmOptions__270;
struct builtin_CallingConvention_X86RegparmOptions__270 {
 struct anon__lazy_248 incoming_stack_alignment;
 uint8_t register_params;
};
struct builtin_CallingConvention_ArmInterruptOptions__272;
struct builtin_CallingConvention_ArmInterruptOptions__272 {
 struct anon__lazy_248 incoming_stack_alignment;
 uint8_t type;
};
struct builtin_CallingConvention_MipsInterruptOptions__274;
struct builtin_CallingConvention_MipsInterruptOptions__274 {
 struct anon__lazy_248 incoming_stack_alignment;
 uint8_t mode;
};
struct builtin_CallingConvention_RiscvInterruptOptions__276;
struct builtin_CallingConvention_RiscvInterruptOptions__276 {
 struct anon__lazy_248 incoming_stack_alignment;
 uint8_t mode;
};
struct builtin_CallingConvention__266 {
 union {
  struct builtin_CallingConvention_CommonOptions__268 x86_64_sysv;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_regcall_v3_sysv;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_regcall_v4_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_vectorcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_64_interrupt;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_sysv;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_win;
  struct builtin_CallingConvention_X86RegparmOptions__270 x86_stdcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_fastcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_thiscall;
  struct builtin_CallingConvention_CommonOptions__268 x86_thiscall_mingw;
  struct builtin_CallingConvention_CommonOptions__268 x86_regcall_v3;
  struct builtin_CallingConvention_CommonOptions__268 x86_regcall_v4_win;
  struct builtin_CallingConvention_CommonOptions__268 x86_vectorcall;
  struct builtin_CallingConvention_CommonOptions__268 x86_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs_darwin;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_aapcs_win;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_vfabi;
  struct builtin_CallingConvention_CommonOptions__268 aarch64_vfabi_sve;
  struct builtin_CallingConvention_CommonOptions__268 arm_aapcs;
  struct builtin_CallingConvention_CommonOptions__268 arm_aapcs_vfp;
  struct builtin_CallingConvention_ArmInterruptOptions__272 arm_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 mips64_n64;
  struct builtin_CallingConvention_CommonOptions__268 mips64_n32;
  struct builtin_CallingConvention_MipsInterruptOptions__274 mips64_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 mips_o32;
  struct builtin_CallingConvention_MipsInterruptOptions__274 mips_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 riscv64_lp64;
  struct builtin_CallingConvention_CommonOptions__268 riscv64_lp64_v;
  struct builtin_CallingConvention_RiscvInterruptOptions__276 riscv64_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 riscv32_ilp32;
  struct builtin_CallingConvention_CommonOptions__268 riscv32_ilp32_v;
  struct builtin_CallingConvention_RiscvInterruptOptions__276 riscv32_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 sparc64_sysv;
  struct builtin_CallingConvention_CommonOptions__268 sparc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf_altivec;
  struct builtin_CallingConvention_CommonOptions__268 powerpc64_elf_v2;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_sysv_altivec;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_aix;
  struct builtin_CallingConvention_CommonOptions__268 powerpc_aix_altivec;
  struct builtin_CallingConvention_CommonOptions__268 wasm_mvp;
  struct builtin_CallingConvention_CommonOptions__268 arc_sysv;
  struct builtin_CallingConvention_CommonOptions__268 bpf_std;
  struct builtin_CallingConvention_CommonOptions__268 csky_sysv;
  struct builtin_CallingConvention_CommonOptions__268 csky_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 hexagon_sysv;
  struct builtin_CallingConvention_CommonOptions__268 hexagon_sysv_hvx;
  struct builtin_CallingConvention_CommonOptions__268 lanai_sysv;
  struct builtin_CallingConvention_CommonOptions__268 loongarch64_lp64;
  struct builtin_CallingConvention_CommonOptions__268 loongarch32_ilp32;
  struct builtin_CallingConvention_CommonOptions__268 m68k_sysv;
  struct builtin_CallingConvention_CommonOptions__268 m68k_gnu;
  struct builtin_CallingConvention_CommonOptions__268 m68k_rtd;
  struct builtin_CallingConvention_CommonOptions__268 m68k_interrupt;
  struct builtin_CallingConvention_CommonOptions__268 msp430_eabi;
  struct builtin_CallingConvention_CommonOptions__268 propeller_sysv;
  struct builtin_CallingConvention_CommonOptions__268 s390x_sysv;
  struct builtin_CallingConvention_CommonOptions__268 s390x_sysv_vx;
  struct builtin_CallingConvention_CommonOptions__268 ve_sysv;
  struct builtin_CallingConvention_CommonOptions__268 xcore_xs1;
  struct builtin_CallingConvention_CommonOptions__268 xcore_xs2;
  struct builtin_CallingConvention_CommonOptions__268 xtensa_call0;
  struct builtin_CallingConvention_CommonOptions__268 xtensa_windowed;
  struct builtin_CallingConvention_CommonOptions__268 amdgcn_device;
  struct builtin_CallingConvention_CommonOptions__268 amdgcn_cs;
 } payload;
 uint8_t tag;
};
struct std_Options__4313;
typedef struct anon__lazy_362 nav__97_39;
struct std_Options__4313 {
 struct anon__lazy_362 page_size_min;
 struct anon__lazy_362 page_size_max;
 uintptr_t fmt_max_depth;
 bool enable_segfault_handler;
 uint8_t log_level;
 bool crypto_always_getrandom;
 bool crypto_fork_safety;
 bool keep_sigpipe;
 bool http_disable_tls;
 bool http_enable_ssl_key_log_file;
 uint8_t side_channels_mitigations;
};
typedef struct anon__lazy_362 nav__4518_38;
struct json_stringify_WriteStream__union_8283__8283;
struct json_stringify_WriteStream__union_8283__8283 {
 union {
  uintptr_t checked_to_fixed_depth;
 } payload;
 uint8_t tag;
};
typedef struct anon__lazy_110 nav__4529_43;
typedef struct anon__lazy_58 nav__4649_40;
static uint8_t const __anon_1704[21];
static uint8_t const __anon_1769[21];
static char const __anon_3403[33];
static uint8_t const __anon_2400[62];
static uint8_t const __anon_3484[34];
static uint8_t const __anon_4002[49];
static uint8_t const __anon_4385[1];
static uint8_t const __anon_4501[68];
static uint8_t const __anon_4624[82];
static uint8_t const __anon_4678[32];
static uint8_t const __anon_4811[22];
static uint8_t const __anon_2108[15];
static uint8_t const __anon_2113[14];
static uint8_t const __anon_2118[4];
static uint8_t const __anon_2123[13];
static uint8_t const __anon_2131[4];
static uint8_t const __anon_2146[19];
static uint8_t const __anon_5166[8];
static uint8_t const __anon_5345[3];
static uint8_t const __anon_5419[25];
static uint8_t const __anon_5485[7];
static uint8_t const __anon_5511[2];
static uint8_t const __anon_5564[23];
static uint8_t const __anon_5589[11];
static uint8_t const __anon_5645[17];
static uint8_t const __anon_5667[10];
static uint8_t const __anon_5737[31];
static uint8_t const __anon_5792[24];
static uint8_t const __anon_5908[23];
static uint8_t const __anon_5913[2];
static uint8_t const __anon_6005[12];
static uint8_t const __anon_6015[1];
static uint8_t const __anon_6034[10];
static uint8_t const __anon_6066[15];
static uint8_t const __anon_6094[13];
static uint8_t const __anon_7231[10];
static uint8_t const __anon_7246[1];
static uint8_t const __anon_7259[1];
static uint8_t const __anon_8103[24];
static uint8_t const __anon_8178[11];
static uint8_t const __anon_8255[13];
static uint8_t const __anon_8272[5];
static uint8_t const __anon_8445[5];
static uint8_t const __anon_8449[17];
static uint8_t const __anon_8550[61];
static uint8_t const __anon_8644[62];
static uint8_t const __anon_8711[55];
static uint8_t const __anon_8724[24];
static uint8_t const __anon_8732[17];
static uint8_t const __anon_8883[63];
static uint8_t const __anon_8956[59];
static uint8_t const __anon_9166[3];
static uint8_t const __anon_9125[3];
static uint8_t const __anon_9130[3];
static uint8_t const __anon_9134[3];
static uint8_t const __anon_9138[3];
static uint8_t const __anon_9142[3];
static uint8_t const __anon_9146[3];
static uint8_t const __anon_9150[3];
static uint8_t const __anon_9366[5];
static uint8_t const __anon_9370[9];
static uint8_t const __anon_9483[5];
static uint8_t const __anon_5003[4];
static uint8_t const __anon_9580[7];
static uint8_t const __anon_9582[7];
#define c_instrument_hooks_set_feature__237 instrument_hooks_set_feature
zig_extern void instrument_hooks_set_feature(uint64_t, bool);
static void features_set_feature__330(uint64_t, bool);
static void bit_set_IntegerBitSet_2864_29_set__369(uint64_t *, uintptr_t);
static void bit_set_IntegerBitSet_2864_29_unset__371(uint64_t *, uintptr_t);
static void debug_assert__180(bool);
static uint64_t bit_set_IntegerBitSet_2864_29_maskBit__390(uintptr_t);
static nav__786_39 instruments_valgrind_ValgrindInstrument_init__786(struct mem_Allocator__599);
static nav__1280_39 fifo_posix_Pipe_openPipe__1280(nav__1280_41);
static nav__1282_39 fifo_posix_Pipe_openWrite__1282(struct mem_Allocator__599, nav__1282_42);
static nav__1281_39 fifo_posix_Pipe_openRead__1281(struct mem_Allocator__599, nav__1281_42);
static nav__1129_39 runner_fifo_RunnerFifo_init__1129(struct mem_Allocator__599);
static uint16_t fifo_posix_Pipe_Writer_sendCmd__1286(struct fifo_posix_Pipe_Writer__791 *, struct shared_Command__1795);
static uint16_t fifo_posix_waitReadable__1268(int32_t, uint64_t);
static nav__5031_38 bincode_deserializeInt__anon_2223__5031(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089);
static nav__5030_38 bincode_deserializeAlloc__anon_2212__5030(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5073_38 bincode_deserializeInt__anon_2332__5073(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089);
static nav__5072_38 bincode_deserializeAlloc__anon_2327__5072(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5075_40 bincode_deserializePointerAlloc__anon_2362__5075(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5074_40 bincode_deserializeAlloc__anon_2349__5074(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5071_39 bincode_deserializeStructAlloc__anon_2311__5071(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5070_39 bincode_deserializeAlloc__anon_2282__5070(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static uint16_t bincode_deserializeAlloc__anon_2422__5079(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5081_39 bincode_deserializeStructAlloc__anon_2488__5081(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5080_39 bincode_deserializeAlloc__anon_2469__5080(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5087_38 bincode_deserializeInt__anon_2610__5087(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089);
static nav__5086_38 bincode_deserializeAlloc__anon_2605__5086(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5085_39 bincode_deserializeUnionAlloc__anon_2592__5085(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5084_39 bincode_deserializeAlloc__anon_2559__5084(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5083_39 bincode_deserializeStructAlloc__anon_2541__5083(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5082_39 bincode_deserializeAlloc__anon_2527__5082(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5089_38 bincode_deserializeEnum__anon_2695__5089(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089);
static nav__5088_38 bincode_deserializeAlloc__anon_2665__5088(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5029_39 bincode_deserializeUnionAlloc__anon_2207__5029(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__5028_39 bincode_deserializeAlloc__anon_2097__5028(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089, struct mem_Allocator__599);
static nav__1291_39 fifo_posix_Pipe_Reader_recvCmd__1291(struct fifo_posix_Pipe_Reader__793 *);
static nav__1292_39 fifo_posix_Pipe_Reader_waitForResponse__1292(struct fifo_posix_Pipe_Reader__793 *, nav__1292_42);
static uint16_t fifo_posix_Pipe_Reader_waitForAck__1293(struct fifo_posix_Pipe_Reader__793 *, nav__1293_40);
static uint16_t runner_fifo_RunnerFifo_send_version__1139(struct runner_fifo_RunnerFifo__631 *, uint64_t);
static uint16_t runner_fifo_RunnerFifo_validate_protocol_version__1130(struct runner_fifo_RunnerFifo__631 *);
static nav__1140_38 runner_fifo_RunnerFifo_get_integration_mode__1140(struct runner_fifo_RunnerFifo__631 *);
static nav__820_39 instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__820(struct mem_Allocator__599);
static nav__807_39 instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_init__807(struct mem_Allocator__599);
static nav__743_39 instruments_root_Instrument_init__743(struct mem_Allocator__599);
static nav__730_39 instrument_hooks_InstrumentHooks_init__730(struct mem_Allocator__599);
#define c_instrument_hooks_init__238 instrument_hooks_init
zig_extern struct InstrumentHooks *instrument_hooks_init(void);
static nav__4541_40 mem_Allocator_create__anon_1051__4541(struct mem_Allocator__599);
static uint16_t fs_accessAbsolute__1348(nav__1348_39, struct fs_File_OpenFlags__1718);
static nav__1345_39 fs_openFileAbsolute__1345(nav__1345_41, struct fs_File_OpenFlags__1718);
static struct fifo_posix_Pipe_Writer__791 fifo_posix_Pipe_Writer_init__1283(struct fs_File__799, struct mem_Allocator__599);
static struct fifo_posix_Pipe_Reader__793 fifo_posix_Pipe_Reader_init__1288(struct fs_File__799, struct mem_Allocator__599);
static void array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1562(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *);
static struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *);
static uint16_t bincode_serialize__anon_1849__4913(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__1795);
static uint8_t const (*mem_asBytes__anon_1887__4914(uint32_t const *))[4];
static uint16_t fs_File_writeAll__1439(struct fs_File__799, nav__1439_40);
static nav__1964_38 posix_poll__1964(nav__1964_41, int32_t);
static nav__1429_38 fs_File_readAll__1429(struct fs_File__799, nav__1429_41);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_resize__1559(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, uintptr_t);
static struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(nav__4994_40);
static struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4983(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *);
static nav__5027_38 io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027(void const *, nav__5027_41);
static nav__5069_39 io_Reader_readBytesNoEof__anon_2269__5069(struct io_Reader__2238);
static nav__5076_39 io_Reader_readBytesNoEof__anon_2375__5076(struct io_Reader__2238);
static nav__5077_40 mem_Allocator_alloc__anon_2384__5077(struct mem_Allocator__599, uintptr_t);
static nav__5034_38 io_Reader_readAll__5034(struct io_Reader__2238, nav__5034_41);
static zig_noreturn void bincode_invalidProtocol__anon_2407__5078(void);
static void shared_Command_deinit__4786(struct shared_Command__1795, struct mem_Allocator__599);
static void logger_debug__anon_2729__5099(nav__5099_39);
static void logger_err__anon_2753__5100(void);
static void logger_err__anon_2762__5101(void);
static zig_noreturn void posix_exit__1747(uint8_t);
static void logger_err__anon_2776__5102(nav__5102_40);
static void runner_fifo_RunnerFifo_deinit__1131(struct runner_fifo_RunnerFifo__631 *);
static struct environment_root_Environment__559 environment_root_Environment_init__773(struct mem_Allocator__599);
static nav__5103_39 mem_Allocator_allocBytesWithAlignment__anon_2823__5103(struct mem_Allocator__599, uintptr_t, uintptr_t);
static bool fs_path_isAbsolute__5138(nav__5138_39);
static struct fs_Dir__1720 fs_cwd__1340(void);
static uint16_t fs_Dir_access__4742(struct fs_Dir__1720, nav__4742_40, struct fs_File_OpenFlags__1718);
static nav__4678_39 fs_Dir_openFile__4678(struct fs_Dir__1720, nav__4678_42, struct fs_File_OpenFlags__1718);
static struct array_list_ArrayListAligned_28u8_2cnull_29__811 array_list_ArrayListAligned_28u8_2cnull_29_init__1527(struct mem_Allocator__599);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, uintptr_t);
static uint16_t bincode_serializeUnion__anon_2865__5201(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__1795);
static nav__1438_38 fs_File_write__1438(struct fs_File__799, nav__1438_41);
static nav__5202_38 math_cast__anon_2885__5202(uintptr_t);
static uint16_t posix_errno__anon_3044__5205(int);
static uint16_t posix_unexpectedErrno__2021(uint16_t);
static nav__1428_38 fs_File_read__1428(struct fs_File__799, nav__1428_41);
static nav__4986_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4986(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *, nav__4986_42);
static uint16_t io_Reader_readNoEof__5036(struct io_Reader__2238, nav__5036_40);
static nav__5206_39 mem_Allocator_allocWithSizeAndAlignment__anon_3080__5206(struct mem_Allocator__599, uintptr_t, uintptr_t);
static nav__5035_38 io_Reader_readAtLeast__5035(struct io_Reader__2238, nav__5035_41, uintptr_t);
static void utils_print__anon_3102__5217(void);
static void mem_Allocator_free__anon_3116__5296(struct mem_Allocator__599, nav__5296_40);
static void logger_logWithPrefix__anon_3124__5297(nav__5297_39);
static void logger_logWithPrefix__anon_3126__5298(void);
static void logger_logWithPrefix__anon_3127__5299(void);
static void logger_logWithPrefix__anon_3132__5301(nav__5301_40);
static void fifo_posix_Pipe_Writer_deinit__1287(struct fifo_posix_Pipe_Writer__791 *);
static void fifo_posix_Pipe_Reader_deinit__1294(struct fifo_posix_Pipe_Reader__793 *);
static struct environment_linked_libraries_root_LinkedLibraries__849 environment_linked_libraries_root_LinkedLibraries_init__3643(struct mem_Allocator__599);
static uint8_t mem_Alignment_fromByteUnits__1113(uintptr_t);
static bool fs_path_isAbsolutePosix__5144(nav__5144_39);
static nav__2022_39 posix_toPosixPath__2022(nav__2022_41);
static uint16_t fs_Dir_accessZ__4743(struct fs_Dir__1720, uint8_t const *, struct fs_File_OpenFlags__1718);
static nav__4679_39 fs_Dir_openFileZ__4679(struct fs_Dir__1720, uint8_t const *, struct fs_File_OpenFlags__1718);
static uintptr_t array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__5372(uintptr_t, uintptr_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1565(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, uintptr_t);
static uint16_t bincode_serialize__anon_3242__5373(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, uint32_t);
static uint16_t bincode_serialize__anon_3244__5374(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__struct_1798__1798);
static uint16_t bincode_serialize__anon_3246__5375(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t bincode_serialize__anon_3248__5376(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__struct_1799__1799);
static uint16_t bincode_serialize__anon_3250__5377(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__struct_1800__1800);
static uint16_t bincode_serialize__anon_3252__5378(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, uint64_t);
static uint16_t bincode_serialize__anon_3254__5379(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, uint8_t);
static nav__1757_38 posix_write__1757(int32_t, nav__1757_40);
static nav__1749_38 posix_read__1749(int32_t, nav__1749_40);
static nav__5445_38 math_mul__anon_3296__5445(uintptr_t, uintptr_t);
static nav__5446_39 mem_Allocator_allocBytesWithAlignment__anon_3298__5446(struct mem_Allocator__599, uintptr_t, uintptr_t);
static nav__5033_38 io_Reader_read__5033(struct io_Reader__2238, nav__5033_41);
static nav__5448_39 mem_sliceAsBytes__anon_3427__5448(nav__5448_39);
static uint16_t fmt_format__anon_3892__5741(struct io_Writer__3878, nav__5741_40);
static nav__5608_40 fmt_bufPrint__anon_3480__5608(nav__5608_39, nav__5608_42);
static uint16_t fmt_format__anon_4401__6109(struct io_Writer__3878);
static nav__5610_40 fmt_bufPrint__anon_3502__5610(nav__5610_39);
static uint16_t fmt_format__anon_4518__6111(struct io_Writer__3878);
static nav__5611_40 fmt_bufPrint__anon_3515__5611(nav__5611_39);
static uint16_t fmt_format__anon_4630__6112(struct io_Writer__3878, nav__6112_41);
static nav__5612_40 fmt_bufPrint__anon_3528__5612(nav__5612_39, nav__5612_43);
static void array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(struct array_list_ArrayListAligned_28u8_2cnull_29__811);
static void fs_File_close__1385(struct fs_File__799);
static struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_init__4347(struct mem_Allocator__599);
static bool math_isPowerOfTwo__anon_3550__5613(uintptr_t);
static uint16_t posix_faccessatZ__1913(int32_t, uint8_t const *, uint32_t, uint32_t);
static nav__1769_38 posix_openatZ__1769(int32_t, uint8_t const *, uint32_t, uintptr_t);
static uint16_t posix_flock__1930(int32_t, int32_t);
static void posix_close__1727(int32_t);
static nav__1575_39 array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(struct array_list_ArrayListAligned_28u8_2cnull_29__811);
static nav__5700_39 mem_Allocator_remap__anon_3698__5700(struct mem_Allocator__599, nav__5700_39, uintptr_t);
static nav__5701_40 mem_Allocator_alignedAlloc__anon_3703__5701(struct mem_Allocator__599, uintptr_t);
static void mem_Allocator_free__anon_3705__5702(struct mem_Allocator__599, nav__5702_40);
static uint16_t bincode_serializeInt__anon_3710__5703(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, uint32_t);
static uint16_t bincode_serializeStruct__anon_3715__5704(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__struct_1798__1798);
static uint16_t bincode_serializeStruct__anon_3716__5705(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__struct_1799__1799);
static uint16_t bincode_serializeStruct__anon_3717__5706(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_Command__struct_1800__1800);
static uint16_t bincode_serializeInt__anon_3718__5707(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, uint64_t);
static uint16_t bincode_serializeEnum__anon_3721__5708(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, uint8_t);
static uint16_t posix_errno__anon_3732__5710(intptr_t);
static struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *);
static nav__5724_38 io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724(void const *, nav__5724_41);
static uint16_t io_Writer_writeAll__5732(struct io_Writer__3878, nav__5732_40);
static uint16_t shared_Command_format__anon_5054__6122(struct shared_Command__1795, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatType__anon_4378__6108(struct shared_Command__1795, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static nav__4992_39 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065);
static uint16_t fmt_formatType__anon_4509__6110(uint64_t, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static uint16_t fmt_formatType__anon_4782__6113(nav__6113_39, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static nav__6117_39 mem_sliceAsBytes__anon_4942__6117(nav__6117_39);
static nav__6118_39 mem_bytesAsSlice__anon_4955__6118(nav__6118_39);
static nav__4912_38 io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912(void const *, nav__4912_41);
static uint16_t bincode_serialize__anon_4986__6119(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, int32_t);
static uint16_t bincode_serialize__anon_4988__6120(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, nav__6120_40);
static uint16_t bincode_serialize__anon_4990__6121(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_MarkerType__1802);
static nav__4987_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4987(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *, nav__4987_42);
static nav__5731_38 io_Writer_write__5731(struct io_Writer__3878, nav__5731_41);
static uint16_t io_Writer_print__anon_5074__6123(struct io_Writer__3878, nav__6123_41);
static uint16_t io_Writer_print__anon_5096__6124(struct io_Writer__3878, nav__6124_41);
static uint16_t io_Writer_print__anon_5108__6125(struct io_Writer__3878, nav__6125_40);
static uint16_t io_Writer_print__anon_5117__6126(struct io_Writer__3878, nav__6126_39);
static uint16_t io_Writer_print__anon_5130__6127(struct io_Writer__3878, nav__6127_39);
static nav__5753_38 unicode_utf8ByteSequenceLength__5753(uint8_t);
static nav__5769_38 unicode_utf8CountCodepoints__5769(nav__5769_40);
static uint16_t fmt_formatBuf__anon_5176__6129(nav__6129_39, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatInt__anon_5237__6139(uint64_t, uint8_t, uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatIntValue__anon_5220__6138(uint64_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatValue__anon_5151__6128(uint64_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static nav__1553_38 array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1553(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, nav__1553_42);
static uint16_t bincode_serializeInt__anon_5184__6130(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, int32_t);
static uint16_t bincode_serializePointer__anon_5189__6131(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, nav__6131_40);
static uint16_t bincode_serializeUnion__anon_5190__6132(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct shared_MarkerType__1802);
static uint16_t fmt_format__anon_5194__6133(struct io_Writer__3878, nav__6133_41);
static uint16_t fmt_format__anon_5198__6134(struct io_Writer__3878, nav__6134_41);
static uint16_t fmt_format__anon_5202__6135(struct io_Writer__3878, nav__6135_40);
static uint16_t fmt_format__anon_5206__6136(struct io_Writer__3878, nav__6136_39);
static uint16_t fmt_format__anon_5210__6137(struct io_Writer__3878, nav__6137_39);
static nav__5509_39 fmt_digits2__5509(uint8_t);
static uint8_t fmt_digitToChar__5525(uint8_t, uint8_t);
static nav__5759_38 unicode_utf8Decode__5759(nav__5759_40);
static nav__5752_38 unicode_utf8CodepointSequenceLength__5752(uint32_t);
static nav__6145_38 unicode_utf8EncodeImpl__anon_5823__6145(uint32_t, nav__6145_40);
static nav__5754_38 unicode_utf8Encode__5754(uint32_t, nav__5754_40);
static uint16_t io_Writer_writeBytesNTimes__5736(struct io_Writer__3878, nav__5736_40, uintptr_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1547(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, nav__1547_41);
static uint16_t fmt_formatType__anon_5471__6140(int32_t, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static uint16_t fmt_formatType__anon_5502__6141(nav__6141_39, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static uint16_t shared_MarkerType_format__anon_5867__6147(struct shared_MarkerType__1802, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatType__anon_5677__6142(struct shared_MarkerType__1802, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static uint16_t fmt_formatType__anon_5746__6143(uint64_t, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static uint16_t fmt_formatType__anon_5800__6144(uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static nav__5761_38 unicode_utf8Decode2__5761(nav__5761_40);
static nav__5763_38 unicode_utf8Decode3__5763(nav__5763_40);
static nav__5767_38 unicode_utf8Decode4__5767(nav__5767_40);
static bool unicode_isSurrogateCodepoint__5832(uint32_t);
static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1566(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, uintptr_t);
static void array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1548(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *, nav__1548_41);
static uint16_t fmt_formatInt__anon_5955__6154(int32_t, uint8_t, uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatIntValue__anon_5944__6153(int32_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatValue__anon_5848__6146(int32_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t io_Writer_print__anon_5881__6148(struct io_Writer__3878, nav__6148_39);
static uint16_t io_Writer_print__anon_5886__6149(struct io_Writer__3878, nav__6149_39);
static uint16_t io_Writer_print__anon_5893__6150(struct io_Writer__3878, nav__6150_39);
static uint16_t io_Writer_print__anon_5899__6151(struct io_Writer__3878, nav__6151_39);
static uint16_t fmt_formatIntValue__anon_5982__6159(uint64_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatValue__anon_5901__6152(uint64_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static nav__5765_38 unicode_utf8Decode3AllowSurrogateHalf__5765(nav__5765_40);
static nav__1491_38 array_list_addOrOom__1491(uintptr_t, uintptr_t);
static uint16_t fmt_format__anon_5966__6155(struct io_Writer__3878, nav__6155_39);
static uint16_t fmt_format__anon_5970__6156(struct io_Writer__3878, nav__6156_39);
static uint16_t fmt_format__anon_5974__6157(struct io_Writer__3878, nav__6157_39);
static uint16_t fmt_format__anon_5978__6158(struct io_Writer__3878, nav__6158_39);
#define c_instrument_hooks_deinit__239 instrument_hooks_deinit
zig_extern void instrument_hooks_deinit(struct InstrumentHooks *);
static void instrument_hooks_InstrumentHooks_deinit__731(struct InstrumentHooks *);
static void mem_Allocator_destroy__anon_6116__6160(struct mem_Allocator__599, struct InstrumentHooks *);
static void instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_deinit__808(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *);
static void instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__821(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *);
static void environment_root_Environment_deinit__774(struct environment_root_Environment__559 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3800(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880);
static nav__6161_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6161(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4011(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901);
static nav__6163_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6163(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 *);
static void environment_root_Environment_freeEnvironmentValue__777(struct environment_root_Environment__559 *, struct environment_root_EnvironmentValue__856);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4003(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct mem_Allocator__599);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3792(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct mem_Allocator__599);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4190(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934);
static nav__6165_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6165(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4182(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct mem_Allocator__599);
static void environment_linked_libraries_root_LinkedLibraries_deinit__3644(struct environment_linked_libraries_root_LinkedLibraries__849 *);
static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889);
static nav__6180_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231);
static nav__6181_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231);
static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906);
static nav__6189_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380);
static nav__6190_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380);
static void mem_Allocator_free__anon_6426__6191(struct mem_Allocator__599, nav__6191_42);
static void debug_SafetyLock_assertUnlocked__3984(void);
static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4129(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *, struct mem_Allocator__599);
static void array_hash_map_IndexHeader_free__3979(struct array_hash_map_IndexHeader__892 *, struct mem_Allocator__599);
static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3935(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *, struct mem_Allocator__599);
static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939);
static nav__6199_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459);
static nav__6200_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459);
static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4308(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *, struct mem_Allocator__599);
static void environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__3645(struct mem_Allocator__599, struct environment_linked_libraries_root_LibraryInfo__946);
static void array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_deinit__4349(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948);
static nav__6272_39 mem_sliceAsBytes__anon_6643__6272(nav__6272_41);
static nav__4161_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4161(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906);
static void mem_Allocator_free__anon_6654__6273(struct mem_Allocator__599, nav__6273_40);
static uintptr_t array_hash_map_capacityIndexSize__3742(uint8_t);
static uintptr_t array_hash_map_IndexHeader_length__3975(struct array_hash_map_IndexHeader__892);
static void mem_Allocator_free__anon_6672__6274(struct mem_Allocator__599, nav__6274_40);
static nav__3967_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3967(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889);
static nav__4340_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4340(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939);
static nav__4395_40 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__4395(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948);
static void mem_Allocator_free__anon_6701__6275(struct mem_Allocator__599, nav__6275_41);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4160(uintptr_t);
static nav__6276_39 mem_sliceAsBytes__anon_6720__6276(nav__6276_39);
static uint8_t array_hash_map_capacityIndexType__3741(uint8_t);
static nav__6292_39 mem_sliceAsBytes__anon_6754__6292(nav__6292_39);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3966(uintptr_t);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4339(uintptr_t);
static nav__6293_39 mem_sliceAsBytes__anon_6767__6293(nav__6293_42);
#define c_instrument_hooks_is_instrumented__240 instrument_hooks_is_instrumented
zig_extern bool instrument_hooks_is_instrumented(struct InstrumentHooks *);
static zig_cold uint16_t runner_fifo_RunnerFifo_start_benchmark__1134(struct runner_fifo_RunnerFifo__631 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_start_benchmark__810(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__823(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *);
#define c_instrument_hooks_start_benchmark__241 instrument_hooks_start_benchmark
zig_extern uint8_t instrument_hooks_start_benchmark(struct InstrumentHooks *);
static bool features_is_feature_enabled__331(uint64_t);
static bool bit_set_IntegerBitSet_2864_29_isSet__366(uint64_t, uintptr_t);
static zig_cold uint16_t runner_fifo_RunnerFifo_stop_benchmark__1135(struct runner_fifo_RunnerFifo__631 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_stop_benchmark__811(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__824(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *);
#define c_instrument_hooks_stop_benchmark__242 instrument_hooks_stop_benchmark
zig_extern uint8_t instrument_hooks_stop_benchmark(struct InstrumentHooks *);
static uint16_t runner_fifo_RunnerFifo_set_executed_benchmark__1136(struct runner_fifo_RunnerFifo__631 *, int32_t, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_executed_benchmark__812(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *, int32_t, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__825(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *, int32_t, uint8_t const *);
#define c_instrument_hooks_set_executed_benchmark__243 instrument_hooks_set_executed_benchmark
zig_extern uint8_t instrument_hooks_set_executed_benchmark(struct InstrumentHooks *, int32_t, char const *);
static uint8_t const *c_toU8__236(char const *);
static nav__6298_39 mem_span__anon_6923__6298(uint8_t const *);
static uintptr_t mem_len__anon_6940__6299(uint8_t const *);
static uintptr_t mem_indexOfSentinel__anon_6947__6300(uint8_t const *);
static bool math_isPowerOfTwo__anon_6958__6301(void);
static nav__6302_38 simd_firstTrue__anon_7009__6302(nav__6302_40);
static uintptr_t mem_alignForward__anon_7012__6303(uintptr_t, uintptr_t);
static bool mem_isAligned__982(uintptr_t, uintptr_t);
static bool mem_isValidAlignGeneric__anon_7054__6304(uintptr_t);
static uintptr_t mem_alignBackward__anon_7055__6305(uintptr_t, uintptr_t);
static bool mem_isAlignedGeneric__anon_7059__6306(uint64_t, uint64_t);
static uint64_t mem_alignBackward__anon_7061__6307(uint64_t, uint64_t);
static bool mem_isValidAlignGeneric__anon_7063__6308(uint64_t);
static bool math_isPowerOfTwo__anon_7064__6309(uint64_t);
#define c_instrument_hooks_executed_benchmark__244 instrument_hooks_executed_benchmark
zig_extern uint8_t instrument_hooks_executed_benchmark(struct InstrumentHooks *, int32_t, char const *);
static uint16_t runner_fifo_RunnerFifo_set_integration__1137(struct runner_fifo_RunnerFifo__631 *, uint8_t const *, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_integration__813(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *, uint8_t const *, uint8_t const *);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__826(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *, uint8_t const *, uint8_t const *);
#define c_instrument_hooks_set_integration__245 instrument_hooks_set_integration
zig_extern uint8_t instrument_hooks_set_integration(struct InstrumentHooks *, char const *, char const *);
static nav__6310_40 fmt_allocPrintZ__anon_7099__6310(struct mem_Allocator__599, nav__6310_43);
static void mem_Allocator_free__anon_7101__6311(struct mem_Allocator__599, nav__6311_40);
static nav__6312_40 fmt_allocPrint__anon_7136__6312(struct mem_Allocator__599, nav__6312_43);
static nav__6313_39 mem_sliceAsBytes__anon_7143__6313(nav__6313_39);
static uint16_t fmt_format__anon_7206__6355(struct io_Writer__3878, nav__6355_40);
static uint64_t fmt_count__anon_7150__6314(nav__6314_39);
static nav__6315_38 math_cast__anon_7152__6315(uint64_t);
static nav__6316_40 fmt_bufPrint__anon_7155__6316(nav__6316_39, nav__6316_42);
static struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 io_counting_writer_countingWriter__anon_7176__6341(void);
static struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__6340(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *);
static nav__6354_38 io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__6354(void const *, nav__6354_41);
static uint16_t fmt_formatType__anon_7240__6356(uint8_t const *, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static nav__6339_38 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__6339(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *, nav__6339_42);
static nav__4890_38 io_dummyWrite__4890(nav__4890_40);
static uint16_t runner_fifo_RunnerFifo_add_marker__1138(struct runner_fifo_RunnerFifo__631 *, int32_t, struct shared_MarkerType__1802);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_add_marker__814(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *, int32_t, struct shared_MarkerType__1802);
static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__827(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *, int32_t, struct shared_MarkerType__1802);
#define c_instrument_hooks_add_marker__250 instrument_hooks_add_marker
zig_extern uint8_t instrument_hooks_add_marker(struct InstrumentHooks *, int32_t, uint8_t, uint64_t);
#define c_instrument_hooks_current_timestamp__254 instrument_hooks_current_timestamp
zig_extern uint64_t instrument_hooks_current_timestamp(void);
static nav__1938_39 posix_clock_gettime__1938(uint32_t);
static uint16_t environment_root_Environment_setIntegrationEnvironmentValue__778(struct environment_root_Environment__559 *, nav__778_41, nav__778_41, struct environment_root_EnvironmentValue__856);
static uint16_t environment_root_Environment_setIntegrationEnvironment__775(struct environment_root_Environment__559 *, nav__775_41, nav__775_41, nav__775_41);
#define c_instrument_hooks_set_environment__255 instrument_hooks_set_environment
zig_extern uint8_t instrument_hooks_set_environment(struct InstrumentHooks *, char const *, char const *, char const *);
static nav__6358_40 mem_Allocator_dupe__anon_7358__6358(struct mem_Allocator__599, nav__6358_43);
static nav__3802_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3802(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct mem_Allocator__599, nav__3802_44);
static nav__4013_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4013(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct mem_Allocator__599, nav__4013_44);
static nav__3803_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3803(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct mem_Allocator__599, nav__3803_44);
static nav__4014_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4014(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct mem_Allocator__599, nav__4014_44);
static nav__6359_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6359(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct mem_Allocator__599, nav__6359_44);
static nav__6360_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6360(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct mem_Allocator__599, nav__6360_44);
static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3812(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct mem_Allocator__599, uintptr_t);
static nav__6361_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6361(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6361_41);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6362(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, nav__6362_42);
static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4023(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct mem_Allocator__599, uintptr_t);
static nav__6363_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6363(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6363_41);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6364(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, nav__6364_42);
static void debug_SafetyLock_lock__3982(void *);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3959(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *, struct mem_Allocator__599, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3955(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *, struct mem_Allocator__599, uintptr_t);
static void debug_SafetyLock_unlock__3983(void *);
static uint32_t array_hash_map_IndexHeader_capacity__3974(struct array_hash_map_IndexHeader__892);
static nav__3977_38 array_hash_map_IndexHeader_findBitIndex__3977(uintptr_t);
static nav__3978_40 array_hash_map_IndexHeader_alloc__3978(struct mem_Allocator__599, uint8_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3904(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct array_hash_map_IndexHeader__892 *);
static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(nav__6365_40);
static nav__6366_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231);
static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(nav__6367_40, nav__6367_40, uintptr_t);
static uint8_t array_hash_map_IndexHeader_capacityIndexType__3973(struct array_hash_map_IndexHeader__892);
static nav__6368_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6368(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6368_41, struct array_hash_map_IndexHeader__892 *);
static nav__6369_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6369(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6369_41, struct array_hash_map_IndexHeader__892 *);
static nav__6370_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6370(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6370_41, struct array_hash_map_IndexHeader__892 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6371(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, nav__6371_42, struct array_hash_map_IndexHeader__892 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6372(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, nav__6372_42, struct array_hash_map_IndexHeader__892 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6373(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, nav__6373_42, struct array_hash_map_IndexHeader__892 *);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4153(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *, struct mem_Allocator__599, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4149(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *, struct mem_Allocator__599, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4115(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct array_hash_map_IndexHeader__892 *);
static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(nav__6374_40);
static nav__6375_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380);
static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(nav__6376_40, nav__6376_40, uintptr_t);
static nav__6377_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6377(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6377_41, struct array_hash_map_IndexHeader__892 *);
static nav__6378_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6378(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6378_41, struct array_hash_map_IndexHeader__892 *);
static nav__6379_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6379(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6379_41, struct array_hash_map_IndexHeader__892 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6380(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, nav__6380_42, struct array_hash_map_IndexHeader__892 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6381(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, nav__6381_42, struct array_hash_map_IndexHeader__892 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6382(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, nav__6382_42, struct array_hash_map_IndexHeader__892 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3957(uintptr_t, uintptr_t);
static nav__6383_40 mem_Allocator_alignedAlloc__anon_7569__6383(struct mem_Allocator__599, uintptr_t);
static uint8_t math_log2_int_ceil__anon_7722__6384(uintptr_t);
static nav__6385_40 mem_Allocator_alignedAlloc__anon_7727__6385(struct mem_Allocator__599, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6386(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct array_hash_map_IndexHeader__892 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6387(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct array_hash_map_IndexHeader__892 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6388(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *, struct array_hash_map_IndexHeader__892 *);
static uint32_t array_hash_map_StringContext_hash__3773(nav__3773_40);
static bool array_hash_map_StringContext_eql__3774(nav__3774_40, nav__3774_40, uintptr_t);
static nav__6389_40 array_hash_map_IndexHeader_indexes__anon_7759__6389(struct array_hash_map_IndexHeader__892 *);
static nav__6390_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6390(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6390_41, struct array_hash_map_IndexHeader__892 *, nav__6390_47);
static nav__6391_40 array_hash_map_IndexHeader_indexes__anon_7769__6391(struct array_hash_map_IndexHeader__892 *);
static nav__6392_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6392(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6392_41, struct array_hash_map_IndexHeader__892 *, nav__6392_47);
static nav__6393_40 array_hash_map_IndexHeader_indexes__anon_7775__6393(struct array_hash_map_IndexHeader__892 *);
static nav__6394_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6394(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880, nav__6394_41, struct array_hash_map_IndexHeader__892 *, nav__6394_47);
static uintptr_t array_hash_map_safeTruncate__anon_7783__6395(uint32_t);
static uintptr_t array_hash_map_IndexHeader_constrainIndex__3971(struct array_hash_map_IndexHeader__892, uintptr_t);
static bool array_hash_map_Index_28u8_29_isEmpty__6280(struct array_hash_map_Index_28u8_29__6735);
static bool array_hash_map_Index_28u16_29_isEmpty__6285(struct array_hash_map_Index_28u16_29__6739);
static bool array_hash_map_Index_28u32_29_isEmpty__6290(struct array_hash_map_Index_28u32_29__6743);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4151(uintptr_t, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6396(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct array_hash_map_IndexHeader__892 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6397(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct array_hash_map_IndexHeader__892 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6398(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *, struct array_hash_map_IndexHeader__892 *);
static nav__6399_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6399(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6399_41, struct array_hash_map_IndexHeader__892 *, nav__6399_47);
static nav__6400_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6400(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6400_41, struct array_hash_map_IndexHeader__892 *, nav__6400_47);
static nav__6401_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6401(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901, nav__6401_41, struct array_hash_map_IndexHeader__892 *, nav__6401_47);
static nav__6402_39 mem_Allocator_allocWithSizeAndAlignment__anon_7842__6402(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint8_t math_log2_int__anon_7845__6403(uintptr_t);
static nav__6404_39 mem_Allocator_allocWithSizeAndAlignment__anon_7849__6404(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint32_t array_hash_map_hashString__3736(nav__3736_39);
static bool array_hash_map_eqlString__3735(nav__3735_39, nav__3735_39);
static uint32_t array_hash_map_IndexHeader_mask__3976(struct array_hash_map_IndexHeader__892);
static nav__6405_39 mem_Allocator_allocBytesWithAlignment__anon_7875__6405(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint64_t hash_wyhash_Wyhash_hash__6465(uint64_t, nav__6465_39);
static bool mem_eql__anon_7887__6466(nav__6466_39, nav__6466_39);
static struct hash_wyhash_Wyhash__7880 hash_wyhash_Wyhash_init__6453(uint64_t);
static bool mem_eqlBytes__856(nav__856_39, nav__856_39);
static uint16_t environment_root_Environment_setIntegrationEnvironmentList__776(struct environment_root_Environment__559 *, nav__776_41, nav__776_41, nav__776_43);
#define c_instrument_hooks_set_environment_list__256 instrument_hooks_set_environment_list
zig_extern uint8_t instrument_hooks_set_environment_list(struct InstrumentHooks *, char const *, char const *, char const *const *, uint32_t);
static nav__6470_42 mem_Allocator_alloc__anon_7987__6470(struct mem_Allocator__599, uintptr_t);
static void mem_Allocator_free__anon_7990__6471(struct mem_Allocator__599, nav__6471_42);
static nav__6472_39 mem_Allocator_allocWithSizeAndAlignment__anon_7998__6472(struct mem_Allocator__599, uintptr_t, uintptr_t);
static nav__6473_39 mem_sliceAsBytes__anon_8005__6473(nav__6473_43);
#define c_instrument_hooks_write_environment__257 instrument_hooks_write_environment
zig_extern uint8_t instrument_hooks_write_environment(struct InstrumentHooks *, int32_t);
static uint16_t environment_linked_libraries_root_LinkedLibraries_collect__3647(struct environment_linked_libraries_root_LinkedLibraries__849 *);
static uint16_t environment_root_Environment_populateLinkedLibraries__779(struct environment_root_Environment__559 *);
static uint8_t environment_root_Environment_writeEnvironment__780(struct environment_root_Environment__559 *, int32_t);
static void array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_clearRetainingCapaci__4382(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *);
static int environment_linked_libraries_root_LinkedLibraries_collect__struct_8048_callback__6929(struct os_linux_dl_phdr_info__8025 *, uintptr_t, void *);
static void logger_err__anon_8063__6930(void);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4185(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *);
static nav__4192_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4192(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct mem_Allocator__599, nav__4192_44);
static void logger_err__anon_8090__6931(void);
static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3797(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880);
static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4187(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934);
static nav__6933_39 mem_span__anon_8113__6933(uint8_t const *);
static uint16_t fmt_format__anon_8220__6990(struct io_Writer__3878, nav__6990_41);
static nav__6934_40 fmt_bufPrint__anon_8122__6934(nav__6934_39, nav__6934_43);
static void logger_err__anon_8129__6935(void);
static nav__6957_40 json_stringify_stringifyAlloc__anon_8144__6957(struct mem_Allocator__599, struct environment_root_EnvironmentJson__845, struct json_stringify_StringifyOptions__8133);
static void logger_err__anon_8151__6958(void);
static nav__1351_39 fs_createFileAbsolute__1351(nav__1351_41, struct fs_File_CreateFlags__8153);
static void logger_err__anon_8170__6959(void);
static bool mem_startsWith__anon_8183__6960(nav__6960_39, nav__6960_39);
static struct environment_linked_libraries_elf_view__8184 environment_linked_libraries_elf_view_init__6966(uintptr_t, struct elf_Elf64_Phdr__8033 const *, uint16_t);
static nav__3646_39 environment_linked_libraries_root_LinkedLibraries_extractLibraryInfo__3646(struct mem_Allocator__599, struct environment_linked_libraries_elf_view__8184, nav__3646_43);
static uint16_t array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_append__4363(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *, struct environment_linked_libraries_root_LibraryInfo__946);
static void logger_logWithPrefix__anon_8206__6987(void);
static nav__4193_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4193(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct mem_Allocator__599, nav__4193_44);
static void logger_logWithPrefix__anon_8218__6988(void);
static uintptr_t mem_len__anon_8219__6989(uint8_t const *);
static void logger_logWithPrefix__anon_8278__6991(void);
static uint16_t json_stringify_stringifyArbitraryDepth__anon_8305__7021(struct mem_Allocator__599, struct environment_root_EnvironmentJson__845, struct json_stringify_StringifyOptions__8133, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static nav__1533_40 array_list_ArrayListAligned_28u8_2cnull_29_toOwnedSlice__1533(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *);
static void logger_logWithPrefix__anon_8310__7022(void);
static nav__4681_39 fs_Dir_createFile__4681(struct fs_Dir__1720, nav__4681_42, struct fs_File_CreateFlags__8153);
static void logger_logWithPrefix__anon_8315__7023(void);
static nav__6975_39 environment_linked_libraries_elf_view_buildId__6975(struct environment_linked_libraries_elf_view__8184);
static nav__3638_39 environment_linked_libraries_root_hexEncode__3638(struct mem_Allocator__599, nav__3638_39);
static nav__6970_40 environment_linked_libraries_elf_view_dynamicEntries__6970(struct environment_linked_libraries_elf_view__8184);
static uint8_t const *environment_linked_libraries_elf_view_strtab__6972(nav__6972_41);
static nav__6974_39 environment_linked_libraries_elf_view_soname__6974(nav__6974_42, uint8_t const *);
static nav__4388_40 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_addOne__4388(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *);
static uint16_t fmt_format__anon_8481__7032(struct io_Writer__3878);
static nav__7024_40 fmt_bufPrint__anon_8364__7024(nav__7024_39);
static nav__7025_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7025(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct mem_Allocator__599, nav__7025_44);
static uint16_t fmt_format__anon_8574__7035(struct io_Writer__3878);
static nav__7026_40 fmt_bufPrint__anon_8381__7026(nav__7026_39);
static uint16_t fmt_format__anon_8650__7036(struct io_Writer__3878);
static nav__7027_40 fmt_bufPrint__anon_8390__7027(nav__7027_39);
static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 json_stringify_writeStreamArbitraryDepth__anon_8398__7028(struct mem_Allocator__599, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct json_stringify_StringifyOptions__8133);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7029(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, struct environment_root_EnvironmentJson__845);
static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6998(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static void array_list_ArrayListAligned_28u8_2cnull_29_clearAndFree__1563(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *);
static uint16_t fmt_format__anon_8814__7039(struct io_Writer__3878);
static nav__7030_40 fmt_bufPrint__anon_8421__7030(nav__7030_39);
static nav__4682_39 fs_Dir_createFileZ__4682(struct fs_Dir__1720, uint8_t const *, struct fs_File_CreateFlags__8153);
static uint16_t fmt_format__anon_8889__7040(struct io_Writer__3878);
static nav__7031_40 fmt_bufPrint__anon_8436__7031(nav__7031_39);
static nav__6968_39 environment_linked_libraries_elf_view_slice__6968(struct environment_linked_libraries_elf_view__8184, uintptr_t, uintptr_t);
static nav__6969_39 environment_linked_libraries_elf_view_findPhdr__6969(struct environment_linked_libraries_elf_view__8184, uint32_t);
static nav__6971_38 environment_linked_libraries_elf_view_dynVal__6971(nav__6971_41, int64_t);
static nav__6973_39 environment_linked_libraries_elf_view_strFromTable__6973(uint8_t const *, uintptr_t);
static uint16_t array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_ensureTotalCapacity__4384(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *, uintptr_t);
static struct environment_linked_libraries_root_LibraryInfo__946 *array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_addOneAssumeCapacity__4389(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *);
static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4202(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct mem_Allocator__599, uintptr_t);
static nav__7033_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7033(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7033_41);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7034(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, nav__7034_42);
static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6997(struct mem_Allocator__599, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839, struct json_stringify_StringifyOptions__8133);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7005(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7008(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7003(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, uint8_t);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7007(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t fmt_formatBuf__anon_9211__7115(nav__7115_39, struct fmt_FormatOptions__4300, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t fmt_formatInt__anon_9201__7114(uint32_t, uint8_t, uint8_t, struct fmt_FormatOptions__4300, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t fmt_formatIntValue__anon_9176__7113(uint32_t, struct fmt_FormatOptions__4300, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t fmt_formatInt__anon_9228__7117(uint16_t, uint8_t, uint8_t, struct fmt_FormatOptions__4300, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t fmt_formatIntValue__anon_9224__7116(uint16_t, struct fmt_FormatOptions__4300, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t json_stringify_outputUnicodeEscape__anon_9160__7112(uint32_t, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t json_stringify_outputSpecialEscape__anon_9120__7111(uint8_t, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t json_stringify_encodeJsonStringChars__anon_9110__7110(nav__7110_39, struct json_stringify_StringifyOptions__8133, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t json_stringify_encodeJsonString__anon_9100__7109(nav__7109_39, struct json_stringify_StringifyOptions__8133, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, nav__7015_41);
static uint16_t json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29_jsonStringif__7118(struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7037(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861);
static uint16_t json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29_jsonStringify__anon_9275__7120(struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7038(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uintptr_t array_list_ArrayListAlignedUnmanaged_28environment_linked_libraries_root_LibraryInfo_2cnull_29_growCapacit__7099(uintptr_t, uintptr_t);
static uint16_t array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_ensureTotalCapacityP__4385(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4332(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *, struct mem_Allocator__599, uintptr_t);
static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4328(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *, struct mem_Allocator__599, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4294(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct array_hash_map_IndexHeader__892 *);
static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(nav__7100_40);
static nav__7101_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459);
static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(nav__7102_40, nav__7102_40, uintptr_t);
static nav__7103_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7103(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7103_41, struct array_hash_map_IndexHeader__892 *);
static nav__7104_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7104(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7104_41, struct array_hash_map_IndexHeader__892 *);
static nav__7105_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7105(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7105_41, struct array_hash_map_IndexHeader__892 *);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7106(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, nav__7106_42, struct array_hash_map_IndexHeader__892 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7107(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, nav__7107_42, struct array_hash_map_IndexHeader__892 *);
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7108(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, nav__7108_42, struct array_hash_map_IndexHeader__892 *);
static nav__7010_38 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7010(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *);
static bool json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7011(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *);
static uint16_t io_Writer_writeByte__5734(struct io_Writer__3878, uint8_t);
static uint16_t io_Writer_writeByteNTimes__5735(struct io_Writer__3878, uint8_t, uintptr_t);
static uint16_t json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_jsonStringify__anon_9358__7130(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7119(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7121(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, struct environment_linked_libraries_root_LibraryEntry__864);
static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7004(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, uint8_t);
static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static nav__7122_40 mem_Allocator_remap__anon_9292__7122(struct mem_Allocator__599, nav__7122_40, uintptr_t);
static nav__7123_41 mem_Allocator_alignedAlloc__anon_9296__7123(struct mem_Allocator__599, uintptr_t);
static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4330(uintptr_t, uintptr_t);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7124(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct array_hash_map_IndexHeader__892 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7125(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct array_hash_map_IndexHeader__892 *);
static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7126(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *, struct array_hash_map_IndexHeader__892 *);
static nav__7127_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7127(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7127_41, struct array_hash_map_IndexHeader__892 *, nav__7127_47);
static nav__7128_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7128(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7128_41, struct array_hash_map_IndexHeader__892 *, nav__7128_47);
static nav__7129_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7129(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934, nav__7129_41, struct array_hash_map_IndexHeader__892 *, nav__7129_47);
static uint16_t environment_root_EnvironmentValue_jsonStringify__anon_9438__7136(struct environment_root_EnvironmentValue__856, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7131(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, struct environment_root_EnvironmentValue__856);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7020(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, nav__7020_41);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7132(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, nav__7132_41);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7133(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, nav__7133_41);
static nav__7134_40 mem_bytesAsSlice__anon_9401__7134(nav__7134_42);
static nav__7135_39 mem_Allocator_allocWithSizeAndAlignment__anon_9402__7135(struct mem_Allocator__599, uintptr_t, uintptr_t);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7137(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, nav__7137_43);
static bool unicode_utf8ValidateSlice__5770(nav__5770_39);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6999(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7138(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *, uint8_t);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7001(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7139(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *);
static bool unicode_utf8ValidateSliceImpl__anon_9465__7140(nav__7140_39);
static uint16_t io_Writer_print__anon_9479__7141(struct io_Writer__3878, nav__7141_39);
static uint16_t fmt_format__anon_9488__7142(struct io_Writer__3878, nav__7142_39);
static uint16_t fmt_formatType__anon_9506__7143(uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878, uintptr_t);
static uint16_t fmt_formatInt__anon_9519__7146(uint8_t, uint8_t, uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatIntValue__anon_9515__7145(uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint16_t fmt_formatValue__anon_9511__7144(uint8_t, struct fmt_FormatOptions__4300, struct io_Writer__3878);
static uint64_t const builtin_zig_backend__262;
static bool const start_simplified_logic__109;
static uint8_t const builtin_output_mode__263;
static uint8_t const builtin_link_mode__264;
static uint64_t features_features__329;
static uintptr_t const bit_set_IntegerBitSet_2864_29_bit_length__360;
static struct Target_Os__639 const builtin_os__270;
static bool const builtin_link_libc__274;
static bool const posix_use_libc__1590;
static uint8_t const c_native_os__2032;
static uint8_t const builtin_mode__273;
static bool const debug_runtime_safety__159;
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3775;
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4165;
static struct Target_DynamicLinker__988 const Target_DynamicLinker_none__4413;
static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3986;
static bool const builtin_is_test__266;
static uint8_t *heap_CAllocator_alloc__4537(void *, uintptr_t, uint8_t, uintptr_t);
static bool heap_CAllocator_resize__4538(void *, nav__4538_40, uint8_t, uintptr_t, uintptr_t);
static uint8_t *heap_CAllocator_remap__4539(void *, nav__4539_40, uint8_t, uintptr_t, uintptr_t);
static void heap_CAllocator_free__4540(void *, nav__4540_40, uint8_t, uintptr_t);
static uint8_t *heap_CAllocator_alignedAlloc__4534(uintptr_t, uint8_t);
static uintptr_t heap_CAllocator_alignedAllocSize__4536(uint8_t *);
static void heap_CAllocator_alignedFree__4535(uint8_t *);
static uintptr_t mem_Alignment_toByteUnits__1112(uint8_t);
static struct mem_Allocator__599 const heap_c_allocator__4502;
static struct mem_Allocator__599 const c_allocator__235;
static struct Target_Cpu_Feature_Set__1005 const Target_Cpu_Feature_Set_empty__4458;
static struct Target_Cpu__973 const builtin_cpu__269;
static uint8_t const builtin_abi__268;
static uint8_t const builtin_object_format__272;
static struct Target__637 const builtin_target__271;
static struct builtin_CallingConvention__266 const builtin_CallingConvention_c__470;
zig_extern uint8_t running_on_valgrind(void);
static uint8_t const (*const shared_RUNNER_CTL_FIFO__801)[21];
static uint8_t const (*const shared_RUNNER_ACK_FIFO__802)[21];
static uint64_t const runner_fifo_PROTOCOL_VERSION__1125;
static uint8_t const mem_native_endian__836;
static uint8_t const posix_native_os__1588;
zig_extern int poll(struct os_linux_pollfd__1921 *, uintptr_t, int);
static uint8_t const os_linux_native_arch__2829;
zig_extern zig_noreturn void exit(int);
static uint8_t const fs_path_native_os__5117;
static uint8_t const fs_native_os__1303;
static uint8_t const fs_Dir_native_os__4784;
static bool const fs_File_is_windows__1479;
static uint8_t const c_native_abi__2030;
zig_extern int *zig_e___errno_location(void) zig_mangled(zig_e___errno_location, "__errno_location");
static bool const posix_unexpected_error_tracing__2019;
zig_extern int printf(char const *, ...);
static uint8_t logger_max_level__5093;
zig_extern int printf(char const *, ...);
zig_extern int flock(int32_t, int);
static bool const fs_Dir_have_flock__4785;
zig_extern intptr_t write(int32_t, uint8_t const *, uintptr_t);
zig_extern intptr_t read(int32_t, uint8_t *, uintptr_t);
static uint16_t const fmt_max_format_args__5471;
static bool const debug_default_enable_segfault_handler__205;
static uint8_t const log_default_level__6042;
static struct std_Options__4313 const std_options__97;
zig_extern int faccessat(int32_t, uint8_t const *, unsigned int, unsigned int);
static bool const posix_lfs64_abi__2018;
zig_extern int openat64(int, uint8_t const *, uint32_t, ...);
zig_extern int close(int32_t);
static uint8_t const (*const fmt_ANY__5474)[4];
static uint8_t const unicode_native_endian__5750;
static uint32_t const unicode_replacement_character__5751;
zig_extern void callgrind_zero_stats(void);
zig_extern void callgrind_start_instrumentation(void);
zig_extern void callgrind_stop_instrumentation(void);
zig_extern void callgrind_dump_stats_at(uint8_t const *);
static bool const mem_backend_supports_vectors__864;
static bool const builtin_valgrind_support__277;
static nav__4518_38 const heap_page_size_min_default__4518;
static uintptr_t const heap_page_size_min__4497;
static uint8_t const c_MARKER_TYPE_SAMPLE_START__246;
static uint8_t const c_MARKER_TYPE_SAMPLE_END__247;
static uint8_t const c_MARKER_TYPE_BENCHMARK_START__248;
static uint8_t const c_MARKER_TYPE_BENCHMARK_END__249;
zig_extern int clock_gettime(uint32_t, struct os_linux_timespec__struct_7321__7321 *);
static uint32_t const array_hash_map_index_capacities__3749[33];
static uint8_t const array_hash_map_Index_28u8_29_empty_sentinel__6278;
static uint16_t const array_hash_map_Index_28u16_29_empty_sentinel__6283;
static uint32_t const array_hash_map_Index_28u32_29_empty_sentinel__6288;
static uint64_t const hash_wyhash_Wyhash_secret__6452[4];
static bool const builtin_fuzz__279;
static bool const mem_eqlBytes_allowed__854;
zig_extern int dl_iterate_phdr(int (*)(struct os_linux_dl_phdr_info__8025 *, uintptr_t, void *), void *);
zig_extern uint8_t const *getenv(uint8_t const *);
static struct array_hash_map_Index_28u8_29__6735 const array_hash_map_Index_28u8_29_empty__6279;
static struct array_hash_map_Index_28u16_29__6739 const array_hash_map_Index_28u16_29_empty__6284;
static struct array_hash_map_Index_28u32_29__6743 const array_hash_map_Index_28u32_29_empty__6289;
static bool const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6993;
static struct json_stringify_WriteStream__union_8283__8283 const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6994;
static uint8_t const environment_linked_libraries_elf_view_native_endian__6964;
static struct mem_Allocator_VTable__602 const heap_CAllocator_vtable__4529;
zig_extern uintptr_t malloc_usable_size(void const *);
static bool const heap_CAllocator_supports_malloc_size__4530;
static bool const heap_CAllocator_supports_posix_memalign__4532;
zig_extern int posix_memalign(void **, uintptr_t, uintptr_t);
zig_extern void free(void *);
static struct Target_Cpu_Model__1000 const Target_x86_cpu_x86_64__4649;
enum {
 zig_error_ModeError = 1u,
 zig_error_OutOfMemory = 2u,
 zig_error_NotInstrumented = 3u,
 zig_error_PermissionDenied = 4u,
 zig_error_FileNotFound = 5u,
 zig_error_NameTooLong = 6u,
 zig_error_InputOutput = 7u,
 zig_error_SystemResources = 8u,
 zig_error_BadPathName = 9u,
 zig_error_FileBusy = 10u,
 zig_error_SymLinkLoop = 11u,
 zig_error_ReadOnlyFileSystem = 12u,
 zig_error_InvalidUtf8 = 13u,
 zig_error_InvalidWtf8 = 14u,
 zig_error_Unexpected = 15u,
 zig_error_SharingViolation = 16u,
 zig_error_PathAlreadyExists = 17u,
 zig_error_AccessDenied = 18u,
 zig_error_PipeBusy = 19u,
 zig_error_NoDevice = 20u,
 zig_error_NetworkNotFound = 21u,
 zig_error_AntivirusInterference = 22u,
 zig_error_ProcessFdQuotaExceeded = 23u,
 zig_error_SystemFdQuotaExceeded = 24u,
 zig_error_FileTooBig = 25u,
 zig_error_IsDir = 26u,
 zig_error_NoSpaceLeft = 27u,
 zig_error_NotDir = 28u,
 zig_error_DeviceBusy = 29u,
 zig_error_FileLocksNotSupported = 30u,
 zig_error_WouldBlock = 31u,
 zig_error_DiskQuota = 32u,
 zig_error_InvalidArgument = 33u,
 zig_error_BrokenPipe = 34u,
 zig_error_OperationAborted = 35u,
 zig_error_NotOpenForWriting = 36u,
 zig_error_LockViolation = 37u,
 zig_error_ConnectionResetByPeer = 38u,
 zig_error_ProcessNotFound = 39u,
 zig_error_NetworkSubsystemFailed = 40u,
 zig_error_AckTimeout = 41u,
 zig_error_ConnectionTimedOut = 42u,
 zig_error_NotOpenForReading = 43u,
 zig_error_SocketNotConnected = 44u,
 zig_error_Canceled = 45u,
 zig_error_NotReady = 46u,
 zig_error_UnexpectedEof = 47u,
 zig_error_EndOfStream = 48u,
 zig_error_UnexpectedError = 49u,
 zig_error_UnexpectedResponse = 50u,
 zig_error_Overflow = 51u,
 zig_error_Utf8ExpectedContinuation = 52u,
 zig_error_Utf8OverlongEncoding = 53u,
 zig_error_Utf8EncodesSurrogateHalf = 54u,
 zig_error_Utf8CodepointTooLarge = 55u,
 zig_error_Utf8InvalidStartByte = 56u,
 zig_error_TruncatedInput = 57u,
 zig_error_Utf8CannotEncodeSurrogateHalf = 58u,
 zig_error_CodepointTooLarge = 59u,
 zig_error_UnsupportedClock = 60u,
 zig_error_DlIteratePhdrFailed = 61u,
};
static uint8_t const zig_errorName_ModeError[10] = "ModeError";
static uint8_t const zig_errorName_OutOfMemory[12] = "OutOfMemory";
static uint8_t const zig_errorName_NotInstrumented[16] = "NotInstrumented";
static uint8_t const zig_errorName_PermissionDenied[17] = "PermissionDenied";
static uint8_t const zig_errorName_FileNotFound[13] = "FileNotFound";
static uint8_t const zig_errorName_NameTooLong[12] = "NameTooLong";
static uint8_t const zig_errorName_InputOutput[12] = "InputOutput";
static uint8_t const zig_errorName_SystemResources[16] = "SystemResources";
static uint8_t const zig_errorName_BadPathName[12] = "BadPathName";
static uint8_t const zig_errorName_FileBusy[9] = "FileBusy";
static uint8_t const zig_errorName_SymLinkLoop[12] = "SymLinkLoop";
static uint8_t const zig_errorName_ReadOnlyFileSystem[19] = "ReadOnlyFileSystem";
static uint8_t const zig_errorName_InvalidUtf8[12] = "InvalidUtf8";
static uint8_t const zig_errorName_InvalidWtf8[12] = "InvalidWtf8";
static uint8_t const zig_errorName_Unexpected[11] = "Unexpected";
static uint8_t const zig_errorName_SharingViolation[17] = "SharingViolation";
static uint8_t const zig_errorName_PathAlreadyExists[18] = "PathAlreadyExists";
static uint8_t const zig_errorName_AccessDenied[13] = "AccessDenied";
static uint8_t const zig_errorName_PipeBusy[9] = "PipeBusy";
static uint8_t const zig_errorName_NoDevice[9] = "NoDevice";
static uint8_t const zig_errorName_NetworkNotFound[16] = "NetworkNotFound";
static uint8_t const zig_errorName_AntivirusInterference[22] = "AntivirusInterference";
static uint8_t const zig_errorName_ProcessFdQuotaExceeded[23] = "ProcessFdQuotaExceeded";
static uint8_t const zig_errorName_SystemFdQuotaExceeded[22] = "SystemFdQuotaExceeded";
static uint8_t const zig_errorName_FileTooBig[11] = "FileTooBig";
static uint8_t const zig_errorName_IsDir[6] = "IsDir";
static uint8_t const zig_errorName_NoSpaceLeft[12] = "NoSpaceLeft";
static uint8_t const zig_errorName_NotDir[7] = "NotDir";
static uint8_t const zig_errorName_DeviceBusy[11] = "DeviceBusy";
static uint8_t const zig_errorName_FileLocksNotSupported[22] = "FileLocksNotSupported";
static uint8_t const zig_errorName_WouldBlock[11] = "WouldBlock";
static uint8_t const zig_errorName_DiskQuota[10] = "DiskQuota";
static uint8_t const zig_errorName_InvalidArgument[16] = "InvalidArgument";
static uint8_t const zig_errorName_BrokenPipe[11] = "BrokenPipe";
static uint8_t const zig_errorName_OperationAborted[17] = "OperationAborted";
static uint8_t const zig_errorName_NotOpenForWriting[18] = "NotOpenForWriting";
static uint8_t const zig_errorName_LockViolation[14] = "LockViolation";
static uint8_t const zig_errorName_ConnectionResetByPeer[22] = "ConnectionResetByPeer";
static uint8_t const zig_errorName_ProcessNotFound[16] = "ProcessNotFound";
static uint8_t const zig_errorName_NetworkSubsystemFailed[23] = "NetworkSubsystemFailed";
static uint8_t const zig_errorName_AckTimeout[11] = "AckTimeout";
static uint8_t const zig_errorName_ConnectionTimedOut[19] = "ConnectionTimedOut";
static uint8_t const zig_errorName_NotOpenForReading[18] = "NotOpenForReading";
static uint8_t const zig_errorName_SocketNotConnected[19] = "SocketNotConnected";
static uint8_t const zig_errorName_Canceled[9] = "Canceled";
static uint8_t const zig_errorName_NotReady[9] = "NotReady";
static uint8_t const zig_errorName_UnexpectedEof[14] = "UnexpectedEof";
static uint8_t const zig_errorName_EndOfStream[12] = "EndOfStream";
static uint8_t const zig_errorName_UnexpectedError[16] = "UnexpectedError";
static uint8_t const zig_errorName_UnexpectedResponse[19] = "UnexpectedResponse";
static uint8_t const zig_errorName_Overflow[9] = "Overflow";
static uint8_t const zig_errorName_Utf8ExpectedContinuation[25] = "Utf8ExpectedContinuation";
static uint8_t const zig_errorName_Utf8OverlongEncoding[21] = "Utf8OverlongEncoding";
static uint8_t const zig_errorName_Utf8EncodesSurrogateHalf[25] = "Utf8EncodesSurrogateHalf";
static uint8_t const zig_errorName_Utf8CodepointTooLarge[22] = "Utf8CodepointTooLarge";
static uint8_t const zig_errorName_Utf8InvalidStartByte[21] = "Utf8InvalidStartByte";
static uint8_t const zig_errorName_TruncatedInput[15] = "TruncatedInput";
static uint8_t const zig_errorName_Utf8CannotEncodeSurrogateHalf[30] = "Utf8CannotEncodeSurrogateHalf";
static uint8_t const zig_errorName_CodepointTooLarge[18] = "CodepointTooLarge";
static uint8_t const zig_errorName_UnsupportedClock[17] = "UnsupportedClock";
static uint8_t const zig_errorName_DlIteratePhdrFailed[20] = "DlIteratePhdrFailed";
static struct anon__lazy_58 const zig_errorName[62] = {{zig_errorName_ModeError, 9ul}, {zig_errorName_OutOfMemory, 11ul}, {zig_errorName_NotInstrumented, 15ul}, {zig_errorName_PermissionDenied, 16ul}, {zig_errorName_FileNotFound, 12ul}, {zig_errorName_NameTooLong, 11ul}, {zig_errorName_InputOutput, 11ul}, {zig_errorName_SystemResources, 15ul}, {zig_errorName_BadPathName, 11ul}, {zig_errorName_FileBusy, 8ul}, {zig_errorName_SymLinkLoop, 11ul}, {zig_errorName_ReadOnlyFileSystem, 18ul}, {zig_errorName_InvalidUtf8, 11ul}, {zig_errorName_InvalidWtf8, 11ul}, {zig_errorName_Unexpected, 10ul}, {zig_errorName_SharingViolation, 16ul}, {zig_errorName_PathAlreadyExists, 17ul}, {zig_errorName_AccessDenied, 12ul}, {zig_errorName_PipeBusy, 8ul}, {zig_errorName_NoDevice, 8ul}, {zig_errorName_NetworkNotFound, 15ul}, {zig_errorName_AntivirusInterference, 21ul}, {zig_errorName_ProcessFdQuotaExceeded, 22ul}, {zig_errorName_SystemFdQuotaExceeded, 21ul}, {zig_errorName_FileTooBig, 10ul}, {zig_errorName_IsDir, 5ul}, {zig_errorName_NoSpaceLeft, 11ul}, {zig_errorName_NotDir, 6ul}, {zig_errorName_DeviceBusy, 10ul}, {zig_errorName_FileLocksNotSupported, 21ul}, {zig_errorName_WouldBlock, 10ul}, {zig_errorName_DiskQuota, 9ul}, {zig_errorName_InvalidArgument, 15ul}, {zig_errorName_BrokenPipe, 10ul}, {zig_errorName_OperationAborted, 16ul}, {zig_errorName_NotOpenForWriting, 17ul}, {zig_errorName_LockViolation, 13ul}, {zig_errorName_ConnectionResetByPeer, 21ul}, {zig_errorName_ProcessNotFound, 15ul}, {zig_errorName_NetworkSubsystemFailed, 22ul}, {zig_errorName_AckTimeout, 10ul}, {zig_errorName_ConnectionTimedOut, 18ul}, {zig_errorName_NotOpenForReading, 17ul}, {zig_errorName_SocketNotConnected, 18ul}, {zig_errorName_Canceled, 8ul}, {zig_errorName_NotReady, 8ul}, {zig_errorName_UnexpectedEof, 13ul}, {zig_errorName_EndOfStream, 11ul}, {zig_errorName_UnexpectedError, 15ul}, {zig_errorName_UnexpectedResponse, 18ul}, {zig_errorName_Overflow, 8ul}, {zig_errorName_Utf8ExpectedContinuation, 24ul}, {zig_errorName_Utf8OverlongEncoding, 20ul}, {zig_errorName_Utf8EncodesSurrogateHalf, 24ul}, {zig_errorName_Utf8CodepointTooLarge, 21ul}, {zig_errorName_Utf8InvalidStartByte, 20ul}, {zig_errorName_TruncatedInput, 14ul}, {zig_errorName_Utf8CannotEncodeSurrogateHalf, 29ul}, {zig_errorName_CodepointTooLarge, 17ul}, {zig_errorName_UnsupportedClock, 16ul}, {zig_errorName_DlIteratePhdrFailed, 19ul}};
static struct anon__lazy_58 zig_tag_name_shared_IntegrationMode__573(uint8_t const tag) {
 switch (tag) {
  case UINT8_C(0): {
   static uint8_t const name[9] = "Walltime";
   return (struct anon__lazy_58){name, (uintptr_t)8ul};
  }
  case UINT8_C(1): {
   static uint8_t const name[11] = "Simulation";
   return (struct anon__lazy_58){name, (uintptr_t)10ul};
  }
  case UINT8_C(2): {
   static uint8_t const name[9] = "Analysis";
   return (struct anon__lazy_58){name, (uintptr_t)8ul};
  }
 }
 while (true) zig_breakpoint();
}

static uint8_t const __anon_1704[21] = "/tmp/runner.ctl.fifo";

static uint8_t const __anon_1769[21] = "/tmp/runner.ack.fifo";

static char const __anon_3403[33] = {73,110,118,97,108,105,100,32,112,114,111,116,111,99,111,108,32,100,101,116,101,99,116,101,100,58,32,37,46,42,115,10,0};

static uint8_t const __anon_2400[62] = "The stream end was found before all required bytes were read.";

static uint8_t const __anon_3484[34] = "[ERROR] logger formatting failed\n";

static uint8_t const __anon_4002[49] = "[DEBUG] waitForAck received unexpected response: ";

static uint8_t const __anon_4385[1] = "\n";

static uint8_t const __anon_4501[68] = "[ERROR] instrument-hooks: CodSpeed runner rejected protocol version ";

static uint8_t const __anon_4624[82] = "[ERROR] instrument-hooks: please update the CodSpeed action to the latest version\n";

static uint8_t const __anon_4678[32] = "[ERROR] instrument-hooks: error ";

static uint8_t const __anon_4811[22] = " during version check\n";

static uint8_t const __anon_2108[15] = "StartBenchmark";

static uint8_t const __anon_2113[14] = "StopBenchmark";

static uint8_t const __anon_2118[4] = "Ack";

static uint8_t const __anon_2123[13] = "PingProfiler";

static uint8_t const __anon_2131[4] = "Err";

static uint8_t const __anon_2146[19] = "GetIntegrationMode";

static uint8_t const __anon_5166[8] = "{ ... }";

static uint8_t const __anon_5345[3] = "\357\277\275";

static uint8_t const __anon_5419[25] = "ExecutedBenchmark { pid: ";

static uint8_t const __anon_5485[7] = ", uri: ";

static uint8_t const __anon_5511[2] = " }";

static uint8_t const __anon_5564[23] = "SetIntegration { name: ";

static uint8_t const __anon_5589[11] = ", version: ";

static uint8_t const __anon_5645[17] = "AddMarker { pid: ";

static uint8_t const __anon_5667[10] = ", marker: ";

static uint8_t const __anon_5737[31] = "SetVersion { protocol_version: ";

static uint8_t const __anon_5792[24] = "IntegrationModeResponse ";

static uint8_t const __anon_5908[23] = "shared.IntegrationMode";

static uint8_t const __anon_5913[2] = ".";

static uint8_t const __anon_6005[12] = "SampleStart(";

static uint8_t const __anon_6015[1] = ")";

static uint8_t const __anon_6034[10] = "SampleEnd(";

static uint8_t const __anon_6066[15] = "BenchmarkStart(";

static uint8_t const __anon_6094[13] = "BenchmarkEnd(";

static uint8_t const __anon_7231[10] = "Metadata: ";

static uint8_t const __anon_7246[1] = " ";

static uint8_t const __anon_7259[1] = "\000";

static uint8_t const __anon_8103[24] = "CODSPEED_PROFILE_FOLDER";

static uint8_t const __anon_8178[11] = "linux-vdso";

static uint8_t const __anon_8255[13] = "/environment-";

static uint8_t const __anon_8272[5] = ".json";

static uint8_t const __anon_8445[5] = "GNU\000";

static uint8_t const __anon_8449[17] = "0123456789abcdef";

static uint8_t const __anon_8550[61] = "[ERROR] instrument-hooks: failed to collect linked libraries\n";

static uint8_t const __anon_8644[62] = "[ERROR] instrument-hooks: failed to populate linked libraries\n";

static uint8_t const __anon_8711[55] = "[ERROR] instrument-hooks: profile folder path too long\n";

static uint8_t const __anon_8724[24] = "integration_environment";

static uint8_t const __anon_8732[17] = "linked_libraries";

static uint8_t const __anon_8883[63] = "[ERROR] instrument-hooks: failed to serialize environment JSON\n";

static uint8_t const __anon_8956[59] = "[ERROR] instrument-hooks: failed to write environment.json\n";

static uint8_t const __anon_9166[3] = "\\u";

static uint8_t const __anon_9125[3] = "\\\\";

static uint8_t const __anon_9130[3] = "\\\"";

static uint8_t const __anon_9134[3] = "\\b";

static uint8_t const __anon_9138[3] = "\\f";

static uint8_t const __anon_9142[3] = "\\n";

static uint8_t const __anon_9146[3] = "\\r";

static uint8_t const __anon_9150[3] = "\\t";

static uint8_t const __anon_9366[5] = "path";

static uint8_t const __anon_9370[9] = "build_id";

static uint8_t const __anon_9483[5] = "null";

static uint8_t const __anon_5003[4] = "any";

static uint8_t const __anon_9580[7] = "x86_64";

static uint8_t const __anon_9582[7] = "x86-64";

void c_instrument_hooks_set_feature__237(uint64_t const a0, bool const a1) {
 uint64_t t0;
 t0 = a0;
 features_set_feature__330(t0, a1);
 return;
}

static void features_set_feature__330(uint64_t const a0, bool const a1) {
 uint64_t t0;
 uintptr_t t1;
 if (a1) {
  t0 = a0;
  t1 = t0;
  bit_set_IntegerBitSet_2864_29_set__369(&features_features__329, t1);
  goto zig_block_0;
 }
 t0 = a0;
 t1 = t0;
 bit_set_IntegerBitSet_2864_29_unset__371(&features_features__329, t1);
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void bit_set_IntegerBitSet_2864_29_set__369(uint64_t *const a0, uintptr_t const a1) {
 uint64_t *const *t1;
 uint64_t t2;
 uint64_t t5;
 uint64_t *t4;
 uint64_t *t0;
 bool t3;
 t0 = a0;
 t1 = (uint64_t *const *)&t0;
 t2 = a1;
 t3 = t2 < UINT64_C(64);
 debug_assert__180(t3);
 t4 = (*t1);
 t4 = (uint64_t *)((uint8_t *)t4 + (uintptr_t)0ul);
 t2 = (*t4);
 t5 = bit_set_IntegerBitSet_2864_29_maskBit__390(a1);
 t5 = t2 | t5;
 (*t4) = t5;
 return;
}

static void bit_set_IntegerBitSet_2864_29_unset__371(uint64_t *const a0, uintptr_t const a1) {
 uint64_t *const *t1;
 uint64_t t2;
 uint64_t t5;
 uint64_t *t4;
 uint64_t *t0;
 bool t3;
 t0 = a0;
 t1 = (uint64_t *const *)&t0;
 t2 = a1;
 t3 = t2 < UINT64_C(64);
 debug_assert__180(t3);
 t4 = (*t1);
 t4 = (uint64_t *)((uint8_t *)t4 + (uintptr_t)0ul);
 t2 = (*t4);
 t5 = bit_set_IntegerBitSet_2864_29_maskBit__390(a1);
 t5 = zig_not_u64(t5, UINT8_C(64));
 t5 = t2 & t5;
 (*t4) = t5;
 return;
}

static void debug_assert__180(bool const a0) {
 bool t0;
 t0 = !a0;
 if (t0) {
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static uint64_t bit_set_IntegerBitSet_2864_29_maskBit__390(uintptr_t const a0) {
 uint64_t t1;
 uint8_t t0;
 t0 = (uint8_t)a0;
 t1 = zig_shlw_u64(UINT64_C(1), t0, UINT8_C(64));
 return t1;
}

static nav__786_39 instruments_valgrind_ValgrindInstrument_init__786(struct mem_Allocator__599 const a0) {
 struct instruments_valgrind_ValgrindInstrument__564 t2;
 nav__786_39 t3;
 uint8_t t0;
 bool t1;
 t0 = running_on_valgrind();
 t1 = t0 > UINT8_C(0);
 t1 = !t1;
 if (t1) {
  return (nav__786_39){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},zig_error_NotInstrumented};
 }
 goto zig_block_0;

 zig_block_0:;
 t2.allocator = a0;
 t3.payload = t2;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__1280_39 fifo_posix_Pipe_openPipe__1280(nav__1280_41 const a0) {
 nav__1280_39 t1;
 struct fs_File__799 t2;
 uint16_t t0;
 t0 = fs_accessAbsolute__1348(a0, (struct fs_File_OpenFlags__1718){UINT8_C(2),UINT8_C(0),false,false});
 if (t0) {
  t1.payload = (struct fs_File__799){-INT32_C(0x55555556)};
  t1.error = t0;
  return t1;
 }
 t1 = fs_openFileAbsolute__1345(a0, (struct fs_File_OpenFlags__1718){UINT8_C(2),UINT8_C(0),false,false});
 if (t1.error) {
  t0 = t1.error;
  t1.payload = (struct fs_File__799){-INT32_C(0x55555556)};
  t1.error = t0;
  return t1;
 }
 t2 = t1.payload;
 t1.payload = t2;
 t1.error = UINT16_C(0);
 return t1;
}

static nav__1282_39 fifo_posix_Pipe_openWrite__1282(struct mem_Allocator__599 const a0, nav__1282_42 const a1) {
 nav__1282_39 t2;
 struct fifo_posix_Pipe_Writer__791 t4;
 nav__1282_59 t0;
 struct fs_File__799 t3;
 uint16_t t1;
 t0 = fifo_posix_Pipe_openPipe__1280(a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct fifo_posix_Pipe_Writer__791){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = fifo_posix_Pipe_Writer_init__1283(t3, a0);
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__1281_39 fifo_posix_Pipe_openRead__1281(struct mem_Allocator__599 const a0, nav__1281_42 const a1) {
 nav__1281_39 t2;
 struct fifo_posix_Pipe_Reader__793 t4;
 nav__1281_59 t0;
 struct fs_File__799 t3;
 uint16_t t1;
 t0 = fifo_posix_Pipe_openPipe__1280(a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct fifo_posix_Pipe_Reader__793){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = fifo_posix_Pipe_Reader_init__1288(t3, a0);
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__1129_39 runner_fifo_RunnerFifo_init__1129(struct mem_Allocator__599 const a0) {
 struct runner_fifo_RunnerFifo__631 *t1;
 struct mem_Allocator__599 *t2;
 struct fifo_posix_Pipe_Writer__791 *t3;
 nav__1129_66 t4;
 nav__1129_39 t6;
 nav__1129_39 t0;
 struct fifo_posix_Pipe_Writer__791 t7;
 struct fifo_posix_Pipe_Reader__793 *t8;
 nav__1129_71 t9;
 struct fifo_posix_Pipe_Reader__793 t10;
 uint16_t t5;
 t0.error = UINT16_C(0);
 t1 = &t0.payload;
 t2 = (struct mem_Allocator__599 *)&t1->allocator;
 (*t2) = a0;
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t1->writer;
 t4 = fifo_posix_Pipe_openWrite__1282(a0, (nav__1129_64){(uint8_t const *)&__anon_1704,(uintptr_t)20ul});
 if (t4.error) {
  t5 = t4.error;
  t6.payload = (struct runner_fifo_RunnerFifo__631){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}};
  t6.error = t5;
  return t6;
 }
 t7 = t4.payload;
 (*t3) = t7;
 t8 = (struct fifo_posix_Pipe_Reader__793 *)&t1->reader;
 t9 = fifo_posix_Pipe_openRead__1281(a0, (nav__1129_64){(uint8_t const *)&__anon_1769,(uintptr_t)20ul});
 if (t9.error) {
  t5 = t9.error;
  t6.payload = (struct runner_fifo_RunnerFifo__631){{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}};
  t6.error = t5;
  return t6;
 }
 t10 = t9.payload;
 (*t8) = t10;
 return t0;
}

static uint16_t fifo_posix_Pipe_Writer_sendCmd__1286(struct fifo_posix_Pipe_Writer__791 *const a0, struct shared_Command__1795 const a1) {
 struct fifo_posix_Pipe_Writer__791 *const *t1;
 struct fifo_posix_Pipe_Writer__791 *t2;
 struct fifo_posix_Pipe_Writer__791 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t6;
 nav__1286_65 t7;
 struct fs_File__799 *t8;
 uintptr_t t10;
 uint32_t const *t13;
 uint8_t const (*t14)[4];
 nav__1286_44 t15;
 struct fs_File__799 t9;
 uint32_t t11;
 uint32_t t12;
 uint16_t t5;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Writer__791 *const *)&t0;
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t2->buffer;
 array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1562(t3);
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t2->buffer;
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(t3);
 t5 = bincode_serialize__anon_1849__4913(t4, a1);
 if (t5) {
  return t5;
 }
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&a0->buffer;
 t6 = (*t3);
 t7 = t6.items;
 t2 = (*t1);
 t8 = (struct fs_File__799 *)&t2->file;
 t9 = (*t8);
 t10 = t7.len;
 t11 = (uint32_t)t10;
 t12 = t11;
 t13 = (uint32_t const *)&t12;
 t14 = mem_asBytes__anon_1887__4914(t13);
 t15.ptr = &(*t14)[(uintptr_t)0ul];
 t15.len = (uintptr_t)4ul;
 t5 = fs_File_writeAll__1439(t9, t15);
 if (t5) {
  return t5;
 }
 t2 = (*t1);
 t8 = (struct fs_File__799 *)&t2->file;
 t9 = (*t8);
 memcpy(&t15, &t7, sizeof(nav__1286_44));
 t5 = fs_File_writeAll__1439(t9, t15);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t fifo_posix_waitReadable__1268(int32_t const a0, uint64_t const a1) {
 uint64_t t3;
 uintptr_t t7;
 uintptr_t t10;
 nav__1268_43 t8;
 nav__1268_46 t9;
 struct os_linux_pollfd__1921 t1;
 struct os_linux_pollfd__1921 t2[1];
 struct os_linux_pollfd__1921 t0[1];
 int32_t t4;
 int32_t t6;
 bool t5;
 t1.fd = a0;
 t1.events = INT16_C(1);
 t1.revents = INT16_C(0);
 t2[0] = t1;
 memcpy((char *)&t0, t2, sizeof(struct os_linux_pollfd__1921[1]));
 t3 = a1 / UINT64_C(1000000);
 t5 = t3 > UINT64_C(2147483647);
 if (t5) {
  t4 = INT32_MAX;
  goto zig_block_0;
 }
 t6 = (int32_t)t3;
 t4 = t6;
 goto zig_block_0;

 zig_block_0:;
 t8.ptr = (struct os_linux_pollfd__1921 *)&t0[(uintptr_t)0ul];
 t8.len = (uintptr_t)1ul;
 t9 = posix_poll__1964(t8, t4);
 t5 = t9.error == UINT16_C(0);
 if (t5) {
  t10 = t9.payload;
  t7 = t10;
  goto zig_block_1;
 }
 return zig_error_AckTimeout;

 zig_block_1:;
 t3 = t7;
 t5 = t3 == UINT64_C(0);
 if (t5) {
  return zig_error_AckTimeout;
 }
 goto zig_block_2;

 zig_block_2:;
 return 0;
}

static nav__5031_38 bincode_deserializeInt__anon_2223__5031(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t8;
 void const *t9;
 nav__5031_50 (**t10)(void const *, nav__5031_52);
 struct io_Reader__2238 t11;
 struct io_Reader__2238 t6;
 struct io_Reader__2238 t12;
 struct io_Reader__2238 t15;
 struct io_Reader__2238 const *t13;
 uint8_t const (*t21)[4];
 nav__5031_38 t14;
 nav__5031_38 t18;
 uint32_t t22;
 nav__5031_64 t16;
 uint16_t t17;
 uint8_t t19[4];
 uint8_t t20[4];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5031_50 (**)(void const *, nav__5031_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2238 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2238 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2269__5069(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = UINT32_C(0xaaaaaaaa);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[4]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[4]));
 t21 = (uint8_t const (*)[4])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[4]));
 memcpy(&t22, &t19, sizeof(uint32_t));
 t22 = zig_wrap_u32(t22, UINT8_C(32));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__5031_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = UINT32_C(0xaaaaaaaa);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__5030_38 bincode_deserializeAlloc__anon_2212__5030(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5030_38 t0;
 uint32_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2223__5031(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT32_C(0xaaaaaaaa);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5073_38 bincode_deserializeInt__anon_2332__5073(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t8;
 void const *t9;
 nav__5073_50 (**t10)(void const *, nav__5073_52);
 struct io_Reader__2238 t11;
 struct io_Reader__2238 t6;
 struct io_Reader__2238 t12;
 struct io_Reader__2238 t15;
 struct io_Reader__2238 const *t13;
 uint8_t const (*t21)[4];
 nav__5073_38 t14;
 nav__5073_38 t18;
 int32_t t22;
 nav__5073_64 t16;
 uint16_t t17;
 uint8_t t19[4];
 uint8_t t20[4];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5073_50 (**)(void const *, nav__5073_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2238 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2238 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2269__5069(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = -INT32_C(0x55555556);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[4]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[4]));
 t21 = (uint8_t const (*)[4])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[4]));
 memcpy(&t22, &t19, sizeof(int32_t));
 t22 = zig_wrap_i32(t22, UINT8_C(32));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__5073_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = -INT32_C(0x55555556);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__5072_38 bincode_deserializeAlloc__anon_2327__5072(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5072_38 t0;
 int32_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2332__5073(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = -INT32_C(0x55555556);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5075_40 bincode_deserializePointerAlloc__anon_2362__5075(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t6;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t7;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t32;
 struct mem_Allocator__599 const *t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t5;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t31;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *t8;
 void const **t10;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t11;
 void const *t12;
 nav__5075_59 (**t13)(void const *, nav__5075_39);
 struct io_Reader__2238 t14;
 struct io_Reader__2238 t9;
 struct io_Reader__2238 t15;
 struct io_Reader__2238 t18;
 struct io_Reader__2238 t33;
 struct io_Reader__2238 t34;
 struct io_Reader__2238 const *t16;
 nav__5075_69 t17;
 nav__5075_69 t21;
 uint8_t const (*t24)[8];
 uint64_t t25;
 uint64_t t38;
 nav__5075_40 t26;
 uintptr_t t27;
 uintptr_t t37;
 struct mem_Allocator__599 t29;
 struct mem_Allocator__599 t2;
 nav__5075_39 t30;
 nav__5075_39 t28;
 nav__5075_59 t35;
 nav__5075_59 t36;
 nav__5075_72 t19;
 uint16_t t20;
 uint8_t t22[8];
 uint8_t t23[8];
 bool t39;
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t1);
 t5 = t4;
 t6 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t5;
 t7 = t6;
 t8 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *)&t7;
 t10 = (void const **)&t9.context;
 t6 = (*t8);
 t11 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t6->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__5075_59 (**)(void const *, nav__5075_39))&t9.readFn;
 (*t13) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027;
 t14 = t9;
 t15 = t14;
 t16 = (struct io_Reader__2238 const *)&t15;
 t14 = (*t16);
 t18 = t14;
 t16 = (struct io_Reader__2238 const *)&t18;
 t14 = (*t16);
 t19 = io_Reader_readBytesNoEof__anon_2375__5076(t14);
 if (t19.error) {
  t20 = t19.error;
  t21.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t21.error = t20;
  t17 = t21;
  goto zig_block_0;
 }
 memcpy(t22, t19.payload, sizeof(uint8_t[8]));
 memcpy((char *)&t23, t22, sizeof(uint8_t[8]));
 t24 = (uint8_t const (*)[8])&t23;
 memcpy(t22, (const char *)t24, sizeof(uint8_t[8]));
 memcpy(&t25, &t22, sizeof(uint64_t));
 t25 = zig_wrap_u64(t25, UINT8_C(64));
 t21.payload = t25;
 t21.error = UINT16_C(0);
 t17 = t21;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t21, &t17, sizeof(nav__5075_69));
 if (t21.error) {
  t20 = t21.error;
  t26.payload = (nav__5075_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t26.error = t20;
  return t26;
 }
 t25 = t21.payload;
 t27 = t25;
 t29 = (*t3);
 t26 = mem_Allocator_alloc__anon_2384__5077(t29, t27);
 if (t26.error) {
  t20 = t26.error;
  t26.payload = (nav__5075_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t26.error = t20;
  return t26;
 }
 t30 = t26.payload;
 t28 = t30;
 t4 = (*t1);
 t30 = t28;
 t31 = t4;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t31;
 t32 = t1;
 t8 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *)&t32;
 t10 = (void const **)&t33.context;
 t1 = (*t8);
 t11 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t1->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__5075_59 (**)(void const *, nav__5075_39))&t33.readFn;
 (*t13) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027;
 t14 = t33;
 t34 = t14;
 t16 = (struct io_Reader__2238 const *)&t34;
 t14 = (*t16);
 t35 = io_Reader_readAll__5034(t14, t30);
 memcpy(&t36, &t35, sizeof(nav__5075_59));
 t37 = t36.payload;
 t25 = t37;
 t38 = t27;
 t39 = t25 != t38;
 if (t39) {
  bincode_invalidProtocol__anon_2407__5078();
  zig_unreachable();
 }
 goto zig_block_1;

 zig_block_1:;
 t30 = t28;
 t26.payload = t30;
 t26.error = UINT16_C(0);
 return t26;
}

static nav__5074_40 bincode_deserializeAlloc__anon_2349__5074(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5074_57 t0;
 nav__5074_40 t2;
 nav__5074_55 t3;
 nav__5074_39 t4;
 uint16_t t1;
 t0 = bincode_deserializePointerAlloc__anon_2362__5075(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (nav__5074_39){(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 memcpy(&t4, &t3, sizeof(nav__5074_39));
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__5071_39 bincode_deserializeStructAlloc__anon_2311__5071(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 int32_t *t1;
 nav__5071_39 t4;
 nav__5071_44 *t6;
 nav__5071_60 t7;
 nav__5071_44 t8;
 struct shared_Command__struct_1798__1798 t9;
 struct shared_Command__struct_1798__1798 t0;
 nav__5071_57 t2;
 int32_t t5;
 uint16_t t3;
 t1 = (int32_t *)&t0.pid;
 t2 = bincode_deserializeAlloc__anon_2327__5072(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1798__1798){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = (nav__5071_44 *)&t0.uri;
 t7 = bincode_deserializeAlloc__anon_2349__5074(a0, a1);
 if (t7.error) {
  t3 = t7.error;
  t4.payload = (struct shared_Command__struct_1798__1798){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t8 = t7.payload;
 (*t6) = t8;
 t9 = t0;
 t4.payload = t9;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__5070_39 bincode_deserializeAlloc__anon_2282__5070(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5070_39 t0;
 struct shared_Command__struct_1798__1798 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2311__5071(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_1798__1798){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static uint16_t bincode_deserializeAlloc__anon_2422__5079(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 (void)a0;
 (void)a1;
 return 0;
}

static nav__5081_39 bincode_deserializeStructAlloc__anon_2488__5081(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5081_44 *t1;
 nav__5081_57 t2;
 nav__5081_39 t4;
 nav__5081_44 t5;
 struct shared_Command__struct_1799__1799 t6;
 struct shared_Command__struct_1799__1799 t0;
 uint16_t t3;
 t1 = (nav__5081_44 *)&t0.name;
 t2 = bincode_deserializeAlloc__anon_2349__5074(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1799__1799){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t1 = (nav__5081_44 *)&t0.version;
 t2 = bincode_deserializeAlloc__anon_2349__5074(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1799__1799){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = t0;
 t4.payload = t6;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__5080_39 bincode_deserializeAlloc__anon_2469__5080(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5080_39 t0;
 struct shared_Command__struct_1799__1799 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2488__5081(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_1799__1799){{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5087_38 bincode_deserializeInt__anon_2610__5087(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0) {
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *t4;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t2;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t0;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t3;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *t5;
 void const **t7;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t8;
 void const *t9;
 nav__5087_50 (**t10)(void const *, nav__5087_52);
 struct io_Reader__2238 t11;
 struct io_Reader__2238 t6;
 struct io_Reader__2238 t12;
 struct io_Reader__2238 t15;
 struct io_Reader__2238 const *t13;
 nav__5087_38 t14;
 nav__5087_38 t18;
 uint8_t const (*t21)[8];
 uint64_t t22;
 nav__5087_64 t16;
 uint16_t t17;
 uint8_t t19[8];
 uint8_t t20[8];
 t0 = a0;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5087_50 (**)(void const *, nav__5087_52))&t6.readFn;
 (*t10) = &io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Reader__2238 const *)&t12;
 t11 = (*t13);
 t15 = t11;
 t13 = (struct io_Reader__2238 const *)&t15;
 t11 = (*t13);
 t16 = io_Reader_readBytesNoEof__anon_2375__5076(t11);
 if (t16.error) {
  t17 = t16.error;
  t18.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t18.error = t17;
  t14 = t18;
  goto zig_block_0;
 }
 memcpy(t19, t16.payload, sizeof(uint8_t[8]));
 memcpy((char *)&t20, t19, sizeof(uint8_t[8]));
 t21 = (uint8_t const (*)[8])&t20;
 memcpy(t19, (const char *)t21, sizeof(uint8_t[8]));
 memcpy(&t22, &t19, sizeof(uint64_t));
 t22 = zig_wrap_u64(t22, UINT8_C(64));
 t18.payload = t22;
 t18.error = UINT16_C(0);
 t14 = t18;
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t18, &t14, sizeof(nav__5087_38));
 if (t18.error) {
  t17 = t18.error;
  t18.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t18.error = t17;
  return t18;
 }
 t22 = t18.payload;
 t18.payload = t22;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__5086_38 bincode_deserializeAlloc__anon_2605__5086(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5086_38 t0;
 uint64_t t2;
 uint16_t t1;
 (void)a1;
 t0 = bincode_deserializeInt__anon_2610__5087(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT64_C(0xaaaaaaaaaaaaaaaa);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5085_39 bincode_deserializeUnionAlloc__anon_2592__5085(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5085_39 t2;
 nav__5085_56 t6;
 uint64_t t7;
 struct shared_MarkerType__1802 t8;
 nav__5085_54 t0;
 uint32_t t3;
 uint16_t t1;
 uint8_t t4;
 bool t5;
 t0 = bincode_deserializeAlloc__anon_2212__5030(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct shared_MarkerType__1802){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t5 = t4 == UINT8_C(0);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2605__5086(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1802){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(0);
  t8.payload.SampleStart = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t4 == UINT8_C(1);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2605__5086(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1802){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(1);
  t8.payload.SampleEnd = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = t4 == UINT8_C(2);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2605__5086(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1802){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(2);
  t8.payload.BenchmarkStart = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_2;

 zig_block_2:;
 t5 = t4 == UINT8_C(3);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2605__5086(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_MarkerType__1802){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(3);
  t8.payload.BenchmarkEnd = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_3;

 zig_block_3:;
 zig_unreachable();
}

static nav__5084_39 bincode_deserializeAlloc__anon_2559__5084(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5084_39 t0;
 struct shared_MarkerType__1802 t2;
 uint16_t t1;
 t0 = bincode_deserializeUnionAlloc__anon_2592__5085(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_MarkerType__1802){{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5083_39 bincode_deserializeStructAlloc__anon_2541__5083(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 int32_t *t1;
 nav__5083_39 t4;
 struct shared_MarkerType__1802 *t6;
 nav__5083_60 t7;
 struct shared_MarkerType__1802 t8;
 struct shared_Command__struct_1800__1800 t9;
 struct shared_Command__struct_1800__1800 t0;
 nav__5083_57 t2;
 int32_t t5;
 uint16_t t3;
 t1 = (int32_t *)&t0.pid;
 t2 = bincode_deserializeAlloc__anon_2327__5072(a0, a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct shared_Command__struct_1800__1800){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t5 = t2.payload;
 (*t1) = t5;
 t6 = (struct shared_MarkerType__1802 *)&t0.marker;
 t7 = bincode_deserializeAlloc__anon_2559__5084(a0, a1);
 if (t7.error) {
  t3 = t7.error;
  t4.payload = (struct shared_Command__struct_1800__1800){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 t8 = t7.payload;
 (*t6) = t8;
 t9 = t0;
 t4.payload = t9;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__5082_39 bincode_deserializeAlloc__anon_2527__5082(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5082_39 t0;
 struct shared_Command__struct_1800__1800 t2;
 uint16_t t1;
 t0 = bincode_deserializeStructAlloc__anon_2541__5083(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__struct_1800__1800){{{UINT64_C(0xaaaaaaaaaaaaaaaa)},UINT8_C(0x2)},-INT32_C(0x55555556)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5089_38 bincode_deserializeEnum__anon_2695__5089(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0) {
 nav__5089_46 t0;
 uint32_t t3;
 uint16_t t1;
 nav__5089_38 t2;
 uint8_t t4;
 t0 = bincode_deserializeInt__anon_2223__5031(a0);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = UINT8_C(0x2);
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__5088_38 bincode_deserializeAlloc__anon_2665__5088(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5088_38 t0;
 uint16_t t1;
 uint8_t t2;
 (void)a1;
 t0 = bincode_deserializeEnum__anon_2695__5089(a0);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = UINT8_C(0x2);
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__5029_39 bincode_deserializeUnionAlloc__anon_2207__5029(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5029_39 t2;
 nav__5029_68 t6;
 struct shared_Command__struct_1798__1798 t7;
 struct shared_Command__1795 t8;
 nav__5029_70 t9;
 struct shared_Command__struct_1799__1799 t10;
 nav__5029_72 t11;
 struct shared_Command__struct_1800__1800 t12;
 nav__5029_74 t13;
 uint64_t t14;
 nav__5029_66 t0;
 uint32_t t3;
 uint16_t t1;
 nav__5029_76 t15;
 uint8_t t4;
 bool t5;
 t0 = bincode_deserializeAlloc__anon_2212__5030(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t2.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t2.error = t1;
  return t2;
 }
 t3 = t0.payload;
 t4 = (uint8_t)t3;
 t5 = t4 == UINT8_C(0);
 if (t5) {
  t6 = bincode_deserializeAlloc__anon_2282__5070(a0, a1);
  if (t6.error) {
   t1 = t6.error;
   t2.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t7 = t6.payload;
  t8.tag = UINT8_C(0);
  t8.payload.ExecutedBenchmark = t7;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = t4 == UINT8_C(1);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2422__5079(a0, a1);
  return (nav__5029_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(1)},0};
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = t4 == UINT8_C(2);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2422__5079(a0, a1);
  return (nav__5029_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(2)},0};
 }
 goto zig_block_2;

 zig_block_2:;
 t5 = t4 == UINT8_C(3);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2422__5079(a0, a1);
  return (nav__5029_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(3)},0};
 }
 goto zig_block_3;

 zig_block_3:;
 t5 = t4 == UINT8_C(4);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2422__5079(a0, a1);
  return (nav__5029_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(4)},0};
 }
 goto zig_block_4;

 zig_block_4:;
 t5 = t4 == UINT8_C(5);
 if (t5) {
  t9 = bincode_deserializeAlloc__anon_2469__5080(a0, a1);
  if (t9.error) {
   t1 = t9.error;
   t2.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t10 = t9.payload;
  t8.tag = UINT8_C(5);
  t8.payload.SetIntegration = t10;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_5;

 zig_block_5:;
 t5 = t4 == UINT8_C(6);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2422__5079(a0, a1);
  return (nav__5029_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(6)},0};
 }
 goto zig_block_6;

 zig_block_6:;
 t5 = t4 == UINT8_C(7);
 if (t5) {
  t11 = bincode_deserializeAlloc__anon_2527__5082(a0, a1);
  if (t11.error) {
   t1 = t11.error;
   t2.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t12 = t11.payload;
  t8.tag = UINT8_C(7);
  t8.payload.AddMarker = t12;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_7;

 zig_block_7:;
 t5 = t4 == UINT8_C(8);
 if (t5) {
  t13 = bincode_deserializeAlloc__anon_2605__5086(a0, a1);
  if (t13.error) {
   t1 = t13.error;
   t2.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t14 = t13.payload;
  t8.tag = UINT8_C(8);
  t8.payload.SetVersion = t14;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_8;

 zig_block_8:;
 t5 = t4 == UINT8_C(9);
 if (t5) {
  (void)bincode_deserializeAlloc__anon_2422__5079(a0, a1);
  return (nav__5029_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(9)},0};
 }
 goto zig_block_9;

 zig_block_9:;
 t5 = t4 == UINT8_C(10);
 if (t5) {
  t15 = bincode_deserializeAlloc__anon_2665__5088(a0, a1);
  if (t15.error) {
   t1 = t15.error;
   t2.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t2.error = t1;
   return t2;
  }
  t4 = t15.payload;
  t8.tag = UINT8_C(10);
  t8.payload.IntegrationModeResponse = t4;
  t2.payload = t8;
  t2.error = UINT16_C(0);
  return t2;
 }
 goto zig_block_10;

 zig_block_10:;
 zig_unreachable();
}

static nav__5028_39 bincode_deserializeAlloc__anon_2097__5028(struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 const a0, struct mem_Allocator__599 const a1) {
 nav__5028_39 t0;
 struct shared_Command__1795 t2;
 uint16_t t1;
 t0 = bincode_deserializeUnionAlloc__anon_2207__5029(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static nav__1291_39 fifo_posix_Pipe_Reader_recvCmd__1291(struct fifo_posix_Pipe_Reader__793 *const a0) {
 struct fifo_posix_Pipe_Reader__793 *const *t1;
 uintptr_t t3;
 uintptr_t t10;
 struct fifo_posix_Pipe_Reader__793 *t4;
 struct fifo_posix_Pipe_Reader__793 *t0;
 struct fs_File__799 *t5;
 nav__1291_68 t7;
 nav__1291_76 t8;
 nav__1291_39 t12;
 nav__1291_39 t13;
 uint64_t t14;
 uint64_t t20;
 uint8_t const (*t15)[4];
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t18;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t19;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t22;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t21;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t23;
 struct mem_Allocator__599 *t24;
 struct mem_Allocator__599 t25;
 struct shared_Command__1795 t26;
 struct fs_File__799 t6;
 uint32_t t17;
 uint16_t t11;
 bool t9;
 uint8_t t16[4];
 uint8_t t2[4];
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__793 *const *)&t0;
 t4 = (*t1);
 t5 = (struct fs_File__799 *)&t4->file;
 t6 = (*t5);
 t7.ptr = &t2[(uintptr_t)0ul];
 t7.len = (uintptr_t)4ul;
 t8 = fs_File_readAll__1429(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_0;
 }
 t11 = t8.error;
 switch (t11) {
  case zig_error_WouldBlock:
  case zig_error_BrokenPipe: {
   t12 = (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_NotReady};
   goto zig_block_1;
  }
  default: {
   t13.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t13.error = t11;
   t12 = t13;
   goto zig_block_1;
  }
 }

 zig_block_1:;
 return t12;

 zig_block_0:;
 t14 = t3;
 t9 = t14 < UINT64_C(4);
 if (t9) {
  return (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_UnexpectedEof};
 }
 goto zig_block_2;

 zig_block_2:;
 t15 = (uint8_t const (*)[4])&t2;
 memcpy(t16, (const char *)t15, sizeof(uint8_t[4]));
 memcpy(&t17, &t16, sizeof(uint32_t));
 t17 = zig_wrap_u32(t17, UINT8_C(32));
 t4 = (*t1);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t4->buffer;
 t3 = (uintptr_t)t17;
 t11 = array_list_ArrayListAligned_28u8_2cnull_29_resize__1559(t18, t3);
 if (t11) {
  t13.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t13.error = t11;
  return t13;
 }
 t4 = (*t1);
 t5 = (struct fs_File__799 *)&t4->file;
 t6 = (*t5);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&a0->buffer;
 t19 = (*t18);
 t7 = t19.items;
 t8 = fs_File_readAll__1429(t6, t7);
 t9 = t8.error == UINT16_C(0);
 if (t9) {
  t10 = t8.payload;
  t3 = t10;
  goto zig_block_3;
 }
 t11 = t8.error;
 switch (t11) {
  case zig_error_WouldBlock:
  case zig_error_BrokenPipe: {
   t13 = (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_NotReady};
   goto zig_block_4;
  }
  default: {
   t12.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
   t12.error = t11;
   t13 = t12;
   goto zig_block_4;
  }
 }

 zig_block_4:;
 return t13;

 zig_block_3:;
 t14 = t3;
 t20 = (uint64_t)t17;
 t9 = t14 < t20;
 if (t9) {
  return (nav__1291_39){{{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)},zig_error_UnexpectedEof};
 }
 goto zig_block_5;

 zig_block_5:;
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&a0->buffer;
 t19 = (*t18);
 t7 = t19.items;
 t22 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(t7);
 t21 = t22;
 t23 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4983(&t21);
 t24 = (struct mem_Allocator__599 *)&a0->allocator;
 t25 = (*t24);
 t13 = bincode_deserializeAlloc__anon_2097__5028(t23, t25);
 if (t13.error) {
  t11 = t13.error;
  t13.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t13.error = t11;
  return t13;
 }
 t26 = t13.payload;
 t13.payload = t26;
 t13.error = UINT16_C(0);
 return t13;
}

static nav__1292_39 fifo_posix_Pipe_Reader_waitForResponse__1292(struct fifo_posix_Pipe_Reader__793 *const a0, nav__1292_42 const a1) {
 struct fifo_posix_Pipe_Reader__793 *const *t1;
 uint64_t t2;
 uint64_t t4;
 struct fs_File__799 *t5;
 nav__1292_39 t9;
 nav__1292_39 t11;
 struct fifo_posix_Pipe_Reader__793 *t10;
 struct fifo_posix_Pipe_Reader__793 *t0;
 struct fs_File__799 t6;
 int32_t t7;
 uint16_t t8;
 bool t3;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__793 *const *)&t0;
 t3 = a1.is_null != true;
 if (t3) {
  t4 = a1.payload;
  t2 = t4;
  goto zig_block_0;
 }
 t2 = UINT64_C(1000000000);
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__799 *)&a0->file;
 t6 = (*t5);
 t7 = t6.handle;
 t8 = fifo_posix_waitReadable__1268(t7, t2);
 if (t8) {
  t9.payload = (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(0xa)};
  t9.error = t8;
  return t9;
 }
 t10 = (*t1);
 t9 = fifo_posix_Pipe_Reader_recvCmd__1291(t10);
 memcpy(&t11, &t9, sizeof(nav__1292_39));
 return t11;
}

static uint16_t fifo_posix_Pipe_Reader_waitForAck__1293(struct fifo_posix_Pipe_Reader__793 *const a0, nav__1293_40 const a1) {
 struct fifo_posix_Pipe_Reader__793 *const *t1;
 struct fifo_posix_Pipe_Reader__793 *t2;
 struct fifo_posix_Pipe_Reader__793 *t0;
 nav__1293_60 t3;
 struct shared_Command__1795 t5;
 struct shared_Command__1795 t6;
 struct shared_Command__1795 const *t7;
 struct mem_Allocator__599 *t9;
 struct mem_Allocator__599 t10;
 nav__1293_77 t11;
 uint16_t t4;
 uint8_t t8;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__793 *const *)&t0;
 t2 = (*t1);
 t3 = fifo_posix_Pipe_Reader_waitForResponse__1292(t2, a1);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 t6 = t5;
 t7 = (struct shared_Command__1795 const *)&t6;
 t8 = t5.tag;
 switch (t8) {
  case UINT8_C(3): {
   t5 = (*t7);
   t9 = (struct mem_Allocator__599 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__4786(t5, t10);
   return 0;
  }
  case UINT8_C(6): {
   t5 = (*t7);
   t9 = (struct mem_Allocator__599 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__4786(t5, t10);
   return zig_error_UnexpectedError;
  }
  default: {
   t11.f0 = t5;
   logger_debug__anon_2729__5099(t11);
   t5 = (*t7);
   t9 = (struct mem_Allocator__599 *)&a0->allocator;
   t10 = (*t9);
   shared_Command_deinit__4786(t5, t10);
   return zig_error_UnexpectedResponse;
  }
 }
}

static uint16_t runner_fifo_RunnerFifo_send_version__1139(struct runner_fifo_RunnerFifo__631 *const a0, uint64_t const a1) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 struct shared_Command__1795 t4;
 struct fifo_posix_Pipe_Reader__793 *t6;
 uint16_t t5;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4.tag = UINT8_C(8);
 t4.payload.SetVersion = a1;
 t5 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t4);
 if (t5) {
  return t5;
 }
 t2 = (*t1);
 t6 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t5 = fifo_posix_Pipe_Reader_waitForAck__1293(t6, (nav__1139_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t runner_fifo_RunnerFifo_validate_protocol_version__1130(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 nav__1130_43 t5;
 nav__1130_45 t6;
 uint16_t t3;
 bool t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = runner_fifo_RunnerFifo_send_version__1139(t2, UINT64_C(2));
 t4 = t3 == UINT16_C(0);
 if (t4) {
  goto zig_block_0;
 }
 switch (t3) {
  case zig_error_AckTimeout: {
   return 0;
  }
  case zig_error_UnexpectedError: {
   logger_err__anon_2753__5100();
   logger_err__anon_2762__5101();
   posix_exit__1747(UINT8_C(1));
   zig_unreachable();
  }
  default: {
   t5 = zig_errorName[t3 - 1];
   t6.f0 = t5;
   logger_err__anon_2776__5102(t6);
   posix_exit__1747(UINT8_C(1));
   zig_unreachable();
  }
 }

 zig_block_0:;
 return 0;
}

static nav__1140_38 runner_fifo_RunnerFifo_get_integration_mode__1140(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 struct fifo_posix_Pipe_Reader__793 *t6;
 nav__1140_81 t7;
 struct shared_Command__1795 t8;
 struct shared_Command__1795 t9;
 struct shared_Command__1795 const *t10;
 struct mem_Allocator__599 *t13;
 struct mem_Allocator__599 t14;
 uint16_t t4;
 nav__1140_38 t5;
 uint8_t t11;
 bool t12;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(9)});
 if (t4) {
  t5.payload = UINT8_C(0x2);
  t5.error = t4;
  return t5;
 }
 t2 = (*t1);
 t6 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t7 = fifo_posix_Pipe_Reader_waitForResponse__1292(t6, (nav__1140_79){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t7.error) {
  t4 = t7.error;
  t5.payload = UINT8_C(0x2);
  t5.error = t4;
  return t5;
 }
 t8 = t7.payload;
 t9 = t8;
 t10 = (struct shared_Command__1795 const *)&t9;
 t11 = t8.tag;
 t12 = t11 == UINT8_C(10);
 if (t12) {
  t11 = t8.payload.IntegrationModeResponse;
  t8 = (*t10);
  t13 = (struct mem_Allocator__599 *)&a0->allocator;
  t14 = (*t13);
  shared_Command_deinit__4786(t8, t14);
  t5.payload = t11;
  t5.error = UINT16_C(0);
  return t5;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (*t10);
 t13 = (struct mem_Allocator__599 *)&a0->allocator;
 t14 = (*t13);
 shared_Command_deinit__4786(t8, t14);
 return (nav__1140_38){zig_error_UnexpectedResponse,UINT8_C(0x2)};
}

static nav__820_39 instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__820(struct mem_Allocator__599 const a0) {
 nav__820_62 t1;
 nav__820_39 t3;
 struct runner_fifo_RunnerFifo__631 t4;
 struct runner_fifo_RunnerFifo__631 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 t9;
 uint16_t t2;
 nav__820_65 t6;
 uint8_t t5;
 uint8_t t8;
 bool t7;
 t1 = runner_fifo_RunnerFifo_init__1129(a0);
 if (t1.error) {
  t2 = t1.error;
  t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
  t3.error = t2;
  return t3;
 }
 t4 = t1.payload;
 t0 = t4;
 (void)runner_fifo_RunnerFifo_validate_protocol_version__1130(&t0);
 t6 = runner_fifo_RunnerFifo_get_integration_mode__1140(&t0);
 t7 = t6.error == UINT16_C(0);
 if (t7) {
  t8 = t6.payload;
  t5 = t8;
  goto zig_block_0;
 }
 t2 = t6.error;
 runner_fifo_RunnerFifo_deinit__1131(&t0);
 t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
 t3.error = t2;
 return t3;

 zig_block_0:;
 t7 = t5 != UINT8_C(2);
 if (t7) {
  runner_fifo_RunnerFifo_deinit__1131(&t0);
  return (nav__820_39){{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}},zig_error_ModeError};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t0;
 t9.fifo = t4;
 t3.payload = t9;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__807_39 instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_init__807(struct mem_Allocator__599 const a0) {
 nav__807_62 t1;
 nav__807_39 t3;
 struct runner_fifo_RunnerFifo__631 t4;
 struct runner_fifo_RunnerFifo__631 t0;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 t9;
 uint16_t t2;
 nav__807_65 t6;
 uint8_t t5;
 uint8_t t8;
 bool t7;
 t1 = runner_fifo_RunnerFifo_init__1129(a0);
 if (t1.error) {
  t2 = t1.error;
  t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
  t3.error = t2;
  return t3;
 }
 t4 = t1.payload;
 t0 = t4;
 (void)runner_fifo_RunnerFifo_validate_protocol_version__1130(&t0);
 t6 = runner_fifo_RunnerFifo_get_integration_mode__1140(&t0);
 t7 = t6.error == UINT16_C(0);
 if (t7) {
  t8 = t6.payload;
  t5 = t8;
  goto zig_block_0;
 }
 t2 = t6.error;
 runner_fifo_RunnerFifo_deinit__1131(&t0);
 t3.payload = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582){{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}};
 t3.error = t2;
 return t3;

 zig_block_0:;
 t7 = t5 != UINT8_C(0);
 if (t7) {
  runner_fifo_RunnerFifo_deinit__1131(&t0);
  return (nav__807_39){{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}},{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},{{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},(uintptr_t)0xaaaaaaaaaaaaaaaaul,{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}},{-INT32_C(0x55555556)}}}},zig_error_ModeError};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t0;
 t9.fifo = t4;
 t3.payload = t9;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__743_39 instruments_root_Instrument_init__743(struct mem_Allocator__599 const a0) {
 nav__743_69 t0;
 struct instruments_valgrind_ValgrindInstrument__564 t2;
 struct instruments_root_Instrument__554 t3;
 nav__743_39 t4;
 nav__743_71 t5;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 t6;
 nav__743_73 t7;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 t8;
 bool t1;
 t0 = instruments_valgrind_ValgrindInstrument_init__786(a0);
 t1 = t0.error == UINT16_C(0);
 if (t1) {
  t2 = t0.payload;
  t3.tag = UINT8_C(0);
  t3.payload.valgrind = t2;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_init__820(a0);
 t1 = t5.error == UINT16_C(0);
 if (t1) {
  t6 = t5.payload;
  t3.tag = UINT8_C(2);
  t3.payload.analysis = t6;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_1;

 zig_block_1:;
 t7 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_init__807(a0);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t8 = t7.payload;
  t3.tag = UINT8_C(1);
  t3.payload.walltime = t8;
  t4.payload = t3;
  t4.error = UINT16_C(0);
  return t4;
 }
 goto zig_block_2;

 zig_block_2:;
 return (nav__743_39){{{{{((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)}}},UINT8_C(3)},0};
}

static nav__730_39 instrument_hooks_InstrumentHooks_init__730(struct mem_Allocator__599 const a0) {
 struct InstrumentHooks *t1;
 struct instruments_root_Instrument__554 *t2;
 nav__730_100 t3;
 struct instruments_root_Instrument__554 t4;
 struct environment_root_Environment__559 *t5;
 struct environment_root_Environment__559 t6;
 nav__730_39 t0;
 t0.error = UINT16_C(0);
 t1 = &t0.payload;
 t2 = (struct instruments_root_Instrument__554 *)&t1->instrument;
 t3 = instruments_root_Instrument_init__743(a0);
 t4 = t3.payload;
 (*t2) = t4;
 t5 = (struct environment_root_Environment__559 *)&t1->environment;
 t6 = environment_root_Environment_init__773(a0);
 (*t5) = t6;
 return t0;
}

struct InstrumentHooks *c_instrument_hooks_init__238(void) {
 struct InstrumentHooks *t0;
 struct InstrumentHooks *t3;
 nav__238_46 t1;
 nav__238_107 t4;
 struct InstrumentHooks t5;
 bool t2;
 t1 = mem_Allocator_create__anon_1051__4541((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)});
 t2 = t1.error == UINT16_C(0);
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 return NULL;

 zig_block_0:;
 t4 = instrument_hooks_InstrumentHooks_init__730((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)});
 t5 = t4.payload;
 (*t0) = t5;
 t3 = (struct InstrumentHooks *)t0;
 return t3;
}

static nav__4541_40 mem_Allocator_create__anon_1051__4541(struct mem_Allocator__599 const a0) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 uintptr_t t3;
 nav__4541_51 t4;
 nav__4541_40 t6;
 uint8_t *t7;
 struct InstrumentHooks *t8;
 uint16_t t5;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t4 = mem_Allocator_allocBytesWithAlignment__anon_2823__5103(t2, (uintptr_t)288ul, t3);
 if (t4.error) {
  t5 = t4.error;
  t6.payload = ((struct InstrumentHooks *)(uintptr_t)0xaaaaaaaaaaaaaaaaul);
  t6.error = t5;
  return t6;
 }
 t7 = t4.payload;
 t8 = (struct InstrumentHooks *)t7;
 t6.payload = t8;
 t6.error = UINT16_C(0);
 return t6;
}

static uint16_t fs_accessAbsolute__1348(nav__1348_39 const a0, struct fs_File_OpenFlags__1718 const a1) {
 struct fs_Dir__1720 const *t3;
 struct fs_Dir__1720 t1;
 struct fs_Dir__1720 t2;
 uint16_t t4;
 bool t0;
 t0 = fs_path_isAbsolute__5138(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1340();
 t2 = t1;
 t3 = (struct fs_Dir__1720 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_access__4742(t1, a0, a1);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__1345_39 fs_openFileAbsolute__1345(nav__1345_41 const a0, struct fs_File_OpenFlags__1718 const a1) {
 struct fs_Dir__1720 const *t3;
 struct fs_Dir__1720 t1;
 struct fs_Dir__1720 t2;
 nav__1345_39 t4;
 bool t0;
 t0 = fs_path_isAbsolute__5138(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1340();
 t2 = t1;
 t3 = (struct fs_Dir__1720 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_openFile__4678(t1, a0, a1);
 return t4;
}

static struct fifo_posix_Pipe_Writer__791 fifo_posix_Pipe_Writer_init__1283(struct fs_File__799 const a0, struct mem_Allocator__599 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t0;
 struct fs_File__799 *t5;
 struct mem_Allocator__599 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t7;
 struct fifo_posix_Pipe_Writer__791 t4;
 uint16_t t2;
 bool t3;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(a1);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(&t0, (uintptr_t)1024ul);
 t3 = t2 == UINT16_C(0);
 if (t3) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__799 *)&t4.file;
 (*t5) = a0;
 t6 = (struct mem_Allocator__599 *)&t4.allocator;
 (*t6) = a1;
 t7 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t4.buffer;
 t1 = t0;
 (*t7) = t1;
 return t4;
}

static struct fifo_posix_Pipe_Reader__793 fifo_posix_Pipe_Reader_init__1288(struct fs_File__799 const a0, struct mem_Allocator__599 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t0;
 struct fs_File__799 *t5;
 struct mem_Allocator__599 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t7;
 struct fifo_posix_Pipe_Reader__793 t4;
 uint16_t t2;
 bool t3;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(a1);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(&t0, (uintptr_t)1024ul);
 t3 = t2 == UINT16_C(0);
 if (t3) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (struct fs_File__799 *)&t4.file;
 (*t5) = a0;
 t6 = (struct mem_Allocator__599 *)&t4.allocator;
 (*t6) = a1;
 t7 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t4.buffer;
 t1 = t0;
 (*t7) = t1;
 return t4;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_clearRetainingCapacity__1562(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 nav__1562_43 *t3;
 uintptr_t *t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__1562_43 *)&t2->items;
 t4 = &t3->len;
 (*t4) = (uintptr_t)0ul;
 return;
}

static struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 **t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static uint16_t bincode_serialize__anon_1849__4913(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__1795 const a1) {
 uint16_t t0;
 t0 = bincode_serializeUnion__anon_2865__5201(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint8_t const (*mem_asBytes__anon_1887__4914(uint32_t const *const a0))[4] {
 uint8_t const (*t0)[4];
 t0 = (uint8_t const (*)[4])a0;
 return t0;
}

static uint16_t fs_File_writeAll__1439(struct fs_File__799 const a0, nav__1439_40 const a1) {
 struct fs_File__799 const *t1;
 nav__1439_40 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 nav__1439_40 t11;
 nav__1439_40 t2;
 uint8_t const *t12;
 nav__1439_47 t14;
 struct fs_File__799 t10;
 struct fs_File__799 t0;
 uint16_t t15;
 bool t9;
 t0 = a0;
 t1 = (struct fs_File__799 const *)&t0;
 t2 = a1;
 t3 = (nav__1439_40 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 < t8;
 if (t9) {
  t6 = t4;
  t10 = (*t1);
  t5 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t const *)(((uintptr_t)t12) + (t5*sizeof(uint8_t)));
  t13 = t11.len;
  t5 = t13 - t5;
  t11.ptr = t12;
  t11.len = t5;
  t14 = fs_File_write__1438(t10, t11);
  if (t14.error) {
   t15 = t14.error;
   return t15;
  }
  t5 = t14.payload;
  t5 = t6 + t5;
  t4 = t5;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 return 0;
}

static nav__1964_38 posix_poll__1964(nav__1964_41 const a0, int32_t const a1) {
 uintptr_t t0;
 uintptr_t t1;
 nav__1964_46 t2;
 struct os_linux_pollfd__1921 *t4;
 nav__1964_38 t7;
 int t5;
 uint16_t t6;
 bool t3;
 zig_loop_5:
 t1 = a0.len;
 t2 = math_cast__anon_2885__5202(t1);
 t3 = t2.is_null != true;
 if (t3) {
  t1 = t2.payload;
  t0 = t1;
  goto zig_block_1;
 }
 return (nav__1964_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};

 zig_block_1:;
 t4 = a0.ptr;
 t5 = a1;
 t5 = poll(t4, t0, t5);
 t6 = posix_errno__anon_3044__5205(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(12): {
   return (nav__1964_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  default: {
   t6 = posix_unexpectedErrno__2021(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_0:;
 goto zig_loop_5;
}

static nav__1429_38 fs_File_readAll__1429(struct fs_File__799 const a0, nav__1429_41 const a1) {
 struct fs_File__799 const *t1;
 nav__1429_41 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 nav__1429_41 t11;
 nav__1429_41 t2;
 uint8_t *t12;
 nav__1429_38 t13;
 struct fs_File__799 t10;
 struct fs_File__799 t0;
 uint16_t t14;
 bool t9;
 t0 = a0;
 t1 = (struct fs_File__799 const *)&t0;
 t2 = a1;
 t3 = (nav__1429_41 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 != t8;
 if (t9) {
  t10 = (*t1);
  t6 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t *)(((uintptr_t)t12) + (t6*sizeof(uint8_t)));
  t5 = t11.len;
  t6 = t5 - t6;
  t11.ptr = t12;
  t11.len = t6;
  t13 = fs_File_read__1428(t10, t11);
  if (t13.error) {
   t14 = t13.error;
   t13.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t13.error = t14;
   return t13;
  }
  t6 = t13.payload;
  t8 = t6;
  t9 = t8 == UINT64_C(0);
  if (t9) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t5 = t4;
  t6 = t5 + t6;
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 t6 = t4;
 t13.payload = t6;
 t13.error = UINT16_C(0);
 return t13;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_resize__1559(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 nav__1559_43 *t4;
 uintptr_t *t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(t2, a1);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (nav__1559_43 *)&t2->items;
 t5 = &t4->len;
 (*t5) = a1;
 return 0;
}

static struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(nav__4994_40 const a0) {
 nav__4994_40 *t1;
 uintptr_t *t2;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 t1 = (nav__4994_40 *)&t0.buffer;
 (*t1) = a0;
 t2 = (uintptr_t *)&t0.pos;
 (*t2) = (uintptr_t)0ul;
 return t0;
}

static struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_reader__4983(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 **t1;
 struct io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29__2089 t0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__5027_38 io_GenericReader_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7b_7d_2c_28function_20_27read_27_29_29_typeEras__5027(void const *const a0, nav__5027_41 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t0;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *t1;
 nav__5027_38 t2;
 nav__5027_38 t3;
 t0 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)a0;
 t1 = (*t0);
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4986(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__5027_38));
 return t3;
}

static nav__5069_39 io_Reader_readBytesNoEof__anon_2269__5069(struct io_Reader__2238 const a0) {
 struct io_Reader__2238 const *t1;
 struct io_Reader__2238 t3;
 struct io_Reader__2238 t0;
 nav__5069_46 t4;
 uint16_t t5;
 nav__5069_39 t6;
 uint8_t t7[4];
 uint8_t t2[4];
 t0 = a0;
 t1 = (struct io_Reader__2238 const *)&t0;
 t3 = (*t1);
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)4ul;
 t5 = io_Reader_readNoEof__5036(t3, t4);
 if (t5) {
  memcpy(t6.payload, "\252\252\252\252", sizeof(uint8_t[4]));
  t6.error = t5;
  return t6;
 }
 memcpy(t7, (const char *)&t2, sizeof(uint8_t[4]));
 memcpy(t6.payload, t7, sizeof(uint8_t[4]));
 t6.error = UINT16_C(0);
 return t6;
}

static nav__5076_39 io_Reader_readBytesNoEof__anon_2375__5076(struct io_Reader__2238 const a0) {
 struct io_Reader__2238 const *t1;
 struct io_Reader__2238 t3;
 struct io_Reader__2238 t0;
 nav__5076_46 t4;
 uint16_t t5;
 nav__5076_39 t6;
 uint8_t t7[8];
 uint8_t t2[8];
 t0 = a0;
 t1 = (struct io_Reader__2238 const *)&t0;
 t3 = (*t1);
 t4.ptr = &t2[(uintptr_t)0ul];
 t4.len = (uintptr_t)8ul;
 t5 = io_Reader_readNoEof__5036(t3, t4);
 if (t5) {
  memcpy(t6.payload, "\252\252\252\252\252\252\252\252", sizeof(uint8_t[8]));
  t6.error = t5;
  return t6;
 }
 memcpy(t7, (const char *)&t2, sizeof(uint8_t[8]));
 memcpy(t6.payload, t7, sizeof(uint8_t[8]));
 t6.error = UINT16_C(0);
 return t6;
}

static nav__5077_40 mem_Allocator_alloc__anon_2384__5077(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__5077_40 t4;
 nav__5077_40 t8;
 nav__5077_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__5077_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_3080__5206(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__5077_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static nav__5034_38 io_Reader_readAll__5034(struct io_Reader__2238 const a0, nav__5034_41 const a1) {
 uintptr_t t0;
 nav__5034_38 t1;
 t0 = a1.len;
 t1 = io_Reader_readAtLeast__5035(a0, a1, t0);
 return t1;
}

static zig_noreturn void bincode_invalidProtocol__anon_2407__5078(void) {
 utils_print__anon_3102__5217();
 posix_exit__1747(UINT8_C(1));
 zig_unreachable();
}

static void shared_Command_deinit__4786(struct shared_Command__1795 const a0, struct mem_Allocator__599 const a1) {
 struct mem_Allocator__599 const *t1;
 struct shared_Command__struct_1799__1799 t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t0;
 nav__4786_43 t5;
 struct shared_Command__struct_1798__1798 t6;
 uint8_t t2;
 t0 = a1;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(5): {
   t3 = a0.payload.SetIntegration;
   t4 = (*t1);
   t5 = t3.name;
   mem_Allocator_free__anon_3116__5296(t4, t5);
   t4 = (*t1);
   t5 = t3.version;
   mem_Allocator_free__anon_3116__5296(t4, t5);
   goto zig_block_0;
  }
  case UINT8_C(0): {
   t6 = a0.payload.ExecutedBenchmark;
   t4 = (*t1);
   t5 = t6.uri;
   mem_Allocator_free__anon_3116__5296(t4, t5);
   goto zig_block_0;
  }
  case UINT8_C(8): {
   goto zig_block_0;
  }
  case UINT8_C(9): {
   goto zig_block_0;
  }
  case UINT8_C(10): {
   goto zig_block_0;
  }
  default: {
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return;
}

static void logger_debug__anon_2729__5099(nav__5099_39 const a0) {
 logger_logWithPrefix__anon_3124__5297(a0);
 return;
}

static void logger_err__anon_2753__5100(void) {
 logger_logWithPrefix__anon_3126__5298();
 return;
}

static void logger_err__anon_2762__5101(void) {
 logger_logWithPrefix__anon_3127__5299();
 return;
}

static zig_noreturn void posix_exit__1747(uint8_t const a0) {
 int t0;
 t0 = (int)a0;
 exit(t0);
 zig_unreachable();
}

static void logger_err__anon_2776__5102(nav__5102_40 const a0) {
 logger_logWithPrefix__anon_3132__5301(a0);
 return;
}

static void runner_fifo_RunnerFifo_deinit__1131(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 struct fifo_posix_Pipe_Reader__793 *t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 fifo_posix_Pipe_Writer_deinit__1287(t3);
 t2 = (*t1);
 t4 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 fifo_posix_Pipe_Reader_deinit__1294(t4);
 return;
}

static struct environment_root_Environment__559 environment_root_Environment_init__773(struct mem_Allocator__599 const a0) {
 struct mem_Allocator__599 *t1;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t2;
 struct environment_linked_libraries_root_LinkedLibraries__849 t3;
 struct environment_root_EnvironmentJson__845 *t4;
 struct environment_root_Environment__559 t0;
 t1 = (struct mem_Allocator__599 *)&t0.allocator;
 (*t1) = a0;
 t2 = (struct environment_linked_libraries_root_LinkedLibraries__849 *)&t0.libs;
 t3 = environment_linked_libraries_root_LinkedLibraries_init__3643(a0);
 (*t2) = t3;
 t4 = (struct environment_root_EnvironmentJson__845 *)&t0.data;
 (*t4) = (struct environment_root_EnvironmentJson__845){{{{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul,(uintptr_t)0ul},NULL}},{{{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul,(uintptr_t)0ul},NULL}}};
 return t0;
}

static nav__5103_39 mem_Allocator_allocBytesWithAlignment__anon_2823__5103(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t7;
 struct mem_Allocator_VTable__602 const *const *t8;
 struct mem_Allocator_VTable__602 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__5103_52 t17;
 nav__5103_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__5103_39){(uint8_t *)(uintptr_t)0xfffffffffffffff8ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t7 = t5;
 t1 = (struct mem_Allocator__599 const *)&t7;
 t8 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__5103_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t14 = (uint8_t *)t4;
 t18.payload = t14;
 t18.error = UINT16_C(0);
 return t18;
}

static bool fs_path_isAbsolute__5138(nav__5138_39 const a0) {
 bool t0;
 t0 = fs_path_isAbsolutePosix__5144(a0);
 return t0;
}

static struct fs_Dir__1720 fs_cwd__1340(void) {
 struct fs_Dir__1720 t0;
 t0 = (struct fs_Dir__1720){-INT32_C(100)};
 return t0;
}

static uint16_t fs_Dir_access__4742(struct fs_Dir__1720 const a0, nav__4742_40 const a1, struct fs_File_OpenFlags__1718 const a2) {
 struct fs_Dir__1720 const *t1;
 uint8_t const (*t6)[4096];
 uint8_t const *t8;
 struct fs_Dir__1720 t7;
 struct fs_Dir__1720 t0;
 nav__4742_49 t2;
 uint16_t t3;
 uint8_t t4[4096];
 uint8_t t5[4096];
 t0 = a0;
 t1 = (struct fs_Dir__1720 const *)&t0;
 t2 = posix_toPosixPath__2022(a1);
 if (t2.error) {
  t3 = t2.error;
  return t3;
 }
 memcpy(t4, t2.payload, sizeof(uint8_t[4096]));
 memcpy((char *)&t5, t4, sizeof(uint8_t[4096]));
 t6 = (uint8_t const (*)[4096])&t5;
 t7 = (*t1);
 t8 = (uint8_t const *)t6;
 t3 = fs_Dir_accessZ__4743(t7, t8, a2);
 return t3;
}

static nav__4678_39 fs_Dir_openFile__4678(struct fs_Dir__1720 const a0, nav__4678_42 const a1, struct fs_File_OpenFlags__1718 const a2) {
 struct fs_Dir__1720 const *t1;
 uint8_t const (*t8)[4096];
 uint8_t const *t10;
 nav__4678_39 t5;
 struct fs_Dir__1720 t9;
 struct fs_Dir__1720 t0;
 nav__4678_53 t3;
 uint16_t t4;
 uint8_t t6[4096];
 uint8_t t7[4096];
 struct fs_File_OpenFlags__1718 t2;
 t0 = a0;
 t1 = (struct fs_Dir__1720 const *)&t0;
 t2 = a2;
 t3 = posix_toPosixPath__2022(a1);
 if (t3.error) {
  t4 = t3.error;
  t5.payload = (struct fs_File__799){-INT32_C(0x55555556)};
  t5.error = t4;
  return t5;
 }
 memcpy(t6, t3.payload, sizeof(uint8_t[4096]));
 memcpy((char *)&t7, t6, sizeof(uint8_t[4096]));
 t8 = (uint8_t const (*)[4096])&t7;
 t9 = (*t1);
 t10 = (uint8_t const *)t8;
 t5 = fs_Dir_openFileZ__4679(t9, t10, a2);
 return t5;
}

static struct array_list_ArrayListAligned_28u8_2cnull_29__811 array_list_ArrayListAligned_28u8_2cnull_29_init__1527(struct mem_Allocator__599 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t0;
 t0.items = (nav__1527_42){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 t0.capacity = (uintptr_t)0ul;
 t0.allocator = a0;
 return t0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__5372(t3, a1);
 t7 = (*t1);
 t8 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1565(t7, t3);
 return t8;
}

static uint16_t bincode_serializeUnion__anon_2865__5201(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__1795 const a1) {
 struct shared_Command__struct_1798__1798 t4;
 struct shared_Command__struct_1799__1799 t5;
 struct shared_Command__struct_1800__1800 t6;
 uint64_t t7;
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 bool t3;
 t0 = a1.tag;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3242__5373(a0, t1);
 if (t2) {
  return t2;
 }
 t0 = a1.tag;
 t3 = t0 == UINT8_C(0);
 if (t3) {
  t4 = a1.payload.ExecutedBenchmark;
  t2 = bincode_serialize__anon_3244__5374(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(1);
 if (t3) {
  t2 = bincode_serialize__anon_3246__5375(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(2);
 if (t3) {
  t2 = bincode_serialize__anon_3246__5375(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(3);
 if (t3) {
  t2 = bincode_serialize__anon_3246__5375(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(4);
 if (t3) {
  t2 = bincode_serialize__anon_3246__5375(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_4;
 }
 goto zig_block_4;

 zig_block_4:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(5);
 if (t3) {
  t5 = a1.payload.SetIntegration;
  t2 = bincode_serialize__anon_3248__5376(a0, t5);
  if (t2) {
   return t2;
  }
  goto zig_block_5;
 }
 goto zig_block_5;

 zig_block_5:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(6);
 if (t3) {
  t2 = bincode_serialize__anon_3246__5375(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_6;
 }
 goto zig_block_6;

 zig_block_6:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(7);
 if (t3) {
  t6 = a1.payload.AddMarker;
  t2 = bincode_serialize__anon_3250__5377(a0, t6);
  if (t2) {
   return t2;
  }
  goto zig_block_7;
 }
 goto zig_block_7;

 zig_block_7:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(8);
 if (t3) {
  t7 = a1.payload.SetVersion;
  t2 = bincode_serialize__anon_3252__5378(a0, t7);
  if (t2) {
   return t2;
  }
  goto zig_block_8;
 }
 goto zig_block_8;

 zig_block_8:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(9);
 if (t3) {
  t2 = bincode_serialize__anon_3246__5375(a0);
  if (t2) {
   return t2;
  }
  goto zig_block_9;
 }
 goto zig_block_9;

 zig_block_9:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(10);
 if (t3) {
  t0 = a1.payload.IntegrationModeResponse;
  t2 = bincode_serialize__anon_3254__5379(a0, t0);
  if (t2) {
   return t2;
  }
  goto zig_block_10;
 }
 goto zig_block_10;

 zig_block_10:;
 return 0;
}

static nav__1438_38 fs_File_write__1438(struct fs_File__799 const a0, nav__1438_41 const a1) {
 nav__1438_38 t1;
 int32_t t0;
 t0 = a0.handle;
 t1 = posix_write__1757(t0, a1);
 return t1;
}

static nav__5202_38 math_cast__anon_2885__5202(uintptr_t const a0) {
 uintptr_t t0;
 nav__5202_38 t1;
 t0 = a0;
 t1.is_null = false;
 t1.payload = t0;
 return t1;
}

static uint16_t posix_errno__anon_3044__5205(int const a0) {
 int *t3;
 int32_t t1;
 int t4;
 uint16_t t0;
 uint16_t t5;
 bool t2;
 t1 = a0;
 t2 = t1 == -INT32_C(1);
 if (t2) {
  t3 = zig_e___errno_location();
  t4 = (*t3);
  t5 = (uint16_t)t4;
  t0 = t5;
  goto zig_block_0;
 }
 t0 = UINT16_C(0);
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static uint16_t posix_unexpectedErrno__2021(uint16_t const a0) {
 (void)a0;
 return zig_error_Unexpected;
}

static nav__1428_38 fs_File_read__1428(struct fs_File__799 const a0, nav__1428_41 const a1) {
 nav__1428_38 t1;
 int32_t t0;
 t0 = a0.handle;
 t1 = posix_read__1749(t0, a1);
 return t1;
}

static nav__4986_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_read__4986(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const a0, nav__4986_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t1;
 nav__4986_42 const *t3;
 uintptr_t t4;
 uintptr_t t7;
 uintptr_t t9;
 nav__4986_42 *t5;
 nav__4986_42 t6;
 nav__4986_42 t13;
 nav__4986_42 t2;
 uintptr_t *t8;
 uint64_t t10;
 uint8_t *t11;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *t12;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *t0;
 nav__4986_38 t14;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t0;
 t2 = a1;
 t3 = (nav__4986_42 const *)&t2;
 t4 = a1.len;
 t5 = (nav__4986_42 *)&a0->buffer;
 t6 = (*t5);
 t7 = t6.len;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t9 = t7 - t9;
 t9 = (t4 < t9) ? t4 : t9;
 t10 = t9;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t4 = t10;
 t4 = t9 + t4;
 t6 = (*t3);
 t11 = t6.ptr;
 t11 = (uint8_t *)(((uintptr_t)t11) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t9 = t10;
 t6.ptr = t11;
 t6.len = t9;
 t12 = (*t1);
 t5 = (nav__4986_42 *)&t12->buffer;
 t8 = (uintptr_t *)&a0->pos;
 t9 = (*t8);
 t13 = (*t5);
 t11 = t13.ptr;
 t11 = (uint8_t *)(((uintptr_t)t11) + (t9*sizeof(uint8_t)));
 t9 = t4 - t9;
 t13.ptr = t11;
 t13.len = t9;
 t11 = t13.ptr;
 if (t6.len != 0) memcpy(t6.ptr, t11, t6.len * sizeof(uint8_t));
 t12 = (*t1);
 t8 = (uintptr_t *)&t12->pos;
 (*t8) = t4;
 t4 = t10;
 t14.payload = t4;
 t14.error = UINT16_C(0);
 return t14;
}

static uint16_t io_Reader_readNoEof__5036(struct io_Reader__2238 const a0, nav__5036_40 const a1) {
 struct io_Reader__2238 const *t1;
 struct io_Reader__2238 t2;
 struct io_Reader__2238 t0;
 nav__5036_43 t3;
 uintptr_t t5;
 uintptr_t t6;
 uint64_t t7;
 uint64_t t8;
 uint16_t t4;
 bool t9;
 t0 = a0;
 t1 = (struct io_Reader__2238 const *)&t0;
 t2 = (*t1);
 t3 = io_Reader_readAll__5034(t2, a1);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 < t8;
 if (t9) {
  return zig_error_EndOfStream;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static nav__5206_39 mem_Allocator_allocWithSizeAndAlignment__anon_3080__5206(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__5206_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__5206_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3296__5445((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__5206_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_3298__5446(t6, t2, a2);
 return t7;
}

static nav__5035_38 io_Reader_readAtLeast__5035(struct io_Reader__2238 const a0, nav__5035_41 const a1, uintptr_t const a2) {
 struct io_Reader__2238 const *t1;
 nav__5035_41 const *t3;
 uintptr_t t4;
 uintptr_t t12;
 uintptr_t t8;
 uint64_t t5;
 uint64_t t6;
 struct io_Reader__2238 t9;
 struct io_Reader__2238 t0;
 nav__5035_41 t10;
 nav__5035_41 t2;
 uint8_t *t11;
 nav__5035_38 t13;
 uint16_t t14;
 bool t7;
 t0 = a0;
 t1 = (struct io_Reader__2238 const *)&t0;
 t2 = a1;
 t3 = (nav__5035_41 const *)&t2;
 t4 = a1.len;
 t5 = a2;
 t6 = t4;
 t7 = t5 <= t6;
 debug_assert__180(t7);
 t8 = (uintptr_t)0ul;
 zig_loop_17:
 t4 = t8;
 t6 = t4;
 t5 = a2;
 t7 = t6 < t5;
 if (t7) {
  t9 = (*t1);
  t4 = t8;
  t10 = (*t3);
  t11 = t10.ptr;
  t11 = (uint8_t *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
  t12 = t10.len;
  t4 = t12 - t4;
  t10.ptr = t11;
  t10.len = t4;
  t13 = io_Reader_read__5033(t9, t10);
  if (t13.error) {
   t14 = t13.error;
   t13.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t13.error = t14;
   return t13;
  }
  t4 = t13.payload;
  t5 = t4;
  t7 = t5 == UINT64_C(0);
  if (t7) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t12 = t8;
  t4 = t12 + t4;
  t8 = t4;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_17;

 zig_block_0:;
 t12 = t8;
 t13.payload = t12;
 t13.error = UINT16_C(0);
 return t13;
}

static void utils_print__anon_3102__5217(void) {
 (void)printf((char const *)&__anon_3403, 61, (uint8_t const *)&__anon_2400);
 return;
}

static void mem_Allocator_free__anon_3116__5296(struct mem_Allocator__599 const a0, nav__5296_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5296_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t const *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 nav__5296_51 t10;
 struct mem_Allocator__599 t11;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__5296_51, uint8_t, uintptr_t);
 void (*t17)(void *, nav__5296_51, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_3427__5448(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t11 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t11;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__5296_51, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t11.ptr;
 t17(t18, t10, t12, t3);
 return;
}

static void logger_logWithPrefix__anon_3124__5297(nav__5297_39 const a0) {
 nav__5297_59 t3;
 nav__5297_59 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__5297_63 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(0) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3480__5608(t6, a0);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void logger_logWithPrefix__anon_3126__5298(void) {
 nav__5298_41 t3;
 nav__5298_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__5298_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3502__5610(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void logger_logWithPrefix__anon_3127__5299(void) {
 nav__5299_41 t3;
 nav__5299_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__5299_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3515__5611(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void logger_logWithPrefix__anon_3132__5301(nav__5301_40 const a0) {
 nav__5301_47 t3;
 nav__5301_47 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__5301_51 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_3528__5612(t6, a0);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static void fifo_posix_Pipe_Writer_deinit__1287(struct fifo_posix_Pipe_Writer__791 *const a0) {
 struct fifo_posix_Pipe_Writer__791 *const *t1;
 struct fifo_posix_Pipe_Writer__791 *t2;
 struct fifo_posix_Pipe_Writer__791 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t3;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t4;
 struct fs_File__799 *t5;
 struct fs_File__799 t6;
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Writer__791 *const *)&t0;
 t2 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t2->buffer;
 t4 = (*t3);
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t4);
 t2 = (*t1);
 t5 = (struct fs_File__799 *)&t2->file;
 t6 = (*t5);
 fs_File_close__1385(t6);
 return;
}

static void fifo_posix_Pipe_Reader_deinit__1294(struct fifo_posix_Pipe_Reader__793 *const a0) {
 struct fifo_posix_Pipe_Reader__793 *const *t1;
 struct fs_File__799 *t4;
 uintptr_t t9;
 uintptr_t t13;
 nav__1294_61 t10;
 nav__1294_64 t11;
 uint64_t t14;
 struct fifo_posix_Pipe_Reader__793 *t15;
 struct fifo_posix_Pipe_Reader__793 *t0;
 nav__1294_53 t16;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t17;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t18;
 struct fs_File__799 t5;
 int32_t t6;
 struct os_linux_pollfd__1921 t7;
 struct os_linux_pollfd__1921 t8[1];
 struct os_linux_pollfd__1921 t3[1];
 bool t12;
 uint8_t t2[4096];
 t0 = a0;
 t1 = (struct fifo_posix_Pipe_Reader__793 *const *)&t0;
 zig_loop_7:
 t4 = (struct fs_File__799 *)&a0->file;
 t5 = (*t4);
 t6 = t5.handle;
 t7.fd = t6;
 t7.events = INT16_C(1);
 t7.revents = INT16_C(0);
 t8[0] = t7;
 memcpy((char *)&t3, t8, sizeof(struct os_linux_pollfd__1921[1]));
 t10.ptr = (struct os_linux_pollfd__1921 *)&t3[(uintptr_t)0ul];
 t10.len = (uintptr_t)1ul;
 t11 = posix_poll__1964(t10, INT32_C(0));
 t12 = t11.error == UINT16_C(0);
 if (t12) {
  t13 = t11.payload;
  t9 = t13;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t14 = t9;
 t12 = t14 == UINT64_C(0);
 if (t12) {
  goto zig_block_0;
 }
 goto zig_block_2;

 zig_block_2:;
 t15 = (*t1);
 t4 = (struct fs_File__799 *)&t15->file;
 t5 = (*t4);
 t16.ptr = &t2[(uintptr_t)0ul];
 t16.len = (uintptr_t)4096ul;
 t11 = fs_File_read__1428(t5, t16);
 t12 = t11.error == UINT16_C(0);
 if (t12) {
  t13 = t11.payload;
  t9 = t13;
  goto zig_block_3;
 }
 goto zig_block_0;

 zig_block_3:;
 t14 = t9;
 t12 = t14 == UINT64_C(0);
 if (t12) {
  goto zig_block_0;
 }
 goto zig_block_4;

 zig_block_4:;
 goto zig_loop_7;

 zig_block_0:;
 t15 = (*t1);
 t17 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *)&t15->buffer;
 t18 = (*t17);
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t18);
 t15 = (*t1);
 t4 = (struct fs_File__799 *)&t15->file;
 t5 = (*t4);
 fs_File_close__1385(t5);
 return;
}

static struct environment_linked_libraries_root_LinkedLibraries__849 environment_linked_libraries_root_LinkedLibraries_init__3643(struct mem_Allocator__599 const a0) {
 struct mem_Allocator__599 *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t2;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t3;
 struct environment_linked_libraries_root_LinkedLibraries__849 t0;
 t1 = (struct mem_Allocator__599 *)&t0.allocator;
 (*t1) = a0;
 t2 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *)&t0.libraries;
 t3 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_init__4347(a0);
 (*t2) = t3;
 return t0;
}

static uint8_t mem_Alignment_fromByteUnits__1113(uintptr_t const a0) {
 bool t0;
 uint8_t t1;
 t0 = math_isPowerOfTwo__anon_3550__5613(a0);
 debug_assert__180(t0);
 t1 = zig_ctz_u64(a0, UINT8_C(64));
 return t1;
}

static bool fs_path_isAbsolutePosix__5144(nav__5144_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 bool t2;
 bool t3;
 uint8_t t4;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 > UINT64_C(0);
 if (t2) {
  t4 = a0.ptr[(uintptr_t)0ul];
  t2 = t4 == UINT8_C(47);
  t3 = t2;
  goto zig_block_0;
 }
 t3 = false;
 goto zig_block_0;

 zig_block_0:;
 return t3;
}

static nav__2022_39 posix_toPosixPath__2022(nav__2022_41 const a0) {
 uintptr_t t1;
 uint64_t t2;
 uint8_t *t4;
 nav__2022_47 t5;
 uint8_t const *t6;
 nav__2022_39 t8;
 bool t3;
 uint8_t t7[4096];
 uint8_t t0[4096];
 t1 = a0.len;
 t2 = t1;
 t3 = t2 >= UINT64_C(4096);
 if (t3) {
  return (nav__2022_39){zig_error_NameTooLong,{'\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa','\xaa'}};
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0.len;
 t4 = (uint8_t *)&t0;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t1;
 t6 = a0.ptr;
 if (t5.len != 0) memcpy(t5.ptr, t6, t5.len * sizeof(uint8_t));
 t1 = a0.len;
 t4 = (uint8_t *)&t0[t1];
 (*t4) = UINT8_C(0);
 memcpy(t7, (const char *)&t0, sizeof(uint8_t[4096]));
 memcpy(t8.payload, t7, sizeof(uint8_t[4096]));
 t8.error = UINT16_C(0);
 return t8;
}

static uint16_t fs_Dir_accessZ__4743(struct fs_Dir__1720 const a0, uint8_t const *const a1, struct fs_File_OpenFlags__1718 const a2) {
 uint32_t t2;
 int32_t t3;
 struct fs_Dir__1720 t0;
 uint16_t t4;
 uint8_t t1;
 t0 = a0;
 t1 = a2.mode;
 switch (t1) {
  case UINT8_C(0): {
   t2 = UINT32_C(0);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t2 = UINT32_C(2);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = UINT32_C(6);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t3 = a0.fd;
 t4 = posix_faccessatZ__1913(t3, a1, t2, UINT32_C(0));
 return t4;
}

static nav__4679_39 fs_Dir_openFileZ__4679(struct fs_Dir__1720 const a0, uint8_t const *const a1, struct fs_File_OpenFlags__1718 const a2) {
 uint32_t *t2;
 struct fs_File__799 *t14;
 int32_t *t15;
 int32_t t6;
 int32_t t11;
 int32_t t12;
 uint32_t t7;
 uint32_t t1;
 nav__4679_50 t8;
 nav__4679_39 t10;
 nav__4679_39 t13;
 struct fs_Dir__1720 t0;
 uint16_t t9;
 uint8_t t3;
 uint8_t t4;
 bool t5;
 t0 = a0;
 t2 = (uint32_t *)&t1;
 t3 = a2.mode;
 switch (t3) {
  case UINT8_C(0): {
   t4 = UINT8_C(0);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t4 = UINT8_C(1);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t4 = UINT8_C(2);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffc)), zig_shl_u32((uint32_t)t4, UINT8_C(0)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffc3)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(2)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffbf)), zig_shl_u32((uint32_t)false, UINT8_C(6)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffff7f)), zig_shl_u32((uint32_t)false, UINT8_C(7)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)false, UINT8_C(8)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffdff)), zig_shl_u32((uint32_t)false, UINT8_C(9)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffbff)), zig_shl_u32((uint32_t)false, UINT8_C(10)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffff7ff)), zig_shl_u32((uint32_t)false, UINT8_C(11)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffefff)), zig_shl_u32((uint32_t)false, UINT8_C(12)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffdfff)), zig_shl_u32((uint32_t)false, UINT8_C(13)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffbfff)), zig_shl_u32((uint32_t)false, UINT8_C(14)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff7fff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(15)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffeffff)), zig_shl_u32((uint32_t)false, UINT8_C(16)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffdffff)), zig_shl_u32((uint32_t)false, UINT8_C(17)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffbffff)), zig_shl_u32((uint32_t)false, UINT8_C(18)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff7ffff)), zig_shl_u32((uint32_t)false, UINT8_C(19)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffefffff)), zig_shl_u32((uint32_t)false, UINT8_C(20)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffdfffff)), zig_shl_u32((uint32_t)false, UINT8_C(21)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffbfffff)), zig_shl_u32((uint32_t)false, UINT8_C(22)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0x7fffff)), zig_shl_u32((uint32_t)UINT16_C(0), UINT8_C(23)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff7ffff)), zig_shl_u32((uint32_t)true, UINT8_C(19)));
 t2 = (uint32_t *)&t1;
 t5 = a2.allow_ctty;
 t5 = !t5;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)t5, UINT8_C(8)));
 t6 = a0.fd;
 t7 = t1;
 t8 = posix_openatZ__1769(t6, a1, t7, (uintptr_t)0ul);
 if (t8.error) {
  t9 = t8.error;
  t10.payload = (struct fs_File__799){-INT32_C(0x55555556)};
  t10.error = t9;
  return t10;
 }
 t6 = t8.payload;
 t4 = a2.lock;
 t5 = t4 != UINT8_C(0);
 if (t5) {
  t5 = a2.lock_nonblocking;
  if (t5) {
   t11 = INT32_C(4);
   goto zig_block_2;
  }
  t11 = INT32_C(0);
  goto zig_block_2;

  zig_block_2:;
  t4 = a2.lock;
  switch (t4) {
   case UINT8_C(0): {
    zig_unreachable();
   }
   case UINT8_C(1): {
    t11 = INT32_C(1) | t11;
    t12 = t11;
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t11 = INT32_C(2) | t11;
    t12 = t11;
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t9 = posix_flock__1930(t6, t12);
  if (t9) {
   posix_close__1727(t6);
   t10.payload = (struct fs_File__799){-INT32_C(0x55555556)};
   t10.error = t9;
   return t10;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t13.error = UINT16_C(0);
 t14 = &t13.payload;
 t15 = (int32_t *)&t14->handle;
 (*t15) = t6;
 return t13;
}

static uintptr_t array_list_ArrayListAlignedUnmanaged_28u8_2cnull_29_growCapacity__5372(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)128ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacityPrecise__1565(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t8;
 nav__1565_43 t9;
 nav__1565_43 t12;
 nav__1565_43 t20;
 nav__1565_43 t21;
 nav__1565_43 t18;
 struct mem_Allocator__599 *t10;
 struct mem_Allocator__599 t11;
 nav__1565_43 *t13;
 uint8_t **t14;
 uint8_t *t15;
 nav__1565_55 t16;
 nav__1565_43 const *t19;
 uint16_t t17;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t8 = (*t7);
 t9 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t8);
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 t12 = mem_Allocator_remap__anon_3698__5700(t11, t9, a1);
 t6 = t12.ptr != NULL;
 if (t6) {
  t9 = t12;
  t7 = (*t1);
  t13 = (nav__1565_43 *)&t7->items;
  t14 = &t13->ptr;
  t15 = t9.ptr;
  (*t14) = t15;
  t7 = (*t1);
  t2 = (uintptr_t *)&t7->capacity;
  t3 = t9.len;
  (*t2) = t3;
  goto zig_block_1;
 }
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 t16 = mem_Allocator_alignedAlloc__anon_3703__5701(t11, a1);
 if (t16.error) {
  t17 = t16.error;
  return t17;
 }
 t12 = t16.payload;
 t18 = t12;
 t19 = (nav__1565_43 const *)&t18;
 t13 = (nav__1565_43 *)&a0->items;
 t20 = (*t13);
 t3 = t20.len;
 t20 = (*t19);
 t15 = t20.ptr;
 t15 = (uint8_t *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t20.ptr = t15;
 t20.len = t3;
 t13 = (nav__1565_43 *)&a0->items;
 t21 = (*t13);
 t15 = t21.ptr;
 if (t20.len != 0) memcpy(t20.ptr, t15, t20.len * sizeof(uint8_t));
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 mem_Allocator_free__anon_3705__5702(t11, t9);
 t7 = (*t1);
 t13 = (nav__1565_43 *)&t7->items;
 t14 = &t13->ptr;
 t15 = t12.ptr;
 (*t14) = t15;
 t7 = (*t1);
 t2 = (uintptr_t *)&t7->capacity;
 t3 = t12.len;
 (*t2) = t3;
 goto zig_block_1;

 zig_block_1:;
 return 0;
}

static uint16_t bincode_serialize__anon_3242__5373(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, uint32_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_3710__5703(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3244__5374(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__struct_1798__1798 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3715__5704(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3246__5375(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0) {
 (void)a0;
 return 0;
}

static uint16_t bincode_serialize__anon_3248__5376(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__struct_1799__1799 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3716__5705(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3250__5377(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__struct_1800__1800 const a1) {
 uint16_t t0;
 t0 = bincode_serializeStruct__anon_3717__5706(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3252__5378(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, uint64_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_3718__5707(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_3254__5379(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, uint8_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeEnum__anon_3721__5708(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static nav__1757_38 posix_write__1757(int32_t const a0, nav__1757_40 const a1) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t const *t3;
 intptr_t t5;
 nav__1757_38 t7;
 uint32_t t4;
 uint16_t t6;
 bool t2;
 t0 = a1.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__1757_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_16:
 t3 = a1.ptr;
 t0 = a1.len;
 t0 = ((uintptr_t)2147479552ul < t0) ? (uintptr_t)2147479552ul : t0;
 t4 = (uint32_t)t0;
 t0 = (uintptr_t)t4;
 t5 = write(a0, t3, t0);
 t6 = posix_errno__anon_3732__5710(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(4): {
   goto zig_block_1;
  }
  case UINT16_C(22): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InvalidArgument};
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(2): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessNotFound};
  }
  case UINT16_C(11): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_WouldBlock};
  }
  case UINT16_C(9): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NotOpenForWriting};
  }
  case UINT16_C(89): {
   zig_unreachable();
  }
  case UINT16_C(122): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DiskQuota};
  }
  case UINT16_C(27): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_FileTooBig};
  }
  case UINT16_C(5): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InputOutput};
  }
  case UINT16_C(28): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
  }
  case UINT16_C(13): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_AccessDenied};
  }
  case UINT16_C(1): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_AccessDenied};
  }
  case UINT16_C(32): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_BrokenPipe};
  }
  case UINT16_C(104): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionResetByPeer};
  }
  case UINT16_C(16): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_DeviceBusy};
  }
  case UINT16_C(6): {
   return (nav__1757_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoDevice};
  }
  default: {
   t6 = posix_unexpectedErrno__2021(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_1:;
 goto zig_loop_16;
}

static nav__1749_38 posix_read__1749(int32_t const a0, nav__1749_40 const a1) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3;
 intptr_t t5;
 nav__1749_38 t7;
 uint32_t t4;
 uint16_t t6;
 bool t2;
 t0 = a1.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__1749_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_16:
 t3 = a1.ptr;
 t0 = a1.len;
 t0 = ((uintptr_t)2147479552ul < t0) ? (uintptr_t)2147479552ul : t0;
 t4 = (uint32_t)t0;
 t0 = (uintptr_t)t4;
 t5 = read(a0, t3, t0);
 t6 = posix_errno__anon_3732__5710(t5);
 switch (t6) {
  case UINT16_C(0): {
   t0 = (uintptr_t)t5;
   t7.payload = t0;
   t7.error = UINT16_C(0);
   return t7;
  }
  case UINT16_C(4): {
   goto zig_block_1;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(2): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ProcessNotFound};
  }
  case UINT16_C(11): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_WouldBlock};
  }
  case UINT16_C(125): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Canceled};
  }
  case UINT16_C(9): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NotOpenForReading};
  }
  case UINT16_C(5): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_InputOutput};
  }
  case UINT16_C(21): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_IsDir};
  }
  case UINT16_C(105): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  case UINT16_C(12): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SystemResources};
  }
  case UINT16_C(107): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_SocketNotConnected};
  }
  case UINT16_C(104): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionResetByPeer};
  }
  case UINT16_C(110): {
   return (nav__1749_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_ConnectionTimedOut};
  }
  default: {
   t6 = posix_unexpectedErrno__2021(t6);
   t7.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
   t7.error = t6;
   return t7;
  }
 }

 zig_block_1:;
 goto zig_loop_16;
}

static nav__5445_38 math_mul__anon_3296__5445(uintptr_t const a0, uintptr_t const a1) {
 nav__5445_42 t0;
 uintptr_t t3;
 nav__5445_38 t4;
 uint8_t t1;
 bool t2;
 t0.f1 = zig_mulo_u64(&t0.f0, a0, a1, UINT8_C(64));
 t1 = t0.f1;
 t2 = t1 != UINT8_C(0);
 if (t2) {
  return (nav__5445_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_Overflow};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t0.f0;
 t4.payload = t3;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__5446_39 mem_Allocator_allocBytesWithAlignment__anon_3298__5446(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t7;
 struct mem_Allocator_VTable__602 const *const *t8;
 struct mem_Allocator_VTable__602 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__5446_52 t17;
 nav__5446_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__5446_39){(uint8_t *)UINTPTR_MAX,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t7 = t5;
 t1 = (struct mem_Allocator__599 const *)&t7;
 t8 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__5446_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t18.payload = t4;
 t18.error = UINT16_C(0);
 return t18;
}

static nav__5033_38 io_Reader_read__5033(struct io_Reader__2238 const a0, nav__5033_41 const a1) {
 struct io_Reader__2238 const *t1;
 nav__5033_38 (*const *t2)(void const *, nav__5033_41);
 nav__5033_38 (*t3)(void const *, nav__5033_41);
 void const *t4;
 nav__5033_38 t5;
 struct io_Reader__2238 t0;
 t0 = a0;
 t1 = (struct io_Reader__2238 const *)&t0;
 t2 = (nav__5033_38 (*const *)(void const *, nav__5033_41))&t1->readFn;
 t3 = (*t2);
 t4 = a0.context;
 t5 = t3(t4, a1);
 return t5;
}

static nav__5448_39 mem_sliceAsBytes__anon_3427__5448(nav__5448_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t const *t4;
 uint8_t const *t5;
 uint8_t const *const *t6;
 nav__5448_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__5448_39){(uint8_t const *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t const *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t const *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint16_t fmt_format__anon_3892__5741(struct io_Writer__3878 const a0, nav__5741_40 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 struct shared_Command__1795 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__5741_45){(uint8_t const *)&__anon_4002,(uintptr_t)49ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_4378__6108(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__5741_45){(uint8_t const *)&__anon_4385,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5608_40 fmt_bufPrint__anon_3480__5608(nav__5608_39 const a0, nav__5608_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__5608_71 (**t11)(void const *, nav__5608_48);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__5608_39 t15;
 nav__5608_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5608_71 (**)(void const *, nav__5608_48))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_3892__5741(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5608_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4401__6109(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6109_43){(uint8_t const *)&__anon_4501,(uintptr_t)68ul});
 if (t3) {
  return t3;
 }
 t3 = fmt_formatType__anon_4509__6110(UINT64_C(2), (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6109_43){(uint8_t const *)&__anon_4385,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5610_40 fmt_bufPrint__anon_3502__5610(nav__5610_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__5610_54 (**t11)(void const *, nav__5610_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__5610_39 t15;
 nav__5610_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5610_54 (**)(void const *, nav__5610_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_4401__6109(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5610_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4518__6111(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6111_43){(uint8_t const *)&__anon_4624,(uintptr_t)82ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5611_40 fmt_bufPrint__anon_3515__5611(nav__5611_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__5611_54 (**t11)(void const *, nav__5611_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__5611_39 t15;
 nav__5611_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5611_54 (**)(void const *, nav__5611_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_4518__6111(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5611_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_4630__6112(struct io_Writer__3878 const a0, nav__6112_41 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 nav__6112_40 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6112_40){(uint8_t const *)&__anon_4678,(uintptr_t)32ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_4782__6113(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6112_40){(uint8_t const *)&__anon_4811,(uintptr_t)22ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5612_40 fmt_bufPrint__anon_3528__5612(nav__5612_39 const a0, nav__5612_43 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__5612_59 (**t11)(void const *, nav__5612_42);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__5612_39 t15;
 nav__5612_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__5612_59 (**)(void const *, nav__5612_42))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_4630__6112(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__5612_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(struct array_list_ArrayListAligned_28u8_2cnull_29__811 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 const *t1;
 struct mem_Allocator__599 const *t2;
 struct mem_Allocator__599 t3;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t0;
 nav__1529_41 t5;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 const *)&t0;
 t2 = (struct mem_Allocator__599 const *)&t1->allocator;
 t3 = (*t2);
 t4 = (*t1);
 t5 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t4);
 mem_Allocator_free__anon_3705__5702(t3, t5);
 return;
}

static void fs_File_close__1385(struct fs_File__799 const a0) {
 int32_t t0;
 t0 = a0.handle;
 posix_close__1727(t0);
 return;
}

static struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_init__4347(struct mem_Allocator__599 const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t0;
 t0.items = (nav__4347_43){(struct environment_linked_libraries_root_LibraryInfo__946 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 t0.capacity = (uintptr_t)0ul;
 t0.allocator = a0;
 return t0;
}

static bool math_isPowerOfTwo__anon_3550__5613(uintptr_t const a0) {
 uint64_t t0;
 uintptr_t t2;
 bool t1;
 t0 = a0;
 t1 = t0 > UINT64_C(0);
 debug_assert__180(t1);
 t2 = a0 - (uintptr_t)1ul;
 t2 = a0 & t2;
 t0 = t2;
 t1 = t0 == UINT64_C(0);
 return t1;
}

static uint16_t posix_faccessatZ__1913(int32_t const a0, uint8_t const *const a1, uint32_t const a2, uint32_t const a3) {
 unsigned int t0;
 unsigned int t1;
 int t2;
 uint16_t t3;
 t0 = a2;
 t1 = a3;
 t2 = faccessat(a0, a1, t0, t1);
 t3 = posix_errno__anon_3044__5205(t2);
 switch (t3) {
  case UINT16_C(0): {
   return 0;
  }
  case UINT16_C(13): {
   return zig_error_PermissionDenied;
  }
  case UINT16_C(1): {
   return zig_error_PermissionDenied;
  }
  case UINT16_C(30): {
   return zig_error_ReadOnlyFileSystem;
  }
  case UINT16_C(40): {
   return zig_error_SymLinkLoop;
  }
  case UINT16_C(26): {
   return zig_error_FileBusy;
  }
  case UINT16_C(20): {
   return zig_error_FileNotFound;
  }
  case UINT16_C(2): {
   return zig_error_FileNotFound;
  }
  case UINT16_C(36): {
   return zig_error_NameTooLong;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(5): {
   return zig_error_InputOutput;
  }
  case UINT16_C(12): {
   return zig_error_SystemResources;
  }
  case UINT16_C(84): {
   t3 = posix_unexpectedErrno__2021(t3);
   return t3;
  }
  default: {
   t3 = posix_unexpectedErrno__2021(t3);
   return t3;
  }
 }
}

static nav__1769_38 posix_openatZ__1769(int32_t const a0, uint8_t const *const a1, uint32_t const a2, uintptr_t const a3) {
 unsigned long t1;
 int t0;
 int32_t t3;
 nav__1769_38 t4;
 uint16_t t2;
 zig_loop_11:
 t0 = a0;
 t1 = a3;
 t0 = openat64(t0, a1, a2, t1);
 t2 = posix_errno__anon_3044__5205(t0);
 switch (t2) {
  case UINT16_C(0): {
   t3 = t0;
   t4.payload = t3;
   t4.error = UINT16_C(0);
   return t4;
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(22): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_BadPathName};
  }
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(13): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_AccessDenied};
  }
  case UINT16_C(27): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileTooBig};
  }
  case UINT16_C(75): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileTooBig};
  }
  case UINT16_C(21): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_IsDir};
  }
  case UINT16_C(40): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_SymLinkLoop};
  }
  case UINT16_C(24): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_ProcessFdQuotaExceeded};
  }
  case UINT16_C(36): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NameTooLong};
  }
  case UINT16_C(23): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_SystemFdQuotaExceeded};
  }
  case UINT16_C(19): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NoDevice};
  }
  case UINT16_C(2): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileNotFound};
  }
  case UINT16_C(12): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_SystemResources};
  }
  case UINT16_C(28): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NoSpaceLeft};
  }
  case UINT16_C(20): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NotDir};
  }
  case UINT16_C(1): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_AccessDenied};
  }
  case UINT16_C(17): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_PathAlreadyExists};
  }
  case UINT16_C(16): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_DeviceBusy};
  }
  case UINT16_C(95): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileLocksNotSupported};
  }
  case UINT16_C(11): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_WouldBlock};
  }
  case UINT16_C(26): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_FileBusy};
  }
  case UINT16_C(6): {
   return (nav__1769_38){-INT32_C(0x55555556),zig_error_NoDevice};
  }
  case UINT16_C(84): {
   t2 = posix_unexpectedErrno__2021(t2);
   t4.payload = -INT32_C(0x55555556);
   t4.error = t2;
   return t4;
  }
  default: {
   t2 = posix_unexpectedErrno__2021(t2);
   t4.payload = -INT32_C(0x55555556);
   t4.error = t2;
   return t4;
  }
 }

 zig_block_0:;
 goto zig_loop_11;
}

static uint16_t posix_flock__1930(int32_t const a0, int32_t const a1) {
 int t0;
 uint16_t t1;
 zig_loop_3:
 t0 = a1;
 t0 = flock(a0, t0);
 t1 = posix_errno__anon_3044__5205(t0);
 switch (t1) {
  case UINT16_C(0): {
   return 0;
  }
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   goto zig_block_0;
  }
  case UINT16_C(22): {
   zig_unreachable();
  }
  case UINT16_C(37): {
   return zig_error_SystemResources;
  }
  case UINT16_C(11): {
   return zig_error_WouldBlock;
  }
  case UINT16_C(95): {
   return zig_error_FileLocksNotSupported;
  }
  default: {
   t1 = posix_unexpectedErrno__2021(t1);
   return t1;
  }
 }

 zig_block_0:;
 goto zig_loop_3;
}

static void posix_close__1727(int32_t const a0) {
 int t0;
 uint16_t t1;
 t0 = close(a0);
 t1 = posix_errno__anon_3044__5205(t0);
 switch (t1) {
  case UINT16_C(9): {
   zig_unreachable();
  }
  case UINT16_C(4): {
   return;
  }
  default: {
   return;
  }
 }
}

static nav__1575_39 array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(struct array_list_ArrayListAligned_28u8_2cnull_29__811 const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 const *t1;
 nav__1575_39 const *t2;
 uint8_t *const *t3;
 uintptr_t t4;
 uint8_t *t5;
 nav__1575_39 t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t0;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 const *)&t0;
 t2 = (nav__1575_39 const *)&t1->items;
 t3 = &t2->ptr;
 t4 = a0.capacity;
 t5 = (*t3);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t6.ptr = t5;
 t6.len = t4;
 return t6;
}

static nav__5700_39 mem_Allocator_remap__anon_3698__5700(struct mem_Allocator__599 const a0, nav__5700_39 const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 nav__5700_39 const *t3;
 uint64_t t4;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t15;
 nav__5700_39 t7;
 nav__5700_39 t10;
 nav__5700_39 t2;
 uint8_t *t8;
 uint8_t *t20;
 uint8_t *t21;
 uint8_t *t22;
 void *t9;
 uintptr_t t11;
 uintptr_t t13;
 nav__5700_50 t12;
 struct mem_Allocator_VTable__602 const *const *t16;
 struct mem_Allocator_VTable__602 const *t17;
 uint8_t *(*const *t18)(void *, nav__5700_39, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*t19)(void *, nav__5700_39, uint8_t, uintptr_t, uintptr_t);
 uint8_t *const *t23;
 bool t5;
 uint8_t t14;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = (nav__5700_39 const *)&t2;
 t4 = a2;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  t6 = (*t1);
  mem_Allocator_free__anon_3705__5702(t6, a1);
  t7 = (*t3);
  t8 = t7.ptr;
  t8 = (uint8_t *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t9 = (void *)t8;
  t7.ptr = t9;
  t7.len = (uintptr_t)0ul;
  t10 = t7;
  return t10;
 }
 goto zig_block_0;

 zig_block_0:;
 t11 = a1.len;
 t4 = t11;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return (nav__5700_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 }
 goto zig_block_1;

 zig_block_1:;
 t10 = mem_sliceAsBytes__anon_4942__6117(a1);
 t12 = math_mul__anon_3296__5445((uintptr_t)1ul, a2);
 t5 = t12.error == UINT16_C(0);
 if (t5) {
  t13 = t12.payload;
  t11 = t13;
  goto zig_block_2;
 }
 return (nav__5700_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_2:;
 t6 = (*t1);
 t14 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t13 = (uintptr_t)zig_return_address();
 t15 = t6;
 t1 = (struct mem_Allocator__599 const *)&t15;
 t16 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t17 = (*t16);
 t18 = (uint8_t *(*const *)(void *, nav__5700_39, uint8_t, uintptr_t, uintptr_t))&t17->remap;
 t19 = (*t18);
 t9 = t6.ptr;
 t20 = t19(t9, t10, t14, t11, t13);
 t5 = t20 != NULL;
 if (t5) {
  t21 = t20;
  t8 = t21;
  goto zig_block_3;
 }
 return (nav__5700_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_3:;
 t22 = t8;
 t23 = (uint8_t *const *)&t22;
 t8 = (*t23);
 t8 = (uint8_t *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t8;
 t10.len = t11;
 memcpy(&t7, &t10, sizeof(nav__5700_39));
 t7 = mem_bytesAsSlice__anon_4955__6118(t7);
 t10 = t7;
 return t10;
}

static nav__5701_40 mem_Allocator_alignedAlloc__anon_3703__5701(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__5701_40 t4;
 nav__5701_40 t8;
 nav__5701_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__5701_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_3080__5206(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__5701_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void mem_Allocator_free__anon_3705__5702(struct mem_Allocator__599 const a0, nav__5702_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__5702_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t12;
 struct mem_Allocator_VTable__602 const *const *t13;
 struct mem_Allocator_VTable__602 const *t14;
 void (*const *t15)(void *, nav__5702_40, uint8_t, uintptr_t);
 void (*t16)(void *, nav__5702_40, uint8_t, uintptr_t);
 void *t17;
 bool t5;
 uint8_t t11;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_4942__6117(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t11 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t12 = t10;
 t1 = (struct mem_Allocator__599 const *)&t12;
 t13 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t14 = (*t13);
 t15 = (void (*const *)(void *, nav__5702_40, uint8_t, uintptr_t))&t14->free;
 t16 = (*t15);
 t17 = t10.ptr;
 t16(t17, t2, t11, t3);
 return;
}

static uint16_t bincode_serializeInt__anon_3710__5703(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, uint32_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t8;
 void const *t9;
 nav__5703_48 (**t10)(void const *, nav__5703_50);
 struct io_Writer__3878 t11;
 struct io_Writer__3878 t6;
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t14;
 struct io_Writer__3878 const *t13;
 nav__5703_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[4];
 uint8_t t15[4];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5703_48 (**)(void const *, nav__5703_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__3878 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__3878 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[4]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[4]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)4ul;
 t18 = io_Writer_writeAll__5732(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3715__5704(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__struct_1798__1798 const a1) {
 nav__5704_45 t2;
 int32_t t0;
 uint16_t t1;
 t0 = a1.pid;
 t1 = bincode_serialize__anon_4986__6119(a0, t0);
 if (t1) {
  return t1;
 }
 t2 = a1.uri;
 t1 = bincode_serialize__anon_4988__6120(a0, t2);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3716__5705(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__struct_1799__1799 const a1) {
 nav__5705_45 t0;
 uint16_t t1;
 t0 = a1.name;
 t1 = bincode_serialize__anon_4988__6120(a0, t0);
 if (t1) {
  return t1;
 }
 t0 = a1.version;
 t1 = bincode_serialize__anon_4988__6120(a0, t0);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeStruct__anon_3717__5706(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_Command__struct_1800__1800 const a1) {
 struct shared_MarkerType__1802 t2;
 int32_t t0;
 uint16_t t1;
 t0 = a1.pid;
 t1 = bincode_serialize__anon_4986__6119(a0, t0);
 if (t1) {
  return t1;
 }
 t2 = a1.marker;
 t1 = bincode_serialize__anon_4990__6121(a0, t2);
 if (t1) {
  return t1;
 }
 return 0;
}

static uint16_t bincode_serializeInt__anon_3718__5707(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, uint64_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t8;
 void const *t9;
 nav__5707_48 (**t10)(void const *, nav__5707_50);
 struct io_Writer__3878 t11;
 struct io_Writer__3878 t6;
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t14;
 struct io_Writer__3878 const *t13;
 nav__5707_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[8];
 uint8_t t15[8];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__5707_48 (**)(void const *, nav__5707_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__3878 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__3878 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[8]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[8]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)8ul;
 t18 = io_Writer_writeAll__5732(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializeEnum__anon_3721__5708(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, uint8_t const a1) {
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 t0 = a1;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3242__5373(a0, t1);
 if (t2) {
  return t2;
 }
 return 0;
}

static uint16_t posix_errno__anon_3732__5710(intptr_t const a0) {
 int64_t t1;
 int *t3;
 int t4;
 uint16_t t0;
 uint16_t t5;
 bool t2;
 t1 = a0;
 t2 = t1 == -INT64_C(1);
 if (t2) {
  t3 = zig_e___errno_location();
  t4 = (*t3);
  t5 = (uint16_t)t4;
  t0 = t5;
  goto zig_block_0;
 }
 t0 = UINT16_C(0);
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 **t1;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__5724_38 io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724(void const *const a0, nav__5724_41 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t0;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *t1;
 nav__5724_38 t2;
 nav__5724_38 t3;
 t0 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)a0;
 t1 = (*t0);
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4987(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__5724_38));
 return t3;
}

static uint16_t io_Writer_writeAll__5732(struct io_Writer__3878 const a0, nav__5732_40 const a1) {
 struct io_Writer__3878 const *t1;
 nav__5732_40 const *t3;
 uintptr_t t5;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t4;
 uint64_t t7;
 uint64_t t8;
 struct io_Writer__3878 t10;
 struct io_Writer__3878 t0;
 nav__5732_40 t11;
 nav__5732_40 t2;
 uint8_t const *t12;
 nav__5732_43 t14;
 uint16_t t15;
 bool t9;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = a1;
 t3 = (nav__5732_40 const *)&t2;
 t4 = (uintptr_t)0ul;
 zig_loop_11:
 t5 = t4;
 t6 = a1.len;
 t7 = t5;
 t8 = t6;
 t9 = t7 != t8;
 if (t9) {
  t6 = t4;
  t10 = (*t1);
  t5 = t4;
  t11 = (*t3);
  t12 = t11.ptr;
  t12 = (uint8_t const *)(((uintptr_t)t12) + (t5*sizeof(uint8_t)));
  t13 = t11.len;
  t5 = t13 - t5;
  t11.ptr = t12;
  t11.len = t5;
  t14 = io_Writer_write__5731(t10, t11);
  if (t14.error) {
   t15 = t14.error;
   return t15;
  }
  t5 = t14.payload;
  t5 = t6 + t5;
  t4 = t5;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_11;

 zig_block_0:;
 return 0;
}

static uint16_t shared_Command_format__anon_5054__6122(struct shared_Command__1795 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 struct io_Writer__3878 const *t1;
 struct shared_Command__struct_1798__1798 t3;
 struct io_Writer__3878 t4;
 struct io_Writer__3878 t0;
 nav__6122_44 t6;
 nav__6122_44 t10;
 nav__6122_66 t7;
 struct shared_Command__struct_1799__1799 t9;
 nav__6122_76 t11;
 struct shared_Command__struct_1800__1800 t12;
 struct shared_MarkerType__1802 t13;
 nav__6122_78 t14;
 uint64_t t15;
 nav__6122_80 t16;
 int32_t t5;
 uint16_t t8;
 uint8_t t2;
 nav__6122_84 t17;
 (void)a1;
 t0 = a2;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.ExecutedBenchmark;
   t4 = (*t1);
   t5 = t3.pid;
   t6 = t3.uri;
   t7.f0 = t5;
   t7.f1 = t6;
   t8 = io_Writer_print__anon_5074__6123(t4, t7);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5732(t4, (nav__6122_44){(uint8_t const *)&__anon_2108,(uintptr_t)14ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5732(t4, (nav__6122_44){(uint8_t const *)&__anon_2113,(uintptr_t)13ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5732(t4, (nav__6122_44){(uint8_t const *)&__anon_2118,(uintptr_t)3ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5732(t4, (nav__6122_44){(uint8_t const *)&__anon_2123,(uintptr_t)12ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(5): {
   t9 = a0.payload.SetIntegration;
   t4 = (*t1);
   t6 = t9.name;
   t10 = t9.version;
   t11.f0 = t6;
   t11.f1 = t10;
   t8 = io_Writer_print__anon_5096__6124(t4, t11);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(6): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5732(t4, (nav__6122_44){(uint8_t const *)&__anon_2131,(uintptr_t)3ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(7): {
   t12 = a0.payload.AddMarker;
   t4 = (*t1);
   t5 = t12.pid;
   t13 = t12.marker;
   t14.f0 = t5;
   t14.f1 = t13;
   t8 = io_Writer_print__anon_5108__6125(t4, t14);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(8): {
   t15 = a0.payload.SetVersion;
   t4 = (*t1);
   t16.f0 = t15;
   t8 = io_Writer_print__anon_5117__6126(t4, t16);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(9): {
   t4 = (*t1);
   t8 = io_Writer_writeAll__5732(t4, (nav__6122_44){(uint8_t const *)&__anon_2146,(uintptr_t)18ul});
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  case UINT8_C(10): {
   t2 = a0.payload.IntegrationModeResponse;
   t4 = (*t1);
   t17.f0 = t2;
   t8 = io_Writer_print__anon_5130__6127(t4, t17);
   if (t8) {
    return t8;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatType__anon_4378__6108(struct shared_Command__1795 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct shared_Command__1795 const *t1;
 struct shared_Command__1795 t3;
 struct shared_Command__1795 t0;
 struct io_Writer__3878 t2;
 uint16_t t4;
 (void)a3;
 t0 = a0;
 t1 = (struct shared_Command__1795 const *)&t0;
 t2 = a2;
 t3 = (*t1);
 t4 = shared_Command_format__anon_5054__6122(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__4992_39 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 const *t1;
 nav__4992_39 const *t2;
 uintptr_t t3;
 nav__4992_39 t4;
 uint8_t *t5;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 const *)&t0;
 t2 = (nav__4992_39 const *)&t1->buffer;
 t3 = a0.pos;
 t4 = (*t2);
 t5 = t4.ptr;
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t5;
 t4.len = t3;
 return t4;
}

static uint16_t fmt_formatType__anon_4509__6110(uint64_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 uint64_t t0;
 struct io_Writer__3878 t1;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5151__6128(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_4782__6113(nav__6113_39 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct io_Writer__3878 const *t2;
 uint64_t t3;
 struct io_Writer__3878 t5;
 struct io_Writer__3878 t1;
 nav__6113_39 t7;
 nav__6113_39 t0;
 uint16_t t6;
 uint16_t t8;
 bool t4;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__3878 const *)&t1;
 t3 = a3;
 t4 = t3 == UINT64_C(0);
 if (t4) {
  t5 = (*t2);
  t6 = io_Writer_writeAll__5732(t5, (nav__6113_39){(uint8_t const *)&__anon_5166,(uintptr_t)7ul});
  return t6;
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(&t7, &a0, sizeof(nav__6113_39));
 t6 = fmt_formatBuf__anon_5176__6129(t7, a1, a2);
 memcpy(&t8, &t6, sizeof(uint16_t));
 return t8;
}

static nav__6117_39 mem_sliceAsBytes__anon_4942__6117(nav__6117_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__6117_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6117_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static nav__6118_39 mem_bytesAsSlice__anon_4955__6118(nav__6118_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__6118_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6118_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t0 = t0 / (uintptr_t)1ul;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static nav__4912_38 io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912(void const *const a0, nav__4912_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t1;
 nav__4912_38 t2;
 nav__4912_38 t3;
 t0 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)a0;
 t1 = (*t0);
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1553(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__4912_38));
 return t3;
}

static uint16_t bincode_serialize__anon_4986__6119(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, int32_t const a1) {
 uint16_t t0;
 t0 = bincode_serializeInt__anon_5184__6130(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_4988__6120(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, nav__6120_40 const a1) {
 uint16_t t0;
 t0 = bincode_serializePointer__anon_5189__6131(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static uint16_t bincode_serialize__anon_4990__6121(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_MarkerType__1802 const a1) {
 uint16_t t0;
 t0 = bincode_serializeUnion__anon_5190__6132(a0, a1);
 if (t0) {
  return t0;
 }
 return 0;
}

static nav__4987_38 io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_write__4987(struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const a0, nav__4987_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t1;
 nav__4987_42 const *t3;
 uintptr_t t4;
 uintptr_t t10;
 uint64_t t5;
 uint64_t t11;
 uintptr_t *t7;
 nav__4987_50 *t8;
 nav__4987_50 t9;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *t12;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *t0;
 uint8_t *t13;
 nav__4987_42 t14;
 nav__4987_42 t2;
 uint8_t const *t15;
 nav__4987_38 t16;
 bool t6;
 t0 = a0;
 t1 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t0;
 t2 = a1;
 t3 = (nav__4987_42 const *)&t2;
 t4 = a1.len;
 t5 = t4;
 t6 = t5 == UINT64_C(0);
 if (t6) {
  return (nav__4987_38){(uintptr_t)0ul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (uintptr_t *)&a0->pos;
 t4 = (*t7);
 t8 = (nav__4987_50 *)&a0->buffer;
 t9 = (*t8);
 t10 = t9.len;
 t5 = t4;
 t11 = t10;
 t6 = t5 >= t11;
 if (t6) {
  return (nav__4987_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
 }
 goto zig_block_1;

 zig_block_1:;
 t8 = (nav__4987_50 *)&a0->buffer;
 t9 = (*t8);
 t10 = t9.len;
 t7 = (uintptr_t *)&a0->pos;
 t4 = (*t7);
 t4 = t10 - t4;
 t10 = a1.len;
 t10 = (t4 < t10) ? t4 : t10;
 t11 = t10;
 t12 = (*t1);
 t8 = (nav__4987_50 *)&t12->buffer;
 t7 = (uintptr_t *)&a0->pos;
 t10 = (*t7);
 t9 = (*t8);
 t13 = t9.ptr;
 t13 = (uint8_t *)(((uintptr_t)t13) + (t10*sizeof(uint8_t)));
 t10 = t11;
 t9.ptr = t13;
 t9.len = t10;
 t14 = (*t3);
 t15 = t14.ptr;
 t15 = (uint8_t const *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10 = t11;
 t14.ptr = t15;
 t14.len = t10;
 t15 = t14.ptr;
 if (t9.len != 0) memcpy(t9.ptr, t15, t9.len * sizeof(uint8_t));
 t12 = (*t1);
 t7 = (uintptr_t *)&t12->pos;
 t10 = (*t7);
 t4 = t11;
 t4 = t10 + t4;
 (*t7) = t4;
 t6 = t11 == UINT64_C(0);
 if (t6) {
  return (nav__4987_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_NoSpaceLeft};
 }
 goto zig_block_2;

 zig_block_2:;
 t4 = t11;
 t16.payload = t4;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__5731_38 io_Writer_write__5731(struct io_Writer__3878 const a0, nav__5731_41 const a1) {
 struct io_Writer__3878 const *t1;
 nav__5731_38 (*const *t2)(void const *, nav__5731_41);
 nav__5731_38 (*t3)(void const *, nav__5731_41);
 void const *t4;
 nav__5731_38 t5;
 struct io_Writer__3878 t0;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (nav__5731_38 (*const *)(void const *, nav__5731_41))&t1->writeFn;
 t3 = (*t2);
 t4 = a0.context;
 t5 = t3(t4, a1);
 return t5;
}

static uint16_t io_Writer_print__anon_5074__6123(struct io_Writer__3878 const a0, nav__6123_41 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5194__6133(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5096__6124(struct io_Writer__3878 const a0, nav__6124_41 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5198__6134(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5108__6125(struct io_Writer__3878 const a0, nav__6125_40 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5202__6135(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5117__6126(struct io_Writer__3878 const a0, nav__6126_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5206__6136(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5130__6127(struct io_Writer__3878 const a0, nav__6127_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5210__6137(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__5753_38 unicode_utf8ByteSequenceLength__5753(uint8_t const a0) {
 nav__5753_38 t0;
 switch (a0) {
  default: if ((a0 >= UINT8_C(0) && a0 <= UINT8_C(127))) {
   t0 = (nav__5753_38){0,UINT8_C(1)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(192) && a0 <= UINT8_C(223))) {
   t0 = (nav__5753_38){0,UINT8_C(2)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(224) && a0 <= UINT8_C(239))) {
   t0 = (nav__5753_38){0,UINT8_C(3)};
   goto zig_block_0;
  }if ((a0 >= UINT8_C(240) && a0 <= UINT8_C(247))) {
   t0 = (nav__5753_38){0,UINT8_C(4)};
   goto zig_block_0;
  }{
   t0 = (nav__5753_38){zig_error_Utf8InvalidStartByte,UINT8_C(0x2)};
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return t0;
}

static nav__5769_38 unicode_utf8CountCodepoints__5769(nav__5769_40 const a0) {
 nav__5769_40 const *t1;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t2;
 uintptr_t t3;
 uint64_t t6;
 uint64_t t7;
 nav__5769_40 t9;
 nav__5769_40 t0;
 uint8_t const *t10;
 uint8_t const (*t11)[8];
 nav__5769_38 t16;
 nav__5769_50 t17;
 nav__5769_48 t14;
 uint16_t t15;
 bool t8;
 uint8_t t12[8];
 uint8_t t13;
 t0 = a0;
 t1 = (nav__5769_40 const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = (uintptr_t)0ul;
 zig_loop_9:
 t4 = t3;
 t5 = a0.len;
 t6 = t4;
 t7 = t5;
 t8 = t6 < t7;
 if (t8) {
  zig_loop_18:
  t5 = t3;
  t5 = t5 + (uintptr_t)8ul;
  t4 = a0.len;
  t7 = t5;
  t6 = t4;
  t8 = t7 <= t6;
  if (t8) {
   t4 = t3;
   t9 = (*t1);
   t10 = t9.ptr;
   t10 = (uint8_t const *)(((uintptr_t)t10) + (t4*sizeof(uint8_t)));
   t11 = (uint8_t const (*)[8])t10;
   memcpy(t12, (const char *)t11, sizeof(uint8_t[8]));
   memcpy(&t4, &t12, sizeof(uintptr_t));
   t4 = zig_wrap_u64(t4, UINT8_C(64));
   t4 = t4 & (uintptr_t)9259542123273814144ul;
   t6 = t4;
   t8 = t6 != UINT64_C(0);
   if (t8) {
    goto zig_block_2;
   }
   goto zig_block_4;

   zig_block_4:;
   t4 = t2;
   t4 = t4 + (uintptr_t)8ul;
   t2 = t4;
   t4 = t3;
   t4 = t4 + (uintptr_t)8ul;
   t3 = t4;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_18;

  zig_block_2:;
  t5 = t3;
  t4 = a0.len;
  t7 = t5;
  t6 = t4;
  t8 = t7 < t6;
  if (t8) {
   t4 = t3;
   t13 = a0.ptr[t4];
   t14 = unicode_utf8ByteSequenceLength__5753(t13);
   if (t14.error) {
    t15 = t14.error;
    t16.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
    t16.error = t15;
    return t16;
   }
   t13 = t14.payload;
   t4 = t3;
   t5 = (uintptr_t)t13;
   t5 = t4 + t5;
   t4 = a0.len;
   t6 = t5;
   t7 = t4;
   t8 = t6 > t7;
   if (t8) {
    return (nav__5769_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_TruncatedInput};
   }
   goto zig_block_6;

   zig_block_6:;
   switch (t13) {
    case UINT8_C(1): {
     goto zig_block_7;
    }
    default: {
     t4 = t3;
     t9 = (*t1);
     t10 = t9.ptr;
     t10 = (uint8_t const *)(((uintptr_t)t10) + (t4*sizeof(uint8_t)));
     t4 = (uintptr_t)t13;
     t9.ptr = t10;
     t9.len = t4;
     t17 = unicode_utf8Decode__5759(t9);
     if (t17.error) {
      t15 = t17.error;
      t16.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
      t16.error = t15;
      return t16;
     }
     goto zig_block_7;
    }
   }

   zig_block_7:;
   t4 = t3;
   t5 = (uintptr_t)t13;
   t5 = t4 + t5;
   t3 = t5;
   t5 = t2;
   t5 = t5 + (uintptr_t)1ul;
   t2 = t5;
   goto zig_block_5;
  }
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_9;

 zig_block_0:;
 t5 = t2;
 t16.payload = t5;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_formatBuf__anon_5176__6129(nav__6129_39 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 struct io_Writer__3878 const *t1;
 nav__6129_44 t2;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 nav__6129_48 t6;
 uint64_t t8;
 uint64_t t9;
 struct io_Writer__3878 t10;
 struct io_Writer__3878 t0;
 nav__6129_57 t15;
 nav__6129_39 t17;
 nav__6129_39 t20;
 uint8_t *t19;
 uint32_t t14;
 uint16_t t11;
 uint16_t t12;
 nav__6129_60 t16;
 bool t3;
 uint8_t t18;
 uint8_t t13[4];
 t0 = a2;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = a1.width;
 t3 = t2.is_null != true;
 if (t3) {
  t4 = t2.payload;
  t6 = unicode_utf8CountCodepoints__5769(a0);
  t3 = t6.error == UINT16_C(0);
  if (t3) {
   t7 = t6.payload;
   t5 = t7;
   goto zig_block_1;
  }
  t7 = a0.len;
  t5 = t7;
  goto zig_block_1;

  zig_block_1:;
  t8 = t5;
  t9 = t4;
  t3 = t8 < t9;
  if (t3) {
   t5 = t4 - t5;
   t7 = t5;
   goto zig_block_2;
  }
  t7 = (uintptr_t)0ul;
  goto zig_block_2;

  zig_block_2:;
  t9 = t7;
  t3 = t9 == UINT64_C(0);
  if (t3) {
   t10 = (*t1);
   t11 = io_Writer_writeAll__5732(t10, a0);
   memcpy(&t12, &t11, sizeof(uint16_t));
   return t12;
  }
  goto zig_block_3;

  zig_block_3:;
  t14 = a1.fill;
  t15.ptr = &t13[(uintptr_t)0ul];
  t15.len = (uintptr_t)4ul;
  t16 = unicode_utf8Encode__5754(t14, t15);
  t3 = t16.error == UINT16_C(0);
  if (t3) {
   t18 = t16.payload;
   t19 = (uint8_t *)&t13;
   t19 = (uint8_t *)(((uintptr_t)t19) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t5 = (uintptr_t)t18;
   t15.ptr = t19;
   t15.len = t5;
   memcpy(&t20, &t15, sizeof(nav__6129_39));
   t17 = t20;
   goto zig_block_4;

  }
  t12 = t16.error;
  switch (t12) {
   case zig_error_Utf8CannotEncodeSurrogateHalf:
   case zig_error_CodepointTooLarge: {
    t17 = (nav__6129_39){(uint8_t const *)&__anon_5345,(uintptr_t)3ul};
    goto zig_block_4;
   }
   default: zig_unreachable();
  }

  zig_block_4:;
  t18 = a1.alignment;
  switch (t18) {
   case UINT8_C(0): {
    t10 = (*t1);
    t12 = io_Writer_writeAll__5732(t10, a0);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5736(t10, t17, t7);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   case UINT8_C(1): {
    t5 = t7 / (uintptr_t)2ul;
    t7 = t7 + (uintptr_t)1ul;
    t7 = t7 / (uintptr_t)2ul;
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5736(t10, t17, t5);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeAll__5732(t10, a0);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5736(t10, t17, t7);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   case UINT8_C(2): {
    t10 = (*t1);
    t12 = io_Writer_writeBytesNTimes__5736(t10, t17, t7);
    if (t12) {
     return t12;
    }
    t10 = (*t1);
    t12 = io_Writer_writeAll__5732(t10, a0);
    if (t12) {
     return t12;
    }
    goto zig_block_6;
   }
   default: zig_unreachable();
  }

  zig_block_6:;
  goto zig_block_0;
 }
 t10 = (*t1);
 t12 = io_Writer_writeAll__5732(t10, a0);
 if (t12) {
  return t12;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatInt__anon_5237__6139(uint64_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4300 const a3, struct io_Writer__3878 const a4) {
 uint64_t t4;
 uint64_t t11;
 uint64_t t2;
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__6139_58 t12;
 nav__6139_47 t13;
 uint16_t t14;
 uint16_t t15;
 bool t0;
 uint8_t t8;
 nav__6139_56 t9;
 uint8_t t10[2];
 uint8_t t1[65];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)65ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_19:
  t4 = t2;
  t0 = t4 >= UINT64_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT64_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__5509(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT64_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_19;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT64_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__5509(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_74:
 t4 = t2;
 t11 = (uint64_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__5525(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint64_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT64_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_74;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)65ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__6139_47));
 t14 = fmt_formatBuf__anon_5176__6129(t13, a3, a4);
 memcpy(&t15, &t14, sizeof(uint16_t));
 return t15;
}

static uint16_t fmt_formatIntValue__anon_5220__6138(uint64_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5237__6139(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5151__6128(uint64_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5220__6138(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__1553_38 array_list_ArrayListAligned_28u8_2cnull_29_appendWrite__1553(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, nav__1553_42 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 nav__1553_38 t4;
 uintptr_t t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1547(t2, a1);
 if (t3) {
  t4.payload = (uintptr_t)0xaaaaaaaaaaaaaaaaul;
  t4.error = t3;
  return t4;
 }
 t5 = a1.len;
 t4.payload = t5;
 t4.error = UINT16_C(0);
 return t4;
}

static uint16_t bincode_serializeInt__anon_5184__6130(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, int32_t const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t8;
 void const *t9;
 nav__6130_48 (**t10)(void const *, nav__6130_50);
 struct io_Writer__3878 t11;
 struct io_Writer__3878 t6;
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t14;
 struct io_Writer__3878 const *t13;
 nav__6130_50 t17;
 uint16_t t18;
 uint16_t t19;
 uint8_t t16[4];
 uint8_t t15[4];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t3;
 t4 = t1;
 t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t4;
 t7 = (void const **)&t6.context;
 t1 = (*t5);
 t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t9 = (void const *)t8;
 (*t7) = t9;
 t10 = (nav__6130_48 (**)(void const *, nav__6130_50))&t6.writeFn;
 (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t11 = t6;
 t12 = t11;
 t13 = (struct io_Writer__3878 const *)&t12;
 t11 = (*t13);
 t14 = t11;
 t13 = (struct io_Writer__3878 const *)&t14;
 memcpy(&t16, &a1, sizeof(uint8_t[4]));
 memcpy((char *)&t15, t16, sizeof(uint8_t[4]));
 t11 = (*t13);
 t17.ptr = &t15[(uintptr_t)0ul];
 t17.len = (uintptr_t)4ul;
 t18 = io_Writer_writeAll__5732(t11, t17);
 memcpy(&t19, &t18, sizeof(uint16_t));
 if (t19) {
  return t19;
 }
 return 0;
}

static uint16_t bincode_serializePointer__anon_5189__6131(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, nav__6131_40 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t23;
 uintptr_t t3;
 uint64_t t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t8;
 void const **t10;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t11;
 void const *t12;
 nav__6131_51 (**t13)(void const *, nav__6131_40);
 struct io_Writer__3878 t14;
 struct io_Writer__3878 t9;
 struct io_Writer__3878 t15;
 struct io_Writer__3878 t17;
 struct io_Writer__3878 t25;
 struct io_Writer__3878 t26;
 struct io_Writer__3878 const *t16;
 nav__6131_40 t20;
 uint16_t t21;
 uint16_t t22;
 uint8_t t19[8];
 uint8_t t18[8];
 t0 = a0;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = t3;
 t5 = t2;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t5;
 t7 = t6;
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t7;
 t10 = (void const **)&t9.context;
 t6 = (*t8);
 t11 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t6->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__6131_51 (**)(void const *, nav__6131_40))&t9.writeFn;
 (*t13) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t14 = t9;
 t15 = t14;
 t16 = (struct io_Writer__3878 const *)&t15;
 t14 = (*t16);
 t17 = t14;
 t16 = (struct io_Writer__3878 const *)&t17;
 memcpy(&t19, &t4, sizeof(uint8_t[8]));
 memcpy((char *)&t18, t19, sizeof(uint8_t[8]));
 t14 = (*t16);
 t20.ptr = &t18[(uintptr_t)0ul];
 t20.len = (uintptr_t)8ul;
 t21 = io_Writer_writeAll__5732(t14, t20);
 memcpy(&t22, &t21, sizeof(uint16_t));
 if (t22) {
  return t22;
 }
 t2 = (*t1);
 t23 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t23;
 t24 = t1;
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t24;
 t10 = (void const **)&t25.context;
 t1 = (*t8);
 t11 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t12 = (void const *)t11;
 (*t10) = t12;
 t13 = (nav__6131_51 (**)(void const *, nav__6131_40))&t25.writeFn;
 (*t13) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t14 = t25;
 t26 = t14;
 t16 = (struct io_Writer__3878 const *)&t26;
 t14 = (*t16);
 t22 = io_Writer_writeAll__5732(t14, a1);
 memcpy(&t21, &t22, sizeof(uint16_t));
 if (t21) {
  return t21;
 }
 return 0;
}

static uint16_t bincode_serializeUnion__anon_5190__6132(struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a0, struct shared_MarkerType__1802 const a1) {
 uint64_t t4;
 uint32_t t1;
 uint16_t t2;
 uint8_t t0;
 bool t3;
 t0 = a1.tag;
 t1 = (uint32_t)t0;
 t2 = bincode_serialize__anon_3242__5373(a0, t1);
 if (t2) {
  return t2;
 }
 t0 = a1.tag;
 t3 = t0 == UINT8_C(0);
 if (t3) {
  t4 = a1.payload.SampleStart;
  t2 = bincode_serialize__anon_3252__5378(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(1);
 if (t3) {
  t4 = a1.payload.SampleEnd;
  t2 = bincode_serialize__anon_3252__5378(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(2);
 if (t3) {
  t4 = a1.payload.BenchmarkStart;
  t2 = bincode_serialize__anon_3252__5378(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a1.tag;
 t3 = t0 == UINT8_C(3);
 if (t3) {
  t4 = a1.payload.BenchmarkEnd;
  t2 = bincode_serialize__anon_3252__5378(a0, t4);
  if (t2) {
   return t2;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 return 0;
}

static uint16_t fmt_format__anon_5194__6133(struct io_Writer__3878 const a0, nav__6133_41 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 nav__6133_40 t5;
 int32_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6133_40){(uint8_t const *)&__anon_5419,(uintptr_t)25ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5471__6140(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6133_40){(uint8_t const *)&__anon_5485,(uintptr_t)7ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5502__6141(t5, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6133_40){(uint8_t const *)&__anon_5511,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5198__6134(struct io_Writer__3878 const a0, nav__6134_41 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 nav__6134_40 t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6134_40){(uint8_t const *)&__anon_5564,(uintptr_t)23ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5502__6141(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6134_40){(uint8_t const *)&__anon_5589,(uintptr_t)11ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f1;
 t3 = fmt_formatType__anon_5502__6141(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6134_40){(uint8_t const *)&__anon_5511,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5202__6135(struct io_Writer__3878 const a0, nav__6135_40 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 struct shared_MarkerType__1802 t5;
 int32_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6135_45){(uint8_t const *)&__anon_5645,(uintptr_t)17ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5471__6140(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6135_45){(uint8_t const *)&__anon_5667,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5677__6142(t5, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6135_45){(uint8_t const *)&__anon_5511,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5206__6136(struct io_Writer__3878 const a0, nav__6136_39 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6136_44){(uint8_t const *)&__anon_5737,(uintptr_t)31ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5746__6143(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6136_44){(uint8_t const *)&__anon_5511,(uintptr_t)2ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5210__6137(struct io_Writer__3878 const a0, nav__6137_39 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 uint8_t t4;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6137_44){(uint8_t const *)&__anon_5792,(uintptr_t)24ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5800__6144(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__5509_39 fmt_digits2__5509(uint8_t const a0) {
 uint8_t *t1;
 uint8_t t2;
 nav__5509_39 t3;
 uint8_t t0[2];
 t1 = (uint8_t *)&t0[(uintptr_t)0ul];
 t2 = a0 / UINT8_C(10);
 t2 = UINT8_C(48) + t2;
 (*t1) = t2;
 t1 = (uint8_t *)&t0[(uintptr_t)1ul];
 t2 = a0 % UINT8_C(10);
 t2 = UINT8_C(48) + t2;
 (*t1) = t2;
 memcpy(t3.array, t0, sizeof(uint8_t[2]));
 return t3;
}

static uint8_t fmt_digitToChar__5525(uint8_t const a0, uint8_t const a1) {
 uint8_t t0;
 uint8_t t1;
 bool t2;
 switch (a0) {
  default: if ((a0 >= UINT8_C(0) && a0 <= UINT8_C(9))) {
   t1 = a0 + UINT8_C(48);
   t0 = t1;
   goto zig_block_0;
  }if ((a0 >= UINT8_C(10) && a0 <= UINT8_C(35))) {
   t2 = a1 == UINT8_C(1);
   if (t2) {
    t1 = UINT8_C(65);
    goto zig_block_1;
   }
   t1 = UINT8_C(97);
   goto zig_block_1;

   zig_block_1:;
   t1 = t1 - UINT8_C(10);
   t1 = a0 + t1;
   t0 = t1;
   goto zig_block_0;
  }{
   zig_unreachable();
  }
 }

 zig_block_0:;
 return t0;
}

static nav__5759_38 unicode_utf8Decode__5759(nav__5759_40 const a0) {
 nav__5759_40 const *t1;
 uintptr_t t2;
 nav__5759_40 t7;
 nav__5759_40 t0;
 uint8_t const *t8;
 uint8_t const (*t9)[2];
 uint8_t const (*t13)[3];
 uint8_t const (*t16)[4];
 nav__5759_38 t3;
 nav__5759_38 t6;
 nav__5759_38 t12;
 uint32_t t5;
 uint8_t t4;
 uint8_t t10[2];
 nav__5759_48 t11;
 uint8_t t14[3];
 nav__5759_52 t15;
 uint8_t t17[4];
 nav__5759_56 t18;
 t0 = a0;
 t1 = (nav__5759_40 const *)&t0;
 t2 = a0.len;
 switch (t2) {
  case (uintptr_t)1ul: {
   t4 = a0.ptr[(uintptr_t)0ul];
   t5 = (uint32_t)t4;
   t6.payload = t5;
   t6.error = UINT16_C(0);
   t3 = t6;
   goto zig_block_0;
  }
  case (uintptr_t)2ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t9 = (uint8_t const (*)[2])t8;
   memcpy(t10, (const char *)t9, sizeof(uint8_t[2]));
   memcpy(t11.array, t10, sizeof(nav__5759_48));
   t6 = unicode_utf8Decode2__5761(t11);
   memcpy(&t12, &t6, sizeof(nav__5759_38));
   t3 = t12;
   goto zig_block_0;
  }
  case (uintptr_t)3ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t13 = (uint8_t const (*)[3])t8;
   memcpy(t14, (const char *)t13, sizeof(uint8_t[3]));
   memcpy(t15.array, t14, sizeof(nav__5759_52));
   t12 = unicode_utf8Decode3__5763(t15);
   memcpy(&t6, &t12, sizeof(nav__5759_38));
   t3 = t6;
   goto zig_block_0;
  }
  case (uintptr_t)4ul: {
   t7 = (*t1);
   t8 = t7.ptr;
   t8 = (uint8_t const *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t16 = (uint8_t const (*)[4])t8;
   memcpy(t17, (const char *)t16, sizeof(uint8_t[4]));
   memcpy(t18.array, t17, sizeof(nav__5759_56));
   t12 = unicode_utf8Decode4__5767(t18);
   memcpy(&t6, &t12, sizeof(nav__5759_38));
   t3 = t6;
   goto zig_block_0;
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 return t3;
}

static nav__5752_38 unicode_utf8CodepointSequenceLength__5752(uint32_t const a0) {
 bool t0;
 t0 = a0 < UINT32_C(128);
 if (t0) {
  return (nav__5752_38){0,UINT8_C(1)};
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 < UINT32_C(2048);
 if (t0) {
  return (nav__5752_38){0,UINT8_C(2)};
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a0 < UINT32_C(65536);
 if (t0) {
  return (nav__5752_38){0,UINT8_C(3)};
 }
 goto zig_block_2;

 zig_block_2:;
 t0 = a0 < UINT32_C(1114112);
 if (t0) {
  return (nav__5752_38){0,UINT8_C(4)};
 }
 goto zig_block_3;

 zig_block_3:;
 return (nav__5752_38){zig_error_CodepointTooLarge,UINT8_C(0x2)};
}

static nav__6145_38 unicode_utf8EncodeImpl__anon_5823__6145(uint32_t const a0, nav__6145_40 const a1) {
 nav__6145_40 const *t1;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 nav__6145_40 t9;
 nav__6145_40 t0;
 uint8_t *t10;
 uint32_t t12;
 nav__6145_38 t2;
 uint16_t t3;
 uint8_t t4;
 uint8_t t11;
 bool t8;
 t0 = a1;
 t1 = (nav__6145_40 const *)&t0;
 t2 = unicode_utf8CodepointSequenceLength__5752(a0);
 if (t2.error) {
  t3 = t2.error;
  t2.payload = UINT8_C(0x2);
  t2.error = t3;
  return t2;
 }
 t4 = t2.payload;
 t5 = a1.len;
 t6 = t5;
 t7 = (uint64_t)t4;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 switch (t4) {
  case UINT8_C(1): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t11 = (uint8_t)a0;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = UINT32_C(192) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t8 = unicode_isSurrogateCodepoint__5832(a0);
   if (t8) {
    return (nav__6145_38){zig_error_Utf8CannotEncodeSurrogateHalf,UINT8_C(0x2)};
   }
   goto zig_block_1;

   zig_block_1:;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(12));
   t12 = UINT32_C(224) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)2ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)0ul];
   t12 = zig_shr_u32(a0, UINT8_C(18));
   t12 = UINT32_C(240) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)1ul];
   t12 = zig_shr_u32(a0, UINT8_C(12));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)2ul];
   t12 = zig_shr_u32(a0, UINT8_C(6));
   t12 = t12 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   t9 = (*t1);
   t10 = &t9.ptr[(uintptr_t)3ul];
   t12 = a0 & UINT32_C(63);
   t12 = UINT32_C(128) | t12;
   t11 = (uint8_t)t12;
   (*t10) = t11;
   goto zig_block_0;
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t2.payload = t4;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__5754_38 unicode_utf8Encode__5754(uint32_t const a0, nav__5754_40 const a1) {
 nav__5754_38 t0;
 nav__5754_38 t1;
 t0 = unicode_utf8EncodeImpl__anon_5823__6145(a0, a1);
 memcpy(&t1, &t0, sizeof(nav__5754_38));
 return t1;
}

static uint16_t io_Writer_writeBytesNTimes__5736(struct io_Writer__3878 const a0, nav__5736_40 const a1, uintptr_t const a2) {
 struct io_Writer__3878 const *t1;
 uintptr_t t3;
 uintptr_t t2;
 uint64_t t4;
 uint64_t t5;
 struct io_Writer__3878 t7;
 struct io_Writer__3878 t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (uintptr_t)0ul;
 zig_loop_9:
 t3 = t2;
 t4 = t3;
 t5 = a2;
 t6 = t4 < t5;
 if (t6) {
  t7 = (*t1);
  t8 = io_Writer_writeAll__5732(t7, a1);
  if (t8) {
   return t8;
  }
  t3 = t2;
  t3 = t3 + (uintptr_t)1ul;
  t2 = t3;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_9;

 zig_block_0:;
 return 0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_appendSlice__1547(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, nav__1547_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 uintptr_t t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1566(t2, t3);
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1548(t2, a1);
 return 0;
}

static uint16_t fmt_formatType__anon_5471__6140(int32_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct io_Writer__3878 t1;
 int32_t t0;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5848__6146(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_5502__6141(nav__6141_39 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct io_Writer__3878 const *t2;
 uint64_t t3;
 struct io_Writer__3878 t5;
 struct io_Writer__3878 t1;
 nav__6141_39 t0;
 uint16_t t6;
 uint16_t t7;
 bool t4;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__3878 const *)&t1;
 t3 = a3;
 t4 = t3 == UINT64_C(0);
 if (t4) {
  t5 = (*t2);
  t6 = io_Writer_writeAll__5732(t5, (nav__6141_39){(uint8_t const *)&__anon_5166,(uintptr_t)7ul});
  return t6;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = fmt_formatBuf__anon_5176__6129(a0, a1, a2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t shared_MarkerType_format__anon_5867__6147(struct shared_MarkerType__1802 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 struct io_Writer__3878 const *t1;
 uint64_t t3;
 struct io_Writer__3878 t4;
 struct io_Writer__3878 t0;
 nav__6147_56 t5;
 uint16_t t6;
 uint8_t t2;
 (void)a1;
 t0 = a2;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.SampleStart;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5881__6148(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t3 = a0.payload.SampleEnd;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5886__6149(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t3 = a0.payload.BenchmarkStart;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5893__6150(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t3 = a0.payload.BenchmarkEnd;
   t4 = (*t1);
   t5.f0 = t3;
   t6 = io_Writer_print__anon_5899__6151(t4, t5);
   if (t6) {
    return t6;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatType__anon_5677__6142(struct shared_MarkerType__1802 const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct shared_MarkerType__1802 const *t1;
 struct shared_MarkerType__1802 t3;
 struct shared_MarkerType__1802 t0;
 struct io_Writer__3878 t2;
 uint16_t t4;
 (void)a3;
 t0 = a0;
 t1 = (struct shared_MarkerType__1802 const *)&t0;
 t2 = a2;
 t3 = (*t1);
 t4 = shared_MarkerType_format__anon_5867__6147(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t fmt_formatType__anon_5746__6143(uint64_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 uint64_t t0;
 struct io_Writer__3878 t1;
 uint16_t t2;
 uint16_t t3;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_5901__6152(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatType__anon_5800__6144(uint8_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct io_Writer__3878 const *t2;
 struct io_Writer__3878 t3;
 struct io_Writer__3878 t1;
 nav__6144_47 t5;
 nav__6144_47 t6;
 uint16_t t4;
 uint8_t t0;
 (void)a1;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = (struct io_Writer__3878 const *)&t1;
 t3 = (*t2);
 t4 = io_Writer_writeAll__5732(t3, (nav__6144_47){(uint8_t const *)&__anon_5908,(uintptr_t)22ul});
 if (t4) {
  return t4;
 }
 t3 = (*t2);
 t4 = io_Writer_writeAll__5732(t3, (nav__6144_47){(uint8_t const *)&__anon_5913,(uintptr_t)1ul});
 if (t4) {
  return t4;
 }
 t3 = (*t2);
 t5 = zig_tag_name_shared_IntegrationMode__573(a0);
 memcpy(&t6, &t5, sizeof(nav__6144_47));
 t4 = io_Writer_writeAll__5732(t3, t6);
 if (t4) {
  return t4;
 }
 return 0;
}

static nav__5761_38 unicode_utf8Decode2__5761(nav__5761_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__5761_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(224);
 t1 = t0 == UINT8_C(192);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(31);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5761_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t4 = t2;
 t1 = t4 < UINT32_C(128);
 if (t1) {
  return (nav__5761_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t5.payload = t4;
 t5.error = UINT16_C(0);
 return t5;
}

static nav__5763_38 unicode_utf8Decode3__5763(nav__5763_40 const a0) {
 nav__5763_38 t1;
 uint32_t t3;
 uint16_t t2;
 nav__5763_40 t0;
 bool t4;
 bool t5;
 memcpy(t0.array, a0.array, sizeof(nav__5763_40));
 t1 = unicode_utf8Decode3AllowSurrogateHalf__5765(t0);
 if (t1.error) {
  t2 = t1.error;
  t1.payload = UINT32_C(0xaaaaa);
  t1.error = t2;
  return t1;
 }
 t3 = t1.payload;
 t4 = UINT32_C(55296) <= t3;
 if (t4) {
  t4 = t3 <= UINT32_C(57343);
  t5 = t4;
  goto zig_block_1;
 }
 t5 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t5) {
  return (nav__5763_38){UINT32_C(0xaaaaa),zig_error_Utf8EncodesSurrogateHalf};
 }
 goto zig_block_0;

 zig_block_0:;
 t1.payload = t3;
 t1.error = UINT16_C(0);
 return t1;
}

static nav__5767_38 unicode_utf8Decode4__5767(nav__5767_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__5767_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(248);
 t1 = t0 == UINT8_C(240);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(7);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5767_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5767_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t4 = zig_shlw_u32(t4, UINT8_C(6), UINT8_C(21));
 t2 = t4;
 t4 = t2;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(63);
 t3 = (uint32_t)t0;
 t3 = t4 | t3;
 t2 = t3;
 t0 = a0.array[(uintptr_t)3ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5767_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)3ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t4 = t2;
 t1 = t4 < UINT32_C(65536);
 if (t1) {
  return (nav__5767_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_3;

 zig_block_3:;
 t4 = t2;
 t1 = t4 > UINT32_C(1114111);
 if (t1) {
  return (nav__5767_38){UINT32_C(0xaaaaa),zig_error_Utf8CodepointTooLarge};
 }
 goto zig_block_4;

 zig_block_4:;
 t4 = t2;
 t5.payload = t4;
 t5.error = UINT16_C(0);
 return t5;
}

static bool unicode_isSurrogateCodepoint__5832(uint32_t const a0) {
 bool t0;
 switch (a0) {
  default: if ((a0 >= UINT32_C(55296) && a0 <= UINT32_C(57343))) {
   t0 = true;
   goto zig_block_0;
  }{
   t0 = false;
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return t0;
}

static uint16_t array_list_ArrayListAligned_28u8_2cnull_29_ensureUnusedCapacity__1566(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 nav__1566_43 *t3;
 nav__1566_43 t4;
 uintptr_t t5;
 nav__1566_52 t6;
 uint16_t t7;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__1566_43 *)&a0->items;
 t4 = (*t3);
 t5 = t4.len;
 t6 = array_list_addOrOom__1491(t5, a1);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t5 = t6.payload;
 t7 = array_list_ArrayListAligned_28u8_2cnull_29_ensureTotalCapacity__1564(t2, t5);
 return t7;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_appendSliceAssumeCapacity__1548(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0, nav__1548_41 const a1) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 nav__1548_47 *t2;
 nav__1548_47 t3;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 uintptr_t *t6;
 uint64_t t8;
 uint64_t t9;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 uint8_t *t12;
 uint8_t const *t13;
 bool t10;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (nav__1548_47 *)&a0->items;
 t3 = (*t2);
 t4 = t3.len;
 t5 = a1.len;
 t5 = t4 + t5;
 t6 = (uintptr_t *)&a0->capacity;
 t7 = (*t6);
 t8 = t5;
 t9 = t7;
 t10 = t8 <= t9;
 debug_assert__180(t10);
 t11 = (*t1);
 t2 = (nav__1548_47 *)&t11->items;
 t6 = &t2->len;
 (*t6) = t5;
 t11 = (*t1);
 t2 = (nav__1548_47 *)&t11->items;
 t5 = a1.len;
 t3 = (*t2);
 t12 = t3.ptr;
 t12 = (uint8_t *)(((uintptr_t)t12) + (t4*sizeof(uint8_t)));
 t3.ptr = t12;
 t3.len = t5;
 t13 = a1.ptr;
 if (t3.len != 0) memcpy(t3.ptr, t13, t3.len * sizeof(uint8_t));
 return;
}

static uint16_t fmt_formatInt__anon_5955__6154(int32_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4300 const a3, struct io_Writer__3878 const a4) {
 uintptr_t t5;
 uintptr_t t4;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__6154_41 t12;
 uint64_t t14;
 nav__6154_58 t15;
 nav__6154_47 t16;
 uint32_t t1;
 uint32_t t11;
 uint32_t t3;
 uint16_t t17;
 uint16_t t18;
 bool t0;
 bool t13;
 uint8_t t8;
 nav__6154_56 t9;
 uint8_t t10[2];
 uint8_t t2[33];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t1 = zig_abs_i32(a0);
 t3 = t1;
 t4 = (uintptr_t)33ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t1 = t3;
  t0 = t1 >= UINT32_C(100);
  if (t0) {
   t5 = t4;
   t5 = t5 - (uintptr_t)2ul;
   t4 = t5;
   t5 = t4;
   t6 = (uint8_t *)&t2;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t1 = t3;
   t1 = t1 % UINT32_C(100);
   t8 = (uint8_t)t1;
   t9 = fmt_digits2__5509(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t1 = t3;
   t1 = t1 / UINT32_C(100);
   t3 = t1;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t1 = t3;
  t0 = t1 < UINT32_C(10);
  if (t0) {
   t5 = t4;
   t5 = t5 - (uintptr_t)1ul;
   t4 = t5;
   t5 = t4;
   t6 = (uint8_t *)&t2[t5];
   t1 = t3;
   t8 = (uint8_t)t1;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t4;
  t5 = t5 - (uintptr_t)2ul;
  t4 = t5;
  t5 = t4;
  t6 = (uint8_t *)&t2;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t1 = t3;
  t8 = (uint8_t)t1;
  t9 = fmt_digits2__5509(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t1 = t3;
 t11 = (uint32_t)a1;
 t11 = t1 % t11;
 t5 = t4;
 t5 = t5 - (uintptr_t)1ul;
 t4 = t5;
 t5 = t4;
 t6 = (uint8_t *)&t2[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__5525(t8, a2);
 (*t6) = t8;
 t11 = t3;
 t1 = (uint32_t)a1;
 t1 = t11 / t1;
 t3 = t1;
 t1 = t3;
 t0 = t1 == UINT32_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 < INT32_C(0);
 if (t0) {
  t5 = t4;
  t5 = t5 - (uintptr_t)1ul;
  t4 = t5;
  t5 = t4;
  t6 = (uint8_t *)&t2[t5];
  (*t6) = UINT8_C(45);
  goto zig_block_6;
 }
 t12 = a3.width;
 t0 = t12.is_null == true;
 if (t0) {
  t13 = true;
  goto zig_block_8;
 }
 t12 = a3.width;
 t5 = t12.payload;
 t14 = t5;
 t0 = t14 == UINT64_C(0);
 t13 = t0;
 goto zig_block_8;

 zig_block_8:;
 if (t13) {
  goto zig_block_7;
 }
 t5 = t4;
 t5 = t5 - (uintptr_t)1ul;
 t4 = t5;
 t5 = t4;
 t6 = (uint8_t *)&t2[t5];
 (*t6) = UINT8_C(43);
 goto zig_block_7;

 zig_block_7:;
 goto zig_block_6;

 zig_block_6:;
 t5 = t4;
 t6 = (uint8_t *)&t2;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)33ul - t5;
 t15.ptr = t6;
 t15.len = t5;
 memcpy(&t16, &t15, sizeof(nav__6154_47));
 t17 = fmt_formatBuf__anon_5176__6129(t16, a3, a4);
 memcpy(&t18, &t17, sizeof(uint16_t));
 return t18;
}

static uint16_t fmt_formatIntValue__anon_5944__6153(int32_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5955__6154(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5848__6146(int32_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5944__6153(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5881__6148(struct io_Writer__3878 const a0, nav__6148_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5966__6155(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5886__6149(struct io_Writer__3878 const a0, nav__6149_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5970__6156(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5893__6150(struct io_Writer__3878 const a0, nav__6150_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5974__6157(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t io_Writer_print__anon_5899__6151(struct io_Writer__3878 const a0, nav__6151_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_5978__6158(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatIntValue__anon_5982__6159(uint64_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_5237__6139(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_5901__6152(uint64_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_5982__6159(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static nav__5765_38 unicode_utf8Decode3AllowSurrogateHalf__5765(nav__5765_40 const a0) {
 uint32_t t3;
 uint32_t t4;
 uint32_t t2;
 nav__5765_38 t5;
 uint8_t t0;
 bool t1;
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(240);
 t1 = t0 == UINT8_C(224);
 debug_assert__180(t1);
 t0 = a0.array[(uintptr_t)0ul];
 t0 = t0 & UINT8_C(15);
 t3 = (uint32_t)t0;
 t2 = t3;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5765_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = t2;
 t3 = zig_shlw_u32(t3, UINT8_C(6), UINT8_C(21));
 t2 = t3;
 t3 = t2;
 t0 = a0.array[(uintptr_t)1ul];
 t0 = t0 & UINT8_C(63);
 t4 = (uint32_t)t0;
 t4 = t3 | t4;
 t2 = t4;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(192);
 t1 = t0 != UINT8_C(128);
 if (t1) {
  return (nav__5765_38){UINT32_C(0xaaaaa),zig_error_Utf8ExpectedContinuation};
 }
 goto zig_block_1;

 zig_block_1:;
 t4 = t2;
 t4 = zig_shlw_u32(t4, UINT8_C(6), UINT8_C(21));
 t2 = t4;
 t4 = t2;
 t0 = a0.array[(uintptr_t)2ul];
 t0 = t0 & UINT8_C(63);
 t3 = (uint32_t)t0;
 t3 = t4 | t3;
 t2 = t3;
 t3 = t2;
 t1 = t3 < UINT32_C(2048);
 if (t1) {
  return (nav__5765_38){UINT32_C(0xaaaaa),zig_error_Utf8OverlongEncoding};
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t5.payload = t3;
 t5.error = UINT16_C(0);
 return t5;
}

static nav__1491_38 array_list_addOrOom__1491(uintptr_t const a0, uintptr_t const a1) {
 nav__1491_42 t2;
 uintptr_t t3;
 uintptr_t t0;
 uintptr_t const *t5;
 uint8_t const *t6;
 nav__1491_38 t8;
 uint8_t t4;
 uint8_t t1;
 bool t7;
 t2.f1 = zig_addo_u64(&t2.f0, a0, a1, UINT8_C(64));
 t3 = t2.f0;
 t0 = t3;
 t4 = t2.f1;
 t1 = t4;
 t5 = (uintptr_t const *)&t0;
 t6 = (uint8_t const *)&t1;
 t4 = (*t6);
 t7 = t4 != UINT8_C(0);
 if (t7) {
  return (nav__1491_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,zig_error_OutOfMemory};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t5);
 t8.payload = t3;
 t8.error = UINT16_C(0);
 return t8;
}

static uint16_t fmt_format__anon_5966__6155(struct io_Writer__3878 const a0, nav__6155_39 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6155_44){(uint8_t const *)&__anon_6005,(uintptr_t)12ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5746__6143(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6155_44){(uint8_t const *)&__anon_6015,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5970__6156(struct io_Writer__3878 const a0, nav__6156_39 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6156_44){(uint8_t const *)&__anon_6034,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5746__6143(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6156_44){(uint8_t const *)&__anon_6015,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5974__6157(struct io_Writer__3878 const a0, nav__6157_39 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6157_44){(uint8_t const *)&__anon_6066,(uintptr_t)15ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5746__6143(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6157_44){(uint8_t const *)&__anon_6015,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint16_t fmt_format__anon_5978__6158(struct io_Writer__3878 const a0, nav__6158_39 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint64_t t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6158_44){(uint8_t const *)&__anon_6094,(uintptr_t)13ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_5746__6143(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6158_44){(uint8_t const *)&__anon_6015,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

void c_instrument_hooks_deinit__239(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t4;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 bool t0;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t4 = (*t3);
  instrument_hooks_InstrumentHooks_deinit__731(t4);
  mem_Allocator_destroy__anon_6116__6160((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)}, t1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void instrument_hooks_InstrumentHooks_deinit__731(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *const *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *t0;
 struct instruments_root_Instrument__554 *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *const *t5;
 struct instruments_root_Instrument__554 t6;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t9;
 struct environment_root_Environment__559 *t10;
 uint8_t t7;
 t0 = a0;
 t1 = (struct InstrumentHooks *const *)&t0;
 t2 = (*t1);
 t3 = (struct instruments_root_Instrument__554 *)&t2->instrument;
 t4 = t3;
 t5 = (struct instruments_root_Instrument__554 *const *)&t4;
 t6 = (*t3);
 t7 = t6.tag;
 switch (t7) {
  case UINT8_C(0): {
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t3 = (*t5);
   t8 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t3->payload.walltime;
   instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_deinit__808(t8);
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t3 = (*t5);
   t9 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t3->payload.analysis;
   instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__821(t9);
   goto zig_block_0;
  }
  case UINT8_C(3): {
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t2 = (*t1);
 t10 = (struct environment_root_Environment__559 *)&t2->environment;
 environment_root_Environment_deinit__774(t10);
 return;
}

static void mem_Allocator_destroy__anon_6116__6160(struct mem_Allocator__599 const a0, struct InstrumentHooks *const a1) {
 struct mem_Allocator__599 const *t1;
 uint8_t *t2;
 uint8_t *t3;
 uint8_t *const *t4;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t10;
 uint8_t (*t6)[288];
 nav__6160_52 t7;
 uintptr_t t9;
 struct mem_Allocator_VTable__602 const *const *t11;
 struct mem_Allocator_VTable__602 const *t12;
 void (*const *t13)(void *, nav__6160_52, uint8_t, uintptr_t);
 void (*t14)(void *, nav__6160_52, uint8_t, uintptr_t);
 void *t15;
 uint8_t t8;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (uint8_t *)a1;
 t3 = t2;
 t4 = (uint8_t *const *)&t3;
 t5 = (*t1);
 t2 = (*t4);
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t6 = (uint8_t (*)[288])t2;
 t7.ptr = &(*t6)[(uintptr_t)0ul];
 t7.len = (uintptr_t)288ul;
 t8 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t9 = (uintptr_t)zig_return_address();
 t10 = t5;
 t1 = (struct mem_Allocator__599 const *)&t10;
 t11 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t12 = (*t11);
 t13 = (void (*const *)(void *, nav__6160_52, uint8_t, uintptr_t))&t12->free;
 t14 = (*t13);
 t15 = t5.ptr;
 t14(t15, t7, t8, t9);
 return;
}

static void instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_deinit__808(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 runner_fifo_RunnerFifo_deinit__1131(t3);
 return;
}

static void instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_deinit__821(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 runner_fifo_RunnerFifo_deinit__1131(t3);
 return;
}

static void environment_root_Environment_deinit__774(struct environment_root_Environment__559 *const a0) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t3;
 struct environment_root_Environment__559 *t0;
 struct environment_root_EnvironmentJson__845 *t4;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *t5;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 t2;
 nav__774_87 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 t12;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 const *t13;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t15;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t17;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t18;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 t19;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 t14;
 nav__774_104 t20;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6181 t21;
 struct mem_Allocator__599 *t22;
 struct mem_Allocator__599 t23;
 nav__774_44 *t24;
 nav__774_44 t25;
 struct environment_root_EnvironmentValue__856 *t26;
 struct environment_root_EnvironmentValue__856 t27;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *t29;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t28;
 nav__774_122 t33;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213 t34;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t35;
 bool t10;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *)&t4->integration_environment;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *)&t5->map;
 t7 = (*t6);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3800(t7);
 t2 = t8;
 zig_loop_14:
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6161(&t2);
 t10 = t9.is_null != true;
 if (t10) {
  t11 = t9.payload;
  t12 = t11;
  t13 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 const *)&t12;
  t15 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *)&t13->value_ptr;
  t16 = (*t15);
  t17 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *)&t16->map;
  t18 = (*t17);
  t19 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4011(t18);
  t14 = t19;
  zig_loop_32:
  t20 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6163(&t14);
  t10 = t20.is_null != true;
  if (t10) {
   t21 = t20.payload;
   t3 = (*t1);
   t22 = (struct mem_Allocator__599 *)&t3->allocator;
   t23 = (*t22);
   t24 = t21.key_ptr;
   t25 = (*t24);
   mem_Allocator_free__anon_3116__5296(t23, t25);
   t3 = (*t1);
   t26 = t21.value_ptr;
   t27 = (*t26);
   environment_root_Environment_freeEnvironmentValue__777(t3, t27);
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_32;

  zig_block_2:;
  t15 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *)&t13->value_ptr;
  t16 = (*t15);
  t17 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *)&t16->map;
  t22 = (struct mem_Allocator__599 *)&a0->allocator;
  t23 = (*t22);
  array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4003(t17, t23);
  t3 = (*t1);
  t22 = (struct mem_Allocator__599 *)&t3->allocator;
  t23 = (*t22);
  t24 = t11.key_ptr;
  t25 = (*t24);
  mem_Allocator_free__anon_3116__5296(t23, t25);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_14;

 zig_block_0:;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *)&t4->integration_environment;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *)&t5->map;
 t22 = (struct mem_Allocator__599 *)&a0->allocator;
 t23 = (*t22);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3792(t6, t23);
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
 t29 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *)&t4->linked_libraries;
 t30 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *)&t29->map;
 t31 = (*t30);
 t32 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4190(t31);
 t28 = t32;
 zig_loop_91:
 t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6165(&t28);
 t10 = t33.is_null != true;
 if (t10) {
  t34 = t33.payload;
  t3 = (*t1);
  t22 = (struct mem_Allocator__599 *)&t3->allocator;
  t23 = (*t22);
  t24 = t34.key_ptr;
  t25 = (*t24);
  mem_Allocator_free__anon_3116__5296(t23, t25);
  goto zig_block_5;
 }
 goto zig_block_4;

 zig_block_5:;
 goto zig_loop_91;

 zig_block_4:;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
 t29 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *)&t4->linked_libraries;
 t30 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *)&t29->map;
 t22 = (struct mem_Allocator__599 *)&a0->allocator;
 t23 = (*t22);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4182(t30, t23);
 t3 = (*t1);
 t35 = (struct environment_linked_libraries_root_LinkedLibraries__849 *)&t3->libs;
 environment_linked_libraries_root_LinkedLibraries_deinit__3644(t35);
 return;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3800(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t6;
 nav__3800_42 **t8;
 nav__3800_62 t10;
 nav__3800_42 *t11;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **t12;
 nav__3800_65 t13;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t14;
 uint32_t *t15;
 uintptr_t t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 t7;
 uint32_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *)&t1->entries;
 t3 = (*t2);
 t4 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t3);
 t5 = t4;
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t5;
 t8 = (nav__3800_42 **)&t7.keys;
 t9 = (*t6);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t9);
 t11 = t10.ptr;
 (*t8) = t11;
 t12 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t7.values;
 t9 = (*t6);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t9);
 t14 = t13.ptr;
 (*t12) = t14;
 t15 = (uint32_t *)&t7.len;
 t16 = t4.len;
 t17 = (uint32_t)t16;
 (*t15) = t17;
 t15 = (uint32_t *)&t7.index;
 (*t15) = UINT32_C(0);
 return t7;
}

static nav__6161_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6161(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 *const *t1;
 uint32_t *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 *t0;
 nav__6161_44 **t7;
 uintptr_t t8;
 nav__6161_44 *t9;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **t10;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 t12;
 nav__6161_39 t13;
 uint32_t t3;
 uint32_t t4;
 bool t5;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 *const *)&t0;
 t2 = (uint32_t *)&a0->index;
 t3 = (*t2);
 t2 = (uint32_t *)&a0->len;
 t4 = (*t2);
 t5 = t3 >= t4;
 if (t5) {
  return (nav__6161_39){{((nav__6161_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},true};
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = (*t1);
 t7 = (nav__6161_44 **)&t6->keys;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t9 = (*t7);
 t9 = (nav__6161_44 *)&t9[t8];
 t6 = (*t1);
 t10 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t6->values;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t11 = (*t10);
 t11 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t11[t8];
 t12.key_ptr = t9;
 t12.value_ptr = t11;
 t6 = (*t1);
 t2 = (uint32_t *)&t6->index;
 t4 = (*t2);
 t4 = t4 + UINT32_C(1);
 (*t2) = t4;
 t13.is_null = false;
 t13.payload = t12;
 return t13;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4011(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t6;
 nav__4011_42 **t8;
 nav__4011_62 t10;
 nav__4011_42 *t11;
 struct environment_root_EnvironmentValue__856 **t12;
 nav__4011_65 t13;
 struct environment_root_EnvironmentValue__856 *t14;
 uint32_t *t15;
 uintptr_t t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 t7;
 uint32_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *)&t1->entries;
 t3 = (*t2);
 t4 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t3);
 t5 = t4;
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t5;
 t8 = (nav__4011_42 **)&t7.keys;
 t9 = (*t6);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t9);
 t11 = t10.ptr;
 (*t8) = t11;
 t12 = (struct environment_root_EnvironmentValue__856 **)&t7.values;
 t9 = (*t6);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t9);
 t14 = t13.ptr;
 (*t12) = t14;
 t15 = (uint32_t *)&t7.len;
 t16 = t4.len;
 t17 = (uint32_t)t16;
 (*t15) = t17;
 t15 = (uint32_t *)&t7.index;
 (*t15) = UINT32_C(0);
 return t7;
}

static nav__6163_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6163(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 *const *t1;
 uint32_t *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 *t0;
 nav__6163_44 **t7;
 uintptr_t t8;
 nav__6163_44 *t9;
 struct environment_root_EnvironmentValue__856 **t10;
 struct environment_root_EnvironmentValue__856 *t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6181 t12;
 nav__6163_39 t13;
 uint32_t t3;
 uint32_t t4;
 bool t5;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 *const *)&t0;
 t2 = (uint32_t *)&a0->index;
 t3 = (*t2);
 t2 = (uint32_t *)&a0->len;
 t4 = (*t2);
 t5 = t3 >= t4;
 if (t5) {
  return (nav__6163_39){{((nav__6163_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_root_EnvironmentValue__856 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},true};
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = (*t1);
 t7 = (nav__6163_44 **)&t6->keys;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t9 = (*t7);
 t9 = (nav__6163_44 *)&t9[t8];
 t6 = (*t1);
 t10 = (struct environment_root_EnvironmentValue__856 **)&t6->values;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t11 = (*t10);
 t11 = (struct environment_root_EnvironmentValue__856 *)&t11[t8];
 t12.key_ptr = t9;
 t12.value_ptr = t11;
 t6 = (*t1);
 t2 = (uint32_t *)&t6->index;
 t4 = (*t2);
 t4 = t4 + UINT32_C(1);
 (*t2) = t4;
 t13.is_null = false;
 t13.payload = t12;
 return t13;
}

static void environment_root_Environment_freeEnvironmentValue__777(struct environment_root_Environment__559 *const a0, struct environment_root_EnvironmentValue__856 const a1) {
 struct environment_root_Environment__559 *const *t1;
 nav__777_43 t3;
 struct environment_root_Environment__559 *t4;
 struct environment_root_Environment__559 *t0;
 struct mem_Allocator__599 *t5;
 struct mem_Allocator__599 t6;
 nav__777_46 t7;
 uintptr_t t9;
 uintptr_t t10;
 uintptr_t t8;
 uint64_t t11;
 uint64_t t12;
 uint8_t t2;
 bool t13;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = a1.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a1.payload.string;
   t4 = (*t1);
   t5 = (struct mem_Allocator__599 *)&t4->allocator;
   t6 = (*t5);
   mem_Allocator_free__anon_3116__5296(t6, t3);
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t7 = a1.payload.list;
   t8 = (uintptr_t)0ul;
   t9 = t7.len;
   zig_loop_20:
   t10 = t8;
   t11 = t10;
   t12 = t9;
   t13 = t11 < t12;
   if (t13) {
    t3 = t7.ptr[t10];
    t4 = (*t1);
    t5 = (struct mem_Allocator__599 *)&t4->allocator;
    t6 = (*t5);
    mem_Allocator_free__anon_3116__5296(t6, t3);
    goto zig_block_2;
   }
   goto zig_block_1;

   zig_block_2:;
   t10 = t10 + (uintptr_t)1ul;
   t8 = t10;
   goto zig_loop_20;

   zig_block_1:;
   t4 = (*t1);
   t5 = (struct mem_Allocator__599 *)&t4->allocator;
   t6 = (*t5);
   mem_Allocator_free__anon_6426__6191(t6, t7);
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4003(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t3;
 struct array_hash_map_IndexHeader__892 **t4;
 struct array_hash_map_IndexHeader__892 *t5;
 struct array_hash_map_IndexHeader__892 *t7;
 struct array_hash_map_IndexHeader__892 *t8;
 struct array_hash_map_IndexHeader__892 *const *t9;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 debug_SafetyLock_assertUnlocked__3984();
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t2->entries;
 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4129(t3, a1);
 t4 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t8 = t7;
  t9 = (struct array_hash_map_IndexHeader__892 *const *)&t8;
  t7 = (*t9);
  array_hash_map_IndexHeader_free__3979(t7, a1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3792(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t3;
 struct array_hash_map_IndexHeader__892 **t4;
 struct array_hash_map_IndexHeader__892 *t5;
 struct array_hash_map_IndexHeader__892 *t7;
 struct array_hash_map_IndexHeader__892 *t8;
 struct array_hash_map_IndexHeader__892 *const *t9;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 debug_SafetyLock_assertUnlocked__3984();
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t2->entries;
 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3935(t3, a1);
 t4 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t8 = t7;
  t9 = (struct array_hash_map_IndexHeader__892 *const *)&t8;
  t7 = (*t9);
  array_hash_map_IndexHeader_free__3979(t7, a1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4190(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t6;
 nav__4190_42 **t8;
 nav__4190_62 t10;
 nav__4190_42 *t11;
 struct environment_linked_libraries_root_LibraryEntry__864 **t12;
 nav__4190_65 t13;
 struct environment_linked_libraries_root_LibraryEntry__864 *t14;
 uint32_t *t15;
 uintptr_t t16;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t7;
 uint32_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *)&t1->entries;
 t3 = (*t2);
 t4 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t3);
 t5 = t4;
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t5;
 t8 = (nav__4190_42 **)&t7.keys;
 t9 = (*t6);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t9);
 t11 = t10.ptr;
 (*t8) = t11;
 t12 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t7.values;
 t9 = (*t6);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t9);
 t14 = t13.ptr;
 (*t12) = t14;
 t15 = (uint32_t *)&t7.len;
 t16 = t4.len;
 t17 = (uint32_t)t16;
 (*t15) = t17;
 t15 = (uint32_t *)&t7.index;
 (*t15) = UINT32_C(0);
 return t7;
}

static nav__6165_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6165(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 *const *t1;
 uint32_t *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 *t0;
 nav__6165_44 **t7;
 uintptr_t t8;
 nav__6165_44 *t9;
 struct environment_linked_libraries_root_LibraryEntry__864 **t10;
 struct environment_linked_libraries_root_LibraryEntry__864 *t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213 t12;
 nav__6165_39 t13;
 uint32_t t3;
 uint32_t t4;
 bool t5;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 *const *)&t0;
 t2 = (uint32_t *)&a0->index;
 t3 = (*t2);
 t2 = (uint32_t *)&a0->len;
 t4 = (*t2);
 t5 = t3 >= t4;
 if (t5) {
  return (nav__6165_39){{((nav__6165_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_linked_libraries_root_LibraryEntry__864 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul)},true};
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = (*t1);
 t7 = (nav__6165_44 **)&t6->keys;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t9 = (*t7);
 t9 = (nav__6165_44 *)&t9[t8];
 t6 = (*t1);
 t10 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t6->values;
 t2 = (uint32_t *)&a0->index;
 t4 = (*t2);
 t8 = (uintptr_t)t4;
 t11 = (*t10);
 t11 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t11[t8];
 t12.key_ptr = t9;
 t12.value_ptr = t11;
 t6 = (*t1);
 t2 = (uint32_t *)&t6->index;
 t4 = (*t2);
 t4 = t4 + UINT32_C(1);
 (*t2) = t4;
 t13.is_null = false;
 t13.payload = t12;
 return t13;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4182(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t3;
 struct array_hash_map_IndexHeader__892 **t4;
 struct array_hash_map_IndexHeader__892 *t5;
 struct array_hash_map_IndexHeader__892 *t7;
 struct array_hash_map_IndexHeader__892 *t8;
 struct array_hash_map_IndexHeader__892 *const *t9;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 debug_SafetyLock_assertUnlocked__3984();
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t2->entries;
 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4308(t3, a1);
 t4 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t8 = t7;
  t9 = (struct array_hash_map_IndexHeader__892 *const *)&t8;
  t7 = (*t9);
  array_hash_map_IndexHeader_free__3979(t7, a1);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return;
}

static void environment_linked_libraries_root_LinkedLibraries_deinit__3644(struct environment_linked_libraries_root_LinkedLibraries__849 *const a0) {
 struct environment_linked_libraries_root_LinkedLibraries__849 *const *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t3;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t4;
 nav__3644_50 t5;
 uintptr_t t6;
 uintptr_t t7;
 uintptr_t t2;
 uint64_t t8;
 uint64_t t9;
 struct environment_linked_libraries_root_LibraryInfo__946 t11;
 struct mem_Allocator__599 *t12;
 struct mem_Allocator__599 t13;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t14;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t0;
 bool t10;
 t0 = a0;
 t1 = (struct environment_linked_libraries_root_LinkedLibraries__849 *const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *)&a0->libraries;
 t4 = (*t3);
 t5 = t4.items;
 t6 = t5.len;
 zig_loop_11:
 t7 = t2;
 t8 = t7;
 t9 = t6;
 t10 = t8 < t9;
 if (t10) {
  t11 = t5.ptr[t7];
  t12 = (struct mem_Allocator__599 *)&a0->allocator;
  t13 = (*t12);
  environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__3645(t13, t11);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t7 = t7 + (uintptr_t)1ul;
 t2 = t7;
 goto zig_loop_11;

 zig_block_0:;
 t14 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *)&t14->libraries;
 t4 = (*t3);
 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_deinit__4349(t4);
 return;
}

static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const a0) {
 static uintptr_t const t10[3] = {16ul,32ul,4ul};
 static uintptr_t const t12[3] = {1ul,2ul,0ul};
 uint8_t *(*t1)[3];
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t11;
 uintptr_t t13;
 uintptr_t t7;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *t4;
 uint64_t t8;
 uint8_t **t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t0;
 bool t9;
 t1 = (uint8_t *(*)[3])&t0.ptrs;
 t2 = (uintptr_t *)&t0.len;
 t3 = a0.len;
 (*t2) = t3;
 t2 = (uintptr_t *)&t0.capacity;
 t3 = a0.capacity;
 (*t2) = t3;
 t5 = a0.bytes;
 t6 = (uint8_t *)t5;
 t4 = t6;
 t7 = (uintptr_t)0ul;
 zig_loop_17:
 t3 = t7;
 t8 = t3;
 t9 = t8 < UINT64_C(3);
 if (t9) {
  t11 = t10[t3];
  t13 = t12[t3];
  t1 = (uint8_t *(*)[3])&t0.ptrs;
  t14 = (uint8_t **)&(*t1)[t13];
  t6 = t4;
  (*t14) = t6;
  t6 = t4;
  t13 = a0.capacity;
  t13 = t11 * t13;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t13*sizeof(uint8_t)));
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t7 = t3;
 goto zig_loop_17;

 zig_block_0:;
 t15 = t0;
 return t15;
}

static nav__6180_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 nav__6180_39 *t5;
 nav__6180_39 *t6;
 nav__6180_39 *const *t7;
 nav__6180_41 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6180_41){(nav__6180_39 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)1ul];
 t5 = (nav__6180_39 *)t4;
 t6 = t5;
 t7 = (nav__6180_39 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (nav__6180_39 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(nav__6180_39)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__6181_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t5;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t6;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t7;
 nav__6181_40 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6181_40){(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)2ul];
 t5 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)t4;
 t6 = t5;
 t7 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const a0) {
 static uintptr_t const t10[3] = {16ul,24ul,4ul};
 static uintptr_t const t12[3] = {1ul,2ul,0ul};
 uint8_t *(*t1)[3];
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t11;
 uintptr_t t13;
 uintptr_t t7;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *t4;
 uint64_t t8;
 uint8_t **t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t0;
 bool t9;
 t1 = (uint8_t *(*)[3])&t0.ptrs;
 t2 = (uintptr_t *)&t0.len;
 t3 = a0.len;
 (*t2) = t3;
 t2 = (uintptr_t *)&t0.capacity;
 t3 = a0.capacity;
 (*t2) = t3;
 t5 = a0.bytes;
 t6 = (uint8_t *)t5;
 t4 = t6;
 t7 = (uintptr_t)0ul;
 zig_loop_17:
 t3 = t7;
 t8 = t3;
 t9 = t8 < UINT64_C(3);
 if (t9) {
  t11 = t10[t3];
  t13 = t12[t3];
  t1 = (uint8_t *(*)[3])&t0.ptrs;
  t14 = (uint8_t **)&(*t1)[t13];
  t6 = t4;
  (*t14) = t6;
  t6 = t4;
  t13 = a0.capacity;
  t13 = t11 * t13;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t13*sizeof(uint8_t)));
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t7 = t3;
 goto zig_loop_17;

 zig_block_0:;
 t15 = t0;
 return t15;
}

static nav__6189_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 nav__6189_39 *t5;
 nav__6189_39 *t6;
 nav__6189_39 *const *t7;
 nav__6189_41 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6189_41){(nav__6189_39 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)1ul];
 t5 = (nav__6189_39 *)t4;
 t6 = t5;
 t7 = (nav__6189_39 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (nav__6189_39 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(nav__6189_39)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__6190_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 struct environment_root_EnvironmentValue__856 *t5;
 struct environment_root_EnvironmentValue__856 *t6;
 struct environment_root_EnvironmentValue__856 *const *t7;
 nav__6190_40 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6190_40){(struct environment_root_EnvironmentValue__856 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)2ul];
 t5 = (struct environment_root_EnvironmentValue__856 *)t4;
 t6 = t5;
 t7 = (struct environment_root_EnvironmentValue__856 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (struct environment_root_EnvironmentValue__856 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_root_EnvironmentValue__856)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static void mem_Allocator_free__anon_6426__6191(struct mem_Allocator__599 const a0, nav__6191_42 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6191_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t const *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 nav__6191_54 t10;
 nav__6191_54 t12;
 struct mem_Allocator__599 t11;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t14;
 struct mem_Allocator_VTable__602 const *const *t15;
 struct mem_Allocator_VTable__602 const *t16;
 void (*const *t17)(void *, nav__6191_54, uint8_t, uintptr_t);
 void (*t18)(void *, nav__6191_54, uint8_t, uintptr_t);
 void *t19;
 bool t5;
 uint8_t t13;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6643__6272(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 t11 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t10.ptr = t7;
 t10.len = t3;
 memcpy(&t12, &t10, sizeof(nav__6191_54));
 t13 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t14 = t11;
 t1 = (struct mem_Allocator__599 const *)&t14;
 t15 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t16 = (*t15);
 t17 = (void (*const *)(void *, nav__6191_54, uint8_t, uintptr_t))&t16->free;
 t18 = (*t17);
 t19 = t11.ptr;
 t18(t19, t12, t13, t3);
 return;
}

static void debug_SafetyLock_assertUnlocked__3984(void) {
 return;
}

static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4129(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const a0, struct mem_Allocator__599 const a1) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *t1;
 struct mem_Allocator__599 const *t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 nav__4129_51 t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t3);
 t5 = (*t1);
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4161(t6);
 mem_Allocator_free__anon_6654__6273(t4, t7);
 return;
}

static void array_hash_map_IndexHeader_free__3979(struct array_hash_map_IndexHeader__892 *const a0, struct mem_Allocator__599 const a1) {
 struct array_hash_map_IndexHeader__892 *const *t1;
 struct mem_Allocator__599 const *t3;
 uint8_t *t4;
 uint8_t *t7;
 uintptr_t t6;
 uintptr_t t11;
 uint8_t *const *t8;
 struct array_hash_map_IndexHeader__892 *t9;
 struct array_hash_map_IndexHeader__892 *t0;
 nav__3979_52 t12;
 struct mem_Allocator__599 t13;
 struct mem_Allocator__599 t2;
 struct array_hash_map_IndexHeader__892 t10;
 uint8_t t5;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__892 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uint8_t *)&a0->bit_index;
 t5 = (*t4);
 t6 = array_hash_map_capacityIndexSize__3742(t5);
 t4 = (uint8_t *)a0;
 t7 = t4;
 t8 = (uint8_t *const *)&t7;
 t9 = (*t1);
 t10 = (*t9);
 t11 = array_hash_map_IndexHeader_length__3975(t10);
 t6 = t11 * t6;
 t6 = (uintptr_t)4ul + t6;
 t4 = (*t8);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t4;
 t12.len = t6;
 t13 = (*t3);
 mem_Allocator_free__anon_6672__6274(t13, t12);
 return;
}

static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3935(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const a0, struct mem_Allocator__599 const a1) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *t1;
 struct mem_Allocator__599 const *t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 nav__3935_51 t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t3);
 t5 = (*t1);
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3967(t6);
 mem_Allocator_free__anon_6654__6273(t4, t7);
 return;
}

static struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const a0) {
 static uintptr_t const t10[3] = {16ul,32ul,4ul};
 static uintptr_t const t12[3] = {1ul,2ul,0ul};
 uint8_t *(*t1)[3];
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t11;
 uintptr_t t13;
 uintptr_t t7;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *t4;
 uint64_t t8;
 uint8_t **t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t0;
 bool t9;
 t1 = (uint8_t *(*)[3])&t0.ptrs;
 t2 = (uintptr_t *)&t0.len;
 t3 = a0.len;
 (*t2) = t3;
 t2 = (uintptr_t *)&t0.capacity;
 t3 = a0.capacity;
 (*t2) = t3;
 t5 = a0.bytes;
 t6 = (uint8_t *)t5;
 t4 = t6;
 t7 = (uintptr_t)0ul;
 zig_loop_17:
 t3 = t7;
 t8 = t3;
 t9 = t8 < UINT64_C(3);
 if (t9) {
  t11 = t10[t3];
  t13 = t12[t3];
  t1 = (uint8_t *(*)[3])&t0.ptrs;
  t14 = (uint8_t **)&(*t1)[t13];
  t6 = t4;
  (*t14) = t6;
  t6 = t4;
  t13 = a0.capacity;
  t13 = t11 * t13;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t13*sizeof(uint8_t)));
  t4 = t6;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t7 = t3;
 goto zig_loop_17;

 zig_block_0:;
 t15 = t0;
 return t15;
}

static nav__6199_41 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 nav__6199_39 *t5;
 nav__6199_39 *t6;
 nav__6199_39 *const *t7;
 nav__6199_41 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6199_41){(nav__6199_39 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)1ul];
 t5 = (nav__6199_39 *)t4;
 t6 = t5;
 t7 = (nav__6199_39 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (nav__6199_39 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(nav__6199_39)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__6200_40 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 struct environment_linked_libraries_root_LibraryEntry__864 *t5;
 struct environment_linked_libraries_root_LibraryEntry__864 *t6;
 struct environment_linked_libraries_root_LibraryEntry__864 *const *t7;
 nav__6200_40 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6200_40){(struct environment_linked_libraries_root_LibraryEntry__864 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)2ul];
 t5 = (struct environment_linked_libraries_root_LibraryEntry__864 *)t4;
 t6 = t5;
 t7 = (struct environment_linked_libraries_root_LibraryEntry__864 *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (struct environment_linked_libraries_root_LibraryEntry__864 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryEntry__864)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static void multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4308(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const a0, struct mem_Allocator__599 const a1) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *t1;
 struct mem_Allocator__599 const *t3;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 t2;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 nav__4308_51 t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (*t3);
 t5 = (*t1);
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4340(t6);
 mem_Allocator_free__anon_6654__6273(t4, t7);
 return;
}

static void environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__3645(struct mem_Allocator__599 const a0, struct environment_linked_libraries_root_LibraryInfo__946 const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 nav__3645_46 t3;
 nav__3645_46 t5;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = a1.path;
 mem_Allocator_free__anon_3116__5296(t2, t3);
 t3 = a1.soname;
 t4 = t3.ptr != NULL;
 if (t4) {
  t5 = t3;
  t2 = (*t1);
  mem_Allocator_free__anon_3116__5296(t2, t5);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = a1.build_id;
 t4 = t5.ptr != NULL;
 if (t4) {
  t3 = t5;
  t2 = (*t1);
  mem_Allocator_free__anon_3116__5296(t2, t3);
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 return;
}

static void array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_deinit__4349(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 const *t1;
 struct mem_Allocator__599 const *t2;
 struct mem_Allocator__599 t3;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t4;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t0;
 nav__4349_42 t5;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 const *)&t0;
 t2 = (struct mem_Allocator__599 const *)&t1->allocator;
 t3 = (*t2);
 t4 = (*t1);
 t5 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__4395(t4);
 mem_Allocator_free__anon_6701__6275(t3, t5);
 return;
}

static nav__6272_39 mem_sliceAsBytes__anon_6643__6272(nav__6272_41 const a0) {
 uintptr_t t0;
 uint64_t t1;
 nav__6272_39 const *t4;
 uint8_t const *t5;
 uint8_t const *t6;
 uint8_t const *const *t7;
 nav__6272_39 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6272_39){(uint8_t const *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (uint8_t const *)t4;
 t6 = t5;
 t7 = (uint8_t const *const *)&t6;
 t0 = a0.len;
 t0 = t0 * (uintptr_t)16ul;
 t5 = (*t7);
 t5 = (uint8_t const *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__4161_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4161(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *t1;
 uint8_t *const *t2;
 uintptr_t t3;
 uint8_t *t4;
 nav__4161_39 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t0;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *)&t0;
 t2 = (uint8_t *const *)&t1->bytes;
 t3 = a0.capacity;
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4160(t3);
 t4 = (*t2);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t3;
 return t5;
}

static void mem_Allocator_free__anon_6654__6273(struct mem_Allocator__599 const a0, nav__6273_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6273_40 t2;
 nav__6273_40 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__6273_40, uint8_t, uintptr_t);
 void (*t17)(void *, nav__6273_40, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6720__6276(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__6273_40));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__6273_40, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static uintptr_t array_hash_map_capacityIndexSize__3742(uint8_t const a0) {
 uint8_t t0;
 t0 = array_hash_map_capacityIndexType__3741(a0);
 switch (t0) {
  case UINT8_C(0): {
   return ((uintptr_t)(uintptr_t)0x2ul);
  }
  case UINT8_C(1): {
   return ((uintptr_t)(uintptr_t)0x4ul);
  }
  case UINT8_C(2): {
   return ((uintptr_t)(uintptr_t)0x8ul);
  }
  default: zig_unreachable();
 }
}

static uintptr_t array_hash_map_IndexHeader_length__3975(struct array_hash_map_IndexHeader__892 const a0) {
 uintptr_t t1;
 uint8_t t0;
 t0 = a0.bit_index;
 t1 = zig_shlw_u64((uintptr_t)1ul, t0, UINT8_C(64));
 return t1;
}

static void mem_Allocator_free__anon_6672__6274(struct mem_Allocator__599 const a0, nav__6274_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6274_40 t2;
 nav__6274_40 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__6274_40, uint8_t, uintptr_t);
 void (*t17)(void *, nav__6274_40, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6754__6292(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__6274_40));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)4ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__6274_40, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static nav__3967_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3967(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *t1;
 uint8_t *const *t2;
 uintptr_t t3;
 uint8_t *t4;
 nav__3967_39 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t0;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *)&t0;
 t2 = (uint8_t *const *)&t1->bytes;
 t3 = a0.capacity;
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3966(t3);
 t4 = (*t2);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t3;
 return t5;
}

static nav__4340_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4340(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *t1;
 uint8_t *const *t2;
 uintptr_t t3;
 uint8_t *t4;
 nav__4340_39 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t0;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *)&t0;
 t2 = (uint8_t *const *)&t1->bytes;
 t3 = a0.capacity;
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4339(t3);
 t4 = (*t2);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5.ptr = t4;
 t5.len = t3;
 return t5;
}

static nav__4395_40 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__4395(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 const *t1;
 nav__4395_40 const *t2;
 struct environment_linked_libraries_root_LibraryInfo__946 *const *t3;
 uintptr_t t4;
 struct environment_linked_libraries_root_LibraryInfo__946 *t5;
 nav__4395_40 t6;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t0;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 const *)&t0;
 t2 = (nav__4395_40 const *)&t1->items;
 t3 = &t2->ptr;
 t4 = a0.capacity;
 t5 = (*t3);
 t5 = (struct environment_linked_libraries_root_LibraryInfo__946 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryInfo__946)));
 t6.ptr = t5;
 t6.len = t4;
 return t6;
}

static void mem_Allocator_free__anon_6701__6275(struct mem_Allocator__599 const a0, nav__6275_41 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6275_51 t2;
 nav__6275_51 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__6275_51, uint8_t, uintptr_t);
 void (*t17)(void *, nav__6275_51, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_6767__6293(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__6275_51));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__6275_51, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4160(uintptr_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)44ul * a0;
 return t0;
}

static nav__6276_39 mem_sliceAsBytes__anon_6720__6276(nav__6276_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__6276_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6276_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint8_t array_hash_map_capacityIndexType__3741(uint8_t const a0) {
 bool t0;
 t0 = a0 <= UINT8_C(8);
 if (t0) {
  return UINT8_C(0);
 }
 goto zig_block_0;

 zig_block_0:;
 t0 = a0 <= UINT8_C(16);
 if (t0) {
  return UINT8_C(1);
 }
 goto zig_block_1;

 zig_block_1:;
 t0 = a0 <= UINT8_C(32);
 debug_assert__180(t0);
 return UINT8_C(2);
}

static nav__6292_39 mem_sliceAsBytes__anon_6754__6292(nav__6292_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__6292_39 t7;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6292_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3966(uintptr_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)52ul * a0;
 return t0;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4339(uintptr_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)52ul * a0;
 return t0;
}

static nav__6293_39 mem_sliceAsBytes__anon_6767__6293(nav__6293_42 const a0) {
 uintptr_t t0;
 uint64_t t1;
 struct environment_linked_libraries_root_LibraryInfo__946 *t4;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *const *t7;
 nav__6293_39 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6293_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (uint8_t *)t4;
 t6 = t5;
 t7 = (uint8_t *const *)&t6;
 t0 = a0.len;
 t0 = t0 * (uintptr_t)48ul;
 t5 = (*t7);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

bool c_instrument_hooks_is_instrumented__240(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 t5;
 bool t0;
 bool t7;
 uint8_t t6;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t5 = (*t4);
  t6 = t5.tag;
  switch (t6) {
   case UINT8_C(0): {
    t6 = running_on_valgrind();
    t7 = t6 > UINT8_C(0);
    t0 = t7;
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t0 = true;
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t0 = true;
    goto zig_block_1;
   }
   case UINT8_C(3): {
    t0 = false;
    goto zig_block_1;
   }
   default: zig_unreachable();
  }

  zig_block_1:;
  return t0;
 }
 goto zig_block_0;

 zig_block_0:;
 return false;
}

static zig_cold uint16_t runner_fifo_RunnerFifo_start_benchmark__1134(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 struct fifo_posix_Pipe_Reader__793 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(1)});
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 t5 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t4 = fifo_posix_Pipe_Reader_waitForAck__1293(t5, (nav__1134_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_start_benchmark__810(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_start_benchmark__1134(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__823(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_start_benchmark__1134(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_start_benchmark__241(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t6;
 struct instruments_root_Instrument__554 *const *t7;
 struct instruments_root_Instrument__554 t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t13;
 uint16_t t5;
 uint16_t t11;
 uint16_t t12;
 bool t0;
 uint8_t t9;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t6 = t4;
  t7 = (struct instruments_root_Instrument__554 *const *)&t6;
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(1);
  if (t0) {
   t4 = (*t7);
   t10 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
   t11 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_start_benchmark__810(t10);
   memcpy(&t12, &t11, sizeof(uint16_t));
   t5 = t12;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(0);
  if (t0) {
   t0 = features_is_feature_enabled__331(UINT64_C(0));
   t0 = !t0;
   if (t0) {
    callgrind_zero_stats();
    callgrind_start_instrumentation();
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t5 = 0;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(2);
  if (t0) {
   t4 = (*t7);
   t13 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
   t12 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_start_benchmark__823(t13);
   memcpy(&t11, &t12, sizeof(uint16_t));
   t5 = t11;
   goto zig_block_2;
  }
  goto zig_block_6;

  zig_block_6:;
  goto zig_block_4;

  zig_block_4:;
  goto zig_block_3;

  zig_block_3:;
  t5 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t12, &t5, sizeof(uint16_t));
  t0 = t12 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static bool features_is_feature_enabled__331(uint64_t const a0) {
 uint64_t t0;
 uint64_t t1;
 uintptr_t t2;
 bool t3;
 t0 = (*&features_features__329);
 t1 = a0;
 t2 = t1;
 t3 = bit_set_IntegerBitSet_2864_29_isSet__366(t0, t2);
 return t3;
}

static bool bit_set_IntegerBitSet_2864_29_isSet__366(uint64_t const a0, uintptr_t const a1) {
 uint64_t t0;
 uint64_t t2;
 bool t1;
 t0 = a1;
 t1 = t0 < UINT64_C(64);
 debug_assert__180(t1);
 t0 = zig_wrap_u64((uint64_t)a0, UINT8_C(64));
 t2 = bit_set_IntegerBitSet_2864_29_maskBit__390(a1);
 t2 = t0 & t2;
 t1 = t2 != UINT64_C(0);
 return t1;
}

static zig_cold uint16_t runner_fifo_RunnerFifo_stop_benchmark__1135(struct runner_fifo_RunnerFifo__631 *const a0) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 struct fifo_posix_Pipe_Reader__793 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, (struct shared_Command__1795){{{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},-INT32_C(0x55555556)}},UINT8_C(2)});
 if (t4) {
  return t4;
 }
 t2 = (*t1);
 t5 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t4 = fifo_posix_Pipe_Reader_waitForAck__1293(t5, (nav__1135_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_stop_benchmark__811(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_stop_benchmark__1135(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__824(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_stop_benchmark__1135(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_stop_benchmark__242(struct InstrumentHooks *const a0) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t6;
 struct instruments_root_Instrument__554 *const *t7;
 struct instruments_root_Instrument__554 t8;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t13;
 uint16_t t5;
 uint16_t t11;
 uint16_t t12;
 bool t0;
 uint8_t t9;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t6 = t4;
  t7 = (struct instruments_root_Instrument__554 *const *)&t6;
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(0);
  if (t0) {
   t0 = features_is_feature_enabled__331(UINT64_C(0));
   t0 = !t0;
   if (t0) {
    callgrind_stop_instrumentation();
    goto zig_block_4;
   }
   goto zig_block_4;

   zig_block_4:;
   t5 = 0;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(1);
  if (t0) {
   t4 = (*t7);
   t10 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
   t11 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_stop_benchmark__811(t10);
   memcpy(&t12, &t11, sizeof(uint16_t));
   t5 = t12;
   goto zig_block_2;
  }
  t8 = (*t4);
  t9 = t8.tag;
  t0 = t9 == UINT8_C(2);
  if (t0) {
   t4 = (*t7);
   t13 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
   t12 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_stop_benchmark__824(t13);
   memcpy(&t11, &t12, sizeof(uint16_t));
   t5 = t11;
   goto zig_block_2;
  }
  goto zig_block_6;

  zig_block_6:;
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_3;

  zig_block_3:;
  t5 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t12, &t5, sizeof(uint16_t));
  t0 = t12 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint16_t runner_fifo_RunnerFifo_set_executed_benchmark__1136(struct runner_fifo_RunnerFifo__631 *const a0, int32_t const a1, uint8_t const *const a2) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 nav__1136_61 t4;
 nav__1136_61 t5;
 struct shared_Command__struct_1798__1798 t6;
 struct shared_Command__1795 t7;
 struct fifo_posix_Pipe_Reader__793 *t9;
 uint16_t t8;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4 = mem_span__anon_6923__6298(a2);
 memcpy(&t5, &t4, sizeof(nav__1136_61));
 t6.uri = t5;
 t6.pid = a1;
 t7.tag = UINT8_C(0);
 t7.payload.ExecutedBenchmark = t6;
 t8 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t7);
 if (t8) {
  return t8;
 }
 t2 = (*t1);
 t9 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t8 = fifo_posix_Pipe_Reader_waitForAck__1293(t9, (nav__1136_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t8) {
  return t8;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_executed_benchmark__812(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0, int32_t const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_executed_benchmark__1136(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__825(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0, int32_t const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_executed_benchmark__1136(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_executed_benchmark__243(struct InstrumentHooks *const a0, int32_t const a1, char const *const a2) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t7;
 uint8_t const *t5;
 uint8_t const *t11;
 struct instruments_root_Instrument__554 *const *t8;
 struct instruments_root_Instrument__554 t9;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t12;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t14;
 uint16_t t6;
 uint16_t t13;
 bool t0;
 uint8_t t10;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t5 = c_toU8__236(a2);
  t7 = t4;
  t8 = (struct instruments_root_Instrument__554 *const *)&t7;
  t9 = (*t4);
  t10 = t9.tag;
  switch (t10) {
   case UINT8_C(0): {
    t11 = (uint8_t const *)t5;
    callgrind_dump_stats_at(t11);
    goto zig_block_3;
   }
   case UINT8_C(1): {
    t4 = (*t8);
    t12 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
    t13 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_executed_benchmark__812(t12, a1, t5);
    if (t13) {
     t6 = t13;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t4 = (*t8);
    t14 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
    t13 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_executed_benchmark__825(t14, a1, t5);
    if (t13) {
     t6 = t13;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(3): {
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t6 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t13, &t6, sizeof(uint16_t));
  t0 = t13 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint8_t const *c_toU8__236(char const *const a0) {
 uint8_t const *t0;
 t0 = (uint8_t const *)a0;
 return t0;
}

static nav__6298_39 mem_span__anon_6923__6298(uint8_t const *const a0) {
 uint8_t const *const *t1;
 uintptr_t t2;
 uint8_t const *t3;
 uint8_t const *t0;
 nav__6298_39 t4;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = mem_len__anon_6940__6299(a0);
 t3 = (*t1);
 t3 = (uint8_t const *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t3;
 t4.len = t2;
 return t4;
}

static uintptr_t mem_len__anon_6940__6299(uint8_t const *const a0) {
 uint8_t const *t1;
 uintptr_t t2;
 bool t0;
 t0 = a0 != NULL;
 debug_assert__180(t0);
 t1 = (uint8_t const *)a0;
 t2 = mem_indexOfSentinel__anon_6947__6300(t1);
 return t2;
}

static uintptr_t mem_indexOfSentinel__anon_6947__6300(uint8_t const *const a0) {
 static uint8_t const t11[16] = "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
 static uint8_t const t18[16] = "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000";
 uint8_t const *const *t1;
 uintptr_t t4;
 uintptr_t t6;
 uintptr_t t16;
 uintptr_t t2;
 uint8_t const *t5;
 uint8_t const *t0;
 uint64_t t7;
 uint8_t const (*t8)[16];
 uint8_t const (*t17)[16];
 bool t3;
 uint8_t t9[16];
 uint8_t t10[16];
 bool t12[16];
 nav__6300_45 t13;
 nav__6300_47 t14;
 uint8_t t15;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = math_isPowerOfTwo__anon_6958__6301();
 if (t3) {
  debug_assert__180(true);
  t4 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t4];
  t4 = (uintptr_t)t5;
  t6 = t4 & (uintptr_t)4095ul;
  t7 = t6;
  t3 = t7 <= UINT64_C(4080);
  if (t3) {
   t6 = t2;
   t5 = (*t1);
   t5 = (uint8_t const *)(((uintptr_t)t5) + (t6*sizeof(uint8_t)));
   t8 = (uint8_t const (*)[16])t5;
   memcpy(t9, (const char *)t8, sizeof(uint8_t[16]));
   memcpy(&t10, &t9, sizeof(uint8_t[16]));
   for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
    t12[t6] = t10[t6] == t11[t6];
   }
   t3 = false;
   for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
    t3 |= t12[t6];
   }
   if (t3) {
    t4 = t2;
    memcpy(t13.array, t12, sizeof(nav__6300_45));
    t14 = simd_firstTrue__anon_7009__6302(t13);
    t15 = t14.payload;
    t6 = (uintptr_t)t15;
    t6 = t4 + t6;
    return t6;
   }
   goto zig_block_2;

   zig_block_2:;
   t6 = t2;
   t16 = mem_alignForward__anon_7012__6303(t4, (uintptr_t)16ul);
   t4 = t16 - t4;
   t4 = t4 / (uintptr_t)1ul;
   t4 = t6 + t4;
   t2 = t4;
   goto zig_block_1;
  }
  zig_loop_64:
  t4 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t4];
  t4 = (uintptr_t)t5;
  t4 = t4 & (uintptr_t)15ul;
  t7 = t4;
  t3 = t7 != UINT64_C(0);
  if (t3) {
   t4 = t2;
   t15 = a0[t4];
   t3 = t15 == UINT8_C(0);
   if (t3) {
    t4 = t2;
    return t4;
   }
   goto zig_block_5;

   zig_block_5:;
   t4 = t2;
   t4 = t4 + (uintptr_t)1ul;
   t2 = t4;
   goto zig_block_4;
  }
  goto zig_block_3;

  zig_block_4:;
  goto zig_loop_64;

  zig_block_3:;
  goto zig_block_1;

  zig_block_1:;
  t6 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)&t5[t6];
  t6 = (uintptr_t)t5;
  t3 = mem_isAligned__982(t6, (uintptr_t)16ul);
  debug_assert__180(t3);
  zig_loop_99:
  t6 = t2;
  t5 = (*t1);
  t5 = (uint8_t const *)(((uintptr_t)t5) + (t6*sizeof(uint8_t)));
  t8 = (uint8_t const (*)[16])t5;
  t17 = (uint8_t const (*)[16])t8;
  memcpy(t10, (const char *)t17, sizeof(uint8_t[16]));
  for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
   t12[t6] = t10[t6] == t18[t6];
  }
  t3 = false;
  for (t6 = (uintptr_t)0ul; t6 < (uintptr_t)16ul; t6 += (uintptr_t)1ul) {
   t3 |= t12[t6];
  }
  if (t3) {
   t6 = t2;
   memcpy(t13.array, t12, sizeof(nav__6300_45));
   t14 = simd_firstTrue__anon_7009__6302(t13);
   t15 = t14.payload;
   t4 = (uintptr_t)t15;
   t4 = t6 + t4;
   return t4;
  }
  goto zig_block_6;

  zig_block_6:;
  t6 = t2;
  t6 = t6 + (uintptr_t)16ul;
  t2 = t6;
  goto zig_loop_99;
 }
 goto zig_block_0;

 zig_block_0:;
 zig_loop_130:
 t16 = t2;
 t15 = a0[t16];
 t3 = t15 != UINT8_C(0);
 if (t3) {
  t16 = t2;
  t16 = t16 + (uintptr_t)1ul;
  t2 = t16;
  goto zig_block_8;
 }
 goto zig_block_7;

 zig_block_8:;
 goto zig_loop_130;

 zig_block_7:;
 t16 = t2;
 return t16;
}

static bool math_isPowerOfTwo__anon_6958__6301(void) {
 debug_assert__180(true);
 return true;
}

static nav__6302_38 simd_firstTrue__anon_7009__6302(nav__6302_40 const a0) {
 static uint8_t const t2[16] = {UINT8_C(0),UINT8_C(1),UINT8_C(2),UINT8_C(3),UINT8_C(4),UINT8_C(5),UINT8_C(6),UINT8_C(7),UINT8_C(8),UINT8_C(9),UINT8_C(10),UINT8_C(11),UINT8_C(12),UINT8_C(13),UINT8_C(14),UINT8_C(15)};
 static uint8_t const t3[16] = {UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15),UINT8_C(15)};
 uintptr_t t1;
 bool t0;
 uint8_t t4[16];
 uint8_t t5;
 nav__6302_38 t6;
 t0 = false;
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)16ul; t1 += (uintptr_t)1ul) {
  t0 |= a0.array[t1];
 }
 t0 = !t0;
 if (t0) {
  return (nav__6302_38){true,UINT8_C(0xa)};
 }
 goto zig_block_0;

 zig_block_0:;
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)16ul; t1 += (uintptr_t)1ul) {
  t4[t1] = a0.array[t1] ? t2[t1] : t3[t1];
 }
 t5 = UINT8_C(15);
 for (t1 = (uintptr_t)0ul; t1 < (uintptr_t)16ul; t1 += (uintptr_t)1ul) {
  t5 = t5 < t4[t1] ? t5 : t4[t1];
 }
 t6.is_null = false;
 t6.payload = t5;
 return t6;
}

static uintptr_t mem_alignForward__anon_7012__6303(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_7054__6304(a1);
 debug_assert__180(t0);
 t1 = a1 - (uintptr_t)1ul;
 t1 = a0 + t1;
 t1 = mem_alignBackward__anon_7055__6305(t1, a1);
 return t1;
}

static bool mem_isAligned__982(uintptr_t const a0, uintptr_t const a1) {
 uint64_t t0;
 uint64_t t1;
 bool t2;
 t0 = a0;
 t1 = a1;
 t2 = mem_isAlignedGeneric__anon_7059__6306(t0, t1);
 return t2;
}

static bool mem_isValidAlignGeneric__anon_7054__6304(uintptr_t const a0) {
 uint64_t t0;
 bool t1;
 bool t2;
 t0 = a0;
 t1 = t0 > UINT64_C(0);
 if (t1) {
  t1 = math_isPowerOfTwo__anon_3550__5613(a0);
  t2 = t1;
  goto zig_block_0;
 }
 t2 = false;
 goto zig_block_0;

 zig_block_0:;
 return t2;
}

static uintptr_t mem_alignBackward__anon_7055__6305(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_7054__6304(a1);
 debug_assert__180(t0);
 t1 = a1 - (uintptr_t)1ul;
 t1 = zig_not_u64(t1, UINT8_C(64));
 t1 = a0 & t1;
 return t1;
}

static bool mem_isAlignedGeneric__anon_7059__6306(uint64_t const a0, uint64_t const a1) {
 uint64_t t0;
 bool t1;
 t0 = mem_alignBackward__anon_7061__6307(a0, a1);
 t1 = t0 == a0;
 return t1;
}

static uint64_t mem_alignBackward__anon_7061__6307(uint64_t const a0, uint64_t const a1) {
 uint64_t t1;
 bool t0;
 t0 = mem_isValidAlignGeneric__anon_7063__6308(a1);
 debug_assert__180(t0);
 t1 = a1 - UINT64_C(1);
 t1 = zig_not_u64(t1, UINT8_C(64));
 t1 = a0 & t1;
 return t1;
}

static bool mem_isValidAlignGeneric__anon_7063__6308(uint64_t const a0) {
 bool t0;
 bool t1;
 t0 = a0 > UINT64_C(0);
 if (t0) {
  t0 = math_isPowerOfTwo__anon_7064__6309(a0);
  t1 = t0;
  goto zig_block_0;
 }
 t1 = false;
 goto zig_block_0;

 zig_block_0:;
 return t1;
}

static bool math_isPowerOfTwo__anon_7064__6309(uint64_t const a0) {
 uint64_t t1;
 bool t0;
 t0 = a0 > UINT64_C(0);
 debug_assert__180(t0);
 t1 = a0 - UINT64_C(1);
 t1 = a0 & t1;
 t0 = t1 == UINT64_C(0);
 return t0;
}

uint8_t c_instrument_hooks_executed_benchmark__244(struct InstrumentHooks *const a0, int32_t const a1, char const *const a2) {
 uint8_t t0;
 t0 = c_instrument_hooks_set_executed_benchmark__243(a0, a1, a2);
 return t0;
}

static uint16_t runner_fifo_RunnerFifo_set_integration__1137(struct runner_fifo_RunnerFifo__631 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 nav__1137_61 t4;
 nav__1137_61 t5;
 nav__1137_61 t6;
 struct shared_Command__struct_1799__1799 t7;
 struct shared_Command__1795 t8;
 struct fifo_posix_Pipe_Reader__793 *t10;
 uint16_t t9;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4 = mem_span__anon_6923__6298(a1);
 t5 = mem_span__anon_6923__6298(a2);
 memcpy(&t6, &t4, sizeof(nav__1137_61));
 memcpy(&t4, &t5, sizeof(nav__1137_61));
 t7.name = t6;
 t7.version = t4;
 t8.tag = UINT8_C(5);
 t8.payload.SetIntegration = t7;
 t9 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t8);
 if (t9) {
  return t9;
 }
 t2 = (*t1);
 t10 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t9 = fifo_posix_Pipe_Reader_waitForAck__1293(t10, (nav__1137_76){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t9) {
  return t9;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_integration__813(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_integration__1137(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__826(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0, uint8_t const *const a1, uint8_t const *const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_set_integration__1137(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_integration__245(struct InstrumentHooks *const a0, char const *const a1, char const *const a2) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct instruments_root_Instrument__554 *t4;
 struct instruments_root_Instrument__554 *t8;
 uint8_t const *t5;
 uint8_t const *t6;
 struct instruments_root_Instrument__554 *const *t9;
 struct instruments_root_Instrument__554 t10;
 struct instruments_valgrind_ValgrindInstrument__564 *t12;
 struct instruments_valgrind_ValgrindInstrument__564 t13;
 struct instruments_valgrind_ValgrindInstrument__564 t15;
 struct instruments_valgrind_ValgrindInstrument__564 const *t16;
 struct mem_Allocator__599 t17;
 nav__245_103 t18;
 nav__245_105 t19;
 nav__245_56 t21;
 uint8_t *t22;
 struct mem_Allocator__599 const *t23;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t24;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t25;
 uint16_t t7;
 uint16_t t14;
 uint16_t t20;
 bool t0;
 uint8_t t11;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t5 = c_toU8__236(a1);
  t6 = c_toU8__236(a2);
  t8 = t4;
  t9 = (struct instruments_root_Instrument__554 *const *)&t8;
  t10 = (*t4);
  t11 = t10.tag;
  switch (t11) {
   case UINT8_C(0): {
    t4 = (*t9);
    t12 = (struct instruments_valgrind_ValgrindInstrument__564 *)&t4->payload.valgrind;
    t13 = (*t12);
    t15 = t13;
    t16 = (struct instruments_valgrind_ValgrindInstrument__564 const *)&t15;
    t17 = t13.allocator;
    t18.f0 = t5;
    t18.f1 = t6;
    t19 = fmt_allocPrintZ__anon_7099__6310(t17, t18);
    if (t19.error) {
     t20 = t19.error;
     t14 = t20;
     goto zig_block_4;
    }
    t21 = t19.payload;
    t22 = t21.ptr;
    t6 = (uint8_t const *)t22;
    callgrind_dump_stats_at(t6);
    t23 = (struct mem_Allocator__599 const *)&t16->allocator;
    t17 = (*t23);
    mem_Allocator_free__anon_7101__6311(t17, t21);
    t14 = 0;
    goto zig_block_4;

    zig_block_4:;
    memcpy(&t20, &t14, sizeof(uint16_t));
    if (t20) {
     t7 = t20;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(1): {
    t4 = (*t9);
    t24 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t4->payload.walltime;
    t20 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_set_integration__813(t24, t5, t6);
    if (t20) {
     t7 = t20;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t4 = (*t9);
    t25 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t4->payload.analysis;
    t20 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_set_integration__826(t25, t5, t6);
    if (t20) {
     t7 = t20;
     goto zig_block_2;
    }
    goto zig_block_3;
   }
   case UINT8_C(3): {
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t7 = 0;
  goto zig_block_2;

  zig_block_2:;
  memcpy(&t20, &t7, sizeof(uint16_t));
  t0 = t20 == UINT16_C(0);
  if (t0) {
   goto zig_block_1;
  }
  return UINT8_C(1);

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static nav__6310_40 fmt_allocPrintZ__anon_7099__6310(struct mem_Allocator__599 const a0, nav__6310_43 const a1) {
 nav__6310_40 t0;
 nav__6310_39 t2;
 nav__6310_39 t3;
 nav__6310_39 const *t4;
 uintptr_t t5;
 uint8_t *t6;
 uint16_t t1;
 t0 = fmt_allocPrint__anon_7136__6312(a0, a1);
 if (t0.error) {
  t1 = t0.error;
  t0.payload = (nav__6310_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t0.error = t1;
  return t0;
 }
 t2 = t0.payload;
 t3 = t2;
 t4 = (nav__6310_39 const *)&t3;
 t5 = t2.len;
 t5 = t5 - (uintptr_t)1ul;
 t2 = (*t4);
 t6 = t2.ptr;
 t6 = (uint8_t *)(((uintptr_t)t6) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t6;
 t2.len = t5;
 t0.payload = t2;
 t0.error = UINT16_C(0);
 return t0;
}

static void mem_Allocator_free__anon_7101__6311(struct mem_Allocator__599 const a0, nav__6311_40 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6311_40 t2;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t12;
 struct mem_Allocator_VTable__602 const *const *t13;
 struct mem_Allocator_VTable__602 const *t14;
 void (*const *t15)(void *, nav__6311_40, uint8_t, uintptr_t);
 void (*t16)(void *, nav__6311_40, uint8_t, uintptr_t);
 void *t17;
 bool t5;
 uint8_t t11;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_7143__6313(a1);
 t3 = t2.len;
 t3 = t3 + (uintptr_t)1ul;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t11 = mem_Alignment_fromByteUnits__1113((uintptr_t)1ul);
 t3 = (uintptr_t)zig_return_address();
 t12 = t10;
 t1 = (struct mem_Allocator__599 const *)&t12;
 t13 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t14 = (*t13);
 t15 = (void (*const *)(void *, nav__6311_40, uint8_t, uintptr_t))&t14->free;
 t16 = (*t15);
 t17 = t10.ptr;
 t16(t17, t2, t11, t3);
 return;
}

static nav__6312_40 fmt_allocPrint__anon_7136__6312(struct mem_Allocator__599 const a0, nav__6312_43 const a1) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t6;
 uint64_t t3;
 nav__6312_54 t4;
 struct mem_Allocator__599 t7;
 struct mem_Allocator__599 t0;
 nav__6312_40 t8;
 nav__6312_40 t11;
 nav__6312_39 t10;
 uint16_t t9;
 bool t5;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = fmt_count__anon_7150__6314(a1);
 t4 = math_cast__anon_7152__6315(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t6 = t4.payload;
  t2 = t6;
  goto zig_block_0;
 }
 return (nav__6312_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_OutOfMemory};

 zig_block_0:;
 t7 = (*t1);
 t8 = mem_Allocator_alloc__anon_2384__5077(t7, t2);
 if (t8.error) {
  t9 = t8.error;
  t8.payload = (nav__6312_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t9;
  return t8;
 }
 t10 = t8.payload;
 t8 = fmt_bufPrint__anon_7155__6316(t10, a1);
 t5 = t8.error == UINT16_C(0);
 if (t5) {
  t10 = t8.payload;
  t8.payload = t10;
  t8.error = UINT16_C(0);
  t11 = t8;
  goto zig_block_1;
 }
 t9 = t8.error;
 switch (t9) {
  case zig_error_NoSpaceLeft: {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_1:;
 return t11;
}

static nav__6313_39 mem_sliceAsBytes__anon_7143__6313(nav__6313_39 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3;
 uint8_t *t4;
 uint8_t *t5;
 uint8_t *const *t6;
 nav__6313_39 t7;
 bool t2;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = a0.ptr;
 t4 = (uint8_t *)t3;
 t5 = t4;
 t6 = (uint8_t *const *)&t5;
 t0 = a0.len;
 t4 = (*t6);
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t7.ptr = t4;
 t7.len = t0;
 return t7;
}

static uint16_t fmt_format__anon_7206__6355(struct io_Writer__3878 const a0, nav__6355_40 const a1) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint8_t const *t4;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6355_44){(uint8_t const *)&__anon_7231,(uintptr_t)10ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f0;
 t3 = fmt_formatType__anon_7240__6356(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6355_44){(uint8_t const *)&__anon_7246,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 t4 = a1.f1;
 t3 = fmt_formatType__anon_7240__6356(t4, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__6355_44){(uint8_t const *)&__anon_7259,(uintptr_t)1ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static uint64_t fmt_count__anon_7150__6314(nav__6314_39 const a0) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 t1;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 t0;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 t2;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 t3;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 const *t4;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 const *t5;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 const *const *t6;
 void const **t8;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const *t9;
 void const *t10;
 nav__6314_53 (**t11)(void const *, nav__6314_54);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 uint64_t t15;
 uint16_t t13;
 bool t14;
 t1 = io_counting_writer_countingWriter__anon_7176__6341();
 t0 = t1;
 t2 = io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__6340(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6314_53 (**)(void const *, nav__6314_54))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__6354;
 t12 = t7;
 t13 = fmt_format__anon_7206__6355(t12, a0);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 zig_unreachable();

 zig_block_0:;
 t1 = t0;
 t15 = t1.bytes_written;
 return t15;
}

static nav__6315_38 math_cast__anon_7152__6315(uint64_t const a0) {
 uintptr_t t0;
 nav__6315_38 t1;
 t0 = a0;
 t1.is_null = false;
 t1.payload = t0;
 return t1;
}

static nav__6316_40 fmt_bufPrint__anon_7155__6316(nav__6316_39 const a0, nav__6316_42 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__6316_57 (**t11)(void const *, nav__6316_58);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__6316_39 t15;
 nav__6316_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6316_57 (**)(void const *, nav__6316_58))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_7206__6355(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6316_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 io_counting_writer_countingWriter__anon_7176__6341(void) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 t0;
 t0 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173){UINT64_C(0)};
 return t0;
}

static struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_writer__6340(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const a0) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 **t1;
 struct io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_2cerror_7b_7d_2c_28function_20_27write_27_29_29__7187 t0;
 t1 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 **)&t0.context;
 (*t1) = a0;
 return t0;
}

static nav__6354_38 io_GenericWriter_28_2aio_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWr__6354(void const *const a0, nav__6354_41 const a1) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const *t0;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *t1;
 nav__6354_38 t2;
 nav__6354_38 t3;
 t0 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const *)a0;
 t1 = (*t0);
 t2 = io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__6339(t1, a1);
 memcpy(&t3, &t2, sizeof(nav__6354_38));
 return t3;
}

static uint16_t fmt_formatType__anon_7240__6356(uint8_t const *const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 nav__6356_47 t2;
 nav__6356_47 t3;
 uint8_t const *t0;
 struct io_Writer__3878 t1;
 uint16_t t4;
 uint16_t t5;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = mem_span__anon_6923__6298(a0);
 memcpy(&t3, &t2, sizeof(nav__6356_47));
 t4 = fmt_formatBuf__anon_5176__6129(t3, a1, a2);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static nav__6339_38 io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29_write__6339(struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const a0, nav__6339_42 const a1) {
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const *t1;
 nav__6339_38 t2;
 uintptr_t t3;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *t4;
 struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *t0;
 uint64_t *t5;
 uint64_t t6;
 uint64_t t7;
 t0 = a0;
 t1 = (struct io_counting_writer_CountingWriter_28io_GenericWriter_28void_2cerror_7b_7d_2c_28function_20_27dummyWrite_27_29_29_29__7173 *const *)&t0;
 t2 = io_dummyWrite__4890(a1);
 t3 = t2.payload;
 t4 = (*t1);
 t5 = (uint64_t *)&t4->bytes_written;
 t6 = (*t5);
 t7 = t3;
 t7 = t6 + t7;
 (*t5) = t7;
 t2.payload = t3;
 t2.error = UINT16_C(0);
 return t2;
}

static nav__4890_38 io_dummyWrite__4890(nav__4890_40 const a0) {
 uintptr_t t0;
 nav__4890_38 t1;
 t0 = a0.len;
 t1.payload = t0;
 t1.error = UINT16_C(0);
 return t1;
}

static uint16_t runner_fifo_RunnerFifo_add_marker__1138(struct runner_fifo_RunnerFifo__631 *const a0, int32_t const a1, struct shared_MarkerType__1802 const a2) {
 struct runner_fifo_RunnerFifo__631 *const *t1;
 struct runner_fifo_RunnerFifo__631 *t2;
 struct runner_fifo_RunnerFifo__631 *t0;
 struct fifo_posix_Pipe_Writer__791 *t3;
 struct shared_Command__struct_1800__1800 t4;
 struct shared_Command__1795 t5;
 struct fifo_posix_Pipe_Reader__793 *t7;
 uint16_t t6;
 t0 = a0;
 t1 = (struct runner_fifo_RunnerFifo__631 *const *)&t0;
 t2 = (*t1);
 t3 = (struct fifo_posix_Pipe_Writer__791 *)&t2->writer;
 t4.marker = a2;
 t4.pid = a1;
 t5.tag = UINT8_C(7);
 t5.payload.AddMarker = t4;
 t6 = fifo_posix_Pipe_Writer_sendCmd__1286(t3, t5);
 if (t6) {
  return t6;
 }
 t2 = (*t1);
 t7 = (struct fifo_posix_Pipe_Reader__793 *)&t2->reader;
 t6 = fifo_posix_Pipe_Reader_waitForAck__1293(t7, (nav__1138_77){UINT64_C(0xaaaaaaaaaaaaaaaa),true});
 if (t6) {
  return t6;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_add_marker__814(struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const a0, int32_t const a1, struct shared_MarkerType__1802 const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_add_marker__1138(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

static uint16_t instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__827(struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const a0, int32_t const a1, struct shared_MarkerType__1802 const a2) {
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *t1;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t2;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t0;
 struct runner_fifo_RunnerFifo__631 *t3;
 uint16_t t4;
 t0 = a0;
 t1 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *const *)&t0;
 t2 = (*t1);
 t3 = (struct runner_fifo_RunnerFifo__631 *)&t2->fifo;
 t4 = runner_fifo_RunnerFifo_add_marker__1138(t3, a1, a2);
 if (t4) {
  return t4;
 }
 return 0;
}

uint8_t c_instrument_hooks_add_marker__250(struct InstrumentHooks *const a0, int32_t const a1, uint8_t const a2, uint64_t const a3) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct shared_MarkerType__1802 t4;
 struct shared_MarkerType__1802 t5;
 struct instruments_root_Instrument__554 *t6;
 struct instruments_root_Instrument__554 *t8;
 struct instruments_root_Instrument__554 *const *t9;
 struct instruments_root_Instrument__554 t10;
 struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *t12;
 struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *t15;
 uint16_t t7;
 uint16_t t13;
 uint16_t t14;
 bool t0;
 uint8_t t11;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  switch (a2) {
   case UINT8_C(0): {
    t5.tag = UINT8_C(0);
    t5.payload.SampleStart = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t5.tag = UINT8_C(1);
    t5.payload.SampleEnd = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t5.tag = UINT8_C(2);
    t5.payload.BenchmarkStart = a3;
    t4 = t5;
    goto zig_block_1;
   }
   case UINT8_C(3): {
    t5.tag = UINT8_C(3);
    t5.payload.BenchmarkEnd = a3;
    t4 = t5;
    goto zig_block_1;
   }
   default: {
    return UINT8_C(2);
   }
  }

  zig_block_1:;
  t1 = (*t3);
  t6 = (struct instruments_root_Instrument__554 *)&t1->instrument;
  t8 = t6;
  t9 = (struct instruments_root_Instrument__554 *const *)&t8;
  t10 = (*t6);
  t11 = t10.tag;
  t0 = t11 == UINT8_C(1);
  if (t0) {
   t6 = (*t9);
   t12 = (struct instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29__582 *)&t6->payload.walltime;
   t13 = instruments_fifo_instrument_FifoInstrument_28_Walltime_2cerror_7bModeError_7d_29_add_marker__814(t12, a1, t4);
   memcpy(&t14, &t13, sizeof(uint16_t));
   t7 = t14;
   goto zig_block_3;
  }
  t10 = (*t6);
  t11 = t10.tag;
  t0 = t11 == UINT8_C(2);
  if (t0) {
   t6 = (*t9);
   t15 = (struct instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29__591 *)&t6->payload.analysis;
   t14 = instruments_fifo_instrument_FifoInstrument_28_Analysis_2cerror_7bModeError_7d_29_add_marker__827(t15, a1, t4);
   memcpy(&t13, &t14, sizeof(uint16_t));
   t7 = t13;
   goto zig_block_3;
  }
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_4;

  zig_block_4:;
  t7 = 0;
  goto zig_block_3;

  zig_block_3:;
  memcpy(&t14, &t7, sizeof(uint16_t));
  t0 = t14 == UINT16_C(0);
  if (t0) {
   goto zig_block_2;
  }
  return UINT8_C(1);

  zig_block_2:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

uint64_t c_instrument_hooks_current_timestamp__254(void) {
 struct os_linux_timespec__struct_7321__7321 t0;
 struct os_linux_timespec__struct_7321__7321 t3;
 nav__254_41 t1;
 intptr_t t4;
 uint64_t t5;
 uint64_t t6;
 bool t2;
 t1 = posix_clock_gettime__1938(UINT32_C(1));
 t2 = t1.error == UINT16_C(0);
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 zig_unreachable();

 zig_block_0:;
 t4 = t0.sec;
 t5 = (uint64_t)t4;
 t5 = t5 * UINT64_C(1000000000);
 t4 = t0.nsec;
 t6 = (uint64_t)t4;
 t6 = t5 + t6;
 return t6;
}

static nav__1938_39 posix_clock_gettime__1938(uint32_t const a0) {
 struct os_linux_timespec__struct_7321__7321 t3;
 struct os_linux_timespec__struct_7321__7321 t0;
 nav__1938_39 t4;
 int t1;
 uint16_t t2;
 t1 = clock_gettime(a0, &t0);
 t2 = posix_errno__anon_3044__5205(t1);
 switch (t2) {
  case UINT16_C(0): {
   t3 = t0;
   t4.payload = t3;
   t4.error = UINT16_C(0);
   return t4;
  }
  case UINT16_C(14): {
   zig_unreachable();
  }
  case UINT16_C(22): {
   return (nav__1938_39){{-(intptr_t)0x5555555555555556,-(intptr_t)0x5555555555555556},zig_error_UnsupportedClock};
  }
  default: {
   t2 = posix_unexpectedErrno__2021(t2);
   t4.payload = (struct os_linux_timespec__struct_7321__7321){-(intptr_t)0x5555555555555556,-(intptr_t)0x5555555555555556};
   t4.error = t2;
   return t4;
  }
 }
}

static uint16_t environment_root_Environment_setIntegrationEnvironmentValue__778(struct environment_root_Environment__559 *const a0, nav__778_41 const a1, nav__778_41 const a2, struct environment_root_EnvironmentValue__856 const a3) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t0;
 struct environment_root_EnvironmentJson__845 *t3;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t5;
 struct mem_Allocator__599 *t6;
 struct mem_Allocator__599 t7;
 nav__778_93 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t11;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 const *t12;
 nav__778_41 *t14;
 nav__778_98 t15;
 nav__778_96 t16;
 nav__778_41 t17;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t18;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t19;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t20;
 nav__778_110 t21;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t22;
 struct environment_root_EnvironmentValue__856 *t23;
 struct environment_root_EnvironmentValue__856 t24;
 uint16_t t9;
 bool t13;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (struct environment_root_EnvironmentJson__845 *)&t2->data;
 t4 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *)&t3->integration_environment;
 t5 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *)&t4->map;
 t6 = (struct mem_Allocator__599 *)&a0->allocator;
 t7 = (*t6);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3802(t5, t7, a1);
 if (t8.error) {
  t9 = t8.error;
  return t9;
 }
 t10 = t8.payload;
 t11 = t10;
 t12 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 const *)&t11;
 t13 = t10.found_existing;
 t13 = !t13;
 if (t13) {
  t14 = t10.key_ptr;
  t2 = (*t1);
  t6 = (struct mem_Allocator__599 *)&t2->allocator;
  t7 = (*t6);
  t15 = mem_Allocator_dupe__anon_7358__6358(t7, a1);
  if (t15.error) {
   t9 = t15.error;
   return t9;
  }
  t16 = t15.payload;
  memcpy(&t17, &t16, sizeof(nav__778_41));
  (*t14) = t17;
  t18 = t10.value_ptr;
  (*t18) = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858){{{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul,(uintptr_t)0ul},NULL}};
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t19 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *)&t12->value_ptr;
 t18 = (*t19);
 t20 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *)&t18->map;
 t6 = (struct mem_Allocator__599 *)&a0->allocator;
 t7 = (*t6);
 t21 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4013(t20, t7, a2);
 if (t21.error) {
  t9 = t21.error;
  return t9;
 }
 t22 = t21.payload;
 t13 = t22.found_existing;
 if (t13) {
  t2 = (*t1);
  t23 = t22.value_ptr;
  t24 = (*t23);
  environment_root_Environment_freeEnvironmentValue__777(t2, t24);
  goto zig_block_1;
 }
 t14 = t22.key_ptr;
 t2 = (*t1);
 t6 = (struct mem_Allocator__599 *)&t2->allocator;
 t7 = (*t6);
 t15 = mem_Allocator_dupe__anon_7358__6358(t7, a2);
 if (t15.error) {
  t9 = t15.error;
  return t9;
 }
 t16 = t15.payload;
 memcpy(&t17, &t16, sizeof(nav__778_41));
 (*t14) = t17;
 goto zig_block_1;

 zig_block_1:;
 t23 = t22.value_ptr;
 (*t23) = a3;
 return 0;
}

static uint16_t environment_root_Environment_setIntegrationEnvironment__775(struct environment_root_Environment__559 *const a0, nav__775_41 const a1, nav__775_41 const a2, nav__775_41 const a3) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t3;
 struct environment_root_Environment__559 *t0;
 struct mem_Allocator__599 *t4;
 struct mem_Allocator__599 t5;
 nav__775_81 t6;
 nav__775_79 t8;
 nav__775_41 t9;
 struct environment_root_EnvironmentValue__856 t10;
 uint16_t t7;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (*t1);
 t4 = (struct mem_Allocator__599 *)&t3->allocator;
 t5 = (*t4);
 t6 = mem_Allocator_dupe__anon_7358__6358(t5, a3);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t8 = t6.payload;
 memcpy(&t9, &t8, sizeof(nav__775_41));
 t10.tag = UINT8_C(0);
 t10.payload.string = t9;
 t7 = environment_root_Environment_setIntegrationEnvironmentValue__778(t2, a1, a2, t10);
 if (t7) {
  return t7;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_environment__255(struct InstrumentHooks *const a0, char const *const a1, char const *const a2, char const *const a3) {
 struct InstrumentHooks *t2;
 struct InstrumentHooks *t3;
 struct InstrumentHooks *const *t4;
 struct environment_root_Environment__559 *t5;
 uint8_t const *t6;
 nav__255_100 t7;
 nav__255_100 t8;
 nav__255_100 t9;
 nav__255_100 t10;
 uint16_t t11;
 bool t0;
 bool t1;
 t0 = a1 == NULL;
 if (t0) {
  t1 = true;
  goto zig_block_1;
 }
 t0 = a2 == NULL;
 t1 = t0;
 goto zig_block_1;

 zig_block_1:;
 if (t1) {
  t0 = true;
  goto zig_block_2;
 }
 t1 = a3 == NULL;
 t0 = t1;
 goto zig_block_2;

 zig_block_2:;
 if (t0) {
  return UINT8_C(1);
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0 != NULL;
 if (t1) {
  t2 = a0;
  t3 = t2;
  t4 = (struct InstrumentHooks *const *)&t3;
  t2 = (*t4);
  t5 = (struct environment_root_Environment__559 *)&t2->environment;
  t6 = c_toU8__236(a1);
  t7 = mem_span__anon_6923__6298(t6);
  memcpy(&t8, &t7, sizeof(nav__255_100));
  t6 = c_toU8__236(a2);
  t7 = mem_span__anon_6923__6298(t6);
  memcpy(&t9, &t7, sizeof(nav__255_100));
  t6 = c_toU8__236(a3);
  t7 = mem_span__anon_6923__6298(t6);
  memcpy(&t10, &t7, sizeof(nav__255_100));
  t11 = environment_root_Environment_setIntegrationEnvironment__775(t5, t8, t9, t10);
  t1 = t11 == UINT16_C(0);
  if (t1) {
   goto zig_block_4;
  }
  return UINT8_C(1);

  zig_block_4:;
  return UINT8_C(0);
 }
 goto zig_block_3;

 zig_block_3:;
 return UINT8_C(1);
}

static nav__6358_40 mem_Allocator_dupe__anon_7358__6358(struct mem_Allocator__599 const a0, nav__6358_43 const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 uintptr_t t3;
 nav__6358_40 t4;
 nav__6358_39 t6;
 uint8_t const *t7;
 uint16_t t5;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = a1.len;
 t4 = mem_Allocator_alloc__anon_2384__5077(t2, t3);
 if (t4.error) {
  t5 = t4.error;
  t4.payload = (nav__6358_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t4.error = t5;
  return t4;
 }
 t6 = t4.payload;
 t7 = a1.ptr;
 if (t6.len != 0) memcpy(t6.ptr, t7, t6.len * sizeof(uint8_t));
 t4.payload = t6;
 t4.error = UINT16_C(0);
 return t4;
}

static nav__3802_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3802(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct mem_Allocator__599 const a1, nav__3802_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 nav__3802_39 t3;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3803(t2, a1, a2);
 return t3;
}

static nav__4013_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4013(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct mem_Allocator__599 const a1, nav__4013_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 nav__4013_39 t3;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4014(t2, a1, a2);
 return t3;
}

static nav__3803_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3803(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct mem_Allocator__599 const a1, nav__3803_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 nav__3803_39 t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t5;
 nav__3803_44 *t7;
 uint16_t t4;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6359(t2, a1, a2);
 if (t3.error) {
  t4 = t3.error;
  t3.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360){((nav__3803_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
  t3.error = t4;
  return t3;
 }
 t5 = t3.payload;
 t6 = t5.found_existing;
 t6 = !t6;
 if (t6) {
  t7 = t5.key_ptr;
  (*t7) = a2;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3.payload = t5;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__4014_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4014(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct mem_Allocator__599 const a1, nav__4014_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 nav__4014_39 t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t5;
 nav__4014_44 *t7;
 uint16_t t4;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6360(t2, a1, a2);
 if (t3.error) {
  t4 = t3.error;
  t3.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371){((nav__4014_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_root_EnvironmentValue__856 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
  t3.error = t4;
  return t3;
 }
 t5 = t3.payload;
 t6 = t5.found_existing;
 t6 = !t6;
 if (t6) {
  t7 = t5.key_ptr;
  (*t7) = a2;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3.payload = t5;
 t3.error = UINT16_C(0);
 return t3;
}

static nav__6359_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6359(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct mem_Allocator__599 const a1, nav__6359_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t4;
 uintptr_t t5;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t8;
 nav__6359_67 t9;
 nav__6359_39 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t14;
 nav__6359_73 t15;
 nav__6359_73 t16;
 nav__6359_73 const *t17;
 nav__6359_44 *t18;
 nav__6359_76 t19;
 nav__6359_76 t20;
 nav__6359_76 const *t21;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t22;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t23;
 uint16_t t6;
 bool t7;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&a0->entries;
 t4 = (*t3);
 t5 = t4.len;
 t5 = t5 + (uintptr_t)1ul;
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3812(t2, a1, t5);
 t7 = t6 == UINT16_C(0);
 if (t7) {
  goto zig_block_0;
 }
 t2 = (*t1);
 t8 = (*t2);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6361(t8, a2);
 t7 = t9.is_null != true;
 if (t7) {
  t10 = t9.payload;
  t5 = t10;
  goto zig_block_1;
 }
 t11.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360){((nav__6359_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
 t11.error = t6;
 return t11;

 zig_block_1:;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t2->entries;
 t4 = (*t3);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t4);
 t13 = t12;
 t14 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t13;
 t12 = (*t14);
 t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t12);
 t16 = t15;
 t17 = (nav__6359_73 const *)&t16;
 t15 = (*t17);
 t18 = &t15.ptr[t5];
 t12 = (*t14);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t12);
 t20 = t19;
 t21 = (nav__6359_76 const *)&t20;
 t19 = (*t21);
 t22 = &t19.ptr[t5];
 t23.key_ptr = t18;
 t23.value_ptr = t22;
 t23.index = t5;
 t23.found_existing = true;
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;

 zig_block_0:;
 t2 = (*t1);
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6362(t2, a2);
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;
}

static nav__6360_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6360(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct mem_Allocator__599 const a1, nav__6360_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t4;
 uintptr_t t5;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t8;
 nav__6360_67 t9;
 nav__6360_39 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t14;
 nav__6360_73 t15;
 nav__6360_73 t16;
 nav__6360_73 const *t17;
 nav__6360_44 *t18;
 nav__6360_76 t19;
 nav__6360_76 t20;
 nav__6360_76 const *t21;
 struct environment_root_EnvironmentValue__856 *t22;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t23;
 uint16_t t6;
 bool t7;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&a0->entries;
 t4 = (*t3);
 t5 = t4.len;
 t5 = t5 + (uintptr_t)1ul;
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4023(t2, a1, t5);
 t7 = t6 == UINT16_C(0);
 if (t7) {
  goto zig_block_0;
 }
 t2 = (*t1);
 t8 = (*t2);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6363(t8, a2);
 t7 = t9.is_null != true;
 if (t7) {
  t10 = t9.payload;
  t5 = t10;
  goto zig_block_1;
 }
 t11.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371){((nav__6360_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_root_EnvironmentValue__856 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
 t11.error = t6;
 return t11;

 zig_block_1:;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t2->entries;
 t4 = (*t3);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t4);
 t13 = t12;
 t14 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t13;
 t12 = (*t14);
 t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t12);
 t16 = t15;
 t17 = (nav__6360_73 const *)&t16;
 t15 = (*t17);
 t18 = &t15.ptr[t5];
 t12 = (*t14);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t12);
 t20 = t19;
 t21 = (nav__6360_76 const *)&t20;
 t19 = (*t21);
 t22 = &t19.ptr[t5];
 t23.key_ptr = t18;
 t23.value_ptr = t22;
 t23.index = t5;
 t23.found_existing = true;
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;

 zig_block_0:;
 t2 = (*t1);
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6364(t2, a2);
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;
}

static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3812(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 void *t3;
 uint64_t t4;
 uint64_t t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t6;
 struct array_hash_map_IndexHeader__892 **t8;
 struct array_hash_map_IndexHeader__892 *t9;
 struct array_hash_map_IndexHeader__892 *t10;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t20;
 struct array_hash_map_IndexHeader__892 *const *t12;
 nav__3812_62 t18;
 struct array_hash_map_IndexHeader__892 t13;
 uint32_t t14;
 uint16_t t7;
 nav__3812_60 t16;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3982(t3);
 t4 = a2;
 t5 = t4 <= UINT64_C(8);
 if (t5) {
  t2 = (*t1);
  t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t2->entries;
  t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3955(t6, a1, a2);
  if (t7) {
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3983(t3);
   return t7;
  }
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t10 = t9;
  t11 = t10;
  t12 = (struct array_hash_map_IndexHeader__892 *const *)&t11;
  t10 = (*t12);
  t13 = (*t10);
  t14 = array_hash_map_IndexHeader_capacity__3974(t13);
  t4 = a2;
  t15 = (uint64_t)t14;
  t5 = t4 <= t15;
  if (t5) {
   t2 = (*t1);
   t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t2->entries;
   t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3955(t6, a1, a2);
   if (t7) {
    t2 = (*t1);
    t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
    debug_SafetyLock_unlock__3983(t3);
    return t7;
   }
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3983(t3);
   return 0;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t2->entries;
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3955(t6, a1, a2);
 if (t7) {
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t16 = array_hash_map_IndexHeader_findBitIndex__3977(a2);
 if (t16.error) {
  t7 = t16.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t17 = t16.payload;
 t18 = array_hash_map_IndexHeader_alloc__3978(a1, t17);
 if (t18.error) {
  t7 = t18.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t10 = t18.payload;
 t8 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t19 = t9;
  t20 = t19;
  t12 = (struct array_hash_map_IndexHeader__892 *const *)&t20;
  t19 = (*t12);
  array_hash_map_IndexHeader_free__3979(t19, a1);
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t2 = (*t1);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3904(t2, t10);
 t2 = (*t1);
 t8 = (struct array_hash_map_IndexHeader__892 **)&t2->index_header;
 t19 = (struct array_hash_map_IndexHeader__892 *)t10;
 (*t8) = t19;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3983(t3);
 return 0;
}

static nav__6361_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6361(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6361_41 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_IndexHeader__892 *t3;
 struct array_hash_map_IndexHeader__892 *t5;
 struct array_hash_map_IndexHeader__892 *t24;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t11;
 nav__6361_61 t12;
 nav__6361_64 t13;
 uintptr_t t15;
 uintptr_t t16;
 uintptr_t t14;
 uint64_t t17;
 uint64_t t18;
 nav__6361_41 *t19;
 nav__6361_41 t22;
 nav__6361_38 t23;
 struct array_hash_map_IndexHeader__892 *const *t25;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 uint32_t t6;
 uint32_t t20;
 struct array_hash_map_IndexHeader__892 t26;
 bool t4;
 bool t21;
 uint8_t t27;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t3 = a0.index_header;
 t4 = t3 != NULL;
 if (t4) {
  t5 = t3;
  t2 = t5;
  goto zig_block_0;
 }
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t7 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *)&t1->entries;
 t8 = (*t7);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t8);
 t10 = t9;
 t11 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t10;
 t9 = (*t11);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t9);
 t9 = (*t11);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t9);
 t14 = (uintptr_t)0ul;
 t15 = t13.len;
 zig_loop_32:
 t16 = t14;
 t17 = t16;
 t18 = t15;
 t4 = t17 < t18;
 if (t4) {
  t19 = &t13.ptr[t16];
  t20 = t12.ptr[t16];
  t4 = t20 == t6;
  if (t4) {
   t22 = (*t19);
   t4 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t22, t16);
   t21 = t4;
   goto zig_block_4;
  }
  t21 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t21) {
   t23.is_null = false;
   t23.payload = t16;
   return t23;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t16 = t16 + (uintptr_t)1ul;
 t14 = t16;
 goto zig_loop_32;

 zig_block_1:;
 return (nav__6361_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t24 = t2;
 t25 = (struct array_hash_map_IndexHeader__892 *const *)&t24;
 t5 = (*t25);
 t26 = (*t5);
 t27 = array_hash_map_IndexHeader_capacityIndexType__3973(t26);
 switch (t27) {
  case UINT8_C(0): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6368(t28, a1, t2);
   return t23;
  }
  case UINT8_C(1): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6369(t28, a1, t2);
   return t23;
  }
  case UINT8_C(2): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6370(t28, a1, t2);
   return t23;
  }
  default: zig_unreachable();
 }
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6362(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, nav__6362_42 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t6;
 struct array_hash_map_IndexHeader__892 *t39;
 struct array_hash_map_IndexHeader__892 **t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t13;
 nav__6362_66 t14;
 nav__6362_66 t15;
 nav__6362_66 const *t16;
 nav__6362_69 t17;
 nav__6362_69 t18;
 nav__6362_69 const *t19;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t20;
 uint64_t t23;
 uint64_t t24;
 nav__6362_42 *t25;
 nav__6362_42 t28;
 nav__6362_72 t29;
 nav__6362_72 t30;
 nav__6362_72 t37;
 nav__6362_72 const *t31;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t33;
 uint32_t *const *t34;
 uint32_t *t35;
 nav__6362_42 *const *t36;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t38;
 struct array_hash_map_IndexHeader__892 *const *t40;
 uint32_t t7;
 uint32_t t26;
 struct array_hash_map_IndexHeader__892 t41;
 bool t5;
 bool t27;
 uint8_t t42;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t3 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t4 = (*t3);
 t5 = t4 != NULL;
 if (t5) {
  t6 = t4;
  t2 = t6;
  goto zig_block_0;
 }
 t7 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t8->entries;
 t10 = (*t9);
 t11 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t10);
 t12 = t11;
 t13 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t12;
 t11 = (*t13);
 t14 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t11);
 t15 = t14;
 t16 = (nav__6362_66 const *)&t15;
 t11 = (*t13);
 t17 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t11);
 t18 = t17;
 t19 = (nav__6362_69 const *)&t18;
 t20 = (uintptr_t)0ul;
 t21 = t17.len;
 zig_loop_39:
 t22 = t20;
 t23 = t22;
 t24 = t21;
 t5 = t23 < t24;
 if (t5) {
  t25 = &t17.ptr[t22];
  t26 = t14.ptr[t22];
  t5 = t26 == t7;
  if (t5) {
   t28 = (*t25);
   t5 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t28, t22);
   t27 = t5;
   goto zig_block_4;
  }
  t27 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t27) {
   t11 = (*t13);
   t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t11);
   t30 = t29;
   t31 = (nav__6362_72 const *)&t30;
   t29 = (*t31);
   t32 = &t29.ptr[t22];
   t33.key_ptr = t25;
   t33.value_ptr = t32;
   t33.index = t22;
   t33.found_existing = true;
   return t33;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t22 = t22 + (uintptr_t)1ul;
 t20 = t22;
 goto zig_loop_39;

 zig_block_1:;
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t8->entries;
 t21 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t9);
 t34 = &t16->ptr;
 t35 = (*t34);
 t35 = (uint32_t *)&t35[t21];
 (*t35) = t7;
 t36 = &t19->ptr;
 t25 = (*t36);
 t25 = (nav__6362_42 *)&t25[t21];
 t11 = (*t13);
 t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t11);
 t37 = t29;
 t31 = (nav__6362_72 const *)&t37;
 t38 = &t31->ptr;
 t32 = (*t38);
 t32 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t32[t21];
 t33.key_ptr = t25;
 t33.value_ptr = t32;
 t33.index = t21;
 t33.found_existing = false;
 return t33;

 zig_block_0:;
 t39 = t2;
 t40 = (struct array_hash_map_IndexHeader__892 *const *)&t39;
 t6 = (*t40);
 t41 = (*t6);
 t42 = array_hash_map_IndexHeader_capacityIndexType__3973(t41);
 switch (t42) {
  case UINT8_C(0): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6371(t8, a1, t2);
   return t33;
  }
  case UINT8_C(1): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6372(t8, a1, t2);
   return t33;
  }
  case UINT8_C(2): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6373(t8, a1, t2);
   return t33;
  }
  default: zig_unreachable();
 }
}

static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4023(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 void *t3;
 uint64_t t4;
 uint64_t t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t6;
 struct array_hash_map_IndexHeader__892 **t8;
 struct array_hash_map_IndexHeader__892 *t9;
 struct array_hash_map_IndexHeader__892 *t10;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t20;
 struct array_hash_map_IndexHeader__892 *const *t12;
 nav__4023_62 t18;
 struct array_hash_map_IndexHeader__892 t13;
 uint32_t t14;
 uint16_t t7;
 nav__4023_60 t16;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3982(t3);
 t4 = a2;
 t5 = t4 <= UINT64_C(8);
 if (t5) {
  t2 = (*t1);
  t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t2->entries;
  t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4149(t6, a1, a2);
  if (t7) {
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3983(t3);
   return t7;
  }
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t10 = t9;
  t11 = t10;
  t12 = (struct array_hash_map_IndexHeader__892 *const *)&t11;
  t10 = (*t12);
  t13 = (*t10);
  t14 = array_hash_map_IndexHeader_capacity__3974(t13);
  t4 = a2;
  t15 = (uint64_t)t14;
  t5 = t4 <= t15;
  if (t5) {
   t2 = (*t1);
   t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t2->entries;
   t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4149(t6, a1, a2);
   if (t7) {
    t2 = (*t1);
    t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
    debug_SafetyLock_unlock__3983(t3);
    return t7;
   }
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3983(t3);
   return 0;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t2->entries;
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4149(t6, a1, a2);
 if (t7) {
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t16 = array_hash_map_IndexHeader_findBitIndex__3977(a2);
 if (t16.error) {
  t7 = t16.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t17 = t16.payload;
 t18 = array_hash_map_IndexHeader_alloc__3978(a1, t17);
 if (t18.error) {
  t7 = t18.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t10 = t18.payload;
 t8 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t19 = t9;
  t20 = t19;
  t12 = (struct array_hash_map_IndexHeader__892 *const *)&t20;
  t19 = (*t12);
  array_hash_map_IndexHeader_free__3979(t19, a1);
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t2 = (*t1);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4115(t2, t10);
 t2 = (*t1);
 t8 = (struct array_hash_map_IndexHeader__892 **)&t2->index_header;
 t19 = (struct array_hash_map_IndexHeader__892 *)t10;
 (*t8) = t19;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3983(t3);
 return 0;
}

static nav__6363_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6363(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6363_41 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_IndexHeader__892 *t3;
 struct array_hash_map_IndexHeader__892 *t5;
 struct array_hash_map_IndexHeader__892 *t24;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t11;
 nav__6363_61 t12;
 nav__6363_64 t13;
 uintptr_t t15;
 uintptr_t t16;
 uintptr_t t14;
 uint64_t t17;
 uint64_t t18;
 nav__6363_41 *t19;
 nav__6363_41 t22;
 nav__6363_38 t23;
 struct array_hash_map_IndexHeader__892 *const *t25;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 uint32_t t6;
 uint32_t t20;
 struct array_hash_map_IndexHeader__892 t26;
 bool t4;
 bool t21;
 uint8_t t27;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t3 = a0.index_header;
 t4 = t3 != NULL;
 if (t4) {
  t5 = t3;
  t2 = t5;
  goto zig_block_0;
 }
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t7 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *)&t1->entries;
 t8 = (*t7);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t8);
 t10 = t9;
 t11 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t10;
 t9 = (*t11);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t9);
 t9 = (*t11);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t9);
 t14 = (uintptr_t)0ul;
 t15 = t13.len;
 zig_loop_32:
 t16 = t14;
 t17 = t16;
 t18 = t15;
 t4 = t17 < t18;
 if (t4) {
  t19 = &t13.ptr[t16];
  t20 = t12.ptr[t16];
  t4 = t20 == t6;
  if (t4) {
   t22 = (*t19);
   t4 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t22, t16);
   t21 = t4;
   goto zig_block_4;
  }
  t21 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t21) {
   t23.is_null = false;
   t23.payload = t16;
   return t23;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t16 = t16 + (uintptr_t)1ul;
 t14 = t16;
 goto zig_loop_32;

 zig_block_1:;
 return (nav__6363_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t24 = t2;
 t25 = (struct array_hash_map_IndexHeader__892 *const *)&t24;
 t5 = (*t25);
 t26 = (*t5);
 t27 = array_hash_map_IndexHeader_capacityIndexType__3973(t26);
 switch (t27) {
  case UINT8_C(0): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6377(t28, a1, t2);
   return t23;
  }
  case UINT8_C(1): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6378(t28, a1, t2);
   return t23;
  }
  case UINT8_C(2): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6379(t28, a1, t2);
   return t23;
  }
  default: zig_unreachable();
 }
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6364(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, nav__6364_42 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t6;
 struct array_hash_map_IndexHeader__892 *t39;
 struct array_hash_map_IndexHeader__892 **t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t13;
 nav__6364_66 t14;
 nav__6364_66 t15;
 nav__6364_66 const *t16;
 nav__6364_69 t17;
 nav__6364_69 t18;
 nav__6364_69 const *t19;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t20;
 uint64_t t23;
 uint64_t t24;
 nav__6364_42 *t25;
 nav__6364_42 t28;
 nav__6364_72 t29;
 nav__6364_72 t30;
 nav__6364_72 t37;
 nav__6364_72 const *t31;
 struct environment_root_EnvironmentValue__856 *t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t33;
 uint32_t *const *t34;
 uint32_t *t35;
 nav__6364_42 *const *t36;
 struct environment_root_EnvironmentValue__856 *const *t38;
 struct array_hash_map_IndexHeader__892 *const *t40;
 uint32_t t7;
 uint32_t t26;
 struct array_hash_map_IndexHeader__892 t41;
 bool t5;
 bool t27;
 uint8_t t42;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t3 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t4 = (*t3);
 t5 = t4 != NULL;
 if (t5) {
  t6 = t4;
  t2 = t6;
  goto zig_block_0;
 }
 t7 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t8->entries;
 t10 = (*t9);
 t11 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t10);
 t12 = t11;
 t13 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t12;
 t11 = (*t13);
 t14 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t11);
 t15 = t14;
 t16 = (nav__6364_66 const *)&t15;
 t11 = (*t13);
 t17 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t11);
 t18 = t17;
 t19 = (nav__6364_69 const *)&t18;
 t20 = (uintptr_t)0ul;
 t21 = t17.len;
 zig_loop_39:
 t22 = t20;
 t23 = t22;
 t24 = t21;
 t5 = t23 < t24;
 if (t5) {
  t25 = &t17.ptr[t22];
  t26 = t14.ptr[t22];
  t5 = t26 == t7;
  if (t5) {
   t28 = (*t25);
   t5 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t28, t22);
   t27 = t5;
   goto zig_block_4;
  }
  t27 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t27) {
   t11 = (*t13);
   t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t11);
   t30 = t29;
   t31 = (nav__6364_72 const *)&t30;
   t29 = (*t31);
   t32 = &t29.ptr[t22];
   t33.key_ptr = t25;
   t33.value_ptr = t32;
   t33.index = t22;
   t33.found_existing = true;
   return t33;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t22 = t22 + (uintptr_t)1ul;
 t20 = t22;
 goto zig_loop_39;

 zig_block_1:;
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t8->entries;
 t21 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t9);
 t34 = &t16->ptr;
 t35 = (*t34);
 t35 = (uint32_t *)&t35[t21];
 (*t35) = t7;
 t36 = &t19->ptr;
 t25 = (*t36);
 t25 = (nav__6364_42 *)&t25[t21];
 t11 = (*t13);
 t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t11);
 t37 = t29;
 t31 = (nav__6364_72 const *)&t37;
 t38 = &t31->ptr;
 t32 = (*t38);
 t32 = (struct environment_root_EnvironmentValue__856 *)&t32[t21];
 t33.key_ptr = t25;
 t33.value_ptr = t32;
 t33.index = t21;
 t33.found_existing = false;
 return t33;

 zig_block_0:;
 t39 = t2;
 t40 = (struct array_hash_map_IndexHeader__892 *const *)&t39;
 t6 = (*t40);
 t41 = (*t6);
 t42 = array_hash_map_IndexHeader_capacityIndexType__3973(t41);
 switch (t42) {
  case UINT8_C(0): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6380(t8, a1, t2);
   return t33;
  }
  case UINT8_C(1): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6381(t8, a1, t2);
   return t33;
  }
  case UINT8_C(2): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6382(t8, a1, t2);
   return t33;
  }
  default: zig_unreachable();
 }
}

static void debug_SafetyLock_lock__3982(void *const a0) {
 void *t0;
 t0 = a0;
 return;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3959(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *t1;
 struct mem_Allocator__599 const *t3;
 uintptr_t *t4;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t2;
 nav__3959_54 t10;
 nav__3959_52 t12;
 nav__3959_52 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t18;
 uint8_t **t16;
 uint8_t *t17;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t19;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t20;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t22;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t21;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t23;
 nav__3959_62 t24;
 nav__3959_62 t25;
 uint32_t *t26;
 nav__3959_67 t27;
 nav__3959_67 t28;
 nav__3959_65 *t29;
 nav__3959_72 t30;
 nav__3959_72 t31;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t32;
 uint16_t t11;
 bool t8;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t6 = a2;
 t7 = t5;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 t9 = (*t3);
 t5 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3966(a2);
 t10 = mem_Allocator_alignedAlloc__anon_7569__6383(t9, t5);
 if (t10.error) {
  t11 = t10.error;
  return t11;
 }
 t12 = t10.payload;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = (*t3);
  t13 = (*t1);
  t14 = (*t13);
  t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3967(t14);
  mem_Allocator_free__anon_6654__6273(t9, t15);
  t13 = (*t1);
  t16 = (uint8_t **)&t13->bytes;
  t17 = t12.ptr;
  (*t16) = t17;
  t13 = (*t1);
  t4 = (uintptr_t *)&t13->capacity;
  (*t4) = a2;
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t17 = t12.ptr;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t14.bytes = t17;
 t14.len = t5;
 t14.capacity = a2;
 t18 = t14;
 t13 = (*t1);
 t14 = (*t13);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t14);
 t20 = t19;
 t21 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t20;
 t14 = t18;
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t14);
 t22 = t19;
 t23 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t22;
 t19 = (*t23);
 t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t19);
 t19 = (*t21);
 t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t19);
 t26 = t25.ptr;
 if (t24.len != 0) memcpy(t24.ptr, t26, t24.len * sizeof(uint32_t));
 t19 = (*t23);
 t27 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t19);
 t19 = (*t21);
 t28 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t19);
 t29 = t28.ptr;
 if (t27.len != 0) memcpy(t27.ptr, t29, t27.len * sizeof(nav__3959_65));
 t19 = (*t23);
 t30 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t19);
 t19 = (*t21);
 t31 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t19);
 t32 = t31.ptr;
 if (t30.len != 0) memcpy(t30.ptr, t32, t30.len * sizeof(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858));
 t9 = (*t3);
 t13 = (*t1);
 t14 = (*t13);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3967(t14);
 mem_Allocator_free__anon_6654__6273(t9, t12);
 t14 = t18;
 (*a0) = t14;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3955(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t0;
 uint16_t t8;
 uint16_t t9;
 bool t6;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a2;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3957(t3, a2);
 t8 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3959(t7, a1, t3);
 memcpy(&t9, &t8, sizeof(uint16_t));
 return t9;
}

static void debug_SafetyLock_unlock__3983(void *const a0) {
 void *t0;
 t0 = a0;
 return;
}

static uint32_t array_hash_map_IndexHeader_capacity__3974(struct array_hash_map_IndexHeader__892 const a0) {
 static uint32_t const t2[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};
 uintptr_t t1;
 uint32_t t3;
 uint8_t t0;
 t0 = a0.bit_index;
 t1 = (uintptr_t)t0;
 t3 = t2[t1];
 return t3;
}

static nav__3977_38 array_hash_map_IndexHeader_findBitIndex__3977(uintptr_t const a0) {
 static uint32_t const t5[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};
 static uint32_t const t8[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};
 uint64_t t0;
 uint64_t t7;
 uintptr_t t4;
 uint32_t t6;
 nav__3977_38 t9;
 bool t1;
 uint8_t t3;
 uint8_t t2;
 t0 = a0;
 t1 = t0 > UINT64_C(4294967295);
 if (t1) {
  return (nav__3977_38){zig_error_OutOfMemory,UINT8_C(0xaa)};
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = math_log2_int_ceil__anon_7722__6384(a0);
 t2 = t3;
 t3 = t2;
 t4 = (uintptr_t)t3;
 t6 = t5[t4];
 t0 = a0;
 t7 = (uint64_t)t6;
 t1 = t0 > t7;
 if (t1) {
  t3 = t2;
  t3 = t3 + UINT8_C(1);
  t2 = t3;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t3 = t2;
 t1 = t3 < UINT8_C(5);
 if (t1) {
  t2 = UINT8_C(5);
  goto zig_block_2;
 }
 goto zig_block_2;

 zig_block_2:;
 t3 = t2;
 t4 = (uintptr_t)t3;
 t6 = t8[t4];
 t7 = a0;
 t0 = (uint64_t)t6;
 t1 = t7 <= t0;
 debug_assert__180(t1);
 t3 = t2;
 t9.payload = t3;
 t9.error = UINT16_C(0);
 return t9;
}

static nav__3978_40 array_hash_map_IndexHeader_alloc__3978(struct mem_Allocator__599 const a0, uint8_t const a1) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t3;
 uintptr_t t4;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 nav__3978_53 t6;
 nav__3978_40 t8;
 nav__3978_51 t9;
 nav__3978_51 t12;
 nav__3978_51 t10;
 nav__3978_51 const *t11;
 uint8_t *t13;
 struct array_hash_map_IndexHeader__892 *t14;
 uint16_t t7;
 uint8_t t2;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = zig_shlw_u64((uintptr_t)1ul, t2, UINT8_C(64));
 t4 = array_hash_map_capacityIndexSize__3742(a1);
 t3 = t4 * t3;
 t3 = (uintptr_t)4ul + t3;
 t5 = (*t1);
 t6 = mem_Allocator_alignedAlloc__anon_7727__6385(t5, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = ((struct array_hash_map_IndexHeader__892 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul);
  t8.error = t7;
  return t8;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (nav__3978_51 const *)&t10;
 t12 = (*t11);
 t13 = t12.ptr;
 t13 = (uint8_t *)(((uintptr_t)t13) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = t12.len;
 t3 = t3 - (uintptr_t)4ul;
 t12.ptr = t13;
 t12.len = t3;
 memset(t12.ptr, UINT8_MAX, t12.len);
 t13 = t9.ptr;
 t14 = (struct array_hash_map_IndexHeader__892 *)t13;
 t13 = (uint8_t *)&t14->bit_index;
 (*t13) = a1;
 t8.payload = t14;
 t8.error = UINT16_C(0);
 return t8;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3904(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct array_hash_map_IndexHeader__892 t5;
 uint8_t t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = (*t4);
 t6 = array_hash_map_IndexHeader_capacityIndexType__3973(t5);
 switch (t6) {
  case UINT8_C(0): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6386(t7, a1);
   return;
  }
  case UINT8_C(1): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6387(t7, a1);
   return;
  }
  case UINT8_C(2): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6388(t7, a1);
   return;
  }
  default: zig_unreachable();
 }
}

static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(nav__6365_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_StringContext_hash__3773(a0);
 return t0;
}

static nav__6366_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 uint32_t *t5;
 uint32_t *t6;
 uint32_t *const *t7;
 nav__6366_39 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6366_39){(uint32_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)0ul];
 t5 = (uint32_t *)t4;
 t6 = t5;
 t7 = (uint32_t *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (uint32_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint32_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(nav__6367_40 const a0, nav__6367_40 const a1, uintptr_t const a2) {
 bool t0;
 t0 = array_hash_map_StringContext_eql__3774(a0, a1, a2);
 return t0;
}

static uint8_t array_hash_map_IndexHeader_capacityIndexType__3973(struct array_hash_map_IndexHeader__892 const a0) {
 uint8_t t0;
 t0 = a0.bit_index;
 t0 = array_hash_map_capacityIndexType__3741(t0);
 return t0;
}

static nav__6368_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6368(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6368_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6368_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 nav__6368_38 t8;
 bool t9;
 struct array_hash_map_Index_28u8_29__6735 t11;
 uint8_t t12;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6390(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6368_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__6369_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6369(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6369_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6369_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 nav__6369_38 t8;
 struct array_hash_map_Index_28u16_29__6739 t11;
 uint16_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6392(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6369_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__6370_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6370(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6370_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6370_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 nav__6370_38 t8;
 struct array_hash_map_Index_28u32_29__6743 t11;
 uint32_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6394(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6370_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t0;
 bool t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *const *)&t0;
 t2 = (uintptr_t *)&a0->len;
 t3 = (*t2);
 t2 = (uintptr_t *)&a0->capacity;
 t4 = (*t2);
 t5 = t3;
 t6 = t4;
 t7 = t5 < t6;
 debug_assert__180(t7);
 t2 = (uintptr_t *)&a0->len;
 t4 = (*t2);
 t8 = (*t1);
 t2 = (uintptr_t *)&t8->len;
 t3 = (*t2);
 t3 = t3 + (uintptr_t)1ul;
 (*t2) = t3;
 return t4;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6371(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, nav__6371_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t9;
 nav__6371_66 t10;
 nav__6371_66 t11;
 nav__6371_66 const *t12;
 nav__6371_69 t13;
 nav__6371_69 t14;
 nav__6371_69 const *t15;
 nav__6371_72 t16;
 nav__6371_72 t17;
 nav__6371_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6371_77 t20;
 nav__6371_77 t55;
 nav__6371_77 t21;
 nav__6371_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u8_29__6735 *t35;
 uint8_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__6371_42 **t42;
 nav__6371_42 *const *t43;
 nav__6371_42 *t44;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **t45;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t46;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t47;
 uintptr_t *t48;
 nav__6371_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__892 t32;
 bool t30;
 bool t50;
 struct array_hash_map_Index_28u8_29__6735 t34;
 struct array_hash_map_Index_28u8_29__6735 t33;
 uint8_t t37;
 uint8_t t53;
 uint8_t t27;
 uint8_t t56;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t7);
 t11 = t10;
 t12 = (nav__6371_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t7);
 t14 = t13;
 t15 = (nav__6371_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t7);
 t17 = t16;
 t18 = (nav__6371_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t19);
 t21 = t20;
 t22 = (nav__6371_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT8_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u8_29_isEmpty__6280(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint8_t *)&t35->entry_index;
   t37 = (uint8_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__6371_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__6371_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__6371_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint8_t *)&t35->entry_index;
   t53 = (uint8_t)t24;
   (*t36) = t53;
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT8_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u8_29_isEmpty__6280(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint8_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__6371_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__6371_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint8_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT8_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT8_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6372(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, nav__6372_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t9;
 nav__6372_66 t10;
 nav__6372_66 t11;
 nav__6372_66 const *t12;
 nav__6372_69 t13;
 nav__6372_69 t14;
 nav__6372_69 const *t15;
 nav__6372_72 t16;
 nav__6372_72 t17;
 nav__6372_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6372_77 t20;
 nav__6372_77 t55;
 nav__6372_77 t21;
 nav__6372_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u16_29__6739 *t35;
 uint16_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__6372_42 **t42;
 nav__6372_42 *const *t43;
 nav__6372_42 *t44;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **t45;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t46;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t47;
 uintptr_t *t48;
 nav__6372_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__892 t32;
 struct array_hash_map_Index_28u16_29__6739 t34;
 struct array_hash_map_Index_28u16_29__6739 t33;
 uint16_t t37;
 uint16_t t53;
 uint16_t t27;
 uint16_t t56;
 bool t30;
 bool t50;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t7);
 t11 = t10;
 t12 = (nav__6372_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t7);
 t14 = t13;
 t15 = (nav__6372_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t7);
 t17 = t16;
 t18 = (nav__6372_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t19);
 t21 = t20;
 t22 = (nav__6372_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT16_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u16_29_isEmpty__6285(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint16_t *)&t35->entry_index;
   t37 = (uint16_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__6372_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__6372_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__6372_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint16_t *)&t35->entry_index;
   t53 = (uint16_t)t24;
   (*t36) = t53;
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT16_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u16_29_isEmpty__6285(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint16_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__6372_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__6372_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint16_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT16_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT16_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6373(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, nav__6373_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t9;
 nav__6373_66 t10;
 nav__6373_66 t11;
 nav__6373_66 const *t12;
 nav__6373_69 t13;
 nav__6373_69 t14;
 nav__6373_69 const *t15;
 nav__6373_72 t16;
 nav__6373_72 t17;
 nav__6373_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6373_77 t20;
 nav__6373_77 t53;
 nav__6373_77 t21;
 nav__6373_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t52;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u32_29__6743 *t35;
 uint32_t *t36;
 uint32_t *const *t38;
 bool *t40;
 nav__6373_42 **t41;
 nav__6373_42 *const *t42;
 nav__6373_42 *t43;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **t44;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *const *t45;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t46;
 uintptr_t *t47;
 nav__6373_42 t50;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t39;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7360 t55;
 uint32_t t23;
 uint32_t t37;
 uint32_t t48;
 uint32_t t27;
 uint32_t t54;
 struct array_hash_map_IndexHeader__892 t32;
 struct array_hash_map_Index_28u32_29__6743 t34;
 struct array_hash_map_Index_28u32_29__6743 t33;
 bool t30;
 bool t49;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t7);
 t11 = t10;
 t12 = (nav__6373_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t7);
 t14 = t13;
 t15 = (nav__6373_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6181(t7);
 t17 = t16;
 t18 = (nav__6373_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t19);
 t21 = t20;
 t22 = (nav__6373_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT32_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u32_29_isEmpty__6290(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint32_t *)&t35->entry_index;
   t37 = (uint32_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t25];
   (*t36) = t23;
   t40 = (bool *)&t39.found_existing;
   (*t40) = false;
   t41 = (nav__6373_42 **)&t39.key_ptr;
   t42 = &t15->ptr;
   t43 = (*t42);
   t43 = (nav__6373_42 *)&t43[t25];
   (*t41) = t43;
   t44 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t39.value_ptr;
   t45 = &t18->ptr;
   t46 = (*t45);
   t46 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t46[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t39.index;
   (*t47) = t25;
   return t39;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t48 = t10.ptr[t24];
  t30 = t23 == t48;
  if (t30) {
   t24 = (uintptr_t)t37;
   t50 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t50, t24);
   t49 = t30;
   goto zig_block_4;
  }
  t49 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t49) {
   t40 = (bool *)&t51.found_existing;
   (*t40) = true;
   t41 = (nav__6373_42 **)&t51.key_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t13 = (*t15);
   t43 = &t13.ptr[t25];
   (*t41) = t43;
   t44 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t51.value_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t16 = (*t18);
   t46 = &t16.ptr[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t51.index;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   (*t47) = t25;
   return t51;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t48 = t27;
  t30 = t37 < t48;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3944(t5);
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t24];
   (*t36) = t23;
   t52 = t31;
   t53 = (*t22);
   t35 = &t53.ptr[t52];
   t36 = (uint32_t *)&t35->entry_index;
   t23 = (uint32_t)t24;
   (*t36) = t23;
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t23 = t27;
   (*t36) = t23;
   t34 = t33;
   t23 = t34.distance_from_start_index;
   t27 = t23;
   t34 = t33;
   t23 = t34.entry_index;
   t54 = t23;
   t52 = t26;
   t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
   t26 = t52;
   t23 = t27;
   t23 = t23 + UINT32_C(1);
   t27 = t23;
   zig_loop_200:
   t52 = t26;
   t29 = t52;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t52 = t26;
    t52 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t52);
    t31 = t52;
    t52 = t31;
    t34 = t20.ptr[t52];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u32_29_isEmpty__6290(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint32_t *)&t35->entry_index;
     t23 = t54;
     (*t36) = t23;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t23 = t27;
     (*t36) = t23;
     t40 = (bool *)&t55.found_existing;
     (*t40) = false;
     t41 = (nav__6373_42 **)&t55.key_ptr;
     t42 = &t15->ptr;
     t43 = (*t42);
     t43 = (nav__6373_42 *)&t43[t24];
     (*t41) = t43;
     t44 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 **)&t55.value_ptr;
     t45 = &t18->ptr;
     t46 = (*t45);
     t46 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *)&t46[t24];
     (*t44) = t46;
     t47 = (uintptr_t *)&t55.index;
     (*t47) = t24;
     return t55;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t23 = t34.distance_from_start_index;
    t48 = t27;
    t30 = t23 < t48;
    if (t30) {
     t52 = t31;
     t53 = (*t22);
     t35 = &t53.ptr[t52];
     t36 = (uint32_t *)&t35->entry_index;
     t48 = t54;
     (*t36) = t48;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t48 = t27;
     (*t36) = t48;
     t34 = t33;
     t48 = t34.entry_index;
     t54 = t48;
     t34 = t33;
     t48 = t34.distance_from_start_index;
     t27 = t48;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t52 = t26;
    t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
    t26 = t52;
    t23 = t27;
    t23 = t23 + UINT32_C(1);
    t27 = t23;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t52 = t26;
  t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
  t26 = t52;
  t37 = t27;
  t37 = t37 + UINT32_C(1);
  t27 = t37;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4153(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *t1;
 struct mem_Allocator__599 const *t3;
 uintptr_t *t4;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t2;
 nav__4153_54 t10;
 nav__4153_52 t12;
 nav__4153_52 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t18;
 uint8_t **t16;
 uint8_t *t17;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t19;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t20;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t22;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t21;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t23;
 nav__4153_62 t24;
 nav__4153_62 t25;
 uint32_t *t26;
 nav__4153_67 t27;
 nav__4153_67 t28;
 nav__4153_65 *t29;
 nav__4153_72 t30;
 nav__4153_72 t31;
 struct environment_root_EnvironmentValue__856 *t32;
 uint16_t t11;
 bool t8;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t6 = a2;
 t7 = t5;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 t9 = (*t3);
 t5 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4160(a2);
 t10 = mem_Allocator_alignedAlloc__anon_7569__6383(t9, t5);
 if (t10.error) {
  t11 = t10.error;
  return t11;
 }
 t12 = t10.payload;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = (*t3);
  t13 = (*t1);
  t14 = (*t13);
  t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4161(t14);
  mem_Allocator_free__anon_6654__6273(t9, t15);
  t13 = (*t1);
  t16 = (uint8_t **)&t13->bytes;
  t17 = t12.ptr;
  (*t16) = t17;
  t13 = (*t1);
  t4 = (uintptr_t *)&t13->capacity;
  (*t4) = a2;
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t17 = t12.ptr;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t14.bytes = t17;
 t14.len = t5;
 t14.capacity = a2;
 t18 = t14;
 t13 = (*t1);
 t14 = (*t13);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t14);
 t20 = t19;
 t21 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t20;
 t14 = t18;
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t14);
 t22 = t19;
 t23 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t22;
 t19 = (*t23);
 t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t19);
 t19 = (*t21);
 t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t19);
 t26 = t25.ptr;
 if (t24.len != 0) memcpy(t24.ptr, t26, t24.len * sizeof(uint32_t));
 t19 = (*t23);
 t27 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t19);
 t19 = (*t21);
 t28 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t19);
 t29 = t28.ptr;
 if (t27.len != 0) memcpy(t27.ptr, t29, t27.len * sizeof(nav__4153_65));
 t19 = (*t23);
 t30 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t19);
 t19 = (*t21);
 t31 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t19);
 t32 = t31.ptr;
 if (t30.len != 0) memcpy(t30.ptr, t32, t30.len * sizeof(struct environment_root_EnvironmentValue__856));
 t9 = (*t3);
 t13 = (*t1);
 t14 = (*t13);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4161(t14);
 mem_Allocator_free__anon_6654__6273(t9, t12);
 t14 = t18;
 (*a0) = t14;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4149(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t0;
 uint16_t t8;
 uint16_t t9;
 bool t6;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a2;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4151(t3, a2);
 t8 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4153(t7, a1, t3);
 memcpy(&t9, &t8, sizeof(uint16_t));
 return t9;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4115(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct array_hash_map_IndexHeader__892 t5;
 uint8_t t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = (*t4);
 t6 = array_hash_map_IndexHeader_capacityIndexType__3973(t5);
 switch (t6) {
  case UINT8_C(0): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6396(t7, a1);
   return;
  }
  case UINT8_C(1): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6397(t7, a1);
   return;
  }
  case UINT8_C(2): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6398(t7, a1);
   return;
  }
  default: zig_unreachable();
 }
}

static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(nav__6374_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_StringContext_hash__3773(a0);
 return t0;
}

static nav__6375_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 uint32_t *t5;
 uint32_t *t6;
 uint32_t *const *t7;
 nav__6375_39 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__6375_39){(uint32_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)0ul];
 t5 = (uint32_t *)t4;
 t6 = t5;
 t7 = (uint32_t *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (uint32_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint32_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(nav__6376_40 const a0, nav__6376_40 const a1, uintptr_t const a2) {
 bool t0;
 t0 = array_hash_map_StringContext_eql__3774(a0, a1, a2);
 return t0;
}

static nav__6377_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6377(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6377_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6377_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 nav__6377_38 t8;
 bool t9;
 struct array_hash_map_Index_28u8_29__6735 t11;
 uint8_t t12;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6399(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6377_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__6378_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6378(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6378_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6378_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 nav__6378_38 t8;
 struct array_hash_map_Index_28u16_29__6739 t11;
 uint16_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6400(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6378_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__6379_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6379(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6379_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6379_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 nav__6379_38 t8;
 struct array_hash_map_Index_28u32_29__6743 t11;
 uint32_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6401(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__6379_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t0;
 bool t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *const *)&t0;
 t2 = (uintptr_t *)&a0->len;
 t3 = (*t2);
 t2 = (uintptr_t *)&a0->capacity;
 t4 = (*t2);
 t5 = t3;
 t6 = t4;
 t7 = t5 < t6;
 debug_assert__180(t7);
 t2 = (uintptr_t *)&a0->len;
 t4 = (*t2);
 t8 = (*t1);
 t2 = (uintptr_t *)&t8->len;
 t3 = (*t2);
 t3 = t3 + (uintptr_t)1ul;
 (*t2) = t3;
 return t4;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6380(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, nav__6380_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t9;
 nav__6380_66 t10;
 nav__6380_66 t11;
 nav__6380_66 const *t12;
 nav__6380_69 t13;
 nav__6380_69 t14;
 nav__6380_69 const *t15;
 nav__6380_72 t16;
 nav__6380_72 t17;
 nav__6380_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6380_77 t20;
 nav__6380_77 t55;
 nav__6380_77 t21;
 nav__6380_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u8_29__6735 *t35;
 uint8_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__6380_42 **t42;
 nav__6380_42 *const *t43;
 nav__6380_42 *t44;
 struct environment_root_EnvironmentValue__856 **t45;
 struct environment_root_EnvironmentValue__856 *const *t46;
 struct environment_root_EnvironmentValue__856 *t47;
 uintptr_t *t48;
 nav__6380_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__892 t32;
 bool t30;
 bool t50;
 struct array_hash_map_Index_28u8_29__6735 t34;
 struct array_hash_map_Index_28u8_29__6735 t33;
 uint8_t t37;
 uint8_t t53;
 uint8_t t27;
 uint8_t t56;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t7);
 t11 = t10;
 t12 = (nav__6380_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t7);
 t14 = t13;
 t15 = (nav__6380_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t7);
 t17 = t16;
 t18 = (nav__6380_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t19);
 t21 = t20;
 t22 = (nav__6380_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT8_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u8_29_isEmpty__6280(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint8_t *)&t35->entry_index;
   t37 = (uint8_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__6380_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__6380_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__856 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_root_EnvironmentValue__856 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__6380_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__856 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint8_t *)&t35->entry_index;
   t53 = (uint8_t)t24;
   (*t36) = t53;
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT8_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u8_29_isEmpty__6280(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint8_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__6380_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__6380_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_root_EnvironmentValue__856 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_root_EnvironmentValue__856 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint8_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT8_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT8_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6381(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, nav__6381_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t9;
 nav__6381_66 t10;
 nav__6381_66 t11;
 nav__6381_66 const *t12;
 nav__6381_69 t13;
 nav__6381_69 t14;
 nav__6381_69 const *t15;
 nav__6381_72 t16;
 nav__6381_72 t17;
 nav__6381_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6381_77 t20;
 nav__6381_77 t55;
 nav__6381_77 t21;
 nav__6381_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u16_29__6739 *t35;
 uint16_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__6381_42 **t42;
 nav__6381_42 *const *t43;
 nav__6381_42 *t44;
 struct environment_root_EnvironmentValue__856 **t45;
 struct environment_root_EnvironmentValue__856 *const *t46;
 struct environment_root_EnvironmentValue__856 *t47;
 uintptr_t *t48;
 nav__6381_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__892 t32;
 struct array_hash_map_Index_28u16_29__6739 t34;
 struct array_hash_map_Index_28u16_29__6739 t33;
 uint16_t t37;
 uint16_t t53;
 uint16_t t27;
 uint16_t t56;
 bool t30;
 bool t50;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t7);
 t11 = t10;
 t12 = (nav__6381_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t7);
 t14 = t13;
 t15 = (nav__6381_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t7);
 t17 = t16;
 t18 = (nav__6381_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t19);
 t21 = t20;
 t22 = (nav__6381_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT16_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u16_29_isEmpty__6285(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint16_t *)&t35->entry_index;
   t37 = (uint16_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__6381_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__6381_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__856 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_root_EnvironmentValue__856 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__6381_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_root_EnvironmentValue__856 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint16_t *)&t35->entry_index;
   t53 = (uint16_t)t24;
   (*t36) = t53;
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT16_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u16_29_isEmpty__6285(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint16_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__6381_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__6381_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_root_EnvironmentValue__856 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_root_EnvironmentValue__856 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint16_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT16_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT16_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6382(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, nav__6382_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t9;
 nav__6382_66 t10;
 nav__6382_66 t11;
 nav__6382_66 const *t12;
 nav__6382_69 t13;
 nav__6382_69 t14;
 nav__6382_69 const *t15;
 nav__6382_72 t16;
 nav__6382_72 t17;
 nav__6382_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6382_77 t20;
 nav__6382_77 t53;
 nav__6382_77 t21;
 nav__6382_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t52;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u32_29__6743 *t35;
 uint32_t *t36;
 uint32_t *const *t38;
 bool *t40;
 nav__6382_42 **t41;
 nav__6382_42 *const *t42;
 nav__6382_42 *t43;
 struct environment_root_EnvironmentValue__856 **t44;
 struct environment_root_EnvironmentValue__856 *const *t45;
 struct environment_root_EnvironmentValue__856 *t46;
 uintptr_t *t47;
 nav__6382_42 t50;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t39;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__7371 t55;
 uint32_t t23;
 uint32_t t37;
 uint32_t t48;
 uint32_t t27;
 uint32_t t54;
 struct array_hash_map_IndexHeader__892 t32;
 struct array_hash_map_Index_28u32_29__6743 t34;
 struct array_hash_map_Index_28u32_29__6743 t33;
 bool t30;
 bool t49;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t7);
 t11 = t10;
 t12 = (nav__6382_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t7);
 t14 = t13;
 t15 = (nav__6382_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6190(t7);
 t17 = t16;
 t18 = (nav__6382_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t19);
 t21 = t20;
 t22 = (nav__6382_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT32_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u32_29_isEmpty__6290(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint32_t *)&t35->entry_index;
   t37 = (uint32_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t25];
   (*t36) = t23;
   t40 = (bool *)&t39.found_existing;
   (*t40) = false;
   t41 = (nav__6382_42 **)&t39.key_ptr;
   t42 = &t15->ptr;
   t43 = (*t42);
   t43 = (nav__6382_42 *)&t43[t25];
   (*t41) = t43;
   t44 = (struct environment_root_EnvironmentValue__856 **)&t39.value_ptr;
   t45 = &t18->ptr;
   t46 = (*t45);
   t46 = (struct environment_root_EnvironmentValue__856 *)&t46[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t39.index;
   (*t47) = t25;
   return t39;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t48 = t10.ptr[t24];
  t30 = t23 == t48;
  if (t30) {
   t24 = (uintptr_t)t37;
   t50 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t50, t24);
   t49 = t30;
   goto zig_block_4;
  }
  t49 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t49) {
   t40 = (bool *)&t51.found_existing;
   (*t40) = true;
   t41 = (nav__6382_42 **)&t51.key_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t13 = (*t15);
   t43 = &t13.ptr[t25];
   (*t41) = t43;
   t44 = (struct environment_root_EnvironmentValue__856 **)&t51.value_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t16 = (*t18);
   t46 = &t16.ptr[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t51.index;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   (*t47) = t25;
   return t51;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t48 = t27;
  t30 = t37 < t48;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4138(t5);
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t24];
   (*t36) = t23;
   t52 = t31;
   t53 = (*t22);
   t35 = &t53.ptr[t52];
   t36 = (uint32_t *)&t35->entry_index;
   t23 = (uint32_t)t24;
   (*t36) = t23;
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t23 = t27;
   (*t36) = t23;
   t34 = t33;
   t23 = t34.distance_from_start_index;
   t27 = t23;
   t34 = t33;
   t23 = t34.entry_index;
   t54 = t23;
   t52 = t26;
   t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
   t26 = t52;
   t23 = t27;
   t23 = t23 + UINT32_C(1);
   t27 = t23;
   zig_loop_200:
   t52 = t26;
   t29 = t52;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t52 = t26;
    t52 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t52);
    t31 = t52;
    t52 = t31;
    t34 = t20.ptr[t52];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u32_29_isEmpty__6290(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint32_t *)&t35->entry_index;
     t23 = t54;
     (*t36) = t23;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t23 = t27;
     (*t36) = t23;
     t40 = (bool *)&t55.found_existing;
     (*t40) = false;
     t41 = (nav__6382_42 **)&t55.key_ptr;
     t42 = &t15->ptr;
     t43 = (*t42);
     t43 = (nav__6382_42 *)&t43[t24];
     (*t41) = t43;
     t44 = (struct environment_root_EnvironmentValue__856 **)&t55.value_ptr;
     t45 = &t18->ptr;
     t46 = (*t45);
     t46 = (struct environment_root_EnvironmentValue__856 *)&t46[t24];
     (*t44) = t46;
     t47 = (uintptr_t *)&t55.index;
     (*t47) = t24;
     return t55;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t23 = t34.distance_from_start_index;
    t48 = t27;
    t30 = t23 < t48;
    if (t30) {
     t52 = t31;
     t53 = (*t22);
     t35 = &t53.ptr[t52];
     t36 = (uint32_t *)&t35->entry_index;
     t48 = t54;
     (*t36) = t48;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t48 = t27;
     (*t36) = t48;
     t34 = t33;
     t48 = t34.entry_index;
     t54 = t48;
     t34 = t33;
     t48 = t34.distance_from_start_index;
     t27 = t48;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t52 = t26;
    t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
    t26 = t52;
    t23 = t27;
    t23 = t23 + UINT32_C(1);
    t27 = t23;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t52 = t26;
  t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
  t26 = t52;
  t37 = t27;
  t37 = t37 + UINT32_C(1);
  t27 = t37;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3957(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)4ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static nav__6383_40 mem_Allocator_alignedAlloc__anon_7569__6383(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__6383_40 t4;
 nav__6383_40 t8;
 nav__6383_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__6383_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_7842__6402(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__6383_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static uint8_t math_log2_int_ceil__anon_7722__6384(uintptr_t const a0) {
 uint64_t t0;
 uintptr_t t2;
 bool t1;
 uint8_t t3;
 t0 = a0;
 t1 = t0 != UINT64_C(0);
 debug_assert__180(t1);
 t0 = a0;
 t1 = t0 == UINT64_C(1);
 if (t1) {
  return UINT8_C(0);
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = a0 - (uintptr_t)1ul;
 t3 = math_log2_int__anon_7845__6403(t2);
 t3 = t3 + UINT8_C(1);
 return t3;
}

static nav__6385_40 mem_Allocator_alignedAlloc__anon_7727__6385(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__6385_40 t4;
 nav__6385_40 t8;
 nav__6385_51 t6;
 uint8_t *t9;
 uint8_t *t10;
 uint8_t *const *t11;
 nav__6385_39 t12;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_7849__6404(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__6385_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = t9;
 t11 = (uint8_t *const *)&t10;
 t9 = (*t11);
 t9 = (uint8_t *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t9;
 t12.len = a1;
 t8.payload = t12;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6386(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t9;
 nav__6386_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6386_60 t12;
 nav__6386_60 t33;
 nav__6386_60 t13;
 nav__6386_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u8_29__6735 const *t31;
 struct array_hash_map_Index_28u8_29__6735 *t34;
 uint8_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__892 t28;
 bool t20;
 uint8_t t26;
 uint8_t t36;
 uint8_t t25;
 uint8_t t27;
 struct array_hash_map_Index_28u8_29__6735 t29;
 struct array_hash_map_Index_28u8_29__6735 t32;
 struct array_hash_map_Index_28u8_29__6735 t30;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t11);
 t13 = t12;
 t14 = (nav__6386_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint8_t)t17;
  t25 = t26;
  t27 = UINT8_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u8_29__6735 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u8_29_isEmpty__6280(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint8_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint8_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT8_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6387(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t9;
 nav__6387_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6387_60 t12;
 nav__6387_60 t33;
 nav__6387_60 t13;
 nav__6387_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u16_29__6739 const *t31;
 struct array_hash_map_Index_28u16_29__6739 *t34;
 uint16_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__892 t28;
 uint16_t t26;
 uint16_t t36;
 uint16_t t25;
 uint16_t t27;
 struct array_hash_map_Index_28u16_29__6739 t29;
 struct array_hash_map_Index_28u16_29__6739 t32;
 struct array_hash_map_Index_28u16_29__6739 t30;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t11);
 t13 = t12;
 t14 = (nav__6387_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint16_t)t17;
  t25 = t26;
  t27 = UINT16_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u16_29__6739 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u16_29_isEmpty__6285(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint16_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint16_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT16_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6388(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t9;
 nav__6388_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6388_60 t12;
 nav__6388_60 t32;
 nav__6388_60 t13;
 nav__6388_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u32_29__6743 const *t30;
 struct array_hash_map_Index_28u32_29__6743 *t33;
 uint32_t *t34;
 uint32_t t21;
 uint32_t t35;
 uint32_t t25;
 uint32_t t26;
 struct array_hash_map_IndexHeader__892 t27;
 struct array_hash_map_Index_28u32_29__6743 t28;
 struct array_hash_map_Index_28u32_29__6743 t31;
 struct array_hash_map_Index_28u32_29__6743 t29;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t11);
 t13 = t12;
 t14 = (nav__6388_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t21 = (uint32_t)t17;
  t25 = t21;
  t26 = UINT32_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t27 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t27, t22);
   t28 = t12.ptr[t22];
   t29 = t28;
   t30 = (struct array_hash_map_Index_28u32_29__6743 const *)&t29;
   t31 = (*t30);
   t20 = array_hash_map_Index_28u32_29_isEmpty__6290(t31);
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t21 = t26;
    (*t34) = t21;
    t34 = (uint32_t *)&t33->entry_index;
    t21 = t25;
    (*t34) = t21;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t21 = t28.distance_from_start_index;
   t35 = t26;
   t20 = t21 < t35;
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t35 = t26;
    (*t34) = t35;
    t34 = (uint32_t *)&t33->entry_index;
    t35 = t25;
    (*t34) = t35;
    t35 = t28.distance_from_start_index;
    t26 = t35;
    t35 = t28.entry_index;
    t25 = t35;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t35 = t26;
   t35 = t35 + UINT32_C(1);
   t26 = t35;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static uint32_t array_hash_map_StringContext_hash__3773(nav__3773_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_hashString__3736(a0);
 return t0;
}

static bool array_hash_map_StringContext_eql__3774(nav__3774_40 const a0, nav__3774_40 const a1, uintptr_t const a2) {
 bool t0;
 (void)a2;
 t0 = array_hash_map_eqlString__3735(a0, a1);
 return t0;
}

static nav__6389_40 array_hash_map_IndexHeader_indexes__anon_7759__6389(struct array_hash_map_IndexHeader__892 *const a0) {
 struct array_hash_map_IndexHeader__892 *const *t1;
 uint8_t *t2;
 struct array_hash_map_Index_28u8_29__6735 *t3;
 struct array_hash_map_Index_28u8_29__6735 *t4;
 struct array_hash_map_Index_28u8_29__6735 *const *t5;
 struct array_hash_map_IndexHeader__892 *t6;
 struct array_hash_map_IndexHeader__892 *t0;
 uintptr_t t8;
 nav__6389_40 t9;
 struct array_hash_map_IndexHeader__892 t7;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__892 *const *)&t0;
 t2 = (uint8_t *)a0;
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = (struct array_hash_map_Index_28u8_29__6735 *)t2;
 t4 = t3;
 t5 = (struct array_hash_map_Index_28u8_29__6735 *const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_hash_map_IndexHeader_length__3975(t7);
 t3 = (*t5);
 t3 = (struct array_hash_map_Index_28u8_29__6735 *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(struct array_hash_map_Index_28u8_29__6735)));
 t9.ptr = t3;
 t9.len = t8;
 return t9;
}

static nav__6390_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6390(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6390_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__6390_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t8;
 nav__6390_66 t9;
 nav__6390_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u8_29__6735 const *t23;
 nav__6390_41 t30;
 nav__6390_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__892 t20;
 bool t18;
 bool t25;
 struct array_hash_map_Index_28u8_29__6735 t21;
 struct array_hash_map_Index_28u8_29__6735 t24;
 struct array_hash_map_Index_28u8_29__6735 t22;
 uint8_t t26;
 uint8_t t27;
 uint8_t t15;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT8_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u8_29__6735 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u8_29_isEmpty__6280(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6390_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT8_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6391_40 array_hash_map_IndexHeader_indexes__anon_7769__6391(struct array_hash_map_IndexHeader__892 *const a0) {
 struct array_hash_map_IndexHeader__892 *const *t1;
 uint8_t *t2;
 struct array_hash_map_Index_28u16_29__6739 *t3;
 struct array_hash_map_Index_28u16_29__6739 *t4;
 struct array_hash_map_Index_28u16_29__6739 *const *t5;
 struct array_hash_map_IndexHeader__892 *t6;
 struct array_hash_map_IndexHeader__892 *t0;
 uintptr_t t8;
 nav__6391_40 t9;
 struct array_hash_map_IndexHeader__892 t7;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__892 *const *)&t0;
 t2 = (uint8_t *)a0;
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = (struct array_hash_map_Index_28u16_29__6739 *)t2;
 t4 = t3;
 t5 = (struct array_hash_map_Index_28u16_29__6739 *const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_hash_map_IndexHeader_length__3975(t7);
 t3 = (*t5);
 t3 = (struct array_hash_map_Index_28u16_29__6739 *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(struct array_hash_map_Index_28u16_29__6739)));
 t9.ptr = t3;
 t9.len = t8;
 return t9;
}

static nav__6392_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6392(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6392_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__6392_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t8;
 nav__6392_66 t9;
 nav__6392_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u16_29__6739 const *t23;
 nav__6392_41 t30;
 nav__6392_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__892 t20;
 struct array_hash_map_Index_28u16_29__6739 t21;
 struct array_hash_map_Index_28u16_29__6739 t24;
 struct array_hash_map_Index_28u16_29__6739 t22;
 uint16_t t26;
 uint16_t t27;
 uint16_t t15;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT16_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u16_29__6739 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u16_29_isEmpty__6285(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6392_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT16_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6393_40 array_hash_map_IndexHeader_indexes__anon_7775__6393(struct array_hash_map_IndexHeader__892 *const a0) {
 struct array_hash_map_IndexHeader__892 *const *t1;
 uint8_t *t2;
 struct array_hash_map_Index_28u32_29__6743 *t3;
 struct array_hash_map_Index_28u32_29__6743 *t4;
 struct array_hash_map_Index_28u32_29__6743 *const *t5;
 struct array_hash_map_IndexHeader__892 *t6;
 struct array_hash_map_IndexHeader__892 *t0;
 uintptr_t t8;
 nav__6393_40 t9;
 struct array_hash_map_IndexHeader__892 t7;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__892 *const *)&t0;
 t2 = (uint8_t *)a0;
 t2 = (uint8_t *)(((uintptr_t)t2) + ((uintptr_t)4ul*sizeof(uint8_t)));
 t3 = (struct array_hash_map_Index_28u32_29__6743 *)t2;
 t4 = t3;
 t5 = (struct array_hash_map_Index_28u32_29__6743 *const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_hash_map_IndexHeader_length__3975(t7);
 t3 = (*t5);
 t3 = (struct array_hash_map_Index_28u32_29__6743 *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(struct array_hash_map_Index_28u32_29__6743)));
 t9.ptr = t3;
 t9.len = t8;
 return t9;
}

static nav__6394_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6394(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0, nav__6394_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__6394_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *t8;
 nav__6394_66 t9;
 nav__6394_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u32_29__6743 const *t23;
 nav__6394_41 t29;
 nav__6394_38 t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t0;
 uint32_t t11;
 uint32_t t26;
 uint32_t t27;
 uint32_t t15;
 struct array_hash_map_IndexHeader__892 t20;
 struct array_hash_map_Index_28u32_29__6743 t21;
 struct array_hash_map_Index_28u32_29__6743 t24;
 struct array_hash_map_Index_28u32_29__6743 t22;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__3937(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6231 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6366(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHa__6180(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6365(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT32_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u32_29__6743 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u32_29_isEmpty__6290(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6394_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t26 = t9.ptr[t28];
  t25 = t11 == t26;
  if (t25) {
   t28 = (uintptr_t)t27;
   t29 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6367(a1, t29, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t30.is_null = false;
   t30.payload = t12;
   return t30;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT32_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static uintptr_t array_hash_map_safeTruncate__anon_7783__6395(uint32_t const a0) {
 uintptr_t t0;
 t0 = (uintptr_t)a0;
 return t0;
}

static uintptr_t array_hash_map_IndexHeader_constrainIndex__3971(struct array_hash_map_IndexHeader__892 const a0, uintptr_t const a1) {
 struct array_hash_map_IndexHeader__892 const *t1;
 uintptr_t t4;
 struct array_hash_map_IndexHeader__892 t2;
 struct array_hash_map_IndexHeader__892 t0;
 uint32_t t3;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__892 const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_IndexHeader_mask__3976(t2);
 t4 = (uintptr_t)t3;
 t4 = a1 & t4;
 return t4;
}

static bool array_hash_map_Index_28u8_29_isEmpty__6280(struct array_hash_map_Index_28u8_29__6735 const a0) {
 uint8_t t0;
 bool t1;
 t0 = a0.entry_index;
 t1 = t0 == UINT8_MAX;
 return t1;
}

static bool array_hash_map_Index_28u16_29_isEmpty__6285(struct array_hash_map_Index_28u16_29__6739 const a0) {
 uint16_t t0;
 bool t1;
 t0 = a0.entry_index;
 t1 = t0 == UINT16_MAX;
 return t1;
}

static bool array_hash_map_Index_28u32_29_isEmpty__6290(struct array_hash_map_Index_28u32_29__6743 const a0) {
 uint32_t t0;
 bool t1;
 t0 = a0.entry_index;
 t1 = t0 == UINT32_MAX;
 return t1;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4151(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)5ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6396(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t9;
 nav__6396_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6396_60 t12;
 nav__6396_60 t33;
 nav__6396_60 t13;
 nav__6396_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u8_29__6735 const *t31;
 struct array_hash_map_Index_28u8_29__6735 *t34;
 uint8_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__892 t28;
 bool t20;
 uint8_t t26;
 uint8_t t36;
 uint8_t t25;
 uint8_t t27;
 struct array_hash_map_Index_28u8_29__6735 t29;
 struct array_hash_map_Index_28u8_29__6735 t32;
 struct array_hash_map_Index_28u8_29__6735 t30;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t11);
 t13 = t12;
 t14 = (nav__6396_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint8_t)t17;
  t25 = t26;
  t27 = UINT8_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u8_29__6735 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u8_29_isEmpty__6280(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint8_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint8_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT8_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6397(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t9;
 nav__6397_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6397_60 t12;
 nav__6397_60 t33;
 nav__6397_60 t13;
 nav__6397_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u16_29__6739 const *t31;
 struct array_hash_map_Index_28u16_29__6739 *t34;
 uint16_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__892 t28;
 uint16_t t26;
 uint16_t t36;
 uint16_t t25;
 uint16_t t27;
 struct array_hash_map_Index_28u16_29__6739 t29;
 struct array_hash_map_Index_28u16_29__6739 t32;
 struct array_hash_map_Index_28u16_29__6739 t30;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t11);
 t13 = t12;
 t14 = (nav__6397_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint16_t)t17;
  t25 = t26;
  t27 = UINT16_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u16_29__6739 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u16_29_isEmpty__6285(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint16_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint16_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT16_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6398(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t9;
 nav__6398_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__6398_60 t12;
 nav__6398_60 t32;
 nav__6398_60 t13;
 nav__6398_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u32_29__6743 const *t30;
 struct array_hash_map_Index_28u32_29__6743 *t33;
 uint32_t *t34;
 uint32_t t21;
 uint32_t t35;
 uint32_t t25;
 uint32_t t26;
 struct array_hash_map_IndexHeader__892 t27;
 struct array_hash_map_Index_28u32_29__6743 t28;
 struct array_hash_map_Index_28u32_29__6743 t31;
 struct array_hash_map_Index_28u32_29__6743 t29;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t11);
 t13 = t12;
 t14 = (nav__6398_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t21 = (uint32_t)t17;
  t25 = t21;
  t26 = UINT32_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t27 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t27, t22);
   t28 = t12.ptr[t22];
   t29 = t28;
   t30 = (struct array_hash_map_Index_28u32_29__6743 const *)&t29;
   t31 = (*t30);
   t20 = array_hash_map_Index_28u32_29_isEmpty__6290(t31);
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t21 = t26;
    (*t34) = t21;
    t34 = (uint32_t *)&t33->entry_index;
    t21 = t25;
    (*t34) = t21;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t21 = t28.distance_from_start_index;
   t35 = t26;
   t20 = t21 < t35;
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t35 = t26;
    (*t34) = t35;
    t34 = (uint32_t *)&t33->entry_index;
    t35 = t25;
    (*t34) = t35;
    t35 = t28.distance_from_start_index;
    t26 = t35;
    t35 = t28.entry_index;
    t25 = t35;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t35 = t26;
   t35 = t35 + UINT32_C(1);
   t26 = t35;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static nav__6399_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6399(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6399_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__6399_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t8;
 nav__6399_66 t9;
 nav__6399_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u8_29__6735 const *t23;
 nav__6399_41 t30;
 nav__6399_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__892 t20;
 bool t18;
 bool t25;
 struct array_hash_map_Index_28u8_29__6735 t21;
 struct array_hash_map_Index_28u8_29__6735 t24;
 struct array_hash_map_Index_28u8_29__6735 t22;
 uint8_t t26;
 uint8_t t27;
 uint8_t t15;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT8_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u8_29__6735 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u8_29_isEmpty__6280(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6399_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT8_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6400_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6400(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6400_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__6400_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t8;
 nav__6400_66 t9;
 nav__6400_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u16_29__6739 const *t23;
 nav__6400_41 t30;
 nav__6400_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__892 t20;
 struct array_hash_map_Index_28u16_29__6739 t21;
 struct array_hash_map_Index_28u16_29__6739 t24;
 struct array_hash_map_Index_28u16_29__6739 t22;
 uint16_t t26;
 uint16_t t27;
 uint16_t t15;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT16_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u16_29__6739 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u16_29_isEmpty__6285(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6400_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT16_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6401_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6401(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const a0, nav__6401_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__6401_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *t8;
 nav__6401_66 t9;
 nav__6401_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u32_29__6743 const *t23;
 nav__6401_41 t29;
 nav__6401_38 t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t0;
 uint32_t t11;
 uint32_t t26;
 uint32_t t27;
 uint32_t t15;
 struct array_hash_map_IndexHeader__892 t20;
 struct array_hash_map_Index_28u32_29__6743 t21;
 struct array_hash_map_Index_28u32_29__6743 t24;
 struct array_hash_map_Index_28u32_29__6743 t22;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29__906 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__4131(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6380 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6375(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_Env__6189(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6374(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT32_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u32_29__6743 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u32_29_isEmpty__6290(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__6401_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t26 = t9.ptr[t28];
  t25 = t11 == t26;
  if (t25) {
   t28 = (uintptr_t)t27;
   t29 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6376(a1, t29, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t30.is_null = false;
   t30.payload = t12;
   return t30;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT32_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__6402_39 mem_Allocator_allocWithSizeAndAlignment__anon_7842__6402(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__6402_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__6402_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3296__5445((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__6402_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_2823__5103(t6, t2, a2);
 return t7;
}

static uint8_t math_log2_int__anon_7845__6403(uintptr_t const a0) {
 uint64_t t0;
 uint16_t t3;
 bool t1;
 uint8_t t2;
 t0 = a0;
 t1 = t0 != UINT64_C(0);
 debug_assert__180(t1);
 t2 = zig_clz_u64(a0, UINT8_C(64));
 t3 = (uint16_t)t2;
 t3 = UINT16_C(63) - t3;
 t2 = (uint8_t)t3;
 return t2;
}

static nav__6404_39 mem_Allocator_allocWithSizeAndAlignment__anon_7849__6404(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__6404_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__6404_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3296__5445((uintptr_t)1ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__6404_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_7875__6405(t6, t2, a2);
 return t7;
}

static uint32_t array_hash_map_hashString__3736(nav__3736_39 const a0) {
 uint64_t t0;
 uint32_t t1;
 t0 = hash_wyhash_Wyhash_hash__6465(UINT64_C(0), a0);
 t1 = (uint32_t)t0;
 return t1;
}

static bool array_hash_map_eqlString__3735(nav__3735_39 const a0, nav__3735_39 const a1) {
 bool t0;
 t0 = mem_eql__anon_7887__6466(a0, a1);
 return t0;
}

static uint32_t array_hash_map_IndexHeader_mask__3976(struct array_hash_map_IndexHeader__892 const a0) {
 struct array_hash_map_IndexHeader__892 const *t1;
 uintptr_t t3;
 struct array_hash_map_IndexHeader__892 t2;
 struct array_hash_map_IndexHeader__892 t0;
 uint32_t t4;
 t0 = a0;
 t1 = (struct array_hash_map_IndexHeader__892 const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_IndexHeader_length__3975(t2);
 t3 = t3 - (uintptr_t)1ul;
 t4 = (uint32_t)t3;
 return t4;
}

static nav__6405_39 mem_Allocator_allocBytesWithAlignment__anon_7875__6405(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uint64_t t2;
 uint8_t *t4;
 uint8_t *t13;
 uint8_t *t14;
 uint8_t *t15;
 struct mem_Allocator__599 t5;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t7;
 struct mem_Allocator_VTable__602 const *const *t8;
 struct mem_Allocator_VTable__602 const *t9;
 uint8_t *(*const *t10)(void *, uintptr_t, uint8_t, uintptr_t);
 uint8_t *(*t11)(void *, uintptr_t, uint8_t, uintptr_t);
 void *t12;
 uint8_t *const *t16;
 nav__6405_52 t17;
 nav__6405_39 t18;
 bool t3;
 uint8_t t6;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  return (nav__6405_39){(uint8_t *)(uintptr_t)0xfffffffffffffffcul,0};
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t1);
 t6 = mem_Alignment_fromByteUnits__1113((uintptr_t)4ul);
 t7 = t5;
 t1 = (struct mem_Allocator__599 const *)&t7;
 t8 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t9 = (*t8);
 t10 = (uint8_t *(*const *)(void *, uintptr_t, uint8_t, uintptr_t))&t9->alloc;
 t11 = (*t10);
 t12 = t5.ptr;
 t13 = t11(t12, a1, t6, a2);
 t3 = t13 != NULL;
 if (t3) {
  t14 = t13;
  t4 = t14;
  goto zig_block_1;
 }
 return (nav__6405_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_1:;
 t15 = t4;
 t16 = (uint8_t *const *)&t15;
 t14 = (*t16);
 t14 = (uint8_t *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t17.ptr = t14;
 t17.len = a1;
 t14 = (uint8_t *)t4;
 t18.payload = t14;
 t18.error = UINT16_C(0);
 return t18;
}

static uint64_t hash_wyhash_Wyhash_hash__6465(uint64_t const a0, nav__6465_39 const a1) {
 zig_u128 t41;
 zig_u128 t42;
 nav__6465_39 const *t1;
 nav__6465_39 const *t17;
 nav__6465_39 const *t61;
 struct hash_wyhash_Wyhash__7880 t3;
 struct hash_wyhash_Wyhash__7880 t2;
 uintptr_t t4;
 uintptr_t t10;
 uintptr_t t15;
 uintptr_t t26;
 uintptr_t t58;
 uint64_t t5;
 uint64_t t22;
 uint64_t t38;
 uint64_t t39;
 uint64_t t40;
 uint64_t t47;
 uint64_t t48;
 uint64_t t52;
 uint64_t t53;
 uint64_t t63;
 uint64_t t64;
 uint64_t t70;
 uint64_t t71;
 struct hash_wyhash_Wyhash__7880 *const *t8;
 struct hash_wyhash_Wyhash__7880 *t11;
 struct hash_wyhash_Wyhash__7880 *t7;
 struct hash_wyhash_Wyhash__7880 *t28;
 struct hash_wyhash_Wyhash__7880 *t54;
 struct hash_wyhash_Wyhash__7880 *t55;
 struct hash_wyhash_Wyhash__7880 *t68;
 uint64_t *t12;
 uint64_t *t69;
 nav__6465_39 t13;
 nav__6465_39 t59;
 nav__6465_39 t0;
 nav__6465_39 t9;
 nav__6465_39 t16;
 nav__6465_39 t21;
 nav__6465_39 t23;
 nav__6465_39 t24;
 nav__6465_39 t31;
 nav__6465_39 t35;
 nav__6465_39 t44;
 nav__6465_39 t46;
 nav__6465_39 t50;
 nav__6465_39 t51;
 nav__6465_39 t56;
 nav__6465_39 t57;
 nav__6465_39 t60;
 nav__6465_39 t62;
 nav__6465_39 t65;
 nav__6465_39 t66;
 uint8_t const *t14;
 uint8_t const (*t18)[4];
 uint8_t const (*t27)[48];
 uint8_t const (*t29)[48];
 uint8_t const (*const *t30)[48];
 uint8_t const (*t32)[8];
 uint8_t const (*t34)[40];
 uint64_t (*t36)[3];
 uint64_t t37[3];
 uint8_t const (*t43)[32];
 uint8_t const (*t45)[24];
 uint8_t const (*t49)[16];
 uintptr_t *t67;
 uint32_t t20;
 bool t6;
 uint8_t t19[4];
 uint8_t t25;
 uint8_t t33[8];
 t0 = a1;
 t1 = (nav__6465_39 const *)&t0;
 t3 = hash_wyhash_Wyhash_init__6453(a0);
 t2 = t3;
 t4 = a1.len;
 t5 = t4;
 t6 = t5 <= UINT64_C(16);
 if (t6) {
  t7 = &t2;
  t8 = (struct hash_wyhash_Wyhash__7880 *const *)&t7;
  t9 = a1;
  t1 = (nav__6465_39 const *)&t9;
  t4 = a1.len;
  t5 = t4;
  t6 = t5 <= UINT64_C(16);
  debug_assert__180(t6);
  t4 = a1.len;
  t5 = t4;
  t6 = t5 >= UINT64_C(4);
  if (t6) {
   t4 = a1.len;
   t4 = t4 - (uintptr_t)4ul;
   t10 = a1.len;
   t10 = zig_shr_u64(t10, UINT8_C(3));
   t10 = zig_shlw_u64(t10, UINT8_C(2), UINT8_C(64));
   t11 = (*t8);
   t12 = (uint64_t *)&t11->a;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t15 = t13.len;
   t13.ptr = t14;
   t13.len = t15;
   t16 = t13;
   t17 = (nav__6465_39 const *)&t16;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t5 = (uint64_t)t20;
   t5 = zig_shlw_u64(t5, UINT8_C(32), UINT8_C(64));
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
   t15 = t13.len;
   t15 = t15 - t10;
   t13.ptr = t14;
   t13.len = t15;
   t21 = t13;
   t17 = (nav__6465_39 const *)&t21;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t22 = (uint64_t)t20;
   t22 = t5 | t22;
   (*t12) = t22;
   t11 = (*t8);
   t12 = (uint64_t *)&t11->b;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t4*sizeof(uint8_t)));
   t15 = t13.len;
   t15 = t15 - t4;
   t13.ptr = t14;
   t13.len = t15;
   t23 = t13;
   t17 = (nav__6465_39 const *)&t23;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t22 = (uint64_t)t20;
   t22 = zig_shlw_u64(t22, UINT8_C(32), UINT8_C(64));
   t10 = t4 - t10;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
   t4 = t13.len;
   t10 = t4 - t10;
   t13.ptr = t14;
   t13.len = t10;
   t24 = t13;
   t1 = (nav__6465_39 const *)&t24;
   debug_assert__180(true);
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t18 = (uint8_t const (*)[4])t14;
   memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
   memcpy(&t20, &t19, sizeof(uint32_t));
   t20 = zig_wrap_u32(t20, UINT8_C(32));
   t5 = (uint64_t)t20;
   t5 = t22 | t5;
   (*t12) = t5;
   goto zig_block_1;
  }
  t15 = a1.len;
  t22 = t15;
  t6 = t22 > UINT64_C(0);
  if (t6) {
   t11 = (*t8);
   t12 = (uint64_t *)&t11->a;
   t25 = a1.ptr[(uintptr_t)0ul];
   t22 = (uint64_t)t25;
   t22 = zig_shlw_u64(t22, UINT8_C(16), UINT8_C(64));
   t15 = a1.len;
   t15 = zig_shr_u64(t15, UINT8_C(1));
   t25 = a1.ptr[t15];
   t5 = (uint64_t)t25;
   t5 = zig_shlw_u64(t5, UINT8_C(8), UINT8_C(64));
   t5 = t22 | t5;
   t15 = a1.len;
   t15 = t15 - (uintptr_t)1ul;
   t25 = a1.ptr[t15];
   t22 = (uint64_t)t25;
   t22 = t5 | t22;
   (*t12) = t22;
   t11 = (*t8);
   t12 = (uint64_t *)&t11->b;
   (*t12) = UINT64_C(0);
   goto zig_block_2;
  }
  t11 = (*t8);
  t12 = (uint64_t *)&t11->a;
  (*t12) = UINT64_C(0);
  t11 = (*t8);
  t12 = (uint64_t *)&t11->b;
  (*t12) = UINT64_C(0);
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;

  zig_block_1:;
  goto zig_block_0;
 }
 t26 = (uintptr_t)0ul;
 t15 = a1.len;
 t22 = t15;
 t6 = t22 >= UINT64_C(48);
 if (t6) {
  zig_loop_180:
  t15 = t26;
  t15 = t15 + (uintptr_t)48ul;
  t10 = a1.len;
  t22 = t15;
  t5 = t10;
  t6 = t22 < t5;
  if (t6) {
   t10 = t26;
   t13 = (*t1);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
   t27 = (uint8_t const (*)[48])t14;
   t28 = &t2;
   t8 = (struct hash_wyhash_Wyhash__7880 *const *)&t28;
   t29 = t27;
   t30 = (uint8_t const (*const *)[48])&t29;
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t27 = (uint8_t const (*)[48])t14;
   t13.ptr = &(*t27)[(uintptr_t)0ul];
   t13.len = (uintptr_t)48ul;
   t31 = t13;
   t17 = (nav__6465_39 const *)&t31;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t5, &t33, sizeof(uint64_t));
   t5 = zig_wrap_u64(t5, UINT8_C(64));
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)8ul*sizeof(uint8_t)));
   t34 = (uint8_t const (*)[40])t14;
   t13.ptr = &(*t34)[(uintptr_t)0ul];
   t13.len = (uintptr_t)40ul;
   t35 = t13;
   t17 = (nav__6465_39 const *)&t35;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t22, &t33, sizeof(uint64_t));
   t22 = zig_wrap_u64(t22, UINT8_C(64));
   t11 = (*t8);
   t36 = (uint64_t (*)[3])&t11->state;
   t12 = (uint64_t *)&(*t36)[(uintptr_t)0ul];
   t5 = t5 ^ UINT64_C(16646288086500911323);
   t36 = (uint64_t (*)[3])&t2.state;
   memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
   t38 = t37[(uintptr_t)0ul];
   t38 = t22 ^ t38;
   t39 = t5;
   t40 = t38;
   t38 = t39;
   t41 = zig_make_u128(0, t38);
   t38 = t40;
   t42 = zig_make_u128(0, t38);
   t42 = zig_mulw_u128(t41, t42, UINT8_C(128));
   t38 = zig_lo_u128(t42);
   t39 = t38;
   t42 = zig_shr_u128(t42, UINT8_C(64));
   t38 = zig_lo_u128(t42);
   t40 = t38;
   t38 = t39;
   t5 = t40;
   t5 = t38 ^ t5;
   (*t12) = t5;
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)16ul*sizeof(uint8_t)));
   t43 = (uint8_t const (*)[32])t14;
   t13.ptr = &(*t43)[(uintptr_t)0ul];
   t13.len = (uintptr_t)32ul;
   t44 = t13;
   t17 = (nav__6465_39 const *)&t44;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t5, &t33, sizeof(uint64_t));
   t5 = zig_wrap_u64(t5, UINT8_C(64));
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)24ul*sizeof(uint8_t)));
   t45 = (uint8_t const (*)[24])t14;
   t13.ptr = &(*t45)[(uintptr_t)0ul];
   t13.len = (uintptr_t)24ul;
   t46 = t13;
   t17 = (nav__6465_39 const *)&t46;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t38, &t33, sizeof(uint64_t));
   t38 = zig_wrap_u64(t38, UINT8_C(64));
   t11 = (*t8);
   t36 = (uint64_t (*)[3])&t11->state;
   t12 = (uint64_t *)&(*t36)[(uintptr_t)1ul];
   t5 = t5 ^ UINT64_C(10285213230658275043);
   t36 = (uint64_t (*)[3])&t2.state;
   memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
   t22 = t37[(uintptr_t)1ul];
   t22 = t38 ^ t22;
   t47 = t5;
   t48 = t22;
   t22 = t47;
   t42 = zig_make_u128(0, t22);
   t22 = t48;
   t41 = zig_make_u128(0, t22);
   t41 = zig_mulw_u128(t42, t41, UINT8_C(128));
   t22 = zig_lo_u128(t41);
   t47 = t22;
   t41 = zig_shr_u128(t41, UINT8_C(64));
   t22 = zig_lo_u128(t41);
   t48 = t22;
   t22 = t47;
   t5 = t48;
   t5 = t22 ^ t5;
   (*t12) = t5;
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)32ul*sizeof(uint8_t)));
   t49 = (uint8_t const (*)[16])t14;
   t13.ptr = &(*t49)[(uintptr_t)0ul];
   t13.len = (uintptr_t)16ul;
   t50 = t13;
   t17 = (nav__6465_39 const *)&t50;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t5, &t33, sizeof(uint64_t));
   t5 = zig_wrap_u64(t5, UINT8_C(64));
   t27 = (*t30);
   t14 = (uint8_t const *)t27;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)40ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   t13.ptr = &(*t32)[(uintptr_t)0ul];
   t13.len = (uintptr_t)8ul;
   t51 = t13;
   t17 = (nav__6465_39 const *)&t51;
   debug_assert__180(true);
   t13 = (*t17);
   t14 = t13.ptr;
   t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t32 = (uint8_t const (*)[8])t14;
   memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
   memcpy(&t22, &t33, sizeof(uint64_t));
   t22 = zig_wrap_u64(t22, UINT8_C(64));
   t11 = (*t8);
   t36 = (uint64_t (*)[3])&t11->state;
   t12 = (uint64_t *)&(*t36)[(uintptr_t)2ul];
   t5 = t5 ^ UINT64_C(6384245875588680899);
   t36 = (uint64_t (*)[3])&t2.state;
   memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
   t38 = t37[(uintptr_t)2ul];
   t38 = t22 ^ t38;
   t52 = t5;
   t53 = t38;
   t38 = t52;
   t41 = zig_make_u128(0, t38);
   t38 = t53;
   t42 = zig_make_u128(0, t38);
   t42 = zig_mulw_u128(t41, t42, UINT8_C(128));
   t38 = zig_lo_u128(t42);
   t52 = t38;
   t42 = zig_shr_u128(t42, UINT8_C(64));
   t38 = zig_lo_u128(t42);
   t53 = t38;
   t38 = t52;
   t5 = t53;
   t5 = t38 ^ t5;
   (*t12) = t5;
   t10 = t26;
   t10 = t10 + (uintptr_t)48ul;
   t26 = t10;
   goto zig_block_5;
  }
  goto zig_block_4;

  zig_block_5:;
  goto zig_loop_180;

  zig_block_4:;
  t54 = &t2;
  t8 = (struct hash_wyhash_Wyhash__7880 *const *)&t54;
  t11 = (*t8);
  t36 = (uint64_t (*)[3])&t11->state;
  t12 = (uint64_t *)&(*t36)[(uintptr_t)0ul];
  t38 = (*t12);
  t36 = (uint64_t (*)[3])&t2.state;
  memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
  t22 = t37[(uintptr_t)1ul];
  t36 = (uint64_t (*)[3])&t2.state;
  memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
  t5 = t37[(uintptr_t)2ul];
  t5 = t22 ^ t5;
  t5 = t38 ^ t5;
  (*t12) = t5;
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t15 = t26;
 t55 = &t2;
 t8 = (struct hash_wyhash_Wyhash__7880 *const *)&t55;
 t56 = a1;
 t1 = (nav__6465_39 const *)&t56;
 t10 = a1.len;
 t38 = t10;
 t6 = t38 >= UINT64_C(16);
 debug_assert__180(t6);
 t10 = a1.len;
 t10 = t10 - t15;
 t38 = t10;
 t6 = t38 <= UINT64_C(48);
 debug_assert__180(t6);
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t15*sizeof(uint8_t)));
 t10 = t13.len;
 t15 = t10 - t15;
 t13.ptr = t14;
 t13.len = t15;
 t57 = t13;
 t17 = (nav__6465_39 const *)&t57;
 t58 = (uintptr_t)0ul;
 zig_loop_483:
 t15 = t58;
 t15 = t15 + (uintptr_t)16ul;
 t10 = t13.len;
 t38 = t15;
 t22 = t10;
 t6 = t38 < t22;
 if (t6) {
  t11 = (*t8);
  t36 = (uint64_t (*)[3])&t11->state;
  t12 = (uint64_t *)&(*t36)[(uintptr_t)0ul];
  t10 = t58;
  t59 = (*t17);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
  t15 = t59.len;
  t10 = t15 - t10;
  t59.ptr = t14;
  t59.len = t10;
  t60 = t59;
  t61 = (nav__6465_39 const *)&t60;
  debug_assert__180(true);
  t59 = (*t61);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t32 = (uint8_t const (*)[8])t14;
  memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
  memcpy(&t22, &t33, sizeof(uint64_t));
  t22 = zig_wrap_u64(t22, UINT8_C(64));
  t22 = t22 ^ UINT64_C(16646288086500911323);
  t10 = t58;
  t10 = t10 + (uintptr_t)8ul;
  t59 = (*t17);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + (t10*sizeof(uint8_t)));
  t15 = t59.len;
  t10 = t15 - t10;
  t59.ptr = t14;
  t59.len = t10;
  t62 = t59;
  t61 = (nav__6465_39 const *)&t62;
  debug_assert__180(true);
  t59 = (*t61);
  t14 = t59.ptr;
  t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t32 = (uint8_t const (*)[8])t14;
  memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
  memcpy(&t38, &t33, sizeof(uint64_t));
  t38 = zig_wrap_u64(t38, UINT8_C(64));
  t36 = (uint64_t (*)[3])&t2.state;
  memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
  t5 = t37[(uintptr_t)0ul];
  t5 = t38 ^ t5;
  t63 = t22;
  t64 = t5;
  t5 = t63;
  t42 = zig_make_u128(0, t5);
  t5 = t64;
  t41 = zig_make_u128(0, t5);
  t41 = zig_mulw_u128(t42, t41, UINT8_C(128));
  t5 = zig_lo_u128(t41);
  t63 = t5;
  t41 = zig_shr_u128(t41, UINT8_C(64));
  t5 = zig_lo_u128(t41);
  t64 = t5;
  t5 = t63;
  t22 = t64;
  t22 = t5 ^ t22;
  (*t12) = t22;
  t10 = t58;
  t10 = t10 + (uintptr_t)16ul;
  t58 = t10;
  goto zig_block_7;
 }
 goto zig_block_6;

 zig_block_7:;
 goto zig_loop_483;

 zig_block_6:;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->a;
 t15 = a1.len;
 t15 = t15 - (uintptr_t)16ul;
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t15*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 t13.ptr = &(*t32)[(uintptr_t)0ul];
 t13.len = (uintptr_t)8ul;
 t65 = t13;
 t17 = (nav__6465_39 const *)&t65;
 debug_assert__180(true);
 t13 = (*t17);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
 memcpy(&t38, &t33, sizeof(uint64_t));
 t38 = zig_wrap_u64(t38, UINT8_C(64));
 (*t12) = t38;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->b;
 t15 = a1.len;
 t15 = t15 - (uintptr_t)8ul;
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t15*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 t13.ptr = &(*t32)[(uintptr_t)0ul];
 t13.len = (uintptr_t)8ul;
 t66 = t13;
 t1 = (nav__6465_39 const *)&t66;
 debug_assert__180(true);
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t32 = (uint8_t const (*)[8])t14;
 memcpy(t33, (const char *)t32, sizeof(uint8_t[8]));
 memcpy(&t38, &t33, sizeof(uint64_t));
 t38 = zig_wrap_u64(t38, UINT8_C(64));
 (*t12) = t38;
 goto zig_block_0;

 zig_block_0:;
 t67 = (uintptr_t *)&t2.total_len;
 t15 = a1.len;
 (*t67) = t15;
 t68 = &t2;
 t8 = (struct hash_wyhash_Wyhash__7880 *const *)&t68;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->a;
 t38 = (*t12);
 t38 = t38 ^ UINT64_C(16646288086500911323);
 (*t12) = t38;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->b;
 t38 = (*t12);
 t36 = (uint64_t (*)[3])&t2.state;
 memcpy(t37, (const char *)t36, sizeof(uint64_t[3]));
 t22 = t37[(uintptr_t)0ul];
 t22 = t38 ^ t22;
 (*t12) = t22;
 t11 = (*t8);
 t12 = (uint64_t *)&t11->a;
 t11 = (*t8);
 t69 = (uint64_t *)&t11->b;
 t22 = (*t12);
 t42 = zig_make_u128(0, t22);
 t22 = (*t69);
 t41 = zig_make_u128(0, t22);
 t41 = zig_mulw_u128(t42, t41, UINT8_C(128));
 t22 = zig_lo_u128(t41);
 (*t12) = t22;
 t41 = zig_shr_u128(t41, UINT8_C(64));
 t22 = zig_lo_u128(t41);
 (*t69) = t22;
 t69 = (uint64_t *)&t2.a;
 t22 = (*t69);
 t22 = t22 ^ UINT64_C(11562461410679940143);
 t67 = (uintptr_t *)&t2.total_len;
 t15 = (*t67);
 t38 = t15;
 t38 = t22 ^ t38;
 t69 = (uint64_t *)&t2.b;
 t22 = (*t69);
 t22 = t22 ^ UINT64_C(16646288086500911323);
 t70 = t38;
 t71 = t22;
 t22 = t70;
 t41 = zig_make_u128(0, t22);
 t22 = t71;
 t42 = zig_make_u128(0, t22);
 t42 = zig_mulw_u128(t41, t42, UINT8_C(128));
 t22 = zig_lo_u128(t42);
 t70 = t22;
 t42 = zig_shr_u128(t42, UINT8_C(64));
 t22 = zig_lo_u128(t42);
 t71 = t22;
 t22 = t70;
 t38 = t71;
 t38 = t22 ^ t38;
 return t38;
}

static bool mem_eql__anon_7887__6466(nav__6466_39 const a0, nav__6466_39 const a1) {
 nav__6466_39 t0;
 nav__6466_39 t1;
 bool t2;
 t0 = mem_sliceAsBytes__anon_3427__5448(a0);
 t1 = mem_sliceAsBytes__anon_3427__5448(a1);
 t2 = mem_eqlBytes__856(t0, t1);
 return t2;
}

static struct hash_wyhash_Wyhash__7880 hash_wyhash_Wyhash_init__6453(uint64_t const a0) {
 zig_u128 t6;
 zig_u128 t7;
 uint64_t (*t1)[3];
 uint64_t *t2;
 uint64_t t3;
 uint64_t t8;
 uint64_t t4;
 uint64_t t5;
 struct hash_wyhash_Wyhash__7880 t9;
 struct hash_wyhash_Wyhash__7880 t0;
 uint64_t t10[3];
 t0 = (struct hash_wyhash_Wyhash__7880){UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),{UINT64_C(0xaaaaaaaaaaaaaaaa), UINT64_C(0xaaaaaaaaaaaaaaaa), UINT64_C(0xaaaaaaaaaaaaaaaa)},(uintptr_t)0ul,(uintptr_t)0ul,"\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252"};
 t1 = (uint64_t (*)[3])&t0.state;
 t2 = (uint64_t *)&(*t1)[(uintptr_t)0ul];
 t3 = a0 ^ UINT64_C(11562461410679940143);
 t4 = t3;
 t5 = UINT64_C(16646288086500911323);
 t3 = t4;
 t6 = zig_make_u128(0, t3);
 t3 = t5;
 t7 = zig_make_u128(0, t3);
 t7 = zig_mulw_u128(t6, t7, UINT8_C(128));
 t3 = zig_lo_u128(t7);
 t4 = t3;
 t7 = zig_shr_u128(t7, UINT8_C(64));
 t3 = zig_lo_u128(t7);
 t5 = t3;
 t3 = t4;
 t8 = t5;
 t8 = t3 ^ t8;
 t8 = a0 ^ t8;
 (*t2) = t8;
 t1 = (uint64_t (*)[3])&t0.state;
 t2 = (uint64_t *)&(*t1)[(uintptr_t)1ul];
 t9 = t0;
 memcpy(t10, t9.state, sizeof(uint64_t[3]));
 t8 = t10[(uintptr_t)0ul];
 (*t2) = t8;
 t1 = (uint64_t (*)[3])&t0.state;
 t2 = (uint64_t *)&(*t1)[(uintptr_t)2ul];
 t9 = t0;
 memcpy(t10, t9.state, sizeof(uint64_t[3]));
 t8 = t10[(uintptr_t)0ul];
 (*t2) = t8;
 t9 = t0;
 return t9;
}

static bool mem_eqlBytes__856(nav__856_39 const a0, nav__856_39 const a1) {
 nav__856_39 const *t1;
 nav__856_39 const *t3;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t17;
 uintptr_t t18;
 uintptr_t t16;
 uintptr_t t26;
 uint64_t t6;
 uint64_t t7;
 uint8_t const *t10;
 uint8_t const *t11;
 uintptr_t t19[4];
 nav__856_39 t21;
 nav__856_39 t0;
 nav__856_39 t2;
 uint8_t const (*t22)[4];
 uint8_t const (*t27)[16];
 uint32_t t20;
 uint32_t t24;
 uint32_t t25;
 uint32_t t15;
 bool t8;
 bool t9;
 uint8_t t12;
 uint8_t t13;
 uint8_t t14;
 uint8_t t23[4];
 uint8_t t28[16];
 uint8_t t29[16];
 uint8_t t30[16];
 bool t31[16];
 t0 = a0;
 t1 = (nav__856_39 const *)&t0;
 t2 = a1;
 t3 = (nav__856_39 const *)&t2;
 t4 = a0.len;
 t5 = a1.len;
 t6 = t4;
 t7 = t5;
 t8 = t6 != t7;
 if (t8) {
  return false;
 }
 goto zig_block_0;

 zig_block_0:;
 t5 = a0.len;
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = true;
  goto zig_block_2;
 }
 t10 = a0.ptr;
 t11 = a1.ptr;
 t8 = t10 == t11;
 t9 = t8;
 goto zig_block_2;

 zig_block_2:;
 if (t9) {
  return true;
 }
 goto zig_block_1;

 zig_block_1:;
 t5 = a0.len;
 t7 = t5;
 t9 = t7 <= UINT64_C(16);
 if (t9) {
  t5 = a0.len;
  t7 = t5;
  t9 = t7 < UINT64_C(4);
  if (t9) {
   t12 = a0.ptr[(uintptr_t)0ul];
   t13 = a1.ptr[(uintptr_t)0ul];
   t13 = t12 ^ t13;
   t5 = a0.len;
   t5 = t5 - (uintptr_t)1ul;
   t12 = a0.ptr[t5];
   t5 = a0.len;
   t5 = t5 - (uintptr_t)1ul;
   t14 = a1.ptr[t5];
   t14 = t12 ^ t14;
   t14 = t13 | t14;
   t5 = a0.len;
   t5 = t5 / (uintptr_t)2ul;
   t13 = a0.ptr[t5];
   t5 = a0.len;
   t5 = t5 / (uintptr_t)2ul;
   t12 = a1.ptr[t5];
   t12 = t13 ^ t12;
   t12 = t14 | t12;
   t9 = t12 == UINT8_C(0);
   return t9;
  }
  goto zig_block_4;

  zig_block_4:;
  t15 = UINT32_C(0);
  t16 = (uintptr_t)0ul;
  t5 = a0.len;
  t5 = t5 - (uintptr_t)4ul;
  t4 = a0.len;
  t4 = t4 / (uintptr_t)8ul;
  t4 = t4 * (uintptr_t)4ul;
  t17 = a0.len;
  t17 = t17 - (uintptr_t)4ul;
  t18 = a0.len;
  t18 = t18 / (uintptr_t)8ul;
  t18 = t18 * (uintptr_t)4ul;
  t18 = t17 - t18;
  t19[0] = (uintptr_t)0ul;
  t19[1] = t5;
  t19[2] = t4;
  t19[3] = t18;
  zig_loop_79:
  t18 = t16;
  t7 = t18;
  t9 = t7 < UINT64_C(4);
  if (t9) {
   t4 = t19[t18];
   t20 = t15;
   t21 = (*t1);
   t11 = t21.ptr;
   t11 = (uint8_t const *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
   t22 = (uint8_t const (*)[4])t11;
   memcpy(t23, (const char *)t22, sizeof(uint8_t[4]));
   memcpy(&t24, &t23, sizeof(uint32_t));
   t24 = zig_wrap_u32(t24, UINT8_C(32));
   t21 = (*t3);
   t11 = t21.ptr;
   t11 = (uint8_t const *)(((uintptr_t)t11) + (t4*sizeof(uint8_t)));
   t22 = (uint8_t const (*)[4])t11;
   memcpy(t23, (const char *)t22, sizeof(uint8_t[4]));
   memcpy(&t25, &t23, sizeof(uint32_t));
   t25 = zig_wrap_u32(t25, UINT8_C(32));
   t25 = t24 ^ t25;
   t25 = t20 | t25;
   t15 = t25;
   goto zig_block_6;
  }
  goto zig_block_5;

  zig_block_6:;
  t18 = t18 + (uintptr_t)1ul;
  t16 = t18;
  goto zig_loop_79;

  zig_block_5:;
  t25 = t15;
  t9 = t25 == UINT32_C(0);
  return t9;
 }
 goto zig_block_3;

 zig_block_3:;
 t26 = (uintptr_t)0ul;
 t18 = a0.len;
 t18 = t18 - (uintptr_t)1ul;
 t18 = t18 / (uintptr_t)16ul;
 zig_loop_132:
 t17 = t26;
 t7 = t17;
 t6 = t18;
 t9 = t7 < t6;
 if (t9) {
  t5 = t17 * (uintptr_t)16ul;
  t21 = (*t1);
  t11 = t21.ptr;
  t11 = (uint8_t const *)(((uintptr_t)t11) + (t5*sizeof(uint8_t)));
  t27 = (uint8_t const (*)[16])t11;
  memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
  memcpy(&t29, &t28, sizeof(uint8_t[16]));
  t5 = t17 * (uintptr_t)16ul;
  t21 = (*t3);
  t11 = t21.ptr;
  t11 = (uint8_t const *)(((uintptr_t)t11) + (t5*sizeof(uint8_t)));
  t27 = (uint8_t const (*)[16])t11;
  memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
  memcpy(&t30, &t28, sizeof(uint8_t[16]));
  for (t5 = (uintptr_t)0ul; t5 < (uintptr_t)16ul; t5 += (uintptr_t)1ul) {
   t31[t5] = t29[t5] != t30[t5];
  }
  t9 = false;
  for (t5 = (uintptr_t)0ul; t5 < (uintptr_t)16ul; t5 += (uintptr_t)1ul) {
   t9 |= t31[t5];
  }
  if (t9) {
   return false;
  }
  goto zig_block_9;

  zig_block_9:;
  goto zig_block_8;
 }
 goto zig_block_7;

 zig_block_8:;
 t17 = t17 + (uintptr_t)1ul;
 t26 = t17;
 goto zig_loop_132;

 zig_block_7:;
 t18 = a0.len;
 t18 = t18 - (uintptr_t)16ul;
 t21 = (*t1);
 t11 = t21.ptr;
 t11 = (uint8_t const *)(((uintptr_t)t11) + (t18*sizeof(uint8_t)));
 t27 = (uint8_t const (*)[16])t11;
 memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
 memcpy(&t30, &t28, sizeof(uint8_t[16]));
 t18 = a0.len;
 t18 = t18 - (uintptr_t)16ul;
 t21 = (*t3);
 t11 = t21.ptr;
 t11 = (uint8_t const *)(((uintptr_t)t11) + (t18*sizeof(uint8_t)));
 t27 = (uint8_t const (*)[16])t11;
 memcpy(t28, (const char *)t27, sizeof(uint8_t[16]));
 memcpy(&t29, &t28, sizeof(uint8_t[16]));
 for (t18 = (uintptr_t)0ul; t18 < (uintptr_t)16ul; t18 += (uintptr_t)1ul) {
  t31[t18] = t30[t18] != t29[t18];
 }
 t9 = false;
 for (t18 = (uintptr_t)0ul; t18 < (uintptr_t)16ul; t18 += (uintptr_t)1ul) {
  t9 |= t31[t18];
 }
 t9 = !t9;
 return t9;
}

static uint16_t environment_root_Environment_setIntegrationEnvironmentList__776(struct environment_root_Environment__559 *const a0, nav__776_41 const a1, nav__776_41 const a2, nav__776_43 const a3) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t0;
 struct mem_Allocator__599 *t3;
 struct mem_Allocator__599 t4;
 uintptr_t t5;
 uintptr_t t12;
 uintptr_t t11;
 uintptr_t t20;
 uintptr_t t24;
 nav__776_85 t6;
 nav__776_83 t8;
 nav__776_83 t16;
 nav__776_83 t9;
 nav__776_83 const *t10;
 uint64_t t13;
 uint64_t t14;
 nav__776_41 *t17;
 nav__776_41 t18;
 nav__776_90 t19;
 nav__776_88 t21;
 nav__776_43 t22;
 struct environment_root_EnvironmentValue__856 t23;
 uint16_t t7;
 bool t15;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (struct mem_Allocator__599 *)&t2->allocator;
 t4 = (*t3);
 t5 = a3.len;
 t6 = mem_Allocator_alloc__anon_7987__6470(t4, t5);
 if (t6.error) {
  t7 = t6.error;
  return t7;
 }
 t8 = t6.payload;
 t9 = t8;
 t10 = (nav__776_83 const *)&t9;
 t11 = (uintptr_t)0ul;
 zig_loop_25:
 t5 = t11;
 t12 = a3.len;
 t13 = t5;
 t14 = t12;
 t15 = t13 < t14;
 if (t15) {
  t12 = t11;
  t16 = (*t10);
  t17 = &t16.ptr[t12];
  t2 = (*t1);
  t3 = (struct mem_Allocator__599 *)&t2->allocator;
  t4 = (*t3);
  t12 = t11;
  t18 = a3.ptr[t12];
  t19 = mem_Allocator_dupe__anon_7358__6358(t4, t18);
  if (t19.error) {
   t7 = t19.error;
   t20 = (uintptr_t)0ul;
   t12 = t11;
   t16 = (*t10);
   t17 = t16.ptr;
   t17 = (nav__776_41 *)(((uintptr_t)t17) + ((uintptr_t)0ul*sizeof(nav__776_41)));
   t16.ptr = t17;
   t16.len = t12;
   t12 = t16.len;
   zig_loop_55:
   t5 = t20;
   t14 = t5;
   t13 = t12;
   t15 = t14 < t13;
   if (t15) {
    t18 = t16.ptr[t5];
    t2 = (*t1);
    t3 = (struct mem_Allocator__599 *)&t2->allocator;
    t4 = (*t3);
    mem_Allocator_free__anon_3116__5296(t4, t18);
    goto zig_block_3;
   }
   goto zig_block_2;

   zig_block_3:;
   t5 = t5 + (uintptr_t)1ul;
   t20 = t5;
   goto zig_loop_55;

   zig_block_2:;
   t2 = (*t1);
   t3 = (struct mem_Allocator__599 *)&t2->allocator;
   t4 = (*t3);
   mem_Allocator_free__anon_7990__6471(t4, t8);
   return t7;
  }
  t21 = t19.payload;
  memcpy(&t18, &t21, sizeof(nav__776_41));
  (*t17) = t18;
  t12 = t11;
  t12 = t12 + (uintptr_t)1ul;
  t11 = t12;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_25;

 zig_block_0:;
 t2 = (*t1);
 memcpy(&t22, &t8, sizeof(nav__776_43));
 t23.tag = UINT8_C(1);
 t23.payload.list = t22;
 t7 = environment_root_Environment_setIntegrationEnvironmentValue__778(t2, a1, a2, t23);
 if (t7) {
  t24 = (uintptr_t)0ul;
  t12 = t11;
  t16 = (*t10);
  t17 = t16.ptr;
  t17 = (nav__776_41 *)(((uintptr_t)t17) + ((uintptr_t)0ul*sizeof(nav__776_41)));
  t16.ptr = t17;
  t16.len = t12;
  t12 = t16.len;
  zig_loop_110:
  t5 = t24;
  t14 = t5;
  t13 = t12;
  t15 = t14 < t13;
  if (t15) {
   t18 = t16.ptr[t5];
   t2 = (*t1);
   t3 = (struct mem_Allocator__599 *)&t2->allocator;
   t4 = (*t3);
   mem_Allocator_free__anon_3116__5296(t4, t18);
   goto zig_block_5;
  }
  goto zig_block_4;

  zig_block_5:;
  t5 = t5 + (uintptr_t)1ul;
  t24 = t5;
  goto zig_loop_110;

  zig_block_4:;
  t2 = (*t1);
  t3 = (struct mem_Allocator__599 *)&t2->allocator;
  t4 = (*t3);
  mem_Allocator_free__anon_7990__6471(t4, t8);
  return t7;
 }
 return 0;
}

uint8_t c_instrument_hooks_set_environment_list__256(struct InstrumentHooks *const a0, char const *const a1, char const *const a2, char const *const *const a3, uint32_t const a4) {
 struct InstrumentHooks *t2;
 struct InstrumentHooks *t3;
 struct InstrumentHooks *const *t4;
 nav__256_47 t5;
 nav__256_47 t8;
 nav__256_47 t9;
 uintptr_t t6;
 uintptr_t t13;
 uintptr_t t12;
 nav__256_54 t7;
 nav__256_47 const *t10;
 uint8_t const *const *t11;
 uint64_t t14;
 uint64_t t15;
 uint8_t const *t16;
 nav__256_45 *t17;
 nav__256_45 t18;
 nav__256_45 t19;
 nav__256_45 t21;
 struct environment_root_Environment__559 *t20;
 nav__256_120 t22;
 uint16_t t23;
 bool t0;
 bool t1;
 t0 = a1 == NULL;
 if (t0) {
  t1 = true;
  goto zig_block_1;
 }
 t0 = a2 == NULL;
 t1 = t0;
 goto zig_block_1;

 zig_block_1:;
 if (t1) {
  t0 = true;
  goto zig_block_2;
 }
 t1 = a3 == NULL;
 t0 = t1;
 goto zig_block_2;

 zig_block_2:;
 if (t0) {
  return UINT8_C(1);
 }
 goto zig_block_0;

 zig_block_0:;
 t1 = a0 != NULL;
 if (t1) {
  t2 = a0;
  t3 = t2;
  t4 = (struct InstrumentHooks *const *)&t3;
  t6 = (uintptr_t)a4;
  t7 = mem_Allocator_alloc__anon_7987__6470((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)}, t6);
  t1 = t7.error == UINT16_C(0);
  if (t1) {
   t8 = t7.payload;
   t5 = t8;
   goto zig_block_4;
  }
  return UINT8_C(1);

  zig_block_4:;
  t9 = t5;
  t10 = (nav__256_47 const *)&t9;
  t11 = (uint8_t const *const *)a3;
  t12 = (uintptr_t)0ul;
  t6 = (uintptr_t)a4;
  zig_loop_45:
  t13 = t12;
  t14 = t13;
  t15 = t6;
  t1 = t14 < t15;
  if (t1) {
   t16 = t11[t13];
   t1 = t16 == NULL;
   if (t1) {
    mem_Allocator_free__anon_7990__6471((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)}, t5);
    return UINT8_C(1);
   }
   goto zig_block_7;

   zig_block_7:;
   t8 = (*t10);
   t17 = &t8.ptr[t13];
   t16 = t11[t13];
   t18 = mem_span__anon_6923__6298(t16);
   memcpy(&t19, &t18, sizeof(nav__256_45));
   (*t17) = t19;
   goto zig_block_6;
  }
  goto zig_block_5;

  zig_block_6:;
  t13 = t13 + (uintptr_t)1ul;
  t12 = t13;
  goto zig_loop_45;

  zig_block_5:;
  t2 = (*t4);
  t20 = (struct environment_root_Environment__559 *)&t2->environment;
  t16 = c_toU8__236(a1);
  t19 = mem_span__anon_6923__6298(t16);
  memcpy(&t18, &t19, sizeof(nav__256_45));
  t16 = c_toU8__236(a2);
  t19 = mem_span__anon_6923__6298(t16);
  memcpy(&t21, &t19, sizeof(nav__256_45));
  memcpy(&t22, &t5, sizeof(nav__256_120));
  t23 = environment_root_Environment_setIntegrationEnvironmentList__776(t20, t18, t21, t22);
  t1 = t23 == UINT16_C(0);
  if (t1) {
   goto zig_block_8;
  }
  mem_Allocator_free__anon_7990__6471((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)}, t5);
  return UINT8_C(1);

  zig_block_8:;
  mem_Allocator_free__anon_7990__6471((struct mem_Allocator__599){((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)}, t5);
  return UINT8_C(0);
 }
 goto zig_block_3;

 zig_block_3:;
 return UINT8_C(1);
}

static nav__6470_42 mem_Allocator_alloc__anon_7987__6470(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__6470_42 t4;
 nav__6470_42 t8;
 nav__6470_54 t6;
 uint8_t *t9;
 nav__6470_39 *t10;
 nav__6470_39 *t11;
 nav__6470_39 *const *t12;
 nav__6470_41 t13;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_7998__6472(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__6470_41){(nav__6470_39 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = (nav__6470_39 *)t9;
 t11 = t10;
 t12 = (nav__6470_39 *const *)&t11;
 t10 = (*t12);
 t10 = (nav__6470_39 *)(((uintptr_t)t10) + ((uintptr_t)0ul*sizeof(nav__6470_39)));
 t13.ptr = t10;
 t13.len = a1;
 t8.payload = t13;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static void mem_Allocator_free__anon_7990__6471(struct mem_Allocator__599 const a0, nav__6471_42 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__6471_52 t2;
 nav__6471_52 t11;
 uintptr_t t3;
 uint64_t t4;
 uint8_t *t6;
 uint8_t *t7;
 uint8_t *t8;
 uint8_t *const *t9;
 struct mem_Allocator__599 t10;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t13;
 struct mem_Allocator_VTable__602 const *const *t14;
 struct mem_Allocator_VTable__602 const *t15;
 void (*const *t16)(void *, nav__6471_52, uint8_t, uintptr_t);
 void (*t17)(void *, nav__6471_52, uint8_t, uintptr_t);
 void *t18;
 bool t5;
 uint8_t t12;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = mem_sliceAsBytes__anon_8005__6473(a1);
 t3 = t2.len;
 t4 = t3;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t6 = t2.ptr;
 t7 = (uint8_t *)t6;
 t8 = t7;
 t9 = (uint8_t *const *)&t8;
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 t10 = (*t1);
 t7 = (*t9);
 t7 = (uint8_t *)(((uintptr_t)t7) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t2.ptr = t7;
 t2.len = t3;
 memcpy(&t11, &t2, sizeof(nav__6471_52));
 t12 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t3 = (uintptr_t)zig_return_address();
 t13 = t10;
 t1 = (struct mem_Allocator__599 const *)&t13;
 t14 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t15 = (*t14);
 t16 = (void (*const *)(void *, nav__6471_52, uint8_t, uintptr_t))&t15->free;
 t17 = (*t16);
 t18 = t10.ptr;
 t17(t18, t11, t12, t3);
 return;
}

static nav__6472_39 mem_Allocator_allocWithSizeAndAlignment__anon_7998__6472(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__6472_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__6472_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3296__5445((uintptr_t)16ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__6472_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_2823__5103(t6, t2, a2);
 return t7;
}

static nav__6473_39 mem_sliceAsBytes__anon_8005__6473(nav__6473_43 const a0) {
 uintptr_t t0;
 uint64_t t1;
 nav__6473_41 *t4;
 uint8_t *t5;
 uint8_t *t6;
 uint8_t *const *t7;
 nav__6473_39 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__6473_39){(uint8_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (uint8_t *)t4;
 t6 = t5;
 t7 = (uint8_t *const *)&t6;
 t0 = a0.len;
 t0 = t0 * (uintptr_t)16ul;
 t5 = (*t7);
 t5 = (uint8_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

uint8_t c_instrument_hooks_write_environment__257(struct InstrumentHooks *const a0, int32_t const a1) {
 struct InstrumentHooks *t1;
 struct InstrumentHooks *t2;
 struct InstrumentHooks *const *t3;
 struct environment_root_Environment__559 *t4;
 bool t0;
 uint8_t t5;
 t0 = a0 != NULL;
 if (t0) {
  t1 = a0;
  t2 = t1;
  t3 = (struct InstrumentHooks *const *)&t2;
  t1 = (*t3);
  t4 = (struct environment_root_Environment__559 *)&t1->environment;
  t5 = environment_root_Environment_writeEnvironment__780(t4, a1);
  return t5;
 }
 goto zig_block_0;

 zig_block_0:;
 return UINT8_C(0);
}

static uint16_t environment_linked_libraries_root_LinkedLibraries_collect__3647(struct environment_linked_libraries_root_LinkedLibraries__849 *const a0) {
 struct environment_linked_libraries_root_LinkedLibraries__849 *const *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t3;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t4;
 nav__3647_50 t5;
 uintptr_t t6;
 uintptr_t t7;
 uintptr_t t2;
 uint64_t t8;
 uint64_t t9;
 struct environment_linked_libraries_root_LibraryInfo__946 t11;
 struct mem_Allocator__599 *t12;
 struct mem_Allocator__599 t13;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t14;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t0;
 void *t15;
 int t16;
 int32_t t17;
 bool t10;
 t0 = a0;
 t1 = (struct environment_linked_libraries_root_LinkedLibraries__849 *const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *)&a0->libraries;
 t4 = (*t3);
 t5 = t4.items;
 t6 = t5.len;
 zig_loop_13:
 t7 = t2;
 t8 = t7;
 t9 = t6;
 t10 = t8 < t9;
 if (t10) {
  t11 = t5.ptr[t7];
  t12 = (struct mem_Allocator__599 *)&a0->allocator;
  t13 = (*t12);
  environment_linked_libraries_root_LinkedLibraries_freeLibraryInfo__3645(t13, t11);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t7 = t7 + (uintptr_t)1ul;
 t2 = t7;
 goto zig_loop_13;

 zig_block_0:;
 t14 = (*t1);
 t3 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *)&t14->libraries;
 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_clearRetainingCapaci__4382(t3);
 t15 = (void *)a0;
 t16 = dl_iterate_phdr(&environment_linked_libraries_root_LinkedLibraries_collect__struct_8048_callback__6929, t15);
 t17 = t16;
 t10 = t17 != INT32_C(0);
 if (t10) {
  return zig_error_DlIteratePhdrFailed;
 }
 goto zig_block_2;

 zig_block_2:;
 return 0;
}

static uint16_t environment_root_Environment_populateLinkedLibraries__779(struct environment_root_Environment__559 *const a0) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t3;
 struct environment_root_Environment__559 *t0;
 struct environment_root_EnvironmentJson__845 *t4;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *t5;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t6;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t2;
 nav__779_87 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213 t11;
 struct mem_Allocator__599 *t12;
 struct mem_Allocator__599 t13;
 nav__779_44 *t14;
 nav__779_44 t15;
 nav__779_44 t26;
 nav__779_44 t27;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t17;
 struct environment_linked_libraries_root_LinkedLibraries__849 t18;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t19;
 nav__779_76 t20;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t16;
 uint64_t t23;
 uint64_t t24;
 struct environment_linked_libraries_root_LibraryInfo__946 t25;
 nav__779_95 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t30;
 nav__779_99 t31;
 nav__779_97 t32;
 struct environment_linked_libraries_root_LibraryEntry__864 *t33;
 uint16_t t29;
 bool t10;
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *)&t4->linked_libraries;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *)&t5->map;
 t7 = (*t6);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4190(t7);
 t2 = t8;
 zig_loop_14:
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6165(&t2);
 t10 = t9.is_null != true;
 if (t10) {
  t11 = t9.payload;
  t3 = (*t1);
  t12 = (struct mem_Allocator__599 *)&t3->allocator;
  t13 = (*t12);
  t14 = t11.key_ptr;
  t15 = (*t14);
  mem_Allocator_free__anon_3116__5296(t13, t15);
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_14;

 zig_block_0:;
 t3 = (*t1);
 t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
 t5 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *)&t4->linked_libraries;
 t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *)&t5->map;
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4185(t6);
 t16 = (uintptr_t)0ul;
 t17 = (struct environment_linked_libraries_root_LinkedLibraries__849 *)&a0->libs;
 t18 = (*t17);
 t19 = t18.libraries;
 t20 = t19.items;
 t21 = t20.len;
 zig_loop_46:
 t22 = t16;
 t23 = t22;
 t24 = t21;
 t10 = t23 < t24;
 if (t10) {
  t25 = t20.ptr[t22];
  t26 = t25.soname;
  t10 = t26.ptr != NULL;
  if (t10) {
   t27 = t26;
   t15 = t27;
   goto zig_block_4;
  }
  t26 = t25.path;
  t15 = t26;
  goto zig_block_4;

  zig_block_4:;
  t3 = (*t1);
  t4 = (struct environment_root_EnvironmentJson__845 *)&t3->data;
  t5 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *)&t4->linked_libraries;
  t6 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *)&t5->map;
  t12 = (struct mem_Allocator__599 *)&a0->allocator;
  t13 = (*t12);
  t28 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4192(t6, t13, t15);
  if (t28.error) {
   t29 = t28.error;
   return t29;
  }
  t30 = t28.payload;
  t14 = t30.key_ptr;
  t3 = (*t1);
  t12 = (struct mem_Allocator__599 *)&t3->allocator;
  t13 = (*t12);
  t31 = mem_Allocator_dupe__anon_7358__6358(t13, t15);
  if (t31.error) {
   t29 = t31.error;
   return t29;
  }
  t32 = t31.payload;
  memcpy(&t15, &t32, sizeof(nav__779_44));
  (*t14) = t15;
  t33 = t30.value_ptr;
  t14 = (nav__779_44 *)&t33->path;
  t15 = t25.path;
  (*t14) = t15;
  t14 = (nav__779_44 *)&t33->build_id;
  t15 = t25.build_id;
  (*t14) = t15;
  goto zig_block_3;
 }
 goto zig_block_2;

 zig_block_3:;
 t22 = t22 + (uintptr_t)1ul;
 t16 = t22;
 goto zig_loop_46;

 zig_block_2:;
 return 0;
}

static uint8_t environment_root_Environment_writeEnvironment__780(struct environment_root_Environment__559 *const a0, int32_t const a1) {
 struct environment_root_Environment__559 *const *t1;
 struct environment_root_Environment__559 *t2;
 struct environment_root_Environment__559 *t0;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t3;
 struct environment_root_EnvironmentJson__845 *t6;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t9;
 uintptr_t t10;
 uint64_t t11;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *t13;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t14;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t15;
 uint8_t const *t16;
 uint8_t const *t17;
 uint8_t const *t18;
 nav__780_83 t19;
 nav__780_86 t21;
 nav__780_86 t22;
 nav__780_86 t28;
 nav__780_89 t23;
 nav__780_91 t24;
 struct mem_Allocator__599 *t25;
 struct mem_Allocator__599 t26;
 struct environment_root_EnvironmentJson__845 t27;
 struct fs_File__799 const *t33;
 struct fs_File__799 t29;
 struct fs_File__799 t31;
 struct fs_File__799 t32;
 nav__780_100 t30;
 uint16_t t4;
 bool t5;
 bool t12;
 uint8_t t20[512];
 t0 = a0;
 t1 = (struct environment_root_Environment__559 *const *)&t0;
 t2 = (*t1);
 t3 = (struct environment_linked_libraries_root_LinkedLibraries__849 *)&t2->libs;
 t4 = environment_linked_libraries_root_LinkedLibraries_collect__3647(t3);
 t5 = t4 == UINT16_C(0);
 if (t5) {
  goto zig_block_0;
 }
 logger_err__anon_8063__6930();
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t4 = environment_root_Environment_populateLinkedLibraries__779(t2);
 t5 = t4 == UINT16_C(0);
 if (t5) {
  goto zig_block_1;
 }
 logger_err__anon_8090__6931();
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct environment_root_EnvironmentJson__845 *)&t2->data;
 t7 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 *)&t6->integration_environment;
 t8 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 *)&t7->map;
 t9 = (*t8);
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3797(t9);
 t11 = t10;
 t5 = t11 == UINT64_C(0);
 if (t5) {
  t2 = (*t1);
  t6 = (struct environment_root_EnvironmentJson__845 *)&t2->data;
  t13 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 *)&t6->linked_libraries;
  t14 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *)&t13->map;
  t15 = (*t14);
  t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4187(t15);
  t11 = t10;
  t5 = t11 == UINT64_C(0);
  t12 = t5;
  goto zig_block_3;
 }
 t12 = false;
 goto zig_block_3;

 zig_block_3:;
 if (t12) {
  return UINT8_C(0);
 }
 goto zig_block_2;

 zig_block_2:;
 t17 = getenv((uint8_t const *)&__anon_8103);
 t12 = t17 != NULL;
 if (t12) {
  t18 = t17;
  t16 = t18;
  goto zig_block_4;
 }
 return UINT8_C(0);

 zig_block_4:;
 t19 = mem_span__anon_8113__6933(t16);
 t22.ptr = &t20[(uintptr_t)0ul];
 t22.len = (uintptr_t)512ul;
 t23.f0 = t19;
 t23.f1 = a1;
 t24 = fmt_bufPrint__anon_8122__6934(t22, t23);
 t12 = t24.error == UINT16_C(0);
 if (t12) {
  t22 = t24.payload;
  t21 = t22;
  goto zig_block_5;
 }
 logger_err__anon_8129__6935();
 return UINT8_C(1);

 zig_block_5:;
 t25 = (struct mem_Allocator__599 *)&a0->allocator;
 t26 = (*t25);
 t6 = (struct environment_root_EnvironmentJson__845 *)&a0->data;
 t27 = (*t6);
 t24 = json_stringify_stringifyAlloc__anon_8144__6957(t26, t27, (struct json_stringify_StringifyOptions__8133){UINT8_C(2),true,false,false,false});
 t12 = t24.error == UINT16_C(0);
 if (t12) {
  t28 = t24.payload;
  t22 = t28;
  goto zig_block_6;
 }
 logger_err__anon_8151__6958();
 return UINT8_C(1);

 zig_block_6:;
 memcpy(&t19, &t21, sizeof(nav__780_83));
 t30 = fs_createFileAbsolute__1351(t19, (struct fs_File_CreateFlags__8153){(uintptr_t)438ul,false,true,false,UINT8_C(0),false});
 t12 = t30.error == UINT16_C(0);
 if (t12) {
  t31 = t30.payload;
  t29 = t31;
  goto zig_block_7;
 }
 logger_err__anon_8170__6959();
 t2 = (*t1);
 t25 = (struct mem_Allocator__599 *)&t2->allocator;
 t26 = (*t25);
 mem_Allocator_free__anon_3705__5702(t26, t22);
 return UINT8_C(1);

 zig_block_7:;
 t32 = t29;
 t33 = (struct fs_File__799 const *)&t32;
 t29 = (*t33);
 memcpy(&t19, &t22, sizeof(nav__780_83));
 t4 = fs_File_writeAll__1439(t29, t19);
 t12 = t4 == UINT16_C(0);
 if (t12) {
  goto zig_block_8;
 }
 logger_err__anon_8170__6959();
 t29 = (*t33);
 fs_File_close__1385(t29);
 t2 = (*t1);
 t25 = (struct mem_Allocator__599 *)&t2->allocator;
 t26 = (*t25);
 mem_Allocator_free__anon_3705__5702(t26, t22);
 return UINT8_C(1);

 zig_block_8:;
 t29 = (*t33);
 fs_File_close__1385(t29);
 t2 = (*t1);
 t25 = (struct mem_Allocator__599 *)&t2->allocator;
 t26 = (*t25);
 mem_Allocator_free__anon_3705__5702(t26, t22);
 return UINT8_C(0);
}

static void array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_clearRetainingCapaci__4382(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t2;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t0;
 nav__4382_44 *t3;
 uintptr_t *t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *)&t0;
 t2 = (*t1);
 t3 = (nav__4382_44 *)&t2->items;
 t4 = &t3->len;
 (*t4) = (uintptr_t)0ul;
 return;
}

static int environment_linked_libraries_root_LinkedLibraries_collect__struct_8048_callback__6929(struct os_linux_dl_phdr_info__8025 *const a0, uintptr_t const a1, void *const a2) {
 struct environment_linked_libraries_root_LinkedLibraries__849 *t0;
 struct environment_linked_libraries_root_LinkedLibraries__849 *t1;
 struct environment_linked_libraries_root_LinkedLibraries__849 *const *t2;
 uint8_t const *t3;
 uint8_t const *t5;
 uint8_t const *t7;
 uint8_t const **t4;
 nav__6929_50 t8;
 nav__6929_50 t11;
 uintptr_t t9;
 uint64_t t10;
 uintptr_t *t12;
 struct elf_Elf64_Phdr__8033 **t13;
 struct elf_Elf64_Phdr__8033 *t14;
 struct elf_Elf64_Phdr__8033 const *t15;
 uint16_t *t16;
 struct environment_linked_libraries_elf_view__8184 t18;
 struct environment_linked_libraries_root_LibraryInfo__946 t19;
 struct environment_linked_libraries_root_LibraryInfo__946 t23;
 struct mem_Allocator__599 *t20;
 struct mem_Allocator__599 t21;
 nav__6929_75 t22;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t24;
 uint16_t t17;
 bool t6;
 (void)a1;
 t0 = (struct environment_linked_libraries_root_LinkedLibraries__849 *)a2;
 t1 = t0;
 t2 = (struct environment_linked_libraries_root_LinkedLibraries__849 *const *)&t1;
 t4 = (uint8_t const **)&a0->name;
 t5 = (*t4);
 t6 = t5 != NULL;
 if (t6) {
  t7 = t5;
  t3 = t7;
  goto zig_block_0;
 }
 return 0;

 zig_block_0:;
 t8 = mem_span__anon_8113__6933(t3);
 t9 = t8.len;
 t10 = t9;
 t6 = t10 == UINT64_C(0);
 if (t6) {
  return 0;
 }
 goto zig_block_1;

 zig_block_1:;
 memcpy(&t11, &t8, sizeof(nav__6929_50));
 t6 = mem_startsWith__anon_8183__6960(t11, (nav__6929_50){(uint8_t const *)&__anon_8178,(uintptr_t)10ul});
 if (t6) {
  return 0;
 }
 goto zig_block_2;

 zig_block_2:;
 t12 = (uintptr_t *)&a0->addr;
 t9 = (*t12);
 t13 = (struct elf_Elf64_Phdr__8033 **)&a0->phdr;
 t14 = (*t13);
 t15 = (struct elf_Elf64_Phdr__8033 const *)t14;
 t16 = (uint16_t *)&a0->phnum;
 t17 = (*t16);
 t18 = environment_linked_libraries_elf_view_init__6966(t9, t15, t17);
 t20 = (struct mem_Allocator__599 *)&t0->allocator;
 t21 = (*t20);
 memcpy(&t11, &t8, sizeof(nav__6929_50));
 t22 = environment_linked_libraries_root_LinkedLibraries_extractLibraryInfo__3646(t21, t18, t11);
 t6 = t22.is_null != true;
 if (t6) {
  t23 = t22.payload;
  t19 = t23;
  goto zig_block_3;
 }
 return -1;

 zig_block_3:;
 t0 = (*t2);
 t24 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *)&t0->libraries;
 t17 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_append__4363(t24, t19);
 t6 = t17 == UINT16_C(0);
 if (t6) {
  goto zig_block_4;
 }
 return -1;

 zig_block_4:;
 return 0;
}

static void logger_err__anon_8063__6930(void) {
 logger_logWithPrefix__anon_8206__6987();
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4185(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 void *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t4;
 uintptr_t *t5;
 struct array_hash_map_IndexHeader__892 **t6;
 struct array_hash_map_IndexHeader__892 *t7;
 struct array_hash_map_IndexHeader__892 *t9;
 struct array_hash_map_IndexHeader__892 *t10;
 struct array_hash_map_IndexHeader__892 *const *t11;
 nav__4185_57 t14;
 uintptr_t t15;
 nav__4185_62 t16;
 nav__4185_67 t17;
 struct array_hash_map_IndexHeader__892 t12;
 bool t8;
 uint8_t t13;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3982(t3);
 t2 = (*t1);
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t2->entries;
 t5 = (uintptr_t *)&t4->len;
 (*t5) = (uintptr_t)0ul;
 t6 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t7 = (*t6);
 t8 = t7 != NULL;
 if (t8) {
  t9 = t7;
  t10 = t9;
  t11 = (struct array_hash_map_IndexHeader__892 *const *)&t10;
  t9 = (*t11);
  t12 = (*t9);
  t13 = array_hash_map_IndexHeader_capacityIndexType__3973(t12);
  switch (t13) {
   case UINT8_C(0): {
    t9 = (*t11);
    t14 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t9);
    for (t15 = (uintptr_t)0ul; t15 != t14.len; ++t15) ((struct array_hash_map_Index_28u8_29__6735 *)t14.ptr)[t15] = (struct array_hash_map_Index_28u8_29__6735){UINT8_MAX,UINT8_C(0xaa)};
    goto zig_block_1;
   }
   case UINT8_C(1): {
    t9 = (*t11);
    t16 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t9);
    for (t15 = (uintptr_t)0ul; t15 != t16.len; ++t15) ((struct array_hash_map_Index_28u16_29__6739 *)t16.ptr)[t15] = (struct array_hash_map_Index_28u16_29__6739){UINT16_MAX,UINT16_C(0xaaaa)};
    goto zig_block_1;
   }
   case UINT8_C(2): {
    t9 = (*t11);
    t17 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t9);
    for (t15 = (uintptr_t)0ul; t15 != t17.len; ++t15) ((struct array_hash_map_Index_28u32_29__6743 *)t17.ptr)[t15] = (struct array_hash_map_Index_28u32_29__6743){UINT32_MAX,UINT32_C(0xaaaaaaaa)};
    goto zig_block_1;
   }
   default: zig_unreachable();
  }

  zig_block_1:;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3983(t3);
 return;
}

static nav__4192_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4192(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct mem_Allocator__599 const a1, nav__4192_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 nav__4192_39 t3;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4193(t2, a1, a2);
 return t3;
}

static void logger_err__anon_8090__6931(void) {
 logger_logWithPrefix__anon_8218__6988();
 return;
}

static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3797(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Data_29__889 t0;
 uintptr_t t1;
 t0 = a0.entries;
 t1 = t0.len;
 return t1;
}

static uintptr_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4187(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t0;
 uintptr_t t1;
 t0 = a0.entries;
 t1 = t0.len;
 return t1;
}

static nav__6933_39 mem_span__anon_8113__6933(uint8_t const *const a0) {
 uint8_t const *const *t1;
 uintptr_t t2;
 uint8_t const *t3;
 uint8_t const *t0;
 nav__6933_39 t4;
 t0 = a0;
 t1 = (uint8_t const *const *)&t0;
 t2 = mem_len__anon_8219__6989(a0);
 t3 = (*t1);
 t3 = (uint8_t const *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t3;
 t4.len = t2;
 return t4;
}

static uint16_t fmt_format__anon_8220__6990(struct io_Writer__3878 const a0, nav__6990_41 const a1) {
 struct io_Writer__3878 const *t1;
 nav__6990_40 t2;
 struct io_Writer__3878 t4;
 struct io_Writer__3878 t0;
 int32_t t5;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = a1.f0;
 t3 = fmt_formatType__anon_4782__6113(t2, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t4 = (*t1);
 t3 = io_Writer_writeAll__5732(t4, (nav__6990_40){(uint8_t const *)&__anon_8255,(uintptr_t)13ul});
 if (t3) {
  return t3;
 }
 t5 = a1.f1;
 t3 = fmt_formatType__anon_5471__6140(t5, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t3) {
  return t3;
 }
 t4 = (*t1);
 t3 = io_Writer_writeAll__5732(t4, (nav__6990_40){(uint8_t const *)&__anon_8272,(uintptr_t)5ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__6934_40 fmt_bufPrint__anon_8122__6934(nav__6934_39 const a0, nav__6934_43 const a1) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__6934_59 (**t11)(void const *, nav__6934_42);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__6934_39 t15;
 nav__6934_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__6934_59 (**)(void const *, nav__6934_42))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_8220__6990(t12, a1);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__6934_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static void logger_err__anon_8129__6935(void) {
 logger_logWithPrefix__anon_8278__6991();
 return;
}

static nav__6957_40 json_stringify_stringifyAlloc__anon_8144__6957(struct mem_Allocator__599 const a0, struct environment_root_EnvironmentJson__845 const a1, struct json_stringify_StringifyOptions__8133 const a2) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 nav__6957_40 t4;
 uint16_t t3;
 bool t5;
 t1 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(a0);
 t0 = t1;
 t2 = array_list_ArrayListAligned_28u8_2cnull_29_writer__1552(&t0);
 t3 = json_stringify_stringifyArbitraryDepth__anon_8305__7021(a0, a1, a2, t2);
 if (t3) {
  t1 = t0;
  array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t1);
  t4.payload = (nav__6957_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t4.error = t3;
  return t4;
 }
 t4 = array_list_ArrayListAligned_28u8_2cnull_29_toOwnedSlice__1533(&t0);
 t5 = t4.error == UINT16_C(0);
 if (t5) {
  return t4;
 }
 t1 = t0;
 array_list_ArrayListAligned_28u8_2cnull_29_deinit__1529(t1);
 return t4;
}

static void logger_err__anon_8151__6958(void) {
 logger_logWithPrefix__anon_8310__7022();
 return;
}

static nav__1351_39 fs_createFileAbsolute__1351(nav__1351_41 const a0, struct fs_File_CreateFlags__8153 const a1) {
 struct fs_Dir__1720 const *t3;
 struct fs_Dir__1720 t1;
 struct fs_Dir__1720 t2;
 nav__1351_39 t4;
 bool t0;
 t0 = fs_path_isAbsolute__5138(a0);
 debug_assert__180(t0);
 t1 = fs_cwd__1340();
 t2 = t1;
 t3 = (struct fs_Dir__1720 const *)&t2;
 t1 = (*t3);
 t4 = fs_Dir_createFile__4681(t1, a0, a1);
 return t4;
}

static void logger_err__anon_8170__6959(void) {
 logger_logWithPrefix__anon_8315__7023();
 return;
}

static bool mem_startsWith__anon_8183__6960(nav__6960_39 const a0, nav__6960_39 const a1) {
 nav__6960_39 const *t1;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 nav__6960_39 t8;
 nav__6960_39 t0;
 uint8_t const *t9;
 bool t2;
 bool t7;
 t0 = a0;
 t1 = (nav__6960_39 const *)&t0;
 t3 = a1.len;
 t4 = a0.len;
 t5 = t3;
 t6 = t4;
 t7 = t5 > t6;
 if (t7) {
  t2 = false;
  goto zig_block_0;
 }
 t4 = a1.len;
 t8 = (*t1);
 t9 = t8.ptr;
 t9 = (uint8_t const *)(((uintptr_t)t9) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t8.ptr = t9;
 t8.len = t4;
 t7 = mem_eql__anon_7887__6466(t8, a1);
 t2 = t7;
 goto zig_block_0;

 zig_block_0:;
 return t2;
}

static struct environment_linked_libraries_elf_view__8184 environment_linked_libraries_elf_view_init__6966(uintptr_t const a0, struct elf_Elf64_Phdr__8033 const *const a1, uint16_t const a2) {
 uintptr_t *t1;
 nav__6966_42 *t2;
 struct elf_Elf64_Phdr__8033 const *const *t4;
 struct elf_Elf64_Phdr__8033 const *t5;
 struct elf_Elf64_Phdr__8033 const *t3;
 uintptr_t t6;
 nav__6966_42 t7;
 struct environment_linked_libraries_elf_view__8184 t0;
 t1 = (uintptr_t *)&t0.base_addr;
 (*t1) = a0;
 t2 = (nav__6966_42 *)&t0.phdrs;
 t3 = a1;
 t4 = (struct elf_Elf64_Phdr__8033 const *const *)&t3;
 t5 = (*t4);
 t5 = (struct elf_Elf64_Phdr__8033 const *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct elf_Elf64_Phdr__8033)));
 t6 = (uintptr_t)a2;
 t7.ptr = t5;
 t7.len = t6;
 (*t2) = t7;
 return t0;
}

static nav__3646_39 environment_linked_libraries_root_LinkedLibraries_extractLibraryInfo__3646(struct mem_Allocator__599 const a0, struct environment_linked_libraries_elf_view__8184 const a1, nav__3646_43 const a2) {
 struct mem_Allocator__599 const *t1;
 struct environment_linked_libraries_elf_view__8184 const *t3;
 nav__3646_43 t4;
 nav__3646_43 t6;
 nav__3646_43 t8;
 struct environment_linked_libraries_elf_view__8184 t5;
 struct environment_linked_libraries_elf_view__8184 t2;
 nav__3646_62 t9;
 nav__3646_62 t10;
 nav__3646_62 t11;
 struct environment_linked_libraries_root_LibraryInfo__946 *t13;
 nav__3646_43 *t14;
 struct mem_Allocator__599 t15;
 struct mem_Allocator__599 t0;
 nav__3646_69 t16;
 nav__3646_67 t17;
 nav__3646_67 t21;
 nav__3646_67 t22;
 uint8_t const *t18;
 uint8_t const *t19;
 uint8_t const *t20;
 nav__3646_39 t12;
 nav__3646_39 t23;
 bool t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = (struct environment_linked_libraries_elf_view__8184 const *)&t2;
 t5 = (*t3);
 t6 = environment_linked_libraries_elf_view_buildId__6975(t5);
 t7 = t6.ptr != NULL;
 if (t7) {
  t8 = t6;
  t8 = environment_linked_libraries_root_hexEncode__3638(a0, t8);
  t4 = t8;
  goto zig_block_0;
 }
 t4 = (nav__3646_43){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 goto zig_block_0;

 zig_block_0:;
 t5 = (*t3);
 t10 = environment_linked_libraries_elf_view_dynamicEntries__6970(t5);
 t7 = t10.ptr != NULL;
 if (t7) {
  t11 = t10;
  t9 = t11;
  goto zig_block_1;
 }
 t12.is_null = false;
 t13 = &t12.payload;
 t14 = (nav__3646_43 *)&t13->path;
 t15 = (*t1);
 t16 = mem_Allocator_dupe__anon_7358__6358(t15, a2);
 t7 = t16.error == UINT16_C(0);
 if (t7) {
  t17 = t16.payload;
  memcpy(&t6, &t17, sizeof(nav__3646_43));
  t8 = t6;
  goto zig_block_2;
 }
 return (nav__3646_39){{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}},true};

 zig_block_2:;
 (*t14) = t8;
 t14 = (nav__3646_43 *)&t13->soname;
 (*t14) = (nav__3646_43){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 t14 = (nav__3646_43 *)&t13->build_id;
 (*t14) = t4;
 return t12;

 zig_block_1:;
 t19 = environment_linked_libraries_elf_view_strtab__6972(t9);
 t7 = t19 != NULL;
 if (t7) {
  t20 = t19;
  t18 = t20;
  goto zig_block_3;
 }
 return (nav__3646_39){{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}},true};

 zig_block_3:;
 t8 = environment_linked_libraries_elf_view_soname__6974(t9, t18);
 t7 = t8.ptr != NULL;
 if (t7) {
  t6 = t8;
  t15 = (*t1);
  t16 = mem_Allocator_dupe__anon_7358__6358(t15, t6);
  t7 = t16.error == UINT16_C(0);
  if (t7) {
   t22 = t16.payload;
   t21 = t22;
   goto zig_block_5;
  }
  return (nav__3646_39){{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}},true};

  zig_block_5:;
  t22 = t21;
  t17 = t22;
  goto zig_block_4;

 }
 t17 = (nav__3646_67){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 goto zig_block_4;

 zig_block_4:;
 t23.is_null = false;
 t13 = &t23.payload;
 t14 = (nav__3646_43 *)&t13->path;
 t15 = (*t1);
 t16 = mem_Allocator_dupe__anon_7358__6358(t15, a2);
 t7 = t16.error == UINT16_C(0);
 if (t7) {
  t22 = t16.payload;
  memcpy(&t6, &t22, sizeof(nav__3646_43));
  t8 = t6;
  goto zig_block_7;
 }
 return (nav__3646_39){{{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},{(uint8_t const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul}},true};

 zig_block_7:;
 (*t14) = t8;
 t14 = (nav__3646_43 *)&t13->soname;
 memcpy(&t8, &t17, sizeof(nav__3646_43));
 (*t14) = t8;
 t14 = (nav__3646_43 *)&t13->build_id;
 (*t14) = t4;
 return t23;
}

static uint16_t array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_append__4363(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const a0, struct environment_linked_libraries_root_LibraryInfo__946 const a1) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *t1;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t2;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t0;
 nav__4363_49 t3;
 struct environment_linked_libraries_root_LibraryInfo__946 *t5;
 uint16_t t4;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *)&t0;
 t2 = (*t1);
 t3 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_addOne__4388(t2);
 if (t3.error) {
  t4 = t3.error;
  return t4;
 }
 t5 = t3.payload;
 (*t5) = a1;
 return 0;
}

static void logger_logWithPrefix__anon_8206__6987(void) {
 nav__6987_41 t3;
 nav__6987_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6987_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8364__7024(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static nav__4193_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4193(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct mem_Allocator__599 const a1, nav__4193_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 nav__4193_39 t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t5;
 nav__4193_44 *t7;
 uint16_t t4;
 bool t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = (*t1);
 t3 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7025(t2, a1, a2);
 if (t3.error) {
  t4 = t3.error;
  t3.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075){((nav__4193_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_linked_libraries_root_LibraryEntry__864 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
  t3.error = t4;
  return t3;
 }
 t5 = t3.payload;
 t6 = t5.found_existing;
 t6 = !t6;
 if (t6) {
  t7 = t5.key_ptr;
  (*t7) = a2;
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3.payload = t5;
 t3.error = UINT16_C(0);
 return t3;
}

static void logger_logWithPrefix__anon_8218__6988(void) {
 nav__6988_41 t3;
 nav__6988_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6988_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8381__7026(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static uintptr_t mem_len__anon_8219__6989(uint8_t const *const a0) {
 uintptr_t t0;
 t0 = mem_indexOfSentinel__anon_6947__6300(a0);
 return t0;
}

static void logger_logWithPrefix__anon_8278__6991(void) {
 nav__6991_41 t3;
 nav__6991_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__6991_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8390__7027(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static uint16_t json_stringify_stringifyArbitraryDepth__anon_8305__7021(struct mem_Allocator__599 const a0, struct environment_root_EnvironmentJson__845 const a1, struct json_stringify_StringifyOptions__8133 const a2, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a3) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 t0;
 uint16_t t2;
 t1 = json_stringify_writeStreamArbitraryDepth__anon_8398__7028(a0, a3, a2);
 t0 = t1;
 t2 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7029(&t0, a1);
 if (t2) {
  json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6998(&t0);
  return t2;
 }
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6998(&t0);
 return 0;
}

static nav__1533_40 array_list_ArrayListAligned_28u8_2cnull_29_toOwnedSlice__1533(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct mem_Allocator__599 *t2;
 struct mem_Allocator__599 t3;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t4;
 struct mem_Allocator__599 const *t5;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t6;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t7;
 nav__1533_39 t8;
 nav__1533_39 t11;
 nav__1533_39 *t10;
 uintptr_t t12;
 nav__1533_40 t14;
 uint8_t *t16;
 uint16_t t15;
 bool t13;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (struct mem_Allocator__599 *)&a0->allocator;
 t3 = (*t2);
 t4 = t3;
 t5 = (struct mem_Allocator__599 const *)&t4;
 t6 = (*t1);
 t7 = (*t6);
 t8 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t7);
 t9 = (*t5);
 t10 = (nav__1533_39 *)&a0->items;
 t11 = (*t10);
 t12 = t11.len;
 t8 = mem_Allocator_remap__anon_3698__5700(t9, t8, t12);
 t13 = t8.ptr != NULL;
 if (t13) {
  t11 = t8;
  t7 = array_list_ArrayListAligned_28u8_2cnull_29_init__1527(t3);
  (*a0) = t7;
  t14.payload = t11;
  t14.error = UINT16_C(0);
  return t14;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t5);
 t10 = (nav__1533_39 *)&a0->items;
 t8 = (*t10);
 t12 = t8.len;
 t14 = mem_Allocator_alignedAlloc__anon_3703__5701(t3, t12);
 if (t14.error) {
  t15 = t14.error;
  t14.payload = (nav__1533_39){(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t14.error = t15;
  return t14;
 }
 t8 = t14.payload;
 t10 = (nav__1533_39 *)&a0->items;
 t11 = (*t10);
 t16 = t11.ptr;
 if (t8.len != 0) memcpy(t8.ptr, t16, t8.len * sizeof(uint8_t));
 t6 = (*t1);
 array_list_ArrayListAligned_28u8_2cnull_29_clearAndFree__1563(t6);
 t14.payload = t8;
 t14.error = UINT16_C(0);
 return t14;
}

static void logger_logWithPrefix__anon_8310__7022(void) {
 nav__7022_41 t3;
 nav__7022_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__7022_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8421__7030(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static nav__4681_39 fs_Dir_createFile__4681(struct fs_Dir__1720 const a0, nav__4681_42 const a1, struct fs_File_CreateFlags__8153 const a2) {
 struct fs_Dir__1720 const *t1;
 uint8_t const (*t7)[4096];
 uint8_t const *t9;
 nav__4681_39 t4;
 struct fs_Dir__1720 t8;
 struct fs_Dir__1720 t0;
 nav__4681_53 t2;
 uint16_t t3;
 uint8_t t5[4096];
 uint8_t t6[4096];
 t0 = a0;
 t1 = (struct fs_Dir__1720 const *)&t0;
 t2 = posix_toPosixPath__2022(a1);
 if (t2.error) {
  t3 = t2.error;
  t4.payload = (struct fs_File__799){-INT32_C(0x55555556)};
  t4.error = t3;
  return t4;
 }
 memcpy(t5, t2.payload, sizeof(uint8_t[4096]));
 memcpy((char *)&t6, t5, sizeof(uint8_t[4096]));
 t7 = (uint8_t const (*)[4096])&t6;
 t8 = (*t1);
 t9 = (uint8_t const *)t7;
 t4 = fs_Dir_createFileZ__4682(t8, t9, a2);
 return t4;
}

static void logger_logWithPrefix__anon_8315__7023(void) {
 nav__7023_41 t3;
 nav__7023_41 t6;
 uint8_t *t4;
 uint8_t (*t5)[511];
 nav__7023_45 t7;
 uintptr_t t8;
 char const *t9;
 uint8_t t0;
 bool t1;
 uint8_t t2[512];
 t0 = (*&logger_max_level__5093);
 t1 = UINT8_C(3) < t0;
 if (t1) {
  return;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = (uint8_t *)&t2;
 t4 = (uint8_t *)(((uintptr_t)t4) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t5 = (uint8_t (*)[511])t4;
 t6.ptr = &(*t5)[(uintptr_t)0ul];
 t6.len = (uintptr_t)511ul;
 t7 = fmt_bufPrint__anon_8436__7031(t6);
 t1 = t7.error == UINT16_C(0);
 if (t1) {
  t6 = t7.payload;
  t3 = t6;
  goto zig_block_1;
 }
 (void)printf((char const *)&__anon_3484);
 return;

 zig_block_1:;
 t8 = t3.len;
 t4 = (uint8_t *)&t2[t8];
 (*t4) = UINT8_C(0);
 t4 = t3.ptr;
 t9 = (char const *)t4;
 (void)printf(t9);
 return;
}

static nav__6975_39 environment_linked_libraries_elf_view_buildId__6975(struct environment_linked_libraries_elf_view__8184 const a0) {
 struct environment_linked_libraries_elf_view__8184 const *t1;
 nav__6975_45 t3;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t2;
 uint64_t t6;
 uint64_t t7;
 struct elf_Elf64_Phdr__8033 t9;
 struct environment_linked_libraries_elf_view__8184 t11;
 struct environment_linked_libraries_elf_view__8184 t0;
 nav__6975_39 t14;
 nav__6975_39 t21;
 nav__6975_39 t15;
 nav__6975_39 const *t16;
 uint8_t const *t17;
 uint8_t const (*t18)[4];
 uint32_t t10;
 uint32_t t20;
 bool t8;
 uint8_t t19[4];
 t0 = a0;
 t1 = (struct environment_linked_libraries_elf_view__8184 const *)&t0;
 t2 = (uintptr_t)0ul;
 t3 = a0.phdrs;
 t4 = t3.len;
 zig_loop_9:
 t5 = t2;
 t6 = t5;
 t7 = t4;
 t8 = t6 < t7;
 if (t8) {
  t9 = t3.ptr[t5];
  t10 = t9.p_type;
  t8 = t10 != UINT32_C(4);
  if (t8) {
   goto zig_block_1;
  }
  goto zig_block_2;

  zig_block_2:;
  t11 = (*t1);
  t7 = t9.p_vaddr;
  t12 = t7;
  t7 = t9.p_memsz;
  t13 = t7;
  t14 = environment_linked_libraries_elf_view_slice__6968(t11, t12, t13);
  t15 = t14;
  t16 = (nav__6975_39 const *)&t15;
  t13 = t14.len;
  t7 = t13;
  t8 = t7 < UINT64_C(16);
  if (t8) {
   goto zig_block_1;
  }
  goto zig_block_3;

  zig_block_3:;
  t14 = (*t16);
  t17 = t14.ptr;
  t17 = (uint8_t const *)(((uintptr_t)t17) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t18 = (uint8_t const (*)[4])t17;
  memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
  memcpy(&t10, &t19, sizeof(uint32_t));
  t10 = zig_wrap_u32(t10, UINT8_C(32));
  t8 = t10 != UINT32_C(4);
  if (t8) {
   goto zig_block_1;
  }
  goto zig_block_4;

  zig_block_4:;
  t14 = (*t16);
  t17 = t14.ptr;
  t17 = (uint8_t const *)(((uintptr_t)t17) + ((uintptr_t)4ul*sizeof(uint8_t)));
  t18 = (uint8_t const (*)[4])t17;
  memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
  memcpy(&t10, &t19, sizeof(uint32_t));
  t10 = zig_wrap_u32(t10, UINT8_C(32));
  t14 = (*t16);
  t17 = t14.ptr;
  t17 = (uint8_t const *)(((uintptr_t)t17) + ((uintptr_t)8ul*sizeof(uint8_t)));
  t18 = (uint8_t const (*)[4])t17;
  memcpy(t19, (const char *)t18, sizeof(uint8_t[4]));
  memcpy(&t20, &t19, sizeof(uint32_t));
  t20 = zig_wrap_u32(t20, UINT8_C(32));
  t8 = t20 != UINT32_C(3);
  if (t8) {
   goto zig_block_1;
  }
  goto zig_block_5;

  zig_block_5:;
  t14 = (*t16);
  t17 = t14.ptr;
  t17 = (uint8_t const *)(((uintptr_t)t17) + ((uintptr_t)12ul*sizeof(uint8_t)));
  t18 = (uint8_t const (*)[4])t17;
  t14.ptr = &(*t18)[(uintptr_t)0ul];
  t14.len = (uintptr_t)4ul;
  t8 = mem_eql__anon_7887__6466(t14, (nav__6975_39){(uint8_t const *)&__anon_8445,(uintptr_t)4ul});
  t8 = !t8;
  if (t8) {
   goto zig_block_1;
  }
  goto zig_block_6;

  zig_block_6:;
  t14 = (*t16);
  t17 = t14.ptr;
  t17 = (uint8_t const *)(((uintptr_t)t17) + ((uintptr_t)16ul*sizeof(uint8_t)));
  t5 = (uintptr_t)t10;
  t14.ptr = t17;
  t14.len = t5;
  t21 = t14;
  return t21;
 }
 goto zig_block_0;

 zig_block_1:;
 t5 = t5 + (uintptr_t)1ul;
 t2 = t5;
 goto zig_loop_9;

 zig_block_0:;
 return (nav__6975_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
}

static nav__3638_39 environment_linked_libraries_root_hexEncode__3638(struct mem_Allocator__599 const a0, nav__3638_39 const a1) {
 struct mem_Allocator__599 const *t1;
 nav__3638_50 t2;
 nav__3638_50 t7;
 nav__3638_50 t8;
 struct mem_Allocator__599 t3;
 struct mem_Allocator__599 t0;
 uintptr_t t4;
 uintptr_t t11;
 uintptr_t t15;
 uintptr_t t10;
 nav__3638_52 t5;
 nav__3638_50 const *t9;
 uint64_t t12;
 uint64_t t13;
 uint8_t *t16;
 uint8_t const *t18;
 nav__3638_39 t19;
 nav__3638_39 t20;
 bool t6;
 uint8_t t14;
 uint8_t t17;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = (*t1);
 t4 = a1.len;
 t4 = t4 * (uintptr_t)2ul;
 t5 = mem_Allocator_alloc__anon_2384__5077(t3, t4);
 t6 = t5.error == UINT16_C(0);
 if (t6) {
  t7 = t5.payload;
  t2 = t7;
  goto zig_block_0;
 }
 return (nav__3638_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_0:;
 t8 = t2;
 t9 = (nav__3638_50 const *)&t8;
 t10 = (uintptr_t)0ul;
 t4 = a1.len;
 zig_loop_25:
 t11 = t10;
 t12 = t11;
 t13 = t4;
 t6 = t12 < t13;
 if (t6) {
  t14 = a1.ptr[t11];
  t15 = t11 * (uintptr_t)2ul;
  t7 = (*t9);
  t16 = &t7.ptr[t15];
  t17 = zig_shr_u8(t14, UINT8_C(4));
  t15 = (uintptr_t)t17;
  t18 = (uint8_t const *)&(*&__anon_8449)[t15];
  t17 = (*t18);
  (*t16) = t17;
  t15 = t11 * (uintptr_t)2ul;
  t15 = t15 + (uintptr_t)1ul;
  t7 = (*t9);
  t16 = &t7.ptr[t15];
  t14 = t14 & UINT8_C(15);
  t15 = (uintptr_t)t14;
  t18 = (uint8_t const *)&(*&__anon_8449)[t15];
  t14 = (*t18);
  (*t16) = t14;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t11 = t11 + (uintptr_t)1ul;
 t10 = t11;
 goto zig_loop_25;

 zig_block_1:;
 memcpy(&t19, &t2, sizeof(nav__3638_39));
 t20 = t19;
 return t20;
}

static nav__6970_40 environment_linked_libraries_elf_view_dynamicEntries__6970(struct environment_linked_libraries_elf_view__8184 const a0) {
 struct environment_linked_libraries_elf_view__8184 const *t1;
 struct elf_Elf64_Phdr__8033 t2;
 struct elf_Elf64_Phdr__8033 t6;
 struct environment_linked_libraries_elf_view__8184 t3;
 struct environment_linked_libraries_elf_view__8184 t0;
 nav__6970_52 t4;
 uintptr_t t7;
 uintptr_t t9;
 uint64_t t8;
 struct elf_Elf64_Dyn__8329 const *t10;
 struct elf_Elf64_Dyn__8329 const *t11;
 struct elf_Elf64_Dyn__8329 const *const *t12;
 nav__6970_40 t13;
 nav__6970_40 t14;
 bool t5;
 t0 = a0;
 t1 = (struct environment_linked_libraries_elf_view__8184 const *)&t0;
 t3 = (*t1);
 t4 = environment_linked_libraries_elf_view_findPhdr__6969(t3, UINT32_C(2));
 t5 = t4.is_null != true;
 if (t5) {
  t6 = t4.payload;
  t2 = t6;
  goto zig_block_0;
 }
 return (nav__6970_40){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_0:;
 t7 = a0.base_addr;
 t8 = t2.p_vaddr;
 t9 = t8;
 t9 = t7 + t9;
 t10 = (struct elf_Elf64_Dyn__8329 const *)t9;
 t11 = t10;
 t12 = (struct elf_Elf64_Dyn__8329 const *const *)&t11;
 t8 = t2.p_memsz;
 t8 = t8 / UINT64_C(16);
 t10 = (*t12);
 t10 = (struct elf_Elf64_Dyn__8329 const *)(((uintptr_t)t10) + ((uintptr_t)0ul*sizeof(struct elf_Elf64_Dyn__8329)));
 t9 = t8;
 t13.ptr = t10;
 t13.len = t9;
 t14 = t13;
 return t14;
}

static uint8_t const *environment_linked_libraries_elf_view_strtab__6972(nav__6972_41 const a0) {
 uintptr_t t0;
 uintptr_t t3;
 nav__6972_45 t1;
 uint8_t const *t4;
 bool t2;
 t1 = environment_linked_libraries_elf_view_dynVal__6971(a0, INT64_C(5));
 t2 = t1.is_null != true;
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 return NULL;

 zig_block_0:;
 t4 = (uint8_t const *)t0;
 return t4;
}

static nav__6974_39 environment_linked_libraries_elf_view_soname__6974(nav__6974_42 const a0, uint8_t const *const a1) {
 uintptr_t t0;
 uintptr_t t3;
 nav__6974_47 t1;
 nav__6974_39 t4;
 nav__6974_39 t5;
 bool t2;
 t1 = environment_linked_libraries_elf_view_dynVal__6971(a0, INT64_C(14));
 t2 = t1.is_null != true;
 if (t2) {
  t3 = t1.payload;
  t0 = t3;
  goto zig_block_0;
 }
 return (nav__6974_39){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_0:;
 t4 = environment_linked_libraries_elf_view_strFromTable__6973(a1, t0);
 t5 = t4;
 return t5;
}

static nav__4388_40 array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_addOne__4388(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *t1;
 nav__4388_47 *t2;
 nav__4388_47 t3;
 uintptr_t t4;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t5;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t0;
 nav__4388_40 t7;
 struct environment_linked_libraries_root_LibraryInfo__946 *t8;
 uint16_t t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *)&t0;
 t2 = (nav__4388_47 *)&a0->items;
 t3 = (*t2);
 t4 = t3.len;
 t4 = t4 + (uintptr_t)1ul;
 t5 = (*t1);
 t6 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_ensureTotalCapacity__4384(t5, t4);
 if (t6) {
  t7.payload = ((struct environment_linked_libraries_root_LibraryInfo__946 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul);
  t7.error = t6;
  return t7;
 }
 t5 = (*t1);
 t8 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_addOneAssumeCapacity__4389(t5);
 t7.payload = t8;
 t7.error = UINT16_C(0);
 return t7;
}

static uint16_t fmt_format__anon_8481__7032(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__7032_43){(uint8_t const *)&__anon_8550,(uintptr_t)61ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7024_40 fmt_bufPrint__anon_8364__7024(nav__7024_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__7024_54 (**t11)(void const *, nav__7024_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__7024_39 t15;
 nav__7024_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7024_54 (**)(void const *, nav__7024_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_8481__7032(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7024_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__7025_39 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7025(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct mem_Allocator__599 const a1, nav__7025_44 const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t4;
 uintptr_t t5;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t8;
 nav__7025_67 t9;
 nav__7025_39 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t14;
 nav__7025_73 t15;
 nav__7025_73 t16;
 nav__7025_73 const *t17;
 nav__7025_44 *t18;
 nav__7025_76 t19;
 nav__7025_76 t20;
 nav__7025_76 const *t21;
 struct environment_linked_libraries_root_LibraryEntry__864 *t22;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t23;
 uint16_t t6;
 bool t7;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&a0->entries;
 t4 = (*t3);
 t5 = t4.len;
 t5 = t5 + (uintptr_t)1ul;
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4202(t2, a1, t5);
 t7 = t6 == UINT16_C(0);
 if (t7) {
  goto zig_block_0;
 }
 t2 = (*t1);
 t8 = (*t2);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7033(t8, a2);
 t7 = t9.is_null != true;
 if (t7) {
  t10 = t9.payload;
  t5 = t10;
  goto zig_block_1;
 }
 t11.payload = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075){((nav__7025_44 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct environment_linked_libraries_root_LibraryEntry__864 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0xaaaaaaaaaaaaaaaaul,false};
 t11.error = t6;
 return t11;

 zig_block_1:;
 t2 = (*t1);
 t3 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t2->entries;
 t4 = (*t3);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t4);
 t13 = t12;
 t14 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t13;
 t12 = (*t14);
 t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t12);
 t16 = t15;
 t17 = (nav__7025_73 const *)&t16;
 t15 = (*t17);
 t18 = &t15.ptr[t5];
 t12 = (*t14);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t12);
 t20 = t19;
 t21 = (nav__7025_76 const *)&t20;
 t19 = (*t21);
 t22 = &t19.ptr[t5];
 t23.key_ptr = t18;
 t23.value_ptr = t22;
 t23.index = t5;
 t23.found_existing = true;
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;

 zig_block_0:;
 t2 = (*t1);
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7034(t2, a2);
 t11.payload = t23;
 t11.error = UINT16_C(0);
 return t11;
}

static uint16_t fmt_format__anon_8574__7035(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__7035_43){(uint8_t const *)&__anon_8644,(uintptr_t)62ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7026_40 fmt_bufPrint__anon_8381__7026(nav__7026_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__7026_54 (**t11)(void const *, nav__7026_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__7026_39 t15;
 nav__7026_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7026_54 (**)(void const *, nav__7026_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_8574__7035(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7026_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static uint16_t fmt_format__anon_8650__7036(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__7036_43){(uint8_t const *)&__anon_8711,(uintptr_t)55ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7027_40 fmt_bufPrint__anon_8390__7027(nav__7027_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__7027_54 (**t11)(void const *, nav__7027_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__7027_39 t15;
 nav__7027_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7027_54 (**)(void const *, nav__7027_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_8650__7036(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7027_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 json_stringify_writeStreamArbitraryDepth__anon_8398__7028(struct mem_Allocator__599 const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a1, struct json_stringify_StringifyOptions__8133 const a2) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 t0;
 t0 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6997(a0, a1, a2);
 return t0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7029(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, struct environment_root_EnvironmentJson__845 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 t7;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 t9;
 struct environment_root_EnvironmentJson__845 t2;
 uint16_t t4;
 bool t6;
 bool t5;
 bool t8;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(t3);
 if (t4) {
  return t4;
 }
 t5 = true;
 t6 = t5;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t3, (nav__7029_62){((uint8_t const *)&__anon_8724),(uintptr_t)23ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t7 = a1.integration_environment;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7037(t3, t7);
  if (t4) {
   return t4;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = true;
 t6 = t8;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t3, (nav__7029_62){((uint8_t const *)&__anon_8732),(uintptr_t)16ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t9 = a1.linked_libraries;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7038(t3, t9);
  if (t4) {
   return t4;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6998(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 t0 = a0;
 return;
}

static void array_list_ArrayListAligned_28u8_2cnull_29_clearAndFree__1563(struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const a0) {
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t1;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t2;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *t0;
 struct mem_Allocator__599 *t3;
 struct mem_Allocator__599 t4;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 t5;
 nav__1563_43 t6;
 nav__1563_43 *t7;
 uintptr_t *t8;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t0;
 t2 = (*t1);
 t3 = (struct mem_Allocator__599 *)&t2->allocator;
 t4 = (*t3);
 t2 = (*t1);
 t5 = (*t2);
 t6 = array_list_ArrayListAligned_28u8_2cnull_29_allocatedSlice__1575(t5);
 mem_Allocator_free__anon_3705__5702(t4, t6);
 t2 = (*t1);
 t7 = (nav__1563_43 *)&t2->items;
 t8 = &t7->len;
 (*t8) = (uintptr_t)0ul;
 t2 = (*t1);
 t8 = (uintptr_t *)&t2->capacity;
 (*t8) = (uintptr_t)0ul;
 return;
}

static uint16_t fmt_format__anon_8814__7039(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__7039_43){(uint8_t const *)&__anon_8883,(uintptr_t)63ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7030_40 fmt_bufPrint__anon_8421__7030(nav__7030_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__7030_54 (**t11)(void const *, nav__7030_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__7030_39 t15;
 nav__7030_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7030_54 (**)(void const *, nav__7030_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_8814__7039(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7030_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__4682_39 fs_Dir_createFileZ__4682(struct fs_Dir__1720 const a0, uint8_t const *const a1, struct fs_File_CreateFlags__8153 const a2) {
 uint32_t *t2;
 uintptr_t t7;
 struct fs_File__799 *t14;
 int32_t *t15;
 int32_t t5;
 int32_t t11;
 int32_t t12;
 uint32_t t6;
 uint32_t t1;
 nav__4682_50 t8;
 nav__4682_39 t10;
 nav__4682_39 t13;
 struct fs_Dir__1720 t0;
 uint16_t t9;
 uint8_t t3;
 bool t4;
 t0 = a0;
 t2 = (uint32_t *)&t1;
 t4 = a2.read;
 if (t4) {
  t3 = UINT8_C(2);
  goto zig_block_0;
 }
 t3 = UINT8_C(1);
 goto zig_block_0;

 zig_block_0:;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffffc)), zig_shl_u32((uint32_t)t3, UINT8_C(0)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffbf)), zig_shl_u32((uint32_t)true, UINT8_C(6)));
 t2 = (uint32_t *)&t1;
 t4 = a2.truncate;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffdff)), zig_shl_u32((uint32_t)t4, UINT8_C(9)));
 t2 = (uint32_t *)&t1;
 t4 = a2.exclusive;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffff7f)), zig_shl_u32((uint32_t)t4, UINT8_C(7)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffffc3)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(2)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffeff)), zig_shl_u32((uint32_t)false, UINT8_C(8)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffffbff)), zig_shl_u32((uint32_t)false, UINT8_C(10)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffff7ff)), zig_shl_u32((uint32_t)false, UINT8_C(11)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffefff)), zig_shl_u32((uint32_t)false, UINT8_C(12)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffdfff)), zig_shl_u32((uint32_t)false, UINT8_C(13)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffffbfff)), zig_shl_u32((uint32_t)false, UINT8_C(14)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffff7fff)), zig_shl_u32((uint32_t)UINT8_C(0), UINT8_C(15)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffeffff)), zig_shl_u32((uint32_t)false, UINT8_C(16)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffdffff)), zig_shl_u32((uint32_t)false, UINT8_C(17)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfffbffff)), zig_shl_u32((uint32_t)false, UINT8_C(18)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff7ffff)), zig_shl_u32((uint32_t)false, UINT8_C(19)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffefffff)), zig_shl_u32((uint32_t)false, UINT8_C(20)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffdfffff)), zig_shl_u32((uint32_t)false, UINT8_C(21)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xffbfffff)), zig_shl_u32((uint32_t)false, UINT8_C(22)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0x7fffff)), zig_shl_u32((uint32_t)UINT16_C(0), UINT8_C(23)));
 t2 = (uint32_t *)&t1;
 (*t2) = zig_or_u32(zig_and_u32((*t2), UINT32_C(0xfff7ffff)), zig_shl_u32((uint32_t)true, UINT8_C(19)));
 t5 = a0.fd;
 t6 = t1;
 t7 = a2.mode;
 t8 = posix_openatZ__1769(t5, a1, t6, t7);
 if (t8.error) {
  t9 = t8.error;
  t10.payload = (struct fs_File__799){-INT32_C(0x55555556)};
  t10.error = t9;
  return t10;
 }
 t5 = t8.payload;
 t3 = a2.lock;
 t4 = t3 != UINT8_C(0);
 if (t4) {
  t4 = a2.lock_nonblocking;
  if (t4) {
   t11 = INT32_C(4);
   goto zig_block_2;
  }
  t11 = INT32_C(0);
  goto zig_block_2;

  zig_block_2:;
  t3 = a2.lock;
  switch (t3) {
   case UINT8_C(0): {
    zig_unreachable();
   }
   case UINT8_C(1): {
    t11 = INT32_C(1) | t11;
    t12 = t11;
    goto zig_block_3;
   }
   case UINT8_C(2): {
    t11 = INT32_C(2) | t11;
    t12 = t11;
    goto zig_block_3;
   }
   default: zig_unreachable();
  }

  zig_block_3:;
  t9 = posix_flock__1930(t5, t12);
  if (t9) {
   posix_close__1727(t5);
   t10.payload = (struct fs_File__799){-INT32_C(0x55555556)};
   t10.error = t9;
   return t10;
  }
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t13.error = UINT16_C(0);
 t14 = &t13.payload;
 t15 = (int32_t *)&t14->handle;
 (*t15) = t5;
 return t13;
}

static uint16_t fmt_format__anon_8889__7040(struct io_Writer__3878 const a0) {
 struct io_Writer__3878 const *t1;
 struct io_Writer__3878 t2;
 struct io_Writer__3878 t0;
 uint16_t t3;
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2 = (*t1);
 t3 = io_Writer_writeAll__5732(t2, (nav__7040_43){(uint8_t const *)&__anon_8956,(uintptr_t)59ul});
 if (t3) {
  return t3;
 }
 return 0;
}

static nav__7031_40 fmt_bufPrint__anon_8436__7031(nav__7031_39 const a0) {
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t1;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 t0;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t2;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 t3;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t4;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *t5;
 struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *t6;
 void const **t8;
 struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *t9;
 void const *t10;
 nav__7031_54 (**t11)(void const *, nav__7031_56);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 nav__7031_39 t15;
 nav__7031_40 t16;
 uint16_t t13;
 bool t14;
 t1 = io_fixed_buffer_stream_fixedBufferStream__anon_2068__4994(a0);
 t0 = t1;
 t2 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_writer__4984(&t0);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write_27_29_29__3868 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29__2065 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7031_54 (**)(void const *, nav__7031_56))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aio_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_2cerror_7bNoSpaceLeft_7d_2c_28function_20_27write__5724;
 t12 = t7;
 t13 = fmt_format__anon_8889__7040(t12);
 t14 = t13 == UINT16_C(0);
 if (t14) {
  goto zig_block_0;
 }
 switch (t13) {
  case zig_error_NoSpaceLeft: {
   return (nav__7031_40){{(uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul},zig_error_NoSpaceLeft};
  }
  default: {
   zig_unreachable();
  }
 }

 zig_block_0:;
 t1 = t0;
 t15 = io_fixed_buffer_stream_FixedBufferStream_28_5b_5du8_29_getWritten__4992(t1);
 t16.payload = t15;
 t16.error = UINT16_C(0);
 return t16;
}

static nav__6968_39 environment_linked_libraries_elf_view_slice__6968(struct environment_linked_libraries_elf_view__8184 const a0, uintptr_t const a1, uintptr_t const a2) {
 uintptr_t t0;
 uint8_t const *t1;
 uint8_t const *t2;
 uint8_t const *const *t3;
 nav__6968_39 t4;
 t0 = a0.base_addr;
 t0 = t0 + a1;
 t1 = (uint8_t const *)t0;
 t2 = t1;
 t3 = (uint8_t const *const *)&t2;
 t1 = (*t3);
 t1 = (uint8_t const *)(((uintptr_t)t1) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4.ptr = t1;
 t4.len = a2;
 return t4;
}

static nav__6969_39 environment_linked_libraries_elf_view_findPhdr__6969(struct environment_linked_libraries_elf_view__8184 const a0, uint32_t const a1) {
 nav__6969_45 t1;
 uintptr_t t2;
 uintptr_t t3;
 uintptr_t t0;
 uint64_t t4;
 uint64_t t5;
 struct elf_Elf64_Phdr__8033 t7;
 nav__6969_39 t9;
 uint32_t t8;
 bool t6;
 t0 = (uintptr_t)0ul;
 t1 = a0.phdrs;
 t2 = t1.len;
 zig_loop_7:
 t3 = t0;
 t4 = t3;
 t5 = t2;
 t6 = t4 < t5;
 if (t6) {
  t7 = t1.ptr[t3];
  t8 = t7.p_type;
  t6 = t8 == a1;
  if (t6) {
   t9.is_null = false;
   t9.payload = t7;
   return t9;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t3 = t3 + (uintptr_t)1ul;
 t0 = t3;
 goto zig_loop_7;

 zig_block_0:;
 return (nav__6969_39){{UINT32_C(0xaaaaaaaa),UINT32_C(0xaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa),UINT64_C(0xaaaaaaaaaaaaaaaa)},true};
}

static nav__6971_38 environment_linked_libraries_elf_view_dynVal__6971(nav__6971_41 const a0, int64_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 struct elf_Elf64_Dyn__8329 t6;
 int64_t t7;
 nav__6971_38 t8;
 bool t5;
 t0 = (uintptr_t)0ul;
 t1 = a0.len;
 zig_loop_6:
 t2 = t0;
 t3 = t2;
 t4 = t1;
 t5 = t3 < t4;
 if (t5) {
  t6 = a0.ptr[t2];
  t7 = t6.d_tag;
  t5 = t7 == a1;
  if (t5) {
   t4 = t6.d_val;
   t2 = t4;
   t8.is_null = false;
   t8.payload = t2;
   return t8;
  }
  goto zig_block_2;

  zig_block_2:;
  t7 = t6.d_tag;
  t5 = t7 == INT64_C(0);
  if (t5) {
   goto zig_block_0;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t2 = t2 + (uintptr_t)1ul;
 t0 = t2;
 goto zig_loop_6;

 zig_block_0:;
 return (nav__6971_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
}

static nav__6973_39 environment_linked_libraries_elf_view_strFromTable__6973(uint8_t const *const a0, uintptr_t const a1) {
 uintptr_t t0;
 uint8_t const *t1;
 nav__6973_39 t2;
 nav__6973_39 t3;
 t0 = (uintptr_t)a0;
 t0 = t0 + a1;
 t1 = (uint8_t const *)t0;
 t2 = mem_span__anon_8113__6933(t1);
 memcpy(&t3, &t2, sizeof(nav__6973_39));
 return t3;
}

static uint16_t array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_ensureTotalCapacity__4384(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t7;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t0;
 uint16_t t8;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = array_list_ArrayListAlignedUnmanaged_28environment_linked_libraries_root_LibraryInfo_2cnull_29_growCapacit__7099(t3, a1);
 t7 = (*t1);
 t8 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_ensureTotalCapacityP__4385(t7, t3);
 return t8;
}

static struct environment_linked_libraries_root_LibraryInfo__946 *array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_addOneAssumeCapacity__4389(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const a0) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *t1;
 nav__4389_44 *t2;
 nav__4389_44 *t11;
 nav__4389_44 t3;
 uintptr_t t4;
 uintptr_t t6;
 uintptr_t *t5;
 uint64_t t7;
 uint64_t t8;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t10;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t0;
 struct environment_linked_libraries_root_LibraryInfo__946 *t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *)&t0;
 t2 = (nav__4389_44 *)&a0->items;
 t3 = (*t2);
 t4 = t3.len;
 t5 = (uintptr_t *)&a0->capacity;
 t6 = (*t5);
 t7 = t4;
 t8 = t6;
 t9 = t7 < t8;
 debug_assert__180(t9);
 t10 = (*t1);
 t2 = (nav__4389_44 *)&t10->items;
 t5 = &t2->len;
 t6 = (*t5);
 t6 = t6 + (uintptr_t)1ul;
 (*t5) = t6;
 t10 = (*t1);
 t2 = (nav__4389_44 *)&t10->items;
 t11 = (nav__4389_44 *)&a0->items;
 t3 = (*t11);
 t6 = t3.len;
 t6 = t6 - (uintptr_t)1ul;
 t3 = (*t2);
 t12 = &t3.ptr[t6];
 return t12;
}

static uint16_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4202(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 void *t3;
 uint64_t t4;
 uint64_t t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t6;
 struct array_hash_map_IndexHeader__892 **t8;
 struct array_hash_map_IndexHeader__892 *t9;
 struct array_hash_map_IndexHeader__892 *t10;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t20;
 struct array_hash_map_IndexHeader__892 *const *t12;
 nav__4202_62 t18;
 struct array_hash_map_IndexHeader__892 t13;
 uint32_t t14;
 uint16_t t7;
 nav__4202_60 t16;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_lock__3982(t3);
 t4 = a2;
 t5 = t4 <= UINT64_C(8);
 if (t5) {
  t2 = (*t1);
  t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t2->entries;
  t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4328(t6, a1, a2);
  if (t7) {
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3983(t3);
   return t7;
  }
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t10 = t9;
  t11 = t10;
  t12 = (struct array_hash_map_IndexHeader__892 *const *)&t11;
  t10 = (*t12);
  t13 = (*t10);
  t14 = array_hash_map_IndexHeader_capacity__3974(t13);
  t4 = a2;
  t15 = (uint64_t)t14;
  t5 = t4 <= t15;
  if (t5) {
   t2 = (*t1);
   t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t2->entries;
   t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4328(t6, a1, a2);
   if (t7) {
    t2 = (*t1);
    t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
    debug_SafetyLock_unlock__3983(t3);
    return t7;
   }
   t2 = (*t1);
   t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
   debug_SafetyLock_unlock__3983(t3);
   return 0;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t2 = (*t1);
 t6 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t2->entries;
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4328(t6, a1, a2);
 if (t7) {
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t16 = array_hash_map_IndexHeader_findBitIndex__3977(a2);
 if (t16.error) {
  t7 = t16.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t17 = t16.payload;
 t18 = array_hash_map_IndexHeader_alloc__3978(a1, t17);
 if (t18.error) {
  t7 = t18.error;
  t2 = (*t1);
  t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
  debug_SafetyLock_unlock__3983(t3);
  return t7;
 }
 t10 = t18.payload;
 t8 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t9 = (*t8);
 t5 = t9 != NULL;
 if (t5) {
  t19 = t9;
  t20 = t19;
  t12 = (struct array_hash_map_IndexHeader__892 *const *)&t20;
  t19 = (*t12);
  array_hash_map_IndexHeader_free__3979(t19, a1);
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t2 = (*t1);
 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4294(t2, t10);
 t2 = (*t1);
 t8 = (struct array_hash_map_IndexHeader__892 **)&t2->index_header;
 t19 = (struct array_hash_map_IndexHeader__892 *)t10;
 (*t8) = t19;
 t2 = (*t1);
 t3 = (void *)((uint8_t *)t2 + (uintptr_t)32ul);
 debug_SafetyLock_unlock__3983(t3);
 return 0;
}

static nav__7033_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7033(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7033_41 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_IndexHeader__892 *t3;
 struct array_hash_map_IndexHeader__892 *t5;
 struct array_hash_map_IndexHeader__892 *t24;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t11;
 nav__7033_61 t12;
 nav__7033_64 t13;
 uintptr_t t15;
 uintptr_t t16;
 uintptr_t t14;
 uint64_t t17;
 uint64_t t18;
 nav__7033_41 *t19;
 nav__7033_41 t22;
 nav__7033_38 t23;
 struct array_hash_map_IndexHeader__892 *const *t25;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t28;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 uint32_t t6;
 uint32_t t20;
 struct array_hash_map_IndexHeader__892 t26;
 bool t4;
 bool t21;
 uint8_t t27;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t3 = a0.index_header;
 t4 = t3 != NULL;
 if (t4) {
  t5 = t3;
  t2 = t5;
  goto zig_block_0;
 }
 t6 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t7 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *)&t1->entries;
 t8 = (*t7);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t8);
 t10 = t9;
 t11 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t10;
 t9 = (*t11);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t9);
 t9 = (*t11);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t9);
 t14 = (uintptr_t)0ul;
 t15 = t13.len;
 zig_loop_32:
 t16 = t14;
 t17 = t16;
 t18 = t15;
 t4 = t17 < t18;
 if (t4) {
  t19 = &t13.ptr[t16];
  t20 = t12.ptr[t16];
  t4 = t20 == t6;
  if (t4) {
   t22 = (*t19);
   t4 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t22, t16);
   t21 = t4;
   goto zig_block_4;
  }
  t21 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t21) {
   t23.is_null = false;
   t23.payload = t16;
   return t23;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t16 = t16 + (uintptr_t)1ul;
 t14 = t16;
 goto zig_loop_32;

 zig_block_1:;
 return (nav__7033_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t24 = t2;
 t25 = (struct array_hash_map_IndexHeader__892 *const *)&t24;
 t5 = (*t25);
 t26 = (*t5);
 t27 = array_hash_map_IndexHeader_capacityIndexType__3973(t26);
 switch (t27) {
  case UINT8_C(0): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7103(t28, a1, t2);
   return t23;
  }
  case UINT8_C(1): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7104(t28, a1, t2);
   return t23;
  }
  case UINT8_C(2): {
   t28 = (*t1);
   t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7105(t28, a1, t2);
   return t23;
  }
  default: zig_unreachable();
 }
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7034(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, nav__7034_42 const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t6;
 struct array_hash_map_IndexHeader__892 *t39;
 struct array_hash_map_IndexHeader__892 **t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t9;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t10;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t11;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t12;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t13;
 nav__7034_66 t14;
 nav__7034_66 t15;
 nav__7034_66 const *t16;
 nav__7034_69 t17;
 nav__7034_69 t18;
 nav__7034_69 const *t19;
 uintptr_t t21;
 uintptr_t t22;
 uintptr_t t20;
 uint64_t t23;
 uint64_t t24;
 nav__7034_42 *t25;
 nav__7034_42 t28;
 nav__7034_72 t29;
 nav__7034_72 t30;
 nav__7034_72 t37;
 nav__7034_72 const *t31;
 struct environment_linked_libraries_root_LibraryEntry__864 *t32;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t33;
 uint32_t *const *t34;
 uint32_t *t35;
 nav__7034_42 *const *t36;
 struct environment_linked_libraries_root_LibraryEntry__864 *const *t38;
 struct array_hash_map_IndexHeader__892 *const *t40;
 uint32_t t7;
 uint32_t t26;
 struct array_hash_map_IndexHeader__892 t41;
 bool t5;
 bool t27;
 uint8_t t42;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t3 = (struct array_hash_map_IndexHeader__892 **)&a0->index_header;
 t4 = (*t3);
 t5 = t4 != NULL;
 if (t5) {
  t6 = t4;
  t2 = t6;
  goto zig_block_0;
 }
 t7 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t8->entries;
 t10 = (*t9);
 t11 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t10);
 t12 = t11;
 t13 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t12;
 t11 = (*t13);
 t14 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t11);
 t15 = t14;
 t16 = (nav__7034_66 const *)&t15;
 t11 = (*t13);
 t17 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t11);
 t18 = t17;
 t19 = (nav__7034_69 const *)&t18;
 t20 = (uintptr_t)0ul;
 t21 = t17.len;
 zig_loop_39:
 t22 = t20;
 t23 = t22;
 t24 = t21;
 t5 = t23 < t24;
 if (t5) {
  t25 = &t17.ptr[t22];
  t26 = t14.ptr[t22];
  t5 = t26 == t7;
  if (t5) {
   t28 = (*t25);
   t5 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t28, t22);
   t27 = t5;
   goto zig_block_4;
  }
  t27 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t27) {
   t11 = (*t13);
   t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t11);
   t30 = t29;
   t31 = (nav__7034_72 const *)&t30;
   t29 = (*t31);
   t32 = &t29.ptr[t22];
   t33.key_ptr = t25;
   t33.value_ptr = t32;
   t33.index = t22;
   t33.found_existing = true;
   return t33;
  }
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_2;
 }
 goto zig_block_1;

 zig_block_2:;
 t22 = t22 + (uintptr_t)1ul;
 t20 = t22;
 goto zig_loop_39;

 zig_block_1:;
 t8 = (*t1);
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t8->entries;
 t21 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t9);
 t34 = &t16->ptr;
 t35 = (*t34);
 t35 = (uint32_t *)&t35[t21];
 (*t35) = t7;
 t36 = &t19->ptr;
 t25 = (*t36);
 t25 = (nav__7034_42 *)&t25[t21];
 t11 = (*t13);
 t29 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t11);
 t37 = t29;
 t31 = (nav__7034_72 const *)&t37;
 t38 = &t31->ptr;
 t32 = (*t38);
 t32 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t32[t21];
 t33.key_ptr = t25;
 t33.value_ptr = t32;
 t33.index = t21;
 t33.found_existing = false;
 return t33;

 zig_block_0:;
 t39 = t2;
 t40 = (struct array_hash_map_IndexHeader__892 *const *)&t39;
 t6 = (*t40);
 t41 = (*t6);
 t42 = array_hash_map_IndexHeader_capacityIndexType__3973(t41);
 switch (t42) {
  case UINT8_C(0): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7106(t8, a1, t2);
   return t33;
  }
  case UINT8_C(1): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7107(t8, a1, t2);
   return t33;
  }
  case UINT8_C(2): {
   t8 = (*t1);
   t33 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7108(t8, a1, t2);
   return t33;
  }
  default: zig_unreachable();
 }
}

static struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6997(struct mem_Allocator__599 const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a1, struct json_stringify_StringifyOptions__8133 const a2) {
 struct json_stringify_StringifyOptions__8133 *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t2;
 uintptr_t *t3;
 uint8_t *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 t0;
 (void)a0;
 t1 = (struct json_stringify_StringifyOptions__8133 *)&t0.options;
 (*t1) = a2;
 t2 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t0.stream;
 (*t2) = a1;
 t3 = (uintptr_t *)&t0.indent_level;
 (*t3) = (uintptr_t)0ul;
 t4 = (uint8_t *)&t0.next_punctuation;
 (*t4) = UINT8_C(0);
 return t0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7005(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_StringifyOptions__8133 *t3;
 uintptr_t t6;
 uintptr_t t8;
 uintptr_t *t7;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t9;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t26;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t13;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t14;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t27;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t15;
 void const **t17;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t18;
 void const *t19;
 nav__7005_56 (**t20)(void const *, nav__7005_58);
 struct io_Writer__3878 t21;
 struct io_Writer__3878 t16;
 struct io_Writer__3878 t22;
 struct io_Writer__3878 t28;
 struct io_Writer__3878 t29;
 struct io_Writer__3878 const *t23;
 uint16_t t24;
 uint16_t t25;
 struct json_stringify_StringifyOptions__8133 t4;
 uint8_t t5;
 uint8_t t2;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = UINT8_C(32);
 t3 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
 t4 = (*t3);
 t5 = t4.whitespace;
 switch (t5) {
  case UINT8_C(0): {
   return 0;
  }
  case UINT8_C(1): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)2ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)3ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(4): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)4ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(5): {
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t8 = (uintptr_t)8ul * t8;
   t6 = t8;
   goto zig_block_0;
  }
  case UINT8_C(6): {
   t2 = UINT8_C(9);
   t7 = (uintptr_t *)&a0->indent_level;
   t8 = (*t7);
   t6 = t8;
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t9 = (*t1);
 t10 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t9->stream;
 t11 = (*t10);
 t12 = t11;
 t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t12;
 t14 = t13;
 t15 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t14;
 t17 = (void const **)&t16.context;
 t13 = (*t15);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t13->context;
 t19 = (void const *)t18;
 (*t17) = t19;
 t20 = (nav__7005_56 (**)(void const *, nav__7005_58))&t16.writeFn;
 (*t20) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t21 = t16;
 t22 = t21;
 t23 = (struct io_Writer__3878 const *)&t22;
 t21 = (*t23);
 t24 = io_Writer_writeByte__5734(t21, UINT8_C(10));
 memcpy(&t25, &t24, sizeof(uint16_t));
 if (t25) {
  return t25;
 }
 t9 = (*t1);
 t10 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t9->stream;
 t11 = (*t10);
 t5 = t2;
 t26 = t11;
 t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t26;
 t27 = t13;
 t15 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t27;
 t17 = (void const **)&t28.context;
 t13 = (*t15);
 t18 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t13->context;
 t19 = (void const *)t18;
 (*t17) = t19;
 t20 = (nav__7005_56 (**)(void const *, nav__7005_58))&t28.writeFn;
 (*t20) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t21 = t28;
 t29 = t21;
 t23 = (struct io_Writer__3878 const *)&t29;
 t21 = (*t23);
 t25 = io_Writer_writeByteNTimes__5735(t21, t5, t6);
 memcpy(&t24, &t25, sizeof(uint16_t));
 if (t24) {
  return t24;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7008(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *t3;
 uint8_t *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t23;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t29;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t30;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t13;
 void const **t15;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t16;
 void const *t17;
 nav__7008_56 (**t18)(void const *, nav__7008_58);
 struct io_Writer__3878 t19;
 struct io_Writer__3878 t14;
 struct io_Writer__3878 t20;
 struct io_Writer__3878 t25;
 struct io_Writer__3878 t26;
 struct io_Writer__3878 t31;
 struct io_Writer__3878 t32;
 struct io_Writer__3878 const *t21;
 struct json_stringify_StringifyOptions__8133 *t27;
 uint16_t t7;
 uint16_t t22;
 bool t4;
 uint8_t t6;
 struct json_stringify_StringifyOptions__8133 t28;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *)t2;
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7011(t3);
 t4 = !t4;
 debug_assert__180(t4);
 t5 = (uint8_t *)&a0->next_punctuation;
 t6 = (*t5);
 switch (t6) {
  case UINT8_C(0): {
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t2 = (*t1);
   t7 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7005(t2);
   if (t7) {
    return t7;
   }
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = (*t1);
   t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
   t9 = (*t8);
   t10 = t9;
   t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t10;
   t12 = t11;
   t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t12;
   t15 = (void const **)&t14.context;
   t11 = (*t13);
   t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t11->context;
   t17 = (void const *)t16;
   (*t15) = t17;
   t18 = (nav__7008_56 (**)(void const *, nav__7008_58))&t14.writeFn;
   (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t19 = t14;
   t20 = t19;
   t21 = (struct io_Writer__3878 const *)&t20;
   t19 = (*t21);
   t7 = io_Writer_writeByte__5734(t19, UINT8_C(44));
   memcpy(&t22, &t7, sizeof(uint16_t));
   if (t22) {
    return t22;
   }
   t2 = (*t1);
   t22 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7005(t2);
   if (t22) {
    return t22;
   }
   goto zig_block_0;
  }
  case UINT8_C(3): {
   t2 = (*t1);
   t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
   t9 = (*t8);
   t23 = t9;
   t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t23;
   t24 = t11;
   t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t24;
   t15 = (void const **)&t25.context;
   t11 = (*t13);
   t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t11->context;
   t17 = (void const *)t16;
   (*t15) = t17;
   t18 = (nav__7008_56 (**)(void const *, nav__7008_58))&t25.writeFn;
   (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t19 = t25;
   t26 = t19;
   t21 = (struct io_Writer__3878 const *)&t26;
   t19 = (*t21);
   t22 = io_Writer_writeByte__5734(t19, UINT8_C(58));
   memcpy(&t7, &t22, sizeof(uint16_t));
   if (t7) {
    return t7;
   }
   t27 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
   t28 = (*t27);
   t6 = t28.whitespace;
   t4 = t6 != UINT8_C(0);
   if (t4) {
    t2 = (*t1);
    t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
    t9 = (*t8);
    t29 = t9;
    t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t29;
    t30 = t11;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t30;
    t15 = (void const **)&t31.context;
    t11 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t11->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7008_56 (**)(void const *, nav__7008_58))&t31.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t31;
    t32 = t19;
    t21 = (struct io_Writer__3878 const *)&t32;
    t19 = (*t21);
    t7 = io_Writer_writeByte__5734(t19, UINT8_C(32));
    memcpy(&t22, &t7, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    goto zig_block_1;
   }
   goto zig_block_1;

   zig_block_1:;
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *t3;
 uint16_t t6;
 uint16_t t7;
 nav__7006_43 t4;
 bool t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *)t2;
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7010(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t5 = t4.payload;
  t5 = !t5;
  debug_assert__180(t5);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7008(t2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7003(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, uint8_t const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uintptr_t *t3;
 uintptr_t t4;
 (void)a1;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t *)&t2->indent_level;
 t4 = (*t3);
 t4 = t4 + (uintptr_t)1ul;
 (*t3) = t4;
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t9;
 void const **t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t12;
 void const *t13;
 nav__7000_54 (**t14)(void const *, nav__7000_56);
 struct io_Writer__3878 t15;
 struct io_Writer__3878 t10;
 struct io_Writer__3878 t16;
 struct io_Writer__3878 const *t17;
 uint8_t *t19;
 uint16_t t3;
 uint16_t t18;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(t2);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
 t5 = (*t4);
 t6 = t5;
 t7 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t6;
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t8;
 t11 = (void const **)&t10.context;
 t7 = (*t9);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t7->context;
 t13 = (void const *)t12;
 (*t11) = t13;
 t14 = (nav__7000_54 (**)(void const *, nav__7000_56))&t10.writeFn;
 (*t14) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t15 = t10;
 t16 = t15;
 t17 = (struct io_Writer__3878 const *)&t16;
 t15 = (*t17);
 t3 = io_Writer_writeByte__5734(t15, UINT8_C(123));
 memcpy(&t18, &t3, sizeof(uint16_t));
 if (t18) {
  return t18;
 }
 t2 = (*t1);
 (void)json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7003(t2, UINT8_C(0));
 t2 = (*t1);
 t19 = (uint8_t *)&t2->next_punctuation;
 (*t19) = UINT8_C(1);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7007(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *t3;
 uint16_t t6;
 uint16_t t7;
 nav__7007_43 t4;
 bool t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *)t2;
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7010(t3);
 t5 = t4.is_null != true;
 if (t5) {
  t5 = t4.payload;
  debug_assert__180(t5);
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t2 = (*t1);
 t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7008(t2);
 memcpy(&t7, &t6, sizeof(uint16_t));
 return t7;
}

static uint16_t fmt_formatBuf__anon_9211__7115(nav__7115_39 const a0, struct fmt_FormatOptions__4300 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a2) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t33;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t34;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t38;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t42;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t46;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t50;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t54;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t58;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t62;
 nav__7115_44 t2;
 uintptr_t t4;
 uintptr_t t5;
 uintptr_t t7;
 nav__7115_52 t6;
 uint64_t t8;
 uint64_t t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t32;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t37;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t41;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t45;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t49;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t53;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t57;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t61;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t13;
 void const **t15;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t16;
 void const *t17;
 nav__7115_52 (**t18)(void const *, nav__7115_39);
 struct io_Writer__3878 t19;
 struct io_Writer__3878 t14;
 struct io_Writer__3878 t20;
 struct io_Writer__3878 t35;
 struct io_Writer__3878 t36;
 struct io_Writer__3878 t39;
 struct io_Writer__3878 t40;
 struct io_Writer__3878 t43;
 struct io_Writer__3878 t44;
 struct io_Writer__3878 t47;
 struct io_Writer__3878 t48;
 struct io_Writer__3878 t51;
 struct io_Writer__3878 t52;
 struct io_Writer__3878 t55;
 struct io_Writer__3878 t56;
 struct io_Writer__3878 t59;
 struct io_Writer__3878 t60;
 struct io_Writer__3878 t63;
 struct io_Writer__3878 t64;
 struct io_Writer__3878 const *t21;
 nav__7115_67 t26;
 nav__7115_39 t28;
 nav__7115_39 t31;
 uint8_t *t30;
 uint32_t t25;
 uint16_t t22;
 uint16_t t23;
 nav__7115_70 t27;
 bool t3;
 uint8_t t29;
 uint8_t t24[4];
 t0 = a2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = a1.width;
 t3 = t2.is_null != true;
 if (t3) {
  t4 = t2.payload;
  t6 = unicode_utf8CountCodepoints__5769(a0);
  t3 = t6.error == UINT16_C(0);
  if (t3) {
   t7 = t6.payload;
   t5 = t7;
   goto zig_block_1;
  }
  t7 = a0.len;
  t5 = t7;
  goto zig_block_1;

  zig_block_1:;
  t8 = t5;
  t9 = t4;
  t3 = t8 < t9;
  if (t3) {
   t5 = t4 - t5;
   t7 = t5;
   goto zig_block_2;
  }
  t7 = (uintptr_t)0ul;
  goto zig_block_2;

  zig_block_2:;
  t9 = t7;
  t3 = t9 == UINT64_C(0);
  if (t3) {
   t10 = (*t1);
   t11 = t10;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t11;
   t12 = t1;
   t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t12;
   t15 = (void const **)&t14.context;
   t1 = (*t13);
   t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t17 = (void const *)t16;
   (*t15) = t17;
   t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t14.writeFn;
   (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t19 = t14;
   t20 = t19;
   t21 = (struct io_Writer__3878 const *)&t20;
   t19 = (*t21);
   t22 = io_Writer_writeAll__5732(t19, a0);
   memcpy(&t23, &t22, sizeof(uint16_t));
   memcpy(&t22, &t23, sizeof(uint16_t));
   return t22;
  }
  goto zig_block_3;

  zig_block_3:;
  t25 = a1.fill;
  t26.ptr = &t24[(uintptr_t)0ul];
  t26.len = (uintptr_t)4ul;
  t27 = unicode_utf8Encode__5754(t25, t26);
  t3 = t27.error == UINT16_C(0);
  if (t3) {
   t29 = t27.payload;
   t30 = (uint8_t *)&t24;
   t30 = (uint8_t *)(((uintptr_t)t30) + ((uintptr_t)0ul*sizeof(uint8_t)));
   t5 = (uintptr_t)t29;
   t26.ptr = t30;
   t26.len = t5;
   memcpy(&t31, &t26, sizeof(nav__7115_39));
   t28 = t31;
   goto zig_block_4;

  }
  t23 = t27.error;
  switch (t23) {
   case zig_error_Utf8CannotEncodeSurrogateHalf:
   case zig_error_CodepointTooLarge: {
    t28 = (nav__7115_39){(uint8_t const *)&__anon_5345,(uintptr_t)3ul};
    goto zig_block_4;
   }
   default: zig_unreachable();
  }

  zig_block_4:;
  t29 = a1.alignment;
  switch (t29) {
   case UINT8_C(0): {
    t10 = (*t1);
    t32 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t32;
    t34 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t34;
    t15 = (void const **)&t35.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t35.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t35;
    t36 = t19;
    t21 = (struct io_Writer__3878 const *)&t36;
    t19 = (*t21);
    t23 = io_Writer_writeAll__5732(t19, a0);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    t10 = (*t1);
    t37 = t10;
    t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t37;
    t38 = t1;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t38;
    t15 = (void const **)&t39.context;
    t1 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t39.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t39;
    t40 = t19;
    t21 = (struct io_Writer__3878 const *)&t40;
    t19 = (*t21);
    t22 = io_Writer_writeBytesNTimes__5736(t19, t28, t7);
    memcpy(&t23, &t22, sizeof(uint16_t));
    if (t23) {
     return t23;
    }
    goto zig_block_6;
   }
   case UINT8_C(1): {
    t5 = t7 / (uintptr_t)2ul;
    t7 = t7 + (uintptr_t)1ul;
    t7 = t7 / (uintptr_t)2ul;
    t10 = (*t1);
    t41 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t41;
    t42 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t42;
    t15 = (void const **)&t43.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t43.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t43;
    t44 = t19;
    t21 = (struct io_Writer__3878 const *)&t44;
    t19 = (*t21);
    t23 = io_Writer_writeBytesNTimes__5736(t19, t28, t5);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    t10 = (*t1);
    t45 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t45;
    t46 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t46;
    t15 = (void const **)&t47.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t47.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t47;
    t48 = t19;
    t21 = (struct io_Writer__3878 const *)&t48;
    t19 = (*t21);
    t22 = io_Writer_writeAll__5732(t19, a0);
    memcpy(&t23, &t22, sizeof(uint16_t));
    if (t23) {
     return t23;
    }
    t10 = (*t1);
    t49 = t10;
    t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t49;
    t50 = t1;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t50;
    t15 = (void const **)&t51.context;
    t1 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t51.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t51;
    t52 = t19;
    t21 = (struct io_Writer__3878 const *)&t52;
    t19 = (*t21);
    t23 = io_Writer_writeBytesNTimes__5736(t19, t28, t7);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    goto zig_block_6;
   }
   case UINT8_C(2): {
    t10 = (*t1);
    t53 = t10;
    t33 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t53;
    t54 = t33;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t54;
    t15 = (void const **)&t55.context;
    t33 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t33->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t55.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t55;
    t56 = t19;
    t21 = (struct io_Writer__3878 const *)&t56;
    t19 = (*t21);
    t23 = io_Writer_writeBytesNTimes__5736(t19, t28, t7);
    memcpy(&t22, &t23, sizeof(uint16_t));
    if (t22) {
     return t22;
    }
    t10 = (*t1);
    t57 = t10;
    t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t57;
    t58 = t1;
    t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t58;
    t15 = (void const **)&t59.context;
    t1 = (*t13);
    t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
    t17 = (void const *)t16;
    (*t15) = t17;
    t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t59.writeFn;
    (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t19 = t59;
    t60 = t19;
    t21 = (struct io_Writer__3878 const *)&t60;
    t19 = (*t21);
    t22 = io_Writer_writeAll__5732(t19, a0);
    memcpy(&t23, &t22, sizeof(uint16_t));
    if (t23) {
     return t23;
    }
    goto zig_block_6;
   }
   default: zig_unreachable();
  }

  zig_block_6:;
  goto zig_block_0;
 }
 t10 = (*t1);
 t61 = t10;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t61;
 t62 = t1;
 t13 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t62;
 t15 = (void const **)&t63.context;
 t1 = (*t13);
 t16 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t17 = (void const *)t16;
 (*t15) = t17;
 t18 = (nav__7115_52 (**)(void const *, nav__7115_39))&t63.writeFn;
 (*t18) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t19 = t63;
 t64 = t19;
 t21 = (struct io_Writer__3878 const *)&t64;
 t19 = (*t21);
 t23 = io_Writer_writeAll__5732(t19, a0);
 memcpy(&t22, &t23, sizeof(uint16_t));
 if (t22) {
  return t22;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t fmt_formatInt__anon_9201__7114(uint32_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4300 const a3, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__7114_54 t12;
 nav__7114_57 t13;
 uint32_t t4;
 uint32_t t11;
 uint32_t t2;
 uint16_t t14;
 uint16_t t15;
 bool t0;
 uint8_t t8;
 nav__7114_52 t9;
 uint8_t t10[2];
 uint8_t t1[22];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)22ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t4 = t2;
  t0 = t4 >= UINT32_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT32_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__5509(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT32_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT32_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__5509(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t4 = t2;
 t11 = (uint32_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__5525(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint32_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT32_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)22ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__7114_57));
 t14 = fmt_formatBuf__anon_9211__7115(t13, a3, a4);
 memcpy(&t15, &t14, sizeof(uint16_t));
 return t15;
}

static uint16_t fmt_formatIntValue__anon_9176__7113(uint32_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_9201__7114(a0, UINT8_C(16), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatInt__anon_9228__7117(uint16_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4300 const a3, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__7117_54 t12;
 nav__7117_57 t13;
 uint16_t t4;
 uint16_t t11;
 uint16_t t2;
 bool t0;
 uint8_t t8;
 nav__7117_52 t9;
 uint8_t t10[2];
 uint8_t t1[17];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)17ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t4 = t2;
  t0 = t4 >= UINT16_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT16_C(100);
   t8 = (uint8_t)t4;
   t9 = fmt_digits2__5509(t8);
   memcpy(t10, t9.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t10, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT16_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT16_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t8 = (uint8_t)t4;
   t8 = UINT8_C(48) + t8;
   (*t6) = t8;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = (uint8_t)t4;
  t9 = fmt_digits2__5509(t8);
  memcpy(t10, t9.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t10, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t4 = t2;
 t11 = (uint16_t)a1;
 t11 = t4 % t11;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t8 = (uint8_t)t11;
 t8 = fmt_digitToChar__5525(t8, a2);
 (*t6) = t8;
 t11 = t2;
 t4 = (uint16_t)a1;
 t4 = t11 / t4;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT16_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)17ul - t5;
 t12.ptr = t6;
 t12.len = t5;
 memcpy(&t13, &t12, sizeof(nav__7117_57));
 t11 = fmt_formatBuf__anon_9211__7115(t13, a3, a4);
 memcpy(&t4, &t11, sizeof(uint16_t));
 return t4;
}

static uint16_t fmt_formatIntValue__anon_9224__7116(uint16_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_9228__7117(a0, UINT8_C(16), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t json_stringify_outputUnicodeEscape__anon_9160__7112(uint32_t const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t19;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t20;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t26;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t18;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t25;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t6;
 void const **t8;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t9;
 void const *t10;
 nav__7112_48 (**t11)(void const *, nav__7112_50);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 struct io_Writer__3878 t13;
 struct io_Writer__3878 t21;
 struct io_Writer__3878 t22;
 struct io_Writer__3878 t27;
 struct io_Writer__3878 t28;
 struct io_Writer__3878 const *t14;
 uint32_t t17;
 uint16_t t15;
 uint16_t t16;
 uint16_t t23;
 uint16_t t24;
 bool t2;
 t0 = a1;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = a0 <= UINT32_C(65535);
 if (t2) {
  t3 = (*t1);
  t4 = t3;
  t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t4;
  t5 = t1;
  t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t5;
  t8 = (void const **)&t7.context;
  t1 = (*t6);
  t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
  t10 = (void const *)t9;
  (*t8) = t10;
  t11 = (nav__7112_48 (**)(void const *, nav__7112_50))&t7.writeFn;
  (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
  t12 = t7;
  t13 = t12;
  t14 = (struct io_Writer__3878 const *)&t13;
  t12 = (*t14);
  t15 = io_Writer_writeAll__5732(t12, (nav__7112_50){(uint8_t const *)&__anon_9166,(uintptr_t)2ul});
  memcpy(&t16, &t15, sizeof(uint16_t));
  if (t16) {
   return t16;
  }
  t16 = fmt_formatIntValue__anon_9176__7113(a0, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)4ul,false},UINT32_C(48),UINT8_C(2)}, a1);
  if (t16) {
   return t16;
  }
  goto zig_block_0;
 }
 t2 = a0 <= UINT32_C(1114111);
 debug_assert__180(t2);
 t17 = a0 - UINT32_C(65536);
 t17 = zig_shr_u32(t17, UINT8_C(10));
 t16 = (uint16_t)t17;
 t16 = t16 + UINT16_C(55296);
 t17 = a0 & UINT32_C(1023);
 t15 = (uint16_t)t17;
 t15 = t15 + UINT16_C(56320);
 t3 = (*t1);
 t18 = t3;
 t19 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t18;
 t20 = t19;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t20;
 t8 = (void const **)&t21.context;
 t19 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t19->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7112_48 (**)(void const *, nav__7112_50))&t21.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t12 = t21;
 t22 = t12;
 t14 = (struct io_Writer__3878 const *)&t22;
 t12 = (*t14);
 t23 = io_Writer_writeAll__5732(t12, (nav__7112_50){(uint8_t const *)&__anon_9166,(uintptr_t)2ul});
 memcpy(&t24, &t23, sizeof(uint16_t));
 if (t24) {
  return t24;
 }
 t16 = fmt_formatIntValue__anon_9224__7116(t16, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)4ul,false},UINT32_C(48),UINT8_C(2)}, a1);
 if (t16) {
  return t16;
 }
 t3 = (*t1);
 t25 = t3;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t25;
 t26 = t1;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t26;
 t8 = (void const **)&t27.context;
 t1 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7112_48 (**)(void const *, nav__7112_50))&t27.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t12 = t27;
 t28 = t12;
 t14 = (struct io_Writer__3878 const *)&t28;
 t12 = (*t14);
 t16 = io_Writer_writeAll__5732(t12, (nav__7112_50){(uint8_t const *)&__anon_9166,(uintptr_t)2ul});
 memcpy(&t24, &t16, sizeof(uint16_t));
 if (t24) {
  return t24;
 }
 t15 = fmt_formatIntValue__anon_9224__7116(t15, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)4ul,false},UINT32_C(48),UINT8_C(2)}, a1);
 if (t15) {
  return t15;
 }
 goto zig_block_0;

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_outputSpecialEscape__anon_9120__7111(uint8_t const a0, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a1) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t17;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t21;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t25;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t29;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t33;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t37;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t16;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t20;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t24;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t28;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t32;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t36;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t5;
 void const **t7;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t8;
 void const *t9;
 nav__7111_48 (**t10)(void const *, nav__7111_50);
 struct io_Writer__3878 t11;
 struct io_Writer__3878 t6;
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t18;
 struct io_Writer__3878 t19;
 struct io_Writer__3878 t22;
 struct io_Writer__3878 t23;
 struct io_Writer__3878 t26;
 struct io_Writer__3878 t27;
 struct io_Writer__3878 t30;
 struct io_Writer__3878 t31;
 struct io_Writer__3878 t34;
 struct io_Writer__3878 t35;
 struct io_Writer__3878 t38;
 struct io_Writer__3878 t39;
 struct io_Writer__3878 const *t13;
 uint32_t t40;
 uint16_t t14;
 uint16_t t15;
 t0 = a1;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 switch (a0) {
  case UINT8_C(92): {
   t2 = (*t1);
   t3 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t3;
   t4 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t4;
   t7 = (void const **)&t6.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t6.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t6;
   t12 = t11;
   t13 = (struct io_Writer__3878 const *)&t12;
   t11 = (*t13);
   t14 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9125,(uintptr_t)2ul});
   memcpy(&t15, &t14, sizeof(uint16_t));
   if (t15) {
    return t15;
   }
   goto zig_block_0;
  }
  case UINT8_C(34): {
   t2 = (*t1);
   t16 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t16;
   t17 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t17;
   t7 = (void const **)&t18.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t18.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t18;
   t19 = t11;
   t13 = (struct io_Writer__3878 const *)&t19;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9130,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(8): {
   t2 = (*t1);
   t20 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t20;
   t21 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t21;
   t7 = (void const **)&t22.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t22.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t22;
   t23 = t11;
   t13 = (struct io_Writer__3878 const *)&t23;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9134,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(12): {
   t2 = (*t1);
   t24 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t24;
   t25 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t25;
   t7 = (void const **)&t26.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t26.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t26;
   t27 = t11;
   t13 = (struct io_Writer__3878 const *)&t27;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9138,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(10): {
   t2 = (*t1);
   t28 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t28;
   t29 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t29;
   t7 = (void const **)&t30.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t30.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t30;
   t31 = t11;
   t13 = (struct io_Writer__3878 const *)&t31;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9142,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(13): {
   t2 = (*t1);
   t32 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t32;
   t33 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t33;
   t7 = (void const **)&t34.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t34.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t34;
   t35 = t11;
   t13 = (struct io_Writer__3878 const *)&t35;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9146,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  case UINT8_C(9): {
   t2 = (*t1);
   t36 = t2;
   t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t36;
   t37 = t1;
   t5 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t37;
   t7 = (void const **)&t38.context;
   t1 = (*t5);
   t8 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
   t9 = (void const *)t8;
   (*t7) = t9;
   t10 = (nav__7111_48 (**)(void const *, nav__7111_50))&t38.writeFn;
   (*t10) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
   t11 = t38;
   t39 = t11;
   t13 = (struct io_Writer__3878 const *)&t39;
   t11 = (*t13);
   t15 = io_Writer_writeAll__5732(t11, (nav__7111_50){(uint8_t const *)&__anon_9150,(uintptr_t)2ul});
   memcpy(&t14, &t15, sizeof(uint16_t));
   if (t14) {
    return t14;
   }
   goto zig_block_0;
  }
  default: {
   t40 = (uint32_t)a0;
   t15 = json_stringify_outputUnicodeEscape__anon_9160__7112(t40, a1);
   if (t15) {
    return t15;
   }
   goto zig_block_0;
  }
 }

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_encodeJsonStringChars__anon_9110__7110(nav__7110_39 const a0, struct json_stringify_StringifyOptions__8133 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a2) {
 nav__7110_39 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t16;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t17;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t30;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t39;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t43;
 uintptr_t t7;
 uintptr_t t8;
 uintptr_t t4;
 uintptr_t t5;
 uint64_t t9;
 uint64_t t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t15;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t29;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t38;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t42;
 nav__7110_39 t13;
 nav__7110_39 t0;
 uint8_t const *t14;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t18;
 void const **t20;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t21;
 void const *t22;
 nav__7110_54 (**t23)(void const *, nav__7110_39);
 struct io_Writer__3878 t24;
 struct io_Writer__3878 t19;
 struct io_Writer__3878 t25;
 struct io_Writer__3878 t31;
 struct io_Writer__3878 t32;
 struct io_Writer__3878 t40;
 struct io_Writer__3878 t41;
 struct io_Writer__3878 t44;
 struct io_Writer__3878 t45;
 struct io_Writer__3878 const *t26;
 uint32_t t35;
 uint32_t t37;
 nav__7110_66 t36;
 uint16_t t27;
 uint16_t t28;
 nav__7110_64 t34;
 bool t6;
 uint8_t t11;
 uint8_t t33;
 t0 = a0;
 t1 = (nav__7110_39 const *)&t0;
 t2 = a2;
 t3 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t2;
 t4 = (uintptr_t)0ul;
 t5 = (uintptr_t)0ul;
 t6 = a1.escape_unicode;
 if (t6) {
  zig_loop_16:
  t7 = t5;
  t8 = a0.len;
  t9 = t7;
  t10 = t8;
  t6 = t9 < t10;
  if (t6) {
   t8 = t5;
   t11 = a0.ptr[t8];
   switch (t11) {
    default: if ((t11 >= UINT8_C(32) && t11 <= UINT8_C(33)) || (t11 >= UINT8_C(35) && t11 <= UINT8_C(91)) || (t11 >= UINT8_C(93) && t11 <= UINT8_C(126))) {
     goto zig_block_3;
    }if (t11 == UINT8_C(92) || t11 == UINT8_C(34) || (t11 >= UINT8_C(0) && t11 <= UINT8_C(31))) {
     t12 = (*t3);
     t8 = t4;
     t7 = t5;
     t13 = (*t1);
     t14 = t13.ptr;
     t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
     t8 = t7 - t8;
     t13.ptr = t14;
     t13.len = t8;
     t15 = t12;
     t16 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t15;
     t17 = t16;
     t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t17;
     t20 = (void const **)&t19.context;
     t16 = (*t18);
     t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t16->context;
     t22 = (void const *)t21;
     (*t20) = t22;
     t23 = (nav__7110_54 (**)(void const *, nav__7110_39))&t19.writeFn;
     (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
     t24 = t19;
     t25 = t24;
     t26 = (struct io_Writer__3878 const *)&t25;
     t24 = (*t26);
     t27 = io_Writer_writeAll__5732(t24, t13);
     memcpy(&t28, &t27, sizeof(uint16_t));
     if (t28) {
      return t28;
     }
     t8 = t5;
     t11 = a0.ptr[t8];
     t28 = json_stringify_outputSpecialEscape__anon_9120__7111(t11, a2);
     if (t28) {
      return t28;
     }
     t8 = t5;
     t8 = t8 + (uintptr_t)1ul;
     t4 = t8;
     goto zig_block_3;
    }if ((t11 >= UINT8_C(127) && t11 <= UINT8_MAX)) {
     t12 = (*t3);
     t8 = t4;
     t7 = t5;
     t13 = (*t1);
     t14 = t13.ptr;
     t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
     t8 = t7 - t8;
     t13.ptr = t14;
     t13.len = t8;
     t29 = t12;
     t16 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t29;
     t30 = t16;
     t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t30;
     t20 = (void const **)&t31.context;
     t16 = (*t18);
     t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t16->context;
     t22 = (void const *)t21;
     (*t20) = t22;
     t23 = (nav__7110_54 (**)(void const *, nav__7110_39))&t31.writeFn;
     (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
     t24 = t31;
     t32 = t24;
     t26 = (struct io_Writer__3878 const *)&t32;
     t24 = (*t26);
     t28 = io_Writer_writeAll__5732(t24, t13);
     memcpy(&t27, &t28, sizeof(uint16_t));
     if (t27) {
      return t27;
     }
     t8 = t5;
     t33 = a0.ptr[t8];
     t34 = unicode_utf8ByteSequenceLength__5753(t33);
     t6 = t34.error == UINT16_C(0);
     if (t6) {
      t33 = t34.payload;
      t11 = t33;
      goto zig_block_4;
     }
     zig_unreachable();

     zig_block_4:;
     t8 = t5;
     t13 = (*t1);
     t14 = t13.ptr;
     t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
     t8 = (uintptr_t)t11;
     t13.ptr = t14;
     t13.len = t8;
     t36 = unicode_utf8Decode__5759(t13);
     t6 = t36.error == UINT16_C(0);
     if (t6) {
      t37 = t36.payload;
      t35 = t37;
      goto zig_block_5;
     }
     zig_unreachable();

     zig_block_5:;
     t27 = json_stringify_outputUnicodeEscape__anon_9160__7112(t35, a2);
     if (t27) {
      return t27;
     }
     t8 = t5;
     t11 = t11 - UINT8_C(1);
     t7 = (uintptr_t)t11;
     t7 = t8 + t7;
     t5 = t7;
     t7 = t5;
     t7 = t7 + (uintptr_t)1ul;
     t4 = t7;
     goto zig_block_3;
    }zig_unreachable();
   }

   zig_block_3:;
   t8 = t5;
   t8 = t8 + (uintptr_t)1ul;
   t5 = t8;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_16;

  zig_block_1:;
  goto zig_block_0;
 }
 zig_loop_172:
 t8 = t5;
 t7 = a0.len;
 t10 = t8;
 t9 = t7;
 t6 = t10 < t9;
 if (t6) {
  t7 = t5;
  t33 = a0.ptr[t7];
  switch (t33) {
   default: if ((t33 >= UINT8_C(32) && t33 <= UINT8_C(33)) || (t33 >= UINT8_C(35) && t33 <= UINT8_C(91)) || (t33 >= UINT8_C(93) && t33 <= UINT8_MAX)) {
    goto zig_block_8;
   }if (t33 == UINT8_C(92) || t33 == UINT8_C(34) || (t33 >= UINT8_C(0) && t33 <= UINT8_C(31))) {
    t12 = (*t3);
    t7 = t4;
    t8 = t5;
    t13 = (*t1);
    t14 = t13.ptr;
    t14 = (uint8_t const *)(((uintptr_t)t14) + (t7*sizeof(uint8_t)));
    t7 = t8 - t7;
    t13.ptr = t14;
    t13.len = t7;
    t38 = t12;
    t16 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t38;
    t39 = t16;
    t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t39;
    t20 = (void const **)&t40.context;
    t16 = (*t18);
    t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t16->context;
    t22 = (void const *)t21;
    (*t20) = t22;
    t23 = (nav__7110_54 (**)(void const *, nav__7110_39))&t40.writeFn;
    (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
    t24 = t40;
    t41 = t24;
    t26 = (struct io_Writer__3878 const *)&t41;
    t24 = (*t26);
    t28 = io_Writer_writeAll__5732(t24, t13);
    memcpy(&t27, &t28, sizeof(uint16_t));
    if (t27) {
     return t27;
    }
    t7 = t5;
    t33 = a0.ptr[t7];
    t27 = json_stringify_outputSpecialEscape__anon_9120__7111(t33, a2);
    if (t27) {
     return t27;
    }
    t7 = t5;
    t7 = t7 + (uintptr_t)1ul;
    t4 = t7;
    goto zig_block_8;
   }zig_unreachable();
  }

  zig_block_8:;
  t7 = t5;
  t7 = t7 + (uintptr_t)1ul;
  t5 = t7;
  goto zig_block_7;
 }
 goto zig_block_6;

 zig_block_7:;
 goto zig_loop_172;

 zig_block_6:;
 goto zig_block_0;

 zig_block_0:;
 t12 = (*t3);
 t8 = t4;
 t7 = a0.len;
 t13 = (*t1);
 t14 = t13.ptr;
 t14 = (uint8_t const *)(((uintptr_t)t14) + (t8*sizeof(uint8_t)));
 t8 = t7 - t8;
 t13.ptr = t14;
 t13.len = t8;
 t42 = t12;
 t3 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t42;
 t43 = t3;
 t18 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t43;
 t20 = (void const **)&t44.context;
 t3 = (*t18);
 t21 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t3->context;
 t22 = (void const *)t21;
 (*t20) = t22;
 t23 = (nav__7110_54 (**)(void const *, nav__7110_39))&t44.writeFn;
 (*t23) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t24 = t44;
 t45 = t24;
 t26 = (struct io_Writer__3878 const *)&t45;
 t24 = (*t26);
 t28 = io_Writer_writeAll__5732(t24, t13);
 memcpy(&t27, &t28, sizeof(uint16_t));
 if (t27) {
  return t27;
 }
 return 0;
}

static uint16_t json_stringify_encodeJsonString__anon_9100__7109(nav__7109_39 const a0, struct json_stringify_StringifyOptions__8133 const a1, struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const a2) {
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t1;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t18;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t2;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t17;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t6;
 void const **t8;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t9;
 void const *t10;
 nav__7109_53 (**t11)(void const *, nav__7109_39);
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t7;
 struct io_Writer__3878 t13;
 struct io_Writer__3878 t19;
 struct io_Writer__3878 t20;
 struct io_Writer__3878 const *t14;
 uint16_t t15;
 uint16_t t16;
 t0 = a2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t0;
 t2 = (*t1);
 t3 = t2;
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t3;
 t5 = t4;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t5;
 t8 = (void const **)&t7.context;
 t4 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t4->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7109_53 (**)(void const *, nav__7109_39))&t7.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t12 = t7;
 t13 = t12;
 t14 = (struct io_Writer__3878 const *)&t13;
 t12 = (*t14);
 t15 = io_Writer_writeByte__5734(t12, UINT8_C(34));
 memcpy(&t16, &t15, sizeof(uint16_t));
 if (t16) {
  return t16;
 }
 t16 = json_stringify_encodeJsonStringChars__anon_9110__7110(a0, a1, a2);
 if (t16) {
  return t16;
 }
 t2 = (*t1);
 t17 = t2;
 t1 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t17;
 t18 = t1;
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t18;
 t8 = (void const **)&t19.context;
 t1 = (*t6);
 t9 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t1->context;
 t10 = (void const *)t9;
 (*t8) = t10;
 t11 = (nav__7109_53 (**)(void const *, nav__7109_39))&t19.writeFn;
 (*t11) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t12 = t19;
 t20 = t12;
 t14 = (struct io_Writer__3878 const *)&t20;
 t12 = (*t14);
 t16 = io_Writer_writeByte__5734(t12, UINT8_C(34));
 memcpy(&t15, &t16, sizeof(uint16_t));
 if (t15) {
  return t15;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, nav__7015_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_stringify_StringifyOptions__8133 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t7;
 uint8_t *t8;
 uint16_t t3;
 struct json_stringify_StringifyOptions__8133 t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7007(t2);
 if (t3) {
  return t3;
 }
 t4 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
 t5 = (*t4);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&a0->stream;
 t7 = (*t6);
 t3 = json_stringify_encodeJsonString__anon_9100__7109(a1, t5, t7);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t8 = (uint8_t *)&t2->next_punctuation;
 (*t8) = UINT8_C(3);
 return 0;
}

static uint16_t json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29_jsonStringif__7118(struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a1) {
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Iterator__6152 t6;
 nav__7118_65 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29_Entry__6161 t12;
 nav__7118_56 *t13;
 nav__7118_56 t14;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 *t15;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 t16;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 t0;
 uint16_t t5;
 bool t11;
 t0 = a0;
 t1 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t2;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(t4);
 if (t5) {
  return t5;
 }
 t7 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const *)&t1->map;
 t8 = (*t7);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3800(t8);
 t6 = t9;
 zig_loop_22:
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__6161(&t6);
 t11 = t10.is_null != true;
 if (t11) {
  t12 = t10.payload;
  t4 = (*t3);
  t13 = t12.key_ptr;
  t14 = (*t13);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t4, t14);
  if (t5) {
   return t5;
  }
  t4 = (*t3);
  t15 = t12.value_ptr;
  t16 = (*t15);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7119(t4, t16);
  if (t5) {
   return t5;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_22;

 zig_block_0:;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(t4);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7037(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 const a1) {
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 const *t2;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 t3;
 struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29__861 const *)&t1;
 t3 = (*t2);
 t4 = json_hashmap_ArrayHashMap_28json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_29_jsonStringif__7118(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29_jsonStringify__anon_9275__7120(struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a1) {
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Iterator__6205 t6;
 nav__7120_65 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Entry__6213 t12;
 nav__7120_56 *t13;
 nav__7120_56 t14;
 struct environment_linked_libraries_root_LibraryEntry__864 *t15;
 struct environment_linked_libraries_root_LibraryEntry__864 t16;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 t0;
 uint16_t t5;
 bool t11;
 t0 = a0;
 t1 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t2;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(t4);
 if (t5) {
  return t5;
 }
 t7 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t1->map;
 t8 = (*t7);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4190(t8);
 t6 = t9;
 zig_loop_22:
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__6165(&t6);
 t11 = t10.is_null != true;
 if (t11) {
  t12 = t10.payload;
  t4 = (*t3);
  t13 = t12.key_ptr;
  t14 = (*t13);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t4, t14);
  if (t5) {
   return t5;
  }
  t4 = (*t3);
  t15 = t12.value_ptr;
  t16 = (*t15);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7121(t4, t16);
  if (t5) {
   return t5;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_22;

 zig_block_0:;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(t4);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7038(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 const a1) {
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 const *t2;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 t3;
 struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29__866 const *)&t1;
 t3 = (*t2);
 t4 = json_hashmap_ArrayHashMap_28environment_linked_libraries_root_LibraryEntry_29_jsonStringify__anon_9275__7120(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint8_t *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t11;
 void const **t13;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t14;
 void const *t15;
 nav__7002_55 (**t16)(void const *, nav__7002_57);
 struct io_Writer__3878 t17;
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t18;
 struct io_Writer__3878 const *t19;
 uint16_t t5;
 uint16_t t20;
 uint8_t t4;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7004(t2, UINT8_C(0));
 t3 = (uint8_t *)&a0->next_punctuation;
 t4 = (*t3);
 switch (t4) {
  case UINT8_C(1): {
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7005(t2);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  case UINT8_C(0):
  case UINT8_C(3): {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t2 = (*t1);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
 t7 = (*t6);
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t8;
 t10 = t9;
 t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t10;
 t13 = (void const **)&t12.context;
 t9 = (*t11);
 t14 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t9->context;
 t15 = (void const *)t14;
 (*t13) = t15;
 t16 = (nav__7002_55 (**)(void const *, nav__7002_57))&t12.writeFn;
 (*t16) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t17 = t12;
 t18 = t17;
 t19 = (struct io_Writer__3878 const *)&t18;
 t17 = (*t19);
 t5 = io_Writer_writeByte__5734(t17, UINT8_C(125));
 memcpy(&t20, &t5, sizeof(uint16_t));
 if (t20) {
  return t20;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(t2);
 return 0;
}

static uintptr_t array_list_ArrayListAlignedUnmanaged_28environment_linked_libraries_root_LibraryInfo_2cnull_29_growCapacit__7099(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)2ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static uint16_t array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_ensureTotalCapacityP__4385(struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const a0, uintptr_t const a1) {
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t7;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *t0;
 struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 t8;
 nav__4385_44 t9;
 nav__4385_44 t12;
 nav__4385_44 t20;
 nav__4385_44 t21;
 nav__4385_44 t18;
 struct mem_Allocator__599 *t10;
 struct mem_Allocator__599 t11;
 nav__4385_44 *t13;
 struct environment_linked_libraries_root_LibraryInfo__946 **t14;
 struct environment_linked_libraries_root_LibraryInfo__946 *t15;
 nav__4385_56 t16;
 nav__4385_44 const *t19;
 uint16_t t17;
 bool t6;
 t0 = a0;
 t1 = (struct array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29__948 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a1;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t8 = (*t7);
 t9 = array_list_ArrayListAligned_28environment_linked_libraries_root_LibraryInfo_2cnull_29_allocatedSlice__4395(t8);
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 t12 = mem_Allocator_remap__anon_9292__7122(t11, t9, a1);
 t6 = t12.ptr != NULL;
 if (t6) {
  t9 = t12;
  t7 = (*t1);
  t13 = (nav__4385_44 *)&t7->items;
  t14 = &t13->ptr;
  t15 = t9.ptr;
  (*t14) = t15;
  t7 = (*t1);
  t2 = (uintptr_t *)&t7->capacity;
  t3 = t9.len;
  (*t2) = t3;
  goto zig_block_1;
 }
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 t16 = mem_Allocator_alignedAlloc__anon_9296__7123(t11, a1);
 if (t16.error) {
  t17 = t16.error;
  return t17;
 }
 t12 = t16.payload;
 t18 = t12;
 t19 = (nav__4385_44 const *)&t18;
 t13 = (nav__4385_44 *)&a0->items;
 t20 = (*t13);
 t3 = t20.len;
 t20 = (*t19);
 t15 = t20.ptr;
 t15 = (struct environment_linked_libraries_root_LibraryInfo__946 *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryInfo__946)));
 t20.ptr = t15;
 t20.len = t3;
 t13 = (nav__4385_44 *)&a0->items;
 t21 = (*t13);
 t15 = t21.ptr;
 if (t20.len != 0) memcpy(t20.ptr, t15, t20.len * sizeof(struct environment_linked_libraries_root_LibraryInfo__946));
 t7 = (*t1);
 t10 = (struct mem_Allocator__599 *)&t7->allocator;
 t11 = (*t10);
 mem_Allocator_free__anon_6701__6275(t11, t9);
 t7 = (*t1);
 t13 = (nav__4385_44 *)&t7->items;
 t14 = &t13->ptr;
 t15 = t12.ptr;
 (*t14) = t15;
 t7 = (*t1);
 t2 = (uintptr_t *)&t7->capacity;
 t3 = t12.len;
 (*t2) = t3;
 goto zig_block_1;

 zig_block_1:;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4332(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *t1;
 struct mem_Allocator__599 const *t3;
 uintptr_t *t4;
 uintptr_t t5;
 uint64_t t6;
 uint64_t t7;
 struct mem_Allocator__599 t9;
 struct mem_Allocator__599 t2;
 nav__4332_54 t10;
 nav__4332_52 t12;
 nav__4332_52 t15;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t13;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t14;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t18;
 uint8_t **t16;
 uint8_t *t17;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t19;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t20;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t22;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t21;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t23;
 nav__4332_62 t24;
 nav__4332_62 t25;
 uint32_t *t26;
 nav__4332_67 t27;
 nav__4332_67 t28;
 nav__4332_65 *t29;
 nav__4332_72 t30;
 nav__4332_72 t31;
 struct environment_linked_libraries_root_LibraryEntry__864 *t32;
 uint16_t t11;
 bool t8;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *)&t0;
 t2 = a1;
 t3 = (struct mem_Allocator__599 const *)&t2;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t6 = a2;
 t7 = t5;
 t8 = t6 >= t7;
 debug_assert__180(t8);
 t9 = (*t3);
 t5 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4339(a2);
 t10 = mem_Allocator_alignedAlloc__anon_7569__6383(t9, t5);
 if (t10.error) {
  t11 = t10.error;
  return t11;
 }
 t12 = t10.payload;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t7 = t5;
 t8 = t7 == UINT64_C(0);
 if (t8) {
  t9 = (*t3);
  t13 = (*t1);
  t14 = (*t13);
  t15 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4340(t14);
  mem_Allocator_free__anon_6654__6273(t9, t15);
  t13 = (*t1);
  t16 = (uint8_t **)&t13->bytes;
  t17 = t12.ptr;
  (*t16) = t17;
  t13 = (*t1);
  t4 = (uintptr_t *)&t13->capacity;
  (*t4) = a2;
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t17 = t12.ptr;
 t4 = (uintptr_t *)&a0->len;
 t5 = (*t4);
 t14.bytes = t17;
 t14.len = t5;
 t14.capacity = a2;
 t18 = t14;
 t13 = (*t1);
 t14 = (*t13);
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t14);
 t20 = t19;
 t21 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t20;
 t14 = t18;
 t19 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t14);
 t22 = t19;
 t23 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t22;
 t19 = (*t23);
 t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t19);
 t19 = (*t21);
 t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t19);
 t26 = t25.ptr;
 if (t24.len != 0) memcpy(t24.ptr, t26, t24.len * sizeof(uint32_t));
 t19 = (*t23);
 t27 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t19);
 t19 = (*t21);
 t28 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t19);
 t29 = t28.ptr;
 if (t27.len != 0) memcpy(t27.ptr, t29, t27.len * sizeof(nav__4332_65));
 t19 = (*t23);
 t30 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t19);
 t19 = (*t21);
 t31 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t19);
 t32 = t31.ptr;
 if (t30.len != 0) memcpy(t30.ptr, t32, t30.len * sizeof(struct environment_linked_libraries_root_LibraryEntry__864));
 t9 = (*t3);
 t13 = (*t1);
 t14 = (*t13);
 t12 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4340(t14);
 mem_Allocator_free__anon_6654__6273(t9, t12);
 t14 = t18;
 (*a0) = t14;
 return 0;
}

static uint16_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4328(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const a0, struct mem_Allocator__599 const a1, uintptr_t const a2) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uint64_t t4;
 uint64_t t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t0;
 uint16_t t8;
 uint16_t t9;
 bool t6;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *)&t0;
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t4 = t3;
 t5 = a2;
 t6 = t4 >= t5;
 if (t6) {
  return 0;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t2 = (uintptr_t *)&a0->capacity;
 t3 = (*t2);
 t3 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4330(t3, a2);
 t8 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4332(t7, a1, t3);
 memcpy(&t9, &t8, sizeof(uint16_t));
 return t9;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4294(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct array_hash_map_IndexHeader__892 t5;
 uint8_t t6;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = (*t4);
 t6 = array_hash_map_IndexHeader_capacityIndexType__3973(t5);
 switch (t6) {
  case UINT8_C(0): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7124(t7, a1);
   return;
  }
  case UINT8_C(1): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7125(t7, a1);
   return;
  }
  case UINT8_C(2): {
   t7 = (*t1);
   array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7126(t7, a1);
   return;
  }
  default: zig_unreachable();
 }
}

static uint32_t array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(nav__7100_40 const a0) {
 uint32_t t0;
 t0 = array_hash_map_StringContext_hash__3773(a0);
 return t0;
}

static nav__7101_39 multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t3[3];
 uint8_t *t4;
 uint32_t *t5;
 uint32_t *t6;
 uint32_t *const *t7;
 nav__7101_39 t8;
 bool t2;
 t0 = a0.capacity;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  return (nav__7101_39){(uint32_t *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 memcpy(t3, a0.ptrs, sizeof(uint8_t *[3]));
 t4 = t3[(uintptr_t)0ul];
 t5 = (uint32_t *)t4;
 t6 = t5;
 t7 = (uint32_t *const *)&t6;
 t0 = a0.len;
 t5 = (*t7);
 t5 = (uint32_t *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint32_t)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static bool array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(nav__7102_40 const a0, nav__7102_40 const a1, uintptr_t const a2) {
 bool t0;
 t0 = array_hash_map_StringContext_eql__3774(a0, a1, a2);
 return t0;
}

static nav__7103_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7103(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7103_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7103_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 nav__7103_38 t8;
 bool t9;
 struct array_hash_map_Index_28u8_29__6735 t11;
 uint8_t t12;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7127(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__7103_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__7104_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7104(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7104_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7104_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 nav__7104_38 t8;
 struct array_hash_map_Index_28u16_29__6739 t11;
 uint16_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7128(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__7104_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static nav__7105_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7105(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7105_41 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_IndexHeader__892 *t4;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7105_58 t5;
 uintptr_t t6;
 uintptr_t t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 nav__7105_38 t8;
 struct array_hash_map_Index_28u32_29__6743 t11;
 uint32_t t12;
 bool t9;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t3);
 t5 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t4);
 t7 = (*t1);
 t8 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7129(t7, a1, a2, t5);
 t9 = t8.is_null != true;
 if (t9) {
  t10 = t8.payload;
  t6 = t10;
  goto zig_block_0;
 }
 return (nav__7105_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};

 zig_block_0:;
 t11 = t5.ptr[t6];
 t12 = t11.entry_index;
 t6 = (uintptr_t)t12;
 t8.is_null = false;
 t8.payload = t6;
 return t8;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const a0) {
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *t1;
 uintptr_t *t2;
 uintptr_t t3;
 uintptr_t t4;
 uint64_t t5;
 uint64_t t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t0;
 bool t7;
 t0 = a0;
 t1 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *const *)&t0;
 t2 = (uintptr_t *)&a0->len;
 t3 = (*t2);
 t2 = (uintptr_t *)&a0->capacity;
 t4 = (*t2);
 t5 = t3;
 t6 = t4;
 t7 = t5 < t6;
 debug_assert__180(t7);
 t2 = (uintptr_t *)&a0->len;
 t4 = (*t2);
 t8 = (*t1);
 t2 = (uintptr_t *)&t8->len;
 t3 = (*t2);
 t3 = t3 + (uintptr_t)1ul;
 (*t2) = t3;
 return t4;
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7106(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, nav__7106_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t9;
 nav__7106_66 t10;
 nav__7106_66 t11;
 nav__7106_66 const *t12;
 nav__7106_69 t13;
 nav__7106_69 t14;
 nav__7106_69 const *t15;
 nav__7106_72 t16;
 nav__7106_72 t17;
 nav__7106_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7106_77 t20;
 nav__7106_77 t55;
 nav__7106_77 t21;
 nav__7106_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u8_29__6735 *t35;
 uint8_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__7106_42 **t42;
 nav__7106_42 *const *t43;
 nav__7106_42 *t44;
 struct environment_linked_libraries_root_LibraryEntry__864 **t45;
 struct environment_linked_libraries_root_LibraryEntry__864 *const *t46;
 struct environment_linked_libraries_root_LibraryEntry__864 *t47;
 uintptr_t *t48;
 nav__7106_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__892 t32;
 bool t30;
 bool t50;
 struct array_hash_map_Index_28u8_29__6735 t34;
 struct array_hash_map_Index_28u8_29__6735 t33;
 uint8_t t37;
 uint8_t t53;
 uint8_t t27;
 uint8_t t56;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t7);
 t11 = t10;
 t12 = (nav__7106_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t7);
 t14 = t13;
 t15 = (nav__7106_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t7);
 t17 = t16;
 t18 = (nav__7106_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t19);
 t21 = t20;
 t22 = (nav__7106_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT8_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u8_29_isEmpty__6280(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint8_t *)&t35->entry_index;
   t37 = (uint8_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__7106_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__7106_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__7106_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint8_t *)&t35->entry_index;
   t53 = (uint8_t)t24;
   (*t36) = t53;
   t36 = (uint8_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT8_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u8_29_isEmpty__6280(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint8_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__7106_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__7106_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint8_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint8_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT8_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT8_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7107(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, nav__7107_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t9;
 nav__7107_66 t10;
 nav__7107_66 t11;
 nav__7107_66 const *t12;
 nav__7107_69 t13;
 nav__7107_69 t14;
 nav__7107_69 const *t15;
 nav__7107_72 t16;
 nav__7107_72 t17;
 nav__7107_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7107_77 t20;
 nav__7107_77 t55;
 nav__7107_77 t21;
 nav__7107_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t54;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u16_29__6739 *t35;
 uint16_t *t36;
 uint32_t *const *t38;
 uint32_t *t39;
 bool *t41;
 nav__7107_42 **t42;
 nav__7107_42 *const *t43;
 nav__7107_42 *t44;
 struct environment_linked_libraries_root_LibraryEntry__864 **t45;
 struct environment_linked_libraries_root_LibraryEntry__864 *const *t46;
 struct environment_linked_libraries_root_LibraryEntry__864 *t47;
 uintptr_t *t48;
 nav__7107_42 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t40;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t52;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t57;
 uint32_t t23;
 uint32_t t49;
 struct array_hash_map_IndexHeader__892 t32;
 struct array_hash_map_Index_28u16_29__6739 t34;
 struct array_hash_map_Index_28u16_29__6739 t33;
 uint16_t t37;
 uint16_t t53;
 uint16_t t27;
 uint16_t t56;
 bool t30;
 bool t50;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t7);
 t11 = t10;
 t12 = (nav__7107_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t7);
 t14 = t13;
 t15 = (nav__7107_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t7);
 t17 = t16;
 t18 = (nav__7107_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t19);
 t21 = t20;
 t22 = (nav__7107_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT16_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u16_29_isEmpty__6285(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint16_t *)&t35->entry_index;
   t37 = (uint16_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t25];
   (*t39) = t23;
   t41 = (bool *)&t40.found_existing;
   (*t41) = false;
   t42 = (nav__7107_42 **)&t40.key_ptr;
   t43 = &t15->ptr;
   t44 = (*t43);
   t44 = (nav__7107_42 *)&t44[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t40.value_ptr;
   t46 = &t18->ptr;
   t47 = (*t46);
   t47 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t47[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t40.index;
   (*t48) = t25;
   return t40;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t49 = t10.ptr[t24];
  t30 = t23 == t49;
  if (t30) {
   t24 = (uintptr_t)t37;
   t51 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t51, t24);
   t50 = t30;
   goto zig_block_4;
  }
  t50 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t50) {
   t41 = (bool *)&t52.found_existing;
   (*t41) = true;
   t42 = (nav__7107_42 **)&t52.key_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t13 = (*t15);
   t44 = &t13.ptr[t25];
   (*t42) = t44;
   t45 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t52.value_ptr;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   t16 = (*t18);
   t47 = &t16.ptr[t25];
   (*t45) = t47;
   t48 = (uintptr_t *)&t52.index;
   t34 = t33;
   t37 = t34.entry_index;
   t25 = (uintptr_t)t37;
   (*t48) = t25;
   return t52;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t53 = t27;
  t30 = t37 < t53;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t5);
   t38 = &t12->ptr;
   t39 = (*t38);
   t39 = (uint32_t *)&t39[t24];
   (*t39) = t23;
   t54 = t31;
   t55 = (*t22);
   t35 = &t55.ptr[t54];
   t36 = (uint16_t *)&t35->entry_index;
   t53 = (uint16_t)t24;
   (*t36) = t53;
   t36 = (uint16_t *)&t35->distance_from_start_index;
   t53 = t27;
   (*t36) = t53;
   t34 = t33;
   t53 = t34.distance_from_start_index;
   t27 = t53;
   t34 = t33;
   t53 = t34.entry_index;
   t56 = t53;
   t54 = t26;
   t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
   t26 = t54;
   t53 = t27;
   t53 = t53 + UINT16_C(1);
   t27 = t53;
   zig_loop_200:
   t54 = t26;
   t29 = t54;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t54 = t26;
    t54 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t54);
    t31 = t54;
    t54 = t31;
    t34 = t20.ptr[t54];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u16_29_isEmpty__6285(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint16_t *)&t35->entry_index;
     t53 = t56;
     (*t36) = t53;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t53 = t27;
     (*t36) = t53;
     t41 = (bool *)&t57.found_existing;
     (*t41) = false;
     t42 = (nav__7107_42 **)&t57.key_ptr;
     t43 = &t15->ptr;
     t44 = (*t43);
     t44 = (nav__7107_42 *)&t44[t24];
     (*t42) = t44;
     t45 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t57.value_ptr;
     t46 = &t18->ptr;
     t47 = (*t46);
     t47 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t47[t24];
     (*t45) = t47;
     t48 = (uintptr_t *)&t57.index;
     (*t48) = t24;
     return t57;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t53 = t34.distance_from_start_index;
    t37 = t27;
    t30 = t53 < t37;
    if (t30) {
     t54 = t31;
     t55 = (*t22);
     t35 = &t55.ptr[t54];
     t36 = (uint16_t *)&t35->entry_index;
     t37 = t56;
     (*t36) = t37;
     t36 = (uint16_t *)&t35->distance_from_start_index;
     t37 = t27;
     (*t36) = t37;
     t34 = t33;
     t37 = t34.entry_index;
     t56 = t37;
     t34 = t33;
     t37 = t34.distance_from_start_index;
     t27 = t37;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t54 = t26;
    t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
    t26 = t54;
    t53 = t27;
    t53 = t53 + UINT16_C(1);
    t27 = t53;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t54 = t26;
  t54 = zig_addw_u64(t54, (uintptr_t)1ul, UINT8_C(64));
  t26 = t54;
  t53 = t27;
  t53 = t53 + UINT16_C(1);
  t27 = t53;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7108(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, nav__7108_42 const a1, struct array_hash_map_IndexHeader__892 *const a2) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t9;
 nav__7108_66 t10;
 nav__7108_66 t11;
 nav__7108_66 const *t12;
 nav__7108_69 t13;
 nav__7108_69 t14;
 nav__7108_69 const *t15;
 nav__7108_72 t16;
 nav__7108_72 t17;
 nav__7108_72 const *t18;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7108_77 t20;
 nav__7108_77 t53;
 nav__7108_77 t21;
 nav__7108_77 const *t22;
 uintptr_t t24;
 uintptr_t t25;
 uintptr_t t52;
 uintptr_t t26;
 uintptr_t t31;
 uint64_t t28;
 uint64_t t29;
 struct array_hash_map_Index_28u32_29__6743 *t35;
 uint32_t *t36;
 uint32_t *const *t38;
 bool *t40;
 nav__7108_42 **t41;
 nav__7108_42 *const *t42;
 nav__7108_42 *t43;
 struct environment_linked_libraries_root_LibraryEntry__864 **t44;
 struct environment_linked_libraries_root_LibraryEntry__864 *const *t45;
 struct environment_linked_libraries_root_LibraryEntry__864 *t46;
 uintptr_t *t47;
 nav__7108_42 t50;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t39;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t51;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_GetOrPutResult__8075 t55;
 uint32_t t23;
 uint32_t t37;
 uint32_t t48;
 uint32_t t27;
 uint32_t t54;
 struct array_hash_map_IndexHeader__892 t32;
 struct array_hash_map_Index_28u32_29__6743 t34;
 struct array_hash_map_Index_28u32_29__6743 t33;
 bool t30;
 bool t49;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t7);
 t11 = t10;
 t12 = (nav__7108_66 const *)&t11;
 t7 = (*t9);
 t13 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t7);
 t14 = t13;
 t15 = (nav__7108_69 const *)&t14;
 t7 = (*t9);
 t16 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6200(t7);
 t17 = t16;
 t18 = (nav__7108_72 const *)&t17;
 t19 = (*t3);
 t20 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t19);
 t21 = t20;
 t22 = (nav__7108_77 const *)&t21;
 t23 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t24 = array_hash_map_safeTruncate__anon_7783__6395(t23);
 t25 = t20.len;
 t25 = zig_addw_u64(t24, t25, UINT8_C(64));
 t26 = t24;
 t27 = UINT32_C(0);
 zig_loop_55:
 t24 = t26;
 t28 = t24;
 t29 = t25;
 t30 = t28 != t29;
 if (t30) {
  t19 = (*t3);
  t32 = (*t19);
  t24 = t26;
  t24 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t24);
  t31 = t24;
  t24 = t31;
  t34 = t20.ptr[t24];
  t33 = t34;
  t34 = t33;
  t30 = array_hash_map_Index_28u32_29_isEmpty__6290(t34);
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
   t25 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t5);
   t24 = t31;
   t20 = (*t22);
   t35 = &t20.ptr[t24];
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t37 = t27;
   (*t36) = t37;
   t36 = (uint32_t *)&t35->entry_index;
   t37 = (uint32_t)t25;
   (*t36) = t37;
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t25];
   (*t36) = t23;
   t40 = (bool *)&t39.found_existing;
   (*t40) = false;
   t41 = (nav__7108_42 **)&t39.key_ptr;
   t42 = &t15->ptr;
   t43 = (*t42);
   t43 = (nav__7108_42 *)&t43[t25];
   (*t41) = t43;
   t44 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t39.value_ptr;
   t45 = &t18->ptr;
   t46 = (*t45);
   t46 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t46[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t39.index;
   (*t47) = t25;
   return t39;
  }
  goto zig_block_2;

  zig_block_2:;
  t34 = t33;
  t37 = t34.entry_index;
  t24 = (uintptr_t)t37;
  t48 = t10.ptr[t24];
  t30 = t23 == t48;
  if (t30) {
   t24 = (uintptr_t)t37;
   t50 = t13.ptr[t24];
   t24 = (uintptr_t)t37;
   t30 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t50, t24);
   t49 = t30;
   goto zig_block_4;
  }
  t49 = false;
  goto zig_block_4;

  zig_block_4:;
  if (t49) {
   t40 = (bool *)&t51.found_existing;
   (*t40) = true;
   t41 = (nav__7108_42 **)&t51.key_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t13 = (*t15);
   t43 = &t13.ptr[t25];
   (*t41) = t43;
   t44 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t51.value_ptr;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   t16 = (*t18);
   t46 = &t16.ptr[t25];
   (*t44) = t46;
   t47 = (uintptr_t *)&t51.index;
   t34 = t33;
   t23 = t34.entry_index;
   t25 = (uintptr_t)t23;
   (*t47) = t25;
   return t51;
  }
  goto zig_block_3;

  zig_block_3:;
  t34 = t33;
  t37 = t34.distance_from_start_index;
  t48 = t27;
  t30 = t37 < t48;
  if (t30) {
   t4 = (*t1);
   t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
   t24 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4317(t5);
   t38 = &t12->ptr;
   t36 = (*t38);
   t36 = (uint32_t *)&t36[t24];
   (*t36) = t23;
   t52 = t31;
   t53 = (*t22);
   t35 = &t53.ptr[t52];
   t36 = (uint32_t *)&t35->entry_index;
   t23 = (uint32_t)t24;
   (*t36) = t23;
   t36 = (uint32_t *)&t35->distance_from_start_index;
   t23 = t27;
   (*t36) = t23;
   t34 = t33;
   t23 = t34.distance_from_start_index;
   t27 = t23;
   t34 = t33;
   t23 = t34.entry_index;
   t54 = t23;
   t52 = t26;
   t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
   t26 = t52;
   t23 = t27;
   t23 = t23 + UINT32_C(1);
   t27 = t23;
   zig_loop_200:
   t52 = t26;
   t29 = t52;
   t28 = t25;
   t30 = t29 != t28;
   if (t30) {
    t19 = (*t3);
    t32 = (*t19);
    t52 = t26;
    t52 = array_hash_map_IndexHeader_constrainIndex__3971(t32, t52);
    t31 = t52;
    t52 = t31;
    t34 = t20.ptr[t52];
    t33 = t34;
    t34 = t33;
    t30 = array_hash_map_Index_28u32_29_isEmpty__6290(t34);
    if (t30) {
     t25 = t31;
     t20 = (*t22);
     t35 = &t20.ptr[t25];
     t36 = (uint32_t *)&t35->entry_index;
     t23 = t54;
     (*t36) = t23;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t23 = t27;
     (*t36) = t23;
     t40 = (bool *)&t55.found_existing;
     (*t40) = false;
     t41 = (nav__7108_42 **)&t55.key_ptr;
     t42 = &t15->ptr;
     t43 = (*t42);
     t43 = (nav__7108_42 *)&t43[t24];
     (*t41) = t43;
     t44 = (struct environment_linked_libraries_root_LibraryEntry__864 **)&t55.value_ptr;
     t45 = &t18->ptr;
     t46 = (*t45);
     t46 = (struct environment_linked_libraries_root_LibraryEntry__864 *)&t46[t24];
     (*t44) = t46;
     t47 = (uintptr_t *)&t55.index;
     (*t47) = t24;
     return t55;
    }
    goto zig_block_8;

    zig_block_8:;
    t34 = t33;
    t23 = t34.distance_from_start_index;
    t48 = t27;
    t30 = t23 < t48;
    if (t30) {
     t52 = t31;
     t53 = (*t22);
     t35 = &t53.ptr[t52];
     t36 = (uint32_t *)&t35->entry_index;
     t48 = t54;
     (*t36) = t48;
     t36 = (uint32_t *)&t35->distance_from_start_index;
     t48 = t27;
     (*t36) = t48;
     t34 = t33;
     t48 = t34.entry_index;
     t54 = t48;
     t34 = t33;
     t48 = t34.distance_from_start_index;
     t27 = t48;
     goto zig_block_9;
    }
    goto zig_block_9;

    zig_block_9:;
    t52 = t26;
    t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
    t26 = t52;
    t23 = t27;
    t23 = t23 + UINT32_C(1);
    t27 = t23;
    goto zig_block_7;
   }
   goto zig_block_6;

   zig_block_7:;
   goto zig_loop_200;

   zig_block_6:;
   zig_unreachable();
  }
  goto zig_block_5;

  zig_block_5:;
  t52 = t26;
  t52 = zig_addw_u64(t52, (uintptr_t)1ul, UINT8_C(64));
  t26 = t52;
  t37 = t27;
  t37 = t37 + UINT32_C(1);
  t27 = t37;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_55;

 zig_block_0:;
 zig_unreachable();
}

static nav__7010_38 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7010(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *t0;
 t0 = a0;
 return (nav__7010_38){true,false};
}

static bool json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7011(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 const *const a0) {
 uintptr_t const *t0;
 uintptr_t t1;
 uint64_t t2;
 uint8_t const *t5;
 bool t3;
 bool t4;
 uint8_t t6;
 t0 = (uintptr_t const *)&a0->indent_level;
 t1 = (*t0);
 t2 = t1;
 t3 = t2 == UINT64_C(0);
 if (t3) {
  t5 = (uint8_t const *)&a0->next_punctuation;
  t6 = (*t5);
  t3 = t6 == UINT8_C(2);
  t4 = t3;
  goto zig_block_0;
 }
 t4 = false;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static uint16_t io_Writer_writeByte__5734(struct io_Writer__3878 const a0, uint8_t const a1) {
 struct io_Writer__3878 const *t1;
 uint8_t const (*t4)[1];
 struct io_Writer__3878 t5;
 struct io_Writer__3878 t0;
 nav__5734_43 t6;
 uint16_t t7;
 uint8_t t2[1];
 uint8_t t3[1];
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t2[0] = a1;
 memcpy((char *)&t3, t2, sizeof(uint8_t[1]));
 t4 = (uint8_t const (*)[1])&t3;
 t5 = (*t1);
 t6.ptr = &(*t4)[(uintptr_t)0ul];
 t6.len = (uintptr_t)1ul;
 t7 = io_Writer_writeAll__5732(t5, t6);
 return t7;
}

static uint16_t io_Writer_writeByteNTimes__5735(struct io_Writer__3878 const a0, uint8_t const a1, uintptr_t const a2) {
 struct io_Writer__3878 const *t1;
 uint8_t *t3;
 uint8_t (*t4)[256];
 uintptr_t t6;
 uintptr_t t14;
 uintptr_t t5;
 uint64_t t7;
 struct io_Writer__3878 t10;
 struct io_Writer__3878 t0;
 nav__5735_52 t11;
 nav__5735_43 t12;
 uint16_t t9;
 uint16_t t13;
 bool t8;
 uint8_t t2[256];
 t0 = a0;
 t1 = (struct io_Writer__3878 const *)&t0;
 t3 = (uint8_t *)&t2;
 t3 = (uint8_t *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t4 = (uint8_t (*)[256])t3;
 memset(t4, a1, 256);
 t5 = a2;
 zig_loop_15:
 t6 = t5;
 t7 = t6;
 t8 = t7 > UINT64_C(0);
 if (t8) {
  t6 = t5;
  t6 = ((uintptr_t)256ul < t6) ? (uintptr_t)256ul : t6;
  t9 = (uint16_t)t6;
  t10 = (*t1);
  t3 = (uint8_t *)&t2;
  t3 = (uint8_t *)(((uintptr_t)t3) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t6 = (uintptr_t)t9;
  t11.ptr = t3;
  t11.len = t6;
  memcpy(&t12, &t11, sizeof(nav__5735_43));
  t13 = io_Writer_writeAll__5732(t10, t12);
  if (t13) {
   return t13;
  }
  t6 = t5;
  t14 = (uintptr_t)t9;
  t14 = t6 - t14;
  t5 = t14;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_15;

 zig_block_0:;
 return 0;
}

static uint16_t json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_jsonStringify__anon_9358__7130(struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a1) {
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *t7;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 t8;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 t9;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Iterator__6173 t6;
 nav__7130_65 t10;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29_Entry__6181 t12;
 nav__7130_56 *t13;
 nav__7130_56 t14;
 struct environment_root_EnvironmentValue__856 *t15;
 struct environment_root_EnvironmentValue__856 t16;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 t0;
 uint16_t t5;
 bool t11;
 t0 = a0;
 t1 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t2;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(t4);
 if (t5) {
  return t5;
 }
 t7 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const *)&t1->map;
 t8 = (*t7);
 t9 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__4011(t8);
 t6 = t9;
 zig_loop_22:
 t10 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__6163(&t6);
 t11 = t10.is_null != true;
 if (t11) {
  t12 = t10.payload;
  t4 = (*t3);
  t13 = t12.key_ptr;
  t14 = (*t13);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t4, t14);
  if (t5) {
   return t5;
  }
  t4 = (*t3);
  t15 = t12.value_ptr;
  t16 = (*t15);
  t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7131(t4, t16);
  if (t5) {
   return t5;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_22;

 zig_block_0:;
 t4 = (*t3);
 t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(t4);
 if (t5) {
  return t5;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7119(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 const a1) {
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 const *t2;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 t3;
 struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29__858 const *)&t1;
 t3 = (*t2);
 t4 = json_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_jsonStringify__anon_9358__7130(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7121(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, struct environment_linked_libraries_root_LibraryEntry__864 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 nav__7121_43 t7;
 struct json_stringify_StringifyOptions__8133 *t9;
 struct environment_linked_libraries_root_LibraryEntry__864 t2;
 uint16_t t4;
 bool t6;
 bool t5;
 bool t8;
 struct json_stringify_StringifyOptions__8133 t10;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7000(t3);
 if (t4) {
  return t4;
 }
 t5 = true;
 t6 = t5;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t3, (nav__7121_43){((uint8_t const *)&__anon_9366),(uintptr_t)4ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t7 = a1.path;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7132(t3, t7);
  if (t4) {
   return t4;
  }
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t8 = true;
 t9 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
 t10 = (*t9);
 t6 = t10.emit_null_optional_fields;
 t6 = !t6;
 if (t6) {
  t7 = a1.build_id;
  t6 = t7.ptr == NULL;
  if (t6) {
   t8 = false;
   goto zig_block_2;
  }
  goto zig_block_2;

  zig_block_2:;
  goto zig_block_1;
 }
 goto zig_block_1;

 zig_block_1:;
 t6 = t8;
 if (t6) {
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7015(t3, (nav__7121_43){((uint8_t const *)&__anon_9370),(uintptr_t)8ul});
  if (t4) {
   return t4;
  }
  t3 = (*t1);
  t7 = a1.build_id;
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7133(t3, t7);
  if (t4) {
   return t4;
  }
  goto zig_block_3;
 }
 goto zig_block_3;

 zig_block_3:;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7002(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7004(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, uint8_t const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uintptr_t *t3;
 uintptr_t t4;
 (void)a1;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t *)&t2->indent_level;
 t4 = (*t3);
 t4 = t4 - (uintptr_t)1ul;
 (*t3) = t4;
 return;
}

static void json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint8_t *t3;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = (uint8_t *)&t2->next_punctuation;
 (*t3) = UINT8_C(2);
 return;
}

static nav__7122_40 mem_Allocator_remap__anon_9292__7122(struct mem_Allocator__599 const a0, nav__7122_40 const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 nav__7122_40 const *t3;
 uint64_t t4;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t18;
 nav__7122_40 t7;
 nav__7122_40 t10;
 nav__7122_40 t2;
 struct environment_linked_libraries_root_LibraryInfo__946 *t8;
 void *t9;
 uintptr_t t11;
 uintptr_t t14;
 nav__7122_56 t12;
 nav__7122_56 t16;
 nav__7122_58 t13;
 uint8_t *t15;
 uint8_t *t23;
 uint8_t *t24;
 uint8_t *t25;
 struct mem_Allocator_VTable__602 const *const *t19;
 struct mem_Allocator_VTable__602 const *t20;
 uint8_t *(*const *t21)(void *, nav__7122_56, uint8_t, uintptr_t, uintptr_t);
 uint8_t *(*t22)(void *, nav__7122_56, uint8_t, uintptr_t, uintptr_t);
 uint8_t *const *t26;
 bool t5;
 uint8_t t17;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = a1;
 t3 = (nav__7122_40 const *)&t2;
 t4 = a2;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  t6 = (*t1);
  mem_Allocator_free__anon_6701__6275(t6, a1);
  t7 = (*t3);
  t8 = t7.ptr;
  t8 = (struct environment_linked_libraries_root_LibraryInfo__946 *)(((uintptr_t)t8) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryInfo__946)));
  t9 = (void *)t8;
  t7.ptr = (struct environment_linked_libraries_root_LibraryInfo__946 *)t9;
  t7.len = (uintptr_t)0ul;
  t10 = t7;
  return t10;
 }
 goto zig_block_0;

 zig_block_0:;
 t11 = a1.len;
 t4 = t11;
 t5 = t4 == UINT64_C(0);
 if (t5) {
  return (nav__7122_40){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};
 }
 goto zig_block_1;

 zig_block_1:;
 t12 = mem_sliceAsBytes__anon_6767__6293(a1);
 t13 = math_mul__anon_3296__5445((uintptr_t)48ul, a2);
 t5 = t13.error == UINT16_C(0);
 if (t5) {
  t14 = t13.payload;
  t11 = t14;
  goto zig_block_2;
 }
 return (nav__7122_40){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_2:;
 t6 = (*t1);
 memcpy(&t16, &t12, sizeof(nav__7122_56));
 t17 = mem_Alignment_fromByteUnits__1113((uintptr_t)8ul);
 t14 = (uintptr_t)zig_return_address();
 t18 = t6;
 t1 = (struct mem_Allocator__599 const *)&t18;
 t19 = (struct mem_Allocator_VTable__602 const *const *)&t1->vtable;
 t20 = (*t19);
 t21 = (uint8_t *(*const *)(void *, nav__7122_56, uint8_t, uintptr_t, uintptr_t))&t20->remap;
 t22 = (*t21);
 t9 = t6.ptr;
 t23 = t22(t9, t16, t17, t11, t14);
 t5 = t23 != NULL;
 if (t5) {
  t24 = t23;
  t15 = t24;
  goto zig_block_3;
 }
 return (nav__7122_40){NULL,(uintptr_t)0xaaaaaaaaaaaaaaaaul};

 zig_block_3:;
 t25 = t15;
 t26 = (uint8_t *const *)&t25;
 t15 = (*t26);
 t15 = (uint8_t *)(((uintptr_t)t15) + ((uintptr_t)0ul*sizeof(uint8_t)));
 t12.ptr = t15;
 t12.len = t11;
 memcpy(&t16, &t12, sizeof(nav__7122_56));
 t10 = mem_bytesAsSlice__anon_9401__7134(t16);
 t7 = t10;
 return t7;
}

static nav__7123_41 mem_Allocator_alignedAlloc__anon_9296__7123(struct mem_Allocator__599 const a0, uintptr_t const a1) {
 struct mem_Allocator__599 const *t1;
 struct mem_Allocator__599 t2;
 struct mem_Allocator__599 t0;
 struct mem_Allocator__599 t5;
 uintptr_t t3;
 nav__7123_41 t4;
 nav__7123_41 t8;
 nav__7123_53 t6;
 uint8_t *t9;
 struct environment_linked_libraries_root_LibraryInfo__946 *t10;
 struct environment_linked_libraries_root_LibraryInfo__946 *t11;
 struct environment_linked_libraries_root_LibraryInfo__946 *const *t12;
 nav__7123_40 t13;
 uint16_t t7;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t2 = (*t1);
 t3 = (uintptr_t)zig_return_address();
 t5 = t2;
 t1 = (struct mem_Allocator__599 const *)&t5;
 t2 = (*t1);
 t6 = mem_Allocator_allocWithSizeAndAlignment__anon_9402__7135(t2, a1, t3);
 if (t6.error) {
  t7 = t6.error;
  t8.payload = (nav__7123_40){(struct environment_linked_libraries_root_LibraryInfo__946 *)(uintptr_t)0xaaaaaaaaaaaaaaaaul, (uintptr_t)0xaaaaaaaaaaaaaaaaul};
  t8.error = t7;
  t4 = t8;
  goto zig_block_0;
 }
 t9 = t6.payload;
 t10 = (struct environment_linked_libraries_root_LibraryInfo__946 *)t9;
 t11 = t10;
 t12 = (struct environment_linked_libraries_root_LibraryInfo__946 *const *)&t11;
 t10 = (*t12);
 t10 = (struct environment_linked_libraries_root_LibraryInfo__946 *)(((uintptr_t)t10) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryInfo__946)));
 t13.ptr = t10;
 t13.len = a1;
 t8.payload = t13;
 t8.error = UINT16_C(0);
 t4 = t8;
 goto zig_block_0;

 zig_block_0:;
 return t4;
}

static uintptr_t multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4330(uintptr_t const a0, uintptr_t const a1) {
 uintptr_t t1;
 uintptr_t t2;
 uintptr_t t0;
 uint64_t t3;
 uint64_t t4;
 bool t5;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t0;
 t2 = t2 / (uintptr_t)2ul;
 t2 = t2 + (uintptr_t)4ul;
 t2 = zig_adds_u64(t1, t2, UINT8_C(64));
 t0 = t2;
 t2 = t0;
 t3 = t2;
 t4 = a1;
 t5 = t3 >= t4;
 if (t5) {
  t2 = t0;
  return t2;
 }
 goto zig_block_0;

 zig_block_0:;
 goto zig_loop_6;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7124(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t9;
 nav__7124_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7124_60 t12;
 nav__7124_60 t33;
 nav__7124_60 t13;
 nav__7124_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u8_29__6735 const *t31;
 struct array_hash_map_Index_28u8_29__6735 *t34;
 uint8_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__892 t28;
 bool t20;
 uint8_t t26;
 uint8_t t36;
 uint8_t t25;
 uint8_t t27;
 struct array_hash_map_Index_28u8_29__6735 t29;
 struct array_hash_map_Index_28u8_29__6735 t32;
 struct array_hash_map_Index_28u8_29__6735 t30;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7759__6389(t11);
 t13 = t12;
 t14 = (nav__7124_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint8_t)t17;
  t25 = t26;
  t27 = UINT8_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u8_29__6735 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u8_29_isEmpty__6280(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint8_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint8_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint8_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT8_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7125(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t9;
 nav__7125_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7125_60 t12;
 nav__7125_60 t33;
 nav__7125_60 t13;
 nav__7125_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u16_29__6739 const *t31;
 struct array_hash_map_Index_28u16_29__6739 *t34;
 uint16_t *t35;
 uint32_t t21;
 struct array_hash_map_IndexHeader__892 t28;
 uint16_t t26;
 uint16_t t36;
 uint16_t t25;
 uint16_t t27;
 struct array_hash_map_Index_28u16_29__6739 t29;
 struct array_hash_map_Index_28u16_29__6739 t32;
 struct array_hash_map_Index_28u16_29__6739 t30;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7769__6391(t11);
 t13 = t12;
 t14 = (nav__7125_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t26 = (uint16_t)t17;
  t25 = t26;
  t27 = UINT16_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t28 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t28, t22);
   t29 = t12.ptr[t22];
   t30 = t29;
   t31 = (struct array_hash_map_Index_28u16_29__6739 const *)&t30;
   t32 = (*t31);
   t20 = array_hash_map_Index_28u16_29_isEmpty__6285(t32);
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t26 = t27;
    (*t35) = t26;
    t35 = (uint16_t *)&t34->entry_index;
    t26 = t25;
    (*t35) = t26;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t26 = t29.distance_from_start_index;
   t36 = t27;
   t20 = t26 < t36;
   if (t20) {
    t33 = (*t14);
    t34 = &t33.ptr[t22];
    t35 = (uint16_t *)&t34->distance_from_start_index;
    t36 = t27;
    (*t35) = t36;
    t35 = (uint16_t *)&t34->entry_index;
    t36 = t25;
    (*t35) = t36;
    t36 = t29.distance_from_start_index;
    t27 = t36;
    t36 = t29.entry_index;
    t25 = t36;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t36 = t27;
   t36 = t36 + UINT16_C(1);
   t27 = t36;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static void array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7126(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const a0, struct array_hash_map_IndexHeader__892 *const a1) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t4;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *t0;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t8;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t9;
 nav__7126_56 t10;
 struct array_hash_map_IndexHeader__892 *t11;
 struct array_hash_map_IndexHeader__892 *t2;
 nav__7126_60 t12;
 nav__7126_60 t32;
 nav__7126_60 t13;
 nav__7126_60 const *t14;
 uintptr_t t16;
 uintptr_t t17;
 uintptr_t t22;
 uintptr_t t23;
 uintptr_t t15;
 uintptr_t t24;
 uint64_t t18;
 uint64_t t19;
 struct array_hash_map_Index_28u32_29__6743 const *t30;
 struct array_hash_map_Index_28u32_29__6743 *t33;
 uint32_t *t34;
 uint32_t t21;
 uint32_t t35;
 uint32_t t25;
 uint32_t t26;
 struct array_hash_map_IndexHeader__892 t27;
 struct array_hash_map_Index_28u32_29__6743 t28;
 struct array_hash_map_Index_28u32_29__6743 t31;
 struct array_hash_map_Index_28u32_29__6743 t29;
 bool t20;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 *const *)&t0;
 t2 = a1;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (*t1);
 t5 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 *)&t4->entries;
 t6 = (*t5);
 t7 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t6);
 t8 = t7;
 t9 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t8;
 t7 = (*t9);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t7);
 t11 = (*t3);
 t12 = array_hash_map_IndexHeader_indexes__anon_7775__6393(t11);
 t13 = t12;
 t14 = (nav__7126_60 const *)&t13;
 t15 = (uintptr_t)0ul;
 t16 = t10.len;
 zig_loop_30:
 t17 = t15;
 t18 = t17;
 t19 = t16;
 t20 = t18 < t19;
 if (t20) {
  t21 = t10.ptr[t17];
  t22 = array_hash_map_safeTruncate__anon_7783__6395(t21);
  t23 = t12.len;
  t23 = zig_addw_u64(t22, t23, UINT8_C(64));
  t24 = t22;
  t21 = (uint32_t)t17;
  t25 = t21;
  t26 = UINT32_C(0);
  zig_loop_53:
  t22 = t24;
  t19 = t22;
  t18 = t23;
  t20 = t19 != t18;
  if (t20) {
   t11 = (*t3);
   t27 = (*t11);
   t22 = t24;
   t22 = array_hash_map_IndexHeader_constrainIndex__3971(t27, t22);
   t28 = t12.ptr[t22];
   t29 = t28;
   t30 = (struct array_hash_map_Index_28u32_29__6743 const *)&t29;
   t31 = (*t30);
   t20 = array_hash_map_Index_28u32_29_isEmpty__6290(t31);
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t21 = t26;
    (*t34) = t21;
    t34 = (uint32_t *)&t33->entry_index;
    t21 = t25;
    (*t34) = t21;
    goto zig_block_1;
   }
   goto zig_block_4;

   zig_block_4:;
   t21 = t28.distance_from_start_index;
   t35 = t26;
   t20 = t21 < t35;
   if (t20) {
    t32 = (*t14);
    t33 = &t32.ptr[t22];
    t34 = (uint32_t *)&t33->distance_from_start_index;
    t35 = t26;
    (*t34) = t35;
    t34 = (uint32_t *)&t33->entry_index;
    t35 = t25;
    (*t34) = t35;
    t35 = t28.distance_from_start_index;
    t26 = t35;
    t35 = t28.entry_index;
    t25 = t35;
    goto zig_block_5;
   }
   goto zig_block_5;

   zig_block_5:;
   t22 = t24;
   t22 = zig_addw_u64(t22, (uintptr_t)1ul, UINT8_C(64));
   t24 = t22;
   t35 = t26;
   t35 = t35 + UINT32_C(1);
   t26 = t35;
   goto zig_block_3;
  }
  goto zig_block_2;

  zig_block_3:;
  goto zig_loop_53;

  zig_block_2:;
  zig_unreachable();
 }
 goto zig_block_0;

 zig_block_1:;
 t17 = t17 + (uintptr_t)1ul;
 t15 = t17;
 goto zig_loop_30;

 zig_block_0:;
 return;
}

static nav__7127_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7127(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7127_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__7127_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t8;
 nav__7127_66 t9;
 nav__7127_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u8_29__6735 const *t23;
 nav__7127_41 t30;
 nav__7127_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__892 t20;
 bool t18;
 bool t25;
 struct array_hash_map_Index_28u8_29__6735 t21;
 struct array_hash_map_Index_28u8_29__6735 t24;
 struct array_hash_map_Index_28u8_29__6735 t22;
 uint8_t t26;
 uint8_t t27;
 uint8_t t15;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT8_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u8_29__6735 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u8_29_isEmpty__6280(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__7127_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT8_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__7128_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7128(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7128_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__7128_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t8;
 nav__7128_66 t9;
 nav__7128_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u16_29__6739 const *t23;
 nav__7128_41 t30;
 nav__7128_38 t31;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 uint32_t t11;
 uint32_t t29;
 struct array_hash_map_IndexHeader__892 t20;
 struct array_hash_map_Index_28u16_29__6739 t21;
 struct array_hash_map_Index_28u16_29__6739 t24;
 struct array_hash_map_Index_28u16_29__6739 t22;
 uint16_t t26;
 uint16_t t27;
 uint16_t t15;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT16_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u16_29__6739 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u16_29_isEmpty__6285(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__7128_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t29 = t9.ptr[t28];
  t25 = t11 == t29;
  if (t25) {
   t28 = (uintptr_t)t27;
   t30 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t30, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t31.is_null = false;
   t31.payload = t12;
   return t31;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT16_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static nav__7129_38 array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7129(struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const a0, nav__7129_41 const a1, struct array_hash_map_IndexHeader__892 *const a2, nav__7129_47 const a3) {
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *t1;
 struct array_hash_map_IndexHeader__892 *const *t3;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *t4;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 t5;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t6;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 t7;
 struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *t8;
 nav__7129_66 t9;
 nav__7129_69 t10;
 uintptr_t t12;
 uintptr_t t13;
 uintptr_t t28;
 uintptr_t t14;
 uint64_t t16;
 uint64_t t17;
 struct array_hash_map_IndexHeader__892 *t19;
 struct array_hash_map_IndexHeader__892 *t2;
 struct array_hash_map_Index_28u32_29__6743 const *t23;
 nav__7129_41 t29;
 nav__7129_38 t30;
 struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 t0;
 uint32_t t11;
 uint32_t t26;
 uint32_t t27;
 uint32_t t15;
 struct array_hash_map_IndexHeader__892 t20;
 struct array_hash_map_Index_28u32_29__6743 t21;
 struct array_hash_map_Index_28u32_29__6743 t24;
 struct array_hash_map_Index_28u32_29__6743 t22;
 bool t18;
 bool t25;
 t0 = a0;
 t1 = (struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const *)&t0;
 t2 = a2;
 t3 = (struct array_hash_map_IndexHeader__892 *const *)&t2;
 t4 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29__939 const *)&t1->entries;
 t5 = (*t4);
 t6 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__4310(t5);
 t7 = t6;
 t8 = (struct multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29_Data_29_Slice__6459 const *)&t7;
 t6 = (*t8);
 t9 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__7101(t6);
 t6 = (*t8);
 t10 = multi_array_list_MultiArrayList_28array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_l__6199(t6);
 t11 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7100(a1);
 t12 = array_hash_map_safeTruncate__anon_7783__6395(t11);
 t13 = a3.len;
 t13 = zig_addw_u64(t12, t13, UINT8_C(64));
 t14 = t12;
 t15 = UINT32_C(0);
 zig_loop_37:
 t12 = t14;
 t16 = t12;
 t17 = t13;
 t18 = t16 != t17;
 if (t18) {
  t19 = (*t3);
  t20 = (*t19);
  t12 = t14;
  t12 = array_hash_map_IndexHeader_constrainIndex__3971(t20, t12);
  t21 = a3.ptr[t12];
  t22 = t21;
  t23 = (struct array_hash_map_Index_28u32_29__6743 const *)&t22;
  t24 = (*t23);
  t18 = array_hash_map_Index_28u32_29_isEmpty__6290(t24);
  if (t18) {
   t25 = true;
   goto zig_block_3;
  }
  t26 = t21.distance_from_start_index;
  t27 = t15;
  t18 = t26 < t27;
  t25 = t18;
  goto zig_block_3;

  zig_block_3:;
  if (t25) {
   return (nav__7129_38){(uintptr_t)0xaaaaaaaaaaaaaaaaul,true};
  }
  goto zig_block_2;

  zig_block_2:;
  t27 = t21.entry_index;
  t28 = (uintptr_t)t27;
  t26 = t9.ptr[t28];
  t25 = t11 == t26;
  if (t25) {
   t28 = (uintptr_t)t27;
   t29 = t10.ptr[t28];
   t28 = (uintptr_t)t27;
   t25 = array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__7102(a1, t29, t28);
   t18 = t25;
   goto zig_block_5;
  }
  t18 = false;
  goto zig_block_5;

  zig_block_5:;
  if (t18) {
   t30.is_null = false;
   t30.payload = t12;
   return t30;
  }
  goto zig_block_4;

  zig_block_4:;
  t12 = t14;
  t12 = zig_addw_u64(t12, (uintptr_t)1ul, UINT8_C(64));
  t14 = t12;
  t27 = t15;
  t27 = t27 + UINT32_C(1);
  t15 = t27;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_37;

 zig_block_0:;
 zig_unreachable();
}

static uint16_t environment_root_EnvironmentValue_jsonStringify__anon_9438__7136(struct environment_root_EnvironmentValue__856 const a0, struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 nav__7136_43 t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t4;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 nav__7136_46 t6;
 uint16_t t5;
 uint8_t t2;
 t0 = a1;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a0.tag;
 switch (t2) {
  case UINT8_C(0): {
   t3 = a0.payload.string;
   t4 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7132(t4, t3);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  case UINT8_C(1): {
   t6 = a0.payload.list;
   t4 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7137(t4, t6);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7131(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, struct environment_root_EnvironmentValue__856 const a1) {
 struct environment_root_EnvironmentValue__856 const *t2;
 struct environment_root_EnvironmentValue__856 t3;
 struct environment_root_EnvironmentValue__856 t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint16_t t4;
 uint16_t t5;
 t0 = a0;
 t1 = a1;
 t2 = (struct environment_root_EnvironmentValue__856 const *)&t1;
 t3 = (*t2);
 t4 = environment_root_EnvironmentValue_jsonStringify__anon_9438__7136(t3, a0);
 memcpy(&t5, &t4, sizeof(uint16_t));
 return t5;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7020(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, nav__7020_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_stringify_StringifyOptions__8133 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t7;
 uint16_t t3;
 struct json_stringify_StringifyOptions__8133 t5;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(t2);
 if (t3) {
  return t3;
 }
 t4 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
 t5 = (*t4);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&a0->stream;
 t7 = (*t6);
 t3 = json_stringify_encodeJsonString__anon_9100__7109(a1, t5, t7);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(t2);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7132(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, nav__7132_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_StringifyOptions__8133 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t7;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uintptr_t t11;
 uintptr_t t12;
 uintptr_t t10;
 uint64_t t13;
 uint64_t t14;
 nav__7132_41 t2;
 uint16_t t8;
 uint16_t t9;
 struct json_stringify_StringifyOptions__8133 t4;
 bool t5;
 bool t6;
 uint8_t t15;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a1;
 t3 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
 t4 = (*t3);
 t5 = t4.emit_strings_as_arrays;
 t5 = !t5;
 if (t5) {
  t5 = unicode_utf8ValidateSlice__5770(a1);
  t6 = t5;
  goto zig_block_1;
 }
 t6 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t6) {
  t7 = (*t1);
  t8 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7020(t7, a1);
  memcpy(&t9, &t8, sizeof(uint16_t));
  return t9;
 }
 goto zig_block_0;

 zig_block_0:;
 t7 = (*t1);
 t9 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6999(t7);
 if (t9) {
  return t9;
 }
 t10 = (uintptr_t)0ul;
 t11 = a1.len;
 zig_loop_44:
 t12 = t10;
 t13 = t12;
 t14 = t11;
 t6 = t13 < t14;
 if (t6) {
  t15 = a1.ptr[t12];
  t7 = (*t1);
  t9 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7138(t7, t15);
  if (t9) {
   return t9;
  }
  goto zig_block_3;
 }
 goto zig_block_2;

 zig_block_3:;
 t12 = t12 + (uintptr_t)1ul;
 t10 = t12;
 goto zig_loop_44;

 zig_block_2:;
 t7 = (*t1);
 t9 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7001(t7);
 if (t9) {
  return t9;
 }
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7133(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, nav__7133_41 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 nav__7133_41 t4;
 nav__7133_41 t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t5;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint16_t t6;
 bool t3;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a1;
 t3 = a1.ptr != NULL;
 if (t3) {
  t4 = a1;
  t5 = (*t1);
  t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7132(t5, t4);
  if (t6) {
   return t6;
  }
  return 0;
 }
 t5 = (*t1);
 t6 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7139(t5);
 if (t6) {
  return t6;
 }
 return 0;
}

static nav__7134_40 mem_bytesAsSlice__anon_9401__7134(nav__7134_42 const a0) {
 uintptr_t t0;
 uint64_t t1;
 uint8_t *t4;
 struct environment_linked_libraries_root_LibraryInfo__946 *t5;
 struct environment_linked_libraries_root_LibraryInfo__946 *t6;
 struct environment_linked_libraries_root_LibraryInfo__946 *const *t7;
 nav__7134_40 t8;
 bool t2;
 bool t3;
 t0 = a0.len;
 t1 = t0;
 t2 = t1 == UINT64_C(0);
 if (t2) {
  t3 = true;
  goto zig_block_1;
 }
 t3 = false;
 goto zig_block_1;

 zig_block_1:;
 if (t3) {
  return (nav__7134_40){(struct environment_linked_libraries_root_LibraryInfo__946 *)((void const *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),(uintptr_t)0ul};
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a0.ptr;
 t5 = (struct environment_linked_libraries_root_LibraryInfo__946 *)t4;
 t6 = t5;
 t7 = (struct environment_linked_libraries_root_LibraryInfo__946 *const *)&t6;
 t0 = a0.len;
 t0 = t0 / (uintptr_t)48ul;
 t5 = (*t7);
 t5 = (struct environment_linked_libraries_root_LibraryInfo__946 *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(struct environment_linked_libraries_root_LibraryInfo__946)));
 t8.ptr = t5;
 t8.len = t0;
 return t8;
}

static nav__7135_39 mem_Allocator_allocWithSizeAndAlignment__anon_9402__7135(struct mem_Allocator__599 const a0, uintptr_t const a1, uintptr_t const a2) {
 struct mem_Allocator__599 const *t1;
 uintptr_t t2;
 uintptr_t t5;
 nav__7135_49 t3;
 struct mem_Allocator__599 t6;
 struct mem_Allocator__599 t0;
 nav__7135_39 t7;
 bool t4;
 t0 = a0;
 t1 = (struct mem_Allocator__599 const *)&t0;
 t3 = math_mul__anon_3296__5445((uintptr_t)48ul, a1);
 t4 = t3.error == UINT16_C(0);
 if (t4) {
  t5 = t3.payload;
  t2 = t5;
  goto zig_block_0;
 }
 return (nav__7135_39){((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),zig_error_OutOfMemory};

 zig_block_0:;
 t6 = (*t1);
 t7 = mem_Allocator_allocBytesWithAlignment__anon_2823__5103(t6, t2, a2);
 return t7;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7137(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, nav__7137_43 const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uintptr_t t6;
 uintptr_t t7;
 uintptr_t t5;
 uint64_t t8;
 uint64_t t9;
 nav__7137_41 t11;
 nav__7137_43 t2;
 uint16_t t4;
 bool t10;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6999(t3);
 if (t4) {
  return t4;
 }
 t5 = (uintptr_t)0ul;
 t6 = a1.len;
 zig_loop_28:
 t7 = t5;
 t8 = t7;
 t9 = t6;
 t10 = t8 < t9;
 if (t10) {
  t11 = a1.ptr[t7];
  t3 = (*t1);
  t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7132(t3, t11);
  if (t4) {
   return t4;
  }
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 t7 = t7 + (uintptr_t)1ul;
 t5 = t7;
 goto zig_loop_28;

 zig_block_0:;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7001(t3);
 if (t4) {
  return t4;
 }
 return 0;
}

static bool unicode_utf8ValidateSlice__5770(nav__5770_39 const a0) {
 bool t0;
 t0 = unicode_utf8ValidateSliceImpl__anon_9465__7140(a0);
 return t0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6999(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t9;
 void const **t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t12;
 void const *t13;
 nav__6999_54 (**t14)(void const *, nav__6999_56);
 struct io_Writer__3878 t15;
 struct io_Writer__3878 t10;
 struct io_Writer__3878 t16;
 struct io_Writer__3878 const *t17;
 uint8_t *t19;
 uint16_t t3;
 uint16_t t18;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(t2);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
 t5 = (*t4);
 t6 = t5;
 t7 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t6;
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t8;
 t11 = (void const **)&t10.context;
 t7 = (*t9);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t7->context;
 t13 = (void const *)t12;
 (*t11) = t13;
 t14 = (nav__6999_54 (**)(void const *, nav__6999_56))&t10.writeFn;
 (*t14) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t15 = t10;
 t16 = t15;
 t17 = (struct io_Writer__3878 const *)&t16;
 t15 = (*t17);
 t3 = io_Writer_writeByte__5734(t15, UINT8_C(91));
 memcpy(&t18, &t3, sizeof(uint16_t));
 if (t18) {
  return t18;
 }
 t2 = (*t1);
 (void)json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7003(t2, UINT8_C(1));
 t2 = (*t1);
 t19 = (uint8_t *)&t2->next_punctuation;
 (*t19) = UINT8_C(1);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7138(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0, uint8_t const a1) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t3;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct json_stringify_StringifyOptions__8133 *t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t11;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t12;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t13;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t14;
 void const **t16;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t17;
 void const *t18;
 nav__7138_57 (**t19)(void const *, nav__7138_59);
 struct io_Writer__3878 t20;
 struct io_Writer__3878 t15;
 struct io_Writer__3878 t21;
 struct io_Writer__3878 const *t22;
 uint16_t t4;
 uint16_t t23;
 struct json_stringify_StringifyOptions__8133 t6;
 bool t7;
 nav__7138_51 t10;
 uint8_t t2;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = a1;
 t3 = (*t1);
 t4 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(t3);
 if (t4) {
  return t4;
 }
 t5 = (struct json_stringify_StringifyOptions__8133 *)&a0->options;
 t6 = (*t5);
 t7 = t6.emit_nonportable_numbers_as_strings;
 if (t7) {
  goto zig_block_0;
 }
 goto zig_block_0;

 zig_block_0:;
 t3 = (*t1);
 t8 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t3->stream;
 t9 = (*t8);
 t10.f0 = a1;
 t11 = t9;
 t12 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t11;
 t13 = t12;
 t14 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t13;
 t16 = (void const **)&t15.context;
 t12 = (*t14);
 t17 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t12->context;
 t18 = (void const *)t17;
 (*t16) = t18;
 t19 = (nav__7138_57 (**)(void const *, nav__7138_59))&t15.writeFn;
 (*t19) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t20 = t15;
 t21 = t20;
 t22 = (struct io_Writer__3878 const *)&t21;
 t20 = (*t22);
 t4 = io_Writer_print__anon_9479__7141(t20, t10);
 memcpy(&t23, &t4, sizeof(uint16_t));
 if (t23) {
  return t23;
 }
 t3 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(t3);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7001(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 uint8_t *t3;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t9;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t10;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t11;
 void const **t13;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t14;
 void const *t15;
 nav__7001_55 (**t16)(void const *, nav__7001_57);
 struct io_Writer__3878 t17;
 struct io_Writer__3878 t12;
 struct io_Writer__3878 t18;
 struct io_Writer__3878 const *t19;
 uint16_t t5;
 uint16_t t20;
 uint8_t t4;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7004(t2, UINT8_C(1));
 t3 = (uint8_t *)&a0->next_punctuation;
 t4 = (*t3);
 switch (t4) {
  case UINT8_C(1): {
   goto zig_block_0;
  }
  case UINT8_C(2): {
   t2 = (*t1);
   t5 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7005(t2);
   if (t5) {
    return t5;
   }
   goto zig_block_0;
  }
  case UINT8_C(0):
  case UINT8_C(3): {
   zig_unreachable();
  }
  default: zig_unreachable();
 }

 zig_block_0:;
 t2 = (*t1);
 t6 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
 t7 = (*t6);
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t8;
 t10 = t9;
 t11 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t10;
 t13 = (void const **)&t12.context;
 t9 = (*t11);
 t14 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t9->context;
 t15 = (void const *)t14;
 (*t13) = t15;
 t16 = (nav__7001_55 (**)(void const *, nav__7001_57))&t12.writeFn;
 (*t16) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t17 = t12;
 t18 = t17;
 t19 = (struct io_Writer__3878 const *)&t18;
 t17 = (*t19);
 t5 = io_Writer_writeByte__5734(t17, UINT8_C(93));
 memcpy(&t20, &t5, sizeof(uint16_t));
 if (t20) {
  return t20;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(t2);
 return 0;
}

static uint16_t json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7139(struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const a0) {
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *t1;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t2;
 struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *t0;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *t4;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t5;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 t6;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t7;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *t8;
 struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *t9;
 void const **t11;
 struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *t12;
 void const *t13;
 nav__7139_54 (**t14)(void const *, nav__7139_56);
 struct io_Writer__3878 t15;
 struct io_Writer__3878 t10;
 struct io_Writer__3878 t16;
 struct io_Writer__3878 const *t17;
 uint16_t t3;
 uint16_t t18;
 t0 = a0;
 t1 = (struct json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29_2c__7b_20_checked_to_arbitrary_depth_20_3d_20_7b_7d_20_7d_29__8291 *const *)&t0;
 t2 = (*t1);
 t3 = json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7006(t2);
 if (t3) {
  return t3;
 }
 t2 = (*t1);
 t4 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 *)&t2->stream;
 t5 = (*t4);
 t6 = t5;
 t7 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *)&t6;
 t8 = t7;
 t9 = (struct io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29__1839 const *const *)&t8;
 t11 = (void const **)&t10.context;
 t7 = (*t9);
 t12 = (struct array_list_ArrayListAligned_28u8_2cnull_29__811 *const *)&t7->context;
 t13 = (void const *)t12;
 (*t11) = t13;
 t14 = (nav__7139_54 (**)(void const *, nav__7139_56))&t10.writeFn;
 (*t14) = &io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d_2c_28function_20_27appendWrite_27_29_29___4912;
 t15 = t10;
 t16 = t15;
 t17 = (struct io_Writer__3878 const *)&t16;
 t15 = (*t17);
 t3 = io_Writer_writeAll__5732(t15, (nav__7139_56){(uint8_t const *)&__anon_9483,(uintptr_t)4ul});
 memcpy(&t18, &t3, sizeof(uint16_t));
 if (t18) {
  return t18;
 }
 t2 = (*t1);
 json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__7009(t2);
 return 0;
}

static bool unicode_utf8ValidateSliceImpl__anon_9465__7140(nav__7140_39 const a0) {
 static uint8_t const t9[16] = "\200\200\200\200\200\200\200\200\200\200\200\200\200\200\200\200";
 static uint8_t const t15[256] = "\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\360\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\361\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\002\023\003\003\003\003\003\003\003\003\003\003\003\003#\003\0034\004\004\004D\361\361\361\361\361\361\361\361\361\361\361";
 nav__7140_39 t1;
 nav__7140_39 t0;
 uintptr_t t2;
 uintptr_t t12;
 uintptr_t t17;
 uintptr_t t11;
 uint64_t t3;
 uint64_t t13;
 uint8_t const *t5;
 uint8_t const (*t6)[16];
 bool t4;
 bool t21;
 uint8_t t7[16];
 uint8_t t8[16];
 bool t10[16];
 uint8_t t14;
 uint8_t t16;
 uint8_t t20;
 uint8_t t18;
 uint8_t t19;
 t0 = a0;
 zig_loop_6:
 t1 = t0;
 t2 = t1.len;
 t3 = t2;
 t4 = t3 >= UINT64_C(16);
 if (t4) {
  t1 = t0;
  t5 = t1.ptr;
  t5 = (uint8_t const *)(((uintptr_t)t5) + ((uintptr_t)0ul*sizeof(uint8_t)));
  t6 = (uint8_t const (*)[16])t5;
  memcpy(t7, (const char *)t6, sizeof(uint8_t[16]));
  memcpy(&t8, &t7, sizeof(uint8_t[16]));
  for (t2 = (uintptr_t)0ul; t2 < (uintptr_t)16ul; t2 += (uintptr_t)1ul) {
   t8[t2] = t8[t2] & t9[t2];
  }
  for (t2 = (uintptr_t)0ul; t2 < (uintptr_t)16ul; t2 += (uintptr_t)1ul) {
   t10[t2] = t8[t2] == t9[t2];
  }
  t4 = false;
  for (t2 = (uintptr_t)0ul; t2 < (uintptr_t)16ul; t2 += (uintptr_t)1ul) {
   t4 |= t10[t2];
  }
  if (t4) {
   goto zig_block_0;
  }
  goto zig_block_2;

  zig_block_2:;
  t1 = t0;
  t5 = t1.ptr;
  t5 = (uint8_t const *)(((uintptr_t)t5) + ((uintptr_t)16ul*sizeof(uint8_t)));
  t2 = t1.len;
  t2 = t2 - (uintptr_t)16ul;
  t1.ptr = t5;
  t1.len = t2;
  t0 = t1;
  goto zig_block_1;
 }
 goto zig_block_0;

 zig_block_1:;
 goto zig_loop_6;

 zig_block_0:;
 t1 = t0;
 t2 = t1.len;
 t11 = (uintptr_t)0ul;
 zig_loop_52:
 t12 = t11;
 t3 = t12;
 t13 = t2;
 t4 = t3 < t13;
 if (t4) {
  t1 = t0;
  t12 = t11;
  t14 = t1.ptr[t12];
  t4 = t14 < UINT8_C(128);
  if (t4) {
   t12 = t11;
   t12 = t12 + (uintptr_t)1ul;
   t11 = t12;
   goto zig_block_5;
  }
  goto zig_block_6;

  zig_block_6:;
  t12 = (uintptr_t)t14;
  t14 = t15[t12];
  t4 = t14 == UINT8_C(241);
  if (t4) {
   return false;
  }
  goto zig_block_7;

  zig_block_7:;
  t16 = t14 & UINT8_C(7);
  t12 = t11;
  t17 = (uintptr_t)t16;
  t17 = t12 + t17;
  t13 = t17;
  t3 = t2;
  t4 = t13 > t3;
  if (t4) {
   return false;
  }
  goto zig_block_8;

  zig_block_8:;
  t18 = UINT8_C(128);
  t19 = UINT8_C(191);
  t14 = zig_shr_u8(t14, UINT8_C(4));
  switch (t14) {
   case UINT8_C(0): {
    goto zig_block_9;
   }
   case UINT8_C(1): {
    t18 = UINT8_C(160);
    goto zig_block_9;
   }
   case UINT8_C(2): {
    t19 = UINT8_C(159);
    goto zig_block_9;
   }
   case UINT8_C(3): {
    t18 = UINT8_C(144);
    goto zig_block_9;
   }
   case UINT8_C(4): {
    t19 = UINT8_C(143);
    goto zig_block_9;
   }
   default: {
    zig_unreachable();
   }
  }

  zig_block_9:;
  t1 = t0;
  t17 = t11;
  t17 = t17 + (uintptr_t)1ul;
  t14 = t1.ptr[t17];
  t20 = t18;
  t4 = t14 < t20;
  if (t4) {
   t21 = true;
   goto zig_block_11;
  }
  t20 = t19;
  t4 = t20 < t14;
  t21 = t4;
  goto zig_block_11;

  zig_block_11:;
  if (t21) {
   return false;
  }
  goto zig_block_10;

  zig_block_10:;
  switch (t16) {
   case UINT8_C(2): {
    t17 = t11;
    t17 = t17 + (uintptr_t)2ul;
    t11 = t17;
    goto zig_block_12;
   }
   case UINT8_C(3): {
    t1 = t0;
    t17 = t11;
    t17 = t17 + (uintptr_t)2ul;
    t16 = t1.ptr[t17];
    t21 = t16 < UINT8_C(128);
    if (t21) {
     t4 = true;
     goto zig_block_14;
    }
    t21 = UINT8_C(191) < t16;
    t4 = t21;
    goto zig_block_14;

    zig_block_14:;
    if (t4) {
     return false;
    }
    goto zig_block_13;

    zig_block_13:;
    t17 = t11;
    t17 = t17 + (uintptr_t)3ul;
    t11 = t17;
    goto zig_block_12;
   }
   case UINT8_C(4): {
    t1 = t0;
    t17 = t11;
    t17 = t17 + (uintptr_t)2ul;
    t16 = t1.ptr[t17];
    t21 = t16 < UINT8_C(128);
    if (t21) {
     t4 = true;
     goto zig_block_16;
    }
    t21 = UINT8_C(191) < t16;
    t4 = t21;
    goto zig_block_16;

    zig_block_16:;
    if (t4) {
     return false;
    }
    goto zig_block_15;

    zig_block_15:;
    t1 = t0;
    t17 = t11;
    t17 = t17 + (uintptr_t)3ul;
    t16 = t1.ptr[t17];
    t21 = t16 < UINT8_C(128);
    if (t21) {
     t4 = true;
     goto zig_block_18;
    }
    t21 = UINT8_C(191) < t16;
    t4 = t21;
    goto zig_block_18;

    zig_block_18:;
    if (t4) {
     return false;
    }
    goto zig_block_17;

    zig_block_17:;
    t17 = t11;
    t17 = t17 + (uintptr_t)4ul;
    t11 = t17;
    goto zig_block_12;
   }
   default: {
    zig_unreachable();
   }
  }

  zig_block_12:;
  goto zig_block_5;

  zig_block_5:;
  goto zig_block_4;
 }
 goto zig_block_3;

 zig_block_4:;
 goto zig_loop_52;

 zig_block_3:;
 return true;
}

static uint16_t io_Writer_print__anon_9479__7141(struct io_Writer__3878 const a0, nav__7141_39 const a1) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_format__anon_9488__7142(a0, a1);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_format__anon_9488__7142(struct io_Writer__3878 const a0, nav__7142_39 const a1) {
 struct io_Writer__3878 t0;
 uint16_t t2;
 uint8_t t1;
 t0 = a0;
 t1 = a1.f0;
 t2 = fmt_formatType__anon_9506__7143(t1, (struct fmt_FormatOptions__4300){{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},{(uintptr_t)0xaaaaaaaaaaaaaaaaul,true},UINT32_C(32),UINT8_C(2)}, a0, (uintptr_t)3ul);
 if (t2) {
  return t2;
 }
 return 0;
}

static uint16_t fmt_formatType__anon_9506__7143(uint8_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2, uintptr_t const a3) {
 struct io_Writer__3878 t1;
 uint16_t t2;
 uint16_t t3;
 uint8_t t0;
 (void)a3;
 t0 = a0;
 t1 = a2;
 t2 = fmt_formatValue__anon_9511__7144(a0, a1, a2);
 memcpy(&t3, &t2, sizeof(uint16_t));
 return t3;
}

static uint16_t fmt_formatInt__anon_9519__7146(uint8_t const a0, uint8_t const a1, uint8_t const a2, struct fmt_FormatOptions__4300 const a3, struct io_Writer__3878 const a4) {
 uintptr_t t5;
 uintptr_t t3;
 uint8_t *t6;
 uint8_t (*t7)[2];
 nav__7146_58 t10;
 nav__7146_47 t11;
 uint16_t t12;
 uint16_t t13;
 bool t0;
 uint8_t t4;
 uint8_t t2;
 nav__7146_56 t8;
 uint8_t t9[2];
 uint8_t t1[9];
 t0 = a1 >= UINT8_C(2);
 debug_assert__180(t0);
 t2 = a0;
 t3 = (uintptr_t)9ul;
 t0 = a1 == UINT8_C(10);
 if (t0) {
  zig_loop_20:
  t4 = t2;
  t0 = t4 >= UINT8_C(100);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)2ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1;
   t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
   t7 = (uint8_t (*)[2])t6;
   t4 = t2;
   t4 = t4 % UINT8_C(100);
   t8 = fmt_digits2__5509(t4);
   memcpy(t9, t8.array, sizeof(uint8_t[2]));
   memcpy((char *)t7, t9, sizeof(uint8_t[2]));
   t4 = t2;
   t4 = t4 / UINT8_C(100);
   t2 = t4;
   goto zig_block_2;
  }
  goto zig_block_1;

  zig_block_2:;
  goto zig_loop_20;

  zig_block_1:;
  t4 = t2;
  t0 = t4 < UINT8_C(10);
  if (t0) {
   t5 = t3;
   t5 = t5 - (uintptr_t)1ul;
   t3 = t5;
   t5 = t3;
   t6 = (uint8_t *)&t1[t5];
   t4 = t2;
   t4 = UINT8_C(48) + t4;
   (*t6) = t4;
   goto zig_block_3;
  }
  t5 = t3;
  t5 = t5 - (uintptr_t)2ul;
  t3 = t5;
  t5 = t3;
  t6 = (uint8_t *)&t1;
  t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
  t7 = (uint8_t (*)[2])t6;
  t4 = t2;
  t8 = fmt_digits2__5509(t4);
  memcpy(t9, t8.array, sizeof(uint8_t[2]));
  memcpy((char *)t7, t9, sizeof(uint8_t[2]));
  goto zig_block_3;

  zig_block_3:;
  goto zig_block_0;
 }
 zig_loop_75:
 t4 = t2;
 t4 = t4 % a1;
 t5 = t3;
 t5 = t5 - (uintptr_t)1ul;
 t3 = t5;
 t5 = t3;
 t6 = (uint8_t *)&t1[t5];
 t4 = fmt_digitToChar__5525(t4, a2);
 (*t6) = t4;
 t4 = t2;
 t4 = t4 / a1;
 t2 = t4;
 t4 = t2;
 t0 = t4 == UINT8_C(0);
 if (t0) {
  goto zig_block_4;
 }
 goto zig_block_5;

 zig_block_5:;
 goto zig_loop_75;

 zig_block_4:;
 goto zig_block_0;

 zig_block_0:;
 t5 = t3;
 t6 = (uint8_t *)&t1;
 t6 = (uint8_t *)(((uintptr_t)t6) + (t5*sizeof(uint8_t)));
 t5 = (uintptr_t)9ul - t5;
 t10.ptr = t6;
 t10.len = t5;
 memcpy(&t11, &t10, sizeof(nav__7146_47));
 t12 = fmt_formatBuf__anon_5176__6129(t11, a3, a4);
 memcpy(&t13, &t12, sizeof(uint16_t));
 return t13;
}

static uint16_t fmt_formatIntValue__anon_9515__7145(uint8_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatInt__anon_9519__7146(a0, UINT8_C(10), UINT8_C(0), a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint16_t fmt_formatValue__anon_9511__7144(uint8_t const a0, struct fmt_FormatOptions__4300 const a1, struct io_Writer__3878 const a2) {
 uint16_t t0;
 uint16_t t1;
 t0 = fmt_formatIntValue__anon_9515__7145(a0, a1, a2);
 memcpy(&t1, &t0, sizeof(uint16_t));
 return t1;
}

static uint64_t const builtin_zig_backend__262 = UINT64_C(3);

static bool const start_simplified_logic__109 = false;

static uint8_t const builtin_output_mode__263 = UINT8_C(1);

static uint8_t const builtin_link_mode__264 = UINT8_C(0);

static uint64_t features_features__329 = ((uint64_t)UINT64_C(0));

static uintptr_t const bit_set_IntegerBitSet_2864_29_bit_length__360 = 64ul;

static struct Target_Os__639 const builtin_os__270 = {{ .linux = {{{4ul,19ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},{6ul,13ul,4ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}}},{2ul,28ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},UINT32_C(14)} },UINT8_C(9)};

static bool const builtin_link_libc__274 = true;

static bool const posix_use_libc__1590 = true;

static uint8_t const c_native_os__2032 = UINT8_C(9);

static uint8_t const builtin_mode__273 = UINT8_C(3);

static bool const debug_runtime_safety__159 = false;

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_EnvironmentValue_29_2carray_hash_map_StringContext_2ctrue_29__880 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cjson_hashmap_ArrayHashMap_28environment_root_Environme__3775 = {{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),0ul,0ul},NULL};

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray_hash_map_StringContext_2ctrue_29__934 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_linked_libraries_root_LibraryEntry_2carray__4165 = {{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),0ul,0ul},NULL};

static struct Target_DynamicLinker__988 const Target_DynamicLinker_none__4413 = {"\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252",UINT8_C(0)};

static struct array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_StringContext_2ctrue_29__901 const array_hash_map_ArrayHashMapUnmanaged_28_5b_5dconst_20u8_2cenvironment_root_EnvironmentValue_2carray_hash_map_Str__3986 = {{((uint8_t *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),0ul,0ul},NULL};

static bool const builtin_is_test__266 = false;

static uint8_t *heap_CAllocator_alloc__4537(void *const a0, uintptr_t const a1, uint8_t const a2, uintptr_t const a3) {
 uint64_t t0;
 uint8_t *t2;
 bool t1;
 (void)a0;
 (void)a3;
 t0 = a1;
 t1 = t0 > UINT64_C(0);
 debug_assert__180(t1);
 t2 = heap_CAllocator_alignedAlloc__4534(a1, a2);
 return t2;
}

static bool heap_CAllocator_resize__4538(void *const a0, nav__4538_40 const a1, uint8_t const a2, uintptr_t const a3, uintptr_t const a4) {
 uintptr_t t0;
 uint64_t t1;
 uint64_t t2;
 uint8_t *t4;
 bool t3;
 (void)a0;
 (void)a2;
 (void)a4;
 t0 = a1.len;
 t1 = a3;
 t2 = t0;
 t3 = t1 <= t2;
 if (t3) {
  return true;
 }
 goto zig_block_0;

 zig_block_0:;
 t4 = a1.ptr;
 t0 = heap_CAllocator_alignedAllocSize__4536(t4);
 t2 = a3;
 t1 = t0;
 t3 = t2 <= t1;
 if (t3) {
  return true;
 }
 goto zig_block_1;

 zig_block_1:;
 return false;
}

static uint8_t *heap_CAllocator_remap__4539(void *const a0, nav__4539_40 const a1, uint8_t const a2, uintptr_t const a3, uintptr_t const a4) {
 uint8_t *t0;
 uint8_t *t2;
 uint8_t *t3;
 bool t1;
 t1 = heap_CAllocator_resize__4538(a0, a1, a2, a3, a4);
 if (t1) {
  t2 = a1.ptr;
  t3 = (uint8_t *)t2;
  t0 = t3;
  goto zig_block_0;
 }
 t0 = NULL;
 goto zig_block_0;

 zig_block_0:;
 return t0;
}

static void heap_CAllocator_free__4540(void *const a0, nav__4540_40 const a1, uint8_t const a2, uintptr_t const a3) {
 uint8_t *t0;
 (void)a0;
 (void)a2;
 (void)a3;
 t0 = a1.ptr;
 heap_CAllocator_alignedFree__4535(t0);
 return;
}

static uint8_t *heap_CAllocator_alignedAlloc__4534(uintptr_t const a0, uint8_t const a1) {
 uint8_t const *t1;
 uintptr_t t3;
 uint64_t t4;
 void *t9;
 void *t5;
 uint8_t *t10;
 int t6;
 int32_t t7;
 uint8_t t2;
 uint8_t t0;
 bool t8;
 t0 = a1;
 t1 = (uint8_t const *)&t0;
 t2 = (*t1);
 t3 = mem_Alignment_toByteUnits__1112(t2);
 t3 = ((uintptr_t)8ul > t3) ? (uintptr_t)8ul : t3;
 t4 = t3;
 t3 = t4;
 t6 = posix_memalign(&t5, t3, a0);
 t7 = t6;
 t8 = t7 != INT32_C(0);
 if (t8) {
  return NULL;
 }
 goto zig_block_0;

 zig_block_0:;
 t9 = t5;
 t10 = (uint8_t *)t9;
 return t10;
}

static uintptr_t heap_CAllocator_alignedAllocSize__4536(uint8_t *const a0) {
 void const *t0;
 void const *t1;
 uintptr_t t2;
 t0 = (void const *)a0;
 t1 = t0;
 t2 = malloc_usable_size(t1);
 return t2;
}

static void heap_CAllocator_alignedFree__4535(uint8_t *const a0) {
 void *t0;
 void *t1;
 t0 = (void *)a0;
 t1 = t0;
 free(t1);
 return;
}

static uintptr_t mem_Alignment_toByteUnits__1112(uint8_t const a0) {
 uintptr_t t1;
 uint8_t t0;
 t0 = a0;
 t1 = zig_shlw_u64((uintptr_t)1ul, t0, UINT8_C(64));
 return t1;
}

static struct mem_Allocator__599 const heap_c_allocator__4502 = {((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)};

static struct mem_Allocator__599 const c_allocator__235 = {((void *)(uintptr_t)0xaaaaaaaaaaaaaaaaul),((struct mem_Allocator_VTable__602 const *)&heap_CAllocator_vtable__4529)};

static struct Target_Cpu_Feature_Set__1005 const Target_Cpu_Feature_Set_empty__4458 = {{0ul,0ul,0ul,0ul,0ul}};

static struct Target_Cpu__973 const builtin_cpu__269 = {((struct Target_Cpu_Model__1000 const *)&Target_x86_cpu_x86_64__4649),{{9570149208162320ul,4506899252772864ul,1224979923328827392ul,0ul,0ul}},UINT8_C(42)};

static uint8_t const builtin_abi__268 = UINT8_C(1);

static uint8_t const builtin_object_format__272 = UINT8_C(0);

static struct Target__637 const builtin_target__271 = {{((struct Target_Cpu_Model__1000 const *)&Target_x86_cpu_x86_64__4649),{{9570149208162320ul,4506899252772864ul,1224979923328827392ul,0ul,0ul}},UINT8_C(42)},{{ .linux = {{{4ul,19ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},{6ul,13ul,4ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}}},{2ul,28ul,0ul,{NULL,0xaaaaaaaaaaaaaaaaul},{NULL,0xaaaaaaaaaaaaaaaaul}},UINT32_C(14)} },UINT8_C(9)},UINT8_C(1),UINT8_C(0),{"/lib64/ld-linux-x86-64.so.2\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252",UINT8_C(27)}};

static struct builtin_CallingConvention__266 const builtin_CallingConvention_c__470 = {{ .x86_64_sysv = {{UINT64_C(0xaaaaaaaaaaaaaaaa),true}} },UINT8_C(4)};

static uint8_t const (*const shared_RUNNER_CTL_FIFO__801)[21] = &__anon_1704;

static uint8_t const (*const shared_RUNNER_ACK_FIFO__802)[21] = &__anon_1769;

static uint64_t const runner_fifo_PROTOCOL_VERSION__1125 = UINT64_C(2);

static uint8_t const mem_native_endian__836 = UINT8_C(1);

static uint8_t const posix_native_os__1588 = UINT8_C(9);

static uint8_t const os_linux_native_arch__2829 = UINT8_C(42);

static uint8_t const fs_path_native_os__5117 = UINT8_C(9);

static uint8_t const fs_native_os__1303 = UINT8_C(9);

static uint8_t const fs_Dir_native_os__4784 = UINT8_C(9);

static bool const fs_File_is_windows__1479 = false;

static uint8_t const c_native_abi__2030 = UINT8_C(1);

static bool const posix_unexpected_error_tracing__2019 = false;

static uint8_t logger_max_level__5093 = UINT8_C(0);

static bool const fs_Dir_have_flock__4785 = true;

static uint16_t const fmt_max_format_args__5471 = UINT16_C(32);

static bool const debug_default_enable_segfault_handler__205 = false;

static uint8_t const log_default_level__6042 = UINT8_C(0);

static struct std_Options__4313 const std_options__97 = {{0xaaaaaaaaaaaaaaaaul,true},{0xaaaaaaaaaaaaaaaaul,true},3ul,false,UINT8_C(0),false,true,false,false,false,UINT8_C(2)};

static bool const posix_lfs64_abi__2018 = true;

static uint8_t const (*const fmt_ANY__5474)[4] = &__anon_5003;

static uint8_t const unicode_native_endian__5750 = UINT8_C(1);

static uint32_t const unicode_replacement_character__5751 = UINT32_C(65533);

static bool const mem_backend_supports_vectors__864 = true;

static bool const builtin_valgrind_support__277 = false;

static nav__4518_38 const heap_page_size_min_default__4518 = {4096ul,false};

static uintptr_t const heap_page_size_min__4497 = 4096ul;

static uint8_t const c_MARKER_TYPE_SAMPLE_START__246 = UINT8_C(0);

static uint8_t const c_MARKER_TYPE_SAMPLE_END__247 = UINT8_C(1);

static uint8_t const c_MARKER_TYPE_BENCHMARK_START__248 = UINT8_C(2);

static uint8_t const c_MARKER_TYPE_BENCHMARK_END__249 = UINT8_C(3);

static uint32_t const array_hash_map_index_capacities__3749[33] = {UINT32_C(0),UINT32_C(1),UINT32_C(2),UINT32_C(4),UINT32_C(9),UINT32_C(19),UINT32_C(38),UINT32_C(76),UINT32_C(153),UINT32_C(307),UINT32_C(614),UINT32_C(1228),UINT32_C(2457),UINT32_C(4915),UINT32_C(9830),UINT32_C(19660),UINT32_C(39321),UINT32_C(78643),UINT32_C(157286),UINT32_C(314572),UINT32_C(629145),UINT32_C(1258291),UINT32_C(2516582),UINT32_C(5033164),UINT32_C(10066329),UINT32_C(20132659),UINT32_C(40265318),UINT32_C(80530636),UINT32_C(161061273),UINT32_C(322122547),UINT32_C(644245094),UINT32_C(1288490188),UINT32_MAX};

static uint8_t const array_hash_map_Index_28u8_29_empty_sentinel__6278 = UINT8_MAX;

static uint16_t const array_hash_map_Index_28u16_29_empty_sentinel__6283 = UINT16_MAX;

static uint32_t const array_hash_map_Index_28u32_29_empty_sentinel__6288 = UINT32_MAX;

static uint64_t const hash_wyhash_Wyhash_secret__6452[4] = {UINT64_C(11562461410679940143),UINT64_C(16646288086500911323),UINT64_C(10285213230658275043),UINT64_C(6384245875588680899)};

static bool const builtin_fuzz__279 = false;

static bool const mem_eqlBytes_allowed__854 = true;

static struct array_hash_map_Index_28u8_29__6735 const array_hash_map_Index_28u8_29_empty__6279 = {UINT8_MAX,UINT8_C(0xaa)};

static struct array_hash_map_Index_28u16_29__6739 const array_hash_map_Index_28u16_29_empty__6284 = {UINT16_MAX,UINT16_C(0xaaaa)};

static struct array_hash_map_Index_28u32_29__6743 const array_hash_map_Index_28u32_29_empty__6289 = {UINT32_MAX,UINT32_C(0xaaaaaaaa)};

static bool const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6993 = false;

static struct json_stringify_WriteStream__union_8283__8283 const json_stringify_WriteStream_28io_GenericWriter_28_2aarray_list_ArrayListAligned_28u8_2cnull_29_2cerror_7bOutOfMemory_7d__6994 = {{0xaaaaaaaaaaaaaaaaul},UINT8_C(2)};

static uint8_t const environment_linked_libraries_elf_view_native_endian__6964 = UINT8_C(1);

static struct mem_Allocator_VTable__602 const heap_CAllocator_vtable__4529 = {&heap_CAllocator_alloc__4537,&heap_CAllocator_resize__4538,&heap_CAllocator_remap__4539,&heap_CAllocator_free__4540};

static bool const heap_CAllocator_supports_malloc_size__4530 = true;

static bool const heap_CAllocator_supports_posix_memalign__4532 = true;

static struct Target_Cpu_Model__1000 const Target_x86_cpu_x86_64__4649 = {{(uint8_t const *)&__anon_9580,6ul},{(uint8_t const *)&__anon_9582,6ul},{{9570149208162320ul,4506899252772864ul,1224979648450920448ul,0ul,0ul}}};

#endif
#ifndef WRAPPER_H
#define WRAPPER_H

#include <stdint.h>

#if defined(_WIN32) || defined(__APPLE__)
// Windows and other platforms - provide no-op implementations
uint8_t running_on_valgrind() { return 0; }

void callgrind_dump_stats() {}

void callgrind_dump_stats_at(uint8_t const *pos_str) { (void)pos_str; }

void callgrind_zero_stats() {}

void callgrind_start_instrumentation() {}

void callgrind_stop_instrumentation() {}
#else
#include "callgrind.h"
#include "valgrind.h"

uint8_t running_on_valgrind() { return RUNNING_ON_VALGRIND > 0; }

void callgrind_dump_stats() { CALLGRIND_DUMP_STATS; }

void callgrind_dump_stats_at(uint8_t const *pos_str) {
  CALLGRIND_DUMP_STATS_AT(pos_str);
}

void callgrind_zero_stats() { CALLGRIND_ZERO_STATS; }

void callgrind_start_instrumentation() { CALLGRIND_START_INSTRUMENTATION; }

void callgrind_stop_instrumentation() { CALLGRIND_STOP_INSTRUMENTATION; }

#endif

#endif

